/* 
** isosurf.c : Main source file for Mirage project isosurf.
**
** This is an example to show how the Volume primitive can be used.
*/

#include "mirage.h"

/*********** Prototypes for isosurf.c ************/

int main(int argc, char **argv);
void init_globals(void);
object make_isosurf(object *active_camera);


/*********** Function definitions, isosurf.c *********/

// Initialise Mirage, initialise global variables and create an animation.
int main(int argc, char **argv)
{
  // Set up Mirage using the command line parameters.
  init_system(argc, argv, "isosurf.key");

  // Initialiase global variables for this project.
  init_globals();

  // Set the scene creator routine.
  set_worldmaker(make_isosurf);

  // Create an animation.
  create_animation();
}

// Initialise global variables needed for the animation.
void init_globals(void)
{
  // Set up any globals here (load in texturemaps, splines, etc.).
}

// This routine returns an object that is the scene.
object make_isosurf(object *active_camera)
{
  object cameraobj, the_scene, blob_obj, extrude_obj, revolve_obj;
  isosurface_val *volume_data;
  int x_index, y_index, z_index;
  double distsq, volume_value;
  vector noiseval;

  the_scene = new_nothing();

  // Add a camera (default is at the origin looking along X).
  cameraobj = new_camera(new PinholeOptics(20, 1),
			 new StandardRayTracer,
			 new Antialiaser(0),
			 //			 new SingleRaySampler(300, 495),
			 //			 new PixelSampler,
			 new ColorStorage(1200, 900), NULL);
  *active_camera = cameraobj;
  LINK(the_scene, PLUS, cameraobj);

  // Add point and ambient lights.
  LINK(the_scene, PLUS, shift(-5.0, +5.0, +5.0,
			      new_point_light(0.6, 0.6, 0.6)));
  LINK(the_scene, PLUS, new_ambient_light(0.4, 0.4, 0.4));


  // First we make a blob. This is a volume primitive whose field is determined
  // by a radial dropoff with noise added. This gives a pseudo-spherical
  // surface since the noise perturbs the otherwise spherical isosurface of
  // the field.

  // Allocate an array of field data for the noisy volume.
  volume_data = (isosurface_val *) calloc(CUBE(40), sizeof(isosurface_val));
  if (volume_data == NULL) {
    fatal_error("make_isosurf", "can't allocate volume data",
		"must be low memory.");
  }

  // Fill the field data.
  printf("Creating noisy volume data ...");
  fflush(stdout);
  for (x_index = 0; x_index < 40; x_index++) {
    for (y_index = 0; y_index < 40; y_index++) {
      for (z_index = 0; z_index < 40; z_index++) {
	// Work out how far we are from the center of the blob.
	distsq = SQR((x_index - 19.5) / 19.5) + SQR((y_index - 19.5) / 19.5) +
	  SQR((z_index - 19.5) / 19.5);

	// Get a noise value for this location.
	noiseval[X] = 24 * x_index / 40.0;
	noiseval[Y] = 24 * y_index / 40.0;
	noiseval[Z] = 24 * z_index / 40.0;

	// Combine the distance from the center and the noise value to
	// get 0.5 + 0.2 * (0.65 - dist) + 0.11 * noiseval.
	// This will produce a surface where this value is 0.5.

	// Store the value in the field.
	volume_data[x_index + 40 * y_index + 40 * 40 * z_index] =
	  (unsigned short) (MAX_SHORT * (0.5 + 0.2 * (0.65 - sqrt(distsq))
					 + 0.11 * noise_3(noiseval)));

      }
    }
  }
  printf(" done.\n");

  // Create an object from the blobby field data.
  blob_obj = new_volume(TRUE, FALSE, 40, 40, 40, volume_data, 32767);
  blob_obj->material = new Material(new DiffuseComponent(0.1, 0.1, 0.15),
				    new SpecTransComponent(1.0,
							   0.3, 0.3, 0.3,
							   0.6, 0.6, 0.6),
				    new PhongComponent(40.0,
						       1.0, 1.0, 1.0));
  LINK(the_scene, PLUS, shift(5.0, 1.0, -0.3, 
		shift(-0.5, -0.5, -0.5, blob_obj)));


  // Create an extrusion from the bitmap that we have. The extrusion is
  // along +Z so we have to rotate this to face -X.
  printf("Creating extrusion from file ...\n");
  extrude_obj = new_extrusion("extrusion.pgm", TRUE);
  printf("... created extrusion.\n");
  extrude_obj->material = new Material(new DiffuseComponent(1.0, 0.0, 0.0),
				       new PhongComponent(40.0,
							  1.0, 1.0, 1.0));
  LINK(the_scene, PLUS, shift(5, 0, 0.8, rotate(Z, X, -90, rotate(Y, X, 90, 
       stretch(1.5, 0.3, 0.2, shift(-0.5, -0.5, -0.5, extrude_obj))))));


  // Create a surface of revolution using the cosine function.
  printf("Creating surface of revolution ...");
  fflush(stdout);
  revolve_obj = new_revolution(160, 160, 160, -PI/2, +PI/2, cos, TRUE);
  printf(" done\n");
  revolve_obj->material = white;
  LINK(the_scene, PLUS, shift(5, -1, -0.3, 
       shift(-0.5, -0.5, -0.5, revolve_obj)));

  // Set the background texture to be sky.
  Vector::assign(background_color, 0.15, 0.15, 0.6);

  // Return the scene, with a cluster wrapped around it.
  return new_stupid_cluster(bind(the_scene));
}

