/* logo.c : An example picture for the system.
** 
** This file holds an example picture that shows off the rendering
** system. Also used to test successful system installation.
**
** Chris Butcher, 15 January 1996
*/

#include "mirage.h"


// Material properties used for the Mirage logo.
Material
*logo_sand_material    = new Material(new DiffuseComponent(1, 1, 0.7)),
*logo_leaf_material    = new Material(new DiffuseComponent(0, 0.48, 0),
				      new PhongComponent(40, 1, 1, 1)),
*logo_water_material   = new Material(new DiffuseComponent(0.25, 0.25, 0.70),
				      new PhongComponent(40, 1, 1, 1),
				      new SpecTransComponent(1.33,
							     0.35, 0.35, 0.4,
							     0.3, 0.3, 0.3)),
*logo_coconut_material = new Material(new DiffuseComponent(0.5, 0.3, 0.1),
				      new PhongComponent(40, 1, 1, 1)),
*logo_pyramid_material = new Material(new DiffuseComponent(0.95, 0.92, 0.46)),
*logo_trunk_material   = logo_coconut_material,
*logo_mirage_text_color= new Material(new DiffuseComponent(0.25, 0.25, 0.25),
				      new PhongComponent(40, 1, 1, 1),
				      new SpecTransComponent(1,
							     0, 0, 0,
							     0.75, 0.75, 0.75)
				      );

// Function prototypes
object mirage_logo(object *active_camera);
object logo_mirage_text(void);
object logo_make_oasis(void);
object logo_make_desert(void);
object logo_make_tree(void);
object logo_make_pyramids(void);
object logo_make_bar(double bar_width, double bar_height, double bar_thick,
		     double bar_bevel, double bar_bev_angle);
object logo_make_M(double l_width, double l_height, double l_thick,
		   double l_linew, double l_bevel, double l_bev_angle,
		   double m_slope);
object logo_make_I(double l_width, double l_height, double l_thick,
		   double l_linew, double l_bevel, double l_bev_angle);
object logo_make_R(double l_width, double l_height, double l_thick,
		   double l_linew, double l_bevel, double l_bev_angle,
		   double loop_amount, double r_slope);
object logo_make_A(double l_width, double l_height, double l_thick,
		   double l_linew, double l_bevel, double l_bev_angle,
		   double a_slope);
object logo_make_G(double l_width, double l_height, double l_thick,
		   double l_linew, double l_bevel, double l_bev_angle,
		   double g_loop_amount, double g_end_amount);
object logo_make_E(double l_width, double l_height, double l_thick,
		   double l_linew, double l_bevel, double l_bev_angle,
		   double e_amount);


// Definitions for classes
class LogoBackground : public BackgroundTexture {
public:

  virtual void apply(vector from, vector dir, uint flags, color col,
		     RayTracer *rayengine)
  {
#define CLOUDS_OMEGA          0.5
#define CLOUDS_OCTAVES        8
#define CLOUDS_SCALEFACTOR    8.0
#define CLOUDS_OFFSET         1.0
    
    if (flags & RAY_THROUGH_GATEWAY) {
      // Do desert sky.
      double up_distance, across_distance, theta, horizon;
      color zenith_color = {0.15, 0.15, 0.60},
	    horizon_color = {0.81, 0.93, 1.00};
	    
      // Get the 'horizon' amount ... from 0 at the zenith to 1 at the horizon.
      up_distance = fabs(dir[Z]);
      across_distance = sqrt(SQR(dir[X]) + SQR(dir[Y]));
      theta = atan2(up_distance, across_distance);
      horizon = 1 - RAD2DEG(theta) / 90;
      
      // Interpolate the horizon amount from (0.9 - 1) -> (0 - 1).
      horizon = 10*horizon - 9;
      horizon = cubicInterpolate(0, 1, horizon);
      
      // Mix from sky colour at the zenith to a whiter blue at the horizon.
      Color::mix(zenith_color, horizon_color, horizon, col);
    } else {
      color sky_color   = {0.15, 0.15, 0.60},
            cloud_color = {1.0,  1.0,  1.0};
      vector samplevect, scaled_dir;
      
      double accum, omega_n, lambda, sky_amount, noise_val;
      int i;
    
      Vector::mult(CLOUDS_SCALEFACTOR / CHECK0(Vector::magnitude(dir),
					       "Sky: zero direction vector"),
		   dir, scaled_dir);
      
      scaled_dir[X] += CLOUDS_OFFSET;
      scaled_dir[Y] += CLOUDS_OFFSET;
      scaled_dir[Z] += CLOUDS_OFFSET;
      
      accum = 0;
      omega_n = 1;
      lambda = 1;
      for (i = 0; i < CLOUDS_OCTAVES; i++) {
	Vector::mult(lambda, scaled_dir, samplevect);
	noise_val = 2.0*noise_3(samplevect) - 1.0;
	accum += omega_n * noise_val;
	lambda *= 2;
	omega_n *= CLOUDS_OMEGA;
      }
      
      // Interpolate using a cubic function between 0 = 0 and 1 = 1.
      sky_amount = cubicInterpolate(0.0, 2.0, accum);
      
      Color::mix(sky_color, cloud_color, sky_amount, col);
    }
  }

  virtual void print_tex(void) {
    printf("LogoBackground background texture\n");
  }
};

class PyramidTexture : public Texture {
public:
  double block_size;
  color crack_col;

  PyramidTexture(double b_size, double c_r, double c_g, double c_b, int prec) {
    Color::assign(crack_col, c_r, c_g, c_b);
    block_size = b_size;
    precedence = prec;
  }

  virtual void apply(vector ray_from, vector ray_dir, uint ray_flags,
		     Material *properties, vector hit_point,
		     vector hit_normal, RayTracer *engine,
		     transform tex_transform);

  // Print the texture's parameters for debugging
  virtual void print_tex(void);
  
  virtual ~PyramidTexture(void) {
  }
};

// Function definitions begin here.

int main(int argc, char **argv)
{
  init_system(argc, argv, "logo.key");
  set_worldmaker(mirage_logo);
  create_animation();
}


// ******************************
// Code to build the Mirage logo.
// ******************************

// Return an object that is the logo for the system.
object mirage_logo(object *active_camera)
{
  object the_camera, the_sun, desert_sun, desert_profiling;
  object desert_ambient, the_ambient, the_text;
  object front_scene, back_scene;

  // Point the camera from +5 Y to the origin.
  the_camera = new_camera(new PinholeOptics(20, 1.0),
			  new StandardRayTracer,
			  new Antialiaser(2),
			  //			  new SingleRaySampler(170, 142),
			  new ColorStorage, NULL);

  // Light sources for in front of the gateway.
  the_sun = shift(2.0, 5.0, 1.5, new_point_light(0.8, 0.8, 0.8));
  the_ambient = new_ambient_light(0.2, 0.2, 0.2);

  // Make the scene IN FRONT OF the gateway.
  front_scene = new_nothing();

  // Camera and lighting setup.
  LINK(front_scene, PLUS, shift(0.0, 5.0, 0.0, rotate(X, Y, -90.0,
						      the_camera)));
  LINK(front_scene, PLUS, the_sun);
  LINK(front_scene, PLUS, the_ambient);

  // Make the 'MIRAGE' text.
  the_text = shift(0.0, 2.5, 0.05, scale(0.251, rotate(Y, 180, rotate(Z, Y, 90,
                      logo_mirage_text()))));
  the_text->flags |= FLAGS_ONEWAY_GATEWAY;
  the_text->material = logo_mirage_text_color;
  LINK(front_scene, PLUS, the_text);

  // Place a cluster around the text.
  front_scene = new_octree_cluster(bind(front_scene),
				   -1.2, +2.25, -0.35,
				   +1.2, +2.75, +0.35, 5, 3, TRUE);
  front_scene->flags |= FLAGS_NOT_THROUGH_GATEWAY;

  // Make the scene BEHIND the gateway.
  back_scene = new_nothing();

  // Make the desert lighting.
  desert_sun = new_directional_light( 0.65,  0.65,  0.65,
				      0.3,  -0.3,   1.0);

  // Another directional light that doesn't have shadows, to provide
  // depth cueing in the shadows of the main light.
  desert_profiling = new_directional_light( 0.2, 0.2, 0.2,
                                           -0.3, 0.3, 1.0);
  desert_profiling->flags |= FLAGS_NOSHADOW;
  desert_ambient = new_ambient_light(0.15, 0.15, 0.15);
  LINK(back_scene, PLUS, desert_sun);
  LINK(back_scene, PLUS, desert_ambient);
  LINK(back_scene, PLUS, desert_profiling);

  // Make the oasis (in its own subdivision scheme).
  LINK(back_scene, PLUS, logo_make_oasis());

  // Make the ground (extends to infinity, hf)
  LINK(back_scene, PLUS, logo_make_desert());

  // Make the pyramids in the distance.
  LINK(back_scene, PLUS, logo_make_pyramids());

  // Bind the back scene, and shift it to the correct place relative
  // to the front scene.
  back_scene = shift(-1.3, -9, -1, bind(back_scene));
  back_scene->flags |= FLAGS_ONLY_THROUGH_GATEWAY;

  // Link the front and back scenes together.
  LINK(front_scene, PLUS, back_scene);

  *active_camera = the_camera;

  // Set the sky background texture.
  background_texture = new LogoBackground;

  return new_stupid_cluster(bind(front_scene));
}

// Parameters for the oasis.
#define OASIS_RADIUS          5.5
#define OASIS_CLUSTER_SIZE    OASIS_RADIUS * 1.2
#define OASIS_DEPTH           1.1
#define WATER_DEPTH           0.05
#define TREE_SCALE            0.018
#define NICE_TREE_ON          1

object logo_make_desert(void)
{
  object the_desert;
#define SAND_NOISE_SCALE            20
#define SAND_NOISE_AMOUNT           0.2

  // The desert is flat. It has a cutout for the oasis to go in.
  the_desert = rotate(Z, Y, 180.0001, new_plane());
  LINK(the_desert, MINUS, scale(OASIS_RADIUS * 0.999, new_cube()));
  the_desert = bind(the_desert);

  // Give the desert a 'sand' texture.
  the_desert->material = logo_sand_material;
  the_desert->texture = new Noisetexture(SAND_NOISE_SCALE,
					 SAND_NOISE_AMOUNT, 0);

  return the_desert;
}

object logo_make_pyramids(void)
{
#define PYRAMID_1_OFFSET       35
#define PYRAMID_1_DISTANCE     120.0
#define PYRAMID_1_HEIGHT       13.0
#define PYRAMID_2_OFFSET       42
#define PYRAMID_2_DISTANCE     107.0
#define PYRAMID_2_HEIGHT       6.0

  object the_pyramid, big_pyramid, small_pyramid, pyramids_obj;

  // Make a pyramid.
  the_pyramid = rotate(Z, X, 128, new_plane());
  LINK(the_pyramid, MINUS, rotate(Z, X,  52, new_plane()));
  LINK(the_pyramid, MINUS, rotate(Z, Y, +52, new_plane()));
  LINK(the_pyramid, MINUS, rotate(Z, Y, -52, new_plane()));

  big_pyramid = shift(0, 0, PYRAMID_1_HEIGHT, rotate(X, Y, 60,
						     bind(the_pyramid)));
  big_pyramid = bind(big_pyramid);
  big_pyramid->material = logo_pyramid_material;
  big_pyramid->texture = new PyramidTexture(1.5, 0.90, 0.63, 0.37, 0);

  small_pyramid = shift(0, 0, PYRAMID_2_HEIGHT, rotate(X, Y, 60,
						     bind(the_pyramid)));
  small_pyramid = bind(small_pyramid);
  small_pyramid->material = logo_pyramid_material;
  small_pyramid->texture = new PyramidTexture(1.1, 0.90, 0.63, 0.37, 0);

  // Place the pyramids.
  pyramids_obj = new_nothing();
  LINK(pyramids_obj, PLUS, shift(PYRAMID_1_OFFSET, -PYRAMID_1_DISTANCE, 0,
				 bind(big_pyramid)));
  LINK(pyramids_obj, PLUS, shift(PYRAMID_2_OFFSET, -PYRAMID_2_DISTANCE, 0,
				 bind(small_pyramid)));

  return bind(pyramids_obj);
}

object logo_make_oasis(void)
{
  object the_oasis, the_hf, the_water, the_tree;
  double octree_extent = OASIS_CLUSTER_SIZE;

  the_oasis = new_nothing();

  // Make the sandy basin of the oasis.
  the_hf = new_heightfield(TRUE, TRUE, FALSE, "maps/oasis_hf.pgm");
  the_hf = shift( -OASIS_RADIUS,  -OASIS_RADIUS,  -OASIS_DEPTH,
           stretch(OASIS_RADIUS*2, OASIS_RADIUS*2, OASIS_DEPTH*2,
		   the_hf));
  the_hf = bind(the_hf);

  the_hf->material = logo_sand_material;
  the_hf->texture = new Noisetexture(SAND_NOISE_SCALE,
				     SAND_NOISE_AMOUNT, 0);
  LINK(the_oasis, PLUS, the_hf);

  // Make the water in the oasis.
  the_water = shift(0, 0, -WATER_DEPTH, rotate(Z, X, 180, new_plane()));
  the_water->material = logo_water_material;
  LINK(the_oasis, PLUS, the_water);

  // Make a tree.
  the_tree = scale(TREE_SCALE, logo_make_tree());

  // Place the trees.
  LINK(the_oasis, PLUS, shift(-0.12 * OASIS_RADIUS, -0.62 * OASIS_RADIUS, 0,
  			      rotate(X, Y,  45.0, bind(the_tree))));
  LINK(the_oasis, PLUS, shift(-0.32 * OASIS_RADIUS, +0.47 * OASIS_RADIUS, 0,
  			      rotate(X, Y,  30.0, bind(the_tree))));
  LINK(the_oasis, PLUS, shift(-0.66 * OASIS_RADIUS, -0.34 * OASIS_RADIUS, 0,
  			      rotate(X, Y, -15.0, bind(the_tree))));
  LINK(the_oasis, PLUS, shift(+0.32 * OASIS_RADIUS, -0.64 * OASIS_RADIUS, 0,
  			      rotate(X, Y, 135.0, bind(the_tree))));
  LINK(the_oasis, PLUS, shift(+0.88 * OASIS_RADIUS, +0.00 * OASIS_RADIUS, 0,
			      rotate(X, Y, 180.0, bind(the_tree))));
  LINK(the_oasis, PLUS, shift(+0.42 * OASIS_RADIUS, +0.68 * OASIS_RADIUS, 0,
  			      rotate(X, Y, 120.0, bind(the_tree))));

  return new_octree_cluster(bind(the_oasis),
			    -octree_extent, -octree_extent, -octree_extent,
			    +octree_extent, +octree_extent, +octree_extent,
			    7, 4, TRUE);
}

object logo_mirage_text(void)
{
  double letter_height = 1.7;
  double letter_width = 1.15;
  double letter_thickness = 1.0;
  double letter_linewidth = 0.3;
  double letter_spacing = 0.15;
  double small_spacing = 0.01;
  double letter_bevel = 0.05;
  double letter_bevel_angle = 45.0;
  double m_slope = 25.0;
  double a_slope = 15.0;
  double e_amount = 0.75;
  double r_loop_amount = 0.65;
  double r_slope = 30.0;
  double g_loop_amount = 0.35;
  double g_end_amount = -0.15;

  // The text is 5 letters long, with an 'I' that is only the line width wide,
  // with 3 large spacing gaps and 2 small spacing gaps.
  double text_length = 5 * letter_width + 1 * letter_linewidth
    + 3 * letter_spacing + 2 * small_spacing;

  // Initial place for the left side of the first character.
  double text_offset = - text_length / 2;

  object text;

  text = new_nothing();

  // Make the bar that goes below the text.
  LINK(text, PLUS, shift(0, -letter_height/2 - letter_linewidth/2
			 -letter_spacing, 0,
			 logo_make_bar(text_length, letter_linewidth,
				       letter_thickness, letter_bevel,
				       letter_bevel_angle)));

  // Make the letters spaced out along X, facing towards +Z.
  LINK(text, PLUS, shift(text_offset + letter_width/2, 0, 0,
                logo_make_M(letter_width, letter_height, letter_thickness,
			    letter_linewidth, letter_bevel, letter_bevel_angle,
			    m_slope)));

  text_offset += letter_width + letter_spacing;

  LINK(text, PLUS, shift(text_offset + letter_linewidth/2, 0, 0,
                logo_make_I(letter_width, letter_height, letter_thickness,
			    letter_linewidth, letter_bevel,
			    letter_bevel_angle)));

  // The 'I' is only letter_linewidth wide, not letter_width.
  text_offset += letter_linewidth + letter_spacing;

  LINK(text, PLUS, shift(text_offset + letter_width/2, 0, 0,
                logo_make_R(letter_width, letter_height, letter_thickness,
			    letter_linewidth, letter_bevel, letter_bevel_angle,
			    r_loop_amount, r_slope)));

  text_offset += letter_width + small_spacing;

  LINK(text, PLUS, shift(text_offset + letter_width/2, 0, 0,
	        logo_make_A(letter_width, letter_height, letter_thickness,
			    letter_linewidth, letter_bevel, letter_bevel_angle,
			    a_slope)));

  text_offset += letter_width + small_spacing;

  LINK(text, PLUS, shift(text_offset + letter_width/2, 0, 0,
                logo_make_G(letter_width, letter_height, letter_thickness,
			    letter_linewidth, letter_bevel, letter_bevel_angle,
			    g_loop_amount, g_end_amount)));

  text_offset += letter_width + letter_spacing;

  LINK(text, PLUS, shift(text_offset + letter_width/2, 0, 0,
                logo_make_E(letter_width, letter_height, letter_thickness,
			    letter_linewidth, letter_bevel, letter_bevel_angle,
			    e_amount)));

  return bind(text);
}

object logo_make_bar(double bar_width, double bar_height, double bar_thick,
		     double bar_bevel, double bar_bev_angle)
{
  object the_bar, beveled_cutter;

  // Make the bar a slab of the desired thickness.
  the_bar = shift(0, 0, -bar_thick/2, new_plane());
  LINK(the_bar, MINUS, shift(0, 0, bar_thick/2, new_plane()));

  // Make a beveled cutter that is the plane in -Y with beveled bits on it.
  beveled_cutter = rotate(Z, Y, -90, new_plane());
  LINK(beveled_cutter, PLUS, shift(0, bar_bevel,  bar_thick/2,
               rotate(Z, Y, -bar_bev_angle, new_plane())));

  // Cut four sides off the bar with the beveled cutter.
  LINK(the_bar, MINUS, shift(0, -bar_height/2, 0, bind(beveled_cutter)));
  LINK(the_bar, MINUS, shift(0, +bar_height/2, 0,
			     rotate(Z, 180, bind(beveled_cutter))));
  LINK(the_bar, MINUS, shift(+bar_width/2, 0, 0,
			     rotate(Y, X, -90, bind(beveled_cutter))));
  LINK(the_bar, MINUS, shift(-bar_width/2, 0, 0,
			     rotate(Y, X, +90, bind(beveled_cutter))));

  return bind(the_bar);
}


object logo_make_M(double l_width, double l_height, double l_thick,
		   double l_linew, double l_bevel, double l_bev_angle,
		   double m_slope)
{
  object beveled_line, beveled_cutter, theM, l_diagonal, r_diagonal;
  double l_horiz_width = l_linew / cos(DEG2RAD(m_slope));

  // Make the line a slab of the desired thickness.
  beveled_line = shift(0, 0, -l_thick/2, new_plane());
  LINK(beveled_line, MINUS, shift(0, 0, l_thick/2, new_plane()));

  // Make a beveled cutter that is the plane in -Y with beveled bits on it.
  beveled_cutter = rotate(Z, Y, -90, new_plane());
  LINK(beveled_cutter, PLUS, shift(0, l_bevel,  l_thick/2,
               rotate(Z, Y, -l_bev_angle, new_plane())));

  // Chop the sides of the beveled line off at +/- l_linew/2 in X.
  LINK(beveled_line, MINUS, shift(+l_linew/2, 0, 0,
                       rotate(X, Y, 90, bind(beveled_cutter))));
  LINK(beveled_line, MINUS, shift(-l_linew/2, 0, 0,
                       rotate(Y, X, 90, bind(beveled_cutter))));

  // The M has two beveled lines forming verticals.
  theM = shift(-(l_width/2 - l_linew/2), 0, 0, bind(beveled_line));
  LINK(theM, PLUS, shift(l_width/2 - l_linew/2, 0, 0, bind(beveled_line)));

  // The diagonals of the M slope inwards at m_slope degrees. This means that
  // their width horizontally is l_horiz_width.
  // Form the diagonal lines; chop them off with planes so that they don't
  // extend beyond the right-hand edge of the top surface of the outside bars
  // or beyond the center of the M.
  l_diagonal = shift(-(l_width/2 - l_linew + l_horiz_width/2), +l_height/2, 0,
		rotate(Y, X, -m_slope, bind(beveled_line)));
  LINK(l_diagonal, MINUS, shift(-(l_width/2 - l_linew + l_bevel), 0, 0,
                rotate(Z, X, -90, new_plane())));
  LINK(l_diagonal, MINUS, rotate(Z, X, 90, new_plane()));

  r_diagonal = shift(+(l_width/2 - l_linew + l_horiz_width/2), +l_height/2, 0,
		rotate(Y, X, m_slope, bind(beveled_line)));
  LINK(r_diagonal, MINUS, shift((l_width/2 - l_linew + l_bevel), 0, 0,
                rotate(Z, X, 90, new_plane())));
  LINK(r_diagonal, MINUS, rotate(Z, X, -90, new_plane()));

  // Put the diagonal lines together.
  LINK(l_diagonal, PLUS, bind(r_diagonal));
  
  // Add the diagonal lines to the M.
  LINK(theM, PLUS, bind(l_diagonal));

  // Bevel the top and bottom of the M.
  LINK(theM, MINUS, shift(0, +l_height/2, 0,
                   rotate(Z, 180, bind(beveled_cutter))));
  LINK(theM, MINUS, shift(0, -l_height/2, 0, bind(beveled_cutter)));

  return bind(theM);
}

object logo_make_I(double l_width, double l_height, double l_thick,
		    double l_linew, double l_bevel, double l_bev_angle)
{
  object theI, beveled_cutter;

  // Make the I a slab of the desired thickness.
  theI = shift(0, 0, -l_thick/2, new_plane());
  LINK(theI, MINUS, shift(0, 0, l_thick/2, new_plane()));

  // Make a beveled cutter that is the plane in -Y with beveled bits on it.
  beveled_cutter = rotate(Z, Y, -90, new_plane());
  LINK(beveled_cutter, PLUS, shift(0, l_bevel,  l_thick/2,
               rotate(Z, Y, -l_bev_angle, new_plane())));

  // Cut four sides off the I with the beveled cutter.
  LINK(theI, MINUS, shift(0, -l_height/2, 0, bind(beveled_cutter)));
  LINK(theI, MINUS, shift(0, +l_height/2, 0,
                    rotate(Z, 180, bind(beveled_cutter))));
  LINK(theI, MINUS, shift(+l_linew/2, 0, 0,
                    rotate(Y, X, -90, bind(beveled_cutter))));
  LINK(theI, MINUS, shift(-l_linew/2, 0, 0,
                    rotate(Y, X, +90, bind(beveled_cutter))));

  return bind(theI);
}

object logo_make_R(double l_width, double l_height, double l_thick,
		   double l_linew, double l_bevel, double l_bev_angle,
		   double loop_amount, double r_slope)
{
  object theR, squarebit, cutter, adder, topbevel;
  object bevel_cone, diagonal;
  double looprad = loop_amount * l_height / 2;
  double loop_x = l_width/2 - looprad;
  double loop_y = l_height/2 - looprad;
  double loop_bottom = l_height/2 - 2*looprad;
  double cone_height, cone_scale, diagonal_horiz_width;

  // The top bevel is a plane at the desired angle, crossing the
  // Z = thickness/2 axis at Y = l_bevel. The bottom bevel crosses at Z = -t/2
  // in the opposite orientation. They both don't exist in negative Y.
  topbevel = shift(0, l_bevel, l_thick/2,
              rotate(Z, Y, -l_bev_angle, new_plane()));
  LINK(topbevel, MINUS, shift(0, l_bevel + 0.001, 0,
                      rotate(Z, Y, 90, new_plane())));
  LINK(topbevel, MINUS, shift(0, -0.001, 0,
                      rotate(Z, Y, -90, new_plane())));

  // Construct the circular loop piece, at the origin initially.
  theR = scale(looprad, new_cyl());
  LINK(theR, MINUS, scale(looprad - l_linew, new_cyl()));

  // Find the cone to bevel the outside of the circle.
  // This must have radius looprad - l_bevel at Z = 0, and
  // angle l_bev_angle.
  cone_height = (looprad - l_bevel) * tan(DEG2RAD(l_bev_angle));
  cone_scale = (looprad - l_bevel) / cone_height;
  bevel_cone = stretch(cone_scale, cone_scale, 1,
                   shift(0, 0, cone_height, new_cone()));

  // Bevel the top of the outside of the circle.
  cutter = new_full();
  LINK(cutter, MINUS, shift(0, 0, l_thick/2, bind(bevel_cone)));
  LINK(theR, MINUS, bind(cutter));

  // Find the cone to bevel the inside of the circle.
  // This must have radius looprad - l_linew + l_bevel at Z = 0, and
  // angle l_bev_angle.
  cone_height = (looprad - l_linew + l_bevel) * tan(DEG2RAD(l_bev_angle));
  cone_scale = (looprad - l_linew + l_bevel) / cone_height;
  bevel_cone = stretch(cone_scale, cone_scale, 1,
                   shift(0, 0, cone_height, new_cone()));

  // Bevel the top of the inside of the circle.
  cutter = bind(bevel_cone);
  LINK(cutter, MINUS, shift(0, 0, cone_height, new_plane()));
  LINK(theR, MINUS, shift(0, 0,  l_thick/2, 
                            rotate(Z, X, 180.0, bind(cutter))));

  // Cut off the circular loop piece where it meets the square bit.
  LINK(theR, MINUS, shift(-0.001, 0, 0, rotate(Z, X, -90, new_plane())));

  // Shift the circular loop to the correct place.
  theR = shift(loop_x, loop_y, 0, bind(theR));
  
  // Make a bit that looks like 
  //        _______
  //       | ______|
  //       | |_____
  //       |  _____|
  //       | |     
  //       | |     
  //       | |
  //   with beveled edges.

  // Make the right edge.
  squarebit = shift(l_width/1.999 - looprad, 0, 0,
                     rotate(Z, X, -90, new_plane()));

  // Make the bottom edge.
  LINK(squarebit, MINUS, shift(0, loop_bottom, 0,
                         rotate(Z, Y, -90, new_plane())));

  // Bevel the bottom edge.
  LINK(squarebit, MINUS, shift(0, loop_bottom, 0,
			 bind(topbevel)));

  // --- Make a beveled cutter that cuts the inside of the square
  // --- piece of the R's loop.

  // Make the cutter's Y extents.
  cutter = shift(0, l_height/2 - l_linew, 0, rotate(Z, Y, -90, new_plane()));
  LINK(cutter, MINUS, shift(0, loop_bottom + l_linew, 0,
                           rotate(Z, Y, -90, new_plane())));

  // Bevel the top of the cutter.
  LINK(cutter, PLUS, shift(0, l_height/2 - l_linew, 0, bind(topbevel)));

  // Bevel the bottom of the cutter.
  LINK(cutter, PLUS, shift(0, loop_bottom + l_linew, 0,
			   rotate(Z, 180, bind(topbevel))));

  // --- Cut out the inside of the square piece of the R's loop using this
  // beveled cutter.
  LINK(squarebit, MINUS, bind(cutter));
			
  // Add the left-hand column to the square bit. Bevel it before adding.
  adder = shift(-l_width/2 + l_linew, 0, 0, rotate(Z, X, -90, new_plane()));
  LINK(adder, MINUS, shift(-l_width/2 + l_linew, 0, 0,
                       rotate(X, Y, 90, bind(topbevel))));
  LINK(squarebit, PLUS, bind(adder));

  // Trim and bevel the top of the square bit.
  LINK(squarebit, MINUS, shift(0, l_height/2, 0,
                     rotate(Z, Y, 90, new_plane())));
  LINK(squarebit, MINUS, shift(0, l_height/2, 0,
                     rotate(Z, 180, bind(topbevel))));

  // Trim and bevel the left-hand side.
  LINK(squarebit, MINUS, shift(-l_width/2, 0, 0,
		     rotate(Z, X, -90, new_plane())));
  LINK(squarebit, MINUS, shift(-l_width/2, 0, 0,
		     rotate(Y, X, 90, bind(topbevel))));

  // Construct the diagonal line (initially along Y, at X = 0)
  diagonal = shift(-l_linew/2, 0, 0, rotate(Z, X, 90, new_plane()));
  LINK(diagonal, MINUS, shift(+l_linew/2, 0, 0,
                         rotate(Z, X, 90, new_plane())));

  // Bevel the left of the diagonal line.
  LINK(diagonal, MINUS, shift(-l_linew/2, 0, 0,
                         rotate(Y, X, 90, bind(topbevel))));

  // Bevel the right of the diagonal line.
  LINK(diagonal, MINUS, shift(+l_linew/2, 0, 0,
                         rotate(Y, X, -90, bind(topbevel))));

  // Position the diagonal line in the correct place.
  diagonal_horiz_width = l_linew / cos(DEG2RAD(r_slope));
  diagonal = shift(l_width/2 - diagonal_horiz_width/2, -l_height/2, 0,
		   rotate(Y, X, -r_slope, bind(diagonal)));
  
  // Cut off the diagonal line before it goes through the loop.
  LINK(diagonal, MINUS, shift(0, loop_bottom + l_linew - l_bevel*1.5, 0,
			    rotate(Z, Y, 90, new_plane())));

  // Add the diagonal line to the square piece.
  LINK(squarebit, PLUS, bind(diagonal));

  // Trim the bottom of the square piece.
  LINK(squarebit, MINUS, shift(0, -l_height/2, 0,
	             rotate(Z, Y, -90, new_plane())));
  LINK(squarebit, MINUS, shift(0, -l_height/2, 0, bind(topbevel)));

  // Link the square piece to the loop piece.
  LINK(theR, PLUS, bind(squarebit));

  // Chop off the R in Z.
  LINK(theR, MINUS, shift(0, 0, +l_thick/2, new_plane()));
  LINK(theR, MINUS, shift(0, 0, -l_thick/2, 
                       rotate(Z, X, 180, new_plane())));

  return bind(theR);
}

object logo_make_A(double l_width, double l_height, double l_thick,
		   double l_linew, double l_bevel, double l_bev_angle,
		   double a_slope)
{
  object theA, tempObject, topbevel;
  double a_bar_height = l_height / 3.0;
  double l_horiz_width = l_linew / cos(DEG2RAD(a_slope));

  // The top bevel is a plane at the desired angle, crossing the
  // Z = thickness/2 axis at Y = l_bevel. The bottom bevel crosses at Z = -t/2
  // in the opposite orientation.
  topbevel = shift(0, l_bevel, l_thick/2,
              rotate(Z, Y, -l_bev_angle, new_plane()));

  // Form the bar across the A.
  theA = shift(0.0, -l_height/2 + a_bar_height - l_linew/2, 0.0,
	       rotate(Z, Y, 90.0, new_plane()));
  LINK(theA, MINUS, shift(0.0, -l_height/2 + a_bar_height + l_linew/2, 0.0,
		   rotate(Z, Y, 90.0, new_plane())));

  // Bevel the bottom of the bar.
  LINK(theA, MINUS, shift(0, -l_height/2 + a_bar_height - l_linew/2, 0,
			  bind(topbevel)));

  // Bevel the top of the bar.
  LINK(theA, MINUS, shift(0, -l_height/2 + a_bar_height + l_linew/2, 0,
                          rotate(Z, 180, bind(topbevel))));

  // Form a beveled object that is the plane in +Y, beveled along X.
  tempObject = rotate(Z, Y, 90, new_plane());
  LINK(tempObject, MINUS, bind(topbevel));

  // --- Construct the right side of the A.
  LINK(theA, PLUS, shift(  l_width/2 - l_horiz_width , -l_height/2, 0,
                   rotate(Y, X, 90 - a_slope, bind(tempObject))));

  // --- Construct the left side of the A.
  LINK(theA, PLUS, shift(-(l_width/2 - l_horiz_width), -l_height/2, 0,
                   rotate(Y, X, a_slope - 90, bind(tempObject))));

  // Form a cutter object that is a plane in -Y, with bevels added to it.
  tempObject = rotate(Z, Y, -90, new_plane());
  LINK(tempObject, PLUS, bind(topbevel));

  // Subtract the right side of the A.
  LINK(theA, MINUS, shift( l_width/2, -l_height/2, 0,
                   rotate(X, Y, 90 + a_slope, bind(tempObject))));

  // Subtract the left side of the A.
  LINK(theA, MINUS, shift(-l_width/2, -l_height/2, 0,
                   rotate(Y, X, 90 + a_slope, bind(tempObject))));

  // Chop off the A in Y.
  LINK(theA, MINUS, shift(0, -l_height/2, 0, bind(tempObject)));
  LINK(theA, MINUS, shift(0, +l_height/2, 0,
                          rotate(Z, 180.0, bind(tempObject))));

  // Chop off the A in Z.
  LINK(theA, MINUS, shift(0, 0,  l_thick/2, new_plane()));
  LINK(theA, MINUS, shift(0, 0, -l_thick/2, rotate(Z, X, 180, new_plane())));

  return bind(theA);
}

object logo_make_G(double l_width, double l_height, double l_thick,
		   double l_linew, double l_bevel, double l_bev_angle,
		   double loop_amount, double end_amount)
{
  object theG, straight, curve, topbevel, cutter, bevel_cone;
  object cut_curve, adder;
  double looprad = loop_amount * l_height;
  double cone_height, cone_scale, end_pos = end_amount * l_width/2;

  theG = new_nothing();

  // The top bevel is a plane at the desired angle, crossing the
  // Z = thickness/2 axis at Y = l_bevel. The bottom bevel crosses at Z = -t/2
  // in the opposite orientation. They both don't exist in negative Y.
  topbevel = shift(0, l_bevel, l_thick/2,
              rotate(Z, Y, -l_bev_angle, new_plane()));
  LINK(topbevel, MINUS, shift(0, l_bevel + 0.001, 0,
                      rotate(Z, Y, 90, new_plane())));
  LINK(topbevel, MINUS, shift(0, -0.001, 0,
                      rotate(Z, Y, -90, new_plane())));

  // --- Make the top line of the G.

  // Define the top (beveled) edge.
  straight = shift(0, l_height/2, 0, rotate(Z, Y, -90, new_plane()));
  LINK(straight, MINUS, shift(0, l_height/2, 0,
                           rotate(Z, 180, bind(topbevel))));

  // Define the bottom (beveled) edge.
  LINK(straight, MINUS, shift(0, l_height/2 - l_linew, 0,
                           rotate(Z, Y, -90, new_plane())));
  LINK(straight, MINUS, shift(0, l_height/2 - l_linew, 0, bind(topbevel)));

  // Define the right (beveled) edge.
  LINK(straight, MINUS, shift(l_width/2, 0, 0, rotate(Z, X, 90, new_plane())));
  LINK(straight, MINUS, shift(l_width/2, 0, 0,
			   rotate(X, Y, 90, bind(topbevel))));

  // Chop it off at the place where it meets the curve.
  LINK(straight, MINUS, shift(-l_width/2 + looprad*0.999, 0, 0,
                           rotate(Z, X, -90, new_plane())));

  // --- Link in the top line of the G.
  LINK(theG, PLUS, bind(straight));

  // --- Make the curve of the G.

  // Construct the circular loop piece, at the origin initially.
  curve = scale(looprad, new_cyl());
  LINK(curve, MINUS, scale(looprad - l_linew, new_cyl()));

  // Find the cone to bevel the outside of the circle.
  // This must have radius looprad - l_bevel at Z = 0, and
  // angle l_bev_angle.
  cone_height = (looprad - l_bevel) * tan(DEG2RAD(l_bev_angle));
  cone_scale = (looprad - l_bevel) / cone_height;
  bevel_cone = stretch(cone_scale, cone_scale, 1,
                   shift(0, 0, cone_height, new_cone()));

  // Bevel the top of the outside of the circle.
  cutter = new_full();
  LINK(cutter, MINUS, shift(0, 0, l_thick/2, bind(bevel_cone)));
  LINK(curve, MINUS, bind(cutter));

  // Find the cone to bevel the inside of the circle.
  // This must have radius looprad - l_linew + l_bevel at Z = 0, and
  // angle l_bev_angle.
  cone_height = (looprad - l_linew + l_bevel) * tan(DEG2RAD(l_bev_angle));
  cone_scale = (looprad - l_linew + l_bevel) / cone_height;
  bevel_cone = stretch(cone_scale, cone_scale, 1,
                   shift(0, 0, cone_height, new_cone()));

  // Bevel the top of the inside of the circle.
  cutter = bind(bevel_cone);
  LINK(cutter, MINUS, shift(0, 0, cone_height, new_plane()));
  LINK(curve, MINUS, shift(0, 0,  l_thick/2, 
                            rotate(Z, X, 180.0, bind(cutter))));

  // --- Add the top left curve.
  cut_curve = bind(curve);
  LINK(cut_curve, MINUS, rotate(Z, X,  90, new_plane()));
  LINK(cut_curve, MINUS, rotate(Z, Y, -90, new_plane()));
  LINK(theG, PLUS, shift(-l_width/2 + looprad,  l_height/2 - looprad, 0,
                     bind(cut_curve)));

  // --- Add the bottom left curve.
  cut_curve = bind(curve);
  LINK(cut_curve, MINUS, rotate(Z, X,  90, new_plane()));
  LINK(cut_curve, MINUS, rotate(Z, Y,  90, new_plane()));
  LINK(theG, PLUS, shift(-l_width/2 + looprad, -l_height/2 + looprad, 0,
                     bind(cut_curve)));

  // --- Make the left line of the G.

  // Define the left (beveled) edge.
  straight = shift(-l_width/2, 0, 0, rotate(Z, X, 90, new_plane()));
  LINK(straight, MINUS, shift(-l_width/2, 0, 0,
                           rotate(X, Y, -90, bind(topbevel))));

  // Define the right (beveled) edge.
  LINK(straight, MINUS, shift(-l_width/2 + l_linew, 0, 0,
                           rotate(Z, X, 90, new_plane())));
  LINK(straight, MINUS, shift(-l_width/2 + l_linew, 0, 0,
			   rotate(X, Y, 90, bind(topbevel))));

  // Chop it off at the places where it meets the curves.
  LINK(straight, MINUS, shift(0,  l_height/2 - looprad*0.999, 0,
			   rotate(Z, Y,  90, new_plane())));
  LINK(straight, MINUS, shift(0, -l_height/2 + looprad*0.999, 0,
			   rotate(Z, Y, -90, new_plane())));

  // --- Link in the left line of the G.
  LINK(theG, PLUS, bind(straight));

  // --- Make the bottom right square piece.
  straight = new_nothing();

  // Make the right-hand line with a beveled inner edge and infinite extent
  // to the right.
  adder = shift(l_width/2 - l_linew, 0, 0, rotate(Z, X, 90, new_plane()));
  LINK(adder, MINUS, shift(l_width/2 - l_linew, 0, 0,
                        rotate(Y, X, 90, bind(topbevel))));

  // Add this right-hand line to the square piece.
  LINK(straight, PLUS, bind(adder));

  // Make the bottom line with a beveled inner edge, chopped off where it
  // meets the bottom left curve, and infinite extent to the southeast.
  adder = shift(0, -l_height/2 + l_linew, 0, rotate(Z, Y, -90, new_plane()));
  LINK(adder, MINUS, shift(0, -l_height/2 + l_linew, 0,
                        rotate(Z, 180, bind(topbevel))));
  LINK(adder, MINUS, shift(-l_width/2 + looprad*0.999, 0, 0,
                        rotate(Z, X, -90, new_plane())));

  // Add this bottom line to the square piece.
  LINK(straight, PLUS, bind(adder));

  // Make the top line with beveled bottom and left edges and infinite extent
  // to the northeast.
  adder = shift(0, -l_linew/2, 0, rotate(Z, Y, 90, new_plane()));
  LINK(adder, MINUS, shift(0, -l_linew/2, 0, bind(topbevel)));

  LINK(adder, MINUS, shift(end_pos, 0, 0, rotate(Z, X, -90, new_plane())));
  LINK(adder, MINUS, shift(end_pos, 0, 0,
                        rotate(Y, X, 90, bind(topbevel))));

  // Add this top line to the square piece.
  LINK(straight, PLUS, bind(adder));

  // Trim and bevel the top edge of the square piece.
  LINK(straight, MINUS, shift(0, l_linew/2, 0, rotate(Z, Y, 90, new_plane())));
  LINK(straight, MINUS, shift(0, l_linew/2, 0,
                           rotate(Z, 180, bind(topbevel))));

  // Trim and bevel the right edge of the square piece.
  LINK(straight, MINUS, shift(l_width/2, 0, 0, rotate(Z, X, 90, new_plane())));
  LINK(straight, MINUS, shift(l_width/2, 0, 0,
                           rotate(X, Y, 90, bind(topbevel))));

  // Trim and bevel the bottom edge of the square piece.
  LINK(straight, MINUS, shift(0, -l_height/2, 0,
                           rotate(Z, Y, -90, new_plane())));
  LINK(straight, MINUS, shift(0, -l_height/2, 0, bind(topbevel)));

  // --- Add the bottom right square piece to the G.
  LINK(theG, PLUS, bind(straight));

  // --- Terminate the G in Z at the desired thickness.
  LINK(theG, MINUS, shift(0, 0,  l_thick/2, new_plane()));
  LINK(theG, MINUS, shift(0, 0, -l_thick/2, rotate(X, Z, 180, new_plane())));

  return bind(theG);
}

object logo_make_E(double l_width, double l_height, double l_thick,
		   double l_linew, double l_bevel, double l_bev_angle,
		   double e_amount)
{
  object theE, beveled_cutter, beveled_line, middle_line;

  // Make a beveled cutter that is the plane in -Y with beveled bits on it.
  beveled_cutter = rotate(Z, Y, -90, new_plane());
  LINK(beveled_cutter, PLUS, shift(0, l_bevel,  l_thick/2,
               rotate(Z, Y, -l_bev_angle, new_plane())));

  // Make the beveled line that is just a slab of the desired thickness.
  beveled_line = shift(0, 0, -l_thick/2, new_plane());
  LINK(beveled_line, MINUS, shift(0, 0, l_thick/2, new_plane()));

  // Chop the sides of the beveled line off at +/- l_linew/2 in X.
  LINK(beveled_line, MINUS, shift(+l_linew/2, 0, 0,
                       rotate(X, Y, 90, bind(beveled_cutter))));
  LINK(beveled_line, MINUS, shift(-l_linew/2, 0, 0,
                       rotate(Y, X, 90, bind(beveled_cutter))));

  // Make the top and bottom lines of the E.
  theE = shift(0, +(l_height/2 - l_linew/2), 0,
	       rotate(Y, X, 90, bind(beveled_line)));
  LINK(theE, PLUS, shift(0, -(l_height/2 - l_linew/2), 0,
			 rotate(Y, X, 90, bind(beveled_line))));

  // Make the middle (shorter) line.
  middle_line = rotate(Y, X, 90, bind(beveled_line));
  LINK(middle_line, MINUS, shift(-l_width/2 + l_width * e_amount, 0, 0,
                   rotate(X, Y, 90, bind(beveled_cutter))));

  // Add the middle line.
  LINK(theE, PLUS, bind(middle_line));

  // Now we chop off the lines at the point where the right-hand edge
  // of the left-hand line meets their top surface.
  LINK(theE, MINUS, shift(-(l_width/2) + l_linew - l_bevel, 0, 0,
			  rotate(Z, X, -90, new_plane())));

  // Add the left-hand line.
  LINK(theE, PLUS, shift(-(l_width/2 - l_linew/2), 0, 0, bind(beveled_line)));

  theE = bind(theE);

  // Cut four sides off the E with the beveled cutter.
  LINK(theE, MINUS, shift(0, -l_height/2.001, 0, bind(beveled_cutter)));
  LINK(theE, MINUS, shift(0, +l_height/2.001, 0,
                    rotate(Z, 180, bind(beveled_cutter))));
  LINK(theE, MINUS, shift(+l_width/2.001, 0, 0,
                    rotate(Y, X, -90, bind(beveled_cutter))));
  LINK(theE, MINUS, shift(-l_width/2.001, 0, 0,
                    rotate(Y, X, +90, bind(beveled_cutter))));

  return bind(theE);
}

object logo_make_tree(void)
{
#define TREE_HEIGHT             70.0
#define TREE_TRUNKPIECES          13
#define TREE_ROTATE              1.7
#define TREE_TWIST               2.0
#define TRUNKPIECE_HEIGHT        6.0
#define TRUNKPIECE_BOTTOM_RADIUS 4.0
#define TRUNKPIECE_TOP_RADIUS    2.5
#define TRUNKPIECE_TORUS_SMOOTH  1.0
#define LEAF_LENGTH             55.0
#define LEAF_WIDTH              35.0
#define LEAF_THICKNESS           1.0
#define LEAF_BENDDOWN           10.0
#define LEAF_OUTSLOPE            7.5
  
  object theTree, theTrunkpiece, theLeaf, theCoconut, leaf_extrusion;
  int i;
  object theLeafpiece, cutter, aTorus, reverseLeafpiece;
  double tPieceSlope;
  double torus_center_x, torus_center_z;
  double torusrad;

  // Create a leaf-shaped extrusion from a heightfield minus a plane.
  leaf_extrusion = new_heightfield(TRUE, FALSE, FALSE, "maps/leaf.pgm");
  LINK(leaf_extrusion, MINUS, shift(0, 0, 0.1,
                       rotate(Z, X, 180, new_plane())));

  /* Load in the leaf shape from the file and scale it correctly */
  theLeafpiece = stretch(LEAF_LENGTH, LEAF_WIDTH, LEAF_THICKNESS,
		    shift(0, 0, -1.0, bind(leaf_extrusion)));
  reverseLeafpiece = stretch(1, -1, 1, bind(theLeafpiece));
  
  /* Create the leaf from two leaf shapes joined. */
  theLeaf = new_nothing();
  LINK(theLeaf, PLUS, rotate(Z, X, LEAF_BENDDOWN,
                      rotate(Z, Y,  LEAF_OUTSLOPE, bind(theLeafpiece))));
  LINK(theLeaf, PLUS, rotate(Z, X, LEAF_BENDDOWN,
		      rotate(Z, Y, -LEAF_OUTSLOPE, bind(reverseLeafpiece))));
  
  theLeaf = bind(theLeaf);
  theLeaf->material = logo_leaf_material;

  /* Make a coconut. */
  theCoconut = stretch(5.5, 5.5, 5.5, new_sphere());
  theCoconut->material = logo_coconut_material;
  
  /* Make the top of the tree. */
  theTree = new_nothing();
  LINK(theTree, PLUS, rotate(X, Y, 12.0,
                      shift(0.0, 0.0, 0.6*LEAF_THICKNESS,
                      rotate(Z, X, -4.2, rotate(Z, Y, -2.1, bind(theLeaf))))));
  LINK(theTree, PLUS, rotate(X, Y, 90.0,
                      shift(0.0, 0.0, 0.0*LEAF_THICKNESS, 
                      rotate(Z, Y, +1.8, bind(theLeaf)))));
  LINK(theTree, PLUS, rotate(X, Y, 175.0, 
                      shift(0.0, 0.0, 1.2*LEAF_THICKNESS, 
                      rotate(Z, X, +2.9, bind(theLeaf)))));
  LINK(theTree, PLUS, rotate(X, Y, 270.0, 
                      shift(0.0, 0.0, 0.3*LEAF_THICKNESS, 
                      rotate(Z, X, +5.3, bind(theLeaf)))));
  LINK(theTree, PLUS, rotate(X, Y, 130.0,
                      shift(9.0, 0.0, -7.5, bind(theCoconut))));
  LINK(theTree, PLUS, rotate(X, Y, 250.0,
                      shift(9.0, 0.0, -7.5, bind(theCoconut))));
  LINK(theTree, PLUS, rotate(X, Y, 384.0, 
                      shift(9.0, 0.0, -7.5, bind(theCoconut))));
  theTree = shift(0.0, 0.0, 4.0, bind(theTree));

  /* Make a cone that goes through radius T_BOTTOM_RADIUS, Z 0 
      and radius T_TOP_RADIUS, T_HEIGHT */
  theTrunkpiece = stretch(1.0, 1.0, TRUNKPIECE_HEIGHT/
                          (TRUNKPIECE_BOTTOM_RADIUS - TRUNKPIECE_TOP_RADIUS),
	          shift(0.0, 0.0, TRUNKPIECE_BOTTOM_RADIUS, new_cone()));
  LINK(theTrunkpiece, MINUS, shift(0.0, 0.0, TRUNKPIECE_HEIGHT, new_plane()));

  /* Make a cutter to smoothly cut off the bottom of the trunk piece */
  cutter = rotate(Z, X, 180.0, new_plane());
    /* Make the torus to leave behind on the bototm of the piece */
  tPieceSlope = atan2(TRUNKPIECE_BOTTOM_RADIUS - TRUNKPIECE_TOP_RADIUS,
                            TRUNKPIECE_HEIGHT);
  torus_center_x = TRUNKPIECE_BOTTOM_RADIUS -
                       cos(tPieceSlope)*TRUNKPIECE_TORUS_SMOOTH;
  torus_center_z = - sin(tPieceSlope)*TRUNKPIECE_TORUS_SMOOTH;
  
  torusrad = TRUNKPIECE_TORUS_SMOOTH / torus_center_x;
  aTorus = new_torus(torusrad);
  aTorus = shift(0.0, 0.0, torus_center_z,
           stretch(torus_center_x, torus_center_x, torus_center_x, aTorus));
  LINK(cutter, MINUS, bind(aTorus));
  
  /* Smooth the trunkpiece */
  LINK(theTrunkpiece, MINUS, bind(cutter));
  theTrunkpiece = bind(theTrunkpiece);

  theTrunkpiece->material = logo_trunk_material;
  
  for (i = 0; i < TREE_TRUNKPIECES; i++) {
    theTree = rotate(X, Y, TREE_TWIST,
	      rotate(Z, X, TREE_ROTATE,
	      shift(0.0, 0.0, TREE_HEIGHT/TREE_TRUNKPIECES, bind(theTree))));
    LINK(theTree, PLUS, bind(theTrunkpiece));
  }

  return bind(theTree);
}

// Functions for the PyramidTexture.
void PyramidTexture::apply(vector ray_from, vector ray_dir, uint ray_flags,
			   Material *properties, vector hit_point,
			   vector hit_normal, RayTracer *engine,
			   transform tex_transform)
{
  vector primitive_hitpoint, prim_normal;
  double texture_x, texture_y;
  int x_block, y_block;
  double x_block_frac, y_block_frac;
  double crack_dist, crack_amount;
  DiffuseComponent *diffuse;

  // Get the hitpoint and normal in primitive-space.
  Matrix::apply(tex_transform[AFT], hit_point, primitive_hitpoint);
  Matrix::applytranspose(tex_transform[AFT], hit_normal, prim_normal);

  // Determine the X and Y coordinates that we use to index the texture
  // on the pyramid's wall.
  if (fabs(prim_normal[X]) > fabs(prim_normal[Y])) {
    // This is a west- or east-facing wall. Index it by Y and Z.
    texture_x = primitive_hitpoint[Y] / block_size;
    texture_y = primitive_hitpoint[Z] / block_size;
  } else {
    // This is a north- or south-facing wall. Index it by X and Z.
    texture_x = primitive_hitpoint[X] / block_size;
    texture_y = primitive_hitpoint[Z] / block_size;
  }
  
  y_block = (int) floor(texture_y);
  y_block_frac = texture_y - y_block;

  // Stair-step pattern every second row.
  if (abs(y_block) & 0x01) {
    texture_x += 0.5;
  }

  x_block = (int) floor(texture_x);
  x_block_frac = texture_x - x_block;

  crack_dist = MIN(MIN(x_block_frac, 1 - x_block_frac),
		   MIN(y_block_frac, 1 - y_block_frac));
		   
  crack_amount = cubicInterpolate(1.0, 0.0, crack_dist * 10.0);

  diffuse = get_diffuse_component(properties);
  Color::mix(diffuse->diffuse_color, crack_col, crack_amount,
	     diffuse->diffuse_color);
}

void PyramidTexture::print_tex(void) {
  printf("PyramidTexture: block size %f, precedence %d\n",
	 block_size, precedence);
}
