/* antialias.c : Antialiasing functions.
** 
** This file defines the functions for the Antialiaser class.
**
** Chris Butcher, 16 January 1997
*/

#include "mirage.h"

#ifdef USE_NAMESPACES
namespace Mirage {
#endif

// The Antialiaser casts rays at the corners of each pixel and then
// localises any edges found by casting further rays along the sides
// of the pixels.
void Antialiaser::makePicture(Camera *theCamera,
			      primnode *trace_object)
{
  int x, y, i;
  color ray_colour;
  int redval, grnval, bluval;
  char enlarge_buf[MIRAGE_STRING_LENGTH];
  double enlarge_x, enlarge_y, enlarge_factor;

  // If the antialiasing level is < 0, then use a PixelSampler
  // instead and cast a ray through the center of each pixel.
  if (antialias_level < 0) {
    theCamera->mySampler = new PixelSampler;
    delete this;
    theCamera->mySampler->makePicture(theCamera, trace_object);
    return;
  }

  // Store the scene to be traced (used by get_colour).
  world_scene = trace_object;

  // Set up the variables that describe the picture that
  // we are going to generate.
  myCamera = theCamera;
  theStorage = theCamera->myStorage;
  image_x_size = theStorage->xsize;
  image_y_size = theStorage->ysize;
  theOptics = theCamera->myOptics;
  theEngine = theCamera->myEngine;
  camera_matrix = &(theCamera->world_transform[FORE]);
  my_primary_ray = theCamera->primary_ray;

  // Allocate space for the color data gathered from the first-pass
  // rays at the corners of each pixel. This is separate from
  // the Storage, which stores the final antialiased values.
  x_step = 1;
  y_step = image_x_size + 1;
  image_val = (unsigned char *) calloc((image_y_size + 1) * y_step *
				       num_colorvalues, sizeof(unsigned char));
  if (image_val == NULL) {
    char errbuf[MIRAGE_STRING_LENGTH];
    sprintf(errbuf, "can't allocate %d x %d image data", image_x_size,
	    image_y_size);
    fatal_error("Antialiaser::makePicture", errbuf,
		"You may be requesting a HUGE picture or just be very\n"
		"low on memory. This error should not really occur\n"
		"unless something is leaking memory over a long\n"
		"series of frames.\n");
  }

  // Set the image data to zero.
  memset((void *) image_val, 0,
	 (image_y_size + 1) * y_step * num_colorvalues *
	 sizeof(unsigned char));

  // Determine the image-plane defining attributes.
  if (get_keyfile_data("enlarged-view", enlarge_buf)) {
    if (sscanf(enlarge_buf, "%lf, %lf: x%lf",
	       &enlarge_x, &enlarge_y, &enlarge_factor) != 3) {
      fatal_error("Antialiaser::makePicture", "wrong enlargement format",
		  "Format for the \'enlarged-view\' parameter is:\n"
		  "enlarged-view        203, 102: x5");
      exit(-1);
    } else {
      x_center = image_x_size/2.0 -
                    enlarge_factor * (enlarge_x - image_x_size/2.0);
      y_center = image_y_size/2.0 -
                    enlarge_factor * (enlarge_y - image_y_size/2.0);
      image_scale = 2.0 / (image_x_size * enlarge_factor);
    }
  } else {
    x_center = image_x_size / 2.0;
    y_center = image_y_size / 2.0;
    image_scale = 1.0 / x_center;
  }

  // Find the SFX at the camera point.
  theCamera->find_camerapos_sfx(trace_object);

  // Set up the dummy variables for inserting all the primitive, t-value
  // etc. information into the Storage. This information is returned from
  // the raytracer, but is destroyed by the antialiasing process, which
  // currently only works on colour.
  dummy_ray_primnode = NULL;
  dummy_ray_t = 0;
  Vector::assign(dummy_ray_hit, 0, 0, 0);
  Vector::assign(dummy_ray_norm, 0, 0, 0);

  // Initialise the ray counters.
  counter_primary_rays = counter_shadow_rays = counter_secondary_rays = 0;
  counter_pixels_with_n_edges = (int *) calloc(4 * maximum_subdivision + 1,
					       sizeof(int));
  if (counter_pixels_with_n_edges == NULL) {
    char errbuf[MIRAGE_STRING_LENGTH];
    sprintf(errbuf, "Can't allocate stats memory for antialiasing level %d",
	    antialias_level);
    fatal_error("Antialiaser::makePicture", errbuf,
		"You may have the antialiasing level set insanely high.\n"
		"Sensible values are from 0 to 3. Regardless, this is\n"
		"not a large amount of memory and this error shouldn't\n"
		"really ever happen.");
  }
  for (i = 0; i < 4*maximum_subdivision + 1; i++)
    counter_pixels_with_n_edges[i] = 0;
  maximum_boundaries = 0;

  // Trace rays at each pixel corner.
  for(y = 0; y <= image_y_size; y++) {
    if (show_raytracing_progress && !completely_silent) {
      printf("Ray tracing, first pass: %04d\r", y);
      fflush(stdout);
    }

    for(x = 0; x <= image_x_size; x++) {
      get_colour(x, y, ray_colour);

      Colorvalue::compact(ray_colour, &image_val[(x + y * y_step) *
						num_colorvalues]);
    }
  }
  
  if (!silent && !completely_silent)
    print_current_time("> Antialiasing pass begun");

  // Store the number of initial rays (at the pixel corners).
  counter_initial_rays = counter_primary_rays;

  // Allocate space for the edge-of-pixel colour information.
  left_edge   = (unsigned char *) calloc((maximum_subdivision + 1) *
					 num_colorvalues,
				       sizeof(unsigned char));
  right_edge  = (unsigned char *) calloc((maximum_subdivision + 1) *
					 num_colorvalues,
				       sizeof(unsigned char));
  top_edge    = (unsigned char *) calloc((maximum_subdivision + 1) *
					 num_colorvalues,
				       sizeof(unsigned char));
  bottom_edge = (unsigned char *) calloc((maximum_subdivision + 1) *
					 num_colorvalues,
				       sizeof(unsigned char));
  row_above_bottom_edge = (unsigned char *) calloc((image_x_size + 1) *
						   (maximum_subdivision + 1) *
						   num_colorvalues,
						 sizeof(unsigned char));
  if (left_edge == NULL || right_edge == NULL  ||
      top_edge  == NULL || bottom_edge == NULL ||
      row_above_bottom_edge == NULL) {
    char errbuf[MIRAGE_STRING_LENGTH];
    sprintf(errbuf, "Can't allocate edge memory for antialiasing level %d",
	    antialias_level);
    fatal_error("Antialiaser::makePicture", errbuf,
		"You may have the antialiasing level set insanely high.\n"
		"Sensible values are from 0 to 3. Regardless, this is\n"
		"not a large amount of memory and this error shouldn't\n"
		"really ever happen.");
  }

  // Antialias each pixel in turn and place the antialiased values
  // into the Storage.
  for(y = 0; y < image_y_size; y++) {
    if (show_raytracing_progress && !completely_silent) {
      printf("Ray tracing, final pass: %04d\r", y);
      fflush(stdout);
    }

    for(x = 0; x < image_x_size; x++) {
      antialias_pixel(x, y, &redval, &grnval, &bluval);
    }
  }
  
  // If the user wants us to, display some statistics about rays.
  // This flag is read in from the keyfile toggle "ray-statistics" by
  // main.c.
  if (display_ray_statistics && !completely_silent) {
    counter_antialiasing_rays = counter_primary_rays - counter_initial_rays;
  
    printf("%d primary, %d secondary (of which %d were shadow) rays traced,\n",
	   counter_primary_rays, counter_secondary_rays, counter_shadow_rays);
    printf("%d initial rays, %d antialiasing rays.\n",
	   counter_initial_rays, counter_antialiasing_rays);

    printf("Antialiasing breakdown:  0 boundaries ... %8d rays\n",
	   counter_pixels_with_n_edges[0]);
    for(i = 1; i <= maximum_boundaries; i++)
      printf("                        %2d boundaries ... %8d rays\n",
	     i, counter_pixels_with_n_edges[i]);
  }

  // Deallocate the memory that we've allocated this frame.
  free(image_val);
  free(counter_pixels_with_n_edges);
  free(left_edge);
  free(right_edge);
  free(top_edge);
  free(bottom_edge);
  free(row_above_bottom_edge);
}

void Antialiaser::get_colour(double x_position, double y_position,
			     color ray_col)
{
  double image_x_point, image_y_point;
  vector ray_from, ray_dir, ray_norm, ray_hit;
  primnode *ray_prim;
  double ray_t;

  image_x_point = (x_position - x_center) * image_scale;
  image_y_point = (y_position - y_center) * image_scale;
  theOptics->generateRay(ray_from, ray_dir, image_x_point, image_y_point);

  // Apply the camera positioning matrix to the ray.
  Matrix::apply(*camera_matrix, ray_from, ray_from);
  Matrix::applynoshift(*camera_matrix, ray_dir, ray_dir);

  // Trace the ray generated by the optics model.
  //
  // ray_flags is my_primary_ray, light_t is 0.0 (not used as this isn't
  // a shadow ray).
  // The places to put return values are ray_col (colour), ray_t (the
  // t value of the hit), ray_norm (the object's normal), ray_hit (the
  // hit point in world coordinates), and ray_prim (the primnode that
  // was hit).
  theEngine->traceRay(world_scene, ray_from, ray_dir, my_primary_ray,
		      new_ray_signature(),
		      myCamera->camera_sfxtype, myCamera->camera_sfx,
		       0.0, dummy_ray_contribution,
		      &ray_prim, &ray_t, ray_norm, ray_hit, ray_col,
		      maximum_secondary_ray_depth);
}

void Antialiaser::antialias_pixel(int pixel_x, int pixel_y,
				  int *redval, int *grnval, int *bluval)
{
  int i, index, current_pixel;
  int number_of_boundaries, far_edge;
  int centroid_x_sum, centroid_y_sum, far_edge_sum;
  int current_triangle_side_length;
  int current_triangle_area;
  double total_area;
  color component_sum;
  color_value current_color;

  // Get the index of the current pixel in the image_val table.
  current_pixel = pixel_x * x_step + pixel_y * y_step;

  // First, subdivide the edges of the pixel if necessary.

  // ** Left edge.

  // Set up the colour values at the top and bottom of the left edge, equal
  // to the values of the corner rays.
  Colorvalue::set(&left_edge[0 * num_colorvalues],
		  &image_val[current_pixel * num_colorvalues]);
  Colorvalue::set(&left_edge[maximum_subdivision * num_colorvalues],
		  &image_val[(current_pixel + y_step) * num_colorvalues]);

  if (pixel_x == 0) {
    // We must subdivide the left edge as this is the leftmost pixel; it
    // doesn't share this edge with another pixel.

    // Initialise values to zero.
    for(i = 1; i < maximum_subdivision; i++)
      Colorvalue::zero(&left_edge[i * num_colorvalues]);

    // Subdivide to the given level of precision.
    subdivide_vertical_edge(pixel_x, pixel_y, pixel_y + 1.0, left_edge,
			    0, maximum_subdivision);

  } else {
    // The left edge is equal to the right edge of the last pixel, copy
    // that data.
    for (i = 0; i <= maximum_subdivision; i++)
      Colorvalue::set(&left_edge[i * num_colorvalues],
		      &right_edge[i * num_colorvalues]);
  }

  // ** Top edge.

  // Set up the colour values at the left and right of the top edge as
  // equal to the values of the corner rays.
  Colorvalue::set(&top_edge[0 * num_colorvalues],
		  &left_edge[0 * num_colorvalues]);
  Colorvalue::set(&top_edge[maximum_subdivision * num_colorvalues],
		  &image_val[(current_pixel + x_step) * num_colorvalues]);

  if (pixel_y == 0) {
    // We must subdivide the top edge as this is the topmost pixel; it
    // doesn't share this edge with another pixel.

    // Initialise values to zero.
    for(i = 1; i < maximum_subdivision; i++)
      Colorvalue::zero(&top_edge[i * num_colorvalues]);

    subdivide_horizontal_edge(pixel_y, pixel_x, pixel_x + 1.0, top_edge,
			      0, maximum_subdivision);

  } else {
    // The top edge is the bottom edge of the corresponding pixel
    // in the row above. Copy the data from the correct place in the
    // row above's stored bottom edge data table.
    for (i = 0; i <= maximum_subdivision; i++)
      Colorvalue::set(&top_edge[i * num_colorvalues],
		      &row_above_bottom_edge[(pixel_x*(maximum_subdivision+1) +
					      i) * num_colorvalues]);
  }

  // Initialise the bottom and right-edge in-between values to zero.
  for(i = 1; i < maximum_subdivision; i++) {
    Colorvalue::zero(&right_edge[i * num_colorvalues]);
    Colorvalue::zero(&bottom_edge[i * num_colorvalues]);
  }

  // ** Right edge.
  // We must always subdivide the right edge as this is the first time
  // we've considered this edge (the antialiaser sweeps from left to right).
  
  Colorvalue::set(&right_edge[0 * num_colorvalues],
		  &top_edge[maximum_subdivision * num_colorvalues]);
  Colorvalue::set(&right_edge[maximum_subdivision * num_colorvalues],
		  &image_val[(current_pixel + x_step + y_step) *
			    num_colorvalues]);

  // Subdivide the edge to the given precision.
  subdivide_vertical_edge(pixel_x + 1.0, pixel_y, pixel_y + 1.0, right_edge,
			  0, maximum_subdivision);

  // ** Bottom edge.
  // We must always subdivide the bottom edge as this is the first time
  // we've considered this edge (the antialiaser sweeps from top to bottom).

  // Set the end values to the values of the corresponding corner rays.
  Colorvalue::set(&bottom_edge[0 * num_colorvalues],
		  &left_edge[maximum_subdivision * num_colorvalues]);
  Colorvalue::set(&bottom_edge[maximum_subdivision * num_colorvalues],
		  &right_edge[maximum_subdivision * num_colorvalues]);

  // Subdivide the edge to the given precision.
  subdivide_horizontal_edge(pixel_y + 1.0, pixel_x, pixel_x + 1.0, bottom_edge,
			    0, maximum_subdivision);

  // If this isn't the bottom-most row, then this data will be needed as the
  // top edge of the row below. Store it in the stored edge data table at the
  // correct place.
  if (pixel_y < image_y_size) {
    for (i = 0; i <= maximum_subdivision; i++) {
      Colorvalue::set(&row_above_bottom_edge[(pixel_x * (maximum_subdivision
							+ 1) + i) *
					   num_colorvalues],
		      &bottom_edge[i * num_colorvalues]);
    }
  }

  // We have now subdivided the edges up to the desired subdivision level.
  // Count all of the boundaries between colours and find the centroid
  // of the boundaries [indirectly ... we don't find the centroid, but
  // instead the sum of the X-positions of the edges and the same
  // sum in Y].
  // The boundaries are flagged as being between the two colour values,
  // at position 2i+1 where the boundary colour value is i.
  //
  // Also note that we are storing 2* the position of the centroid, as the
  // boundaries occur between two values -> the position may not be a whole
  // number but 2p is guaranteed to be a whole number.

  // Note that we don't count the far ends of the edges (array element
  // maximum_subdivision) because due to the way that the subdivision
  // code flags colour values as being a boundary when they are on the
  // low side of a discontinuity, the end values can never be boundary values.

  // far_edge holds the (constant) X position of boundaries on the right edge,
  // equal to the Y position of boundaries on the bottom edge.
  far_edge = maximum_subdivision * 2;

  number_of_boundaries = 0;
  centroid_x_sum = 0;
  centroid_y_sum = 0;

  for(i = 0; i < maximum_subdivision; i++) {

    if (IS_BOUNDARY(&left_edge[i * num_colorvalues])) {
      number_of_boundaries++;
      // X position is 0, don't add to centroid X sum.
      centroid_y_sum += 2*i + 1;
    }

    if (IS_BOUNDARY(&top_edge[i * num_colorvalues])) {
      number_of_boundaries++;
      centroid_x_sum += 2*i + 1;
      // Y position is 0, don't add to centroid Y sum.
    }

    if (IS_BOUNDARY(&right_edge[i * num_colorvalues])) {
      number_of_boundaries++;
      centroid_x_sum += far_edge;
      centroid_y_sum += 2*i + 1;
    }

    if (IS_BOUNDARY(&bottom_edge[i * num_colorvalues])) {
      number_of_boundaries++;
      centroid_x_sum += 2*i + 1;
      centroid_y_sum += far_edge;
    }
  }

  // Work out the centroid sum that we would have had if the centroid
  // was on the far edge ... so centroid_x_sum / far_edge_sum is
  // a ratio from 0 to 1.
  far_edge_sum = far_edge * number_of_boundaries;

  // Increment the maximum number of boundaries in a pixel if we have
  // to.
  if (number_of_boundaries > maximum_boundaries)
    maximum_boundaries = number_of_boundaries;

  // Increment the appropriate ray counter.
  counter_pixels_with_n_edges[number_of_boundaries]++;

  // If no colour boundaries, don't try antialiasing, store the value
  // (which we can take from the top left corner as all the corners are
  // approximately the same).
  if (number_of_boundaries == 0) {
    *redval = left_edge[0] + left_edge[3];
    *grnval = left_edge[1] + left_edge[4];
    *bluval = left_edge[2] + left_edge[5];

    store_antialiased_pixel(pixel_x, pixel_y, left_edge);
    return;
  }

  // Iterate through the triangles that make up the antialiased pixel.
  // First set up some totals.
  current_triangle_side_length = 0;
  current_triangle_area = 0;
  total_area = 0;
  Color::set(component_sum, zero_color);

  // Start at the top left corner, then go right, down, left, up.
  Colorvalue::set(current_color, top_edge);

  // Traverse the top edge from left to right.
  for(i = 0; i < maximum_subdivision; i++) {
    if (IS_BOUNDARY(&top_edge[i * num_colorvalues])) {
      // Work out the area of the triangle that we've just finished.
      // Note that the vertex is actually 1 edge unit past the point
      // that is marked as BOUNDARY.
      current_triangle_area += (current_triangle_side_length + 1)
	* centroid_y_sum;
      
      // Update the total area and colour sums.
      total_area += current_triangle_area;
      for (index = 0; index < num_color_components; index++)
	component_sum[index] += current_color[index] * current_triangle_area;

      // Set up for the new triangle. Note that the new triangle already
      // has 1 edge unit worth of side past the vertex to the next
      // sampled edge[] point.
      Colorvalue::set(current_color, &top_edge[(i+1) * num_colorvalues]);
      current_triangle_side_length = 1;
      current_triangle_area = 0;
    } else {
      // This isn't a boundary point ... update the side length that
      // we've traversed so far.
      current_triangle_side_length += 2;
    }
  }

  // Work out the area of the current triangle (that stops at the top right
  // corner) and start a new area segment going down from the top right corner.
  current_triangle_area += current_triangle_side_length * centroid_y_sum;
  current_triangle_side_length = 0;

  // Traverse the right edge from top to bottom.
  for(i = 0; i < maximum_subdivision; i++) {
    if (IS_BOUNDARY(&right_edge[i * num_colorvalues])) {
      // Work out the area of the triangle that we've just finished.
      // Note that the vertex is actually 1 edge unit past the point
      // that is marked as BOUNDARY.
      current_triangle_area += (current_triangle_side_length + 1)
	* (far_edge_sum - centroid_x_sum);
      
      // Update the total area and colour sums.
      total_area += current_triangle_area;
      for (index = 0; index < num_color_components; index++)
	component_sum[index] += current_color[index] * current_triangle_area;

      // Set up for the new triangle. Note that the new triangle already
      // has 1 edge unit worth of side past the vertex to the next
      // sampled edge[] point.
      Colorvalue::set(current_color, &right_edge[(i+1) * num_colorvalues]);
      current_triangle_side_length = 1;
      current_triangle_area = 0;
    } else {
      // This isn't a boundary point ... update the side length that
      // we've traversed so far.
      current_triangle_side_length += 2;
    }
  }

  // Work out the area of the current triangle (that stops at the bottom right
  // corner) and start a new area segment going left from the bottom 
  // right corner.
  current_triangle_area += current_triangle_side_length
    * (far_edge_sum - centroid_x_sum);
  current_triangle_side_length = 0;

  // Traverse the bottom edge from right to left.
  for(i = maximum_subdivision - 1; i >= 0; i--) {
    if (IS_BOUNDARY(&bottom_edge[i * num_colorvalues])) {
      // Work out the area of the triangle that we've just finished.
      // Note that the vertex is actually 1 edge unit before the point
      // that is marked as BOUNDARY, however we maintain current_side_length
      // as the length up to the previous point for reverse traverses, 
      // so the actual side length up to the vertex is current_length + 1.
      current_triangle_area += (current_triangle_side_length + 1)
	* (far_edge_sum - centroid_y_sum);
      
      // Update the total area and colour sums.
      total_area += current_triangle_area;
      for (index = 0; index < num_color_components; index++)
	component_sum[index] += current_color[index] * current_triangle_area;

      // Set up for the new triangle. Note that the new triangle already
      // has 1 edge unit worth of side past the vertex to the current
      // point bottom_edge[i]. This will be the previous point for the next
      // loop so we set current_side_length to 1.
      Colorvalue::set(current_color, &bottom_edge[i * num_colorvalues]);
      current_triangle_side_length = 1;
      current_triangle_area = 0;
    } else {
      // This isn't a boundary point ... update the side length that
      // we've traversed so far.
      current_triangle_side_length += 2;
    }
  }

  // Work out the area of the current triangle (that stops at the bottom right
  // corner) and start a new area segment going left from the bottom 
  // right corner.
  current_triangle_area += current_triangle_side_length
    * (far_edge_sum - centroid_y_sum);
  current_triangle_side_length = 0;

  // Traverse the left edge from bottom to top.
  for(i = maximum_subdivision - 1; i >= 0; i--) {
    if (IS_BOUNDARY(&left_edge[i * num_colorvalues])) {
      // Work out the area of the triangle that we've just finished.
      // Note that the vertex is actually 1 edge unit past the point
      // that is marked as BOUNDARY, however we maintain current_side_length
      // as the length up to the previous point for reverse traverses, 
      // so the actual side length up to the vertex is current_length + 1.
      current_triangle_area += (current_triangle_side_length + 1)
	* centroid_x_sum;
      
      // Update the total area and colour sums.
      total_area += current_triangle_area;
      for (index = 0; index < num_color_components; index++)
	component_sum[index] += current_color[index] *
	  current_triangle_area;

      // Set up for the new triangle. Note that the new triangle already
      // has 1 edge unit worth of side past the vertex to the current
      // point left_edge[i]. This will be the previous point for the next
      // loop so we set current_side_length to 1.
      Colorvalue::set(current_color, &left_edge[i * num_colorvalues]);
      current_triangle_side_length = 1;
      current_triangle_area = 0;
    } else {
      // This isn't a boundary point ... update the side length that
      // we've traversed so far.
      current_triangle_side_length += 2;
    }
  }


  // Add in the contribution of the final triangle, which finishes at the
  // top left corner.
  current_triangle_area += current_triangle_side_length * centroid_x_sum;
  
  total_area += current_triangle_area;
  for (index = 0; index < num_color_components; index++)
    component_sum[index] += current_color[index] * current_triangle_area;
  

  // Work out the antialiased colour.
  if (total_area == 0)
    fatal_error("Antialiaser::antialias_pixel", "total_area == 0!",
		"This error should not happen.");

  Color::mult(1.0 / (total_area * 255), component_sum, component_sum);

  store_antialiased_pixel(pixel_x, pixel_y, component_sum);

  *redval = (int) (255 * (component_sum[R] + component_sum[ER]));
  *bluval = (int) (255 * (component_sum[G] + component_sum[EG]));
  *grnval = (int) (255 * (component_sum[B] + component_sum[EB]));
}

// Store the final antialiased color value in the Camera's Storage.
void Antialiaser::store_antialiased_pixel(int pixel_x, int pixel_y,
					  color_value final_color)
{
  Colorvalue::extract(final_color, antialiased_color);
  theStorage->store(pixel_x, pixel_y, dummy_ray_primnode, dummy_ray_t,
		    dummy_ray_norm, dummy_ray_hit, antialiased_color);
}

// Store the final antialiased color value in the Camera's Storage.
void Antialiaser::store_antialiased_pixel(int pixel_x, int pixel_y,
					  color final_color)
{
  theStorage->store(pixel_x, pixel_y, dummy_ray_primnode, dummy_ray_t,
		    dummy_ray_norm, dummy_ray_hit, final_color);
}

// Subdivide a vertical edge on the image plane casting rays to localise
// color discontinuities.
void Antialiaser::subdivide_vertical_edge(double edge_x, double y_min,
					  double y_max, unsigned char *edge,
					  int sub_min, int sub_max)
{
  int sub_mid;
  double y_mid;
  color mid_col;

  // Check to see if we need to subdivide this edge.
  if (Colorvalue::same_color(&edge[sub_min * num_colorvalues],
			     &edge[sub_max * num_colorvalues]))
    return;

  // The colours are different, we need to subdivide further.

  // If we've reached the bottom level of subdivision, stop and mark the
  // colour value on the low side of the discontinuity as a boundary.
  if (sub_max - sub_min == 1) {
    edge[sub_min * num_colorvalues + FLAGINDEX] |= BOUNDARY_VAL;
    return;
  }

  // Find the midpoint, get its colour by casting a ray and subdivide further.
  sub_mid = (sub_min + sub_max) >> 1;
  y_mid   = (y_min   + y_max  )  / 2;

  get_colour(edge_x, y_mid, mid_col);
  Colorvalue::compact(mid_col, &edge[sub_mid * num_colorvalues]);

  subdivide_vertical_edge(edge_x, y_min, y_mid, edge, sub_min, sub_mid);
  subdivide_vertical_edge(edge_x, y_mid, y_max, edge, sub_mid, sub_max);
}

// Subdivide a horizontal edge on the image plane casting rays to localise
// color discontinuities.
void Antialiaser::subdivide_horizontal_edge(double edge_y, double x_min,
					    double x_max, unsigned char *edge,
					    int sub_min, int sub_max)
{
  int sub_mid;
  double x_mid;
  color mid_col;

  // Check to see if we need to subdivide this edge.
  if (Colorvalue::same_color(&edge[sub_min * num_colorvalues],
			     &edge[sub_max * num_colorvalues]))
    return;

  // The colours are different, we need to subdivide further.

  // If we've reached the bottom level of subdivision, stop and mark the
  // colour value on the low side of the discontinuity as a boundary.
  if (sub_max - sub_min == 1) {
    edge[sub_min * num_colorvalues + FLAGINDEX] |= BOUNDARY_VAL;
    return;
  }

  // Find the midpoint, get its colour by casting a ray and subdivide further.
  sub_mid = (sub_min + sub_max) >> 1;
  x_mid   = (x_min   + x_max  )  / 2;

  get_colour(x_mid, edge_y, mid_col);
  Colorvalue::compact(mid_col, &edge[sub_mid * num_colorvalues]);

  subdivide_horizontal_edge(edge_y, x_min, x_mid, edge, sub_min, sub_mid);
  subdivide_horizontal_edge(edge_y, x_mid, x_max, edge, sub_mid, sub_max);
}












#ifdef XWINDOWS_ON
// All of the following code deals with the X11 interface. If this
// is not working on your system, you may wish to remove it by commenting
// out the XWINDOWS_ON definition in mirage.h.

// The DisplayAntialiaser is the same as the Antialiaser except it
// casts initial rays in a quadtree-like pattern and displays the results in
// an X window.
void DisplayAntialiaser::makePicture(Camera *theCamera,
				     primnode *trace_object)
{
  int x, y, i;
  color ray_colour;
  int r, g, b;
  unsigned char *val_ptr;
  double enlarge_x, enlarge_y, enlarge_factor;
  char enlarge_buf[MIRAGE_STRING_LENGTH];

  // If the antialiasing level is < 0, then use a PixelSampler
  // instead and cast a ray through the center of each pixel. Note that this
  // means that no display is performed; possibly a bad thing. However, using
  // an antialias level of -1 is usually for a specialised purpose anyway,
  // and won't require X display.
  if (antialias_level < 0) {
    theCamera->mySampler = new PixelSampler;
    delete this;
    theCamera->mySampler->makePicture(theCamera, trace_object);
    return;
  }

  // Store the scene to be traced (used by get_colour).
  world_scene = trace_object;

  // Set up the variables that describe the picture that
  // we are going to generate.
  theStorage = theCamera->myStorage;
  image_x_size = theStorage->xsize;
  image_y_size = theStorage->ysize;
  theOptics = theCamera->myOptics;
  theEngine = theCamera->myEngine;
  camera_matrix = &(theCamera->world_transform[FORE]);
  my_primary_ray = theCamera->primary_ray;

  // Initialise the X-Windows display.
  theDisplay = new XDisplay(image_x_size, image_y_size, NULL);

  // Allocate space for the color data from the initial rays (separate from
  // final values, which are placed in the camera's Storage).
  x_step = 1;
  y_step = image_x_size + 1;
  image_val = (unsigned char *) calloc(((image_y_size + 1) * y_step) *
				       num_colorvalues,
				       sizeof(unsigned char));
  if (image_val == NULL) {
    char errbuf[MIRAGE_STRING_LENGTH];
    sprintf(errbuf, "can't allocate %d x %d image data", image_x_size,
	    image_y_size);
    fatal_error("DisplayAntialiaser::makePicture", errbuf,
		"You may be requesting a HUGE picture or just be very\n"
		"low on memory. This error should not really occur\n"
		"unless something is leaking memory over a long\n"
		"series of frames.\n");
  }

  // Initialise the color data to not done. Note that this is necessary so that
  // the quadtree ray-casting process knows that it must cast rays at these
  // pixels.
  for(y = 0; y <= image_y_size; y++) {
    val_ptr = &image_val[(y * y_step) * num_colorvalues + FLAGINDEX];
    for(x = 0; x <= image_x_size; x++) {
      *val_ptr |= NOT_DONE;
      val_ptr += num_colorvalues;
    }
  }

  // Determine the image-plane defining attributes.
  if (get_keyfile_data("enlarged-view", enlarge_buf)) {
    if (sscanf(enlarge_buf, "%lf, %lf: x%lf",
	       &enlarge_x, &enlarge_y, &enlarge_factor) != 3) {
      fatal_error("DisplayAntialiaser::makePicture",
		  "wrong enlargement format",
		  "Format for the \'enlarged-view\' parameter is:\n"
		  "enlarged-view        203, 102: x5");
      exit(-1);
    } else {
      x_center = image_x_size/2.0 -
                    enlarge_factor * (enlarge_x - image_x_size/2.0);
      y_center = image_y_size/2.0 -
                    enlarge_factor * (enlarge_y - image_y_size/2.0);
      image_scale = 2.0 / (image_x_size * enlarge_factor);
    }
  } else {
    x_center = image_x_size / 2.0;
    y_center = image_y_size / 2.0;
    image_scale = 1.0 / x_center;
  }

  // Find the SFX at the camera point.
  theCamera->find_camerapos_sfx(trace_object);

  // Set up the dummy variables for inserting all the primitive, t-value
  // etc. information into the Storage. This information is returned from
  // the raytracer, but is destroyed by the antialiasing process, which
  // currently only works on colour.
  dummy_ray_primnode = NULL;
  dummy_ray_t = 0;
  Vector::assign(dummy_ray_hit, 0, 0, 0);
  Vector::assign(dummy_ray_norm, 0, 0, 0);

  // Initialise the ray counters.
  counter_primary_rays = counter_shadow_rays = counter_secondary_rays = 0;
  counter_pixels_with_n_edges = (int *) calloc(4 * maximum_subdivision + 1,
					       sizeof(int));
  if (counter_pixels_with_n_edges == NULL) {
    char errbuf[MIRAGE_STRING_LENGTH];
    sprintf(errbuf, "Can't allocate stats memory for antialiasing level %d",
	    antialias_level);
    fatal_error("DisplayAntialiaser::makePicture", errbuf,
		"You may have the antialiasing level set insanely high.\n"
		"Sensible values are from 0 to 3. Regardless, this is\n"
		"not a large amount of memory and this error shouldn't\n"
		"really ever happen.");
  }
  for (i = 0; i < 4*maximum_subdivision + 1; i++)
    counter_pixels_with_n_edges[i] = 0;
  maximum_boundaries = 0;

  // Trace rays at each pixel corner, performing an iterative approximation.
  // This uses a BlockQueue structure to store all of the blocks that need to
  // be subdivided, then adds the top-level whole picture block, then tells
  // the BlockQueue to process it.
  theQueue = new BlockQueue(this, (image_x_size+1) * (image_y_size+1));
  theQueue->add(0, 0, image_x_size, image_y_size);
  theQueue->process();
  delete theQueue;
  
  if (!silent)
    print_current_time("> Antialiasing pass begun");

  // Store the number of initial rays (at the pixel corners).
  counter_initial_rays = counter_primary_rays;

  // Allocate space for the edge-of-pixel colour information.
  left_edge   = (unsigned char *) calloc((maximum_subdivision + 1) *
					 num_colorvalues,
				       sizeof(unsigned char));
  right_edge  = (unsigned char *) calloc((maximum_subdivision + 1) *
					 num_colorvalues,
				       sizeof(unsigned char));
  top_edge    = (unsigned char *) calloc((maximum_subdivision + 1) *
					 num_colorvalues,
				       sizeof(unsigned char));
  bottom_edge = (unsigned char *) calloc((maximum_subdivision + 1) *
					 num_colorvalues,
				       sizeof(unsigned char));
  row_above_bottom_edge = (unsigned char *) calloc((image_x_size + 1) *
						   (maximum_subdivision + 1) *
						   num_colorvalues,
						 sizeof(unsigned char));
  if (left_edge == NULL || right_edge == NULL  ||
      top_edge  == NULL || bottom_edge == NULL ||
      row_above_bottom_edge == NULL) {
    char errbuf[MIRAGE_STRING_LENGTH];
    sprintf(errbuf, "Can't allocate edge memory for antialiasing level %d",
	    antialias_level);
    fatal_error("DisplayAntialiaser::makePicture", errbuf,
		"You may have the antialiasing level set insanely high.\n"
		"Sensible values are from 0 to 3. Regardless, this is\n"
		"not a large amount of memory and this error shouldn't\n"
		"really ever happen.");
  }

  // Antialias each pixel in turn and place the antialiased values
  // into the Storage. Draw on screen as well.
  for(y = 0; y < image_y_size; y++) {
    if (show_raytracing_progress && !completely_silent) {
      printf("Ray tracing, final pass: %04d\r", y);
      fflush(stdout);
    }

    for(x = 0; x < image_x_size; x++) {
      // Flash the pixel under consideration white.
      theDisplay->colour_pixel(x, y, 1, 1, 1);

      // Antialias the pixel.
      antialias_pixel(x, y, &r, &g, &b);
      theDisplay->colour_pixel(x, y, r, g, b);
    }
  }
  
  // If the user wants us to, display some statistics about rays.
  if (display_ray_statistics && !completely_silent) {
    counter_antialiasing_rays = counter_primary_rays - counter_initial_rays;
  
    printf("%d primary, %d secondary (of which %d were shadow) rays traced,\n",
	   counter_primary_rays, counter_secondary_rays, counter_shadow_rays);
    printf("%d initial rays, %d antialiasing rays.\n",
	   counter_initial_rays, counter_antialiasing_rays);

    printf("Antialiasing breakdown:  0 boundaries ... %8d rays\n",
	   counter_pixels_with_n_edges[0]);
    for(i = 1; i <= maximum_boundaries; i++)
      printf("                        %2d boundaries ... %8d rays\n",
	     i, counter_pixels_with_n_edges[i]);
  }

  // Deallocate the memory that we've allocated this frame.
  delete theDisplay;
  free(image_val);
  free(counter_pixels_with_n_edges);
  free(left_edge);
  free(right_edge);
  free(top_edge);
  free(bottom_edge);
  free(row_above_bottom_edge);
}

// End of the X11-specific code.
#endif



#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif
