/* antialias.h : Header file for the Antialiaser.
** 
** This file defines the headers and definitions for the antialiasing
** subsystem.
**
** Chris Butcher, 16 January 1997
*/

// In the edge tables, a color_value marks a boundary between colours if it has
// the BOUNDARY_VAL flag set.
#define BOUNDARY_VAL         ((unsigned char) (1 << 7))
#define IS_BOUNDARY(val)     ((val)[FLAGINDEX] & BOUNDARY_VAL)
#define NOT_DONE             ((unsigned char) (1 << 6))
#define IS_DONE(val)         (((val)[FLAGINDEX] & NOT_DONE) == 0)

// The Antialiaser casts rays at the corners of each pixel and then
// localises any edges found by casting further rays along the sides
// of the pixels.
class Antialiaser : public PictureSampler {
public:
  // Should we be silent?
  int silent;

  // Size of the image.
  int image_x_size, image_y_size;

  // Ray casting variables.
  double x_center, y_center, image_scale;
  unsigned int my_primary_ray;

  // Dummy (primary ray has 100% significance) ray_contribution color to pass
  // to the raytracer and the final antialiased color.
  color dummy_ray_contribution, antialiased_color;

  // Variables that define the antialiasing level.
  int antialias_level, maximum_subdivision, maximum_boundaries;

  // Components of the camera.
  Camera *myCamera;
  Storage *theStorage;
  OpticsModel *theOptics;
  RayTracer *theEngine;
  matrix *camera_matrix;

  // Array of color values from rays cast at the corners of each pixel and
  // indices to index it.
  unsigned char *image_val;
  int x_step, y_step;

  // Color values along the edge of the currently considered pixel.
  unsigned char *left_edge, *top_edge, *right_edge, *bottom_edge;
  unsigned char *row_above_bottom_edge;

  // The number of pixels that have had n boundaries in them.
  int *counter_pixels_with_n_edges;

  // The world-scene to be raytraced.
  primnode *world_scene;

  // Dummy variables for storing discarded return values from the raytracer.
  primnode *dummy_ray_primnode;
  double dummy_ray_t;
  vector dummy_ray_norm, dummy_ray_hit;

  // Create an antialiased picture.
  virtual void makePicture(class Camera *generating,
			   primnode *trace_object);

  // Prototypes for routines used within the antialiasing process.
  void subdivide_vertical_edge(double edge_x, double y_min,
			       double y_max, unsigned char *edge,
			       int sub_min, int sub_max);
  void subdivide_horizontal_edge(double edge_y, double x_min,
				 double x_max, unsigned char *edge,
				 int sub_min, int sub_max);
  void store_antialiased_pixel(int pixel_x, int pixel_y,
			       color_value final_color);
  void store_antialiased_pixel(int pixel_x, int pixel_y,
			       color final_color);
  void antialias_pixel(int pixel_x, int pixel_y, int *r, int *g, int *b);
  void get_colour(double x_position, double y_position, color ray_col);

  // Constructor that doesn't specify a subdivision level; it is read from
  // the keyfile.
  Antialiaser(void) {
    char antialiasbuf[MIRAGE_STRING_LENGTH];

    silent = FALSE;
    if (!get_keyfile_data("antialiasing", antialiasbuf)) {
      fatal_error("Antialiaser::constructor", "antialias level not specified!",
		  "You must either initialise an Antialiaser giving it a\n"
		  "specific level of antialiasing, or else set the default\n"
		  "level in the keyfile using the antialiasing key, e.g.\n"
		  "antialiasing               2");
      exit(-1);
    }

    if (sscanf(antialiasbuf, "%d", &antialias_level) != 1) {
      fatal_error("Antialiaser::constructor", "wrong antialiasing format",
		  "The syntax for the antialiasing key is e.g.\n"
		  "antialiasing               2");
      exit(-1);
    }

    maximum_subdivision = (1 << antialias_level);

    // Set up a dummy color that represents the fact that primary
    // rays count for the entire color returned.
    Color::set(dummy_ray_contribution, full_color);
  }

  // Constructor specifies the maximum subdivision level to antialias edges
  // of the pixels to.
  Antialiaser(int my_level) {
    silent = FALSE;
    antialias_level = my_level;
    maximum_subdivision = (1 << my_level);

    // Set up a dummy color that represents the fact that primary
    // rays count for the entire color returned.
    Color::set(dummy_ray_contribution, full_color);
  }

  // Constructor that also specifies the silent flag.
  Antialiaser(int my_level, int be_silent) {
    silent = be_silent;
    antialias_level = my_level;
    maximum_subdivision = (1 << my_level);

    // Set up a dummy color that represents the fact that primary
    // rays count for the entire color returned.
    Color::set(dummy_ray_contribution, full_color);
  }
};

#ifdef XWINDOWS_ON
// All of the following code deals with the X11 interface. If this
// is not working on your system, you may wish to remove it by commenting
// out the XWINDOWS_ON definition in mirage.h.

// The DisplayAntialiaser acts just like the Antialiaser except that the
// initial rays are cast in an quadtree-like pattern and the results are
// displayed in an X window on the screen.
class DisplayAntialiaser : public Antialiaser {
public:
  XDisplay *theDisplay;
  BlockQueue *theQueue;

  virtual void makePicture(class Camera *generating,
			   primnode *trace_object);

  DisplayAntialiaser(int my_level) {
    silent = FALSE;
    antialias_level = my_level;
    maximum_subdivision = (1 << my_level);

    // Set up a dummy color that represents the fact that primary
    // rays count for the entire color returned.
    Color::set(dummy_ray_contribution, full_color);
  }

  DisplayAntialiaser(int my_level, int be_silent) {
    silent = be_silent;
    antialias_level = my_level;
    maximum_subdivision = (1 << my_level);

    // Set up a dummy color that represents the fact that primary
    // rays count for the entire color returned.
    Color::set(dummy_ray_contribution, full_color);
  }
};

// End of the X11-specific code.
#endif
