/* blockqueue.c : Functions to maintain the BlockQueue.
** 
** This file defines the functions for the BlockQueue class, which is used
** by the DisplayAntialiaser to perform a progressive approximation to the
** raytraced picture.
**
** Chris Butcher, 21 January 1997
*/

#include "mirage.h"

#ifdef USE_NAMESPACES
namespace Mirage {
#endif

#ifdef XWINDOWS_ON
// All of the following code deals with the X11 interface. If this
// is not working on your system, you may wish to remove it by commenting
// out the XWINDOWS_ON definition in mirage.h.

void BlockQueue::process(void)
{
  int x_min, y_min, x_max, y_max, x_mid, y_mid;
  color_value ne_val, nw_val, se_val, sw_val;
  int block_r, block_g, block_b, i;
  color ray_col;

  // Keep retrieving blocks from the queue until there are none left.
  while (retrieve(&x_min, &y_min, &x_max, &y_max)) {
    // Get the colour of the NW corner.
    Colorvalue::set(nw_val, &theAA->image_val[(x_min + y_min * theAA->y_step) *
					     num_colorvalues]);
    if (!IS_DONE(nw_val)) {
      // We need to trace a ray here.
      theAA->get_colour(x_min, y_min, ray_col);

      // What is this as a color_value?
      Colorvalue::compact(ray_col, nw_val);

      // Store the returned colour.
      Colorvalue::set(&theAA->image_val[(x_min + y_min * theAA->y_step) *
				       num_colorvalues], nw_val);
    }
    
    // Similarly for the NE corner.
    Colorvalue::set(ne_val, &theAA->image_val[(x_max + y_min * theAA->y_step) *
					     num_colorvalues]);
    if (!IS_DONE(ne_val)) {
      theAA->get_colour(x_max, y_min, ray_col);
      
      // What is this as a color_value?
      Colorvalue::compact(ray_col, ne_val);

      Colorvalue::set(&theAA->image_val[(x_max + y_min * theAA->y_step) *
				       num_colorvalues], ne_val);
    }
    
    // Similarly for the SW corner.
    Colorvalue::set(sw_val, &theAA->image_val[(x_min + y_max * theAA->y_step) *
					     num_colorvalues]);
    if (!IS_DONE(sw_val)) {
      theAA->get_colour(x_min, y_max, ray_col);
      
      // What is this as a color_value?
      Colorvalue::compact(ray_col, sw_val);

      Colorvalue::set(&theAA->image_val[(x_min + y_max * theAA->y_step) *
				       num_colorvalues], sw_val);
    }
    
    // Similarly for the SE corner.
    Colorvalue::set(se_val, &theAA->image_val[(x_max + y_max * theAA->y_step) *
					     num_colorvalues]);
    if (!IS_DONE(se_val)) {
      theAA->get_colour(x_max, y_max, ray_col);
      
      // What is this as a color_value?
      Colorvalue::compact(ray_col, se_val);

      Colorvalue::set(&theAA->image_val[(x_max + y_max * theAA->y_step) *
				       num_colorvalues], se_val);
    }
    
    // Work out the displayed colour.
    block_r = (nw_val[R]  + ne_val[R]  + sw_val[R]  + se_val[R]  +
	       nw_val[ER] + ne_val[ER] + sw_val[ER] + se_val[ER]) / 4;
    block_g = (nw_val[G]  + ne_val[G]  + sw_val[G]  + se_val[G]  +
	       nw_val[EG] + ne_val[EG] + sw_val[EG] + se_val[EG]) / 4;
    block_b = (nw_val[B]  + ne_val[B]  + sw_val[B]  + se_val[B]  +
	       nw_val[EB] + ne_val[EB] + sw_val[EB] + se_val[EB]) / 4;

    x_mid = (x_max + x_min) >> 1;
    y_mid = (y_max + y_min) >> 1;

    if (x_mid == x_min && y_mid == y_min) {
      // This is a single pixel.
      theAA->theDisplay->colour_pixel(x_min, y_min,
				      255 - block_r,
				      255 - block_g,
				      255 - block_b);
      theAA->theDisplay->colour_pixel(x_min, y_min,
				      block_r, block_g, block_b);
    } else {
      // This is a block.
      theAA->theDisplay->colour_block(x_min, y_min, x_max, y_max,
				      255 - block_r,
				      255 - block_g,
				      255 - block_b);
      theAA->theDisplay->colour_block(x_min, y_min, x_max, y_max,
				      block_r, block_g, block_b);
    }
    
    // Add the subdivided blocks to the queue if necessary.
    if (x_mid == x_min) {
      // This block is 1 pixel wide.
      if (y_mid == y_min) {
	// This was an individual pixel, do nothing.
      } else {
	// Subdivide in Y.
	add(x_min, y_min, x_max, y_mid);
	add(x_min, y_mid, x_max, y_max);
      }
    } else {
      // This block needs to be subdivided in X.
      if (y_mid == y_min) {
	// One pixel high, subdivided in X only.
	add(x_min, y_min, x_mid, y_max);
	add(x_mid, y_min, x_max, y_max);
      } else {
	// This needs to be subdivided in X and in Y.
	add(x_min, y_min, x_mid, y_mid);
	add(x_mid, y_min, x_max, y_mid);
	add(x_min, y_mid, x_mid, y_max);
	add(x_mid, y_mid, x_max, y_max);
      }
    }
  }
}

// End of the X11-specific code.
#endif



#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif
