/* blockqueue.h : Header file for the BlockQueue.
** 
** This file defines the headers and definitions for the BlockQueue,
** which is used by the DisplayAntialiaser to store blocks of the image to
** be subdivided + raytraced + displayed.
**
** Chris Butcher, 21 January 1997
*/

#ifdef XWINDOWS_ON
// All of the following code deals with the X11 interface. If this
// is not working on your system, you may wish to remove it by commenting
// out the XWINDOWS_ON definition in mirage.h.

// A block_element has X and Y coordinates for the corners of the block.
typedef struct {
  int x_min;
  int y_min;
  int x_max;
  int y_max;
} block_element;

// The BlockQueue stores a set of blocks and subdivides them when processing.
class BlockQueue {
public:
  block_element *blockStorage;
  int head, tail, empty;
  int queue_size;
  class DisplayAntialiaser *theAA;

  // Initialise a new BlockQueue for the given Antialiaser, maximum number
  // of blocks equal to the_size.
  BlockQueue(class DisplayAntialiaser *theAntialiaser, int the_size) {
    queue_size = the_size;
    theAA = theAntialiaser;

    // Allocate memory for the block_elements.
    blockStorage = (block_element *) calloc(queue_size, sizeof(block_element));
    
    if (blockStorage == NULL) {
      char errbuf[MIRAGE_STRING_LENGTH];

      sprintf(errbuf, "can't allocate blockQueue %d elements long", the_size);
      fatal_error("BlockQueue::constructor", errbuf,
		  "Either a very large picture is being made (hence the\n"
		  "very large BlockQueue needed to store the progressive\n"
		  "refinement) or you are seriously low on memory.");
    }

    // Initially the queue is empty, next element goes at 0.
    head = 0;
    tail = -1;
    empty = TRUE;
  }

  ~BlockQueue(void) {
    // Deallocate the block_element memory.
    delete blockStorage;
  }

  // Add a block to the queue.
  void add(int x0, int y0, int x1, int y1) {
    tail = (tail + 1) % queue_size;
    if (!empty && tail == head) {
      fatal_error("BlockQueue::add", "queue has overflowed!",
		  "This error should not happen.");
    }

    blockStorage[tail].x_min = x0;
    blockStorage[tail].y_min = y0;
    blockStorage[tail].x_max = x1;
    blockStorage[tail].y_max = y1;

    empty = FALSE;
  }

  // Get the next block from the queue (returns TRUE if successful,
  // FALSE if empty.
  int retrieve(int *x0, int *y0, int *x1, int *y1) {
    if (empty)
      return FALSE;

    if (head == tail)
      empty = TRUE;

    *x0 = blockStorage[head].x_min;
    *y0 = blockStorage[head].y_min;
    *x1 = blockStorage[head].x_max;
    *y1 = blockStorage[head].y_max;

    head = (head + 1) % queue_size;

    return TRUE;
  }

  // Process each block in the queue in turn, raytracing, subdividing and
  // displaying it. Subdivided blocks are placed on the tail of the queue
  // for later processing.
  void process(void);
};

// End of X11-specific code.
#endif
