/* camera.c : The camera functions.
** 
** This file defines the functions for the Camera class and some
** component subclasses.
**
** Chris Butcher, 29 November 1996
*/

#include "mirage.h"

#ifdef USE_NAMESPACES
namespace Mirage {
#endif

// Maximum depth of secondary rays to cast
int maximum_secondary_ray_depth;

// The generateImage() function samples the worldscene and produces
// output in the framebuffer.
void Camera::generateImage(primnode *trace_object)
{
  myStorage->allocate();
  mySampler->makePicture(this, trace_object);
  if (myProcess != NULL)
    myProcess->postprocess(this);
}

// The generateRay() procedure generates the ray through the camera's
// eyepoint (the origin) to the correct place on the image plane
// (which is the point image_x, image_y, viewplane_dist).
void PinholeOptics::generateRay(vector ray_from, vector ray_dir,
				double image_x, double image_y)
{
  // All rays cast from eyepoint at origin of camera.
  Vector::assign(ray_from, 0.0, 0.0, 0.0);

  // The camera is pointing along the X axis, with its up-vector oriented
  // towards Z. This means that the X pixel axis corresponds to -Y
  // and the Y pixel axis -Z. All rays are thus cast through the point
  // (viewplane_dist, -image_x, -image_y).
  Vector::assign(ray_dir, viewplane_dist, -image_x, -image_y * aspect_ratio);
  
  // Ray direction vector must be a unit.
  Vector::normalise(ray_dir, 
		    "PinholeOptics::generateRay() - normalise ray_dir");
}

// A constructor for PinholeOptics that allows the field-of-view
// to be specified.
PinholeOptics::PinholeOptics(double field_of_view, double an_aspect)
{
  double radians;

  if (field_of_view < TINY || field_of_view > 90.0 - TINY)
    fatal_error("PinholeOptics", "field of view not from 0 - 90",
		"You cannot specify that a camera should have a\n"
		"zero field of view or greater than 90 degrees\n"
		"with the PinholeOptics camera model. Typical\n"
		"field of view is about 20-30 degrees.");
  else {
    // We want the triangle from the origin to the center of the viewplane
    // to the point (+1, +0) on the viewplane to have an angle of
    // field_of_view, i.e. tan(field_of_view) = 1/viewplane_dist.

    radians = DEG2RAD(field_of_view);
    viewplane_dist = 1 / tan(radians);
    aspect_ratio = an_aspect;
  }
}

// The PixelSampler simply casts a ray at each pixel and stores
// the result in the Camera's Storage.
void PixelSampler::makePicture(Camera *theCamera,
			       primnode *trace_object)
{
  // Components of the camera
  Storage *theStorage;
  OpticsModel *theOptics;
  RayTracer *theEngine;
  matrix *camera_matrix;

  // Picture dimensions
  int picture_x, picture_y;

  // Where we are currently looking (pixel and imageplane coords)
  int x, y;
  double image_x, image_y, image_scale, x_center, y_center;
  vector ray_from, ray_dir;
  color ray_contribution;

  // The return values from the ray tracer
  color ray_col;
  double ray_t;
  vector ray_norm;
  vector ray_hit;
  primnode *ray_prim;

  // Enlargement variables
  double enlarge_x, enlarge_y, enlarge_factor;
  char enlarge_buf[MIRAGE_STRING_LENGTH];


  theStorage = theCamera->myStorage;
  picture_x = theStorage->xsize;
  picture_y = theStorage->ysize;
  theOptics = theCamera->myOptics;
  theEngine = theCamera->myEngine;
  camera_matrix = &(theCamera->world_transform[FORE]);

  Color::set(ray_contribution, full_color);

  // Determine the image-plane defining attributes.
  if (get_keyfile_data("enlarged-view", enlarge_buf)) {
    if (sscanf(enlarge_buf, "%lf, %lf: x%lf",
	       &enlarge_x, &enlarge_y, &enlarge_factor) != 3) {
      fatal_error("PixelSampler::makePicture", "wrong enlargement format",
		  "Format for the \'enlarged-view\' parameter is:\n"
		  "enlarged-view        203, 102: x5");
      exit(-1);
    } else {
      x_center = picture_x/2.0 - enlarge_factor * (enlarge_x - picture_x/2.0);
      y_center = picture_y/2.0 - enlarge_factor * (enlarge_y - picture_y/2.0);
      image_scale = 2.0 / (picture_x * enlarge_factor);
    }
  } else {
    x_center = picture_x / 2.0;
    y_center = picture_y / 2.0;
    image_scale = 1.0 / x_center;
  }

  theCamera->find_camerapos_sfx(trace_object);

  counter_primary_rays = counter_shadow_rays = counter_secondary_rays = 0;

  for (y = 0; y < picture_y; y++) {
    if (show_raytracing_progress && !completely_silent) {
      printf("Ray tracing: %04d\r", y);
      fflush(stdout);
    }

    for (x = 0; x < picture_x; x++) {
      if (pixel_sampler_print_every_pixel && !completely_silent)
	printf("Tracing pixel %d, %d\n", x, y);
      image_x = (x - x_center) * image_scale;
      image_y = (y - y_center) * image_scale;
      theOptics->generateRay(ray_from, ray_dir, image_x, image_y);
      
      // Apply the camera positioning matrix to the ray.
      Matrix::apply(*camera_matrix, ray_from, ray_from);
      Matrix::applynoshift(*camera_matrix, ray_dir, ray_dir);

      // trace the ray given by ray_from and ray_dir.
      // ray_flags is primary_ray, light_t is 0.0 (not used as this isn't
      // a shadow ray).
      // The places to put return values are ray_col (colour), ray_t (the
      // t value of the hit), ray_norm (the object's normal), ray_hit (the
      // hit point in world coordinates), and ray_prim (the primnode that
      // was hit).
      theEngine->traceRay(trace_object, ray_from, ray_dir,
			  theCamera->primary_ray, new_ray_signature(),
			  theCamera->camera_sfxtype, theCamera->camera_sfx,
			  0.0, ray_contribution,
			  &ray_prim, &ray_t, ray_norm, ray_hit, ray_col,
			  maximum_secondary_ray_depth);
      // Store the resulting colour.
      theStorage->store(x, y, ray_prim, ray_t, ray_norm, ray_hit, ray_col);

    }
  }
  
  if (display_ray_statistics && !completely_silent) {
    printf("%d primary, %d secondary (of which %d were shadow) rays traced\n",
	   counter_primary_rays, counter_secondary_rays, counter_shadow_rays);
  }
}

// The SingleRaySampler simply casts a single ray.
void SingleRaySampler::makePicture(Camera *theCamera,
				   primnode *trace_object)
{
  // Components of the camera
  Storage *theStorage;
  OpticsModel *theOptics;
  RayTracer *theEngine;
  matrix *camera_matrix;

  char enlarge_buf[MIRAGE_STRING_LENGTH];
  double enlarge_x, enlarge_y, enlarge_factor;
  double x_center, y_center, image_x, image_y, image_scale;
  int x, y, picture_x, picture_y;
  vector ray_from, ray_dir;
  color ray_contribution;

  // The return values from the ray tracer
  color ray_col;
  double ray_t;
  vector ray_norm;
  vector ray_hit;
  primnode *ray_prim;


  theStorage = theCamera->myStorage;
  picture_x = theStorage->xsize;
  picture_y = theStorage->ysize;
  theOptics = theCamera->myOptics;
  theEngine = theCamera->myEngine;
  camera_matrix = &(theCamera->world_transform[FORE]);

  theCamera->find_camerapos_sfx(trace_object);

  // Determine the image-plane defining attributes.
  if (get_keyfile_data("enlarged-view", enlarge_buf)) {
    if (sscanf(enlarge_buf, "%lf, %lf: x%lf",
	       &enlarge_x, &enlarge_y, &enlarge_factor) != 3) {
      fatal_error("SingleRaySampler::makePicture", "wrong enlargement format",
		  "Format for the \'enlarged-view\' parameter is:\n"
		  "enlarged-view        203, 102: x5");
      exit(-1);
    } else {
      x_center = picture_x/2.0 - enlarge_factor * (enlarge_x - picture_x/2.0);
      y_center = picture_y/2.0 - enlarge_factor * (enlarge_y - picture_y/2.0);
      image_scale = 2.0 / (picture_x * enlarge_factor);
    }
  } else {
    x_center = picture_x / 2.0;
    y_center = picture_y / 2.0;
    image_scale = 1.0 / x_center;
  }

  image_x = (pix_x - x_center) * image_scale;
  image_y = (pix_y - y_center) * image_scale;


  Color::set(ray_contribution, full_color);

  theCamera->find_camerapos_sfx(trace_object);

  theOptics->generateRay(ray_from, ray_dir, image_x, image_y);

  // Apply the camera positioning matrix to the ray.
  Matrix::apply(*camera_matrix, ray_from, ray_from);
  Matrix::applynoshift(*camera_matrix, ray_dir, ray_dir);

  // trace the ray given by ray_from and ray_dir.
  // ray_flags is primary_ray, light_t is 0.0 (not used as this isn't
  // a shadow ray).
  // The places to put return values are ray_col (colour), ray_t (the
  // t value of the hit), ray_norm (the object's normal), ray_hit (the
  // hit point in world coordinates), and ray_prim (the primnode that
  // was hit).

  counter_primary_rays = counter_shadow_rays = counter_secondary_rays = 0;

  theEngine->traceRay(trace_object, ray_from, ray_dir, theCamera->primary_ray,
		      new_ray_signature(), 
		      theCamera->camera_sfxtype, theCamera->camera_sfx,
		      0.0, ray_contribution,
		      &ray_prim, &ray_t, ray_norm, ray_hit, ray_col,
		      maximum_secondary_ray_depth);

  for (x = 0; x < picture_x; x++) {
    for (y = 0; y < picture_y; y++) {
      theStorage->store(x, y, ray_prim, ray_t, ray_norm, ray_hit, ray_col);
    }
  }
  
  if (display_ray_statistics && !completely_silent) {
    printf("%d primary, %d secondary (of which %d were shadow) rays traced\n",
	   counter_primary_rays, counter_secondary_rays, counter_shadow_rays);
  }
}

// Allocate storage space for the ColorStorage (an array of 3 * x * y unsigned
// char)
void ColorStorage::allocate(void)
{
  colorStorage = (unsigned char *) calloc(3 * xsize * ysize,
					  sizeof(unsigned char));
  if (colorStorage == NULL) {
    char tempbuf[MIRAGE_STRING_LENGTH];

    sprintf(tempbuf, "can't allocate ColorStorage (%d x %d)", xsize, ysize);
    fatal_error("ColorStorage::allocate()", tempbuf,
	       "Out of memory. Try specifying a smaller picture size\n"
	       "or a smaller scene or getting more memory. However,\n"
	       "this error really shouldn't happen in any case.\n"
	       "Something very odd is going on.");
  }
}

// Store a value in the colorStorage as unsigned char.
void ColorStorage::store(int x, int y, primnode *store_prim, double store_t,
			 vector store_norm, vector store_hit, color store_col)
{
  int offset;

  if (colorStorage == NULL) {
    fatal_error("ColorStorage::store", "no storage allocated",
		"Must allocate storage space before attempting to\n"
		"store data!");
  }

  // Determine where in colorStorage to store data.
  offset = 3 * (y * xsize + x);

  // Store the emitted and diffuse colours together.
  store_col[R] += store_col[ER];
  store_col[G] += store_col[EG];
  store_col[B] += store_col[EB];
  
  // Store the color as unsigned char.
  if (store_col[R] > 1.0)
    colorStorage[offset++] = 255;
  else if (store_col[R] < 0.0)
    colorStorage[offset++] = 0;
  else
    colorStorage[offset++] = (unsigned char) (255 * store_col[R]);

  if (store_col[G] > 1.0)
    colorStorage[offset++] = 255;
  else if (store_col[G] < 0.0)
    colorStorage[offset++] = 0;
  else
    colorStorage[offset++] = (unsigned char) (255 * store_col[G]);

  if (store_col[B] > 1.0)
    colorStorage[offset++] = 255;
  else if (store_col[B] < 0.0)
    colorStorage[offset++] = 0;
  else
    colorStorage[offset++] = (unsigned char) (255 * store_col[B]);
}

// Retrieve a color value from colorStorage.
void ColorStorage::retrieve(int x, int y, color store_col)
{
  int offset;

  if (colorStorage == NULL) {
    fatal_error("ColorStorage::retrieve", "no storage allocated",
		"Must allocate storage space before attempting to\n"
		"retrieve data!");
  }

  // Determine where in colorStorage to get data from.
  offset = 3 * (y * xsize + x);
  
  // Retrieve the color.
  store_col[R] = colorStorage[offset++] / 255.0;
  store_col[G] = colorStorage[offset++] / 255.0;
  store_col[B] = colorStorage[offset] / 255.0;
}

// Allocate storage space for the TrackStorage (an array of x * y color_values)
void TrackStorage::allocate(void)
{
  trackStorage = (unsigned char *) calloc(xsize * ysize * num_colorvalues,
					  sizeof(unsigned char));
  if (trackStorage == NULL) {
    char tempbuf[MIRAGE_STRING_LENGTH];

    sprintf(tempbuf, "can't allocate TrackStorage (%d x %d)", xsize, ysize);
    fatal_error("TrackStorage::allocate()", tempbuf,
	       "Out of memory. Try specifying a smaller picture size\n"
	       "or a smaller scene or getting more memory. However,\n"
	       "this error really shouldn't happen in any case.\n"
	       "You must be critically low on memory.");
  }
}

// Store a color in the trackStorage as a color_value.
void TrackStorage::store(int x, int y, primnode *store_prim, double store_t,
			 vector store_norm, vector store_hit, color store_col)
{
  if (trackStorage == NULL) {
    fatal_error("TrackStorage::store", "no storage allocated",
		"Must allocate storage space before attempting to\n"
		"store data!");
  }

  Colorvalue::compact(store_col, &trackStorage[(y * xsize + x) *
					      num_colorvalues]);
}

// Retrieve a color value from trackStorage.
void TrackStorage::retrieve(int x, int y, color store_col)
{
  if (trackStorage == NULL) {
    fatal_error("ColorStorage::retrieve", "no storage allocated",
		"Must allocate storage space before attempting to\n"
		"retrieve data!");
  }

  Colorvalue::extract(&trackStorage[(y * xsize + x) * num_colorvalues],
		      store_col);
}

// Create a default camera (so that users don't have to worry about
// most of the parameters to the camera). This function takes a picture
// size also.
object create_default_camera(int x_size, int y_size, int antialiasing,
			     double field_of_view)
{
  Camera *aCamera;

  aCamera = new Camera(new PinholeOptics(field_of_view, 1.0),
		       new StandardRayTracer, new Antialiaser(antialiasing),
		       new ColorStorage(x_size, y_size), NULL);
  return Dagnode::newCamera(aCamera);
}

// Create a default camera (so that users don't have to worry about
// most of the parameters to the camera). This function takes no picture size.
object create_default_camera(int antialiasing, double field_of_view)
{
  Camera *aCamera;

  aCamera = new Camera(new PinholeOptics(field_of_view, 1.0),
		       new StandardRayTracer, new Antialiaser(antialiasing),
		       new ColorStorage, NULL);
  return Dagnode::newCamera(aCamera);
}

void Camera::find_camerapos_sfx(primnode *world_scene)
{
  vector camera_pos;

  Matrix::apply(world_transform[FORE], zero_vector, camera_pos);

  find_point_sfx(world_scene, camera_pos, primary_ray,
		 &camera_sfxtype, &camera_sfx);
}

#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif
