/* camera.h : Header file for the camera objects.
** 
** This file defines the headers for camera class and its component subclasses.
**
** Chris Butcher, 29 November 1996
*/

// An OpticsModel takes an imageplane location (in terms of X and Y
// coordinates where X from +1 to -1 is the width of the picture and
// Y has the same scale from +k to -k where k = height/width.
class OpticsModel : public DeletableObject {
public:
  // generateRay() forms a ray from and direction given the image
  // plane X and Y.
  virtual void generateRay(vector ray_from, vector ray_dir,
			   double image_x, double image_y) = 0;

  virtual ~OpticsModel(void) {
  }
};

// The PinholeOptics optics model models an ideal 'graphics camera'
// which has the viewplane in front of the eyepoint. It is created
// with the parameters field-of-view (which defines the angle in
// degrees from the middle of the image plane to +1 in X on the image
// plane) that determines the amount of 'fish-eye' to apply,
// and aspect_ratio, which changes the height of the pixels relative
// to the width of the pixels.
// 
// The camera that is modeled is pointed along X, up-vector +Z.
class PinholeOptics : public OpticsModel {
public:
  double viewplane_dist, aspect_ratio;

  virtual void generateRay(vector ray_from, vector ray_dir,
			   double image_x, double image_y);

  // Constructor function allows specifying of field of view and
  // aspect ratio (height / width of the pixels).
  PinholeOptics(double field_of_view, double an_aspect);

  virtual ~PinholeOptics(void) {
  }
};

// The RayTracer traces the given ray from and direction and returns
// some data giving the color, and other parameters of the place hit.
class RayTracer : public DeletableObject {
public:
  // traceRay() traces the ray from ray_from in the unit direction
  // ray_dir with flags ray_flags (what type of ray, etc.) and unique
  // signature ray_signature (ensures primitives are never intersected
  // twice with the same ray).
  // The ray is traced against the object world_scene.
  // ray_contribution is the contribution that this ray's colour
  // will eventually make to the final image (if the value drops
  // below a cutoff then the ray isn't bothered with).
  // Return values are placed in primnode *hit_object, double hit_t,
  // vector hit_normal, vector hit_point, color hit_color.
  virtual void traceRay(primnode *world_scene,
			vector ray_from, vector ray_dir,
			uint ray_flags, signature ray_signature,
			SfxType start_sfxtype, void *start_sfx,
			double light_t, color ray_contribution,
			primnode **hit_object,
			double *hit_t, vector hit_normal,
			vector hit_point, color hit_color,
			int max_ray_depth) = 0;

  virtual ~RayTracer(void) {
  }
};

// The PictureSampler generates a picture of the correct size when
// asked to, using the RayTracer on the worldscene. It does anti-
// aliasing if necessary (or desired).
class PictureSampler : public DeletableObject {
public:
  // makePicture() generates the picture (after telling the Storage
  // to allocate the necessary storage).
  virtual void makePicture(class Camera *generating,
			   primnode *trace_object) = 0;

  virtual ~PictureSampler(void) {
  }
};

// The PixelSampler simply casts a ray at the middle of each pixel and
// stores the result in the Storage.
class PixelSampler : public PictureSampler {
public:
  virtual void makePicture(class Camera *generating,
			   primnode *trace_object);

  virtual ~PixelSampler(void) {
  }
};

// The SingleRaySampler casts only one ray and stores this at
// every place in the Storage.
class SingleRaySampler : public PictureSampler {
public:
  // The coordinates of the ray to cast.
  int pix_x, pix_y;

  virtual void makePicture(class Camera *generating,
			   primnode *trace_object);

  SingleRaySampler(int p_x, int p_y) {
    pix_x = p_x;
    pix_y = p_y;
  }

  virtual ~SingleRaySampler(void) {
  }
};

// The Storage holds the frame buffer (may have color information only
// or more specific G-buffer e.g. object hit, normal, etc.)
class Storage : public DeletableObject {
public:
  int xsize, ysize;

  // The magic number stores which kind of Storage object this is ...
  // e.g. COLORSTORAGE_MAGIC etc.
  int magic;

  // Store the data returned from the ray. Most of this will be ignored
  // e.g. if the buffer is only set up to store color.
  virtual void store(int x, int y, primnode *store_prim, double store_t,
		     vector store_norm, vector store_hit, color store_col) = 0;

  // Retrieve the color returned from the ray (for file writing or for
  // postprocessing).
  virtual void retrieve(int x, int y, color getcol) = 0;

  // Allocate the storage space that will be required.
  virtual void allocate(void) = 0;

  virtual ~Storage(void) {
  }
};

// The PostProcessor operates on data in the framebuffer of the camera
// after the image has been generated. It may do anything ... including
// calling another camera to do a cross-dissolve if you want.
class PostProcessor : public DeletableObject {
public:

  // Postprocesses the framebuffer by the given effect.
  virtual void postprocess(class Camera *generated) = 0;

  virtual ~PostProcessor(void) {
  }
};

// The ColorStorage subclass merely stores colors with 8-bit precision
// in R, G, B.
class ColorStorage : public Storage {
public:
  // Storage for color data
  unsigned char *colorStorage;

  // Store a color at a particular pixel.
  void store(int x, int y, primnode *store_prim, double store_t,
	     vector store_norm, vector store_hit, color store_col);

  // Retrieve color from a particular pixel (used for postprocessing).
  void retrieve(int x, int y, color storecol);

  // Allocate storage prior to storing colors
  void allocate(void);

  // Destructor for ColorStorage
  virtual ~ColorStorage(void)
  {
    free(colorStorage);
  }

  // Set the size of the picture. Note that the space for the picture
  // isn't actually allocated unless the camera is called, which means
  // that we avoid allocating unused space for all the cameras in the
  // scene that aren't used.
  ColorStorage(int img_x, int img_y)
  {
    xsize = img_x;
    ysize = img_y;

    // Storage space not allocated yet
    colorStorage = NULL;

    // Signify that this is a ColorStorage object using the magic
    // number field.
    magic = COLORSTORAGE_MAGIC;
  }

  // Set the size of the picture from parameters specified in the keyfile.
  ColorStorage(void)
  {
    char sizebuf[MIRAGE_STRING_LENGTH];

    if (!get_keyfile_data("picture-size", sizebuf)) {
      fatal_error("ColorStorage::constructor", "no picture size specified!",
		  "You must either initialise a camera giving it a specific\n"
		  "size, or else set the default size in the keyfile using\n"
		  "the picture-size key, e.g.\n"
		  "picture-size         400, 300");
      exit(-1);
    }

    if (sscanf(sizebuf, "%d, %d", &xsize, &ysize) != 2) {
      fatal_error("ColorStorage::constructor", "wrong picture-size format",
		  "The syntax for the picture-size key is e.g.\n"
		  "picture-size            400, 300");
      exit(-1);
    }

     // Storage space not allocated yet
    colorStorage = NULL;

    // Signify that this is a ColorStorage object using the magic
    // number field.
    magic = COLORSTORAGE_MAGIC;
  }
};

// The TrackStorage subclass stores color_values that describe the
// color at a pixel in terms of its indirect RGB, emitted RGB and its
// alpha values.
class TrackStorage : public Storage {
public:
  // Storage for color data
  unsigned char *trackStorage;

  // Store a color at a particular pixel.
  void store(int x, int y, primnode *store_prim, double store_t,
	     vector store_norm, vector store_hit, color store_col);

  // Retrieve color from a particular pixel (used for postprocessing).
  void retrieve(int x, int y, color storecol);

  // Allocate storage prior to storing colors
  void allocate(void);

  // Destructor for TrackStorage
  virtual ~TrackStorage(void)
  {
    free(trackStorage);
  }

  TrackStorage(int img_x, int img_y)
  {
    xsize = img_x;
    ysize = img_y;

    // Storage space not allocated yet
    trackStorage = NULL;

    // Signify that this is a TrackStorage object using the magic
    // number field.
    magic = TRACKSTORAGE_MAGIC;
  }
};

//   A Camera generates a view of the world. Cameras are created and
// then placed in the DAG, specifying their position and orientation
// in world-space. When the DAG is converted into a renderable
// representation by the cluster, any cameras contained therein
// are placed into the global cameralist with a transformation matrix
// describing their orientation in world-space.
class Camera : public DeletableObject {
public:
  // Each camera has a transformation matrix to transform rays from
  // camera space into worldspace ... this effectively defines
  // the camera's position.
  transform world_transform;

  // OpticsModel generates ray start points and directions given the
  // point on the image plane. Can be used for modelling simple pinhole
  // cameras or strange Picasso-esque distortions.
  class OpticsModel *myOptics;

  // The RayTracer traces the given ray at the object being viewed.
  // Essentially consists of a call to ray_intersect on the object
  // (which must be a cluster that traces a ray through some internal
  // representation or else a single primitive).
  class RayTracer *myEngine;

  // The PictureSampler, when asked to generate a picture of the
  // given size, calls the RayTracer repeatedly to sample the image
  // and produces a picture from the ray samples. This allows
  // simple methods or more complex antialiasing / edge-following ones.
  class PictureSampler *mySampler;

  // A camera's Storage holds the frame buffer. Alternatively, it
  // may store more detailed information such as surface normal, distance
  // to first hit, and object hit, if desired.
  class Storage *myStorage;

  // The PostProcessor performs post-processing on the information in
  // the frame buffer if it is necessary.
  class PostProcessor *myProcess;

  // The value of ray_flags for all primary rays cast from the camera.
  // This is set when the camera is located in worldspace, when we know
  // if the camera is on the far side of the gateway or not.
  uint primary_ray;

  // The SFX parameters of the worldscene, at the camera's position.
  // Set by find_camerapos_sfx().
  SfxType camera_sfxtype;
  void *camera_sfx;

  // Ray trace the image.
  void generateImage(primnode *trace_object);

  // Find the SFX parameter at the camera's location.
  void find_camerapos_sfx(primnode *world_scene);

  // Constructor that initialises the components of the camera
  Camera(class OpticsModel *anOptics, class RayTracer *anEngine,
	 class PictureSampler *aSampler, class Storage *aStorage,
	 class PostProcessor *aProcess)
  {
    myOptics = anOptics;
    myEngine = anEngine;
    mySampler = aSampler;
    myStorage = aStorage;
    myProcess = aProcess;

    // The camera initially has a unit transform; placed in the default
    // orientation specified by the OpticsModel. This will be changed by
    // the preparation process when the scene is traversed looking for
    // and placing cameras.
    Transform::unit(world_transform);

    // No SFX unless they are found by find_camerapos_sfx.
    camera_sfxtype = NOSFX;
    camera_sfx = NULL;

    mark_for_deletion();
  }

  // Destructor that frees the parts of the camera.
  ~Camera(void)
  {
    delete myOptics;
    delete myEngine;
    delete mySampler;
    delete myStorage;
    if (myProcess != NULL)
      delete myProcess;
  }
};

