/* color.h : color manipulation utilities
** 
** The RGB class described here should never be instantiated. It is a
** toolkit of color functions that is used by the rendering system.
** They are accessed as Color::function(a, b, c) e.g.
** Color::set(col, r, g, b);
**
** Chris Butcher, 28 November 1996
*/

// Define the color storage. Note that a 'color' is a different thing to the
// 'Color' toolkit class ...
// Colors are stored as an array of doubles. They are indexed as
// R, G, B : red, green, blue indirect light.
// ER, EG, EB : red, green, blue emitted light.
// A : alpha channel.
const int num_color_components = 7;
typedef double color[num_color_components];

extern color zero_color, full_color;

// This class should never be instantiated. It is a toolkit only.
class Color {
public:
  
  // Is a color zero?
  static int is_zero(color col)
  {
    for (int i = 0; i < num_color_components; i++)
      if (col[i] != 0)
	return FALSE;

    return TRUE;
  }

  // Sets color result to color col1.
  static void set(color result, color col1)
  {
    for (int i = 0; i < num_color_components; i++)
      result[i]  = col1[i];
  }
  
  // Assign color values to color result.
  static void assign(color result, double r, double g, double b,
		     double er = 0, double eg = 0, double eb = 0,
		     double a = 0)
  {
    result[R]  = r;
    result[G]  = g;
    result[B]  = b;
    result[ER] = er;
    result[EG] = eg;
    result[EB] = eb;
    result[A]  = a;
  }

  // Adds two colors col1 and col2 to produce result.
  static void add(color col1, color col2, color result)
  {
    for (int i = 0; i < num_color_components; i++)
      result[i] = col1[i]  + col2[i];
  }

  // Adds a fraction of color col1 to color result.
  static void addfrac(double frac, color col1, color result)
  {
    for (int i = 0; i < num_color_components; i++)
      result[i] += frac * col1[i];
  }

  // Multiply col by a scalar value to produce result.
  static void mult(double val, color col, color result)
  {
    for (int i = 0; i < num_color_components; i++)
      result[i] = val * col[i];
  }

  // Multiply two colors together.
  static void mult(color col1, color col2, color result)
  {
    for (int i = 0; i < num_color_components; i++)
      result[i] = col1[i] * col2[i];
  }

  // Mix two colours together.
  static void mix(color col1, color col2, double fraction, color result)
  {
    double reverse_frac;

    if (fraction <= 0) {
      Color::set(result, col1);
    } else if (fraction >= 1) {
      Color::set(result, col2);
    } else {
      reverse_frac = 1.0 - fraction;

      for (int i = 0; i < num_color_components; i++)
	result[i] = col1[i] * reverse_frac + col2[i] * fraction;
    }
  }

  // Clamp a color from 0 to 1 in R, G and B.
  static void clamp(color col)
  {
    for (int i = 0; i < num_color_components; i++)
      CLAMP(0, col[i],  1);
  }
};

