/* colorvalue.h : color value manipulation utilities
** 
** A color_value is a packed representation of a color. It has 8 bits
** for each of the seven color components, and 8 flag bits.
**
** The colorvalue class described here should never be instantiated. It is a
** toolkit of colorvalue functions that is used by the rendering system.
** They are accessed as Colorvalue::function(a, b, c) e.g.
** val = Colorvalue::compact(col);
**
** Chris Butcher, 23 April 1998
*/

// color_value is a compacted data type that represents a color.
const int num_colorvalues = num_color_components + 1;
typedef unsigned char color_value[num_colorvalues];
const int FLAGINDEX = num_colorvalues - 1;
const unsigned char colorvalue_same_tolerance = 0xFF / 20;

// This class should never be instantiated. It is a toolkit only.
class Colorvalue {
public:
  
  // Zero a color value.
  static void zero(color_value val)
  {
    for (int i = 0; i < num_color_components; i++)
      val[i] = 0x00;
    val[FLAGINDEX] = 0x00;
  }

  // Set one to another.
  static void set(color_value dest, color_value src)
  {
    for (int i = 0; i < num_color_components; i++)
      dest[i] = src[i];
    dest[FLAGINDEX] = src[FLAGINDEX];
  }

  // Compacts a color_value.
  static void compact(color source, color_value result)
  {
    for (int i = 0; i < num_color_components; i++) {
      if (source[i] < 0)
	result[i] = 0x00;
      else if (source[i] > 1)
	result[i] = 0xFF;
      else
	result[i] = (unsigned char) (source[i] * 0xFF);
    }

    // No flags so far.
    result[FLAGINDEX] = 0x00;
  }
  
  // Extracts a color_value.
  static void extract(color_value source, color result)
  {
    for (int i = 0; i < num_color_components; i++)
      result[i] = source[i] / 255.0;
  }

  // Checks to see if two colour values are within the antialiaser's
  // tolerance.
  static int same_color(color_value val1, color_value val2)
  {
    int difference;

    for (int i = 0; i < num_color_components; i++) {
      difference = val1[i] - val2[i];
      if ((difference >  colorvalue_same_tolerance) ||
	  (difference < -colorvalue_same_tolerance))
	return FALSE;
    }
    
    return TRUE;
  }
};
