/* Constants.h : Constants and constant #defines
** 
** This file contains all of the constants, enumerated types and constant
** #defines that are needed in Mirage. Macros are found in macros.h.
**
** Chris Butcher, 27 November 1996
*/

#ifndef FALSE
#define FALSE            0
#define TRUE             ~FALSE
#endif

#ifndef M_PI
#define M_PI          3.14159265358979323846
#endif
#define PI            M_PI

#ifndef MAXINT
#define MAXINT        INT_MAX
#endif

// Very small value used for checking for zero values.
#define TINY             (double)1.0e-250

// Very large value used for flagging infinity.
#ifndef INFINITY
#define INFINITY   1e+35
#endif

// Default length of fixed-length strings.
#define MIRAGE_STRING_LENGTH   256

// Default 'flags' value given to new objects.
#define FLAGS_DEFAULT          0

// -- start defining bitflags for dag objects (and primnodes also).
//      flags to say what kind of dagnode this is.
#define FLAGS_DAG                    (1 << 0)
#define FLAGS_PRIM                   (1 << 1)
#define FLAGS_CAMERA                 (1 << 2)
#define FLAGS_LIGHT                  (1 << 3)
//      mask to remove the previous four flags ... all 1's except for the
//        bottom four bits.
#define FLAGS_NOTYPE                 ((~((uint) 0)) << 4)
//      this flag must be set for an object to have a partially transparent
//        shadow.
#define FLAGS_TRANSLUCENT            (1 << 4)
//      flags to say which types of rays this object doesn't appear in.
//        NOPRIMARY  = can't be seen unless it's a reflected or transmitted or
//                    shadow ray.
//        NOSHADOW   = doesn't cast a shadow. (lights: doesn't cast shadows)
//        NOREFLECT  = doesn't appear in mirrors.
//        NOTRANSMIT = doesn't appear in transmitted rays.
#define FLAGS_NOPRIMARY              (1 << 5)
#define FLAGS_NOSHADOW               (1 << 6)
#define FLAGS_NOREFLECT              (1 << 7)
#define FLAGS_NOTRANSMIT             (1 << 8)
#define FLAGS_ONLYTRANSMIT           (1 << 9)
#define FLAGS_GATEWAY                (1 << 10)
#define FLAGS_ONEWAY_GATEWAY         (1 << 11)
#define FLAGS_NOT_THROUGH_GATEWAY    (1 << 12)
#define FLAGS_ONLY_THROUGH_GATEWAY   (1 << 13)
//      the LABELED flag indicates that when we are traversing the dag-graph
//        propagating matrices to find the positions of all labeled objects,
//        this object is one of those that needs to have its position
//        placed in the global label-list. This flag has no meaning for
//        primnodes.
#define FLAGS_LABELED                (1 << 14)
//      MULTITEXTURE designates that there is more than one texture applied
//        to a given primnode, and that the texcol pointer should be used
//        rather than the texture pointer. Set automatically when the primnode
//        is created.
#define FLAGS_MULTITEXTURE           (1 << 15)
// -- end bitflag definitions.

// -- start bitflag definitions for ray_flags indicating the type of ray
#define RAY_SHADOW             (1 << 0)
#define RAY_REFLECTED          (1 << 1)
#define RAY_TRANSMITTED        (1 << 2)
#define RAY_THROUGH_GATEWAY    (1 << 3)
//      primary mask is used to mask out all ray type flags. Ray is a primary
//        ray if (ray_flags & RAY_PRIMARYMASK) == 0.
#define RAY_PRIMARYMASK        (RAY_SHADOW | RAY_REFLECTED | RAY_TRANSMITTED)
//      macro to tell if a ray is a primary ray.
#define IS_PRIMARY(f)          ((f & RAY_PRIMARYMASK) == 0)
//      value of ray_flags used for primary rays (cast by camera).
#define PRIMARY_RAY            0
// -- end bitflag definitions for ray_flags


// Magic numbers to determine the type of a Storage class.
#define COLORSTORAGE_MAGIC     498
#define TRACKSTORAGE_MAGIC     1203

// Size of the global label list.
#define LABEL_LIST_SIZE 2000

// Maximum number of textures that may be overlaid on a single object.
#define MAX_OVERLAID_TEXTURES  10

// When tracing a ray, indicate that the background texture has been hit by
// returning a very large value for t.
#define HIT_BACKGROUND         1e30

// Maximum number of lights in a lightList.
#define MAX_LIGHTS             10

// Rays cast from a point on a surface must not intersect that surface.
// This could cause erroneous self-shadowing ('cancer'). So we discard any
// intersections that have t-values less than a cutoff.
#define MIN_TVAL_CUTOFF                 1.0e-06

// Don't trace object rays if their final contribution to the image is
// smaller than this cutoff value.
#define RAY_CONTRIBUTION_CUTOFF 0.01

// Constants used for the spline code ... maximum number of approximation
// iterations and the tolerance to iterate to. Also the timestep to use
// as an instant for approximating velocity, and the minimum distance moved
// to be considered as moving, the timestep to use when looking for movement,
// and the maximum time to look ahead / back when finding movement.
#define SPLINE_MAX_ITERATIONS     100000
#define SPLINE_TIME_TOLERANCE     0.001
#define SPLINE_VEL_DELTA_T        0.005
#define SPLINE_MIN_DELTA_X        0.00005
#define SPLINE_DIR_DELTA_T        0.05
#define SPLINE_MAX_LOOKAHEAD      100.0
#define SPLINE_CLOSEST_SAMPLES    10.0
#define SPLINE_CLOSEST_PRECISION  0.00001
#define SPLINE_MOVE_DELTA_T       0.05

// Sizes of the blocks of working memory that are allocated. The larger the
// block, in general the more efficient working memory will be (and the
// larger the overhead of unused space will be).
#define PERMANENT_DAG_BLOCKSIZE   10000
#define DAG_BLOCKSIZE             10000
#define PRIMNODE_BLOCKSIZE        10000
#define RCSG_BLOCKSIZE            10000
#define TEX_BLOCKSIZE             1000
#define TEXCOL_BLOCKSIZE          100
#define DEL_BLOCKSIZE             10000
#define PRIMDEL_BLOCKSIZE         10000
#define OCTREE_BLOCKSIZE          10000
#define LIST_BLOCKSIZE            10
#define INT_BLOCKSIZE             100

// The maximum number of blocks of each type that may be allocated. This
// is used as the size for an array that holds the addresses of each block of
// the given kind.
#define MAX_PERMANENTDAGBLOCKS    10000
#define MAX_DAGBLOCKS             10000
#define MAX_PRIMNODEBLOCKS        10000
#define MAX_RCSGBLOCKS            10000
#define MAX_TEXBLOCKS             10000
#define MAX_TEXCOLBLOCKS          10000
#define MAX_DELBLOCKS             10000
#define MAX_PRIMDELBLOCKS         10000
#define MAX_OCTREEBLOCKS          10000
#define MAX_LISTBLOCKS            10000
#define MAX_INTBLOCKS             10000

// A sentinel number used to indicate that there is no next element in the
// free-memory chains managed by the working memory manager.
#define NO_FREE_MEMORY            ~((unsigned int) 0)

// A sentinel pointer used to indicate that there is currently no free
// IntersectionList for allocation.
#define NO_FREE_LIST_MEMORY       ((IntersectionList *) NULL)

// A sentinel pointer used to indicate that there is currently no free
// intersection for allocation.
#define NO_FREE_INT_MEMORY       ((intersection *) NULL)

// The number used to seed the random number generator for jittering the
// values read in from a heightfield pixmap.
#define HEIGHTFIELD_RANDOM_SEED   8488

// The minimum t-value to travel in the heightfield before checking for
// intersections.
#define HF_MIN_CELL_TVAL          0.001

// The maximum value that the heightfield takes (stored in an unsigned short).
#define MAX_SHORT                 65535
#define HF_MAXVAL                 (MAX_SHORT - 1)

// Maximum number of 'flares' generated by a single light when applying
// a lensflare effect.
#define LENSFLARE_MAX_FLARES       15

// Maximum number of material components in a material.
#define MAX_MATERIAL_COMPONENTS    10

#ifdef USE_OWN_DRAND48
// Coefficients for drand48 etc. code.
#define	MIRAGE_RAND48_SEED_0	(0x330e)
#define	MIRAGE_RAND48_SEED_1	(0xabcd)
#define	MIRAGE_RAND48_SEED_2	(0x1234)
#define	MIRAGE_RAND48_MULT_0	(0xe66d)
#define	MIRAGE_RAND48_MULT_1	(0xdeec)
#define	MIRAGE_RAND48_MULT_2	(0x0005)
#define	MIRAGE_RAND48_ADD	(0x000b)
#endif

// These prototypes are used in the definitions for many classes and so have
// to come very early on.
int get_keyfile_data(char *key_item, char *databuf);
void fatal_error(char *location, char *errtype, char *description);

const char axis_str[][2] = {"X", "Y", "Z", "W"};

