/* dagnode.c : Functions for the directed acyclic graph classes.
** 
** This file defines operations on objects (rotate, shift, stretch, new etc.)
**
** Chris Butcher, 28 November 1996
*/

#include "mirage.h"

#ifdef USE_NAMESPACES
namespace Mirage {
#endif

// When bind is called on an object it creates a new dagnode that
// has a vertical pointer to the object: done so that a set of dagnodes
// which are linked together can be manipulated as one object.
// Syntax: myObject = myObject->bind();
object Dagnode::bind(void)
{
  object newobj;

  // Allocate a dagnode from working memory, and initialise it.
  newobj = new_dag();
  initialise(newobj, FLAGS_DAG, FALSE);

  // Details of the new object is that it contains a vertical pointer
  // to the object that bind() was called on.
  newobj->details  = (void *) this;

  return newobj;
}
  
// newPrimitive creates an object from the given Primitive subclass
// (set of intersection routines).
object Dagnode::newPrimitive(Primitive *aPrim)
{
  object newobj;

  // Allocate a dagnode from working memory, and initialise it.
  newobj = new_dag();
  initialise(newobj, FLAGS_PRIM, aPrim->is_cluster);

  // Details of the new object is that it contains a vertical pointer
  // to the primitive that we called it with.
  newobj->details  = (void *) aPrim;

  return newobj;
}

// newCamera creates an object that represents the given Camera.
// This object can then be positioned in the scene to change the camera's
// view point and view direction.
// Note that scaling the camera is a bad idea, generally.
object Dagnode::newCamera(Camera *aCamera)
{
  object newobj;

  // Allocate a dagnode from working memory, and initialise it.
  newobj = new_dag();
  initialise(newobj, FLAGS_CAMERA, FALSE);

  // Details of the new object is that it contains a vertical pointer
  // to the camera that it represents.
  newobj->details  = (void *) aCamera;

  return newobj;
}

// newLight creates an object that represents the given Light.
// This object can then be positioned in the scene to change the light's
// position and direction.
object Dagnode::newLight(Light *aLight)
{
  object newobj;

  // Allocate a dagnode from working memory, and initialise it.
  newobj = new_dag();
  initialise(newobj, FLAGS_LIGHT, FALSE);

  // Details of the new object is that it contains a vertical pointer
  // to the light that it represents.
  newobj->details  = (void *) aLight;

  return newobj;
}

// rotate(axis, axis, angle) rotates the given object from one axis
// to the next axis by the angle specified.
void Dagnode::rotate(axis from, axis to, double angle)
{
  transform rotation;
  double radians, sine, cosine;

  Transform::unit(rotation);

  radians = DEG2RAD(angle);
  sine = sin(radians);
  cosine = cos(radians);

  // Construct the right rotation matrices.
  rotation[FORE][from][from] = cosine;
  rotation[FORE][from][to  ] = -sine;
  rotation[FORE][to  ][from] = +sine;
  rotation[FORE][to  ][to  ] = cosine;
  rotation[AFT][from][from] = cosine;
  rotation[AFT][from][to  ] = +sine;
  rotation[AFT][to  ][from] = -sine;
  rotation[AFT][to  ][to  ] = cosine;

  // Apply the constructed rotation matrices
  Transform::combine(this->position, rotation, this->position);
}

// rotate(axis, angle) rotates the given object around the axis
// by the angle specified.
void Dagnode::rotate(axis around, double angle)
{
  transform rotation;
  int from, to;
  double radians, sine, cosine;

  from = (around + 1) % 3;
  to = (around + 2) % 3;

  Transform::unit(rotation);

  radians = DEG2RAD(angle);
  sine = sin(radians);
  cosine = cos(radians);

  // Construct the right rotation matrices
  rotation[FORE][from][from] = cosine;
  rotation[FORE][from][to  ] = -sine;
  rotation[FORE][to  ][from] = +sine;
  rotation[FORE][to  ][to  ] = cosine;
  rotation[AFT][from][from] = cosine;
  rotation[AFT][from][to  ] = +sine;
  rotation[AFT][to  ][from] = -sine;
  rotation[AFT][to  ][to  ] = cosine;

  // Apply the constructed rotation matrices
  Transform::combine(this->position, rotation, this->position);
}

// rotate(forward, up) rotates the given object so that its X axis
// points along the forward vector and its Z axis points in the direction of
// the up vector.
void Dagnode::rotate(vector given_fwd, vector given_up)
{
  transform rotation;
  vector new_fwd, new_up, new_left;
  double length;
  axis a;

  length = Vector::magnitude(given_up);
  if (fabs(length) < TINY)
    // We have a zero-length up-vector, use +Z instead.
    Vector::set(new_up, plus_Z);
  else
    Vector::set(new_up, given_up);

  length = Vector::magnitude(given_fwd);
  if (fabs(length) < TINY) {
    output_warning("rotate(vector fwd, vector up)",
		   "Forward-vector is zero!",
		   "You must specify a direction to point the object in.");
    return;
  }
  Vector::mult(1.0 / length, given_fwd, new_fwd);

  Vector::cross(given_up, new_fwd, new_left);
  length = Vector::magnitude(new_left);
  if (fabs(length) < TINY) {
    // The forward and up vectors are in the same direction. This means that
    // there is no 'right' way to position the object. However when the
    // object's up-vector is along +Z we wish the behaviour to be:
    //
    // fwd_vector == +Z: up points towards -X
    // fwd_vector == -Z: up points towards +X
    //
    // so that the results are the same as if we had just rotated from X to Z
    // by +90 or -90.
    //
    // We do this by saying that we just swap the Y and Z components of the
    // up-vector and setting its X component to zero to get the left-vector.
    // This then gives us a vector that is guaranteed to be perpendicular to
    // the up and fwd vectors as long as the up-vector isn't along +X. If
    // it is then we use left-vector = +Y.
    if ((fabs(given_up[Y]) < TINY) &&
	(fabs(given_up[Z]) < TINY)) {
      // The up-vector is along +X; use left-vector along +Y.
      Vector::assign(new_left, 0, 1, 0);
    } else {
      // We form the left-vector from the up-vector's components in such a
      // way that it is guaranteed to be perpendicular to the up-vector
      // and it wil be +Y if the up-vector is +Z.
      Vector::assign(new_left, 0, new_up[Z], -new_up[Y]);
      Vector::normalise(new_left,
			"rotate(vec, vec): error in internal arithmetic!!!!");
    }
  }
  Vector::normalise(new_left, "Dagnode::rotate(vec, vec): zero left-vector");

  Vector::cross(new_fwd, new_left, new_up);
  Vector::normalise(new_up, "Dagnode::rotate(vec, vec): zero up-vector");


  Transform::unit(rotation);

  for (a = X; a <= Z; a = (axis) (a + 1)) {
    rotation[FORE][a][X] = new_fwd[a];
    rotation[FORE][a][Y] = new_left[a];
    rotation[FORE][a][Z] = new_up[a];

    rotation[AFT][X][a] = new_fwd[a];
    rotation[AFT][Y][a] = new_left[a];
    rotation[AFT][Z][a] = new_up[a];
  }

  // Apply the constructed rotation matrices
  Transform::combine(this->position, rotation, this->position);
}

// shift(xshift, yshift, zshift) translates the given object.
void Dagnode::shift(double xshift, double yshift, double zshift)
{
  transform shift;

  Transform::unit(shift);

  // Construct the right shift matrices
  shift[FORE][X][3] = xshift;
  shift[FORE][Y][3] = yshift;
  shift[FORE][Z][3] = zshift;
  shift[AFT][X][3] = -xshift;
  shift[AFT][Y][3] = -yshift;
  shift[AFT][Z][3] = -zshift;

  // Apply the constructed rotation matrices
  Transform::combine(this->position, shift, this->position);
}

// shift(shift_vec) translates the given object.
void Dagnode::shift(vector shift_vec)
{
  transform shift;

  Transform::unit(shift);

  // Construct the right shift matrices
  shift[FORE][X][3] = shift_vec[X];
  shift[FORE][Y][3] = shift_vec[Y];
  shift[FORE][Z][3] = shift_vec[Z];
  shift[AFT][X][3] = -shift_vec[X];
  shift[AFT][Y][3] = -shift_vec[Y];
  shift[AFT][Z][3] = -shift_vec[Z];

  // Apply the constructed rotation matrices
  Transform::combine(this->position, shift, this->position);
}

// stretch(xstretch, ystretch, zstretch) stretches the given object.
void Dagnode::stretch(double xstretch, double ystretch, double zstretch)
{
  transform stretch;

  Transform::unit(stretch);

  if (xstretch == 0) {
    output_warning("Dagnode::stretch", "X stretch factor is zero!",
		   "Remember, a unit stretch is 1.0 not 0.0. The X stretch\n"
		   "factor has been replaced with 1.0.");
    xstretch = 1;
  }
  if (ystretch == 0) {
    output_warning("Dagnode::stretch", "Y stretch factor is zero!",
		   "Remember, a unit stretch is 1.0 not 0.0. The Y stretch\n"
		   "factor has been replaced with 1.0.");
    ystretch = 1;
  }
  if (zstretch == 0) {
    output_warning("Dagnode::stretch", "Z stretch factor is zero!",
		   "Remember, a unit stretch is 1.0 not 0.0. The Z stretch\n"
		   "factor has been replaced with 1.0.");
    zstretch = 1;
  }

  // Construct the right shift matrices
  stretch[FORE][X][X] = xstretch;
  stretch[FORE][Y][Y] = ystretch;
  stretch[FORE][Z][Z] = zstretch;
  stretch[AFT][X][X] = 1.0 / xstretch;
  stretch[AFT][Y][Y] = 1.0 / ystretch;
  stretch[AFT][Z][Z] = 1.0 / zstretch;

  // Apply the constructed rotation matrices
  Transform::combine(this->position, stretch, this->position);
}

// stretch(stretchvec) stretches the given object.
void Dagnode::stretch(vector stretchvec)
{
  transform stretch;

  if (stretchvec[X] == 0) {
    output_warning("Dagnode::stretch", "X stretch factor is zero!",
		   "Remember, a unit stretch is 1.0 not 0.0. The X stretch\n"
		   "factor has been replaced with 1.0.");
    stretchvec[X] = 1;
  }
  if (stretchvec[Y] == 0) {
    output_warning("Dagnode::stretch", "Y stretch factor is zero!",
		   "Remember, a unit stretch is 1.0 not 0.0. The Y stretch\n"
		   "factor has been replaced with 1.0.");
    stretchvec[Y] = 1;
  }
  if (stretchvec[Z] == 0) {
    output_warning("Dagnode::stretch", "Z stretch factor is zero!",
		   "Remember, a unit stretch is 1.0 not 0.0. The Z stretch\n"
		   "factor has been replaced with 1.0.");
    stretchvec[Z] = 1;
  }

  Transform::unit(stretch);

  // Construct the right shift matrices
  stretch[FORE][X][X] = stretchvec[X];
  stretch[FORE][Y][Y] = stretchvec[Y];
  stretch[FORE][Z][Z] = stretchvec[Z];
  stretch[AFT][X][X] = 1.0 / stretchvec[X];
  stretch[AFT][Y][Y] = 1.0 / stretchvec[Y];
  stretch[AFT][Z][Z] = 1.0 / stretchvec[Z];

  // Apply the constructed rotation matrices
  Transform::combine(this->position, stretch, this->position);
}

// reflect(the_axis) reflects the given object in the given axis..
void Dagnode::reflect(axis the_axis)
{
  transform reflect;

  Transform::unit(reflect);

  // Construct the right shift matrices
  reflect[FORE][the_axis][the_axis] = -1.0;
  reflect[AFT] [the_axis][the_axis] = -1.0;

  // Apply the constructed rotation matrices
  Transform::combine(this->position, reflect, this->position);
}

// scale(scale_factor) scales the given object.
void Dagnode::scale(double scale_factor)
{
  transform scale;

  if (scale_factor == 0) {
    output_warning("Dagnode::scale", "Scale factor is zero!",
		   "Remember, a unit scale is 1.0 not 0.0. The scale factor\n"
		   "has been replaced with 1.0.");
    scale_factor = 1;
  }

  Transform::unit(scale);

  // Construct the right shift matrices
  scale[FORE][X][X] = scale_factor;
  scale[FORE][Y][Y] = scale_factor;
  scale[FORE][Z][Z] = scale_factor;
  scale[AFT][X][X] = 1.0 / scale_factor;
  scale[AFT][Y][Y] = 1.0 / scale_factor;
  scale[AFT][Z][Z] = 1.0 / scale_factor;

  // Apply the constructed rotation matrices
  Transform::combine(this->position, scale, this->position);
}

// shear(fromaxis, towardsaxis, amount) shears the given object.
// The point at distance 1 along fromaxis goes to the same point
// that is (amount) moved along the towardsaxis.
// e.g. shear(X, Z, 0.5) takes (1, 0, 0) to (1, 0, 0.5).
void Dagnode::shear(axis from, axis to, double amount)
{
  transform shear;

  Transform::unit(shear);

  // Construct the shear matrices
  shear[FORE][to][from] = amount;
  shear[AFT][to][from] = -amount;

  // Apply the constructed rotation matrices
  Transform::combine(this->position, shear, this->position);
}

// Link another object to this object and return a pointer to the
// object on the end of the chain. NB must be called with a 'pure'
// dagnode (that hasn't been linked to anything yet).
// May be linked with the modes PLUS, MINUS, PAINTMINUS, AND, or SURF_AND.
// Usage: myObj = myObj->link(PLUS, myObj2);
object Dagnode::link(linktype mode, object tolink)
{
  if (tolink->linkptr != NULL) {
    // Can't link this part on.
    output_warning("link", "attempt to link impure object",
		   "Objects are linked in a chain. The link command\n"
		   "when applied to a (possibly already chained) object\n"
		   "adds the object you pass it to the end of the chain\n"
		   "and returns the new end of the chain. This means that\n"
		   "the object to be added must not be part of a chain\n"
		   "(a \'pure\' object). To turn a chain of objects into\n"
		   "a pure object, use the bind() command to create a\n"
		   "pointer that points to the chain as a single object.");
    return this;
  } else {
    // Link the object onto the end of the chain and return
    // a pointer to it.
    tolink->linkmode = mode;
    tolink->linkptr = this;

    return tolink;
  }
}

// Place a wrapper Dagnode around a Light, Camera or Primitive if necessary
// before linking them into the DAG graph.
object Dagnode::link(linktype mode, Light *l)
{
  return link(mode, Dagnode::newLight(l));
}

object Dagnode::link(linktype mode, Camera *c)
{
  return link(mode, Dagnode::newCamera(c));
}

object Dagnode::link(linktype mode, Primitive *p)
{
  return link(mode, Dagnode::newPrimitive(p));
}

// Place this object at a certain position.
void Dagnode::place(transform t)
{
  Transform::set(position, t);
}

// Add this object to the global label list.
void Dagnode::label(char *labelname)
{
  world_label_list->insert(labelname, this);
}

// Set this object to be the source of a light.
void Dagnode::lightsource(object x) {
  if ((x->flags & FLAGS_LIGHT) == 0) {
    fatal_error("Dagnode::set_light", "object to set to is not a light!",
		"You must pass an object that is a light if you want\n"
		"to set another object to be the source of this light.");
  }
  light = (Light *) x->details;
}

// Set this object to be the source of a light.
void Dagnode::lightsource(Light *the_light) {
  light = the_light;
}

void Dagnode::initialise(object newobj, uint dag_type, int is_cluster)
{
  // Set up values for a blank dagnode.
  newobj->flags              = FLAGS_DEFAULT | dag_type;

  Transform::unit(newobj->position);
  if ((dag_type == FLAGS_PRIM) && (!is_cluster))
    // Primitives that aren't clusters are objects and have the
    // default material given to them.
    newobj->material = default_material;
  else
    newobj->material           = NULL;
  newobj->texture            = NULL;
  newobj->light              = NULL;
  newobj->linkmode           = PLUS;
  newobj->linkptr            = NULL;
  newobj->gaseous            = NULL;
  newobj->surface_distortion = NULL;
  newobj->affect             = NULL;
}




#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif
