/* dagnode.c : Type definitions for the directed acyclic graph classes.
** 
** This file defines operations on objects (rotate, shift, stretch, new etc.)
**
** Chris Butcher, 26 January 1998
*/

// The Affect class stores a function that modifies the dagnode
// corresponding to an object. It may do basically anything to this node,
// and is usually used in conjunction with the global world_label_list ... 
// i.e. it is used to change objects depending on the worldspace positions
// of other objects.
//
// An example application might be to apply labels to a camera and an object
// to be pointed at. You could then make an Affect function to look at
// the absolute positions of these two objects and point the camera at the
// object.
class Affect : public DeletableObject {
public:
  // Alter the dagnode that we are attached to.
  virtual void affect_object(class Dagnode *my_object) = 0;
};

// A Dagnode holds either a primitive, a camera, or an object (a
// chain of other dagnodes plus possible a vertical pointer).
// A (D)irected (A)cyclic (G)raph is formed of dagnodes that represents
// the CSG structure of the model (the dagnodes are linked by CSG
// operators e.g. PLUS, MINUS).
class Dagnode {
public:
  // flags holds information about the dagnode, such as whether it is
  // a camera, an object, or a primitive.
  uint flags;

  // position holds a pair of matrices that specify transformations to
  // apply to this object.
  transform position;
  
  // material holds the properties of the object and a method for
  // applying the material properties by casting additional rays.
  Material *material;

  // texture holds a texture to be applied to the object that modifies
  // the material properties before they are used, based on the hit
  // point, direction etc.
  class Texture *texture;

  // light holds a pointer to a Light object if this object is the
  // source of a light. If so, then a light appears at the origin of
  // the object. The object is lit via the lightSource() function of the
  // Light, other objects by the lightOn() function of the light.
  Light *light;

  // details is a pointer to more information about the dagnode. If the
  // dagnode is a camera (specified by flags), then it holds a
  // pointer to the Camera data structure. If the dagnode is a
  // primitive, then it holds a pointer to the Primitive data
  // structure. If the dagnode is an object then it is a dagnode pointer
  // down to the next level of the graph.
  void *details;

  // linkmode indicates the type of link to the next dagnode; the
  // kind of CSG operation to perform, in other words. May be PLUS,
  // MINUS, PAINTMINUS, SURF_AND or AND.
  linktype linkmode;

  // link is a horizontal pointer to the object that is linked to this
  // one (may be NULL for no link). The mode of the link is indicated
  // by linkmode.
  object linkptr;

  // surface_distortion is a distortion applied to the primitive's
  // surface. This is applied when the intersection is inserted into
  // the intersection list, by the IntersectionList::insert() procedure.
  // The transform distortion_transform is used to take the hitpoint
  // back into the space of the primitive that it was applied from.
  //
  // This is called with a transform that specifies which primitive space the
  // distortion should be applied in.

  void (*surface_distortion)(primnode *node, transform space_trans,
			     vector normal, double *t_val, vector ray_from,
			     vector ray_dir, uint ray_flags);

  // affect contains a pointer to an Affect instance. When traversing the
  // dag-graph in order to convert it to an RCSG graph, if any dagnode
  // has a non-NULL affect field, then its affect_object procedure is called.
  // This means that the dag-graph may be changed however you desire
  // at this stage ... that is, once the absolute positions of all the labeled
  // objects in worldspace are known. These affect routines may do such useful
  // things as make objects point to another specified object, change colour
  // depending on how far away another object is, move the objects towards
  // another object, etc, etc, the list is endless.
  //
  // The most common use is for camera and light control; making them point
  // towards a specific labelled object.
  Affect *affect;

  // The gas that this object contains (this makes the material properties
  // field meaningless).
  class Gas *gaseous;

  // Prototypes for operations that may be performed on a Dagnode.
  static void initialise(object newobj, uint dag_type, int is_cluster);
  static object newPrimitive(Primitive *aPrimitive);
  static object newCamera(class Camera *aCamera);
  static object newLight(Light *aLight);
  object bind(void);
  void rotate(axis around, double angle);
  void rotate(axis from, axis to, double angle);
  void rotate(vector given_fwd, vector given_up);
  void shift(double xshift, double yshift, double zshift);
  void shift(vector shift_vec);
  void stretch(double xstretch, double ystretch, double zstretch);
  void stretch(vector stretchvec);
  void reflect(axis the_axis);
  void scale(double scale_factor);
  void shear(axis from, axis to, double amount);
  object link(linktype mode, object tolink);
  object link(linktype mode, Light *l);
  object link(linktype mode, Camera *c);
  object link(linktype mode, Primitive *p);
  void label(char *labelname);
  void lightsource(Light *the_light);
  void lightsource(object x);
  void place(transform t);

  // Flag setting/getting operations. Inlined for efficiency.
  inline void set_flag(uint the_flag) {
    flags |= the_flag;
  }
  inline void clear_flag(uint the_flag) {
    flags &= ~the_flag;
  }
  inline void toggle_flag(uint the_flag) {
    flags ^= the_flag;
  }
  inline int get_flag(uint the_flag) {
    if (flags & the_flag)
      return TRUE;
    else
      return FALSE;
  }
};
