/* flare.c : Lensflare code.
** 
** This file holds code to initialise and apply a foreground texture
** that simulates lensflare for a camera.
**
** Chris Butcher, 27 January 1997
*/

#include "mirage.h"

#ifdef USE_NAMESPACES
namespace Mirage {
#endif

flare_parameters default_flare_parameters;

// Construct the lensflare (passed a camera, uses default flare).
Lensflare::Lensflare(Camera *the_camera, int num_lights)
{
  if (!completely_silent)
    printf("Constructing lensflare, %d lights.\n", num_lights);
  view_camera = the_camera;
  flare_param = &default_flare_parameters;

  // Construct the default lensflare. This set of parameters was created
  // by Stu Smith.
  flare_param->num_flares = 8;
  flare_param->flare_storage[0] = new CircularBrightEdgeFlare(-1.00, 0.80, 0,
							    0.12, 0.27, 0.06);
  flare_param->flare_storage[1] = new LargeRingFlare(0.50, 0.13);
  flare_param->flare_storage[2] = new CircularFlare(-0.45, 2.50, 2,
						    0.00, 0.00, 0.30);
  flare_param->flare_storage[3] = new FuzzyBlobFlare(-0.65, 7.50, 3,
						     0.60, 0.60, 0.64);
  flare_param->flare_storage[4] = new CircularBrightEdgeFlare(-0.75, 1.00, 4,
							    0.08, 0.05, 0.29);
  flare_param->flare_storage[5] = new CircularFlare(-0.91, 1.20, 5,
						    0.10, 0.00, 0.19);
  flare_param->flare_storage[6] = new CircularFlare(0.33, 2.2, 6,
						    0.03, 0.00, 0.10);
  flare_param->flare_storage[7] = new CircularFlare(0.40, 2.6, 7,
						    0.07, 0.00, 0.05);
  flare_param->light_effect = new SmoothLightEffect;

  max_lights = num_lights;
  current_lights = 0;
  found_flares = FALSE;

  if (flare_param->num_flares > LENSFLARE_MAX_FLARES)
    fatal_error("Lensflare::initialise_lensflare",
		"too many lensflares per light!",
		"You have exceeded the maximum number of flares per\n"
		"light in the flare_parameters that was passed to\n"
		"the Lensflare. Either use less flares or change\n"
		"the #define LENSFLARE_MAX_FLARES and recompile.");

  // Allocate memory for the lights' positions etc.
  light_active       =    (int *) calloc(max_lights, sizeof(int));
  light_label        =  (char **) calloc(max_lights, sizeof(char *));
  light_position     = (vector *) calloc(max_lights, sizeof(vector));
  light_radius_angle = (double *) calloc(max_lights, sizeof(double));
  light_brightness   = (double *) calloc(max_lights, sizeof(double));

  if (light_active == NULL || light_position == NULL ||
      light_radius_angle == NULL || light_brightness == NULL ||
      light_label == NULL) {
    char errbuf[MIRAGE_STRING_LENGTH];

    sprintf(errbuf, "can't allocate space for %d lights", max_lights);
    fatal_error("Lensflare::constructor", errbuf,
		"This is not a large amount of memory usually. You may\n"
		"have requested an insanely large number of lights\n"
		"or be very low on memory.");
  }

  mark_for_deletion();
}

// Construct the lensflare (passed a camera object, uses default flare).
Lensflare::Lensflare(object the_camera, int num_lights)
{
  if ((the_camera->flags & FLAGS_CAMERA) == 0)
    fatal_error("Lensflare::constructor", "passed an object not a camera",
		"You can pass a camera as an object or as a camera\n"
		"pointer to this constructor; but it must be a camera\n"
		"object.");
  view_camera = (Camera *) (the_camera->details);
  flare_param = &default_flare_parameters;

  // Construct the default lensflare. This set of parameters was created
  // by Stu Smith.
  flare_param->num_flares = 8;
  flare_param->flare_storage[0] = new CircularBrightEdgeFlare(-1.00, 0.80, 0,
							    0.12, 0.27, 0.06);
  flare_param->flare_storage[1] = new LargeRingFlare(0.50, 0.13);
  flare_param->flare_storage[2] = new CircularFlare(-0.45, 2.50, 2,
						    0.00, 0.00, 0.30);
  flare_param->flare_storage[3] = new FuzzyBlobFlare(-0.65, 7.50, 3,
						     0.60, 0.60, 0.64);
  flare_param->flare_storage[4] = new CircularBrightEdgeFlare(-0.75, 1.00, 4,
							    0.08, 0.05, 0.29);
  flare_param->flare_storage[5] = new CircularFlare(-0.91, 1.20, 5,
						    0.10, 0.00, 0.19);
  flare_param->flare_storage[6] = new CircularFlare(0.33, 2.2, 6,
						    0.03, 0.00, 0.10);
  flare_param->flare_storage[7] = new CircularFlare(0.40, 2.6, 7,
						    0.07, 0.00, 0.05);
  flare_param->light_effect = new SmoothLightEffect;

  max_lights = num_lights;
  current_lights = 0;
  found_flares = FALSE;

  if (flare_param->num_flares > LENSFLARE_MAX_FLARES)
    fatal_error("Lensflare::initialise_lensflare",
		"too many lensflares per light!",
		"You have exceeded the maximum number of flares per\n"
		"light in the flare_parameters that was passed to\n"
		"the Lensflare. Either use less flares or change\n"
		"the #define LENSFLARE_MAX_FLARES and recompile.");

  // Allocate memory for the lights' positions etc.
  light_active       =    (int *) calloc(max_lights, sizeof(int));
  light_label        =  (char **) calloc(max_lights, sizeof(char *));
  light_position     = (vector *) calloc(max_lights, sizeof(vector));
  light_radius_angle = (double *) calloc(max_lights, sizeof(double));
  light_brightness   = (double *) calloc(max_lights, sizeof(double));

  if (light_active == NULL || light_position == NULL ||
      light_radius_angle == NULL || light_brightness == NULL ||
      light_label == NULL) {
    char errbuf[MIRAGE_STRING_LENGTH];

    sprintf(errbuf, "can't allocate space for %d lights", max_lights);
    fatal_error("Lensflare::constructor", errbuf,
		"This is not a large amount of memory usually. You may\n"
		"have requested an insanely large number of lights\n"
		"or be very low on memory.");
  }

  mark_for_deletion();
}

/// Construct the lensflare (passed a camera, custom flare parameters).
Lensflare::Lensflare(flare_parameters *flare_p, Camera *the_camera,
		     int num_lights)
{
  if (!completely_silent)
    printf("Constructing lensflare, %d lights.\n", num_lights);
  view_camera = the_camera;
  flare_param = flare_p;
  max_lights = num_lights;
  current_lights = 0;
  found_flares = FALSE;

  if (flare_param->num_flares > LENSFLARE_MAX_FLARES)
    fatal_error("Lensflare::initialise_lensflare",
		"too many lensflares per light!",
		"You have exceeded the maximum number of flares per\n"
		"light in the flare_parameters that was passed to\n"
		"the Lensflare. Either use less flares or change\n"
		"the #define LENSFLARE_MAX_FLARES and recompile.");

  // Allocate memory for the lights' positions etc.
  light_active       =    (int *) calloc(max_lights, sizeof(int));
  light_label        =  (char **) calloc(max_lights, sizeof(char *));
  light_position     = (vector *) calloc(max_lights, sizeof(vector));
  light_radius_angle = (double *) calloc(max_lights, sizeof(double));
  light_brightness   = (double *) calloc(max_lights, sizeof(double));

  if (light_active == NULL || light_position == NULL ||
      light_radius_angle == NULL || light_brightness == NULL ||
      light_label == NULL) {
    char errbuf[MIRAGE_STRING_LENGTH];

    sprintf(errbuf, "can't allocate space for %d lights", max_lights);
    fatal_error("Lensflare::constructor", errbuf,
		"This is not a large amount of memory usually. You may\n"
		"have requested an insanely large number of lights\n"
		"or be very low on memory.");
  }

  mark_for_deletion();
}

// Construct the lensflare (passed a camera object, custom flare parameters).
Lensflare::Lensflare(flare_parameters *flare_p, object the_camera,
		     int num_lights)
{
  if ((the_camera->flags & FLAGS_CAMERA) == 0)
    fatal_error("Lensflare::constructor", "passed an object not a camera",
		"You can pass a camera as an object or as a camera\n"
		"pointer to this constructor; but it must be a camera\n"
		"object.");
  view_camera = (Camera *) (the_camera->details);
  flare_param = flare_p;
  max_lights = num_lights;
  current_lights = 0;
  found_flares = FALSE;

  if (flare_param->num_flares > LENSFLARE_MAX_FLARES)
    fatal_error("Lensflare::initialise_lensflare",
		"too many lensflares per light!",
		"You have exceeded the maximum number of flares per\n"
		"light in the flare_parameters that was passed to\n"
		"the Lensflare. Either use less flares or change\n"
		"the #define LENSFLARE_MAX_FLARES and recompile.");

  // Allocate memory for the lights' positions etc.
  light_active       =    (int *) calloc(max_lights, sizeof(int));
  light_label        =  (char **) calloc(max_lights, sizeof(char *));
  light_position     = (vector *) calloc(max_lights, sizeof(vector));
  light_radius_angle = (double *) calloc(max_lights, sizeof(double));
  light_brightness   = (double *) calloc(max_lights, sizeof(double));

  if (light_active == NULL || light_position == NULL ||
      light_radius_angle == NULL || light_brightness == NULL ||
      light_label == NULL) {
    char errbuf[MIRAGE_STRING_LENGTH];

    sprintf(errbuf, "can't allocate space for %d lights", max_lights);
    fatal_error("Lensflare::constructor", errbuf,
		"This is not a large amount of memory usually. You may\n"
		"have requested an insanely large number of lights\n"
		"or be very low on memory.");
  }

  mark_for_deletion();
}

// Deallocate the memory allocated in the constructor.
Lensflare::~Lensflare(void)
{
  free(light_active);
  free(light_label);
  free(light_position);
  free(light_radius_angle);
  free(light_brightness);
}

// Add a new light to the list of lights to be lensflared.
void Lensflare::add_light(vector light_pos, double light_rad,
			  double brightness)
{
  if (current_lights >= max_lights) {
    char errbuf[MIRAGE_STRING_LENGTH];

    sprintf(errbuf, "exceeded maximum lights (%d)", max_lights);
    fatal_error("Lensflare::add_light", errbuf,
		"You are trying to add a lensflare for more lights\n"
		"than the amount stated when the Lensflare was\n"
		"initialised. Change the max_lights parameter.");
  }

  // Store the lightsource's position.
  Vector::set(light_position[current_lights], light_pos);

  // Store the light's radius. Note that this is currently in world-
  // coordinates, not in radians as it will eventually be.
  light_radius_angle[current_lights] = light_rad;

  // Store the light's brightness. 1.0 is standard.
  light_brightness[current_lights] = brightness;

  // The light isn't specified by a label but rather directly as a position.
  light_label[current_lights] = NULL;

  // This light is inactive (doesn't cause flares) unless otherwise
  // activated.
  light_active[current_lights] = FALSE;

  current_lights++;
}

// Add a new light to the list of lights to be lensflared.
void Lensflare::add_light(char *light_name, double light_rad,
			  double brightness)
{
  if (current_lights >= max_lights) {
    char errbuf[MIRAGE_STRING_LENGTH];

    sprintf(errbuf, "exceeded maximum lights (%d)", max_lights);
    fatal_error("Lensflare::add_light", errbuf,
		"You are trying to add a lensflare for more lights\n"
		"than the amount stated when the Lensflare was\n"
		"initialised. Change the max_lights parameter.");
  }

  // Store the light's radius. Note that this is currently in world-
  // coordinates, not in radians as it will eventually be.
  light_radius_angle[current_lights] = light_rad;

  // Store the light's brightness. 1.0 is standard.
  light_brightness[current_lights] = brightness;

  // The light is specified by a label.
  light_label[current_lights] = light_name;

  // This light is inactive (doesn't cause flares) unless otherwise
  // activated.
  light_active[current_lights] = FALSE;

  current_lights++;
}

// Initialise the lensflare; determine the camera's position and that of all
// the lights. This step must only be called once the labels have all been
// found ... it is called from Lensflare::apply() the first time it is
// applied, i.e. when the first ray is cast.
void Lensflare::initialise_lensflare(void)
{
  int light_num, find_result, flare_num, total_flares;
  object light_obj;
  transform light_trans;
  double brightness_correction, light_distance;
  double light_image_x, light_image_y;
  double imageplane_height;
  Storage *camera_storage;
  Flare *the_flare;

  // Find the position of the camera.
  Transform::set(view_transform, view_camera->world_transform);

  // Determine the focal length (viewplane distance) of the camera.
  // This assumes a PinholeOptics and will crash if it isn't.
  viewplane_dist = ((PinholeOptics *) view_camera->myOptics)
    ->viewplane_dist;
    
  // Determine the imageplane height in Y (the imageplane is defined
  // to go from -1 to +1 in X). This is 0.75 for 400x300.
  camera_storage = view_camera->myStorage;
  imageplane_height = ((double) camera_storage->ysize) / camera_storage->xsize;
  
  // Allocate space to store the flare positions in.
  flare_point = (double *) calloc(flare_param->num_flares * current_lights,
				  2 * sizeof(double));
  if (flare_point == NULL) {
    char errbuf[MIRAGE_STRING_LENGTH];

    sprintf(errbuf, "can't allocate %d flares for %d lights",
	    flare_param->num_flares, current_lights);
    fatal_error("Lensflare::initialise_lensflare", errbuf,
		"You may have too many flares for each light,\n"
		"too many lights causing flares, or be fatally\n"
		"low on memory.");
  }

  // For each light that causes, flares, determine its position in the
  // image and where its flares will appear.
  for (light_num = 0; light_num < current_lights; light_num++) {
    if (light_label[light_num] != NULL) {
      // We have to find the worldspace coordinates of the light.
      find_result = world_label_list->find(light_label[light_num],
					   &light_obj, light_trans);

      // Now light_trans[FORE] holds a matrix from the light's primitive
      // space to worldspace. We want to know where the light's origin
      // goes to in worldspace. This is given by the numbers
      // light_trans[FORE][X|Y|Z][3].
      Vector::assign(light_position[light_num],
		     light_trans[FORE][X][3],
		     light_trans[FORE][Y][3],
		     light_trans[FORE][Z][3]);

    } else {
      // We were given the worldspace coordinates of the light when
      // it was added.
    }

    // We must now find the coordinates of the light relative to the
    // camera's coordinate system so that we can find the imageplane
    // coordinates of the flare points.
    Matrix::apply(view_transform[AFT],
		  light_position[light_num], light_position[light_num]);

    // Work out how far away the light is.
    light_distance = CHECK0(Vector::magnitude(light_position[light_num]),
				"Lensflare: camera + light coincident!");

    // Find the radius that the light subtends as an angle. We have been
    // given light_radius_angle[light_num] in worldunits and so we need
    // to work out how many radians its radius is.
    light_radius_angle[light_num] = atan2(light_radius_angle[light_num],
					  light_distance);
					  
    // Determine how close the light is to the camera's center; use this
    // to adjust the brightness. We say full brightness when the light
    // is directly in front of the camera, zero brightness when the X
    // component of the direction is 0.8 * the total direction.
    brightness_correction = light_position[light_num][X] / light_distance;
    brightness_correction = (brightness_correction - 0.8) / (1.0 - 0.8);
    CLAMP(0, brightness_correction, 1);
    light_brightness[light_num] *= brightness_correction;

    if (light_brightness[light_num] > 0) {
      // Determine the position of the light in image coordinates.
      light_image_x = - viewplane_dist * (light_position[light_num][Y]
					  / light_position[light_num][X]);
      light_image_y = - viewplane_dist * (light_position[light_num][Z]
					  / light_position[light_num][X]);

      // If the light is inside the viewable area then it will cause
      // flares. So determine the positions of the flares on the viewplane.
      // Otherwise the light is inactive by default.
      if (fabs(light_image_x) < 1.0 &&
	  fabs(light_image_y) < imageplane_height) {

	// Activate this light.
	light_active[light_num] = TRUE;
      
	// Calculate the flares' positions ... they appear in a straight
	// line through the center of the imageplane, at positions specified
	// by flare_z (+1.0 = the light position, 0.0 = the center).
	total_flares = flare_param->num_flares;

	for (flare_num = 0; flare_num < total_flares; flare_num++) {
	  the_flare = flare_param->flare_storage[flare_num];

	  flare_point[X + 2*(flare_num + total_flares * light_num)] =
	    light_image_x * the_flare->flare_z_position;
	  flare_point[Y + 2*(flare_num + total_flares * light_num)] =
	    light_image_y * the_flare->flare_z_position;
	}
      }
    }
  }

  found_flares = TRUE;
}

void Lensflare::get_flare(vector direction, color diffcol, color transcol)
{
  int light_num, flare_num, total_flares, i;
  vector camera_direction;
  double image_x, image_y;
  double dx, dy, radius_sq, theta;
  color current_flare_color, light_color, trans_amount;
  double current_flare_amount, light_amount;
  Flare *the_flare;

  // Initially the lensflare is nothing.
  Color::set(diffcol,  zero_color);
  Color::set(transcol, full_color);

  // Must assign values to these so that flares which don't won't crash
  // the system.
  current_flare_amount = 0;
  Color::set(current_flare_color, zero_color);
  
  // If we haven't called the initialisation procedure, do so.
  if (!found_flares)
    initialise_lensflare();

  // Find the direction in the camera's primitive space.
  Matrix::applynoshift(view_transform[AFT],
		       direction, camera_direction);

  // Calculate the imageplane coordinates.
  if (camera_direction[X] <= 0) {
    // Don't apply any flare.
    return;
  } else {
    image_x = - viewplane_dist * (camera_direction[Y] /
				  camera_direction[X]);
    image_y = - viewplane_dist * (camera_direction[Z] /
				  camera_direction[X]);
  }

  // Get the number of flares for each light.
  total_flares = flare_param->num_flares;

  // Apply all of the lensflares in succession.
  for (light_num = 0; light_num < current_lights; light_num++) {
    if (light_active[light_num]) {
      for (flare_num = 0; flare_num < flare_param->num_flares; flare_num++) {
	// Get the current flare.
	the_flare = flare_param->flare_storage[flare_num];

	// Calculate where in the "flare's primitive space" this ray is
	// (dx, dy go from -1 to +1 across the flare).
	dx = (image_x - flare_point[X + 2*(flare_num + 
					   light_num*total_flares)]) *
	  the_flare->flare_size_adjust / 0.05;
	dy = (image_y - flare_point[Y + 2*(flare_num +
					   light_num*total_flares)]) *
	  the_flare->flare_size_adjust / 0.05;
	
	// Compute the distance squared from the center of the flare.
	radius_sq = SQR(dx) + SQR(dy);

	if (radius_sq < 1.0) {
	  // Determine how much each flare affects the current point.
	  the_flare->apply_flare(dx, dy, radius_sq, &current_flare_amount,
				 current_flare_color);

	  current_flare_amount *= light_brightness[light_num];

	  // Apply the current_flare_color and current_flare_amount.
	  Color::mult(current_flare_color, transcol, trans_amount);
	  Color::addfrac(current_flare_amount, transcol, diffcol);

	  for (i = 0; i < num_color_components; i++)
	    transcol[i] *= (1.0 - current_flare_amount *
			    current_flare_color[i]);
	}
      }
    }
  
    // Apply an effect at the source of the current light.
    if ((flare_param->light_effect != NULL) &&
	(light_radius_angle[light_num] > 0)) {
      // Work out how far from the light as an angle we are. At this point
      // theta = cos(theta).
      theta = Vector::dot(camera_direction, light_position[light_num]) /
	(Vector::magnitude(camera_direction)
	 * Vector::magnitude(light_position[light_num]));
      if (theta > 0) {
	// Work out the actual angle.
	theta = acos(theta);
	
	if (theta < light_radius_angle[light_num]) {
	  // Get a value that goes from 1 in the middle of the light to
	  // 0 at the edge.
	  light_amount = 1 - theta / light_radius_angle[light_num];
	  CLAMP(0, light_amount, 1);
	  
	  // Determine the light effect.
	  flare_param->light_effect->apply_effect(light_amount, &light_amount,
						  light_color);
	  
	  // Apply the light effect.
	  Color::mult(transcol, light_color, trans_amount);
	  Color::addfrac(light_amount, trans_amount, diffcol);
	  Color::mult(1 - light_amount, transcol, transcol);
	}
      }
    }
  }
}

void SmoothLightEffect::apply_effect(double theta, double *light_amount,
				     color light_color)
{
  // Determine how much the light is present.
  *light_amount = 1 - CUBE(1 - SQR(theta));

  // Get a white color varying to red at the edges.
  Color::assign(light_color, 1.0,
		0.7 + *light_amount * 0.3,
		0.7 + *light_amount * 0.3);
}

void LargeRingFlare::apply_flare(double dx, double dy, double radius_sq,
				 double *flare_amount, color flare_color)
{
  if (radius_sq > 0.9) {
    *flare_amount = (radius_sq - 0.9) / 0.1;
    Color::assign(flare_color, 0.25 *      *flare_amount, 0,
		               0.25 * (1 - *flare_amount));
  } else {
    *flare_amount = 0;
  }
}

void FuzzyBlobFlare::apply_flare(double dx, double dy, double radius_sq,
				 double *flare_amount, color flare_color)
{
  vector noisept;
  double extra_red;

  // Get a small amount of red for use as noise.
  noisept[X] = 4*dx + noise_offset_val + 10;
  noisept[Y] = 2*dy                    + 10;
  extra_red = noise_2(noisept);
  if (extra_red < 0.5)
    extra_red = 0;
  else
    extra_red *= 0.1;
	    
  // This flare is brightest in the middle.
  *flare_amount = CUBE(1 - radius_sq);
	    
  Color::assign(flare_color, blob_color[R] + extra_red, blob_color[G],
		blob_color[B]);
}

void CircularFlare::apply_flare(double dx, double dy, double radius_sq,
				 double *flare_amount, color applied_color)
{
  vector noisept;
  double extra_red;

  // Get a small amount of red for use as noise.
  noisept[X] = 4*dx + noise_offset_val + 10;
  noisept[Y] = 2*dy                    + 10;
  extra_red = noise_2(noisept);
  if (extra_red < 0.5)
    extra_red = 0;
  else
    extra_red *= 0.1;
	    
  // Fade off the flare towards the edge (at 80%) evenly.
  *flare_amount = (radius_sq <= 0.8) ? 1
                                     : (1 - radius_sq) / 0.2;
	    
  Color::assign(applied_color,
		flare_color[R] + extra_red,
		flare_color[G],
		flare_color[B]);
}

void CircularBrightEdgeFlare::apply_flare(double dx, double dy,
					  double radius_sq,
					  double *flare_amount,
					  color applied_color)
{
  vector noisept;
  double extra_red, edge_amount;

  // Get a small amount of red for use as noise.
  noisept[X] = 4*dx + noise_offset_val + 10;
  noisept[Y] = 2*dy                    + 10;
  extra_red = noise_2(noisept);
  if (extra_red < 0.5)
    extra_red = 0;
  else
    extra_red *= 0.1;
	    
  // Fade off the flare towards the edge (at 80%) evenly.
  *flare_amount = (radius_sq <= 0.8) ? 1
                                     : (1 - radius_sq) / 0.2;
	    
  // The flare is brighter just in from the edge (64 - 80% radius).
  edge_amount = (radius_sq - 0.64) / (0.64 - 0.8);
  CLAMP(0, edge_amount, 1);
  *flare_amount *= (0.75 + 0.25 * edge_amount);
	    
  Color::assign(applied_color,
		flare_color[R] + extra_red,
		flare_color[G],
		flare_color[B]);
}




	

				  




#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif
