/* flare.h : Definitions and structures for lensflares.
** 
** This file defines all the structures necessary to set up and apply
** a lensflare. 
**
** Chris Butcher, 27 December 1996
*/

// A Flare holds the procedure + parameters to provide a flare from a
// simulated camera lens. It always has position and size information,
// things like color, etc. are defined in the subclasses of Flare.
class Flare : public DeletableObject {
public:
  double flare_z_position;
  double flare_size_adjust;
  
  // Apply the lensflare.
  virtual void apply_flare(double dx, double dy, double radius_sq,
			   double *flare_amount, color flare_color) = 0;

  virtual ~Flare(void) {
  }
};

class LargeRingFlare : public Flare {
public:

  // Generate a blue/red ring.
  virtual void apply_flare(double dx, double dy, double radius_sq,
			   double *flare_amount, color flare_color);

  LargeRingFlare(double z_pos, double size) {
    flare_z_position = z_pos;
    flare_size_adjust = size;

    mark_for_deletion();
  }

  virtual ~LargeRingFlare(void) {
  }
};

class FuzzyBlobFlare : public Flare {
public:
  color blob_color;
  int noise_offset_val;

  // Generate a fuzzy blob (usually small).
  virtual void apply_flare(double dx, double dy, double radius_sq,
			   double *flare_amount, color flare_color);

  FuzzyBlobFlare(double z_pos, double size, int offset,
		 double r, double g, double b) {
    flare_z_position = z_pos;
    flare_size_adjust = size;
    Color::assign(blob_color, r, g, b);
    noise_offset_val = offset;

    mark_for_deletion();
  }

  virtual ~FuzzyBlobFlare(void) {
  }
};

class CircularFlare : public Flare {
public:
  color flare_color;
  int noise_offset_val;

  // Generate a circular flare.
  virtual void apply_flare(double dx, double dy, double radius_sq,
			   double *flare_amount, color flare_color);

  CircularFlare(double z_pos, double size, int offset,
		 double r, double g, double b) {
    flare_z_position = z_pos;
    flare_size_adjust = size;
    Color::assign(flare_color, r, g, b);
    noise_offset_val = offset;

    mark_for_deletion();
  }

  virtual ~CircularFlare(void) {
  }
};

class CircularBrightEdgeFlare : public Flare {
public:
  color flare_color;
  int noise_offset_val;

  // Generate a circular flare with a bright edge.
  virtual void apply_flare(double dx, double dy, double radius_sq,
			   double *flare_amount, color flare_color);

  CircularBrightEdgeFlare(double z_pos, double size, int offset,
			  double r, double g, double b) {
    flare_z_position = z_pos;
    flare_size_adjust = size;
    Color::assign(flare_color, r, g, b);
    noise_offset_val = offset;

    mark_for_deletion();
  }

  virtual ~CircularBrightEdgeFlare(void) {
  }
};

// The LightEffect class holds an effect that is applied to a ray that is
// due to coming close to the lightsource.
class LightEffect : public DeletableObject {
public:

  // Apply the light's effect, from 0 at the center to 1 at the edge in theta.
  virtual void apply_effect(double theta, double *light_amount,
			    color light_color) = 0;

  virtual ~LightEffect(void) {
  }
};

class SmoothLightEffect : public LightEffect {
public:

  // Apply a smoothly varying reddish glare.
  virtual void apply_effect(double theta, double *light_amount,
			    color light_color);

  // Constructor has no parameters.
  SmoothLightEffect(void) {
    mark_for_deletion();
  }

  virtual ~SmoothLightEffect(void) {
  }
};

// The FlareParameters structure holds a number of Flares and a LightEffect.
typedef struct FlareParameters {
  int num_flares;

  Flare *flare_storage[LENSFLARE_MAX_FLARES];
  LightEffect *light_effect;
} flare_parameters;

class Lensflare : public DeletableObject {
public:
  flare_parameters *flare_param;
  int max_lights;
  int current_lights;
  Camera *view_camera;
  transform view_transform;
  double viewplane_dist;
  int found_flares;

  // This holds the positions of the flares in image coordinates.
  // They are indexed as [2*(flare_num + num_flares * light_num) + (X|Y)].
  double *flare_point;

  // Arrays of length max_lights, allocated in constructor.
  int *light_active;
  char **light_label;
  vector *light_position;
  double *light_radius_angle;
  double *light_brightness;

  // Constructor: uses the default lensflare effect.
  Lensflare(Camera *the_camera, int num_lights);
  Lensflare(object the_camera,  int num_lights);

  // Constructor: given a camera and parameter type.
  Lensflare(flare_parameters *flare_p, Camera *the_camera, int num_lights);
  Lensflare(flare_parameters *flare_p, object the_camera,  int num_lights);

  // Set up the lens flare to be applied. Must be called only once the
  // positions of the camera and lights are known; this is called automatically
  // from the FlareForeground texture.
  void initialise_lensflare(void);

  // Add a light that needs to be lensflared.
  void add_light(vector position, double light_radius, double brightness);
  void add_light(char *labelname, double light_radius, double brightness);

  // Determine the lensflare's contribution for a ray cast from the camera
  // in a given direction.
  void get_flare(vector direction, color diffcol, color transcol);

  // Free the memory allocated by the constructor.
  ~Lensflare(void);
};

// A foreground texture that applies a Lensflare.
class FlareForeground : public BackgroundTexture {
public:
  Lensflare *the_flare;

  virtual void apply(vector from, vector dir, uint ray_flags, color col,
		     RayTracer *rayengine) {
    color flare_diff, flare_trans;

    if (IS_PRIMARY(ray_flags)) {
      the_flare->get_flare(dir, flare_diff, flare_trans);

      col[R] = col[R] * flare_trans[R] + flare_diff[R];
      col[G] = col[G] * flare_trans[G] + flare_diff[G];
      col[B] = col[B] * flare_trans[B] + flare_diff[B];
    }
  }

  FlareForeground(Lensflare *a_flare) {
    the_flare = a_flare;
  }

  virtual void print_tex(void) {
    printf("FlareForeground texture.\n");
  }
};

