/* gaseous.h : Definitions for the Gas object.
** 
** This file has prototypes for the Gas object and an example gas.
**
** Chris Butcher, 28 November 1996
*/

// The gasses all subclass the generic Gas.
class Gas : public DeletableObject {
public:
  virtual void traverse_gas(transform gas_transform, 
			    vector ray_from, vector ray_dir,
			    color hit_color, double t_entry,
			    double t_exit) = 0;

  virtual ~Gas(void) {
  }
};

// The GasSFX class contains details of the gaseous object attached to
// a set of primnodes.
class GasSFX : public DeletableObject {
public:
  // Construct the object from the parameters provided.
  GasSFX(Gas *the_gas, transform gas_trans) {
    gas = the_gas;
    Transform::set(gas_transform, gas_trans);
    
    mark_for_deletion();
  }

  // The gas that is contained in the object.
  class Gas *gas;

  // A transform specifying the worldspace of the gas.
  transform gas_transform;

  virtual ~GasSFX(void) {};
};

// The DistortSFX class contains details of the surface distortion attached to
// a set of primnodes.
class DistortSFX : public DeletableObject {
public:
  // Construct the object from the parameters provided.
  DistortSFX(void (*distortion)(PrimitiveNode *node, transform space_trans,
				vector normal, double *t_val, vector ray_from,
				vector ray_dir, uint ray_flags),
	     transform dis_trans) {
    surface_distortion = distortion;
    Transform::set(distort_transform, dis_trans);

    mark_for_deletion();
  }

  // surface_distortion is a distortion applied to the primitive's
  // surface. This is applied when the intersection is inserted into
  // the intersection list, by the IntersectionList::insert() procedure.
  // The transform distortion_transform is used to take the hitpoint
  // back into the space of the primitive that it was applied from.
  void (*surface_distortion)(PrimitiveNode *node, transform space_trans,
			     vector normal, double *t_val, vector ray_from,
			     vector ray_dir, uint ray_flags);

  // A transform that specifies which primitive space the distortion
  // should be applied in.
  transform distort_transform;

  virtual ~DistortSFX(void) {};
};

// The Fog gas.
class Fog : public Gas {
public:
  color fogColor;
  double fogFactor;

  virtual void traverse_gas(transform gas_transform, vector ray_from,
			    vector ray_dir, color hit_color,
			    double t_entry, double t_exit)
  {
    double fogDist, fogAmount;
    vector p, d;

    Matrix::apply(gas_transform[AFT], ray_from, p);
    Matrix::applynoshift(gas_transform[AFT], ray_dir, d);

    // Don't apply any fogging that occurs before the ray start point.
    if (t_entry < 0) {
      t_entry = 0;
      if (t_entry > t_exit)
	return;
    }

    if (t_exit == HIT_BACKGROUND) {
      // Return the fog colour (completely fogged out).
      Color::set(hit_color, fogColor);
      return;
    }

    // Work out how far the ray traveled in the primitive fog space.
    fogDist = (t_exit - t_entry) * Vector::magnitude(ray_dir);

    // Work out how much light from the background can be seen.
    fogAmount = pow(fogFactor, fogDist);
  
    // Mix from the ray's color to the fog color.
    Color::mix(fogColor, hit_color, fogAmount, hit_color);
  }

  Fog(double r, double g, double b, double fog_f) {
    Color::assign(fogColor, r, g, b);
    fogFactor = fog_f;
  }

  Fog(double r, double g, double b, double a, double fog_f) {
    Color::assign(fogColor, r, g, b, 0, 0, 0, a);
    fogFactor = fog_f;
  }

  Fog(double r, double g, double b,
      double er, double eg, double eb,
      double a, double fog_f) {
    Color::assign(fogColor, r, g, b, er, eg, eb, a);
    fogFactor = fog_f;
  }

  // Parameters passed to this are the background color and the
  // strength of the fog; 0.25 is 25% of light remains after one world-unit,
  // 75% is the background colour. After two world units it would be 6%
  // to 94%. A good value is 0.95 for scenes on the ten world-unit scale
  // of things.
  virtual ~Fog(void) {
  }
};
