/* grid.c : Uniform grid traversal routines
** 
** This file contains functions used by all of the different cluster classes
** to cast rays into a uniform-grid subdivision.
**
** Chris Butcher, 14 January 1997
*/

#include "mirage.h"

#ifdef USE_NAMESPACES
namespace Mirage {
#endif

// This routine traces a ray through the uniform grid structure.
int UniformClusterPrimitive::grid_ray_intersect(
			 primnode *cluster_primnode,
			 primnode *validate_primnode,
			 IntersectionList *above_list, vector ray_from,
			 vector ray_dir, uint ray_flags,
			 signature ray_signature)
{
  vector ray_entry;
  double t_in_x,  t_in_y,  t_in_z;
  double t_out_x, t_out_y, t_out_z;
  double t_entry, t_exit;
  double dir_x, dir_y, dir_z;
  double start_cell_x, start_cell_y, start_cell_z;
  double ray_length;
  double t_next_x, t_next_y, t_next_z;
  double delta_t_per_x_cell, delta_t_per_y_cell, delta_t_per_z_cell;
  double cell_frac_x, cell_frac_y, cell_frac_z;
  int x_cell_no, y_cell_no, z_cell_no;
  int x_incr, y_incr, z_incr;
  int starts_inside_bounds;
  double cell_entry_t, cell_exit_t;
  unsigned int array_index;
  thing this_cell;
  primnode *the_primnode;
  int found_solid, cell_solid;
  intersection *list_position;

  dir_x = ray_dir[X];
  dir_y = ray_dir[Y];
  dir_z = ray_dir[Z];

  // Calculate the points where the ray enters and exits the
  // cluster's bounding volume.

  if (dir_x == 0) {
    // The ray has no X direction.
    if (ray_from[X] >= maxbound[X] || ray_from[X] <= minbound[X]) {
      // The ray is outside the X bounds, no intersections possible.
      return FALSE;
    } else {
      // The ray is inside the X bounds for its whole length.
      t_in_x  = -INFINITY;
      t_out_x = +INFINITY;
    }
  } else if (dir_x > 0) {
    // Calculate the in and out t-values where the ray enters and exits.
    t_in_x  = (minbound[X] - ray_from[X]) / dir_x;
    t_out_x = (maxbound[X] - ray_from[X]) / dir_x;
  } else {
    // Calculate the in and out t-values where the ray enters and exits.
    t_in_x  = (maxbound[X] - ray_from[X]) / dir_x;
    t_out_x = (minbound[X] - ray_from[X]) / dir_x;
  }

  if (dir_y == 0) {
    // The ray has no Y direction.
    if (ray_from[Y] >= maxbound[Y] || ray_from[Y] <= minbound[Y]) {
      // The ray is outside the Y bounds, no intersections possible.
      return FALSE;
    } else {
      // The ray is inside the Y bounds for its whole length.
      t_in_y  = -INFINITY;
      t_out_y = +INFINITY;
    }
  } else if (dir_y > 0) {
    // Calculate the in and out t-values where the ray enters and exits.
    t_in_y  = (minbound[Y] - ray_from[Y]) / dir_y;
    t_out_y = (maxbound[Y] - ray_from[Y]) / dir_y;
  } else {
    // Calculate the in and out t-values where the ray enters and exits.
    t_in_y  = (maxbound[Y] - ray_from[Y]) / dir_y;
    t_out_y = (minbound[Y] - ray_from[Y]) / dir_y;
  }

  if (dir_z == 0) {
    // The ray has no Z direction.
    if (ray_from[Z] >= maxbound[Z] || ray_from[Z] <= minbound[Z]) {
      // The ray is outside the Z bounds, no intersections possible.
      return FALSE;
    } else {
      // The ray is inside the Z bounds for its whole length.
      t_in_z  = -INFINITY;
      t_out_z = +INFINITY;
    }
  } else if (dir_z > 0) {
    // Calculate the in and out t-values where the ray enters and exits.
    t_in_z  = (minbound[Z] - ray_from[Z]) / dir_z;
    t_out_z = (maxbound[Z] - ray_from[Z]) / dir_z;
  } else {
    // Calculate the in and out t-values where the ray enters and exits.
    t_in_z  = (maxbound[Z] - ray_from[Z]) / dir_z;
    t_out_z = (minbound[Z] - ray_from[Z]) / dir_z;
  }

  // Work out the correct t-values where the ray enters and exits the
  // bounding volume.
  t_entry = MAX3(t_in_x,  t_in_y,  t_in_z);
  t_exit  = MIN3(t_out_x, t_out_y, t_out_z);
  
  if (t_exit <= t_entry)
    // The ray doesn't intersect the bounding box.
    return FALSE;

  if (t_exit <= 0)
    // The ray is outside the bounding box, pointing away.
    return FALSE;

  // Find the coordinates of the point where the ray enters the bounding
  // box (or where the ray starts).

  Vector::set(ray_entry, ray_from);
  if (t_entry > 0) {
    starts_inside_bounds = FALSE;
    // The ray enters the bounding box at positive t, set ray_entry
    // to be this point rather than ray_from.
    Vector::addfrac(t_entry, ray_dir, ray_entry);
  } else {
    starts_inside_bounds = TRUE;
  }

  // Get the coordinates of the ray_entry point in terms of cells
  // from the minimum corner.
  start_cell_x = (ray_entry[X] - minbound[X]) / cell_size[X];
  start_cell_y = (ray_entry[Y] - minbound[Y]) / cell_size[Y];
  start_cell_z = (ray_entry[Z] - minbound[Z]) / cell_size[Z];

  x_cell_no = (int) start_cell_x;
  y_cell_no = (int) start_cell_y;
  z_cell_no = (int) start_cell_z;

  cell_frac_x = start_cell_x - x_cell_no;
  cell_frac_y = start_cell_y - y_cell_no;
  cell_frac_z = start_cell_z - z_cell_no;

  // Ensure that we start inside the cluster (not in a boundary cell).
  if (x_cell_no < 0) {
    x_cell_no = 0;
    cell_frac_x = 0;
  } else if (x_cell_no > array_size - 1) {
    x_cell_no = array_size - 1;
    cell_frac_x = 1;
  }

  if (y_cell_no < 0) {
    y_cell_no = 0;
    cell_frac_y = 0;
  } else if (y_cell_no > array_size - 1) {
    y_cell_no = array_size - 1;
    cell_frac_y = 1;
  }

  if (z_cell_no < 0) {
    z_cell_no = 0;
    cell_frac_z = 0;
  } else if (z_cell_no > array_size - 1) {
    z_cell_no = array_size - 1;
    cell_frac_z = 1;
  }

  // Set up the t values to the next X, Y or Z cell boundary
  // Also determine the delta-t that the ray takes to cross
  // an X, Y or Z cell.
  ray_length = CHECK0(Vector::magnitude(ray_dir),
		      "grid_ray_intersect:zero ray direction");

  if (dir_x == 0) {
    // The ray is parallel to the X boundaries and never crosses them.
    t_next_x = INFINITY;
    delta_t_per_x_cell = INFINITY;
  } else if (dir_x > 0) {
    // We are travelling from low X to high X across the cells.

    // Work out how long in t it takes to cross an X cell.
    delta_t_per_x_cell = cell_size[X] / dir_x;

    // Work out how long in t until we cross the next boundary.
    t_next_x = (1.0 - cell_frac_x) * delta_t_per_x_cell;

    // If it's going to be a looong time, flag it as being so.
    if (t_next_x >= INFINITY) {
      t_next_x = INFINITY;
      delta_t_per_x_cell = INFINITY;
    }

  } else {
    // We are travelling from high X to low X across the cells.

    // Work out how long in t it takes to cross an X cell.
    delta_t_per_x_cell = cell_size[X] / -dir_x;

    // Work out how long in t until we cross the next boundary.
    t_next_x = cell_frac_x * delta_t_per_x_cell;

    // If it's going to be a looong time, flag it as being so.
    if (t_next_x >= INFINITY) {
      t_next_x = INFINITY;
      delta_t_per_x_cell = INFINITY;
    }
  }

  if (dir_y == 0) {
    // The ray is parallel to the Y boundaries and never crosses them.
    t_next_y = INFINITY;
    delta_t_per_y_cell = INFINITY;

  } else if (dir_y > 0) {
    // We are travelling from low Y to high Y across the cells.

    // Work out how long in t it takes to cross a Y cell.
    delta_t_per_y_cell = cell_size[Y] / dir_y;

    // Work out how long in t until we cross the next boundary.
    t_next_y = (1.0 - cell_frac_y) * delta_t_per_y_cell;

    // If it's going to be a looong time, flag it as being so.
    if (t_next_y >= INFINITY) {
      t_next_y = INFINITY;
      delta_t_per_y_cell = INFINITY;
    }

  } else {
    // We are travelling from high Y to low Y across the cells.

    // Work out how long in t it takes to cross a Y cell.
    delta_t_per_y_cell = cell_size[Y] / -dir_y;

    // Work out how long in t until we cross the next boundary.
    t_next_y = cell_frac_y * delta_t_per_y_cell;

    // If it's going to be a looong time, flag it as being so.
    if (t_next_y >= INFINITY) {
      t_next_y = INFINITY;
      delta_t_per_y_cell = INFINITY;
    }
  }

  if (dir_z == 0) {
    // The ray is parallel to the Z boundaries and never crosses them.
    t_next_z = INFINITY;
    delta_t_per_z_cell = INFINITY;

  } else if (dir_z > 0) {
    // We are travelling from low Z to high Z across the cells.

    // Work out how long in t it takes to cross a Z cell.
    delta_t_per_z_cell = cell_size[Z] / dir_z;

    // Work out how long in t until we cross the next boundary.
    t_next_z = (1.0 - cell_frac_z) * delta_t_per_z_cell;

    // If it's going to be a looong time, flag it as being so.
    if (t_next_z >= INFINITY) {
      t_next_z = INFINITY;
      delta_t_per_z_cell = INFINITY;
    }

  } else {
    // We are travelling from high Z to low Z across the cells.

    // Work out how long in t it takes to cross a Z cell.
    delta_t_per_z_cell = cell_size[Z] / -dir_z;

    // Work out how long in t until we cross the next boundary.
    t_next_z = cell_frac_z * delta_t_per_z_cell;

    // If it's going to be a looong time, flag it as being so.
    if (t_next_z >= INFINITY) {
      t_next_z = INFINITY;
      delta_t_per_z_cell = INFINITY;
    }
  }

  // Currently t_next_x, t_next_y and t_next_z are measured relative
  // to the point where we started traversal. If the ray starts outside
  // the bounds of the cluster, then we must add t_entry.
  if (!starts_inside_bounds) {
    t_next_x += t_entry;
    t_next_y += t_entry;
    t_next_z += t_entry;
  }

  // Get the increments that we wish to step by when we cross
  // a cell border.
  x_incr = (dir_x < 0) ? -1 : +1;
  y_incr = (dir_y < 0) ? -y_step : +y_step;
  z_incr = (dir_z < 0) ? -z_step : +z_step;

  // Get the initial index for the array (cell position
  // start_cell_x, start_cell_y, start_cell_z).
  array_index = (x_cell_no + 1) +
                (y_cell_no + 1) * y_step +
                (z_cell_no + 1) * z_step;

  // The initial entry is at zero t (relative to ray_entry here)
  cell_entry_t = 0;
  found_solid = FALSE;

  // **********************************************************

  // Step through cells looking for intersections.
  while (1) {

    this_cell = my_array[array_index];

    //    printf("grid: cell\n");

    if (!IS_EMPTY(this_cell)) {
      // This cell isn't empty.
      
      //      printf("grid: cell not empty\n");

      if (IS_CLUSTER_BORDER(this_cell)) {
	// We have hit the edge of the cluster's bounds.
	// Now we need to delete any intersections which were put back
	// into the intersection list by validate_intersections, in
	// the hope that they were valid in a later cell. They will still
	// be in the intersection list.
	//	printf("grid: hit border\n");

	tidy_intersections(above_list, cluster_primnode);
	return found_solid;
      }

      // Find the t-value at which we will exit this current cell.
      cell_exit_t = MIN3(t_next_x, t_next_y, t_next_z);

      if (IS_PRIM(this_cell)) {
	// If this cell is full of the primitive, then
	// we don't need to intersect with it.
	if (IS_FULL(this_cell)) {
	  //	  printf("grid: full primnode\n");

	  cell_solid = FALSE;
	} else {
	  //	  printf("grid: surface primnode\n");

	  // This cell contains a primnode.
	  the_primnode = GET_PRIMNODE(this_cell);
	  
	  if (check_ray_flags(ray_flags, the_primnode->flags)) {
	    //	    printf("grid: passed c_r_flags\n");

	    if (the_primnode->last_ray_signature == ray_signature) {
	      // We've already intersected this ray with this
	      // primitive.
	      cell_solid = the_primnode->last_ray_hit;

	      //	      printf("grid: last_ray_hit\n");

	      if (cell_solid) {
		cell_solid = validate_intersections(above_list,
						    cluster_primnode,
						    validate_primnode,
						    ray_from, ray_dir,
						    ray_flags,
						    VALIDATE_PRIMNODE,
						    this_cell,
						    cell_entry_t,
						    cell_exit_t);
		
		//		printf("grid: validated prim, %s\n", (cell_solid) ? "T" : "F");
	      }

	    } else {
	      // Intersect with the primitive.
	      list_position = above_list->get_position();
	      cell_solid = the_primnode->prim->ray_intersect(above_list,
							     the_primnode,
							     cluster_primnode,
							     ray_from,
							     ray_dir,
							     ray_flags,
							     ray_signature);
	      above_list->set_position(list_position);

	      //	      printf("grid: done r_i w/ prim\n");
	    }

	    if (cell_solid) {
	      cell_solid = validate_intersections(above_list,
						  cluster_primnode,
						  validate_primnode,
						  ray_from, ray_dir,
						  ray_flags,
						  VALIDATE_PRIMNODE,
						  this_cell,
						  cell_entry_t,
						  cell_exit_t);

	      //	      printf("grid: validated prim, %s\n", (cell_solid) ? "T" : "F");
	    }
	  } else {
	    //	    printf("grid: prim r_i return false\n");
	    cell_solid = FALSE;
	  }
	}
      } else {
	// This cell contains an RCSG node.
	//	printf("grid: cell is RCSG\n");
	list_position = above_list->get_position();
	cell_solid = rcsg_ray_intersect(above_list, this_cell,
					cluster_primnode,
					ray_from, ray_dir,
					ray_flags, ray_signature);
	above_list->set_position(list_position);
	
	//	printf("grid: RCSG r_i done\n");

	if (cell_solid) {
	  cell_solid = validate_intersections(above_list,
					      cluster_primnode,
					      validate_primnode,
					      ray_from, ray_dir,
					      ray_flags,
					      VALIDATE_RCSG, this_cell,
					      cell_entry_t,
					      cell_exit_t);

	  //	  printf("grid: RCSG validates as %s\n", cell_solid ? "T" : "F");
	} else {
	  //	  printf("grid: RCSG r_i return false\n");
	  cell_solid = FALSE;
	}
      }

      if (cell_solid) {
	if (stop_at_first_intersection) {
	  tidy_intersections(above_list, cluster_primnode);
	  return TRUE;
	} else {
	  found_solid = TRUE;
	}
      }
    }

    // Skip to the next cell ...
    if (t_next_x <= t_next_y) {
      if (t_next_x <= t_next_z) {
	// The next step is in X.

	cell_entry_t = t_next_x;
	t_next_x += delta_t_per_x_cell;
	array_index += x_incr;
      } else {
	// The next step is in Z.

	cell_entry_t = t_next_z;
	t_next_z += delta_t_per_z_cell;
	array_index += z_incr;
      }
    } else {
      if (t_next_y <= t_next_z) {
	// The next step is in Y.

	cell_entry_t = t_next_y;
	t_next_y += delta_t_per_y_cell;
	array_index += y_incr;
      } else {
	// The next step is in Z.

	cell_entry_t = t_next_z;
	t_next_z += delta_t_per_z_cell;
	array_index += z_incr;
      }
    }
  }
}

// Convert the octree structure into a uniform-grid array.
void UniformClusterPrimitive::flatten_octree(thing the_thing, thing *the_array,
					     int x_min, int y_min, int z_min,
					     int x_max, int y_max, int z_max)
{
  // If we have a primnode in the octree and we're not at the minimum
  // level of subdivision yet, we must subdivide the primnode further.
  if (IS_PRIM(the_thing) && (x_max - x_min > 1)) {
    thing octree_thing, *octree_children;
    int x_mid, y_mid, z_mid;
    double world_x_min, world_y_min, world_z_min;
    double world_x_max, world_y_max, world_z_max;
   
    // Work out the world-corners of the cell that needs subdividing.
    world_x_min = minbound[X] + x_min * cell_size[X];
    world_y_min = minbound[Y] + y_min * cell_size[Y];
    world_z_min = minbound[Z] + z_min * cell_size[Z];
    world_x_max = minbound[X] + x_max * cell_size[X];
    world_y_max = minbound[Y] + y_max * cell_size[Y];
    world_z_max = minbound[Z] + z_max * cell_size[Z];

    // Subdivide the primnode.
    octree_thing = create_octree_from_prim(the_thing,
				   world_x_min, world_y_min, world_z_min,
				   world_x_max, world_y_max, world_z_max);

    octree_children = GET_OCTREE(octree_thing)->child;

    x_mid = (x_min + x_max) / 2;
    y_mid = (y_min + y_max) / 2;
    z_mid = (z_min + z_max) / 2;

    // For each child, place the child node of the primnode's octree
    // into the array in the correct place.
    flatten_octree(octree_children[0], the_array,
		   x_min, y_min, z_min, x_mid, y_mid, z_mid);

    flatten_octree(octree_children[1], the_array,
		   x_mid, y_min, z_min, x_max, y_mid, z_mid);

    flatten_octree(octree_children[2], the_array,
		   x_min, y_mid, z_min, x_mid, y_max, z_mid);

    flatten_octree(octree_children[3], the_array,
		   x_mid, y_mid, z_min, x_max, y_max, z_mid);

    flatten_octree(octree_children[4], the_array,
		   x_min, y_min, z_mid, x_mid, y_mid, z_max);

    flatten_octree(octree_children[5], the_array,
		   x_mid, y_min, z_mid, x_max, y_mid, z_max);

    flatten_octree(octree_children[6], the_array,
		   x_min, y_mid, z_mid, x_mid, y_max, z_max);

    flatten_octree(octree_children[7], the_array,
		   x_mid, y_mid, z_mid, x_max, y_max, z_max);

    free_octreenode(octree_thing);
    return;
  } else if (IS_OCTREE(the_thing)) {
    int x_mid, y_mid, z_mid;
    thing *octree_children;

    octree_children = GET_OCTREE(the_thing)->child;

    x_mid = (x_min + x_max) / 2;
    y_mid = (y_min + y_max) / 2;
    z_mid = (z_min + z_max) / 2;

    // For each child of this octree, place it into the array in the
    // correct place.
    flatten_octree(octree_children[0], the_array,
		   x_min, y_min, z_min, x_mid, y_mid, z_mid);

    flatten_octree(octree_children[1], the_array,
		   x_mid, y_min, z_min, x_max, y_mid, z_mid);

    flatten_octree(octree_children[2], the_array,
		   x_min, y_mid, z_min, x_mid, y_max, z_mid);

    flatten_octree(octree_children[3], the_array,
		   x_mid, y_mid, z_min, x_max, y_max, z_mid);

    flatten_octree(octree_children[4], the_array,
		   x_min, y_min, z_mid, x_mid, y_mid, z_max);

    flatten_octree(octree_children[5], the_array,
		   x_mid, y_min, z_mid, x_max, y_mid, z_max);

    flatten_octree(octree_children[6], the_array,
		   x_min, y_mid, z_mid, x_mid, y_max, z_max);

    flatten_octree(octree_children[7], the_array,
		   x_mid, y_mid, z_mid, x_max, y_max, z_max);

    return;
  } else {
    // We have a thing that isn't an octree node. It is either an
    // RCSG node (in which case we must have reached the bottom
    // level of subdivision since it would be subdivided if it could)
    // or a primnode (in which case we must have reached the bottom
    // level of subdivision otherwise it would have been subdivided
    // above).
    //
    // Store it in the correct cell in the array (account for the
    // one cell's padding around the edges).

    my_array[(x_min + 1) + (y_min + 1) * y_step + (z_min + 1) * z_step] =
      the_thing;

    return;
  }
}

#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif
