/* heightfield.c : The Heightfield primitive.
** 
** This file defines the functions for the Heightfield primitive.
**
** Chris Butcher, 7 April 1997
*/

#include "mirage.h"

#ifdef USE_NAMESPACES
namespace Mirage {
#endif

int Heightfield::ray_intersect(IntersectionList *the_list, 
			       struct PrimitiveNode *node,
			       struct PrimitiveNode *cluster_node,
			       vector ray_from, vector ray_dir,
			       uint ray_flags, signature ray_signature)
{
  double p_x, p_y, p_z, d_x, d_y, d_z;
  double t_x_entry, t_x_exit, t_y_entry, t_y_exit, t_z_entry, t_z_exit;
  double scaled_x, scaled_y, entry_height;
  int entry_x, entry_y, exit_x, exit_y;
  vector norm;
  int found_intersection = FALSE;
  double initial_height = 0, final_height, exit_height;
  intersect_result start_inout;

  validate_node = cluster_node;

  // Store the signature of this ray that we're intersecting with.
  node->last_ray_signature = ray_signature;

  // This apply is an inline function defined in matrix.h
  // Transform the ray startpoint into primitive space
  Matrix::apply(node->position[AFT], ray_from, &p_x, &p_y, &p_z);

  // This apply is also an inline function defined in matrix.h
  // Transform the ray direction into primitive space (note: shift does
  // not apply to this, only apply rotation + stretch sections of the matrix)
  Matrix::applynoshift(node->position[AFT], ray_dir, &d_x, &d_y, &d_z);

  if (d_x == 0) {
    // The ray has no X direction component.
    if (p_x <= 0 || p_x >= 1) {
      // The ray is outside the heightfield in X always.
      node->last_ray_hit = FALSE;
      return node->last_ray_hit;
    } else {
      // The ray is inside the heightfield in X always.
      t_x_entry = -INFINITY;
      t_x_exit = +INFINITY;
    }
  } else if (d_x > 0) {
    // The ray is travelling in +X.
    t_x_entry = (0 - p_x) / d_x;
    t_x_exit  = (1 - p_x) / d_x;
  } else {
    // The ray is travelling in -X.
    t_x_entry = (1 - p_x) / d_x;
    t_x_exit  = (0 - p_x) / d_x;
  }

  if (d_y == 0) {
    // The ray has no Y direction component.
    if (p_y <= 0 || p_y >= 1) {
      // The ray is outside the heightfield in Y always.
      node->last_ray_hit = FALSE;
      return node->last_ray_hit;
    } else {
      // The ray is inside the heightfield in Y always.
      t_y_entry = -INFINITY;
      t_y_exit = +INFINITY;
    }
  } else if (d_y > 0) {
    // The ray is travelling in +Y.
    t_y_entry = (0 - p_y) / d_y;
    t_y_exit  = (1 - p_y) / d_y;
  } else {
    // The ray is travelling in -Y.
    t_y_entry = (1 - p_y) / d_y;
    t_y_exit  = (0 - p_y) / d_y;
  }

  if (d_z == 0) {
    // The ray has no Z direction component.
    if (p_z <= -1 || p_z >= 1) {
      // The ray is outside the heightfield in Z always.
      node->last_ray_hit = FALSE;
      return node->last_ray_hit;
    } else {
      // The ray is inside the heightfield in Z always.
      t_z_entry = -INFINITY;
      t_z_exit = +INFINITY;
    }
  } else if (d_z > 0) {
    // The ray is travelling in +Z.
    t_z_entry = (-1 - p_z) / d_z;
    t_z_exit  = ( 1 - p_z) / d_z;
  } else {
    // The ray is travelling in -Z.
    t_z_entry = ( 1 - p_z) / d_z;
    t_z_exit  = (-1 - p_z) / d_z;
  }

  // Work out when the ray enters and exits the heightfield's bounding box.
  if (t_x_entry > t_y_entry) {
    if (t_x_entry > t_z_entry) {
      t_entry = t_x_entry;
      entry_plane = X;
    } else {
      t_entry = t_z_entry;
      entry_plane = Z;
    }
  } else {
    if (t_y_entry > t_z_entry) {
      t_entry = t_y_entry;
      entry_plane = Y;
    } else {
      t_entry = t_z_entry;
      entry_plane = Z;
    }
  }

  if (t_x_exit < t_y_exit) {
    if (t_x_exit < t_z_exit) {
      t_exit = t_x_exit;
      exit_plane = X;
    } else {
      t_exit = t_z_exit;
      exit_plane = Z;
    }
  } else {
    if (t_y_exit < t_z_exit) {
      t_exit = t_y_exit;
      exit_plane = Y;
    } else {
      t_exit = t_z_exit;
      exit_plane = Z;
    }
  }

  if (t_entry > t_exit) {
    // The ray never enters the bounding volume.
    node->last_ray_hit = FALSE;
    return node->last_ray_hit;
  }

  // Classify the ray's entry intersection with the bounding volume.

  // The entry plane intersection has to be checked to see
  // if it is a valid intersection (i.e. value above the surface threshold).
  // In the course of this we also work out what initial_value is.

  if (entry_plane == X) {
    // Work out the place on the X plane where the ray enters.
    scaled_y     = (p_y + t_entry * d_y) * (y_res - 1);
    entry_height = (p_z + t_entry * d_z) * HF_MAXVAL;

    // Is this a low-X or high-X entry?
    if (d_x > 0) {
      // This enters the volume through the -X bounding plane. Determine
      // the field value at this point if we have to.
      entry_x = 0;
      if (entry_height > 0) {
	// We have to check whether or not the heightfield's surface is
	// above the entry point.
	initial_height = x_face_val(entry_x, scaled_y);
      } else {
	// The heightfield's surface value doesn't matter, it will definitely
	// be above the ray's entry.
	initial_height = 0;
      }
      
      // If this is above the threshold and not too close to the from point,
      // then we need to insert an entry intersection.
      if (initial_height > entry_height) {
	start_inout = IN;
	if (t_entry > MIN_TVAL_CUTOFF) {
	  norm[X] = -1;
	  norm[Y] = 0;
	  norm[Z] = 0;
	  
	  found_intersection = TRUE;
	  
	  insert_entering(the_list, node, validate_node, norm, t_entry,
			  ray_from, ray_dir, ray_flags);
	}
      } else {
	// The ray enters above the surface.
	start_inout = OUT;
      }
    } else {
      // This enters the volume through the +X bounding plane. Determine
      // the field value at this point.
      entry_x = x_res - 1;
      if (entry_height > 0) {
	// We have to check whether or not the heightfield's surface is
	// above the entry point.
	initial_height = x_face_val(entry_x, scaled_y);
      } else {
	// The heightfield's surface value doesn't matter, it will definitely
	// be above the ray's entry.
	initial_height = 0;
      }
      
      // If this is above the threshold and not too close to the from point,
      // then we need to insert an entry intersection.
      if (initial_height > entry_height) {
	start_inout = IN;
	if (t_entry > MIN_TVAL_CUTOFF) {
	  norm[X] = +1;
	  norm[Y] = 0;
	  norm[Z] = 0;
	  
	  found_intersection = TRUE;
	  
	  insert_entering(the_list, node, validate_node, norm, t_entry,
			  ray_from, ray_dir, ray_flags);
	}
      } else {
	start_inout = OUT;
      }
    }
  } else if (entry_plane == Y) {
    // Work out the place on the Y plane where the ray enters.
    scaled_x     = (p_x + t_entry * d_x) * (x_res - 1);
    entry_height = (p_z + t_entry * d_z) * HF_MAXVAL;

    // Is this a low-Y or high-Y entry?
    if (d_y > 0) {
      // This enters the volume through the -Y bounding plane. Determine
      // the field value at this point if we have to.
      entry_y = 0;
      if (entry_height > 0) {
	// We have to check whether or not the heightfield's surface is
	// above the entry point.
	initial_height = y_face_val(scaled_x, entry_y);
      } else {
	// The heightfield's surface value doesn't matter, it will definitely
	// be above the ray's entry.
	initial_height = 0;
      }
      
      // If this is above the threshold and not too close to the from point,
      // then we need to insert an entry intersection.
      if (initial_height > entry_height) {
	start_inout = IN;
	if (t_entry > MIN_TVAL_CUTOFF) {
	  norm[X] = 0;
	  norm[Y] = -1;
	  norm[Z] = 0;
	  
	  found_intersection = TRUE;
	  
	  insert_entering(the_list, node, validate_node, norm, t_entry,
			  ray_from, ray_dir, ray_flags);
	}
      } else {
	start_inout = OUT;
      }
    } else {
      // This enters the volume through the +Y bounding plane. Determine
      // the field value at this point.
      entry_y = y_res - 1;
      if (entry_height > 0) {
	// We have to check whether or not the heightfield's surface is
	// above the entry point.
	initial_height = y_face_val(scaled_x, entry_y);
      } else {
	// The heightfield's surface value doesn't matter, it will definitely
	// be above the ray's entry.
	initial_height = 0;
      }
      
      // If this is above the threshold and not too close to the from point,
      // then we need to insert an entry intersection.
      if (initial_height > entry_height) {
	start_inout = IN;
	if (t_entry > MIN_TVAL_CUTOFF) {
	  norm[X] = 0;
	  norm[Y] = +1;
	  norm[Z] = 0;
	  
	  found_intersection = TRUE;
	  
	  insert_entering(the_list, node, validate_node, norm, t_entry,
			  ray_from, ray_dir, ray_flags);
	}
      } else {
	start_inout = OUT;
      }
    }
  } else if (entry_plane == Z) {
    // Is this a low-Z or high-Z entry?
    if (d_z > 0) {
      // This enters the volume through the -Z bounding plane. If this is not
      // too close to the from point, then we need to insert an entry
      // intersection.
      start_inout = IN;
      if (t_entry > MIN_TVAL_CUTOFF) {
	norm[X] = 0;
	norm[Y] = 0;
	norm[Z] = -1;

	found_intersection = TRUE;

	insert_entering(the_list, node, validate_node, norm, t_entry,
			ray_from, ray_dir, ray_flags);
      }
    } else {
      // The ray enters through the +Z surface.
      start_inout = OUT;
    }
  } else {
    // This will never happen - keep the compiler happy.
    start_inout = OUT;
  }

  // Classify the ray's exit intersection with the bounding volume.

  // The exit plane intersection has to be checked to see
  // if it is a valid intersection (i.e. value above the surface threshold).
  if (exit_plane == X) {
    // Work out the place on the X plane where the ray exits.
    scaled_y    = (p_y + t_exit * d_y) * (y_res - 1);
    exit_height = (p_z + t_exit * d_z) * HF_MAXVAL;

    // Is this a low-X or high-X exit?
    if (d_x < 0) {
      // This exits the volume through the -X bounding plane. Determine
      // the field value at this point if we have to.
      exit_x = 0;
      if (exit_height > 0) {
	// We have to check whether or not the heightfield's surface is
	// above the entry point.
	final_height = x_face_val(exit_x, scaled_y);
      } else {
	// The heightfield's surface value doesn't matter, it will definitely
	// be above the ray's entry.
	final_height = 0;
      }
      
      // If this is above the threshold and not too close to the from point,
      // then we need to insert an exit intersection.
      if ((final_height > exit_height) && (t_exit > MIN_TVAL_CUTOFF)) {
	norm[X] = -1;
	norm[Y] = 0;
	norm[Z] = 0;

	insert_leaving(the_list, node, validate_node, norm, t_exit,
		       ray_from, ray_dir, ray_flags);
      }
    } else {
      // This exits the volume through the +X bounding plane. Determine
      // the field value at this point.
      exit_x = x_res - 1;
      if (exit_height > 0) {
	// We have to check whether or not the heightfield's surface is
	// above the exit point.
	final_height = x_face_val(exit_x, scaled_y);
      } else {
	// The heightfield's surface value doesn't matter, it will definitely
	// be above the ray's exit.
	final_height = 0;
      }
      
      // If this is above the threshold and not too close to the from point,
      // then we need to insert an exit intersection.
      if ((final_height > exit_height) && (t_exit > MIN_TVAL_CUTOFF)) {
	norm[X] = +1;
	norm[Y] = 0;
	norm[Z] = 0;

	insert_leaving(the_list, node, validate_node, norm, t_exit,
		       ray_from, ray_dir, ray_flags);
      }
    }
  } else if (exit_plane == Y) {
    // Work out the place on the Y plane where the ray exits.
    scaled_x    = (p_x + t_exit * d_x) * (x_res - 1);
    exit_height = (p_z + t_exit * d_z) * HF_MAXVAL;

    // Is this a low-Y or high-Y exit?
    if (d_y < 0) {
      // This exits the volume through the -Y bounding plane. Determine
      // the field value at this point if we have to.
      exit_y = 0;
      if (exit_height > 0) {
	// We have to check whether or not the heightfield's surface is
	// above the entry point.
	final_height = y_face_val(scaled_x, exit_y);
      } else {
	// The heightfield's surface value doesn't matter, it will definitely
	// be above the ray's entry.
	final_height = 0;
      }
      
      // If this is above the threshold and not too close to the from point,
      // then we need to insert an exit intersection.
      if ((final_height > exit_height) && (t_exit > MIN_TVAL_CUTOFF)) {
	norm[X] = 0;
	norm[Y] = -1;
	norm[Z] = 0;

	insert_leaving(the_list, node, validate_node, norm, t_exit,
		       ray_from, ray_dir, ray_flags);
      }
    } else {
      // This exits the volume through the +Y bounding plane. Determine
      // the field value at this point.
      exit_y = y_res - 1;
      if (exit_height > 0) {
	// We have to check whether or not the heightfield's surface is
	// above the exit point.
	final_height = y_face_val(scaled_x, exit_y);
      } else {
	// The heightfield's surface value doesn't matter, it will definitely
	// be above the ray's exit.
	final_height = 0;
      }
      
      // If this is above the threshold and not too close to the from point,
      // then we need to insert an exit intersection.
      if ((final_height > exit_height) && (t_exit > MIN_TVAL_CUTOFF)) {
	norm[X] = 0;
	norm[Y] = +1;
	norm[Z] = 0;

	insert_leaving(the_list, node, validate_node, norm, t_exit,
		       ray_from, ray_dir, ray_flags);
      }
    }
  } else if (exit_plane == Z) {
    // Is this a low-Z or high-Z exit?
    if (d_z < 0) {
      // This exits the heightfield through the -Z bounding plane. If this is
      // not too close to the from point, then we need to insert an exit
      // intersection.
      if (t_exit > MIN_TVAL_CUTOFF) {
	norm[X] = 0;
	norm[Y] = 0;
	norm[Z] = -1;

	insert_leaving(the_list, node, validate_node, norm, t_exit,
		       ray_from, ray_dir, ray_flags);
      }
    }
  }

  // If we've found an intersection, stop if necessary.
  if (stop_at_first_intersection && found_intersection) {
    node->last_ray_hit = TRUE;
    return node->last_ray_hit;
  }

  // If the ray enters below Z = 0 then we must work out whether or not
  // it goes through the 'real' part of the heightfield, from Z = 0 to 1.
  if (p_z + t_entry * d_z < 0) {
    t_entry = (0 - p_z) / d_z;
    entry_plane = Z;
    if (t_entry > t_exit) {
      // The ray doesn't intersect the active part of the heightfield.
      node->last_ray_hit = found_intersection;
      return node->last_ray_hit;
    }
  }

  // If the ray exits below Z = 0 then we must work out the real t-value
  // when it exits the active part of the heightfield (at Z = 0).
  if (p_z + t_exit * d_z < 0) {
    t_exit = (0 - p_z) / d_z;
    exit_plane = Z;
  }

  // Intersect with the surface here.
  if (intersect_with_surface(p_x, p_y, p_z, d_x, d_y, d_z,
			     t_entry, t_exit, start_inout,
			     the_list, node, ray_from, ray_dir,
			     ray_flags, ray_signature))
    node->last_ray_hit = TRUE;
  else
    node->last_ray_hit = found_intersection;

  return node->last_ray_hit;
}

int Heightfield::ray_intersect(vector ray_from, vector ray_dir,
			       rayintersect *int_storage)
{
  // This routine is just too damn complicated to copy. It's not
  // likely to be needed for the interactive-CSG system, either.
  fatal_error("Heightfield::ray_intersect", "Incorrect r_i call!",
	      "This error should not happen.");
  return 0;
}

int Heightfield::intersect_with_surface(double p_x, double p_y, double p_z,
					double d_x, double d_y, double d_z,
					double t_entry, double t_exit,
					intersect_result start_inout,
					IntersectionList *the_list,
					primnode *node, vector ray_from,
					vector ray_dir, uint ray_flags,
					signature ray_signature)
{
  double entry_x, entry_y, entry_z;
  double start_cell_x, start_cell_y, start_height;
  double cell_frac_x, cell_frac_y;
  double ray_length, t_next_x, t_next_y, cell_entry_t, cell_exit_t;
  double delta_t_per_x_cell, delta_t_per_y_cell;
  int x_cell_no, y_cell_no, x_incr, y_incr;
  int x_leave_side, y_leave_side;
  intersect_result cell_entry_inout, cell_exit_inout;
  int found_intersection, this_cell_found, ray_starts_inside;

  // Find the position where the ray enters the heightfield (alternatively,
  // if the ray starts inside the heightfield, the position where the ray
  // starts).
  if (t_entry > 0) {
    // The ray starts outside the volume and hits it in +t.
    ray_starts_inside = FALSE;
    entry_x = p_x + t_entry * d_x;
    entry_y = p_y + t_entry * d_y;
    entry_z = p_z + t_entry * d_z;
  } else if (t_exit < 0) {
    // The ray starts heading away from the volume.
    return FALSE;
  } else {
    // The ray startsinside the volume and exits it in +t.
    ray_starts_inside = TRUE;
    entry_x = p_x;
    entry_y = p_y;
    entry_z = p_z;
  }

  // Get the coordintes of the ray entry in terms of cells.
  start_cell_x = entry_x * (x_res - 1);
  start_cell_y = entry_y * (y_res - 1);
  start_height = entry_z * HF_MAXVAL;

  x_cell_no = (int) start_cell_x;
  y_cell_no = (int) start_cell_y;

  cell_frac_x = start_cell_x - x_cell_no;
  cell_frac_y = start_cell_y - y_cell_no;

  // Ensure that we start in the right heightfield cell.
  if (x_cell_no < 0) {
    x_cell_no = 0;
    cell_frac_x = 0;
  } else if (x_cell_no > x_res - 2) {
    x_cell_no = x_res - 2;
    cell_frac_x = 1;
  }

  if (y_cell_no < 0) {
    y_cell_no = 0;
    cell_frac_y = 0;
  } else if (y_cell_no > y_res - 2) {
    y_cell_no = y_res - 2;
    cell_frac_y = 1;
  }

  // Set up the t values to the next X and Y cell boundaries. Also determine
  // the delta-t that the ray takes to cross an X or Y cell.
  ray_length = sqrt(SQR(d_x) + SQR(d_y));

  if (d_x == 0) {
    // The ray is parallel to the X boundaries and never crosses them.
    t_next_x = INFINITY;
    delta_t_per_x_cell = INFINITY;
  } else if (d_x > 0) {
    // We are travelling from low X to high X across the cells.

    // Work out how long in t it takes to cross an X cell.
    delta_t_per_x_cell = 1.0 / (d_x * (x_res - 1));

    // Work out how long in t until we cross the next boundary.
    t_next_x = (1.0 - cell_frac_x) * delta_t_per_x_cell;

    // If it's going to be a looong time, flag it as being so.
    if (t_next_x >= INFINITY) {
      t_next_x = INFINITY;
      delta_t_per_x_cell = INFINITY;
    }

  } else {
    // We are travelling from high X to low X across the cells.

    // Work out how long in t it takes to cross an X cell.
    delta_t_per_x_cell = 1.0 / (-d_x * (x_res - 1));

    // Work out how long in t until we cross the next boundary.
    t_next_x = cell_frac_x * delta_t_per_x_cell;

    // If it's going to be a looong time, flag it as being so.
    if (t_next_x >= INFINITY) {
      t_next_x = INFINITY;
      delta_t_per_x_cell = INFINITY;
    }
  }

  if (d_y == 0) {
    // The ray is parallel to the Y boundaries and never crosses them.
    t_next_y = INFINITY;
    delta_t_per_y_cell = INFINITY;
  } else if (d_y > 0) {
    // We are travelling from low Y to high Y across the cells.

    // Work out how long in t it takes to cross a Y cell.
    delta_t_per_y_cell = 1.0 / (d_y * (y_res - 1));

    // Work out how long in t until we cross the next boundary.
    t_next_y = (1.0 - cell_frac_y) * delta_t_per_y_cell;

    // If it's going to be a looong time, flag it as being so.
    if (t_next_y >= INFINITY) {
      t_next_y = INFINITY;
      delta_t_per_y_cell = INFINITY;
    }
  } else {
    // We are travelling from high Y to low Y across the cells.

    // Work out how long in t it takes to cross a Y cell.
    delta_t_per_y_cell = 1.0 / (-d_y * (y_res - 1));

    // Work out how long in t until we cross the next boundary.
    t_next_y = cell_frac_y * delta_t_per_y_cell;

    // If it's going to be a looong time, flag it as being so.
    if (t_next_y >= INFINITY) {
      t_next_y = INFINITY;
      delta_t_per_y_cell = INFINITY;
    }
  }

  // We are currently measuring t_next_x and t_next_y relative to the
  // point where traversal starts. If the ray starts outside the bounds
  // of the heightfield, we must add t_entry.
  if (!ray_starts_inside) {
    t_next_x += t_entry;
    t_next_y += t_entry;
    cell_entry_t = t_entry;
  } else {
    // The ray starts inside the volume. We must recalculate start_inout
    // to work out whether the start point is inside or outside the surface.
    start_inout = point_inout(start_cell_x, start_cell_y, start_height);
    cell_entry_t = 0;
  }

  // Get the directions to step in.
  x_incr = (d_x < 0) ? -1 : +1;
  y_incr = (d_y < 0) ? -1 : +1;

  // Get the sides of the current cell to check when leaving.
  x_leave_side = (d_x < 0) ? 0 : 1;
  y_leave_side = (d_y < 0) ? 0 : 1;

  found_intersection = FALSE;

  cell_entry_inout = start_inout;

  // Setup complete.

  // ****************************************************************
  
  // Step through cells looking for intersections.

  while (1) {

    // Work out the next cell that we will go to ... and process the current
    // cell for intersections if we find ourselves on the opposite side of the
    // surface.
    if (t_next_x <= t_next_y) {
      // The next step is in X.
      cell_exit_t = t_next_x;
	
      // When we leave the cell, are we in or out?
      cell_exit_inout = x_face_inout(x_cell_no + x_leave_side,
				     (p_y + cell_exit_t * d_y) * (y_res - 1),
				     (p_z + cell_exit_t * d_z) * HF_MAXVAL);
	
      // If we've crossed the surface, intersect with it in the current
      // cell. We must not do this if we started in this cell (cell entry t
      // was 0), otherwise we may get erroneous self-shadowing.
      if ((cell_entry_t > HF_MIN_CELL_TVAL) &&
	  (cell_exit_inout != cell_entry_inout)) {
	
	this_cell_found = get_intersection(x_cell_no, y_cell_no,
					   cell_entry_t, cell_exit_t,
					   cell_entry_inout, cell_exit_inout,
					   p_x, p_y, p_z, d_x, d_y, d_z, 
					   the_list, node, ray_from, ray_dir,
					   ray_flags, ray_signature);
	if (this_cell_found)
	  found_intersection = TRUE;
	  
	if (stop_at_first_intersection && found_intersection)
	  return TRUE;
      }
    
      if ((t_next_x > t_exit) && (t_next_y > t_exit) && (exit_plane == Z)) {
	// We have gone through the top (or bottom) of the heightfield. Exit.
	return found_intersection;
      }
    
      t_next_x += delta_t_per_x_cell;
      x_cell_no += x_incr;
    } else {
      // The next step is in Y.
      cell_exit_t = t_next_y;
	
      // When we leave the cell, what is the surface value?
      cell_exit_inout = y_face_inout((p_x + cell_exit_t * d_x) * (x_res - 1),
				     y_cell_no + y_leave_side,
				     (p_z + cell_exit_t * d_z) * HF_MAXVAL);
	
      // If we've crossed the surface, intersect with it in the current
      // cell. We must not do this if we started in this cell (cell entry t
      // was 0), otherwise we may get erroneous self-shadowing.
      if ((cell_entry_t > HF_MIN_CELL_TVAL) &&
	  (cell_exit_inout != cell_entry_inout)) {

	this_cell_found = get_intersection(x_cell_no, y_cell_no,
					   cell_entry_t, cell_exit_t,
					   cell_entry_inout, cell_exit_inout,
					   p_x, p_y, p_z, d_x, d_y, d_z, 
					   the_list, node, ray_from, ray_dir,
					   ray_flags, ray_signature);

	if (this_cell_found)
	  found_intersection = TRUE;
	  
	if (stop_at_first_intersection && found_intersection)
	  return TRUE;
      }

      if ((t_next_x > t_exit) && (t_next_y > t_exit) && (exit_plane == Z)) {
	// We have gone through the top (or bottom) of the heightfield. Exit.
	return found_intersection;
      }
    
      t_next_y += delta_t_per_y_cell;
      y_cell_no += y_incr;
    }

    // Check to see if we just left the heightfield. If so, stop traversing.
    if ((x_cell_no < 0         && x_incr < 0) ||
	(y_cell_no < 0         && y_incr < 0) ||
	(x_cell_no > x_res - 2 && x_incr > 0) ||
	(y_cell_no > y_res - 2 && y_incr > 0))
      return found_intersection;

    // Set up for the traversal of the next cell.
    cell_entry_t = cell_exit_t;
    cell_entry_inout = cell_exit_inout;
  }

  // We never get here.
  return FALSE;
}

int Heightfield::get_intersection(int x_index, int y_index,
				  double cell_entry_t, double cell_exit_t,
				  intersect_result cell_entry_inout,
				  intersect_result cell_exit_inout,
				  double from_x, double from_y,
				  double from_z, double dir_x,
				  double dir_y, double dir_z, 
				  IntersectionList *the_list,
				  primnode *node, vector ray_from,
				  vector ray_dir, uint ray_flags,
				  signature ray_signature)
{
  double hit_x, hit_y, hit_z, px, py, pz, dx, dy, dz;
  double a, b, c;
  double q_coeff[3], q_sols[3];
  double val0, val1, val2, val3;
  int array_index, num_sols, current_sol, found_intersection = FALSE;;
  double intersect_t;
  vector norm;

  array_index = x_index * x_step + y_index * y_step;

  // Get the values of the field at the vertices.
  val0 = height_data[array_index                  ];
  val1 = height_data[array_index + x_step         ];
  val2 = height_data[array_index          + y_step];
  val3 = height_data[array_index + x_step + y_step];

  // Get the from point and direction vector in the 'primitive space' of the
  // cell ... i.e. the edges of the cell map to 0-1 X and Y, 0-HF_MAXVAL Z.
  px = from_x * (x_res - 1) - x_index;
  py = from_y * (y_res - 1) - y_index;
  pz = from_z *  HF_MAXVAL;
  dx = dir_x * (x_res - 1);
  dy = dir_y * (y_res - 1);
  dz = dir_z *  HF_MAXVAL;

  // Calculate coefficients that describe the surface in the cell.
  a = val0 - val1 - val2 + val3;
  b = -val0 + val1;
  c = -val0 + val2;

  // Form the quadratic equation.
  q_coeff[2] = a * dx * dy;
  q_coeff[1] = dx * (b + (py * a)) + dy * (c + px * a) - dz;
  q_coeff[0] = val0 + b * px + c * py + a * px * py - pz;

  // Solve the quadratic equation.
  num_sols = solveQuadric(q_coeff, q_sols);

  for (current_sol = 0; current_sol < num_sols; current_sol++) {
    intersect_t = q_sols[current_sol];

    if ((intersect_t >= cell_entry_t) &&
	(intersect_t <= cell_exit_t) &&
	(intersect_t > MIN_TVAL_CUTOFF)) {

      // Get the hitpoint in the cell's primitive space.
      hit_x = px + dx * intersect_t;
      hit_y = py + dy * intersect_t;
      hit_z = pz + dz * intersect_t;

      if (interpolate_normals) {
	// Get the hitpoint in terms of cells.
	hit_x += x_index;
	hit_y += y_index;

	// Get the interpolated normal of the heightfield at this point.
	find_cell_normal(hit_x, hit_y, norm);
	Vector::normalise(norm, "zero normal in heightfield:get_intersection");
      } else {
	// Get the gradient of the pseudo-linear field in this
	// single cell.
	norm[X] = - (x_res - 1) * (a * hit_y + b);
	norm[Y] = - (y_res - 1) * (a * hit_x + c);
	norm[Z] = HF_MAXVAL;
	Vector::normalise(norm, "zero normal in heightfield:get_intersection");
      }

      if (cell_entry_inout == OUT) {
	// We started this cell outside the surface.
	insert_entering(the_list, node, validate_node, norm, intersect_t,
			ray_from, ray_dir, ray_flags);
      } else {
	// We started this cell inside the surface.
	insert_leaving(the_list, node, validate_node, norm, intersect_t,
		       ray_from, ray_dir, ray_flags);
      }

      found_intersection = TRUE;
    }
  }

  return found_intersection;
}

// Test to see if a single point is inside the heightfield (a point in
// primitive space).
intersect_result Heightfield::point_intersect(primnode *node, vector apoint,
					      PointInfo *info,
					      uint ray_flags)
{
  double p_x, p_y, p_z;

  // Transform the point into primitivespace.
  Matrix::apply(node->position[AFT], apoint, &p_x, &p_y, &p_z);

  // Check to see if the point is inside the volume's bounding box.
  if ((p_z > -1) && (p_z < 1) &&
      (p_x >  0) && (p_x < 1) &&
      (p_y >  0) && (p_y < 1)) {
    
    if (p_z < 0) {
      // We are inside the block portion of the heightfield.
      // Return an IN result.
      set_point_contents(node, info);
      return IN;
    }

    // The point lies inside the volume. Sample the field at this point.
    if (p_z * HF_MAXVAL < interpolate_height(p_x * (x_res - 1),
					     p_y * (y_res - 1))) {
      // Return an IN result.
      set_point_contents(node, info);
      return IN;
    } else {
      // Return an OUT result.
      return OUT;
    }
  }
  
  // The point is outside the heightfield's bounding box.
  return OUT;
}

intersect_result Heightfield::point_intersect(vector apoint)
{
  // This routine is just too damn complicated to copy. It's not
  // likely to be needed for the interactive-CSG system, either.
  fatal_error("Heightfield::point_intersect", "Incorrect p_i call!",
	      "This error should not happen.");
  return OUT;
}

// Calculate the value of the surface at a point.
double Heightfield::interpolate_height(double px, double py)
{
  int cell_x, cell_y;
  double f_x, f_y, b_x, b_y;
  int array_index;
  double found_val;

  // Calculate the cell that we are in.
  cell_x = (int) px;
  cell_y = (int) py;

  // Calculate where we are in the cell (forward and backward fractions).
  f_x = px - cell_x;
  f_y = py - cell_y;
  b_x = 1.0 - f_x;
  b_y = 1.0 - f_y;

  // Calculate the array index at the base of the cell (SW corner).
  array_index = cell_x + cell_y * y_step;

  // Interpolate the values from the four vertices of the current cell.
  found_val  = b_x * b_y * height_data[array_index];
  found_val += f_x * b_y * height_data[array_index + x_step];
  found_val += b_x * f_y * height_data[array_index + y_step];
  found_val += f_x * f_y * height_data[array_index + x_step + y_step];

  return found_val;
}

double Heightfield::x_face_val(int cell_x, double py)
{
  int cell_y;
  double f_y, b_y;
  int array_index;
  double found_val;

  // Calculate the cell that we are in.
  cell_y = (int) py;

  // Calculate where we are in the cell (forward and backward fractions).
  f_y = py - cell_y;
  b_y = 1.0 - f_y;

  // Calculate the array index at the base of the face (W edge).
  array_index = cell_x + cell_y * y_step;

  // Interpolate the values from the two vertices of the current edge.
  found_val  = b_y * height_data[array_index];
  found_val += f_y * height_data[array_index + y_step];

  return found_val;
}

double Heightfield::y_face_val(double px, int cell_y)
{
  int cell_x;
  double f_x, b_x;
  int array_index;
  double found_val;

  // Calculate the cell that we are in.
  cell_x = (int) px;

  // Calculate where we are in the cell (forward and backward fractions).
  f_x = px - cell_x;
  b_x = 1.0 - f_x;

  // Calculate the array index at the base of the face (S edge).
  array_index = cell_x + cell_y * y_step;

  // Interpolate the values from the two vertices of the current edge.
  found_val  = b_x * height_data[array_index];
  found_val += f_x * height_data[array_index + x_step];

  return found_val;
}

intersect_result Heightfield::point_inout(double px, double py,
					  double height)
{
  double hf_height;

  hf_height = interpolate_height(px, py);
  if (hf_height > height)
    return IN;
  else
    return OUT;
}

intersect_result Heightfield::x_face_inout(int cell_x, double py,
					   double height)
{
  double hf_height;

  hf_height = x_face_val(cell_x, py);
  if (hf_height > height)
    return IN;
  else
    return OUT;
}

intersect_result Heightfield::y_face_inout(double px, int cell_y,
					   double height)
{
  double hf_height;

  hf_height = y_face_val(px, cell_y);
  if (hf_height > height)
    return IN;
  else
    return OUT;
}

void Heightfield::find_cell_normal(double px, double py, vector norm)
{
  int cell_x, cell_y;
  double f_x, f_y, b_x, b_y;
  vector vertex_normal[4];

  // Determine which cell of the volume we are in.
  cell_x = (int) px;
  cell_y = (int) py;

  // Determine the normal of the heightfield at each of the four corners of
  // the current cell.
  get_vertex_normal(cell_x,     cell_y,     vertex_normal[0]);
  get_vertex_normal(cell_x + 1, cell_y,     vertex_normal[1]);
  get_vertex_normal(cell_x,     cell_y + 1, vertex_normal[2]);
  get_vertex_normal(cell_x + 1, cell_y + 1, vertex_normal[3]);

  // Calculate where we are in the cell (forward and backward fractions).
  f_x = px - cell_x;
  f_y = py - cell_y;
  b_x = 1.0 - f_x;
  b_y = 1.0 - f_y;

  // Initialise the normal to zero.
  Vector::assign(norm, 0, 0, 0);

  // Interpolate the normal from the four corner normals.
  Vector::addfrac(b_x * b_y, vertex_normal[0], norm);
  Vector::addfrac(f_x * b_y, vertex_normal[1], norm);
  Vector::addfrac(b_x * f_y, vertex_normal[2], norm);
  Vector::addfrac(f_x * f_y, vertex_normal[3], norm);
}

void Heightfield::get_vertex_normal(int vx, int vy, vector norm)
{
  long array_index;

  // Determine the position in the heightfield data array.
  array_index = vx * x_step + vy * y_step;
  
  // Calculate an approximation to the partial derivative of the field
  // in X. We go one cell on either side unless this would take us outside
  // the bounds of the volume.
  if (vx == 0)
    norm[X] = height_data[array_index + x_step] - height_data[array_index];
  else if (vx == (x_res - 1))
    norm[X] = height_data[array_index] - height_data[array_index - x_step];
  else
    norm[X] = (height_data[array_index + x_step]
                   - height_data[array_index - x_step]) / 2.0;

  // Calculate an approximation to the partial derivative of the field
  // in Y. We go one cell on either side unless this would take us outside
  // the bounds of the volume.
  if (vy == 0)
    norm[Y] = height_data[array_index + y_step] - height_data[array_index];
  else if (vy == (y_res - 1))
    norm[Y] = height_data[array_index] - height_data[array_index - y_step];
  else
    norm[Y] = (height_data[array_index + y_step]
                   - height_data[array_index - y_step]) / 2.0;

  norm[X] = -norm[X] * x_res;
  norm[Y] = -norm[Y] * y_res;
  norm[Z] = HF_MAXVAL;
  Vector::normalise(norm, "zero normal in heightfield:get_vertex_normal");
}

// Test to see if a voxel intersects with the heightfield's bounding
// volume (X, Y from 0 - 1, Z from -1 - 1).
intersect_result Heightfield::voxel_intersect(box v)
{
  box cubeprim_box;
  int i;
  intersect_result cube_result;

  // Calculate the equivalent box relative to a -1 - +1 cube in
  // X, Y and Z.
  for (i = 0; i < 8; i++) {
    cubeprim_box[i][X] = 2*v[i][X] - 1;
    cubeprim_box[i][Y] = 2*v[i][Y] - 1;
    cubeprim_box[i][Z] = v[i][Z];
  }

  cube_result = cube_voxel_intersect(cubeprim_box);

  // If the voxel doesn't intersect the bounding box, return OUT.
  if (cube_result == OUT)
    return OUT;

  // Otherwise, the heightfield will need to be tested against.
  // This is not a real voxel intersection test; ideally this would actually
  // check the voxel against the cells of the heightfield to see if it
  // intersected the surface.
  return SURFACE;
}

// Work out the height of a point in the heightfield's primitive space.
double Heightfield::get_height(double x, double y)
{
  double px, py;

  // What is this in cell coordinates?
  px = x * (x_res - 1);
  py = y * (y_res - 1);

  if (px < 0)
    px = 0;
  if (px >= (x_res - 1))
    px = x_res - 1.0001;
  if (py < 0)
    py = 0;
  if (py >= (y_res - 1))
    py = y_res - 1.0001;

  // Work out the height of the heightfield at this point.
  return interpolate_height(px, py) / HF_MAXVAL;
}

// Work out the normal at a point on the heightfield's surface (in its
// primitive space).
void Heightfield::get_normal(double x, double y, vector norm)
{
  double px, py;

  // What is this in cell coordinates?
  px = x * (x_res - 1);
  py = y * (y_res - 1);

  if (px < 0)
    px = 0;
  if (px >= (x_res - 1))
    px = x_res - 1.0001;
  if (py < 0)
    py = 0;
  if (py >= (y_res - 1))
    py = y_res - 1.0001;

  // Work out the normal at this point.
  find_cell_normal(px, py, norm);
}

#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif
