/* heightfield.h : Header file for the Heightfield primitive.
** 
** This file defines the prototypes and definitions for the Heightfield
** primitive.
**
** Chris Butcher, 7 April 1997
*/

//   The Heightfield primitive is defined by a field of values in the region
// (0, 0) to (1, 1). The surface of the object is obtained as a height
// specified at points in this region.
//   Currently the only method of field storage supported is a uniform array
// of field values stored at the vertices of cells.
class Heightfield : public Primitive {
public:
  // Variables that hold the data for the Heightfield primitive.
  height_val *height_data;
  int height_data_allocated_internally;

  int x_res, y_res;
  int x_step, y_step;
  int stop_at_first_intersection;
  int interpolate_normals;
  axis entry_plane, exit_plane;

  // Stores the primnode which will validate our intersections.
  struct PrimitiveNode *validate_node;

  // Entry/exit points of the ray under consideration.
  double t_entry, t_exit;

  // Intersection interface (subclassed from Primitive)
  virtual int ray_intersect(IntersectionList *the_list, 
			    struct PrimitiveNode *node,
			    struct PrimitiveNode *cluster_node,
			    vector ray_from, vector ray_dir, uint ray_flags,
			    signature ray_signature);

  virtual intersect_result point_intersect(primnode *node, vector apoint,
					   PointInfo *info,
					   uint ray_flags);
  virtual int ray_intersect(vector ray_from,
			    vector ray_dir, rayintersect *int_storage);
  virtual intersect_result point_intersect(vector apoint);
  virtual intersect_result voxel_intersect(box v);

  // Internal routines used for intersection.
  int intersect_with_surface(double p_x, double p_y, double p_z,
			     double d_x, double d_y, double d_z,
			     double t_entry, double t_exit,
			     intersect_result start_inout,
			     IntersectionList *the_list,
			     primnode *node, vector ray_from,
			     vector ray_dir, uint ray_flags,
			     signature ray_signature);
  int get_intersection(int x_index, int y_index,
		       double cell_entry_t, double cell_exit_t,
		       intersect_result cell_entry_inout,
		       intersect_result cell_exit_inout,
		       double from_x, double from_y, double from_z,
		       double dir_x, double dir_y, double dir_z, 
		       IntersectionList *the_list,
		       primnode *node, vector ray_from,
		       vector ray_dir, uint ray_flags,
		       signature ray_signature);

  intersect_result point_inout(double px, double py, double height);
  intersect_result x_face_inout(int cell_x, double py, double height);
  intersect_result y_face_inout(double px, int cell_y, double height);

  // Note: these routines are Heightfield internal ONLY.
  double interpolate_height(double px, double py);
  double x_face_val(int cell_x, double py);
  double y_face_val(double px, int cell_y);
  void find_cell_normal(double px, double py, vector norm);
  void get_vertex_normal(int vx, int vy, vector norm);

  // Routine that users can use to get the height of a point on the
  // the heightfield's surface.
  double get_height(double x, double y);

  // Routine that users can use to get the normal at a point on the
  // the heightfield's surface.
  void get_normal(double x, double y, vector norm);

  // Return the name of the primitive.
  virtual char *prim_name(void) {
    return "Heightfield";
  }

  // This primitive may take a filename when created.
  Heightfield(int smooth_normals, int should_stop, int do_jitter,
		 char *filename) {
    GenericPixmap *fileImage;
    height_val *val_ptr;
    unsigned char *gray_ptr;
    int i, j;

    // Read in the pixmap file.
    fileImage = new GenericPixmap(filename);
    x_res = fileImage->width;
    y_res = fileImage->height;

    height_data_allocated_internally = TRUE;
    height_data = (height_val *) calloc(x_res * y_res, sizeof(height_val));
    if (height_data == NULL)
      fatal_error("Heightfield::constructor", "can't allocate field memory",
		  "Allocation of memory to store the heightfield values in\n"
		  "failed. The heightfield may be huge, or else you are\n"
		  "critically low on memory.");

    if (fileImage->type != GRAYSCALE)
      fatal_error("Heightfield::constructor", "Image must be grayscale",
		  "The pixmap that the heightfield is constructed from must\n"
		  "be a grayscale PNM file (e.g. height1.pgm). This is a\n"
		  "different kind of PNM and can't be read.");

    if (do_jitter)
      srand48(HEIGHTFIELD_RANDOM_SEED);

    // Copy data from the pixmap into the heightfield. Note that we must
    // flip the pixmap in Y because the X, Y coordinate system on an image
    // is different to that in X, Y in primitivespace.
    gray_ptr = fileImage->storage;
    for (j = 0; j < y_res; j++) {
      val_ptr = &height_data[x_res * ((y_res - 1) - j)];
      for (i = 0; i < x_res; i++) {
	if (do_jitter) {
	  *val_ptr++ = (height_val) (((*gray_ptr++ + drand48())
				    * ((HF_MAXVAL - 2) / 256.0)) + 1);
	} else {
	  *val_ptr++ = (height_val) (((*gray_ptr++ + 0.5      )
				    * ((HF_MAXVAL - 2) / 256.0)) + 1);
	}
      }
    }

    stop_at_first_intersection = should_stop;
    interpolate_normals = smooth_normals;

    // Work out increments for stepping in X and Y.
    x_step = 1;
    y_step = x_res;
    
    is_cluster = 0;
    mark_for_deletion();
  }

  // This primitive takes a size and array of height_val when created.
  Heightfield(int smooth_normals, int should_stop,
		 int x_size, int y_size,
		 height_val *field_data) {

    interpolate_normals = smooth_normals;
    stop_at_first_intersection = should_stop;
    x_res = x_size;
    y_res = y_size;

    height_data_allocated_internally = FALSE;
    height_data = field_data;

    // Work out increments for stepping in X and Y.
    x_step = 1;
    y_step = x_res;

    is_cluster = 0;
    mark_for_deletion();
  }

  virtual ~Heightfield(void) {
    if (height_data_allocated_internally)
      free(height_data);
  }
}; 


