/* imagemap.c : Functions that define the ImageTexture textures.
** 
** These functions define textures that apply images to objects according
** to the desired parameters (projection type, placement, etc.)
**
** Chris Butcher, 20 December 1996
*/

#include "mirage.h"

#ifdef USE_NAMESPACES
namespace Mirage {
#endif

void ImageTexture::apply(vector ray_from, vector ray_dir, uint ray_flags,
			 Material *properties, vector hit_point,
			 vector hit_normal, RayTracer *engine,
			 transform tex_transform)
{
  vector prim_hitpoint;
  double image_x, image_y;
  double pixel_x, pixel_y;
  color image_col;
  double image_grey;
  SpecTransComponent *st_component;
  DiffuseComponent *diffuse;

  Matrix::apply(tex_transform[AFT], hit_point, prim_hitpoint);

  image_x = (prim_hitpoint[texture_x] - texture_x_min)
                     / (texture_x_max - texture_x_min);
  image_y = (prim_hitpoint[texture_y] - texture_y_min)
                     / (texture_y_max - texture_y_min);

  if (my_tile == SINGLE) {
    // If the hitpoint is outside the bounds of the image, no effect.
    if (image_x < 0 || image_x >= 1 || image_y < 0 || image_y >= 1)
      return;
  } else if (my_tile == TILED) {
    // If the hitpoint is outside the bounds of the image, tile it so
    // that it is inside.
    if (image_x >= 1)
      image_x = image_x - ((int) image_x);
    if (image_x < 0)
      image_x = 1.0 - fmod(-image_x, 1.0);
    if (image_y >= 1)
      image_y = image_y - ((int) image_y);
    if (image_y < 0)
      image_y = 1.0 - fmod(-image_y, 1.0);
  }

  pixel_x = image_x * (image->width - 1);
  pixel_y = image_y * (image->height - 1);

  // Zero the specular and transparent properties if we have to.
  if ((zero_specular) || (zero_transparent)) {
    st_component = get_spectrans_component(properties);
    if (st_component != NULL) {
      if (zero_specular)
	Color::set(st_component->specular_color, zero_color);
      if (zero_transparent)
	Color::set(st_component->transmissive_color, zero_color);
    }
  }

  diffuse = get_diffuse_component(properties);
  if (diffuse == NULL)
    fatal_error("ImageTexture::apply", "No diffuse component!",
		"Can't apply a texture to an object\n"
		"without a diffuse component to its\n"
		"material. Add one.");

  if (bilinear_filtered && image->type != B_W) {
    // If we are set to bilinear filter and we don't have a
    // bitmap image (black and white only),
    if (image->type == COLOR) {
      ColorPixmap *color_image = (ColorPixmap *) image;

      // Get the colour (bilinearly interpolated).
      color_image->get_bilinear(pixel_x, pixel_y, image_col);

      if (my_method == TEXMETHOD_OPAQUE) {
	// Apply the image's color opaquely.
	Color::set(diffuse->diffuse_color, image_col);
	return;
      } else if (my_method == TEXMETHOD_TRANSPARENT) {
	// Apply the image's colour as a multiplier for the diffuse color.
	Color::mult(image_col, diffuse->diffuse_color, diffuse->diffuse_color);
	return;
      }

    } else if (image->type == GRAYSCALE) {
      GrayscalePixmap *gray_image = (GrayscalePixmap *) image;

      // Get the gray value (bilinearly interpolated).
      image_grey = gray_image->get_bilinear(pixel_x, pixel_y) / 255.0;

      if (my_method == TEXMETHOD_OPAQUE) {
	// Apply the image's grayvalue opaquely.
	Color::assign(diffuse->diffuse_color,
		      image_grey, image_grey, image_grey);
	return;
      } else if (my_method == TEXMETHOD_TRANSPARENT) {
	// Apply the image's greyvalue as a multiplier for the diffuse color.
	Color::mult(image_grey,
		    diffuse->diffuse_color, diffuse->diffuse_color);
	return;
      }
    }
  } else {
    // No bilinear filtering.
    if (image->type == COLOR) {
      ColorPixmap *color_image = (ColorPixmap *) image;

      // Get the colour of the pixel.
      color_image->get((int) pixel_x, (int) pixel_y, image_col);

      if (my_method == TEXMETHOD_OPAQUE) {
	// Apply the image's color opaquely.
	Color::set(diffuse->diffuse_color, image_col);
	return;
      } else if (my_method == TEXMETHOD_TRANSPARENT) {
	// Apply the image's colour as a multiplier for the diffuse color.
	Color::mult(image_col, diffuse->diffuse_color, diffuse->diffuse_color);
	return;
      }

    } else if (image->type == GRAYSCALE) {
      GrayscalePixmap *gray_image = (GrayscalePixmap *) image;

      // Get the gray value of the pixel and scale to 0 - 1.
      image_grey = gray_image->get((int) pixel_x, (int) pixel_y) / 255.0;

      if (my_method == TEXMETHOD_OPAQUE) {
	// Apply the image's grayvalue opaquely.
	Color::assign(diffuse->diffuse_color,
		      image_grey, image_grey, image_grey);
	return;
      } else if (my_method == TEXMETHOD_TRANSPARENT) {
	// Apply the image's greyvalue as a multiplier for the diffuse color.
	Color::mult(image_grey,
		    diffuse->diffuse_color, diffuse->diffuse_color);
	return;
      }

    } else if (image->type == B_W) {
      bwPixmap *bw_image = (bwPixmap *) image;

      // Get the value of the pixel (either 1 or 0).
      image_grey = bw_image->get((int) pixel_x, (int) pixel_y);

      if (my_method == TEXMETHOD_OPAQUE) {
	// Apply the image's grayvalue opaquely.
	Color::assign(diffuse->diffuse_color,
		      image_grey, image_grey, image_grey);
	return;
      } else if (my_method == TEXMETHOD_TRANSPARENT) {
	// Apply the image's greyvalue as a multiplier for the diffuse color.
	Color::mult(image_grey,
		    diffuse->diffuse_color, diffuse->diffuse_color);
	return;
      }
    }
  }
}

void ImageTexture::print_tex(void)
{
  printf("ImageTexture\n");
  printf("image x along %s, %f-%f\n", axis_str[texture_x],
	 texture_x_min, texture_x_max);
  printf("image y along %s, %f-%f\n", axis_str[texture_y],
	 texture_y_min, texture_y_max);
  if (my_tile == SINGLE)
    printf("image not tiled\n");
  else
    printf("image is tiled\n");
  if (my_method == TEXMETHOD_OPAQUE)
    printf("image applied opaquely\n");
  else
    printf("image applied transparently\n");
  if (bilinear_filtered)
    printf("image is bilinearly filtered\n");

  if (zero_specular)
    printf("setting zero specular values\n");
  if (zero_transparent)
    printf("setting zero transparent values\n");
}

void CylindricalImageTexture::apply(vector ray_from, vector ray_dir,
				    uint ray_flags, Material *properties,
				    vector hit_point, vector hit_normal,
				    RayTracer *engine, transform tex_transform)
{
  vector prim_hitpoint;
  axis cyl_x, cyl_y;
  double theta_dist, theta_val, axis_val;
  double pixel_x, pixel_y;
  color image_col;
  double image_grey;
  SpecTransComponent *st_component;
  DiffuseComponent *diffuse;

  Matrix::apply(tex_transform[AFT], hit_point, prim_hitpoint);

  cyl_x = (axis) ((texture_axis + 1) % 3);
  cyl_y = (axis) ((texture_axis + 2) % 3);

  // Calculate the image coordinates along the axis of the cylinder.
  axis_val = (prim_hitpoint[texture_axis] - texture_axis_min)
    / (texture_axis_max - texture_axis_min);

  // Calculate theta, the angle of the hitpoint around the cylinder.
  if (fabs(prim_hitpoint[cyl_y]) < TINY &&
      fabs(prim_hitpoint[cyl_x]) < TINY)
    theta_val = 0.0;
  else {
    theta_val = RAD2DEG(atan2(prim_hitpoint[cyl_y], prim_hitpoint[cyl_x]));
    if (theta_val < 0)
      theta_val += 360;
  }

  // We work out the difference between the current point and the image's
  // zero point (in terms of the angle around the cylinder). If this is not
  // the same sign as the distance then we add / subtract 360 to make it the
  // same sign.
  theta_dist = theta_val - texture_theta;

  if (theta_span < 0) {
    if (theta_dist > 0)
      theta_dist -= 360;
  } else {
    if (theta_dist < 0)
      theta_dist += 360;
  }
  
  // We now have the distance of the current point from the zero point in
  // the correct direction around the cylinder. Dividing this by theta_span
  // gives us the fraction around the image.
  if (theta_span == 0)
    fatal_error("CylindricalImageTexture::apply", "zero theta_span",
                "The image cannot occupy zero degrees around an object!");
  theta_val = theta_dist / theta_span;
  
  // Check the hitpoint against the bounds of the image.
  if (axis_tile == SINGLE) {
    // If the hitpoint is outside the bounds of the image, no effect.
    if (axis_val < 0 || axis_val >= 1)
      return;
  } else if (axis_tile == TILED) {
    // If the hitpoint is outside the bounds of the image, tile it so
    // that it is inside.
    if (axis_val >= 1)
      axis_val = axis_val - ((int) axis_val);
    if (axis_val < 0)
      axis_val = 1.0 - fmod(-axis_val, 1.0);
  }
  if (theta_tile == SINGLE) {
    // If the hitpoint is outside the bounds of the image, no effect.
    if (theta_val < 0 || theta_val >= 1)
      return;
  } else if (theta_tile == TILED) {
    // If the hitpoint is outside the bounds of the image, tile it so
    // that it is inside.
    if (theta_val >= 1)
      theta_val = theta_val - ((int) theta_val);
    if (theta_val < 0)
      theta_val = 1.0 - fmod(-theta_val, 1.0);
  }

  if (image_main_axis == X) {
    pixel_x = axis_val  * (image->width - 1);
    pixel_y = theta_val * (image->height - 1);
  } else {
    pixel_x = theta_val * (image->width - 1);
    pixel_y = axis_val  * (image->height - 1);
  }

  // Zero the specular and transparent properties if we have to.
  if ((zero_specular) || (zero_transparent)) {
    st_component = get_spectrans_component(properties);
    if (st_component != NULL) {
      if (zero_specular)
	Color::set(st_component->specular_color, zero_color);
      if (zero_transparent)
	Color::set(st_component->transmissive_color, zero_color);
    }
  }

  diffuse = get_diffuse_component(properties);
  if (diffuse == NULL)
    fatal_error("ImageTexture::apply", "No diffuse component!",
		"Can't apply a texture to an object\n"
		"without a diffuse component to its\n"
		"material. Add one.");

  if (bilinear_filtered && image->type != B_W) {
    // If we are set to bilinear filter and we don't have a
    // bitmap image (black and white only),
    if (image->type == COLOR) {
      ColorPixmap *color_image = (ColorPixmap *) image;

      // Get the colour (bilinearly interpolated).
      color_image->get_bilinear(pixel_x, pixel_y, image_col);

      if (my_method == TEXMETHOD_OPAQUE) {
	// Apply the image's color opaquely.
	Color::set(diffuse->diffuse_color, image_col);
	return;
      } else if (my_method == TEXMETHOD_TRANSPARENT) {
	// Apply the image's colour as a multiplier for the diffuse color.
	Color::mult(image_col, diffuse->diffuse_color, diffuse->diffuse_color);
	return;
      }

    } else if (image->type == GRAYSCALE) {
      GrayscalePixmap *gray_image = (GrayscalePixmap *) image;

      // Get the gray value (bilinearly interpolated).
      image_grey = gray_image->get_bilinear(pixel_x, pixel_y) / 255.0;

      if (my_method == TEXMETHOD_OPAQUE) {
	// Apply the image's grayvalue opaquely.
	Color::assign(diffuse->diffuse_color,
		      image_grey, image_grey, image_grey);
	return;
      } else if (my_method == TEXMETHOD_TRANSPARENT) {
	// Apply the image's greyvalue as a multiplier for the diffuse color.
	Color::mult(image_grey,
		    diffuse->diffuse_color, diffuse->diffuse_color);
	return;
      }
    }
  } else {
    // No bilinear filtering.
    if (image->type == COLOR) {
      ColorPixmap *color_image = (ColorPixmap *) image;

      // Get the colour of the pixel.
      color_image->get((int) pixel_x, (int) pixel_y, image_col);

      if (my_method == TEXMETHOD_OPAQUE) {
	// Apply the image's color opaquely.
	Color::set(diffuse->diffuse_color, image_col);
	return;
      } else if (my_method == TEXMETHOD_TRANSPARENT) {
	// Apply the image's colour as a multiplier for the diffuse color.
	Color::mult(image_col, diffuse->diffuse_color, diffuse->diffuse_color);
	return;
      }

    } else if (image->type == GRAYSCALE) {
      GrayscalePixmap *gray_image = (GrayscalePixmap *) image;

      // Get the gray value of the pixel and scale to 0 - 1.
      image_grey = gray_image->get((int) pixel_x, (int) pixel_y) / 255.0;

      if (my_method == TEXMETHOD_OPAQUE) {
	// Apply the image's grayvalue opaquely.
	Color::assign(diffuse->diffuse_color,
		      image_grey, image_grey, image_grey);
	return;
      } else if (my_method == TEXMETHOD_TRANSPARENT) {
	// Apply the image's greyvalue as a multiplier for the diffuse color.
	Color::mult(image_grey,
		    diffuse->diffuse_color, diffuse->diffuse_color);
	return;
      }

    } else if (image->type == B_W) {
      bwPixmap *bw_image = (bwPixmap *) image;

      // Get the value of the pixel (either 1 or 0).
      image_grey = bw_image->get((int) pixel_x, (int) pixel_y);

      if (my_method == TEXMETHOD_OPAQUE) {
	// Apply the image's grayvalue opaquely.
	Color::assign(diffuse->diffuse_color,
		      image_grey, image_grey, image_grey);
	return;
      } else if (my_method == TEXMETHOD_TRANSPARENT) {
	// Apply the image's greyvalue as a multiplier for the diffuse color.
	Color::mult(image_grey,
		    diffuse->diffuse_color, diffuse->diffuse_color);
	return;
      }
    }
  }
}

void CylindricalImageTexture::print_tex(void)
{
  printf("ImageTexture\n");
  printf("projection axis along %s, %f-%f\n", axis_str[texture_axis],
	 texture_axis_min, texture_axis_max);
  printf("theta start %f span %f\n", texture_theta, theta_span);
  if (axis_tile == SINGLE)
    printf("image not tiled along axis\n");
  else
    printf("image is tiled along axis\n");
  if (theta_tile == SINGLE)
    printf("image not tiled along theta\n");
  else
    printf("image is tiled along theta\n");
  if (my_method == TEXMETHOD_OPAQUE)
    printf("image applied opaquely\n");
  else
    printf("image applied transparently\n");
  printf("projection axis is image %s coordinate axis\n",
	 axis_str[image_main_axis]);
  if (bilinear_filtered)
    printf("image is bilinearly filtered\n");

  if (zero_specular)
    printf("setting zero specular values\n");
  if (zero_transparent)
    printf("setting zero transparent values\n");
}

#if 0

// NB the PatchImageTexture functions ***ONLY WORKS*** on a bilinear patch.
// **DO NOT** use it on a non-patch object!
void PatchImageTexture::apply(vector ray_from, vector ray_dir, uint ray_flags,
			      Material *properties, vector hit_point,
			      vector hit_normal, RayTracer *engine,
			      transform tex_transform)
{
  vector prim_hitpoint;
  double image_x, image_y;
  double pixel_x, pixel_y;
  color image_col, material_col;
  double image_grey;

  Matrix::apply(tex_transform[AFT], hit_point, prim_hitpoint);

  // This is a bilinear patch. We can find the s1 and s2 coordinates of the
  // point that was hit by using the formula
  //
  // s1 = hit_x + (1 - patch_x) * hit_z
  // s2 = hit_y + (1 - patch_y) * hit_z

  image_x = prim_hitpoint[X] + (1 - patch_x) * prim_hitpoint[Z];
  image_y = prim_hitpoint[Y] + (1 - patch_y) * prim_hitpoint[Z];

  // If the hitpoint is outside the bounds of the patch, no effect.
  if (image_x < 0 || image_x >= 1 || image_y < 0 || image_y >= 1)
    return;

  pixel_x = image_x * image->width;
  pixel_y = image_y * image->height;

  if (bilinear_filtered && image->type != B_W) {
    // If we are set to bilinear filter and we don't have a
    // bitmap image (black and white only),
    if (image->type == COLOR) {
      ColorPixmap *color_image = (ColorPixmap *) image;

      // Get the colour (bilinearly interpolated).
      color_image->get_bilinear(pixel_x, pixel_y, image_col);

      if (my_method == TEXMETHOD_OPAQUE) {
	// Apply the image's color opaquely.
#ifndef FIXME
	properties->set_diffuse(image_col);
#endif
	return;
      } else if (my_method == TEXMETHOD_TRANSPARENT) {
	// Apply the image's colour as a multiplier for the diffuse color.
	properties->get_diffuse(material_col);
	Color::mult(image_col, material_col, material_col);
	properties->set_diffuse(material_col);
	return;
      }

    } else if (image->type == GRAYSCALE) {
      GrayscalePixmap *gray_image = (GrayscalePixmap *) image;

      // Get the gray value (bilinearly interpolated).
      image_grey = gray_image->get_bilinear(pixel_x, pixel_y) / 255.0;

      if (my_method == TEXMETHOD_OPAQUE) {
	// Apply the image's grayvalue opaquely.
	properties->set_diffuse(image_grey, image_grey, image_grey);
	return;
      } else if (my_method == TEXMETHOD_TRANSPARENT) {
	// Apply the image's greyvalue as a multiplier for the diffuse color.
	properties->get_diffuse(material_col);
	Color::mult(image_grey, material_col, material_col);
	properties->set_diffuse(material_col);
	return;
      }
    }
  } else {
    // No bilinear filtering.
    if (image->type == COLOR) {
      ColorPixmap *color_image = (ColorPixmap *) image;

      // Get the colour of the pixel.
      color_image->get((int) pixel_x, (int) pixel_y, image_col);

      if (my_method == TEXMETHOD_OPAQUE) {
	// Apply the image's color opaquely.
	properties->set_diffuse(image_col);
	return;
      } else if (my_method == TEXMETHOD_TRANSPARENT) {
	// Apply the image's colour as a multiplier for the diffuse color.
	properties->get_diffuse(material_col);
	Color::mult(image_col, material_col, material_col);
	properties->set_diffuse(material_col);
	return;
      }

    } else if (image->type == GRAYSCALE) {
      GrayscalePixmap *gray_image = (GrayscalePixmap *) image;

      // Get the gray value of the pixel and scale to 0 - 1.
      image_grey = gray_image->get((int) pixel_x, (int) pixel_y) / 255.0;

      if (my_method == TEXMETHOD_OPAQUE) {
	// Apply the image's grayvalue opaquely.
	properties->set_diffuse(image_grey, image_grey, image_grey);
	return;
      } else if (my_method == TEXMETHOD_TRANSPARENT) {
	// Apply the image's greyvalue as a multiplier for the diffuse color.
	properties->get_diffuse(material_col);
	Color::mult(image_grey, material_col, material_col);
	properties->set_diffuse(material_col);
	return;
      }

    } else if (image->type == B_W) {
      bwPixmap *bw_image = (bwPixmap *) image;

      // Get the value of the pixel (either 1 or 0).
      image_grey = bw_image->get((int) pixel_x, (int) pixel_y);

      if (my_method == TEXMETHOD_OPAQUE) {
	// Apply the image's grayvalue opaquely.
	properties->set_diffuse(image_grey, image_grey, image_grey);
	return;
      } else if (my_method == TEXMETHOD_TRANSPARENT) {
	// Apply the image's greyvalue as a multiplier for the diffuse color.
	properties->get_diffuse(material_col);
	Color::mult(image_grey, material_col, material_col);
	properties->set_diffuse(material_col);
	return;
      }
    }
  }
}

void PatchImageTexture::print_tex(void)
{
  printf("PatchImageTexture\n");
  printf("Parameters of patch are: %f %f\n", patch_x, patch_y);
  if (my_method == TEXMETHOD_OPAQUE)
    printf("image applied opaquely\n");
  else
    printf("image applied transparently\n");
  if (bilinear_filtered)
    printf("image is bilinearly filtered\n");
}

#endif

#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif
