/* imagemap.h : Definition of the ImageTexture class.
** 
** Chris Butcher, 20 November 1996
*/

// Maps a texture defined by a GenericPixmap onto an object along the
// given axes.
class ImageTexture : public Texture {
public:
  // The image to apply.
  GenericPixmap *image;

  // The axis and bounds to map to the x coordinate of the image.
  axis texture_x;
  double texture_x_min, texture_x_max;
  
  // The axis and bounds to map to the y coordinate of the image.
  axis texture_y;
  double texture_y_min, texture_y_max;
  
  // Are the specular / transparent values being set to 0 where the image is?
  int zero_specular, zero_transparent;

  // Do we tile the image?
  texture_tiling my_tile;

  // Is the image's colour applied directly to the material properties
  // or is it multiplied by the existing value of the material properties?
  texture_method my_method;

  // Is the image bilinear filtered?
  int bilinear_filtered;

  // Apply the image as a colour map.
  virtual void apply(vector ray_from, vector ray_dir, uint ray_flags,
		     Material *properties, vector hit_point, vector hit_normal,
		     RayTracer *engine, transform tex_transform);

  // Print out the texture (for debugging).
  virtual void print_tex(void);

  // Constructor (reads image from file).
  ImageTexture(int my_precedence, char *filename, axis x_ax, double x_min,
	       double x_max, axis y_ax, double y_min, double y_max,
	       texture_tiling do_tile, texture_method apply_method,
	       int do_filter, int spec_0, int trans_0) {
    precedence = my_precedence;
    image = new GenericPixmap(filename);
    texture_x = x_ax;
    texture_x_min = x_min;
    texture_x_max = x_max;
    texture_y = y_ax;
    texture_y_min = y_min;
    texture_y_max = y_max;
    my_tile = do_tile;
    my_method = apply_method;
    bilinear_filtered = do_filter;
    zero_specular = spec_0;
    zero_transparent = trans_0;

    mark_for_deletion();
  }

  // Constructor (passed a GenericPixmap).
  ImageTexture(int my_precedence, GenericPixmap *my_image, axis x_ax,
	       double x_min, double x_max, axis y_ax, double y_min,
	       double y_max, texture_tiling do_tile,
	       texture_method apply_method,
	       int do_filter, int spec_0, int trans_0) {
    precedence = my_precedence;
    image = my_image;
    texture_x = x_ax;
    texture_x_min = x_min;
    texture_x_max = x_max;
    texture_y = y_ax;
    texture_y_min = y_min;
    texture_y_max = y_max;
    my_tile = do_tile;
    my_method = apply_method;
    bilinear_filtered = do_filter;
    zero_specular = spec_0;
    zero_transparent = trans_0;

    mark_for_deletion();
  }
  
  virtual ~ImageTexture(void) {
    // Don't deallocate the pixmap as the deletionlist will handle it or
    // else it's a static pixmap.
  }
};

// Exactly the same as ImageTexture except this uses a cylindrical
// projection to generate the image coordinates.
class CylindricalImageTexture : public Texture {
public:
  // The image to apply.
  GenericPixmap *image;
   
  // The axis that forms the center of the cylinder, and bounds for the
  // image on this axis.
  axis texture_axis;
  double texture_axis_min, texture_axis_max;
    
  // The angle at which the image's zero boundary is and the distance
  // that it spans around the cylinder.
  double texture_theta, theta_span;
  
  // Do we tile the image along the axis and around in theta?
  texture_tiling axis_tile;
  texture_tiling theta_tile;

  // Is the image's colour applied directly to the material properties
  // or is it multiplied by the existing value of the material properties?
  texture_method my_method;

  // Which image axis goes along the projection axis?
  axis image_main_axis;

  // Is the image bilinear filtered?
  int bilinear_filtered;

  // Are the specular / transparent values being set to 0 where the image is?
  int zero_specular, zero_transparent;

  // Apply the image as a colour map.
  virtual void apply(vector ray_from, vector ray_dir, uint ray_flags,
		     Material *properties, vector hit_point,
		     vector hit_normal, RayTracer *engine,
		     transform tex_transform);

  // Print out the texture (for debugging).
  virtual void print_tex(void);

  // Constructor (reads image from file).
  CylindricalImageTexture(int my_precedence, char *filename, axis cyl_ax,
			  double ax_min, double ax_max, double theta,
			  double span, texture_tiling ax_tile,
			  texture_tiling th_tile, axis image_main,
			  texture_method apply_method,
			  int do_filter, int spec_0, int trans_0) {
    precedence = my_precedence;
    image = new GenericPixmap(filename);
    texture_axis = cyl_ax;
    texture_axis_min = ax_min;
    texture_axis_max = ax_max;
    texture_theta = theta;
    theta_span = span;
    axis_tile = ax_tile;
    theta_tile = th_tile;
    image_main_axis = image_main;
    my_method = apply_method;
    bilinear_filtered = do_filter;
    zero_specular = spec_0;
    zero_transparent = trans_0;

    mark_for_deletion();
  }

  // Constructor (passed a GenericPixmap).
  CylindricalImageTexture(int my_precedence, GenericPixmap *my_img,
			  axis cyl_ax, double ax_min, double ax_max,
			  double theta, double span,
			  texture_tiling ax_tile, texture_tiling th_tile,
			  axis image_main, texture_method apply_method,
			  int do_filter, int spec_0, int trans_0) {
    precedence = my_precedence;
    image = my_img;
    texture_axis = cyl_ax;
    texture_axis_min = ax_min;
    texture_axis_max = ax_max;
    texture_theta = theta;
    theta_span = span;
    axis_tile = ax_tile;
    theta_tile = th_tile;
    image_main_axis = image_main;
    my_method = apply_method;
    bilinear_filtered = do_filter;
    zero_specular = spec_0;
    zero_transparent = trans_0;

    mark_for_deletion();
  }

  virtual ~CylindricalImageTexture() {
    // Don't deallocate the pixmap as the deletionlist will handle it
    // if we need to.
  }
};

#if 0
// Maps a texture defined by a GenericPixmap onto a bilinear patch.
//
// NB the PatchImageTexture functions ***ONLY WORKS*** on a bilinear patch.
// **DO NOT** use it on a non-patch object!
class PatchImageTexture : public Texture {
public:
  // The image to apply.
  GenericPixmap *image;

  // Is the image's colour applied directly to the material properties
  // or is it multiplied by the existing value of the material properties?
  texture_method my_method;

  // Is the image bilinear filtered?
  int bilinear_filtered;

  // The parameters of the bilinear patch.
  double patch_x, patch_y;

  // Apply the image as a colour map.
  virtual void apply(vector ray_from, vector ray_dir, uint ray_flags,
		     Material *properties, vector hit_point, vector hit_normal,
		     RayTracer *engine, transform tex_transform);

  // Print out the texture (for debugging).
  virtual void print_tex(void);

  // Constructor (reads image from file).
  PatchImageTexture(int my_precedence, char *filename,
		    double x, double y, texture_method apply_method,
		    int do_filter) {
    precedence = my_precedence;
    image = new GenericPixmap(filename);
    my_method = apply_method;
    bilinear_filtered = do_filter;
    patch_x = x;
    patch_y = y;

    mark_for_deletion();
  }

  // Constructor (passed a GenericPixmap).
  PatchImageTexture(int my_precedence, GenericPixmap *my_image,
		    double x, double y, texture_method apply_method,
		    int do_filter) {
    precedence = my_precedence;
    image = my_image;
    my_method = apply_method;
    bilinear_filtered = do_filter;
    patch_x = x;
    patch_y = y;

    mark_for_deletion();
  }
  
  virtual ~PatchImageTexture(void) {
    // Don't deallocate the pixmap as the deletionlist will handle it or
    // else it's a static pixmap.
  }
};
#endif

