/* keyfile.c : Reads in a keyfile
** 
** This file contains the routines to read in a keyfile and to
** access the data that it contained.
**
** Chris Butcher, 27 November 1996
*/

#include "mirage.h"

#ifdef USE_NAMESPACES
namespace Mirage {
#endif

// The maximum size of the key database.
#define KEY_DATABASE_SIZE         100

// The maximum length of a key string.
#define KEY_MAXIMUM_LENGTH        30

// The maximum length of the key data strings.
#define KEY_DATA_MAXIMUM_LENGTH   100

// Global storage for the key database.
typedef struct {
  char key[KEY_MAXIMUM_LENGTH];
  char data[KEY_DATA_MAXIMUM_LENGTH];
} key_entry;

int num_keys;
key_entry key_database[KEY_DATABASE_SIZE];

// Read in a new keyfile. Passed a filename for the keyfile.
void read_keyfile(char *keyfilename)
{
  FILE *keyfile;
  char tempbuf[MIRAGE_STRING_LENGTH], *new_fname;
  int i;

  if (keyfilename == NULL) {
    // We don't want to read any keyfile, the database is empty.
    num_keys = 0;
    return;
  }

  // Open the keyfile, fatal error if we can't.
  new_fname = filename_oscheck(keyfilename);
  keyfile = fopen(new_fname, "r");
  if (keyfile == NULL) {
    debugging_messages = 1;
    sprintf(tempbuf, "can't open keyfile %s", new_fname);
    fatal_error("read_keyfile", tempbuf, 
		"Either couldn't find the keyfile or there was an error in\n"
		"opening it. Every project needs a \'key file\' that\n"
		"contains a set of parameters used by the rendering system.\n"
		"It is a text file which has lines \"key   data\" where\n"
		"\'key\' is the name of the parameter, \'data\' is the\n"
		"value. An empty file will do. The name of the file must be\n"
		"<project>.key where <project> is the name of the animation\n"
		"project.");
  }

  // Read in lines from the key file into the key database.
  num_keys = 0;
  do {
    // Read a line; NULL is returned on EOF or error.
    if (fgets(tempbuf, MIRAGE_STRING_LENGTH, keyfile) == NULL)
      break;

    // Only parse lines that have a key on them and that don't begin
    // with a # symbol.
    if ((sscanf(tempbuf, "%s", key_database[num_keys].key) == 1) &&
	(key_database[num_keys].key[0] != '#')) {

      // Strip trailing newline from line if it exists.
      if (tempbuf[strlen(tempbuf) - 1] == '\n')
	tempbuf[strlen(tempbuf) - 1] = '\0';

      // Scan through tempbuf looking for the first whitespace character.
      for (i = 0; tempbuf[i] != ' ' && tempbuf[i] != '\t' &&
	     tempbuf[i] != '\n' && tempbuf[i] != '\0'; i++)
	;
      
      // Scan through the whitespace looking for the first non-whitespace
      // character or end-of-string.
      for ( ; tempbuf[i] == ' ' || tempbuf[i] == '\t' ||
	      tempbuf[i] == '\n'; i++)
	;

      // If we didn't find any non-whitespace chars before the end of the
      // string then there wasn't a data item on that line.
      if (tempbuf[i] == '\0') {
	debugging_messages = 1;
	sprintf(tempbuf, "can't find data item for key %s", 
		key_database[num_keys].key);
	fatal_error("read_keyfile", tempbuf,
		    "The key file must be a text file with each line\n"
		    "following the format \"key    data\" where \'data\'\n"
		    "may be any string (possibly including spaces).");
      } else {
	// Must have found a non-blank character ... the start of the
	// data item, presumably.

	// Check the length restrictions
	if (strlen(tempbuf) - i > KEY_DATA_MAXIMUM_LENGTH) {
	  debugging_messages = 1;
	  sprintf(tempbuf, "data %s too long (max %d)", &tempbuf[i],
		  KEY_DATA_MAXIMUM_LENGTH);
	  fatal_error("read_keyfile", tempbuf,
		      "Data items in the key file have a maximum\n"
		      "limit on their length. To change this limit\n"
		      "the system must be recompiled.");
	}

	// Copy the key data item from tempbuf at position i into
	// the key database.
	strcpy(key_database[num_keys].data, &tempbuf[i]);
      }

      num_keys++;

      // Check for database overruns.
      if (num_keys >= KEY_DATABASE_SIZE) {
	debugging_messages = 1;
	sprintf(tempbuf, "too many key items (max %d)", KEY_DATABASE_SIZE);
	fatal_error("read_keyfile", tempbuf,
		    "The database that holds key items is a fixed size.\n"
		    "To change this size the system must be recompiled.");
      }
    }
    // Keep reading keys forever. Break out of this loop when fgets()
    // returns NULL (indicating EOF).
  } while (1);

  fclose(keyfile);

  if (!completely_silent)
    printf(">> Key file %s read: %d items.\n", new_fname, num_keys);
    
  free(new_fname);
}

int get_keyfile_data(char *key_item, char *databuf)
{
  int i;

  // Search through key database looking for the right key item.
  for (i = 0; i < num_keys && strcmp(key_item, key_database[i].key) != 0; i++)
    ;

  // Did we reach the end without finding the key? If so, return FALSE.
  if (i == num_keys)
    return FALSE;
  else {
    strcpy(databuf, key_database[i].data);
    return TRUE;
  }
}

int get_keyfile_toggle(char *key_item, int *to_toggle)
{
  int find_success;
  char tempbuf[KEY_DATA_MAXIMUM_LENGTH];

  // Look for the key and return failure if necessary
  find_success = get_keyfile_data(key_item, tempbuf);
  if (!find_success)
    return find_success;

  if (strncmp(tempbuf, "on", 2) == 0) {
    *to_toggle = 1;
    return TRUE;
  } else if (strncmp(tempbuf, "off", 3) == 0) {
    *to_toggle = 0;
    return TRUE;
  } else {
    sprintf(tempbuf, "incorrect toggle value for key %s", key_item);
    fatal_error("get_keyfile_toggle", tempbuf,
		"This key item is meant to be a toggle parameter.\n"
		"Valid values are \'on\' or \'off\'.");
    return FALSE;
  }
}

int get_keyfile_number(char *key_item, int *to_get)
{
  int find_success;
  char tempbuf[KEY_DATA_MAXIMUM_LENGTH];

  // Look for the key and return failure if necessary
  find_success = get_keyfile_data(key_item, tempbuf);
  if (!find_success)
    return find_success;

  if (sscanf(tempbuf, "%d", to_get) == 1)
    return TRUE;
  else {
    sprintf(tempbuf, "corrupt number value for key %s", key_item);
    fatal_error("get_keyfile_number", tempbuf,
		"This key item is meant to be a numeric parameter.");
    return FALSE;
  }
}

int get_keyfile_float(char *key_item, double *to_get)
{
  int find_success;
  char tempbuf[KEY_DATA_MAXIMUM_LENGTH];

  // Look for the key and return failure if necessary
  find_success = get_keyfile_data(key_item, tempbuf);
  if (!find_success)
    return find_success;

  if (sscanf(tempbuf, "%lf", to_get) == 1)
    return TRUE;
  else {
    sprintf(tempbuf, "corrupt float value for key %s", key_item);
    fatal_error("get_keyfile_float", tempbuf,
		"This key item is meant to be a floating-point parameter.");
    return FALSE;
  }
}

int get_keyfile_vector(char *key_item, vector to_get)
{
  int find_success;
  char tempbuf[KEY_DATA_MAXIMUM_LENGTH];

  // Look for the key and return failure if necessary
  find_success = get_keyfile_data(key_item, tempbuf);
  if (!find_success)
    return find_success;

  if (sscanf(tempbuf, "%lf %lf %lf", &to_get[X], &to_get[Y], &to_get[Z]) == 3)
    return TRUE;
  else {
    sprintf(tempbuf, "corrupt vector value for key %s", key_item);
    fatal_error("get_keyfile_vector", tempbuf,
		"This key item is meant to be a vector parameter in the\n"
		"form e.g. \'6.7 -8.9 7.8843\'.");
    return FALSE;
  }
}


#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif
