/* label.c : Functions to manage the global label list.
** 
** This file defines the functions for the LabelList class.
**
** Chris Butcher, 20 December 1996
*/

#include "mirage.h"

#ifdef USE_NAMESPACES
namespace Mirage {
#endif

// The global label list.
LabelList *world_label_list = NULL;

// Constructor.
LabelList::LabelList(int my_size) {
  list_size = my_size;
  next_free = 0;
  if ((label_storage = (label *) calloc(list_size, sizeof(label))) == NULL) {
    char tempbuf[MIRAGE_STRING_LENGTH];

    sprintf(tempbuf, "can't allocate label storage for %d listsize", my_size);
    fatal_error("LabelList constructor", tempbuf,
		"There is not enough memory available to construct a"
		"label list of the requested size. Generally this error"
		"should not occur unless the requested intersection list"
		"size is huge. It probably indicates a memory leak"
		"somewhere.");
  }
  mark_for_deletion();
}

// Reset the list.
void LabelList::init(void) {
  next_free = 0;
}

// Insert a label into the list.
void LabelList::insert(char *newname, object newplace)
{
  if (next_free >= list_size)
    fatal_error("LabelList::insert", "list storage space overflow",
		"The list in which labels are stored has a limited size."
		"It's not big enough for the object that you're trying"
		"to create. Either increase the LABEL_LIST_SIZE in"
		"constants.h and recompile, or reduce the number of"
		"labels in the scene.");

  // Mark this object as labeled.
  newplace->flags |= FLAGS_LABELED;

  // Insert into the list.
  strcpy(label_storage[next_free].label_name, newname);
  label_storage[next_free].label_place = newplace;
  label_storage[next_free].label_position_set = FALSE;

  next_free++;
}

// Find a label in the list.
int LabelList::find(char *findname, object *findplace, transform findtrans)
{
  int i;

  // Traverse the list looking for the same name as we were given.
  for (i = 0; i < next_free; i++) {
    if (strcmp(label_storage[i].label_name, findname) == 0) {
      // Found it. Has its position been resolved yet?
      if (label_storage[i].label_position_set == FALSE) {
	char tempbuf[MIRAGE_STRING_LENGTH];

	sprintf(tempbuf, "Tried to retrieve label %s whose position wasn't"
		"resolved!", findname);
	fatal_error("LabelList::find", tempbuf,
		    "All labels, once placed in the label list, must have"
		    "their absolute position found during a pre-processing"
		    "pass of the object graph. Either this label wasn't"
		    "in the object graph when this pass was made or you"
		    "are calling the find() procedure before this pass has"
		    "been made. The find_labels() pass is made when you"
		    "call create_animation()... LabelList::find() is only"
		    "intended to be used after this, e.g. in affect"
		    "functions which are called when the RCSG graph is"
		    "created.");
      }
      *findplace = label_storage[i].label_place;
      Transform::set(findtrans, label_storage[i].label_transform);
      return TRUE;
    }
  }

  // We have reached the end of the list without finding the
  // desired label, return FALSE.
  return FALSE;
}

void LabelList::set_transform(object the_obj, transform obj_position)
{
  int i;

  // Traverse the list looking for the same object as we were given.
  for (i = 0; i < next_free; i++) {
    if (label_storage[i].label_place == the_obj) {
      Transform::set(label_storage[i].label_transform, obj_position);
      label_storage[i].label_position_set = TRUE;
      return;
    }
  }

  // We didn't find the object ... it wasn't placed in the label list
  // and yet for this procedure to be called, its FLAGS_LABELED
  // flag must have been set. Scream and die.
  fatal_error("LabelList::set_transform",
	      "tried to set transform of nonexistent label?!",
	      "This error should not happen.");
}

void recursive_find_labels(object theDag, transform existing_transform)
{
  transform new_transform;

  // Process any objects linked to this one precisely the same as this one
  // will be processed (same transform passed to them).
  if (theDag->linkptr != NULL)
    recursive_find_labels(theDag->linkptr, existing_transform);

  // Work out the new transform that applies to this object.
  Transform::combine(theDag->position, existing_transform, new_transform);

  // Check to see if this object is labeled and so needs its position set.
  if (theDag->flags & FLAGS_LABELED) {
    world_label_list->set_transform(theDag, new_transform);
  }

  if (theDag->flags & FLAGS_DAG) {
    // If this dagnode has a vertical pointer to more objects, recurse
    // them with the new transform that applies to this object.
    recursive_find_labels((object) theDag->details, new_transform);

  } else if (theDag->flags & FLAGS_PRIM) {
    // Check to see if this is a cluster that needs recursing.
    if (((Primitive *) theDag->details)->is_cluster) {

      // Recursively search the cluster for labels.
      ((ClusterPrimitive *) theDag->details)->find_labels(new_transform);
    }
  }
}


#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif
