/* light.c : Contains code for the subclasses of Light and also for
** the lightList.
** 
** This has the lightList class and code to add lights to this list.
** All of the different light types are also in here.
**
** Chris Butcher, 6 December 1996
*/

#include "mirage.h"

#ifdef USE_NAMESPACES
namespace Mirage {
#endif

void lightList::add_light(Light *new_light)
{
  lightStorage[num_lights++] = new_light;
}

// The ambient light casts an equal amount of light on everything.
void AmbientLight::lightOn(color light_amount, vector world_point,
			   vector from_dir, RayTracer *the_engine,
			   primnode *world_scene,
			   uint current_ray_flags,
			   IntersectInfo *int_info)
{
  // Check to see that the light should be applied to this ray.
  if ((current_ray_flags & RAY_REFLECTED && 
       (light_flags & FLAGS_NOREFLECT)) ||
      (current_ray_flags & RAY_TRANSMITTED && 
       (light_flags & FLAGS_NOTRANSMIT)) ||
      (current_ray_flags & RAY_THROUGH_GATEWAY && 
       (light_flags & FLAGS_NOT_THROUGH_GATEWAY)) ||
      ((current_ray_flags & RAY_PRIMARYMASK) == 0 && 
       (light_flags & FLAGS_NOPRIMARY)) ||
      ((current_ray_flags & RAY_TRANSMITTED) == 0 && 
       (light_flags & FLAGS_ONLYTRANSMIT)) ||
      ((current_ray_flags & RAY_THROUGH_GATEWAY) == 0 && 
       (light_flags & FLAGS_ONLY_THROUGH_GATEWAY))) {
    // This light doesn't apply to this ray.
    Vector::assign(from_dir, 0, 0, 0);
    Color::set(light_amount, zero_color);
    return;
  }

  // The direction that the light comes from is zero. This is a special
  // case that means 'directionless'.
  Vector::assign(from_dir, 0.0, 0.0, 0.0);
  Vector::set(light_amount, light_color);
}  

void PointLight::lightOn(color light_amount, vector world_point,
			 vector from_dir, RayTracer *engine,
			 primnode *world_scene, uint current_ray_flags,
			 IntersectInfo *int_info)
{
  vector light_world_position, dummy_hit_normal, dummy_hit_point;
  color shadow_amount, dummy_contribution;
  primnode *dummy_hitobject;
  double dummy_hit_t, light_dist;

  // Check to see that the light should be applied to this ray.
  if ((current_ray_flags & RAY_REFLECTED && 
       (light_flags & FLAGS_NOREFLECT)) ||
      (current_ray_flags & RAY_TRANSMITTED && 
       (light_flags & FLAGS_NOTRANSMIT)) ||
      (current_ray_flags & RAY_THROUGH_GATEWAY && 
       (light_flags & FLAGS_NOT_THROUGH_GATEWAY)) ||
      ((current_ray_flags & RAY_PRIMARYMASK) == 0 && 
       (light_flags & FLAGS_NOPRIMARY)) ||
      ((current_ray_flags & RAY_TRANSMITTED) == 0 && 
       (light_flags & FLAGS_ONLYTRANSMIT)) ||
      ((current_ray_flags & RAY_THROUGH_GATEWAY) == 0 && 
       (light_flags & FLAGS_ONLY_THROUGH_GATEWAY))) {
    // This light doesn't apply to this ray.
    Vector::set(from_dir, zero_vector);
    Color::set(light_amount, zero_color);
    return;
  }

  // Work out where the light is in worldspace.
  light_world_position[X] = world_transform[FORE][0][3];
  light_world_position[Y] = world_transform[FORE][1][3];
  light_world_position[Z] = world_transform[FORE][2][3];

  // Work out the direction from the world-point to the light.
  // This is used in the material color calculations (Lambertian)
  Vector::sub(light_world_position, world_point, from_dir);

  // We normalise this direction to avoid problems with TVAL_MIN_CUTOFF
  // translating to a significant distance in worldspace if the light is
  // a very long way away.
  light_dist = Vector::magnitude(from_dir);
  Vector::mult(1.0 / CHECK0(light_dist,
			    "PointLight::lightOn ... light from direction"),
	       from_dir, from_dir);

  // Do we cast shadows?
  if (light_flags & FLAGS_NOSHADOW) {
    // This light doesn't check for shadows.
    Color::set(shadow_amount, full_color);
  } else {
    // Work out the amount of the light that gets from the light
    // to the world_point (by casting a shadow ray). Note that the
    // shadow rays won't cast secondary rays so the maximum ray depth
    // is set to 0.

    engine->traceRay(world_scene, world_point, from_dir,
		     current_ray_flags | RAY_SHADOW,
		     new_ray_signature(),
		     int_info->from.sfxtype, int_info->from.sfx,
		     light_dist, dummy_contribution,
		     &dummy_hitobject, &dummy_hit_t, dummy_hit_normal,
		     dummy_hit_point, shadow_amount, 0);
  }

  // If totally in shadow, return no light.
  if (Color::is_zero(shadow_amount))
    Color::set(light_amount, zero_color);
  else
    // Work out the light that reaches the point.
    Color::mult(light_color, shadow_amount, light_amount);
}

void PointLight::lightSource(color light_amount, vector world_point)
{
  // This light casts its full light on its source objects.
  Color::set(light_amount, light_color);
}

void DirectionalLight::lightOn(color light_amount, vector world_point,
			       vector from_dir, RayTracer *engine,
			       primnode *world_scene, uint current_ray_flags,
			       IntersectInfo *int_info)
{
  vector dummy_hit_normal, dummy_hit_point;
  color shadow_amount, dummy_contribution;
  primnode *dummy_hitobject;
  double dummy_hit_t;

  // Check to see that the light should be applied to this ray.
  if ((current_ray_flags & RAY_REFLECTED && 
       (light_flags & FLAGS_NOREFLECT)) ||
      (current_ray_flags & RAY_TRANSMITTED && 
       (light_flags & FLAGS_NOTRANSMIT)) ||
      (current_ray_flags & RAY_THROUGH_GATEWAY && 
       (light_flags & FLAGS_NOT_THROUGH_GATEWAY)) ||
      ((current_ray_flags & RAY_PRIMARYMASK) == 0 && 
       (light_flags & FLAGS_NOPRIMARY)) ||
      ((current_ray_flags & RAY_TRANSMITTED) == 0 && 
       (light_flags & FLAGS_ONLYTRANSMIT)) ||
      ((current_ray_flags & RAY_THROUGH_GATEWAY) == 0 && 
       (light_flags & FLAGS_ONLY_THROUGH_GATEWAY))) {
    // This light doesn't apply to this ray.
    Vector::set(from_dir, zero_vector);
    Color::set(light_amount, zero_color);
    return;
  }

  // Work out the direction from the world-point to the light.
  // This is constant.
  Vector::set(from_dir, light_dir);

  // Do we cast shadows?
  if (light_flags & FLAGS_NOSHADOW) {
    // This light doesn't check for shadows.
    Color::set(shadow_amount, full_color);
  } else {
    // Work out the amount of the light that gets from the light
    // to the world_point (by casting a shadow ray). Note that the
    // shadow rays won't cast secondary rays so the maximum ray depth
    // is set to 0. We cast rays back to INFINITY.
    
    engine->traceRay(world_scene, world_point, from_dir,
		     current_ray_flags | RAY_SHADOW,
		     new_ray_signature(),
		     int_info->from.sfxtype, int_info->from.sfx,
		     INFINITY, dummy_contribution,
		     &dummy_hitobject, &dummy_hit_t, dummy_hit_normal,
		     dummy_hit_point, shadow_amount, 0);
  }

  // If totally in shadow, return no light.
  if (Color::is_zero(shadow_amount))
    Color::set(light_amount, zero_color);

  //  printf("DirectionalLight: light %f %f %f has shadow %f %f %f\n",
  //  	 light_color[R], light_color[G], light_color[B],
  //  	 shadow_amount[R], shadow_amount[G], shadow_amount[B]);

  // Work out the light that reaches the point.
  Color::mult(light_color, shadow_amount, light_amount);
}

void DirectionalLight::lightSource(color light_amount, vector world_point)
{
  // This light casts its full light on its source objects.
  Color::set(light_amount, light_color);
}

// LimitedLight light source: info may be found in light.h.
void LimitedLight::lightOn(color light_amount, vector world_point,
			   vector from_dir, RayTracer *engine,
			   primnode *world_scene, uint current_ray_flags,
			   IntersectInfo *int_info)
{
  vector light_world_position, dummy_hit_normal, dummy_hit_point;
  color shadow_amount, dummy_contribution;
  primnode *dummy_hitobject;
  double dummy_hit_t, light_dist;

  // Check to see that the light should be applied to this ray.
  if ((current_ray_flags & RAY_REFLECTED && 
       (light_flags & FLAGS_NOREFLECT)) ||
      (current_ray_flags & RAY_TRANSMITTED && 
       (light_flags & FLAGS_NOTRANSMIT)) ||
      (current_ray_flags & RAY_THROUGH_GATEWAY && 
       (light_flags & FLAGS_NOT_THROUGH_GATEWAY)) ||
      ((current_ray_flags & RAY_PRIMARYMASK) == 0 && 
       (light_flags & FLAGS_NOPRIMARY)) ||
      ((current_ray_flags & RAY_TRANSMITTED) == 0 && 
       (light_flags & FLAGS_ONLYTRANSMIT)) ||
      ((current_ray_flags & RAY_THROUGH_GATEWAY) == 0 && 
       (light_flags & FLAGS_ONLY_THROUGH_GATEWAY))) 
    {
      // This light doesn't apply to this ray.
      Vector::set(from_dir, zero_vector);
      Color::set(light_amount, zero_color);
      return;
    }
  
  // Work out where the light is in worldspace.
  light_world_position[X] = world_transform[FORE][0][3];
  light_world_position[Y] = world_transform[FORE][1][3];
  light_world_position[Z] = world_transform[FORE][2][3];
  
  // Work out the direction from the world-point to the light.
  // This is used in the material color calculations (Lambertian)
  Vector::sub(light_world_position, world_point, from_dir);
  
  // We normalise this direction to avoid problems with TVAL_MIN_CUTOFF
  // translating to a significant distance in worldspace if the light is
  // a very long way away.
  light_dist = Vector::magnitude(from_dir);

  // Don't do anything if we're outside the light's dropoff limit.
  if (light_dist > extent) {
    Vector::set(from_dir, zero_vector);
    Color::set(light_amount, zero_color);
    return;
  }

  Vector::mult(1.0 / CHECK0(light_dist,
			    "LimitedLight::lightOn ... light from direction"),
	       from_dir, from_dir);

  // Do we cast shadows?
  if (light_flags & FLAGS_NOSHADOW) {
    // This light doesn't check for shadows.
    Color::set(shadow_amount, full_color);
  } else {
    // Work out the amount of the light that gets from the light
    // to the world_point (by casting a shadow ray). Note that the
    // shadow rays won't cast secondary rays so the maximum ray depth
    // is set to 0.

    engine->traceRay(world_scene, world_point, from_dir,
		     current_ray_flags | RAY_SHADOW,
		     new_ray_signature(),
		     int_info->from.sfxtype, int_info->from.sfx,
		     light_dist, dummy_contribution,
		     &dummy_hitobject, &dummy_hit_t, dummy_hit_normal,
		     dummy_hit_point, shadow_amount, 0);
  }

  // If totally in shadow, return no light.
  if (Color::is_zero(shadow_amount))
    Color::set(light_amount, shadow_amount);
  else {
    // Work out the light that reaches the point.  
    double t, dropoff_scaling;
    
    switch(dropoff) {
    case LL_NONE:
      dropoff_scaling = 1.0;
      break;
      
    case LL_LINEAR:
      dropoff_scaling = 1 - light_dist / extent;
      break;
      
    case LL_SQUARE:
      t = 1 - light_dist / extent;
      dropoff_scaling = (t * t);
      break;
      
    case LL_CUBIC:
      t = 1 - light_dist / extent;
      dropoff_scaling = (t * t * t);
      break;
      
    case LL_CUSTOM:
      dropoff_scaling = dropoff_func(light_dist / extent);
      break;

    default:
      fatal_error("LimitedLight::LightOn", "Unknown dropoff_scaling!",
		  "This error should not happen.");
      exit(-1);
    }

    Color::mult(light_color, shadow_amount, light_amount);		
    Color::mult(dropoff_scaling, light_amount, light_amount);
  }
}


#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif
