/* light.h : Header file for Light and lightList classes.
** 
** This file defines the Light and lightList classes and prototypes
** for their functions.
**
** Chris Butcher, 5 December 1996
*/

// A Light object specifies the light falling on a certain object.
// In the future it might also have a voxel in/out test to determine
// if it needs to be checked for the given object.
class Light : public DeletableObject {
public:
  // This light's position.
  transform world_transform;

  // The flags given to this light.
  uint light_flags;

  // Work out the amount of light that this light source casts on a
  // given point.
  virtual void lightOn(color light_amount, vector world_point,
		       vector from_dir, class RayTracer *the_engine,
		       struct PrimitiveNode *world_scene,
		       uint current_ray_flags,
		       struct _IntersectInfo *int_info) = 0;

  // Work out the amount of light that this light source casts on a
  // point, given that it is the source of the light (e.g. how bright
  // does the lightbulb look?)
  virtual void lightSource(color light_amount, vector world_point) = 0;
};

// This is a point light source of the given color.
class PointLight : public Light {
public:
  color light_color;

  // The light casts an equal amount of light on everything. It casts a
  // shadow ray to determine if it can be seen.
  void lightOn(color light_amount, vector world_point,
	       vector from_dir, RayTracer *the_engine,
	       struct PrimitiveNode *world_scene,
	       uint current_ray_flags, struct _IntersectInfo *int_info);
  void lightSource(color light_amount, vector world_point);
  
  PointLight(double r_col, double g_col, double b_col) {
    // Set the light's emitted color.
    Color::assign(light_color, 0, 0, 0, r_col, g_col, b_col, 0);

    mark_for_deletion();
  }

  PointLight(double r_col, double g_col, double b_col,
	     double er_col, double eg_col, double eb_col) {
    Color::assign(light_color, r_col, g_col, b_col, er_col, eg_col, eb_col, 0);

    mark_for_deletion();
  }

  PointLight(double r_col, double g_col, double b_col,
	     double er_col, double eg_col, double eb_col,
	     double alpha) {
    Color::assign(light_color, r_col, g_col, b_col, er_col, eg_col, eb_col, alpha);

    mark_for_deletion();
  }

  virtual ~PointLight(void) {
  }
};

// This is a point light source of the given color.
class DirectionalLight : public Light {
public:
  color light_color;
  vector light_dir;

  // The light casts an equal amount of light on everything. It casts a
  // shadow ray to determine if it can be seen. The light comes from a given
  // direction.
  void lightOn(color light_amount, vector world_point,
	       vector from_dir, RayTracer *the_engine,
	       struct PrimitiveNode *world_scene,
	       uint current_ray_flags, struct _IntersectInfo *int_info);
  void lightSource(color light_amount, vector world_point);
  
  DirectionalLight(double r_col, double g_col, double b_col,
		   double dir_x, double dir_y, double dir_z) {
    Color::assign(light_color, r_col, g_col, b_col);
    Vector::assign(light_dir, dir_x, dir_y, dir_z);

    mark_for_deletion();
  }

  DirectionalLight(double r_col, double g_col, double b_col,
		   double er_col, double eg_col, double eb_col,
		   double dir_x, double dir_y, double dir_z) {
    Color::assign(light_color, r_col, g_col, b_col, er_col, eg_col, eb_col, 0);
    Vector::assign(light_dir, dir_x, dir_y, dir_z);

    mark_for_deletion();
  }

  DirectionalLight(double r_col, double g_col, double b_col,
		   double er_col, double eg_col, double eb_col,
		   double alpha,
		   double dir_x, double dir_y, double dir_z) {
    Color::assign(light_color, r_col, g_col, b_col, er_col, eg_col, eb_col, alpha);
    Vector::assign(light_dir, dir_x, dir_y, dir_z);

    mark_for_deletion();
  }

  virtual ~DirectionalLight(void) {
  }
};

// This is an ambient light source of the given color.
class AmbientLight : public Light {
public:
  color light_color;

  // The light casts an equal amount of light on everything.
  void lightOn(color light_amount, vector world_point,
	       vector from_dir, RayTracer *the_engine,
	       struct PrimitiveNode *world_scene,
	       uint current_ray_flags, struct _IntersectInfo *int_info);

  void lightSource(color light_amount, vector world_point) {
    Vector::set(light_amount, light_color);
  }
  
  AmbientLight(double r_col, double g_col, double b_col) {
    Color::assign(light_color, r_col, g_col, b_col);
    mark_for_deletion();
  }

  virtual ~AmbientLight() {
  }
};

// A lightList stores all lights in the scene (if the global lightList)
// or all lights in a specific voxel of the subdivision.
class lightList : public DeletableObject {
public:
  int num_lights;
  Light *lightStorage[MAX_LIGHTS];

  // Add a light to the list.
  void add_light(Light *the_light);

  // Empty the list.
  void init(void) {
    num_lights = 0;
  }
  
  // Create a new (empty) list.
  lightList(void) {
    mark_for_deletion();
    num_lights = 0;
  }

  // Create a new list that contains a subset of the lights in an
  // existing list (all that intersect the given box).
  // Not done yet.
  lightList(lightList *oldlist, box new_box) {
    mark_for_deletion();
    num_lights = 0;
  }
  
  ~lightList() {
  }
};




/*
    LimitedLight is a limited-area light class for Mirage. These lights
  are point-sourced with a specified limited area of effect. They may be
  specified to have dropoff with distance - either linear, squared, cubic or no
  dropoff effect.
    Users can choose which type to use by calling the constructor with one of
  the dropoff_types enumerated below.  Alternatively, a constructor is
  provided to allow custom dropoff functions, which takes a pointer to a custom
  function.

  Peter Ashford, 28/3/98.

  Formatting changes, integration into Mirage: Chris Butcher, 31/3/98
  */

class LimitedLight : public PointLight
{
public:
  double extent;                  // maximum extent of light
  dropoff_type dropoff;           // kind of dropoff 
  double (*dropoff_func)(double); // pointer to dropoff function for
                                  // custom dropoff functions

  //-----------------------------------------------------------------
  //	constructor using default dropoff_types.
  //	r_col, g_col, b_col = rgb color values.
  //	limit = maximum number of world units that light travels.
  //	dot = type of dropoff to use.
  //-----------------------------------------------------------------
  LimitedLight(double r_col, double g_col, double b_col, double limit, 
	       dropoff_type dot) : PointLight(r_col, g_col, b_col)
  {
    if(dot == LL_CUSTOM)
      fatal_error("LimitedLight::LimitedLight - standard constuctor",
		  "user can't specify LL_CUSTOM dropoff type",
		  "If you want a custom dropoff function, you need to\n"
		  "initialise LimitedLight by passing it your custom\n"
		  "dropoff function.\n"
		  "See light.h for details.");

    extent = limit;
    dropoff = dot;
    dropoff_func = NULL;
  }

  LimitedLight(double r_col, double g_col, double b_col, 
	       double er_col, double eg_col, double eb_col, double limit, 
	       dropoff_type dot) : PointLight(r_col, g_col, b_col,
					      er_col, eg_col, eb_col)
  {
    if(dot == LL_CUSTOM)
      fatal_error("LimitedLight::LimitedLight - standard constuctor",
		  "user can't specify LL_CUSTOM dropoff type",
		  "If you want a custom dropoff function, you need to\n"
		  "initialise LimitedLight by passing it your custom\n"
		  "dropoff function.\n"
		  "See light.h for details.");

    extent = limit;
    dropoff = dot;
    dropoff_func = NULL;
  }

  LimitedLight(double r_col, double g_col, double b_col, 
	       double er_col, double eg_col, double eb_col, double alpha,
	       double limit, dropoff_type dot) : PointLight(r_col, g_col, b_col,
							    er_col, eg_col, eb_col,
							    alpha)
  {
    if(dot == LL_CUSTOM)
      fatal_error("LimitedLight::LimitedLight - standard constuctor",
		  "user can't specify LL_CUSTOM dropoff type",
		  "If you want a custom dropoff function, you need to\n"
		  "initialise LimitedLight by passing it your custom\n"
		  "dropoff function.\n"
		  "See light.h for details.");

    extent = limit;
    dropoff = dot;
    dropoff_func = NULL;
  }
  //-------------------------------------------------------------------------
  //	constructor using custom dropoff function.  
  //	r_col, g_col, b_col = rgb color values.
  //	limit = maximum number of world units that light travels.
  //	dropoff_fn = pointer to user's custom dropoff function. The custom
  //	  dropoff function will be given a value from 0.0 - 1.0 indicating
  //	  how far the hit point is away from the light (relative to it's 
  //	  extent). It must return a value between 0.0 and 1.0 which indicates
  //	  the light's intensity at that distance.
  //-------------------------------------------------------------------------
  LimitedLight(double r_col, double g_col, double b_col, double limit, 
	       double (*dropoff_fn)(double)) : PointLight(r_col, g_col, b_col)
  {
    if(dropoff_fn == NULL)
      fatal_error("LimitedLight::LimitedLight - custom constuctor",
		  "user can't specify NULL custom dropoff function",
		  "You can't initialise LimitedLight using a NULL pointer\n"
                  "for a custom dropoff function. See light.h for details.");
    
    extent = limit;
    dropoff = LL_CUSTOM;
    dropoff_func = dropoff_fn; 
  }

  LimitedLight(double r_col, double g_col, double b_col, 
	       double er_col, double eg_col, double eb_col, double limit, 
	       double (*dropoff_fn)(double)) : PointLight(r_col, g_col, b_col,
							  er_col, eg_col, eb_col)
  {
    if(dropoff_fn == NULL)
      fatal_error("LimitedLight::LimitedLight - custom constuctor",
		  "user can't specify NULL custom dropoff function",
		  "You can't initialise LimitedLight using a NULL pointer\n"
                  "for a custom dropoff function. See light.h for details.");
    
    extent = limit;
    dropoff = LL_CUSTOM;
    dropoff_func = dropoff_fn; 
  }

  LimitedLight(double r_col, double g_col, double b_col, 
	       double er_col, double eg_col, double eb_col, double alpha,
	       double limit, double (*dropoff_fn)(double))
    : PointLight(r_col, g_col, b_col, er_col, eg_col, eb_col, alpha)
  {
    if(dropoff_fn == NULL)
      fatal_error("LimitedLight::LimitedLight - custom constuctor",
		  "user can't specify NULL custom dropoff function",
		  "You can't initialise LimitedLight using a NULL pointer\n"
                  "for a custom dropoff function. See light.h for details.");
    
    extent = limit;
    dropoff = LL_CUSTOM;
    dropoff_func = dropoff_fn; 
  }

  void lightOn(color light_amount, vector world_point,
	       vector from_dir, RayTracer *engine,
	       struct PrimitiveNode *world_scene,
	       uint current_ray_flags,
	       struct _IntersectInfo *int_info);

  virtual ~LimitedLight(void) {}
};
