/* list.c : Functions to manage the global intersection list.
** 
** This file defines the functions for the IntersectionList class.
**
** Chris Butcher, 2 December 1996
*/

#include "mirage.h"

#ifdef USE_NAMESPACES
namespace Mirage {
#endif

#define DEBUG_LIST 0
#define DEBUG_INSERTIONS 0

void IntersectionList::delete_intersections(void)
{
  intersection *ptr, *to_free;

  for (ptr = head; ptr != NULL; ) {
    to_free = ptr;
    ptr = ptr->next;
    free_intersection(to_free);
  }
}

// Set up the list for a new ray.
void IntersectionList::init(void) {
  delete_intersections();

  head = tail = current_intersect = NULL;
  currently_at_head = TRUE;
  currently_at_tail = FALSE;

#if DEBUG_LIST
  printf("L: init\n");
#endif
}

// Reset the next intersection pointer back to the start of the list.
void IntersectionList::restart(void) {
  current_intersect = NULL;
  currently_at_head = TRUE;
  currently_at_tail = FALSE;

#if DEBUG_LIST
  printf("L: restart\n");
#endif
}

// Get the current next-intersection position.
intersection *IntersectionList::get_position(void) {

#if DEBUG_LIST
  printf("L: get pos\n");
#endif

  return current_intersect;
}

// Set the next-intersection position.
void IntersectionList::set_position(intersection *new_pos) {
  intersection *ptr;

  if (new_pos == NULL) {
    current_intersect = new_pos;
    currently_at_head = TRUE;
    currently_at_tail = FALSE;
    return;
  }
    
#if DEBUG_LIST
  printf("L: set pos\n");
#endif

  for (ptr = head; ptr != NULL; ptr = ptr->next)
    if (ptr == new_pos)
      break;

  if (ptr == NULL)
    fatal_error("IntersectionList::set_position()", "Can't find position!",
		"This error should not happen.");

#if DEBUG_LIST
  printf("L: pos unknown\n");
#endif

  current_intersect = ptr;
  currently_at_head = FALSE;
  currently_at_tail = FALSE;
}

// Insert an intersection into the list.
void IntersectionList::insert(uint hit_flags, primnode *node, vector norm,
			      double t_value,
			      int from_empty, int to_empty,
			      Material *mat_from, Material *mat_to,
			      SfxType sfxtype1, SfxType sfxtype2,
			      void *sfx1, void *sfx2,
			      int multitex1, int multitex2,
			      void *tex1, void *tex2,
			      vector ray_from, vector ray_dir,
			      uint ray_flags, primnode *real_primnode,
			      primnode *validate_primnode) {
  intersection *the_int, *next_ptr;
  uint int_flags;

#if DEBUG_LIST
  printf("L: insert tval %f\n", t_value);
  //  printf("Before insertion:\n");
  //  print_list();
#endif

  // Leaving intersections with a gateway object aren't classed as
  // gateway intersections.
  if (to_empty)
    int_flags = hit_flags & (~FLAGS_GATEWAY);
  else
    int_flags = hit_flags;

  // All solid intersections may have a surface_distortion applied to them.
  // If so, then apply the distortion procedure to the hitpoint.
  // NB: this procedure can only modify the t_value and the normal.
  if (node == real_primnode) {
    if (sfxtype2 == DISTORTION) {
      DistortSFX *distort = (DistortSFX *) sfx2;

      distort->surface_distortion(node, distort->distort_transform,
				  norm, &t_value, ray_from, ray_dir,
				  ray_flags);
    }
    if (sfxtype1 == DISTORTION) {
      DistortSFX *distort = (DistortSFX *) sfx1;

      distort->surface_distortion(node, distort->distort_transform,
				  norm, &t_value, ray_from, ray_dir,
				  ray_flags);
    }
  }

#if DEBUG_INSERTIONS
  printf("L: insert t %f\n", t_value);
#endif

  the_int = new_intersection(int_flags, node, norm, t_value, 
			     from_empty, to_empty, mat_from, mat_to,
			     sfxtype1, sfxtype2, sfx1, sfx2,
			     multitex1, multitex2, tex1, tex2,
			     ray_from, ray_dir, ray_flags, real_primnode,
			     validate_primnode);

  // Find the place to insert the intersection at.
  for (next_ptr = head; next_ptr != NULL; next_ptr = next_ptr->next)
    if (next_ptr->t_value > t_value)
      break;

  the_int->next = next_ptr;
  
  // Work out the intersection immediately previous to this one.
  if (next_ptr == NULL)
    the_int->prev = tail;
  else
    the_int->prev = next_ptr->prev;

  // Set the link that points back to this one.
  if (next_ptr == NULL)
    tail = the_int;
  else
    the_int->next->prev = the_int;

  // Set the link that points forwards to this one.
  if (the_int->prev == NULL)
    head = the_int;
  else
    the_int->prev->next = the_int;

#if DEBUG_LIST
  //  printf("After insertion:\n");
  //print_list();
#endif
}

// Retrieve the next intersection in t-value order from the list.
intersection *IntersectionList::get_next(void)
{
#if DEBUG_LIST
  printf("L: get next\n");
#endif

  if (currently_at_head) {
#if DEBUG_LIST
    printf("L: at head\n");
#endif
    current_intersect = head;
    currently_at_head = FALSE;
    if (current_intersect == NULL)
      currently_at_tail = TRUE;
    else
      currently_at_tail = FALSE;
    
#if DEBUG_LIST
    if (head == NULL)
      printf("L: list empty, returning NULL\n");
    else
      printf("L: returning head, tval %f\n", current_intersect->t_value);
#endif

    return current_intersect;
  } else if (currently_at_tail) {
#if DEBUG_LIST
    printf("L: already tail, returning NULL\n");
#endif
    current_intersect = NULL;
    return current_intersect;
  } else {
#if DEBUG_LIST
    printf("L: in middle\n");
#endif
    current_intersect = current_intersect->next;

    if (current_intersect == NULL) {
      currently_at_tail = TRUE;
#if DEBUG_LIST
      printf("L: returning NULL, list now at tail\n");
#endif
    } else {
#if DEBUG_LIST
      printf("L: returning middle w/ tval %f\n", current_intersect->t_value);
#endif
    }
    return current_intersect;
  }
}

// Undo the last retrieve operation from the list.
void IntersectionList::undo_last_retrieve(void)
{
#if DEBUG_LIST
  printf("L: undo\n");
#endif

  if (currently_at_head) {
    // Nothing to undo.
  } else if (currently_at_tail) {
    // Go to the last intersection.
    current_intersect = tail;
    currently_at_tail = FALSE;
    if (current_intersect == NULL)
      currently_at_head = TRUE;
  } else {
    current_intersect = current_intersect->prev;
    if (current_intersect == NULL)
      currently_at_head = TRUE;
  }
}

// Is the list empty?
int IntersectionList::empty(void)
{
#if DEBUG_LIST
  //  printf("L: empty?\n");
#endif

  if (currently_at_tail) {
#if DEBUG_LIST
  printf("L: tail, empty!\n");
#endif
    return TRUE;
  } else if (currently_at_head) {
    if (head == NULL) {
#if DEBUG_LIST
      printf("L: head, empty!\n");
#endif
      return TRUE;
    } else {
#if DEBUG_LIST
      printf("L: head, non-empty!\n");
#endif
      return FALSE;
    }
  } else {
    if (current_intersect->next == NULL) {
#if DEBUG_LIST
      printf("L: middle, last int, empty!\n");
#endif
      return TRUE;
    } else {
#if DEBUG_LIST
      printf("L: middle, more ints, non-empty!\n");
#endif
      return FALSE;
    }
  }
}

// Remove the current intersection from the list.
void IntersectionList::remove_intersection(intersection *the_int)
{
#if DEBUG_LIST
  printf("L: remove\n");
#endif

  if (the_int == current_intersect) {
    if (the_int->prev == NULL) {
      // We are deleting the first intersection, which is the
      // current intersection. Therefore we will currently be
      // at the head of the list.
      current_intersect = NULL;
      currently_at_head = TRUE;
    } else {
      // We are deleting an intersection that has one just before
      // it; set the current intersection to be this previous one.
      current_intersect = current_intersect->prev;
    }
  }

  if (the_int->next == NULL)
    tail = the_int->prev;
  else
    the_int->next->prev = the_int->prev;

  if (the_int->prev == NULL)
    head = the_int->next;
  else
    the_int->prev->next = the_int->next;

#if 0
  if (current_intersect == the_int) {
    current_intersect = the_int->next;
    if (current_intersect == NULL)
      currently_at_tail = TRUE;
  }
#endif

  free_intersection(the_int);
}

// Remove the current intersection from the list.
void IntersectionList::print_list(void)
{
  intersection *ptr;
  
  printf("# printbegin\n");

  if (head == NULL)
    printf("head: NULL\n");
  else
    printf("head: %f\n", head->t_value);

  if (tail == NULL)
    printf("tail: NULL\n");
  else
    printf("tail: %f\n", tail->t_value);

  if (current_intersect == NULL)
    printf("current: NULL\n");
  else
    printf("current: %f\n", current_intersect->t_value);

  if (currently_at_head)
    printf("currently at head\n");
  else if (currently_at_tail)
    printf("currently at tail\n");
  else
    printf("currently in middle\n");

  for (ptr = head; ptr != NULL; ptr = ptr->next) {
    printf("node %f: prev ", ptr->t_value);
    if (ptr->prev == NULL)
      printf("NULL, next ");
    else
      printf("%f, next ", ptr->prev->t_value);
    if (ptr->next == NULL)
      printf("NULL.\n");
    else
      printf("%f.\n", ptr->next->t_value);
  }

  printf("# printfinish\n");
}

// Set one IntersectInfo structure to another.
void copy_intersect_info(IntersectInfo *dest, IntersectInfo *src)
{
  copy_point_info(&(dest->from), &(src->from));
  copy_point_info(&(dest->to), &(src->to));
  dest->normal_flipped = src->normal_flipped;
}

// Set one PointInfo structure to another.
void copy_point_info(PointInfo *dest, PointInfo *src)
{
  dest->empty = src->empty;

  dest->mat            = src->mat;
  if ((dest->tex_multi = src->tex_multi) == TRUE)
    dest->tex_col      = src->tex_col;
  else
    dest->texture      = src->texture;
  dest->sfxtype        = src->sfxtype;
  dest->sfx            = src->sfx;
}

// Are two PointInfo structures identical?
int point_info_equal(PointInfo *info1, PointInfo *info2)
{
  if (info1->empty != info2->empty)
    return FALSE;

  if (info1->mat != info2->mat)
    return FALSE;

  if (info1->tex_multi != info2->tex_multi)
    return FALSE;

  if (info1->texture != info2->texture)
    return FALSE;

  if (info1->sfxtype != info2->sfxtype)
    return FALSE;
 
  if (info1->sfx != info2->sfx)
    return FALSE;

  return TRUE;
}

// Set PointInfo from a primitive node.
void set_point_contents(primnode *node, PointInfo *info)
{
  info->empty = FALSE;
  info->mat = node->material;

  if (node->flags & FLAGS_MULTITEXTURE) {
    info->tex_multi = TRUE;
    info->tex_col = node->texcol;
  } else {
    info->tex_multi = FALSE;
    info->texture = node->texture;
  }
  info->sfxtype = node->sfxtype;
  info->sfx = node->sfx;
}

#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif
