/* list.h : Header file for the global intersection list.
** 
** This file defines the Intersection struct and IntersectionList class.
**
** Chris Butcher, 2 December 1996
*/

/*
 * The PointInfo structure contains information about a
 * point in space - material, texture and SFX.
 */
typedef struct _PointInfo {
  int empty;
  Material *mat;
  int tex_multi;
  union {
    SingleTexture *texture;
    TextureCollection *tex_col;
  };
  SfxType sfxtype;
  void *sfx;
} PointInfo;

/*
 * The IntersectInfo structure contains information about an
 * intersection - materials, textures and gases on both sides of
 * the intersection and the normal direction.
 *
 * This structure is also accompanied by an intersect_result.
 * If the result is IN, the primitive places its contents
 * into 'contents' - if OUT no contents are placed - if
 * SURFACE then 'from' and 'to' are filled with the from and to contents,
 * and normal_flipped is set.
 */
typedef struct _IntersectInfo {
  union {
    PointInfo contents;
    PointInfo from;
  };
  PointInfo to;
  int normal_flipped;
} IntersectInfo;

// The Intersection struct holds the data relating to an intersection
// with a primitive.
typedef struct Intersection {
  // Flags relating to the intersection. See constants.h for a list
  // of them. Note that this includes both the HIT_ENTERING flag and the
  // FLAGS_ flags that relate to primitives (TRANSLUCENT through to
  // ONLY_THROUGH_GATEWAY).
  uint hit_flags;

  // The primitive-node that the intersection is with.
  primnode *primitivehit;

  // The t value of the intersection.
  double t_value;

  // The normal vector to the intersection.
  vector normal;

  // The real primitive node that was hit (not the clustered one).
  primnode *real_primnode;

  // The contents of either side of the surface intersection.
  IntersectInfo info;

  // The primitive node representing the cluster that presently needs to
  // validate this intersection.
  primnode *validate_primnode;

  // Pointers for the doubly linked list.
  struct Intersection *prev, *next;
} intersection;

// The IntersectionList class stores the intersections of the current
// ray with every primitive. Intersections are inserted by the primitive
// intersection routines into this list.
class IntersectionList : public DeletableObject {
public:
  // Pointers to both ends of the list.
  intersection *head, *tail;

  // Next intersection to be retrieved.
  intersection *current_intersect;

  // Are we at the head or tail of the list?
  int currently_at_head, currently_at_tail;

  // Set up the linked list. Dump any existing intersections.
  void init(void);

  // Print out the list.
  void print_list(void);

  // Insert an intersection into the list.
  void insert(uint hit_flags, primnode *node, vector norm,
	      double t_value, int from_empty, int to_empty,
	      Material *mat_from, Material *mat_to,
	      SfxType sfxtype1, SfxType sfxtype2,
	      void *sfx1, void *sfx2,
	      int multitex1, int multitex2,
	      void *tex1, void *tex2,
	      vector ray_from, vector ray_dir,
	      uint ray_flags, primnode *real_primnode,
	      primnode *validate_primnode);

  // Get the next intersection from the list.
  intersection *get_next(void);
 
  // Undo the last retrieve operation from the list.
  void undo_last_retrieve(void);

  // Set the next-intersection pointer back to the start of the list.
  void restart(void);

  // Get the current intersection.
  intersection *get_position(void);

  // Set the current intersection.
  void set_position(intersection *new_pos);

  // Is the list empty?
  int empty(void);

  // Remove an intersection.
  void remove_intersection(intersection *the_int);

  // Delete the intersections in the list.
  void delete_intersections(void);

  // Constructor.
  IntersectionList(void) {
    head = tail = current_intersect = NULL;
    currently_at_head = TRUE;
    currently_at_tail = FALSE;
  }

  // Destructor. Delete all intersections.
  ~IntersectionList(void) {
    delete_intersections();
  }
};

// Set one IntersectInfo structure to another.
void copy_intersect_info(IntersectInfo *dest, IntersectInfo *src);

// Set one PointInfo structure to another.
void copy_point_info(PointInfo *dest, PointInfo *src);

// Are two PointInfo structures identical?
int point_info_equal(PointInfo *info1, PointInfo *info2);

// Set PointInfo from a primitive node.
void set_point_contents(primnode *node, PointInfo *info);


