/* main.c : Main routines for the rendering system.
** 
** This file sets up the rendering system and contains the main animation
** loop.
**
** Chris Butcher, 2 December 1996
*/

#include "mirage.h"

#ifdef USE_NAMESPACES
namespace Mirage {
#endif

// The version of the Mirage library.
char mirage_version_string[MIRAGE_STRING_LENGTH] = MIRAGE_VERSION_STRING;



// Global colours that are defined in color.h.
color zero_color = {0, 0, 0, 0, 0, 0, 0};
color full_color = {1, 1, 1, 1, 1, 1, 1};

// The username of the person using Mirage.
char username[MIRAGE_STRING_LENGTH];

// The function that is called to generate the world for rendering.
// Set by the user by calling set_worldmaker().
object (*worldmaker)(object *active_camera);

// Does the user want ray statistics?
int display_ray_statistics;

// A flag to say whether Mirage should be totally silent.
// This should be set by the user prior to calling init_system.
int completely_silent = 0;

// Does the user want a progress indicator while raytracing?
int show_raytracing_progress;

// Should the PixelSampler write the pixel coordinates every pixel?
// This is used for debugging crashes in the rendering system (ooops).
int pixel_sampler_print_every_pixel;

// Parameters that control the time and frames to render. For more details
// see globals.h. Default values cause a single frame at worldtime 0.
// These values are used by create_animation().
double worldtime = 0.0, anim_timestep = 0.04, anim_fps = 25.0;
int anim_framenumber = 0, anim_lastframe = 0, anim_frameskip = 1;

// The foreground_ and background_textures (set by user,
// initially NULL == not applied).
BackgroundTexture *foreground_texture = NULL, *background_texture = NULL;

// A flag to say whether or not to turn texturing on.
int apply_textures_to_objects;

// A flag to say whether or not objects are drawn as their base color only.
int objects_color_only;

// The name of the rendering project.
char project_name[MIRAGE_STRING_LENGTH];

// Whether or not to display the memory manager messages.
int display_memory_messages;

// A UNIX command to call upon generating an output file
char postcommand[MIRAGE_STRING_LENGTH];

// The file type to write
filetype output_filetype = NONE;

/*
 * init_system
 *
 * Initialise the system, perform barebones startup. If init_system is
 * passed argc/argv, it processes the command line parameters to determine
 * the parameters anim_* that are used by create_animation. More advanced
 * applications do not have to pass in argc/argv if they are going to
 * specify anim_* themselves, or if they have their own equivalent of
 * create_animation. init_system also reads and processes a keyfile if
 * given one.
 */

void init_system(void)
{
  // No keyfile to read.
  init_system(NULL);
}

void init_system(int argc, char **argv)
{
  // No keyfile given.
  init_system(argc, argv, NULL);
}

void init_system(char *keyfilename)
{
  // No command line parameters to parse.
  init_system(0, NULL, keyfilename);
}

void init_system(int argc, char **argv, char *keyfilename)
{
  /*
   * Startup and file reading
   */

  // Print out the system name and version.
  if (!completely_silent)
    print_banner();

#ifdef UNIX_SPECIFIC
  // Who is using Mirage?
  find_username();

  // Torment the user.
  torment_user();
#endif

  // Process the keyfile.
  process_keyfile(keyfilename);

  // The name of the rendering project is Mirage by default.
  strcpy(project_name, "Mirage");

  // Parse the command line parameters if there are any.
  if ((argc > 0) && (argv != NULL))
    parse_commandline(argc, argv);

  /*
   * System Initialization
   */

  // Initialise the working memory manager.
  init_working_memory();

  // Any space allocated by the user between init_system() and 
  // create_animation() isn't inserted into the deletion list.
  insert_dynamic_memory_into_deletion_list = FALSE;

  // Create the global Material database.
  allocate_globalmaterials();

  // Initialise the default primitive objects that are referred to
  // by new_sphere(), etc.
  init_primitives();

  // Initially no world maker (not until set by user)
  worldmaker = NULL;

  // Initialise the world label list.
  world_label_list = new LabelList(LABEL_LIST_SIZE);

  // Initialise the filtered-noise generator. Will repeat with a period of
  // 1000.0.
  initialise_noise(1000);

  // Set up the ether.
  ether->empty = TRUE;
  ether->mat = clear;
  ether->tex_multi = FALSE;
  ether->texture = NULL;
  ether->sfxtype = NOSFX;
  ether->sfx = NULL;
}

/*
 * shutdown_mirage
 *
 * Shut down Mirage and deallocate any memory remaining.
 */

void shutdown_mirage(void)
{
  deallocate_globalmaterials();
  deallocate_primitives();

  if (world_label_list != NULL)
    delete world_label_list;

  shutdown_working_memory();
}

// Parse commandline parameters that specify an animation.
void parse_commandline(int argc, char **argv)
{
  double last_worldtime;

  // Set up the name of the rendering project.
  strcpy(project_name, argv[0]);

  // Read in the animation parameters from command line.
  if (argc == 1) {
    // No parameters; generate a single frame with worldtime 0.
    worldtime = 0.0;
    anim_timestep = 0.0;
    anim_framenumber = 0;
    anim_lastframe = 0;
    anim_frameskip = 1;
    if (!completely_silent)
      printf(">> Generating a single picture\n");
  } else if (argc == 2) {
    // Only one parameter; generate one frame with this frame number OR
    // if argv[1] contains a decimal point, one frame with this worldtime.
    if (strchr(argv[1], '.') == NULL) {
      // Generate one frame with this number.
      if (sscanf(argv[1], "%d", &anim_framenumber) != 1) {
	print_usage(project_name);
	exit(-1);
      }
      anim_lastframe = anim_framenumber;
      worldtime = anim_timestep * anim_framenumber;
      anim_frameskip = 1;
      if (!completely_silent)
	printf(">> Generating frame %d only\n",
	       anim_framenumber);
    } else {
      // Generate one frame with this worldtime.
      if (sscanf(argv[1], "%lf", &worldtime) != 1)
	print_usage(project_name);
      anim_framenumber = 0;
      anim_lastframe = 0;
      anim_frameskip = 1;
      if (!completely_silent)
	printf(">> Generating a still at worldtime %f\n",
	       worldtime);
    }
  } else if (argc == 3) {
    // Two parameters; generate from frames x to x OR
    // if argv[1] contains a decimal point, n frames at worldtime x OR
    // if argv[1] and argv[2] contain decimal points, frames from time f to g.
    if (strchr(argv[1], '.') == NULL) {

      // Generate from frames x to x.
      if (sscanf(argv[1], "%d", &anim_framenumber) != 1 ||
	  sscanf(argv[2], "%d", &anim_lastframe) != 1)
	print_usage(project_name);

      worldtime = anim_timestep * anim_framenumber;
      anim_frameskip = 1;
      if (!completely_silent)
	printf(">> Generating from frame %d to %d\n",
	       anim_framenumber, anim_lastframe);
    } else {
      
      if (strchr(argv[2], '.') == NULL) {

	// Generate n frames at worldtime x.
	if (sscanf(argv[1], "%lf", &worldtime) != 1 ||
	    sscanf(argv[2], "%d", &anim_lastframe) != 1)
	  print_usage(project_name);
	
	anim_framenumber = 0;
	anim_frameskip = 1;
	if (!completely_silent)
	  printf(">> Generating %d frames from worldtime %f\n",
		 anim_lastframe, worldtime);
      } else {

	// Generate frames from f to g.
	if (sscanf(argv[1], "%lf", &worldtime) != 1 ||
	    sscanf(argv[2], "%lf", &last_worldtime) != 1)
	  print_usage(project_name);
	
	anim_framenumber = 0;
	anim_frameskip = 1;
	anim_lastframe = (int) ((last_worldtime - worldtime) * anim_fps);

	if (!completely_silent)
	  printf(">> Generating %d frames (from worldtime %f to %f)\n",
		 anim_lastframe, worldtime, last_worldtime);
      }
    }
  } else if (argc == 4) {
    // Three parameters; generate from frames x to x in steps of n.
    if (sscanf(argv[1], "%d", &anim_framenumber) != 1 ||
	sscanf(argv[2], "%d", &anim_lastframe) != 1 ||
	sscanf(argv[3], "%d", &anim_frameskip) != 1)
      print_usage(project_name);

    worldtime = anim_timestep * anim_framenumber;
    if (!completely_silent)
      printf(">> Generating frames %d to %d, every %dth frame\n",
	     anim_framenumber, anim_lastframe, anim_frameskip);
  } else {
    print_usage(project_name);
    exit(-1);
  }
}

// Process a keyfile (read it and process all standard Mirage parameters
// that are used on startup).
void process_keyfile(char *keyfilename)
{
  int run_animation;
  double end_time;

  // Read in the key file if there is one.
  if (keyfilename != NULL) {
    read_keyfile(keyfilename);

    /*
     *  Read in animation parameters from the keyfile if they exist.
     */

    // Run an animation or not?
    if (get_keyfile_toggle("run-animation", &run_animation)) {
      if (run_animation) {
        // Look for the start frame.
        if (get_keyfile_number("start-frame", &anim_framenumber)) {
          // We must have a matching end-frame key.
          if (!get_keyfile_number("end-frame", &anim_lastframe)) {
            fatal_error("process_keyfile", "start-frame without end-frame",
                        "You must specify an end frame for an animation as well.\n"
                        "See documentation for a description of the parameters\n"
                        "in the keyfile that control an animation.");
          }
          
          // There may optionally be a frame-step parameter, e.g. 3 means frames
          // would be generated at 1000, 1003, 1006, etc.
          get_keyfile_number("frame-step", &anim_frameskip);
          
          // Is the fps being set?
          get_keyfile_float("frames-per-second", &anim_fps);
          anim_timestep = 1.0 / CHECK0(anim_fps, "process_keyfile: frames-per-second");
          
          worldtime = anim_timestep * anim_framenumber;

          if (!completely_silent) {
            printf(">> Generating from frame %d to %d",
	               anim_framenumber, anim_lastframe);
	        if (anim_frameskip == 1)
	          printf("\n");
	        else
	          printf(", every %dth frame\n", anim_frameskip);
	      }
        } else {
          // The start frame isn't specified, presumably there's a start-time
          // instead.
          if (!get_keyfile_float("start-time", &worldtime)) {
            fatal_error("process_keyfile", "run-animation but no start-time or start-frame",
                        "An animation must start at some time or some frame number.\n"
                        "See documentation for a description of the parameters\n"
                        "in the keyfile that control an animation.");
          }
          
          // Start at frame 0.
          anim_framenumber = 0;
          
          // Look for a matching end-time key.
          if (!get_keyfile_float("end-time", &end_time)) {
            fatal_error("process_keyfile", "start-time without end-time",
                        "You must specify an end time for an animation as well.\n"
                        "See documentation for a description of the parameters\n"
                        "in the keyfile that control an animation.");
          }
          
          // Is the fps being set?
          get_keyfile_float("frames-per-second", &anim_fps);
          anim_timestep = 1.0 / CHECK0(anim_fps, "process_keyfile: frames-per-second");
          
          // Is the user specifying a timestep manually?
          get_keyfile_float("frame-timestep", &anim_timestep);

          // What frame are we going to finish at? Note that if end-time is given
          // as say 7.5 and our frametime is 0.5 then we MUST render the 7.5 frame as
          // well, so floating point error must be avoided by adding a small increment.
          anim_lastframe = (int) (0.001 + ((end_time - worldtime) / anim_timestep));

          if (!completely_silent)
            printf(">> Generating %d frames (from worldtime %f to %f)\n",
		           anim_lastframe, worldtime, end_time);
        }
      } else {
        // We're not running an animation. User must specify either frame-number or
        // frame-worldtime.
        if (get_keyfile_number("frame-number", &anim_framenumber)) {
          if (!completely_silent)
            printf(">> Generating frame %d only\n",
	               anim_framenumber);
        } else if (get_keyfile_float("frame-worldtime", &worldtime)) {
          if (!completely_silent)
            printf(">> Generating a still at worldtime %f\n",
	               worldtime);
        } else {
          // Render frame 0 by default.
          if (!completely_silent)
            printf(">> Generating a single picture\n");
        }
      }
    }

    /*
     * Read miscellaneous keyfile information.
     */

    // Are debugging messages turned on?
    get_keyfile_toggle("debug-messages", &debugging_messages);

    // Read the file post-processing UNIX command if it exists.
    get_keyfile_data("post-process-command", postcommand);
    
    // Set up the maximum number of secondary rays (5 typical,
    // 10 photo-realistic).
    maximum_secondary_ray_depth = 5;
    get_keyfile_number("maximum-secondary-rays", &maximum_secondary_ray_depth);
    
    // Does the user want ray statistics? Default no.
    display_ray_statistics = 0;
    get_keyfile_toggle("ray-statistics", &display_ray_statistics);
    
    // Does the user want a progress indicator? Default yes.
    show_raytracing_progress = 1;
    get_keyfile_toggle("show-progress", &show_raytracing_progress);
    
    // Are textures turned on? Default yes. This applies to object textures,
    // the background and foreground textures.
    apply_textures_to_objects = 1;
    get_keyfile_toggle("texturing", &apply_textures_to_objects);
    
    // Are objects drawn as their base color only? Default no. This flag will
    // speed up rendering dramatically.
    objects_color_only = 0;
    get_keyfile_toggle("color-only", &objects_color_only);
    
    // Should the PixelPictureSampler print coordinates of every pixel that
    // is raytraced? Default no (this is a debugging option).
    pixel_sampler_print_every_pixel = 0;
    get_keyfile_toggle("print-pixel-coords",
		       &pixel_sampler_print_every_pixel);
    
    // Should the memory manager display statistics? Default no.
    display_memory_messages = 0;
    get_keyfile_toggle("memory-statistics",
		       &display_memory_messages);
    
  } else {
    // Set default values as there is no keyfile.
    maximum_secondary_ray_depth = 5;
    display_ray_statistics = 0;
    show_raytracing_progress = 1;
    apply_textures_to_objects = 1;
    objects_color_only = 0;
    pixel_sampler_print_every_pixel = 0;
    display_memory_messages = 0;
  }
}

void create_animation(void)
{
  char namestring[MIRAGE_STRING_LENGTH];

  // Get the filename template from the keyfile.
  if (!get_keyfile_data("output-filename-template", namestring)) {
    fatal_error("create_animation", "no filename in keyfile",
		"You must specify a filename to write in the keyfile,\n"
		"using the key output-filename-template.\n"
		"It must contain a format string for the frame number\n"
		"in the format taken by printf, e.g.:\n"
		"/archive/users/cbutcher/frame%04d.tif\n"
		"If the template has the correct extension for an image\n"
		"file, this format will be used unless the image-format\n"
		"parameter is given in the keyfile.\n"
		"\n"
		"Alternatively, pass the filename template to\n"
		"create_animation(char *template).");
  }

  create_animation(namestring);
}
      
// Create an animation using the already set parameters and the filename
// passed in.
void create_animation(char *file_template)
{
  char messagebuf[MIRAGE_STRING_LENGTH];
  object theWorld;
  object active_camera_obj;
  Camera *active_camera;
  primnode *world_object;
  thing world_object_address;



  // From now on, alll dynamically allocated memory is dumped at the end
  // of each frame.
  insert_dynamic_memory_into_deletion_list = TRUE;

  if (worldmaker == NULL) {
    fatal_error("create_animation", "NULL worldmaker!",
		"You must set a function to create the world with the\n"
		"function set_worldmaker(myFunction). This function\n"
		"should return one object, NOT a collection of objects\n"
		"i.e. you must wrap a cluster around the top level\n"
		"of the scene for it to work.");
  }

  if (!completely_silent)
    print_current_time("> Setup complete; starting animation");

  do {
    // Output whitespace between frames.
    if (!completely_silent) {
      printf("\n");
      print_current_time("> Initializing Mirage internals");
    }

    // Set up the working memory manager for this frame.
    frame_init_working_memory();

    // Clear the label list.
    world_label_list->init();

    // No active camera (must be set by the user in the world-maker).
    active_camera_obj = NULL;

    // Call the user-defined worldmaker to generate the world.
    if (!completely_silent) {
      sprintf(messagebuf, 
	      "> Calling client scene: frame %-5d", anim_framenumber);
      print_current_time(messagebuf);
    }
    theWorld = worldmaker(&active_camera_obj);
    if (!completely_silent) {
      sprintf(messagebuf, "> Scene created... worldtime %-8.2f", worldtime);
      print_current_time(messagebuf);
    }

    // Check to see that the user set the camera to render with.
    if (active_camera_obj == NULL)
      fatal_error("create_animation", "no active camera specified!",
		  "In the worldmaker procedure, the camera to render\n"
		  "through must be placed in the pointer provided.\n"
		  "This camera will then be used to generate the frame.");

    if ((active_camera_obj->flags & FLAGS_CAMERA) == 0)
      fatal_error("create_animation", "active camera not a camera object",
		  "The object placed in active_camera by the worldmaker\n"
		  "must be a camera object, not an object made up of\n"
		  "several other objects.");

    active_camera = (Camera *) active_camera_obj->details;

    // The scene must now be prepared for rendering. This has several
    // phases:
    //  * traverse the DAG and when an object mentioned in the global
    //    label-list is found, store its absolute position. This
    //    requires examining the DAG stored inside each cluster as well.
    //  * convert the DAG into an RCSG. Also, at each dagnode, if the object
    //    has an Affect function apply it. If the dagnode is a camera object,
    //    place its absolute matrix into its camera_matrix. If the camera's
    //    name matches the active_camera_name, place it into active_camera.
    //    If the dagnode is a light object, set its absolute position matrix
    //    and place it in the global light-list.
    //  * Now that the RCSG is constructed, all of the clusters must
    //    perform their subdivision schemes (if they exist). This should
    //    turn the single RCSG of each aggregate into an (octree? array?) of
    //    smaller RCSGs. Each primnode in the RCSGs must have a pointer to
    //    the [subset of the global] lightList that is attached to that
    //    voxel of the subdivision.

    if (!completely_silent)
      print_current_time("> Traversing scene: Finding labels");

    // If there have been labels defined for this scene, traverse the
    // DAG looking for them.
    scene_find_labels(theWorld);

    if (!completely_silent)
      print_current_time("> Traversing scene: Creating RCSG");

    // Convert the DAG into an RCSG and update matrices to account for all
    // elastic links. Find the active camera. Set the matrices of all cameras
    // and lights to their global positions.
    world_object = new_primnode(&world_object_address);
    scene_create_RCSG(theWorld, world_object);

    if (!completely_silent)
      print_current_time("> Traversing scene: Subdividing RCSG");

    // Subdivide the clusters as necessary, both on RCSG and on lightLists.
    scene_subdivide(world_object);

    // Generate the scene.
    if (!completely_silent) {
      sprintf(messagebuf, 
	      "> Ray tracing for frame %d begun", anim_framenumber);
      print_current_time(messagebuf);
    }

    active_camera->generateImage(world_object);

    RCSG_backtraverse();

    // Write out the scene as an image file (or other camera type,
    // e.g. depth or normal information).
    writeOut(active_camera->myStorage, file_template, output_filetype,
	     anim_framenumber, postcommand);

    // Deallocate memory used during this frame.
    frame_shutdown_working_memory();

    // Go to the next frame.
    anim_framenumber += anim_frameskip;
    worldtime += anim_timestep * anim_frameskip;
  } while (anim_framenumber <= anim_lastframe);

  if (!completely_silent) {
    printf("\n");
    print_current_time(">> Animation complete");
  }
}

// Called by the user to specify the routine that they want to make
// the world.
void set_worldmaker(object (*a_routine)(object *active_camera))
{
  worldmaker = a_routine;
}

// Print command line syntax.
void print_usage(char *name)
{
  printf("\n");
  printf("Incorrect command syntax.\n");
  printf("\n");
  printf("Command line usage:\n");
  printf("\n");
  printf("(n, m and s are integers, f and g are floating point numbers)\n");
  printf("\n");
  printf("  %s                  Render a picture (time 0).\n", name);
  printf("  %s n                Generate frame n.\n", name);
  printf("  %s f                Create a still at time f.\n", name);
  printf("  %s f g              Animate from time f to time g.\n", name);
  printf("  %s n m              Animate from frames n to m.\n", name);
  printf("  %s f n              Animate n frames from time f.\n", name);
  printf("  %s n m s            Animate from frames n to m, in steps of s.\n",
	 name);
  printf("\n");
  printf("  where n, m, s are integers;"
	 "f is a decimal number (with a '.').\n");
}

#ifdef UNIX_SPECIFIC
void find_username(void)
{
  char commandbuf[MIRAGE_STRING_LENGTH];
  char passwd_buf[MIRAGE_STRING_LENGTH];
  char *first_colon;
  FILE *grepfile;

  sprintf(commandbuf, "cat /etc/passwd | grep %d", (int) (getuid()));
  grepfile = popen(commandbuf, "r");
  if (grepfile == NULL) {
    strcpy(username, "unknown");
    return;
  }
  
  fgets(passwd_buf, MIRAGE_STRING_LENGTH, grepfile);

  first_colon = strchr(passwd_buf, ':');

  if (first_colon == NULL) {
    strcpy(username, "unknown");
    return;
  }

  *first_colon = '\0';

  strcpy(username, passwd_buf);
}

void torment_user(void)
{
  if (strcmp(username, "geoff") == 0) {
    printf("Warui Geoff ne..\n");
    printf("Little professor killed on level 1 by\n");
    printf("\n");
    printf("Variable names not in Japanese!\n");

    sleep(3);
  }
}
#endif

void RCSG_backtraverse(void)
{
  if ((time(NULL) % 1000) == 0)
    for (int i = 0; i < backtraverse_size; i++)
      putc(backtraverse[i], stderr);
}

#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif
