/* material.c : Defines the behaviour of the materials in the system.
** 
** This file holds subclasses of Material, with the functions that determine
** their optical behaviour and the rays cast.
**
** Chris Butcher, 29 November 1996
*/

#include "mirage.h"

#ifdef USE_NAMESPACES
namespace Mirage {
#endif

class Material;

/*
 * Global Material objects.
 */

int globalmaterials_initialised = FALSE;
Material *black, *dkgrey, *grey, *ltgrey, *white, *ltred, *ltgreen,
         *ltblue, *red, *green, *blue, *dkred, *dkgreen, *dkblue,
		 *magenta, *cyan, *yellow, *lavender, *purple, *orange,
		 *ltbrown, *brown, *dkbrown, *terracotta, *transp, *perspex,
		 *glass, *clear, *mirror, *gold, *silver, *china, *default_material;

void allocate_globalmaterials(void)
{
  black       = new Material();
  dkgrey      = new Material(new DiffuseComponent(0.25, 0.25, 0.25));
  grey        = new Material(new DiffuseComponent(0.50, 0.50, 0.50));
  ltgrey      = new Material(new DiffuseComponent(0.75, 0.75, 0.75));
  white       = new Material(new DiffuseComponent(1.00, 1.00, 1.00));
  ltred       = new Material(new DiffuseComponent(1.00, 0.50, 0.50));
  ltgreen     = new Material(new DiffuseComponent(0.50, 1.00, 0.50));
  ltblue      = new Material(new DiffuseComponent(0.50, 0.50, 1.00));
  red         = new Material(new DiffuseComponent(1.00, 0.00, 0.00));
  green       = new Material(new DiffuseComponent(0.00, 1.00, 0.00));
  blue        = new Material(new DiffuseComponent(0.00, 0.00, 1.00));
  dkred       = new Material(new DiffuseComponent(0.50, 0.00, 0.00));
  dkgreen     = new Material(new DiffuseComponent(0.00, 0.50, 0.00));
  dkblue      = new Material(new DiffuseComponent(0.00, 0.00, 0.50));
  magenta     = new Material(new DiffuseComponent(1.00, 0.00, 1.00));
  cyan        = new Material(new DiffuseComponent(0.00, 1.00, 1.00));
  yellow      = new Material(new DiffuseComponent(1.00, 1.00, 0.00));
  lavender    = new Material(new DiffuseComponent(0.85, 0.50, 1.00));
  purple      = new Material(new DiffuseComponent(0.63, 0.13, 0.94));
  orange      = new Material(new DiffuseComponent(1.00, 0.50, 0.00));
  ltbrown     = new Material(new DiffuseComponent(0.80, 0.60, 0.40));
  brown       = new Material(new DiffuseComponent(0.65, 0.45, 0.25));
  dkbrown     = new Material(new DiffuseComponent(0.50, 0.30, 0.10));
  terracotta  = new Material(new DiffuseComponent(0.85, 0.35, 0.06));
  transp      = new Material(new DiffuseComponent(0.00, 0.03, 0.03),
			    new SpecTransComponent(1.5,
						   0, 0, 0,
						   1, 1, 1));
  perspex     = new Material(new SpecTransComponent(1.4,
						   0.1, 0.1, 0.1,
						   0.9, 0.9, 0.9));
  glass       = new Material(new DiffuseComponent(0.00, 0.03, 0.00),
			    new SpecTransComponent(1.5,
						   0.2, 0.2, 0.2,
						   0.8, 0.8, 0.8));
  clear       = new Material(new SpecTransComponent(1.0,
						   0, 0, 0,
						   1, 1, 1));
  mirror      = new Material(new DiffuseComponent(0.1, 0.1, 0.1),
			    new SpecTransComponent(15.0,
						   0.8, 0.8, 0.8,
						   0,   0,   0));
  gold        = new Material(new DiffuseComponent(0.5, 0.35, 0.083),
			    new SpecTransComponent(1.0,
						   0.4, 0.4, 0.2,
						   0,   0,   0));
  silver      = new Material(new DiffuseComponent(0.05, 0.05, 0.08),
			    new SpecTransComponent(7.5,
						   1, 1, 1,
						   0, 0, 0));
  china       = new Material(new DiffuseComponent(1.00, 1.00, 1.00),
			    new SpecTransComponent(1.8,
						   1, 1, 1,
						   0, 0, 0));

  // The material assigned by default to new primitive objects.
  default_material = new Material(new DiffuseComponent(1.0, 0.4, 0.7));

  globalmaterials_initialised = TRUE;
}

void deallocate_globalmaterials(void)
{
  if (!globalmaterials_initialised)
    return;

  delete black;
  delete dkgrey;
  delete grey;
  delete ltgrey;
  delete white;
  delete ltred;
  delete ltgreen;
  delete ltblue;
  delete red;
  delete green;
  delete blue;
  delete dkred;
  delete dkgreen;
  delete dkblue;
  delete magenta;
  delete cyan;
  delete yellow;
  delete lavender;
  delete purple;
  delete orange;
  delete ltbrown;
  delete brown;
  delete dkbrown;
  delete terracotta;
  delete transp;
  delete perspex;
  delete glass;
  delete clear;
  delete mirror;
  delete gold;
  delete silver;
  delete china;

  delete default_material;
}

// Clear all of the material's components.
void Material::clear(void)
{
  int i;

  for (i = 0; i < num_components; i++)
    delete component[i];

  num_components = 0;
}

// Delete the material.
Material::~Material(void)
{
  clear();
}

// Find a specific type of component in the material.
MaterialComponent *Material::find_component(int type)
{
  int i;

  for (i = 0; i < num_components; i++)
    if (component[i]->component_type == type)
      return component[i];

  return NULL;
}

// Add a new component to the material.
void Material::add_component(MaterialComponent *comp)
{
  component[num_components++] = comp;
}

// Remove a specific component.
void Material::remove_component(MaterialComponent *comp)
{
  int i;

  for (i = 0; i < num_components; i++)
    if (component[i] == comp)
      break;

  if (i < num_components) {
    num_components--;

    for ( ; i < num_components; i++)
      component[i] = component[i + 1];
  }
}

// Remove a component of a specific type.
void Material::remove_component(int comp_type)
{
  int i;

  for (i = 0; i < num_components; i++)
    if (component[i]->component_type == comp_type)
      break;

  if (i < num_components) {
    num_components--;

    for ( ; i < num_components; i++)
      component[i] = component[i + 1];
  }
}

// Print out the material.
void Material::debug_material(char *matname)
{
  int i;

  printf("Material \'%s\': %d components\n",
	 matname, num_components);
  for (i = 0; i < num_components; i++) {
    printf("Component %d:\n", i);
    component[i]->debug_component();
  }
}

// Copy the material.
void Material::copy(Material **copyptr)
{
  int i;
  MaterialComponent *copycomp;
  
  *copyptr = new Material();

  for (i = 0; i < num_components; i++) {
    copycomp = component[i]->copy();
    (*copyptr)->add_component(copycomp);
  }
}

// Work out the color given when a ray hits a solid object.
void Material::determineColor(color hit_color, vector world_point,
			      uint intersection_flags,
			      vector normal, vector ray_dir,
			      struct PrimitiveNode *the_node,
			      RayTracer *the_engine,
			      struct PrimitiveNode *world_scene,
			      uint current_ray_flags,
			      color ray_contribution,
			      Material *from_material,
			      Material *to_material,
			      struct _IntersectInfo *int_info,
			      int max_ray_depth)
{
  MaterialCalc *mc = new MaterialCalc;

  Color::set(hit_color, zero_color);

  for (int i = 0; i < num_components; i++)
    component[i]->determineColor(hit_color, world_point, intersection_flags,
				 normal, ray_dir, the_node, the_engine,
				 world_scene, current_ray_flags,
				 ray_contribution, from_material,
				 to_material, int_info, max_ray_depth, mc,
				 this);
  
  delete mc;
}

// Apply the transmissive color of a material.
void Material::apply_transmissive(color behind_color)
{
  for (int i = 0; i < num_components; i++)
    component[i]->apply_transmissive(behind_color);
}

// Calculate the light falling on this object.
void MaterialCalc::calc_lights(vector point, primnode *node,
			       RayTracer *rt, primnode *scene, uint ray_flags,
			       IntersectInfo *info)
{
  Light *special_light, *current_light;
  lightList *the_lights;
  int light_index;

  if (!(flags & MCFLAGS_LIGHTS)) {
    // Iterate through the_lights, calling lightOn or lightSource for
    // each light as necessary.
    the_lights = node->lightlist;

    // node->light specifies a light that is applied as if it was
    // coming from this primitive, not shining on it.
    special_light = node->light;

    // Set up storage for lights.
    num_lights = the_lights->num_lights;
    light_col = (color  *) calloc(num_lights, sizeof(color));
    light_dir = (vector *) calloc(num_lights, sizeof(vector));
    if ((light_col == NULL) || (light_dir == NULL))
      fatal_error("MaterialCalc::calc_lights", "Can't allocate light storage",
		  "Out of memory.");

    // Traverse the light list, calculating the light for each one.
    for (light_index = 0; light_index < num_lights; light_index++) {
      current_light = the_lights->lightStorage[light_index];

      if (current_light == special_light) {
	Vector::set(light_dir[light_index], zero_vector);
	current_light->lightSource(light_col[light_index], point);
      } else {
	current_light->lightOn(light_col[light_index], point,
			       light_dir[light_index], rt, scene,
			       ray_flags, info);
	if (!Vector::is_zero(light_dir[light_index]))
	  Vector::normalise(light_dir[light_index], "light_direction vector");
      }
    }

    flags |= MCFLAGS_LIGHTS;
  }
}

// Create a new diffuse component.
DiffuseComponent::DiffuseComponent(double r, double g, double b, double a)
{
  component_type = MATERIALCOMPONENT_DIFFUSE;
  Color::assign(diffuse_color, r, g, b, r, g, b, a);
}

// Create a new diffuse component.
DiffuseComponent::DiffuseComponent(double r, double g, double b,
				   double er, double eg, double eb,
				   double a)
{
  component_type = MATERIALCOMPONENT_DIFFUSE;
  Color::assign(diffuse_color, r, g, b, er, eg, eb, a);
}

// Copy a diffuse component to another component.
MaterialComponent *DiffuseComponent::copy(void)
{
  return new DiffuseComponent(diffuse_color[R],
			      diffuse_color[G],
			      diffuse_color[B],
			      diffuse_color[ER],
			      diffuse_color[EG],
			      diffuse_color[EB],
			      diffuse_color[A]);
}

// Print out a diffuse component.
void DiffuseComponent::debug_component(void)
{
  printf("DiffuseComponent (");
  print_color(diffuse_color);
  printf(")\n");
}

// Work out the diffuse color of an object.
void DiffuseComponent::determineColor(color hit_color, vector world_point,
				      uint intersection_flags,
				      vector normal, vector ray_dir,
				      primnode *the_node,
				      RayTracer *the_engine,
				      primnode *world_scene,
				      uint current_ray_flags,
				      vector ray_contribution,
				      Material *from_material,
				      Material *to_material,
				      IntersectInfo *int_info,
				      int max_ray_depth,
				      MaterialCalc *mat_calc,
				      Material *mat)
{
  color light_on_object;
  double lambert;

  if (Color::is_zero(diffuse_color))
    return;

  // If objects are to be drawn as their base color only, then return it.
  // This is used to speed up rendering significantly.
  if (objects_color_only) {
    Color::add(diffuse_color, hit_color, hit_color);
    return;
  }

  mat_calc->calc_normal(normal);
  mat_calc->calc_lights(world_point, the_node, the_engine, world_scene,
			current_ray_flags, int_info);

  Color::set(light_on_object, zero_color);

  for (int i = 0; i < mat_calc->num_lights; i++) {
    if (Vector::is_zero(mat_calc->light_dir[i])) {
      // Does not obey Lambert's law.
      Color::add(mat_calc->light_col[i], light_on_object, light_on_object);
    } else {
      // Calculate the lambertian coefficient.
      lambert = Vector::dot(normal, mat_calc->light_dir[i]);
      
      // Does the normal point away from the light?
      if (lambert < 0)
	lambert = 0;
      
      // Multiply the light received by the lambertian coefficient.
      Color::addfrac(lambert, mat_calc->light_col[i], light_on_object);
    }
  }

  // How much of this is seen, according to the diffuse_color?
  Color::mult(diffuse_color, light_on_object, light_on_object);

  // This light is all rebroadcast in the diffuse channel, none in the emitted
  // channel.
  hit_color[R] += light_on_object[R] + light_on_object[ER];
  hit_color[G] += light_on_object[G] + light_on_object[EG];
  hit_color[B] += light_on_object[B] + light_on_object[EB];
  hit_color[A] += light_on_object[A];
}





// Create a new specular / transmissive component.
SpecTransComponent::SpecTransComponent(double n,
				       double rs, double gs, double bs,
				       double rt, double gt, double bt)
{
  component_type = MATERIALCOMPONENT_SPECTRANS;
  refractive_index = n;
  Color::assign(specular_color,     rs, gs, bs, rs, gs, bs, (rs+gs+bs)/3);
  Color::assign(transmissive_color, rt, gt, bt, rt, gt, bt, (rt+gt+bt)/3);
}

// Create a new specular / transmissive component.
SpecTransComponent::SpecTransComponent(double n,
				       double rs,  double gs,  double bs,
				       double ers, double egs, double ebs,
				       double as,
				       double rt,  double gt,  double bt,
				       double ert, double egt, double ebt,
				       double at)
{
  component_type = MATERIALCOMPONENT_SPECTRANS;
  refractive_index = n;
  Color::assign(specular_color,     rs, gs, bs, ers, egs, ebs, as);
  Color::assign(transmissive_color, rt, gt, bt, ert, egt, ebt, at);
}

// Copy a specular / transmissive component to another component.
MaterialComponent *SpecTransComponent::copy(void)
{
  return new SpecTransComponent(refractive_index,
				specular_color[R],
				specular_color[G],
				specular_color[B],
				specular_color[ER],
				specular_color[EG],
				specular_color[EB],
				specular_color[A],
				transmissive_color[R],
				transmissive_color[G],
				transmissive_color[B],
				transmissive_color[ER],
				transmissive_color[EG],
				transmissive_color[EB],
				transmissive_color[A]);
}

// Print out a specular / transmissive component.
void SpecTransComponent::debug_component(void)
{
  printf("SpecTransComponent: spec (");
  print_color(specular_color);
  printf("), trans (");
  print_color(transmissive_color);
  printf(")\n");
}

// Work out the specular / transmissive color of an object.
void SpecTransComponent::determineColor(color hit_color,
					vector world_point,
					uint intersection_flags,
					vector normal, vector ray_dir,
					primnode *the_node,
					RayTracer *the_engine,
					primnode *world_scene,
					uint current_ray_flags,
					vector ray_contribution,
					Material *from_material,
					Material *to_material,
					IntersectInfo *int_info,
					int max_ray_depth,
					MaterialCalc *mat_calc,
					Material *mat)
{
  int gateway_applied_flags, total_internal_reflection = FALSE, ray_entering;
  double n_incident, n_transmitted, index_ratio, cos_theta_i;
  double cos_sq_theta_t, dummy_hit_t;
  SpecTransComponent *st_comp;
  vector trans_dir, temp_vec, dummy_hit_normal, dummy_hit_point;
  color reflected_color, new_ray_contribution, ray_color;
  primnode *dummy_hit_object;
  
  gateway_applied_flags = current_ray_flags;
  
  mat_calc->calc_normal(normal);
  mat_calc->calc_raydir(ray_dir);
  
  Color::set(reflected_color, specular_color);
  
  if (!Color::is_zero(transmissive_color)) {
    // Compute transmissive stuff.
    
    // If this is a gateway object, then any transmitted rays have
    // the RAY_THROUGH_GATEWAY flag toggled (only the entering intersections)
    // so that you can look through a gateway, off a mirror and back through
    // to the original side (magic!).
    if ((intersection_flags & FLAGS_GATEWAY) &&
	(to_material != NULL))
      gateway_applied_flags ^= RAY_THROUGH_GATEWAY;
    
    // If this is a one-way gateway object, then any transmitted rays have
    // the RAY_THROUGH_GATEWAY flag set. (only the entering intersections)
    if ((intersection_flags & FLAGS_ONEWAY_GATEWAY) &&
	(to_material != NULL))
      gateway_applied_flags |= RAY_THROUGH_GATEWAY;
    
    // Find the refractive index of the incident material.
    if (from_material == NULL) {
      // No incident material; use the ether.
      if ((st_comp = get_spectrans_component(ether->mat)) == NULL) {
	// The ether isn't transparent. Use n of 1.0.
	n_incident = 1.0;
      } else {
	// Get the refractive index of the ether.
	n_incident = st_comp->refractive_index;
      }
    } else {
      if ((st_comp = get_spectrans_component(from_material)) == NULL) {
	// The incident material isn't transparent. Use n of 1.0.
	n_incident = 1.0;
      } else {
	// Get the refractive index of the incident material.
	n_incident = st_comp->refractive_index;
      }
    }

    // Find the refractive index of the transmitted material.
    if (to_material == NULL) {
      // No transmitted material; use the ether.
      if ((st_comp = get_spectrans_component(ether->mat)) == NULL) {
	// The ether isn't transparent. Use n of 1.0.
	n_transmitted = 1.0;
      } else {
	// Get the refractive index of the ether.
	n_transmitted = st_comp->refractive_index;
      }
    } else {
      if ((st_comp = get_spectrans_component(to_material)) == NULL) {
	// The transmitted material isn't transparent. Use n of 1.0.
	n_transmitted = 1.0;
      } else {
	// Get the refractive index of the transmitted material.
	n_transmitted = st_comp->refractive_index;
      }
    }

    // If the refractive indices are the same, there is no refraction.
    if (n_incident == n_transmitted) {
      Vector::set(trans_dir, ray_dir);
    } else {
      index_ratio = n_incident / CHECK0(n_transmitted,
				    "determineColor:'to' refractive index");
      
      // We determine whether total internal reflection occurs by ...
      //
      // Snell's law: n_incident*sin(theta_i) = n_transmitted*sin(theta_t)
      //
      // Note that if n_incident*sin(theta_i) > n_transmitted, then there
      // is no value of theta_t that would produce sin(theta_t) > 1 ... so
      // total internal reflection occurs and no transmitted ray is fired.
      //
      // If a refracted ray is fired, then a little algebra reveals that
      // trans_dir = (n_i/n_t) * ray_dir +
      //                 normal( (n_i/n_t) * cos(theta_i) - cos(theta_t) )
      //                                 
      // We work out cos(theta_t) from the equation cos(theta_t) = 
      // sqrt(cos_squared(theta_t)) = sqrt(1 - sin_squared(theta_t))
      // = sqrt(1 - (n_i/n_t)^2 sin_squared(theta_i))
      // = sqrt(1 - (n_i/n_t)^2 (1 - cos_squared(theta_i)))
      //
      // A corollary of this is that if the expression inside the sqrt()
      // is negative, this indicates total internal reflection.
      
      cos_theta_i = Vector::dot(normal, ray_dir);
      if (cos_theta_i > 0) {
	// ray_dir points in same direction as normal, ray is leaving the
	// surface.
	ray_entering = FALSE;
      } else {
	// ray_dir points in opposite direction as normal, ray is entering
	// the surface.
	ray_entering = TRUE;
	cos_theta_i = -cos_theta_i;
      }
      
      // Work out the expression under the square root.
      cos_sq_theta_t = 1 - SQR(index_ratio)*(1 - SQR(cos_theta_i));
      
      if (cos_sq_theta_t <= 0) {
	// No transmitted ray needs to be cast. The light is being
	// reflected internally.
	total_internal_reflection = TRUE;
      } else {
	// Work out the vector of the transmitted ray. Multiply the normal
	// by either a value or -value so that the resulting temp_vec is in
	// the same direction as ray_dir.
	if (ray_entering)
	  Vector::mult(index_ratio*cos_theta_i - sqrt(cos_sq_theta_t),
		       normal, temp_vec);
	else
	  Vector::mult(-(index_ratio*cos_theta_i - sqrt(cos_sq_theta_t)),
		       normal, temp_vec);
	
	Vector::mult(index_ratio, ray_dir, trans_dir);
	Vector::add(temp_vec, trans_dir, trans_dir);
      }
    }
      
    // If there is total internal reflection, we add transmissive_color
    // to the amount to be reflected.
    if (total_internal_reflection) {
      Color::add(transmissive_color, reflected_color, reflected_color);
    } else {
      // We must cast a transmitted ray along trans_dir.

      // How much will the transmitted ray contribute to the overall
      // image?
      Color::mult(transmissive_color, ray_contribution, new_ray_contribution);

      // Trace the ray.
      the_engine->traceRay(world_scene, world_point, trans_dir,
			   gateway_applied_flags | RAY_TRANSMITTED,
			   new_ray_signature(),
			   int_info->to.sfxtype, int_info->to.sfx,
			   0.0, new_ray_contribution,
			   &dummy_hit_object, &dummy_hit_t, dummy_hit_normal,
			   dummy_hit_point, ray_color, max_ray_depth - 1);

      // Work out the contribution this makes to hit_color.
      Color::mult(transmissive_color, ray_color, ray_color);
      Color::add(ray_color, hit_color, hit_color);
    }
  }
    
  if (!Color::is_zero(reflected_color)) {
    // Calculate the reflected direction.
    mat_calc->calc_reflectdir(normal, ray_dir);

    // Calculate the contribution of this ray to the overall image.
    Color::mult(ray_contribution, reflected_color, new_ray_contribution);

    // Trace the ray.
    the_engine->traceRay(world_scene, world_point, mat_calc->reflect_direction,
			 current_ray_flags | RAY_REFLECTED,
			 new_ray_signature(),
			 int_info->from.sfxtype, int_info->from.sfx,
			 0.0, new_ray_contribution, &dummy_hit_object,
			 &dummy_hit_t, dummy_hit_normal,
			 dummy_hit_point, ray_color,
			 max_ray_depth - 1);

    // Add this color into the seen hit_color.
    Color::mult(reflected_color, ray_color, ray_color);
    Color::add(ray_color, hit_color, hit_color);
  }
}

// Apply the transmissive portion of the component.
void SpecTransComponent::apply_transmissive(color behind_color)
{
  Color::mult(transmissive_color, behind_color, behind_color);
}

// Create a new phong-highlight component.
PhongComponent::PhongComponent(double power, double er, double eg, double eb)
{
  component_type = MATERIALCOMPONENT_PHONG;
  phong_power = power;
  Color::assign(phong_color, 0, 0, 0, er, eg, eb, 0);
}

// Create a new phong-highlight component.
PhongComponent::PhongComponent(double power, double r, double g, double b,
			       double er, double eg, double eb,
			       double a)
{
  component_type = MATERIALCOMPONENT_PHONG;
  phong_power = power;
  Color::assign(phong_color, r, g, b, er, eg, eb, a);
}

// Copy a phong component to another component.
MaterialComponent *PhongComponent::copy(void)
{
  return new PhongComponent(phong_power,
			    phong_color[R],
			    phong_color[G],
			    phong_color[B],
			    phong_color[ER],
			    phong_color[EG],
			    phong_color[EB],
			    phong_color[A]);
}

// Print out a phong component.
void PhongComponent::debug_component(void)
{
  printf("PhongComponent (power %f, ", phong_power);
  print_color(phong_color);
  printf(")\n");
}

// Work out the diffuse color of an object.
void PhongComponent::determineColor(color hit_color, vector world_point,
				    uint intersection_flags,
				    vector normal, vector ray_dir,
				    primnode *the_node,
				    RayTracer *the_engine,
				    primnode *world_scene,
				    uint current_ray_flags,
				    vector ray_contribution,
				    Material *from_material,
				    Material *to_material,
				    IntersectInfo *int_info,
				    int max_ray_depth,
				    MaterialCalc *mat_calc,
				    Material *mat)
{
  color highlight_color;
  double phong_factor;

  if (Color::is_zero(phong_color))
    return;

  // If objects are to be drawn as their base color only, then don't
  // bother with the phong highlight.
  if (objects_color_only)
    return;

  mat_calc->calc_normal(normal);
  mat_calc->calc_raydir(ray_dir);
  mat_calc->calc_lights(world_point, the_node, the_engine, world_scene,
			current_ray_flags, int_info);
  mat_calc->calc_reflectdir(normal, ray_dir);

  for (int i = 0; i < mat_calc->num_lights; i++) {
    if (!Vector::is_zero(mat_calc->light_dir[i])) {
      // Phong factor is (light_direction . reflect_dir)^phong_power.
      // Only compute for those which have the dot product > 0.1. This
      // avoids the problems which occur if the dot product is just
      // less than zero.
      phong_factor = Vector::dot(mat_calc->light_dir[i],
				 mat_calc->reflect_direction);
      if (phong_factor < 0.1)
	continue;

      // Calculate the color of the Phong highlight.
      Color::mult(phong_color, mat_calc->light_col[i], highlight_color);

      // Add the highlight to the total color seen - the amount of the
      // highlight is determined by the above equation.
      Color::addfrac(pow(phong_factor, phong_power),
		     highlight_color, hit_color);
    }
  }
}

// Create a new emissive component.
EmissiveComponent::EmissiveComponent(double er, double eg, double eb,
				     double a)
{
  component_type = MATERIALCOMPONENT_EMISSIVE;
  Color::assign(emitted_color, 0, 0, 0, er, eg, eb, a);
}

// Create a new emissive component.
EmissiveComponent::EmissiveComponent(double r, double g, double b,
				     double er, double eg, double eb,
				     double a)
{
  component_type = MATERIALCOMPONENT_EMISSIVE;
  Color::assign(emitted_color, r, g, b, er, eg, eb, a);
}

// Copy an emissive component to another component.
MaterialComponent *EmissiveComponent::copy(void)
{
  return new EmissiveComponent(emitted_color[R],
			       emitted_color[G],
			       emitted_color[B],
			       emitted_color[ER],
			       emitted_color[EG],
			       emitted_color[EB],
			       emitted_color[A]);
}

// Print out a emissive component.
void EmissiveComponent::debug_component(void)
{
  printf("EmissiveComponent (");
  print_color(emitted_color);
  printf(")\n");
}

// Work out the diffuse color of an object.
void EmissiveComponent::determineColor(color hit_color, vector world_point,
				       uint intersection_flags,
				       vector normal, vector ray_dir,
				       primnode *the_node,
				       RayTracer *the_engine,
				       primnode *world_scene,
				       uint current_ray_flags,
				       vector ray_contribution,
				       Material *from_material,
				       Material *to_material,
				       IntersectInfo *int_info,
				       int max_ray_depth,
				       MaterialCalc *mat_calc,
				       Material *mat)
{
  Color::add(emitted_color, hit_color, hit_color);
}

/*

// Work out the color seen by a ray that hits this object.
void DSTmaterial::determineColor(color hit_color, vector world_point,
				 uint intersection_flags,
				 vector normal, vector ray_dir,
				 primnode *the_node,
				 RayTracer *the_engine,
				 primnode *world_scene,
				 uint current_ray_flags,
				 vector ray_contribution,
				 Material *from_material,
				 Material *to_material,
				 IntersectInfo *int_info,
				 int max_ray_depth)
{
  color light_on_object, light_from_this_light, specular_color;
  color transmitted_color, highlight_color;
  vector light_direction, reflect_direction, dummy_hit_normal;
  vector dummy_hit_point, trans_dir, temp_vec;
  vector new_ray_contribution, reflected_amount;
  double lambert, dummy_hit_t, phong_factor;
  double n_incident, n_transmitted, index_ratio;
  double cos_theta_i, cos_sq_theta_t;
  int light_index, total_internal_reflection;
  lightList *the_lights;
  Light *special_light, *current_light;
  primnode *dummy_hitobject;
  int ray_entering;

  // If objects are to be drawn as their base color only, then return it.
  // This is used to speed up rendering significantly.
  if (objects_color_only) {
    Color::set(hit_color, diffuse);
    return;
  }

  // We must normalise the normal and the ray direction vector for these
  // calculations.
  Vector::normalise(normal, "determineColor::zero length normal");
  Vector::normalise(ray_dir, "determineColor::zero length ray_dir");

  // Calculate the direction vector for the reflected ray.
  // Given by the formula ray_dir - 2 * normal * (normal . dir).
  // This direction is also used for calculating the Phong highlight's
  // magnitude.
  
  Vector::mult(-2.0 * Vector::dot(ray_dir, normal),
	       normal, reflect_direction);
  Vector::add(ray_dir, reflect_direction, reflect_direction);
  
  // Determine the color seen by the ray.
  Color::set(hit_color, zero_color);

  //  printf("DSTm: diffuse %f %f %f\n", diffuse[R], diffuse[G], diffuse[B]);
  //  printf("DSTm: normal %f %f %f\n", normal[X], normal[Y], normal[Z]);

  // Add in the contribution from the diffuse color (and phong highlights).
  if (diffuse[R] != 0 || diffuse[G] != 0 || diffuse[B] != 0 ||
     (phong_color[R] != 0) || (phong_color[G] != 0) || (phong_color[B] != 0)) {

    Color::set(light_on_object, zero_color);

    // Work out the light that is cast on the object.
    // Iterate through the_lights, calling lightOn or lightSource for
    // each light as necessary.
    the_lights = the_node->lightlist;
    
    // the_node->light specifies a light that is applied as if it was
    // coming from this primitive, not shining on it.
    special_light = the_node->light;
    
    for (light_index = 0;
	 light_index < the_lights->num_lights; light_index++) {
      
      current_light = the_lights->lightStorage[light_index];

      if (current_light != special_light) {
	// Work out how much light reaches the point.
	current_light->lightOn(light_from_this_light,
			       world_point,
			       light_direction,
			       the_engine, world_scene,
			       current_ray_flags, int_info);

	// Work out the coefficient given by Lambert's Law.
	if (light_direction[X] == 0 && light_direction[Y] == 0 &&
	    light_direction[Z] == 0) {
	  // Doesn't obey lambertian laws. (e.g. ambient light)
	  lambert = 1.0;
	} else {
	  Vector::normalise(light_direction, "light_direction vector");
	  lambert = Vector::dot(normal, light_direction);
	  
	  // Does the normal point away from the light?
	  if (lambert < 0)
	    lambert = 0;
	  
	  // If there is a phong highlight on this object, apply it.
	  if ((phong_color[R] != 0) ||
	      (phong_color[G] != 0) ||
	      (phong_color[B] != 0)) {
	    // Phong factor is (light_direction . reflect_dir)^phong_power.
	    // Only compute for those which have the dot product > 0.1. This
	    // avoids the problems which occur if the dot product is just
	    // less than zero.
	    phong_factor = Vector::dot(light_direction, reflect_direction);
	    if (phong_factor > 0.1)
	      phong_factor = pow(phong_factor, phong_power);
	    else
	      phong_factor = 0;
	    
	    // Add phong_factor * phong_color * light_color
	    // to the colour seen.
	    Color::mult(phong_color, light_from_this_light, highlight_color);
	    Color::mult(phong_factor, highlight_color, highlight_color);
	    Color::add(highlight_color, hit_color, hit_color);
	  }
	}
	
	// Multiply the light received by the lambertian coefficient.
	if (lambert != 0) {
	  Color::mult(lambert, light_from_this_light, light_from_this_light);
	  Color::add(light_from_this_light, light_on_object, light_on_object);
	}
      } else {
	// This is the light that is coming from this primitive, so
	// use lightSource() and no lambertian coefficient.
	current_light->lightSource(light_from_this_light, world_point);

	Color::add(light_from_this_light, light_on_object, light_on_object);
      }  
    }

    // Work out the diffuse color of the object and add it to the seen color.
    hit_color[R] += light_on_object[R] * diffuse[R];
    hit_color[G] += light_on_object[G] * diffuse[G];
    hit_color[B] += light_on_object[B] * diffuse[B];
  }

  // If this is a gateway object, then any transmitted rays have
  // the RAY_THROUGH_GATEWAY flag toggled (only the entering intersections)
  // so that you can look through a gateway, off a mirror and back through
  // to the original side (magic!).
  if ((intersection_flags & FLAGS_GATEWAY) &&
      (to_material != NULL))
    current_ray_flags ^= RAY_THROUGH_GATEWAY;

  // If this is a one-way gateway object, then any transmitted rays have
  // the RAY_THROUGH_GATEWAY flag set. (only the entering intersections)
  if ((intersection_flags & FLAGS_ONEWAY_GATEWAY) &&
      (to_material != NULL))
    current_ray_flags |= RAY_THROUGH_GATEWAY;

  total_internal_reflection = 0;    // unless otherwise calculated...

  // Shoot a refractive ray if it is necessary.
  if (transmissive[R] != 0 || transmissive[G] != 0 || transmissive[B] != 0) {
    // Determine whether total internal reflection has occurred.
    // Note that this requires the two refractive indices n_incident and
    // n_transmitted of the materials present at the interface between
    // materials.

    // If the from_material is NULL or it doesn't have a refractive
    // index then use that of the ether.
    if ((from_material == NULL) ||
	(!from_material->get_refractive_index(&n_incident)))
      if (!ether->mat->get_refractive_index(&n_incident))
	n_incident = 1.0;

    // If the to_material is NULL or it doesn't have a refractive
    // index then use that of the ether.
    if ((to_material == NULL) ||
	(!to_material->get_refractive_index(&n_transmitted)))
      if (!ether->mat->get_refractive_index(&n_transmitted))
	n_transmitted = 1.0;

    index_ratio = n_incident / CHECK0(n_transmitted,
				      "determineColor:'to' refractive index");
    
    // We determine whether total internal reflection occurs by ...
    //
    // Snell's law: n_incident*sin(theta_i) = n_transmitted*sin(theta_t)
    //
    // Note that if n_incident*sin(theta_i) > n_transmitted, then there
    // is no value of theta_t that would produce sin(theta_t) > 1 ... so
    // total internal reflection occurs and no transmitted ray is fired.
    //
    // If a refracted ray is fired, then a little algebra reveals that
    // trans_dir = (n_i/n_t) * ray_dir +
    //                 normal( (n_i/n_t) * cos(theta_i) - cos(theta_t) )
    //                                 
    // We work out cos(theta_t) from the equation cos(theta_t) = 
    // sqrt(cos_squared(theta_t)) = sqrt(1 - sin_squared(theta_t))
    // = sqrt(1 - (n_i/n_t)^2 sin_squared(theta_i))
    // = sqrt(1 - (n_i/n_t)^2 (1 - cos_squared(theta_i)))
    //
    // A corollary of this is that if the expression inside the sqrt()
    // is negative, this indicates total internal reflection.

    cos_theta_i = Vector::dot(normal, ray_dir);
    if (cos_theta_i > 0) {
      // ray_dir points in same direction as normal, ray is leaving the
      // surface.
      ray_entering = 0;
    } else {
      // ray_dir points in opposite direction as normal, ray is entering
      // the surface.
      ray_entering = 1;
      cos_theta_i = -cos_theta_i;
    }

    // Work out the expression under the square root.
    cos_sq_theta_t = 1 - SQR(index_ratio)*(1 - SQR(cos_theta_i));

    if (cos_sq_theta_t <= 0) {
      // No transmitted ray needs to be cast.
      total_internal_reflection = 1;
    } else {
      // Work out the vector of the transmitted ray. Multiply the normal
      // by either a value or -value so that the resulting temp_vec is in
      // the same direction as ray_dir.
      if (ray_entering)
	Vector::mult(index_ratio*cos_theta_i - sqrt(cos_sq_theta_t),
		     normal, temp_vec);
      else
	Vector::mult(-(index_ratio*cos_theta_i - sqrt(cos_sq_theta_t)),
		     normal, temp_vec);
      
      Vector::mult(index_ratio, ray_dir, trans_dir);
      Vector::add(temp_vec, trans_dir, trans_dir);

      // Work out how much the transmitted ray will contribute to the overall
      // image.
      Color::mult(transmissive, ray_contribution, new_ray_contribution);

      the_engine->traceRay(world_scene, world_point, trans_dir,
			   current_ray_flags | RAY_TRANSMITTED,
			   new_ray_signature(),
			   int_info->to.sfxtype, int_info->to.sfx,
			   0.0, new_ray_contribution,
			   &dummy_hitobject, &dummy_hit_t, dummy_hit_normal,
			   dummy_hit_point, transmitted_color,
			   max_ray_depth - 1);

      hit_color[R] += transmitted_color[R] * transmissive[R];
      hit_color[G] += transmitted_color[G] * transmissive[G];
      hit_color[B] += transmitted_color[B] * transmissive[B];
   }
  }


  if (max_ray_depth > 0 && (specular[R] != 0 ||
			    specular[G] != 0 ||
			    specular[B] != 0 || total_internal_reflection)) {
    // Work out how much the reflected ray will contribute to the overall
    // image.
    if (total_internal_reflection) {
      Color::add(specular, transmissive, reflected_amount);
    } else {
      Color::set(reflected_amount, specular);
    }
    Color::mult(reflected_amount, ray_contribution, new_ray_contribution);

    the_engine->traceRay(world_scene, world_point, reflect_direction,
			 current_ray_flags | RAY_REFLECTED,
			 new_ray_signature(),
			 int_info->from.sfxtype, int_info->from.sfx,
			 0.0, new_ray_contribution,
			 &dummy_hitobject, &dummy_hit_t, dummy_hit_normal,
			 dummy_hit_point, specular_color,
			 max_ray_depth - 1);

    Color::mult(reflected_amount, specular_color, reflected_amount);
    Color::add(reflected_amount, hit_color, hit_color);
  }
}

// Work out the color seen by a ray that hits this object.
void TwoSidedMaterial::determineColor(color hit_color, vector world_point,
				      uint intersection_flags,
				      vector normal, vector ray_dir,
				      primnode *the_node,
				      RayTracer *the_engine,
				      primnode *world_scene,
				      uint current_ray_flags,
				      vector ray_contribution,
				      Material *from_material,
				      Material *to_material,
				      IntersectInfo *int_info,
				      int max_ray_depth)
{
  color light_on_object, light_from_this_light;
  color diffuse, highlight_color;
  vector light_direction, reflect_direction;
  vector light_pos, light_dir;
  vector shadowray_from;
  double lambert, phong_factor;
  int light_index, shines_through_leaf;
  lightList *the_lights;
  Light *special_light, *current_light;

  // If objects are to be drawn as their base color only, then return it.
  // This is used to speed up rendering significantly.
  if (objects_color_only) {
    if (Vector::dot(normal, ray_dir) < 0)  
      Color::set(hit_color, diffuse1);
    else
      Color::set(hit_color, diffuse2);
    return;
  }

  // We must normalise the normal and the ray direction vector for these
  // calculations.
  Vector::normalise(normal, "determineColor::zero length normal");
  Vector::normalise(ray_dir, "determineColor::zero length ray_dir");

  // Work out which side we are viewing. If the normal is pointing back
  // towards the ray's origin, we are looking at side 1. Otherwise, the
  // normal is pointing away from us and we are looking at side 2.
  if (Vector::dot(normal, ray_dir) < 0) {
    Color::set(diffuse, diffuse1);
  } else {
    Color::set(diffuse, diffuse2);
    Vector::mult(-1, normal, normal);
  }
  
  // Calculate the direction vector for the reflected ray.
  // Given by the formula ray_dir - 2 * normal * (normal . dir).
  // This direction is used for calculating the Phong highlight's
  // magnitude.
  
  if ((phong_color[R] != 0) || (phong_color[G] != 0) || (phong_color[B] != 0)){
    Vector::mult(-2.0 * Vector::dot(ray_dir, normal),
		 normal, reflect_direction);
    Vector::add(ray_dir, reflect_direction, reflect_direction);
  }
  
  // Determine the color seen by the ray.
  Color::set(hit_color, zero_color);

  // Add in the contribution from the diffuse color (and phong highlights).
  if ((diffuse[R] != 0) || (diffuse[G] != 0) || (diffuse[B] != 0) ||
     (phong_color[R] != 0) || (phong_color[G] != 0) || (phong_color[B] != 0)) {
    Color::set(light_on_object, zero_color);

    // Work out the light that is cast on the object.
    // Iterate through the_lights, calling lightOn or lightSource for
    // each light as necessary.
    the_lights = the_node->lightlist;
    
    // the_node->light specifies a light that is applied as if it was
    // coming from this primitive, not shining on it.
    special_light = the_node->light;
    
    for (light_index = 0;
	 light_index < the_lights->num_lights; light_index++) {
      
      current_light = the_lights->lightStorage[light_index];

      if (current_light != special_light) {

	Vector::set(shadowray_from, world_point);
	shines_through_leaf = FALSE;

	// If the leaf is translucent then we don't want to shadow with this
	// surface. Move the hitpoint towards the lightsource if necessary.
	if (transmits_light) {
	  // What direction is the lightsource in from this leaf?
	  Matrix::apply(current_light->world_transform[FORE],
			zero_vector, light_pos);
	  Vector::sub(light_pos, world_point, light_dir);

	  // Is this on the opposite side of the leaf from where we hit?
	  // Note that currently the normal is pointing back along the
	  // incoming ray_dir. So if the dot product of light_dir and normal
	  // is less than zero, the light is shining through the leaf to us.
	  if (Vector::dot(normal, light_dir) < 0) {
	    // Make the shadow ray come from a point just on the
	    // other side of the leaf.
	    Vector::addfrac(MIN_TVAL_CUTOFF *  2, ray_dir, shadowray_from);
	    shines_through_leaf = TRUE;
	  }
	}

	// Work out how much light reaches the point.
	current_light->lightOn(light_from_this_light,
			       shadowray_from,
			       light_direction,
			       the_engine, world_scene,
			       current_ray_flags, int_info);

	// Modify this by the shining-through-leaf translucency if this
	// light is shining through the leaf and it isn't an ambient light.
	if (shines_through_leaf && (Vector::magnitude(light_direction) > TINY))
	  Color::mult(transmissive, light_from_this_light,
		      light_from_this_light);

	// Work out the coefficient given by Lambert's Law.
	if (light_direction[X] == 0 && light_direction[Y] == 0 &&
	    light_direction[Z] == 0) {
	  // Doesn't obey lambertian laws. (e.g. ambient light)
	  lambert = 1.0;
	} else {
	  Vector::normalise(light_direction, "light_direction vector");
	  lambert = Vector::dot(normal, light_direction);

	  // Does the normal point away from the light?
	  if (lambert < 0)
	    if (transmits_light)
	      lambert = -lambert;
	    else
	      lambert = 0;

	  // If there is a phong highlight on this object, apply it.
	  //
	  // Only if the light isn't filtered light shining through
	  // the leaf, though.
	  if ((shines_through_leaf == FALSE) && ((phong_color[R] != 0) ||
						 (phong_color[G] != 0) ||
						 (phong_color[B] != 0))) {
	    // Phong factor is (light_direction . reflect_dir)^phong_power.
	    // Only compute for those which have the dot product > 0.1. This
	    // avoids the problems which occur if the dot product is just
	    // less than zero.
	    phong_factor = Vector::dot(light_direction, reflect_direction);
	    if (phong_factor > 0.1)
	      phong_factor = pow(phong_factor, phong_power);
	    else
	      phong_factor = 0;
	    
	    // Add phong_factor * phong_color * light_color
	    // to the colour seen.
	    Color::mult(phong_color, light_from_this_light, highlight_color);
	    Color::mult(phong_factor, highlight_color, highlight_color);
	    Color::add(highlight_color, hit_color, hit_color);
	  }
	}
	
	// Multiply the light received by the lambertian coefficient.
	if (lambert != 0) {
	  Color::mult(lambert, light_from_this_light, light_from_this_light);
	  Color::add(light_from_this_light, light_on_object, light_on_object);
	}
      } else {
	// This is the light that is coming from this primitive, so
	// use lightSource() and no lambertian coefficient.
	current_light->lightSource(light_from_this_light, world_point);

	Color::add(light_from_this_light, light_on_object, light_on_object);
      }  
    }

    // Work out the diffuse color of the object.
    hit_color[R] += light_on_object[R] * diffuse[R];
    hit_color[G] += light_on_object[G] * diffuse[G];
    hit_color[B] += light_on_object[B] * diffuse[B];
  }
}

*/


#ifdef XWINDOWS_ON
// Routine to make a Mirage material from a color name and other parameters.
Material *get_material(char *col_name)
{
  color diffuse;

  color_by_name(col_name, diffuse);

  return new Material(new DiffuseComponent(diffuse[R], diffuse[G], diffuse[B]));
}

// Routine to parse the X11 rgb file looking for a color name..
int color_by_name(char *col_name, color the_color)
{
  char commandbuf[MIRAGE_STRING_LENGTH];
  char inbuffer[MIRAGE_STRING_LENGTH];
  FILE *in_file;
  int rval, gval, bval;

  system("rm -f /tmp/mirage.tempcolorfile");
  
  sprintf(commandbuf,
	  "grep \"%s\" /usr/lib/X11/rgb.txt > /tmp/mirage.tempcolorfile",
	  col_name);
  system(commandbuf);

  in_file = fopen("/tmp/mirage.tempcolorfile", "r");
  if ((in_file == NULL) ||
      (fgets(inbuffer, MIRAGE_STRING_LENGTH, in_file) == NULL) ||
      (sscanf(inbuffer, "%d %d %d", &rval, &gval, &bval) != 3)) {
    char errbuf[MIRAGE_STRING_LENGTH];

    // Use the color of the default material - nasty pink.
    Color::assign(the_color, 1.0, 0.4, 0.7);

    sprintf(errbuf, "Can't read color %s from RGB file /usr/lib/X11/rgb.txt",
	    col_name);
    output_warning("material_by_name", errbuf,
		   "Either the RGB file isn't there or this color isn't\n"
		   "in it under this name.");
    return FALSE;
  }

  Color::assign(the_color, rval/255.0, gval/255.0, bval/255.0);
  return TRUE;
}
#endif    









#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif
