/* material.h : Structures and prototypes for object materials.
** 
** This file defines subclasses of Material, with the functions that determine
** their optical behaviour and the rays cast.
**
** Chris Butcher, 26 January 1998
*/

// Prototypes for Material routines.
#ifdef XWINDOWS_ON
int color_by_name(char *col_name, color the_color);
Material *get_material(char *col_name);
#endif

// A MaterialComponent is a particular optical property of an object -
// diffuse reflection, specular reflection, etc.
class MaterialComponent : public DeletableObject {
public:
  // The type of this component.
  int component_type;

  // Determine the color produced by this component.
  virtual void determineColor(color hit_color, vector world_point,
			      uint intersection_flags,
			      vector normal, vector ray_dir,
			      struct PrimitiveNode *the_node,
			      RayTracer *the_engine,
			      struct PrimitiveNode *world_scene,
			      uint current_ray_flags,
			      color ray_contribution,
			      Material *from_material,
			      Material *to_material,
			      struct _IntersectInfo *int_info,
			      int max_ray_depth,
			      class MaterialCalc *mat_calc,
			      class Material *mat) = 0;

  // Apply the transmissive portion of this component.
  virtual void apply_transmissive(color behind_color) = 0;

  // Print out the component's properties.
  virtual void debug_component(void) = 0;

  // Make a copy of the material component.
  virtual MaterialComponent *copy(void) = 0;
};

// Allocate and deallocate the database of global Materials.
void allocate_globalmaterials(void);
void deallocate_globalmaterials(void);

// The Material class holds both the optical properties of an object
// and a means for extracting a color from these properties by firing
// rays given the ray tracing engine.
//
// A Material is made up of a set of MaterialComponents that each determine
// one aspect of the object's color.
class Material : public DeletableObject {
public:
  // The number of components in this material.
  int num_components;

  // The array of material components, and their precedences.
  MaterialComponent *component[MAX_MATERIAL_COMPONENTS];

  // Work out the color given when a ray hits a solid object.
  void determineColor(color hit_color, vector world_point,
		      uint intersection_flags,
		      vector normal, vector ray_dir,
		      struct PrimitiveNode *the_node,
		      RayTracer *the_engine,
		      struct PrimitiveNode *world_scene,
		      uint current_ray_flags,
		      color ray_contribution,
		      Material *from_material,
		      Material *to_material,
		      struct _IntersectInfo *int_info,
		      int max_ray_depth);

  // Apply the transmissive component to find out what light is
  // passed through the material.
  void apply_transmissive(color behind_color);

  // Make a copy of the material.
  void copy(Material **copyptr);

  // Print out the material properties.
  void debug_material(char *matname);
  
  // Clear all the material's components.
  void clear(void);

  // Add a new component.
  void add_component(MaterialComponent *comp);

  // Find a specific component.
  MaterialComponent *find_component(int component_type);

  // Remove a specific component.
  void remove_component(int component_type);
  void remove_component(MaterialComponent *comp);

  // Create a new blank material.
  Material(void) {
    num_components = 0;
  }

  // Create a material from a set of components.
  Material(MaterialComponent *comp1) {
    num_components = 0;
    add_component(comp1);
  }
  Material(MaterialComponent *comp1, MaterialComponent *comp2) {
    num_components = 0;
    add_component(comp1);
    add_component(comp2);
  }
  Material(MaterialComponent *comp1, MaterialComponent *comp2,
	   MaterialComponent *comp3) {
    num_components = 0;
    add_component(comp1);
    add_component(comp2);
    add_component(comp3);
  }
  Material(MaterialComponent *comp1, MaterialComponent *comp2,
	   MaterialComponent *comp3, MaterialComponent *comp4) {
    num_components = 0;
    add_component(comp1);
    add_component(comp2);
    add_component(comp3);
    add_component(comp4);
  }
  Material(MaterialComponent *comp1, MaterialComponent *comp2,
	   MaterialComponent *comp3, MaterialComponent *comp4,
	   MaterialComponent *comp5) {
    num_components = 0;
    add_component(comp1);
    add_component(comp2);
    add_component(comp3);
    add_component(comp4);
    add_component(comp5);
  }
  Material(MaterialComponent *comp1, MaterialComponent *comp2,
	   MaterialComponent *comp3, MaterialComponent *comp4,
	   MaterialComponent *comp5, MaterialComponent *comp6) {
    num_components = 0;
    add_component(comp1);
    add_component(comp2);
    add_component(comp3);
    add_component(comp4);
    add_component(comp5);
    add_component(comp6);
  }
  Material(MaterialComponent *comp1, MaterialComponent *comp2,
	   MaterialComponent *comp3, MaterialComponent *comp4,
	   MaterialComponent *comp5, MaterialComponent *comp6,
	   MaterialComponent *comp7) {
    num_components = 0;
    add_component(comp1);
    add_component(comp2);
    add_component(comp3);
    add_component(comp4);
    add_component(comp5);
    add_component(comp6);
    add_component(comp7);
  }
  Material(MaterialComponent *comp1, MaterialComponent *comp2,
	   MaterialComponent *comp3, MaterialComponent *comp4,
	   MaterialComponent *comp5, MaterialComponent *comp6,
	   MaterialComponent *comp7, MaterialComponent *comp8) {
    num_components = 0;
    add_component(comp1);
    add_component(comp2);
    add_component(comp3);
    add_component(comp4);
    add_component(comp5);
    add_component(comp6);
    add_component(comp7);
    add_component(comp8);
  }

  // Delete the material.
  ~Material(void);
};

const unsigned short MCFLAGS_LIGHTS  = (1 << 0);
const unsigned short MCFLAGS_NORMAL  = (1 << 1);
const unsigned short MCFLAGS_RAYDIR  = (1 << 2);
const unsigned short MCFLAGS_REFLECT = (1 << 3);

// Material components don't want to recalculate some values (normal, ray_dir,
// light from each light, etc.) so we store these in a MaterialCalc class.
class MaterialCalc {
public: 
  unsigned short flags;           // What has been calculated?

  int num_lights;
  color *light_col;               // The lights, and where they are,
  vector *light_dir;              // and how much light they give the object.

  vector reflect_direction;       // Specular reflection direction.

  // Calculate the lights on this object.
  void calc_lights(vector point, struct PrimitiveNode *node,
		   RayTracer *rt, struct PrimitiveNode *scene, uint ray_flags,
		   struct _IntersectInfo *info);

  // Calculate the unit normal on this object.
  void calc_normal(vector norm)
  {
    if (!(flags & MCFLAGS_NORMAL)) {
      Vector::normalise(norm, "MaterialCalc::zero length normal");
      flags |= MCFLAGS_NORMAL;
    }
  }


  // Calculate the unit ray_dir on this object.
  void calc_raydir(vector raydir)
  {
    if (!(flags & MCFLAGS_RAYDIR)) {
      Vector::normalise(raydir, "MaterialCalc::zero length raydir");
      flags |= MCFLAGS_RAYDIR;
    }
  }

  // Calculate the reflection direction on this object.
  void calc_reflectdir(vector norm, vector raydir)
  {
    calc_normal(norm);
    calc_raydir(raydir);
    
    if (!(flags & MCFLAGS_REFLECT)) {
      // Given by the formula ray_dir - 2 * normal * (normal . dir).
      Vector::mult(-2.0 * Vector::dot(raydir, norm),
		   norm, reflect_direction);
      Vector::add(raydir, reflect_direction, reflect_direction);

      flags |= MCFLAGS_REFLECT;
    }      
  }

  // Create a new MaterialCalc class.
  MaterialCalc(void) {
    flags = 0;
  }

  // Delete the MaterialCalc class.
  ~MaterialCalc(void) {
    if (flags & MCFLAGS_LIGHTS) {
      free(light_col);
      free(light_dir);
    }
  }
};












// The different types of material components.
const int MATERIALCOMPONENT_ERROR        = 0;
const int MATERIALCOMPONENT_DIFFUSE      = 1;
const int MATERIALCOMPONENT_SPECTRANS    = 2;
const int MATERIALCOMPONENT_PHONG        = 3;
const int MATERIALCOMPONENT_EMISSIVE     = 4;

// The diffuse reflection component of Materials.
class DiffuseComponent : public MaterialComponent {
public:
  color diffuse_color;

  // Determine the color produced by this component.
  virtual void determineColor(color hit_color, vector world_point,
			      uint intersection_flags,
			      vector normal, vector ray_dir,
			      struct PrimitiveNode *the_node,
			      RayTracer *the_engine,
			      struct PrimitiveNode *world_scene,
			      uint current_ray_flags,
			      color ray_contribution,
			      Material *from_material,
			      Material *to_material,
			      struct _IntersectInfo *int_info,
			      int max_ray_depth,
			      MaterialCalc *mat_calc,
			      class Material *mat);

  // No transmissive portion to this component.
  virtual void apply_transmissive(color behind_color) {}

  // Print out the component's properties.
  virtual void debug_component(void);

  // Make a copy of the material component.
  virtual MaterialComponent *copy(void);

  // Create a new DiffuseComponent.
  DiffuseComponent(double r, double g, double b, double a = 0);

  // Create a new DiffuseComponent.
  DiffuseComponent(double r, double g, double b,
		   double er, double eg, double eb, double a = 0);
};

// The specular / transmissive component of Materials.
class SpecTransComponent : public MaterialComponent {
public:
  color specular_color;
  color transmissive_color;
  double refractive_index;

  // Determine the color produced by this component.
  virtual void determineColor(color hit_color, vector world_point,
			      uint intersection_flags,
			      vector normal, vector ray_dir,
			      struct PrimitiveNode *the_node,
			      RayTracer *the_engine,
			      struct PrimitiveNode *world_scene,
			      uint current_ray_flags,
			      color ray_contribution,
			      Material *from_material,
			      Material *to_material,
			      struct _IntersectInfo *int_info,
			      int max_ray_depth,
			      MaterialCalc *mat_calc,
			      class Material *mat);

  // Apply the transmissive portion of this component.
  virtual void apply_transmissive(color behind_color);

  // Print out the component's properties.
  virtual void debug_component(void);

  // Make a copy of the material component.
  virtual MaterialComponent *copy(void);

  // Create a new SpecTransComponent.
  SpecTransComponent(double n,
		     double rs, double gs, double bs,
		     double rt, double gt, double bt);
  SpecTransComponent(double n,
		     double rs,  double gs,  double bs,
		     double ers, double egs, double ebs, double as,
		     double rt,  double gt,  double bt,
		     double ert, double egt, double ebt, double a);
};

// The phong highlight component of Materials.
class PhongComponent : public MaterialComponent {
public:
  color phong_color;
  double phong_power;

  // Determine the color produced by this component.
  virtual void determineColor(color hit_color, vector world_point,
			      uint intersection_flags,
			      vector normal, vector ray_dir,
			      struct PrimitiveNode *the_node,
			      RayTracer *the_engine,
			      struct PrimitiveNode *world_scene,
			      uint current_ray_flags,
			      color ray_contribution,
			      Material *from_material,
			      Material *to_material,
			      struct _IntersectInfo *int_info,
			      int max_ray_depth,
			      MaterialCalc *mat_calc,
			      class Material *mat);

  // No transmissive portion to this component.
  virtual void apply_transmissive(color behind_color) {}

  // Print out the component's properties.
  virtual void debug_component(void);

  // Make a copy of the material component.
  virtual MaterialComponent *copy(void);

  // Create a new PhongComponent.
  PhongComponent(double power, double er, double eg, double eb);
  PhongComponent(double power,
		 double r, double g, double b,
		 double er, double eg, double eb, double a);
};

// The emitted light component of Materials.
class EmissiveComponent : public MaterialComponent {
public:
  color emitted_color;

  // Determine the color produced by this component.
  virtual void determineColor(color hit_color, vector world_point,
			      uint intersection_flags,
			      vector normal, vector ray_dir,
			      struct PrimitiveNode *the_node,
			      RayTracer *the_engine,
			      struct PrimitiveNode *world_scene,
			      uint current_ray_flags,
			      color ray_contribution,
			      Material *from_material,
			      Material *to_material,
			      struct _IntersectInfo *int_info,
			      int max_ray_depth,
			      MaterialCalc *mat_calc,
			      class Material *mat);

  // No transmissive portion to this component.
  virtual void apply_transmissive(color behind_color) {}

  // Print out the component's properties.
  virtual void debug_component(void);

  // Make a copy of the material component.
  virtual MaterialComponent *copy(void);

  // Create a new EmissiveComponent.
  EmissiveComponent(double er, double eg, double eb, double a = 0);
  EmissiveComponent(double r, double g, double b,
		    double er, double eg, double eb, double a = 0);
};

// Routines to retrieve specific components from materials.
inline DiffuseComponent *get_diffuse_component(Material *mat)
{
  return (DiffuseComponent *) mat->find_component(MATERIALCOMPONENT_DIFFUSE);
}

inline SpecTransComponent *get_spectrans_component(Material *mat)
{
  return (SpecTransComponent *)
    mat->find_component(MATERIALCOMPONENT_SPECTRANS);
}

inline PhongComponent *get_phong_component(Material *mat)
{
  return (PhongComponent *) mat->find_component(MATERIALCOMPONENT_PHONG);
}

inline EmissiveComponent *get_emissive_component(Material *mat)
{
  return (EmissiveComponent *) mat->find_component(MATERIALCOMPONENT_EMISSIVE);
}


/*

// DSTmaterial defines an object's optical properties by specifying
// a Diffuse, Specular and Transparent colour, a refractive index and
// a coefficient for the Phong highlight (0 = no highlight).
//
// Rays are fired to each lightsource withing range to determine
// the lighting; this is multiplied by the diffuse colour. The angle
// between each lightsource vector and the reflected-ray vector is used
// to determine the value of the Phong highlight if there is one.
//
// If the refractive index indicates a reflection, a reflection ray
// is fired and its color is multiplied by the object's specular color.
//
// If the refractive index indicates a transmitted (refracted) ray,
// a transmission ray is fired and its color is multiplied by the
// object's transmissive color.
class DSTmaterial : public Material {
public:
  color diffuse, specular, transmissive, phong_color;
  double refr_index, phong_power;

  // Constructor puts in initial values.
  DSTmaterial(double d_r, double d_g, double d_b,
	      double s_r, double s_g, double s_b,
	      double t_r, double t_g, double t_b,
	      double ph_r, double ph_g, double ph_b,
	      double ph_pow, double n)
  {
    Color::assign(diffuse,      d_r, d_g, d_b);
    Color::assign(specular,     s_r, s_g, s_b);
    Color::assign(transmissive, t_r, t_g, t_b);
    Color::assign(phong_color,  ph_r, ph_g, ph_b);
    phong_power = ph_pow;
    refr_index = n;
    mark_for_deletion();
  }
  
  // This is called to make a copy of a DSTmaterial that may be modified
  // by a texture function. Since we don't want to allocate hundreds
  // and then deallocate them all at the end of the frame but rather
  // allocate / deallocate them immediately, we ***don't*** insert them
  // into the deletion-list for deletion at the end of the frame.
  // Instead, they are automagically deleted by the ray-tracing engine
  // once it has determined the colour.
  //
  // DO NOT call this yourself unless you delete the result afterwards,
  // or else they will hang around eating up memory.
  DSTmaterial(DSTmaterial *old_mat)
  {
    Color::set(diffuse, old_mat->diffuse);
    Color::set(specular, old_mat->specular);
    Color::set(transmissive, old_mat->transmissive);
    Color::set(phong_color, old_mat->phong_color);
    refr_index = old_mat->refr_index;
    phong_power = old_mat->phong_power;
  }
  
  // Work out the color seen by a ray that hits this object.
  void determineColor(color hit_color, vector world_point,
		      uint intersection_flags,
		      vector normal, vector ray_dir,
		      struct PrimitiveNode *the_node,
		      RayTracer *the_engine,
		      struct PrimitiveNode *world_scene,
		      uint current_ray_flags,
		      color ray_contribution,
		      Material *from_material,
		      Material *to_material,
		      struct _IntersectInfo *int_info,
		      int max_ray_depth);

  // Apply the transmissive component to find out what light is
  // passed through the material.
  void apply_transmissive(color behind_color) {
    // This function inlined for efficiency.
    Color::mult(transmissive, behind_color, behind_color);
  }

  // Copy the material. Allocates new memory without placing it in
  // the deletion list; see above for details.
  void copy(Material **copyptr) {
    *copyptr = (Material *) new DSTmaterial(this);
  }

  // Print out the material properties for debugging
  void debug_material(char *matname);

  virtual ~DSTmaterial(void) {
    // This would deallocate any data structures dynamically allocated
    // in the constructor, if there were any.
  }

  // Routines that are used to access the DSTmaterial's properties.
  virtual int get_diffuse(color col) {
    Color::set(col, diffuse);
    return TRUE;
  }

  virtual int get_diffuse2(color col) {
    return FALSE;
  }

  virtual int get_specular(color col) {
    Color::set(col, specular);
    return TRUE;
  }

  virtual int get_transmissive(color col) {
    Color::set(col, transmissive);
    return TRUE;
  }

  virtual int get_phong_power( double *power) {
    *power = phong_power;
    return TRUE;
  }

  virtual int get_phong_color(color col) {
    Color::set(col, phong_color);
    return TRUE;
  }

  virtual int get_emissive(color col) {
    return FALSE;
  }

  virtual int get_refractive_index(double *index) {
    *index = refr_index;
    return TRUE;
  }

  // Routines that are used to change the DSTmaterial's properties.
  virtual int set_diffuse(double r, double g, double b) {
    Color::assign(diffuse, r, g, b);
    return TRUE;
  }

  virtual int set_diffuse(color col) {
    Color::set(diffuse, col);
    return TRUE;
  }

  virtual int set_diffuse2(double r, double g, double b) {
    return FALSE;
  }

  virtual int set_diffuse2(color col) {
    return FALSE;
  }

  virtual int set_specular(double r, double g, double b) {
    Color::assign(specular, r, g, b);
    return TRUE;
  }

  virtual int set_specular(color col) {
    Color::set(specular, col);
    return TRUE;
  }

  virtual int set_transmissive(double r, double g, double b) {
    Color::assign(transmissive, r, g, b);
    return TRUE;
  }

  virtual int set_transmissive(color col) {
    Color::set(transmissive, col);
    return TRUE;
  }

  virtual int set_phong_power(double power) {
    phong_power = power;
    return TRUE;
  }

  virtual int set_phong_color(double r, double g, double b) {
    Color::assign(phong_color, r, g, b);
    return TRUE;
  }

  virtual int set_phong_color(color col) {
    Color::set(phong_color, col);
    return TRUE;
  }

  virtual int set_emissive(double r, double g, double b) {
    return FALSE;
  }

  virtual int set_emissive(color col) {
    return FALSE;
  }

  virtual int set_refractive_index(double index) {
    refr_index = index;
    return TRUE;
  }
};

// TwoSidedMaterial has two different sets of diffuse values in it so
// that an object like a TwoSidedPatch can look different from both
// sides. It does not reflect or transmit light, but it may look only
// partially shadowed from underneath, and it has two phong coefficients.
class TwoSidedMaterial : public Material {
public:
  color diffuse1, diffuse2, transmissive, phong_color;
  double phong_power;
  int transmits_light;

  // Constructor puts in initial values.
  TwoSidedMaterial(double d1_r, double d1_g, double d1_b,
		   double d2_r, double d2_g, double d2_b,
		   double t_r, double t_g, double t_b,
		   double ph_r, double ph_g, double ph_b,
		   double ph_pow, int trans_light)
  {
    Color::assign(diffuse1,     d1_r, d1_g, d1_b);
    Color::assign(diffuse2,     d2_r, d2_g, d2_b);
    Color::assign(transmissive, t_r, t_g, t_b);
    Color::assign(phong_color,  ph_r, ph_g, ph_b);
    phong_power = ph_pow;
    transmits_light = trans_light;
    mark_for_deletion();
  }
  
  // This is called to make a copy of a TwoSidedmaterial that may be modified
  // by a texture function. Since we don't want to allocate hundreds
  // and then deallocate them all at the end of the frame but rather
  // allocate / deallocate them immediately, we ***don't*** insert them
  // into the deletion-list for deletion at the end of the frame.
  // Instead, they are automagically deleted by the ray-tracing engine
  // once it has determined the colour.
  //
  // DO NOT call this yourself unless you delete the result afterwards,
  // or else they will hang around eating up memory.
  TwoSidedMaterial(TwoSidedMaterial *old_mat)
  {
    Color::set(diffuse1, old_mat->diffuse1);
    Color::set(diffuse2, old_mat->diffuse2);
    Color::set(transmissive, old_mat->transmissive);
    phong_power = old_mat->phong_power;
    Color::set(phong_color, old_mat->phong_color);
    transmits_light = old_mat->transmits_light;
  }
  
  // Work out the color seen by a ray that hits this object.
  void determineColor(color hit_color, vector world_point,
		      uint intersection_flags,
		      vector normal, vector ray_dir,
		      struct PrimitiveNode *the_node,
		      RayTracer *the_engine,
		      struct PrimitiveNode *world_scene,
		      uint current_ray_flags,
		      color ray_contribution,
		      Material *from_material,
		      Material *to_material,
		      struct _IntersectInfo *int_info,
		      int max_ray_depth);

  // Apply the transmissive component to find out what light is
  // passed through the material.
  void apply_transmissive(color behind_color) {
    // This function inlined for efficiency.
    Color::mult(transmissive, behind_color, behind_color);
  }

  // Copy the material. Allocates new memory without placing it in
  // the deletion list; see above for details.
  void copy(Material **copyptr) {
    *copyptr = (Material *) new TwoSidedMaterial(this);
  }

  // Print out the material properties for debugging
  void debug_material(char *matname);

  virtual ~TwoSidedMaterial(void) {
    // This would deallocate any data structures dynamically allocated
    // in the constructor, if there were any.
  }

  // Routines that are used to access the TwoSidedMaterial's properties.
  virtual int get_diffuse(color col) {
    Color::set(col, diffuse1);
    return TRUE;
  }

  virtual int get_diffuse2(color col) {
    Color::set(col, diffuse2);
    return TRUE;
  }

  virtual int get_specular(color col) {
    return FALSE;
  }

  virtual int get_transmissive(color col) {
    Color::set(col, transmissive);
    return TRUE;
  }

  virtual int get_phong_power(double *power) {
    *power = phong_power;
    return TRUE;
  }

  virtual int get_phong_color(color col) {
    Color::set(col, phong_color);
    return TRUE;
  }

  virtual int get_emissive(color col) {
    return FALSE;
  }

  virtual int get_refractive_index(double *index) {
    return FALSE;
  }

  // Routines that are used to change the TwoSidedMaterial's properties.
  virtual int set_diffuse(double r, double g, double b) {
    Color::assign(diffuse1, r, g, b);
    return TRUE;
  }

  virtual int set_diffuse(color col) {
    Color::set(diffuse1, col);
    return TRUE;
  }

  virtual int set_diffuse2(double r, double g, double b) {
    Color::assign(diffuse2, r, g, b);
    return TRUE;
  }

  virtual int set_diffuse2(color col) {
    Color::set(diffuse2, col);
    return TRUE;
  }

  virtual int set_specular(double r, double g, double b) {
    return FALSE;
  }

  virtual int set_specular(color col) {
    return FALSE;
  }

  virtual int set_transmissive(double r, double g, double b) {
    Color::assign(transmissive, r, g, b);
    return TRUE;
  }

  virtual int set_transmissive(color col) {
    Color::set(transmissive, col);
    return TRUE;
  }

  virtual int set_phong_power(double power) {
    phong_power = power;
    return TRUE;
  }

  virtual int set_phong_color(double r, double g, double b) {
    Color::assign(phong_color, r, g, b);
    return TRUE;
  }

  virtual int set_phong_color(color col) {
    Color::set(phong_color, col);
    return TRUE;
  }

  virtual int set_emissive(double r, double g, double b) {
    return FALSE;
  }

  virtual int set_emissive(color col) {
    return FALSE;
  }

  virtual int set_refractive_index(double index) {
    return FALSE;
  }
};

*/
