/* matrix.h : Matrix manipulation utilities
** 
** The Matrix class described here is a toolkit of matrix functions
** that is used by Mirage. They are accessed as Matrix::function(a, b, c) e.g.
** Matrix::mult(T1, T2, M);
**
** Chris Butcher, 27 November 1996
*/

// Define the matrix storage. Note that a 'matrix' is a different thing to the
// 'Matrix' toolkit class ...
// Matrices are stored as a 3-row, 4-column array of doubles ... doesn't
// represent all transformations on vectors, no support for perspective
// deformations. Indexed as M[r][c].
typedef double matrix[3][4];

// This class should never be instantiated. It is a toolkit only.
class Matrix {
public:

  // Apply a matrix M to a vector v and store the resulting vector in result.
  static void apply(matrix m, vector v, vector result)
  {
    double temp_x, temp_y;

    // We use temporary variables so that if the matrix is applied to a
    // vector and result stored in the same vector, it won't overwrite values
    // prematurely.
    temp_x    = m[0][0] * v[X] + m[0][1] * v[Y] + m[0][2] * v[Z] + m[0][3];
    temp_y    = m[1][0] * v[X] + m[1][1] * v[Y] + m[1][2] * v[Z] + m[1][3];
    result[Z] = m[2][0] * v[X] + m[2][1] * v[Y] + m[2][2] * v[Z] + m[2][3];
    result[Y] = temp_y;
    result[X] = temp_x;
  }    

  // Apply a matrix M to a vector v and store the resulting vector in x,y,z.
  static void apply(matrix m, vector v, double *x, double *y, double *z)
  {
    *x = m[0][0] * v[X] + m[0][1] * v[Y] + m[0][2] * v[Z] + m[0][3];
    *y = m[1][0] * v[X] + m[1][1] * v[Y] + m[1][2] * v[Z] + m[1][3];
    *z = m[2][0] * v[X] + m[2][1] * v[Y] + m[2][2] * v[Z] + m[2][3];
  }    

  // Apply a matrix M to a vector v and store the resulting vector in result.
  // This apply function doesn't use the shift coefficients (m[i][3]).
  static void applynoshift(matrix m, vector v, vector result)
  {
    double temp_x, temp_y;

    // We use temporary variables so that if the matrix is applied to a
    // vector and stored in the same vector, it won't overwrite values
    // prematurely.
    temp_x    = m[0][0] * v[X] + m[0][1] * v[Y] + m[0][2] * v[Z];
    temp_y    = m[1][0] * v[X] + m[1][1] * v[Y] + m[1][2] * v[Z];
    result[Z] = m[2][0] * v[X] + m[2][1] * v[Y] + m[2][2] * v[Z];
    result[Y] = temp_y;
    result[X] = temp_x;
  }    

  // Apply a matrix M to a vector v and store the resulting vector in x,y,z.
  // This apply function doesn't use the shift coefficients (m[i][3]).
  static void applynoshift(matrix m, vector v, double *x, double *y, double *z)
  {
    *x = m[0][0] * v[X] + m[0][1] * v[Y] + m[0][2] * v[Z];
    *y = m[1][0] * v[X] + m[1][1] * v[Y] + m[1][2] * v[Z];
    *z = m[2][0] * v[X] + m[2][1] * v[Y] + m[2][2] * v[Z];
  }    

  // Apply the transpose of a matrix M to a vector v and store the
  // resulting vector in result.
  static void applytranspose(matrix m, vector v, vector result)
  {
    double temp_x, temp_y;

    // We use temporary variables so that if the matrix is applied to a
    // vector and stored in the same vector, it won't overwrite values
    // prematurely.
    temp_x    = m[0][0] * v[X] + m[1][0] * v[Y] + m[2][0] * v[Z];
    temp_y    = m[0][1] * v[X] + m[1][1] * v[Y] + m[2][1] * v[Z];
    result[Z] = m[0][2] * v[X] + m[1][2] * v[Y] + m[2][2] * v[Z];
    result[Y] = temp_y;
    result[X] = temp_x;
  }    

  // Store a unit matrix in the given matrix M.
  static void unit(matrix m)
  {
    static matrix I = {{1.0, 0.0, 0.0, 0.0},
		       {0.0, 1.0, 0.0, 0.0},
		       {0.0, 0.0, 1.0, 0.0}};

    Matrix::set(m, I);
  }

  // Assign values to a matrix.
  static void assign(matrix m,
		     double a00, double a01, double a02, double a03,
		     double a10, double a11, double a12, double a13,
		     double a20, double a21, double a22, double a23)
  {
    m[0][0] = a00;
    m[1][0] = a10;
    m[2][0] = a20;
    m[0][1] = a01;
    m[1][1] = a11;
    m[2][1] = a21;
    m[0][2] = a02;
    m[1][2] = a12;
    m[2][2] = a22;
    m[0][3] = a03;
    m[1][3] = a13;
    m[2][3] = a23;
  }

  // Set matrix a to matrix b.
  static void set(matrix a, matrix b)
  {
    memcpy((void *) a, (void *) b, 12 * sizeof(double));
  }

  // Multiply matrix a by matrix b and store in result.
  // A * B = R ... i.e. A is on the left, B on the right.
  static void mult(matrix a, matrix b, matrix result)
  {
    int row, col, i;
    matrix temp;
    double sum;

    for (row = 0; row < 3; row++) {
      for (col = 0; col < 4; col++) {
	// Since matrices in this representation have no
	// bottom row, we assume it to be (0, 0, 0, 1).
	// This means that the bottom row's contribution is
	// zero to the first three (0, 1, 2) columns of the
	// result, and 1 * a[row][3] for the fourth column.
	if (col == 3)
	  sum = a[row][3];
	else
	  sum = 0.0;

	for (i = 0; i < 3; i++)
	  sum += a[row][i] * b[i][col];
	
	temp[row][col] = sum;
      }
    }

    // Copy the temporary matrix to result.
    // NB this is necessary as otherwise we might overwrite one
    // of the matrices that we're multiplying if a or b == result.
    Matrix::set(result, temp);
  }
};  

// Note that a transform is different from the Transform class.
typedef matrix transform[2];

// A Transform is a toolkit for the transform struct.
// The transform struct has two matrices in it, and is often used for objects
// so that we have a matrix to go from primitive space to worldspace (fore)
// and another to go back to worldspace (aft). This saves us inverting
// matrices whenever we need to do so.
class Transform {
public:
  // Apply the unit transformation.
  inline static void unit(transform a)
  {
    Matrix::unit(a[FORE]);
    Matrix::unit(a[AFT]);
  }

  // Set one transform to another.
  inline static void set(transform a, transform b)
  {
    Matrix::set(a[FORE], b[FORE]);
    Matrix::set(a[AFT],  b[AFT]);
  }

  // Invert a transform.
  inline static void invert(transform t, transform inv)
  {
    Matrix::set(inv[FORE], t[AFT]);
    Matrix::set(inv[AFT],  t[FORE]);
  }

  // The combine for transforms is used a lot everywhere.
  // combine(a, b, c) means that the fore matrix of c (the transform
  // from primitive to world space) is equivalent to a then b. In other
  // words, the primitive has first a done to it, then b. 
  inline static void combine(transform a, transform b, transform result)
  {
    Matrix::mult(b[FORE], a[FORE], result[FORE]);
    Matrix::mult(a[AFT],  b[AFT],  result[AFT]);
  }

  // ********** Following here are functions to rotate, stretch, etc.
  // and are used for composing transformations one step at a time. *********

  // rotate(t, axis, axis, angle) rotates the given transform from one axis
  // to the next axis by the angle specified.
  inline static void rotate(transform t, axis from, axis to, double angle)
  {
    transform rotation;
    double radians, sine, cosine;

    Transform::unit(rotation);

    radians = DEG2RAD(angle);
    sine = sin(radians);
    cosine = cos(radians);

    // Construct the right rotation matrices.
    rotation[FORE][from][from] = cosine;
    rotation[FORE][from][to  ] = -sine;
    rotation[FORE][to  ][from] = +sine;
    rotation[FORE][to  ][to  ] = cosine;
    rotation[AFT][from][from] = cosine;
    rotation[AFT][from][to  ] = +sine;
    rotation[AFT][to  ][from] = -sine;
    rotation[AFT][to  ][to  ] = cosine;
    
    // Apply the constructed rotation matrices
    Transform::combine(t, rotation, t);
  }

  // rotate(t, axis, angle) rotates the transform around the axis
  // by the angle specified.
  inline static void rotate(transform t, axis around, double angle)
  {
    transform rotation;
    int from, to;
    double radians, sine, cosine;
    
    from = (around + 1) % 3;
    to = (around + 2) % 3;

    Transform::unit(rotation);

    radians = DEG2RAD(angle);
    sine = sin(radians);
    cosine = cos(radians);

    // Construct the right rotation matrices
    rotation[FORE][from][from] = cosine;
    rotation[FORE][from][to  ] = -sine;
    rotation[FORE][to  ][from] = +sine;
    rotation[FORE][to  ][to  ] = cosine;
    rotation[AFT][from][from] = cosine;
    rotation[AFT][from][to  ] = +sine;
    rotation[AFT][to  ][from] = -sine;
    rotation[AFT][to  ][to  ] = cosine;

    // Apply the constructed rotation matrices
    Transform::combine(t, rotation, t);
  }

  // rotate(t, forward, up) rotates the given transform so that its X axis
  // points along the forward vector and its Z axis points in the direction of
  // the up vector.
  inline static void rotate(transform t, vector given_fwd, vector given_up)
  {
    transform rotation;
    vector new_fwd, new_up, new_left;
    axis a;

    Vector::set(new_fwd, given_fwd);
    Vector::set(new_up, given_up);
    
    Vector::cross(new_up, new_fwd, new_left);
    Vector::normalise(new_fwd, "Dagnode::rotate(vec, vec): zero fwd vector");
    
    Vector::cross(new_fwd, new_left, new_up);
    Vector::normalise(new_up, "Dagnode::rotate(vec, vec): zero up-vector");
    
    Vector::normalise(new_left, "Dagnode::rotate(vec, vec): zero left-vector");
    
    Transform::unit(rotation);
    
    for (a = X; a <= Z; a = (axis) (a + 1)) {
      rotation[FORE][a][X] = new_fwd[a];
      rotation[FORE][a][Y] = new_left[a];
      rotation[FORE][a][Z] = new_up[a];
      
      rotation[AFT][X][a] = new_fwd[a];
      rotation[AFT][Y][a] = new_left[a];
      rotation[AFT][Z][a] = new_up[a];
    }

    // Apply the constructed rotation matrices
    Transform::combine(t, rotation, t);
  }

  // shift(t, xshift, yshift, zshift) translates the given transform.
  inline static void shift(transform t, double xshift, double yshift,
			   double zshift)
  {
    transform shift;

    Transform::unit(shift);
    
    // Construct the right shift matrices
    shift[FORE][X][3] = xshift;
    shift[FORE][Y][3] = yshift;
    shift[FORE][Z][3] = zshift;
    shift[AFT][X][3] = -xshift;
    shift[AFT][Y][3] = -yshift;
    shift[AFT][Z][3] = -zshift;
    
    // Apply the constructed rotation matrices
    Transform::combine(t, shift, t);
  }

  // shift(t, shift_vec) translates the given transform.
  inline static void shift(transform t, vector shift_vec)
  {
    transform shift;

    Transform::unit(shift);
    
    // Construct the right shift matrices
    shift[FORE][X][3] = shift_vec[X];
    shift[FORE][Y][3] = shift_vec[Y];
    shift[FORE][Z][3] = shift_vec[Z];
    shift[AFT][X][3] = -shift_vec[X];
    shift[AFT][Y][3] = -shift_vec[Y];
    shift[AFT][Z][3] = -shift_vec[Z];
    
    // Apply the constructed rotation matrices
    Transform::combine(t, shift, t);
  }

  // stretch(t, xstretch, ystretch, zstretch) stretches the given transform.
  inline static void stretch(transform t, double xstretch, double ystretch,
			     double zstretch)
  {
    transform stretch;

    Transform::unit(stretch);

    // Check that the reverse transformation exists
    xstretch = CHECK0(xstretch, "Dagnode::stretch: zero X scale");
    ystretch = CHECK0(ystretch, "Dagnode::stretch: zero Y scale");
    zstretch = CHECK0(zstretch, "Dagnode::stretch: zero Z scale");
    
    // Construct the right shift matrices
    stretch[FORE][X][X] = xstretch;
    stretch[FORE][Y][Y] = ystretch;
    stretch[FORE][Z][Z] = zstretch;
    stretch[AFT][X][X] = 1.0 / xstretch;
    stretch[AFT][Y][Y] = 1.0 / ystretch;
    stretch[AFT][Z][Z] = 1.0 / zstretch;
    
    // Apply the constructed rotation matrices
    Transform::combine(t, stretch, t);
  }

  // scale(t, scale_factor) scales the given transform.
  inline static void scale(transform t, double scale_factor)
  {
    transform scale;

    Transform::unit(scale);
    
    // Check that the reverse transformation exists
    scale_factor = CHECK0(scale_factor, "Dagnode::scale: zero scalefactor");
    
    // Construct the right shift matrices
    scale[FORE][X][X] = scale_factor;
    scale[FORE][Y][Y] = scale_factor;
    scale[FORE][Z][Z] = scale_factor;
    scale[AFT][X][X] = 1.0 / scale_factor;
    scale[AFT][Y][Y] = 1.0 / scale_factor;
    scale[AFT][Z][Z] = 1.0 / scale_factor;
    
    // Apply the constructed rotation matrices
    Transform::combine(t, scale, t);
  }

  // shear(t, fromaxis, towardsaxis, amount) shears the given transform.
  // The point at distance 1 along fromaxis goes to the same point
  // that is (amount) moved along the towardsaxis.
  // e.g. shear(X, Z, 0.5) takes (1, 0, 0) to (1, 0, 0.5).
  inline static void shear(transform t, axis from, axis to, double amount)
  {
    transform shear;

    Transform::unit(shear);
    
    // Construct the shear matrices
    shear[FORE][to][from] = amount;
    shear[AFT][to][from] = -amount;
    
    // Apply the constructed rotation matrices
    Transform::combine(t, shear, t);
  }

  // Assign values to a transform.
  static inline void assign(transform t,
			    double f00, double f01, double f02, double f03,
			    double f10, double f11, double f12, double f13,
			    double f20, double f21, double f22, double f23,
			    double a00, double a01, double a02, double a03,
			    double a10, double a11, double a12, double a13,
			    double a20, double a21, double a22, double a23)
  {
    t[FORE][0][0] = f00;
    t[FORE][1][0] = f10;
    t[FORE][2][0] = f20;
    t[FORE][0][1] = f01;
    t[FORE][1][1] = f11;
    t[FORE][2][1] = f21;
    t[FORE][0][2] = f02;
    t[FORE][1][2] = f12;
    t[FORE][2][2] = f22;
    t[FORE][0][3] = f03;
    t[FORE][1][3] = f13;
    t[FORE][2][3] = f23;
    t[AFT] [0][0] = a00;
    t[AFT] [1][0] = a10;
    t[AFT] [2][0] = a20;
    t[AFT] [0][1] = a01;
    t[AFT] [1][1] = a11;
    t[AFT] [2][1] = a21;
    t[AFT] [0][2] = a02;
    t[AFT] [1][2] = a12;
    t[AFT] [2][2] = a22;
    t[AFT] [0][3] = a03;
    t[AFT] [1][3] = a13;
    t[AFT] [2][3] = a23;
  }
};




