/* memory.c : Memory management functions.
** 
** This file defines the functions that are used to allocate and deallocate
** memory.
**
** Chris Butcher, 8 January 1997
*/

#include "mirage.h"

#ifdef USE_NAMESPACES
namespace Mirage {
#endif

#define DEBUG 0

// Have we set up the memory manager?
int memory_manager_initialised = FALSE;

// Do we presently have the frame memory allocated?
int frame_memory_initialised = FALSE;

// Flag to say whether newly created objects should be placed into the
// deletion list for deallocation at the end of the frame. This is FALSE
// before the animation starts, for the user to load in global variables etc.
// and is set to TRUE inside create_animation.
int insert_dynamic_memory_into_deletion_list;

// The structures that hold pointers to all of the blocks that have
// been allocated. RCSG and octree nodes will be created and destroyed
// on the fly so their control structures are quite complex. Similarly for
// intersection lists.
struct rcsg_storage {
  int blocks_used, items_used, items_freed;
  thing next_free_rcsgnode;
  rcsgnode *storage[MAX_RCSGBLOCKS];
} rcsg_memory;

struct octree_storage {
  int blocks_used, items_used, items_freed;
  thing next_free_octreenode;
  octree *storage[MAX_OCTREEBLOCKS];
} octree_memory;

struct trioctree_storage {
  int blocks_used, items_used, items_freed;
  trioctindex next_free_octreenode;
  trioctree *storage[MAX_OCTREEBLOCKS];
} trioctree_memory;

struct list_storage {
  int blocks_used, items_used, items_freed;
  IntersectionList *next_free_intlist;
  IntersectionList *storage[MAX_LISTBLOCKS];
} list_memory;

struct intersect_storage {
  int blocks_used, items_used, items_freed;
  intersection *next_free_int;
  intersection *storage[MAX_INTBLOCKS];
} intersect_memory;

// These kinds of memory structures never need to be deallocated
// dynamically; thus their control structures are simpler.
struct dag_storage {
  int blocks_used, items_used;
  int used_in_block;
  Dagnode *storage[MAX_DAGBLOCKS];
} dag_memory;

struct primnode_storage {
  int blocks_used, items_used;
  int used_in_block;
  primnode *storage[MAX_PRIMNODEBLOCKS];
} primnode_memory;

struct tex_storage {
  int blocks_used, items_used;
  int used_in_block;
  SingleTexture *storage[MAX_TEXBLOCKS];
} tex_memory;

struct texcol_storage {
  int blocks_used, items_used;
  int used_in_block;
  TextureCollection *storage[MAX_TEXCOLBLOCKS];
} texcol_memory;

struct del_storage {
  int blocks_used, items_used;
  int used_in_block;
  DeletableObject **storage[MAX_DELBLOCKS];
} del_memory;

// The permanent objects go in here; all the dagnodes that aren't dumped at
// the end of the frame.
struct permanent_dag_storage {
  int blocks_used, items_used;
  int used_in_block;
  Dagnode *storage[MAX_DAGBLOCKS];
} permanent_dag_memory;


void init_working_memory(void)
{
  // At the start of the animation the permanent dag memory is zeroed.
  // This is the only time.
  permanent_dag_memory.blocks_used = 0;
  permanent_dag_memory.items_used = 0;

  memory_manager_initialised = TRUE;
}  

void shutdown_working_memory(void)
{
  int i;

  if (!memory_manager_initialised)
    return;

  frame_shutdown_working_memory();

  // Deallocate all of the permanent dagblocks.
  for (i = 0; i < permanent_dag_memory.blocks_used; i++) {
#if DEBUG
    printf("Deallocate permanent dag memory block %d\n", i);
#endif
    free(permanent_dag_memory.storage[i]);
  }
}

void frame_init_working_memory(void)
{
  // At the start of each frame no blocks have been allocated.
  dag_memory.blocks_used = 0;
  primnode_memory.blocks_used = 0;
  tex_memory.blocks_used = 0;
  texcol_memory.blocks_used = 0;
  del_memory.blocks_used = 0;

  dag_memory.items_used = 0;
  primnode_memory.items_used = 0;
  tex_memory.items_used = 0;
  texcol_memory.items_used = 0;
  del_memory.items_used = 0;

  intersect_memory.blocks_used = 0;
  list_memory.blocks_used = 0;
  rcsg_memory.blocks_used = 0;
  octree_memory.blocks_used = 0;
  intersect_memory.items_used = 0;
  list_memory.items_used = 0;
  rcsg_memory.items_used = 0;
  octree_memory.items_used = 0;
  intersect_memory.items_freed = 0;
  list_memory.items_freed = 0;
  rcsg_memory.items_freed = 0;
  octree_memory.items_freed = 0;
  intersect_memory.next_free_int = NO_FREE_INT_MEMORY;
  list_memory.next_free_intlist = NO_FREE_LIST_MEMORY;
  rcsg_memory.next_free_rcsgnode = NO_FREE_MEMORY;
  octree_memory.next_free_octreenode = NO_FREE_MEMORY;

  frame_memory_initialised = TRUE;
}

void frame_shutdown_working_memory(void)
{
  int i, j;
  int bytes_used, bytes_allocated, total_bytes_used, total_bytes_allocated;

  if (!frame_memory_initialised)
    return;
  
  // This deallocates all of the working memory that has been
  // allocated during the current frame.

  // We should print out how much memory was used if the user wants us to.
  if (display_memory_messages && !completely_silent) {
    total_bytes_used = 0;
    total_bytes_allocated = 0;
      
    printf("** Memory manager: deallocating memory for frame %d\n",
	   anim_framenumber);

    bytes_used = permanent_dag_memory.items_used * sizeof(Dagnode);
    bytes_allocated = permanent_dag_memory.blocks_used
      * PERMANENT_DAG_BLOCKSIZE * sizeof(Dagnode);
    total_bytes_used += bytes_used;
    total_bytes_allocated += bytes_allocated;
    
    printf("* STATIC  : %3d blocks, %7d nodes used (%9d / %9d bytes)\n",
	   permanent_dag_memory.blocks_used, permanent_dag_memory.items_used,
	   bytes_used, bytes_allocated);

    bytes_used = dag_memory.items_used * sizeof(Dagnode);
    bytes_allocated = dag_memory.blocks_used * DAG_BLOCKSIZE * sizeof(Dagnode);
    total_bytes_used += bytes_used;
    total_bytes_allocated += bytes_allocated;
    
    printf("* DAGNODE : %3d blocks, %7d items used (%9d / %9d bytes)\n",
	   dag_memory.blocks_used, dag_memory.items_used, bytes_used,
	   bytes_allocated);

    bytes_used = primnode_memory.items_used * sizeof(primnode);
    bytes_allocated = primnode_memory.blocks_used * PRIMNODE_BLOCKSIZE
      * sizeof(primnode);
    total_bytes_used += bytes_used;
    total_bytes_allocated += bytes_allocated;
    
    printf("* PRIMNODE: %3d blocks, %7d items used (%9d / %9d bytes)\n",
	   primnode_memory.blocks_used, primnode_memory.items_used,
	   bytes_used, bytes_allocated);
    
    bytes_used = tex_memory.items_used * sizeof(SingleTexture);
    bytes_allocated = tex_memory.blocks_used * TEX_BLOCKSIZE
      * sizeof(SingleTexture);
    total_bytes_used += bytes_used;
    total_bytes_allocated += bytes_allocated;
    
    printf("* TEXTURE : %3d blocks, %7d items used (%9d / %9d bytes)\n",
	   tex_memory.blocks_used, tex_memory.items_used,
	   bytes_used, bytes_allocated);

    bytes_used = texcol_memory.items_used * sizeof(TextureCollection);
    bytes_allocated = texcol_memory.blocks_used * TEXCOL_BLOCKSIZE
      * sizeof(TextureCollection);
    total_bytes_used += bytes_used;
    total_bytes_allocated += bytes_allocated;
    
    printf("* TEX-COL : %3d blocks, %7d items used (%9d / %9d bytes)\n",
	   tex_memory.blocks_used, tex_memory.items_used,
	   bytes_used, bytes_allocated);

    bytes_used = list_memory.items_used * sizeof(IntersectionList);
    bytes_allocated = list_memory.blocks_used * LIST_BLOCKSIZE
      * sizeof(IntersectionList);
    total_bytes_used += bytes_used;
    total_bytes_allocated += bytes_allocated;
    
    printf("* I-LIST  : %3d blocks, %7d items used (%9d / %9d bytes)\n",
	   list_memory.blocks_used, list_memory.items_used,
	   bytes_used, bytes_allocated);
    printf("*                       %7d items freed(%9d             bytes)\n",
	   list_memory.items_freed,
	   (list_memory.items_freed < (int) (MAXINT/sizeof(IntersectionList)))
	   ? (int) (list_memory.items_freed * sizeof(IntersectionList))
	   : 999999999);

    bytes_used = intersect_memory.items_used * sizeof(intersection);
    bytes_allocated = intersect_memory.blocks_used * INT_BLOCKSIZE
      * sizeof(intersection);
    total_bytes_used += bytes_used;
    total_bytes_allocated += bytes_allocated;
    
    printf("* INTERSECT %3d blocks, %7d items used (%9d / %9d bytes)\n",
	   intersect_memory.blocks_used, intersect_memory.items_used,
	   bytes_used, bytes_allocated);
    printf("*                       %7d items freed(%9d             bytes)\n",
	   intersect_memory.items_freed,
	   (intersect_memory.items_freed < (int) (MAXINT/sizeof(intersection)))
	   ? (int) (intersect_memory.items_freed * sizeof(intersection))
	   : 999999999);

    bytes_used = del_memory.items_used * sizeof(void *);
    bytes_allocated = del_memory.blocks_used * DEL_BLOCKSIZE * sizeof(void *);
    total_bytes_used += bytes_used;
    total_bytes_allocated += bytes_allocated;
    
    printf("* DELETION: %3d blocks, %7d items used (%9d / %9d bytes)\n",
	   del_memory.blocks_used, del_memory.items_used,
	   bytes_used, bytes_allocated);
    
    bytes_used = rcsg_memory.items_used * sizeof(rcsgnode);
    bytes_allocated = rcsg_memory.blocks_used * RCSG_BLOCKSIZE
      * sizeof(rcsgnode);
    total_bytes_used += bytes_used;
    total_bytes_allocated += bytes_allocated;
    
    printf("* RCSGNODE: %3d blocks, %7d items used (%9d / %9d bytes)\n",
	   rcsg_memory.blocks_used, rcsg_memory.items_used,
	   bytes_used, bytes_allocated);
    printf("*                       %7d items freed(%9d             bytes)\n",
	   rcsg_memory.items_freed,
	   (int) (rcsg_memory.items_freed * sizeof(rcsgnode)));

    bytes_used = octree_memory.items_used * sizeof(octree);
    bytes_allocated = octree_memory.blocks_used * OCTREE_BLOCKSIZE
      * sizeof(octree);
    total_bytes_used += bytes_used;
    total_bytes_allocated += bytes_allocated;
    
    printf("* OCTREE  : %3d blocks, %7d items used (%9d / %9d bytes)\n",
	   octree_memory.blocks_used, octree_memory.items_used,
	   bytes_used, bytes_allocated);
    printf("*                       %7d items freed(%9d             bytes)\n",
	   octree_memory.items_freed,
	   (int) (octree_memory.items_freed * sizeof(octree)));

    printf("** Overall memory usage is %4.1f%%:          (%9d / %9d bytes)\n",
	   (total_bytes_allocated == 0) ? 0
	             : (total_bytes_used * 100.0) / total_bytes_allocated,
	   total_bytes_used, total_bytes_allocated);
    
  }

  // NB: We do not deallocate permanent_dag_memory as it persists from frame
  // to frame.
  
  // Deallocate all of the dagblocks.
  for (i = 0; i < dag_memory.blocks_used; i++) {
#if DEBUG
    printf("Deallocate dag memory block %d\n", i);
#endif
    free(dag_memory.storage[i]);
  }
  
  // Deallocate all of the primnodeblocks.
  for (i = 0; i < primnode_memory.blocks_used; i++) {
#if DEBUG
    printf("Deallocate primnode memory block %d\n", i);
#endif
    free(primnode_memory.storage[i]);
  }

  // Deallocate all of the listblocks.
  for (i = 0; i < list_memory.blocks_used; i++) {
#if DEBUG
    printf("Deallocate intersectionlist memory block %d\n", i);
#endif
    free(list_memory.storage[i]);
  }

  // Deallocate all of the intersectionblocks.
  for (i = 0; i < intersect_memory.blocks_used; i++) {
#if DEBUG
    printf("Deallocate intersection memory block %d\n", i);
#endif
    free(intersect_memory.storage[i]);
  }

  // Deallocate all of the rcsgblocks.
  for (i = 0; i < rcsg_memory.blocks_used; i++) {
#if DEBUG
    printf("Deallocate rcsgnode memory block %d\n", i);
#endif
    free(rcsg_memory.storage[i]);
  }

  // Deallocate all of the octree blocks.
  for (i = 0; i < octree_memory.blocks_used; i++) {
#if DEBUG
    printf("Deallocate octree memory block %d\n", i);
#endif
    free(octree_memory.storage[i]);
  }

  // Deallocate all of the texblocks.
  for (i = 0; i < tex_memory.blocks_used; i++) {
#if DEBUG
    printf("Deallocate texture memory block %d\n", i);
#endif
    free(tex_memory.storage[i]);
  }

  // Deallocate all of the texcol blocks.
  for (i = 0; i < texcol_memory.blocks_used; i++) {
#if DEBUG
    printf("Deallocate texcol memory block %d\n", i);
#endif
    free(texcol_memory.storage[i]);
  }

  // Process and deallocate all of the deletionblocks.
  for (i = 0; i < del_memory.blocks_used; i++) {
    if (i != del_memory.blocks_used - 1) {
#if DEBUG
      printf("Process and deallocate full deletionblock %d\n", i);
#endif
      // This block is full. Free all of the things contained
      // in the deletion block (as that's what it's there for).
      for (j = 0; j < DEL_BLOCKSIZE; j++) {
	delete del_memory.storage[i][j];
      }
    } else {
#if DEBUG
      printf("Process and deallocate partial deletionblock %d (%d elements)\n",
	     i, del_memory.used_in_block);
#endif
      // This block isn't necessarily full. Free what elements there are.
      for (j = 0; j < del_memory.used_in_block; j++) {
	delete del_memory.storage[i][j];
      }
    }

    // We've used this deletion block, now free it.
    free(del_memory.storage[i]);
  }
}

Dagnode *new_dag(void)
{
  Dagnode *the_block;

  // If we're in the middle of a frame, allocate a temporary dagnode,
  // otherwise we must allocate a permanent dagnode instead.
  if (!frame_memory_initialised)
    return new_permanent_dag();



  // Check to see if we need to allocate a new dagblock.
  if (dag_memory.blocks_used == 0 ||
      dag_memory.used_in_block >= DAG_BLOCKSIZE) {

    if (dag_memory.blocks_used >= MAX_DAGBLOCKS)
      fatal_error("new_dag", "maximum memory size exceeded",
		  "There is a fixed limit on the number of dagnode blocks\n"
		  "that can be allocated. To change this, alter constants.h\n"
		  "and recompile the system.");

#if DEBUG
    printf("Allocate dag memory block %d\n", dag_memory.blocks_used);
#endif

    the_block = (Dagnode *) calloc(DAG_BLOCKSIZE, sizeof(Dagnode));
    if (the_block == NULL)
      fatal_error("new_dag", "can't allocate new dagnode block!",
		  "The system has tried to dynamically expand its\n"
		  "working memory and has failed. This could indicate\n"
		  "that you are running severely low on memory or that\n"
		  "there is a memory leak in the system.");

    // Update the array of dagnode blocks.
    dag_memory.storage[dag_memory.blocks_used++] = the_block;
    dag_memory.used_in_block = 0;
  } else {
    // Find the current dagnode block.
    the_block = dag_memory.storage[dag_memory.blocks_used - 1];
  }

  dag_memory.items_used++;
  
  return &(the_block[dag_memory.used_in_block++]);
}
    
Dagnode *new_permanent_dag(void)
{
  Dagnode *the_block;

  // Check to see if we need to allocate a new permanent dagblock.
  if (permanent_dag_memory.blocks_used == 0 ||
      permanent_dag_memory.used_in_block >= PERMANENT_DAG_BLOCKSIZE) {

    if (permanent_dag_memory.blocks_used >= MAX_PERMANENTDAGBLOCKS)
      fatal_error("new_permanent_dag", "maximum memory size exceeded",
		  "There is a fixed limit on the number of permanent dagnode\n"
		  "blocks that can be allocated. To change this, alter\n"
		  "constants.h and recompile the system.");

#if DEBUG
    printf("Allocate permanent dag memory block %d\n",
	   permanent_dag_memory.blocks_used);
#endif

    the_block = (Dagnode *) calloc(PERMANENT_DAG_BLOCKSIZE, sizeof(Dagnode));
    if (the_block == NULL)
      fatal_error("new_permanent_dag",
		  "can't allocate new permanent dagnode block!",
		  "The system has tried to dynamically expand its\n"
		  "working memory and has failed. This could indicate\n"
		  "that you are running severely low on memory or that\n"
		  "there is a memory leak in the system.");

    // Update the array of permanent dagnode blocks.
    permanent_dag_memory.storage[permanent_dag_memory.blocks_used++]
      = the_block;
    permanent_dag_memory.used_in_block = 0;
  } else {
    // Find the current permanent dagnode block.
    the_block = permanent_dag_memory.storage[permanent_dag_memory.blocks_used
					    - 1];
  }

  permanent_dag_memory.items_used++;
  
  return &(the_block[permanent_dag_memory.used_in_block++]);
}
    
primnode *new_primnode(thing *new_thing)
{
  primnode *the_block;
  int block_no, index_no;

  // Check to see if we need to allocate a new primnode block.
  if (primnode_memory.blocks_used == 0 ||
      primnode_memory.used_in_block >= PRIMNODE_BLOCKSIZE) {

    if (primnode_memory.blocks_used >= MAX_PRIMNODEBLOCKS)
      fatal_error("new_primnode", "maximum memory size exceeded",
		  "There is a fixed limit on the number of primnode blocks\n"
		  "that can be allocated. To change this, alter constants.h\n"
		  "and recompile the system.");

#if DEBUG
    printf("Allocate primnode memory block %d\n", primnode_memory.blocks_used);
#endif

    the_block = (primnode *) calloc(PRIMNODE_BLOCKSIZE, sizeof(primnode));
    if (the_block == NULL)
      fatal_error("new_primnode", "can't allocate new primnode block!",
		  "The system has tried to dynamically expand its\n"
		  "working memory and has failed. This could indicate\n"
		  "that you are running severely low on memory or that\n"
		  "there is a memory leak in the system.");

    // Update the array of primnode blocks.
    block_no = primnode_memory.blocks_used++;
    index_no = 0;
    primnode_memory.storage[block_no] = the_block;
    primnode_memory.used_in_block = 1;

    *new_thing = (PRIMNODE_BLOCKSIZE * block_no) | THING_PRIMNODE;

    primnode_memory.items_used++;

    return the_block;
  } else {
    // Find the current primnode block.
    block_no = primnode_memory.blocks_used - 1;
    index_no = primnode_memory.used_in_block++;

    *new_thing = (PRIMNODE_BLOCKSIZE * block_no + index_no) | THING_PRIMNODE;

    primnode_memory.items_used++;
    
    return &(primnode_memory.storage[block_no][index_no]);
  }
}
    
SingleTexture *alloc_texture(void)
{
  SingleTexture *the_block;

  // Check to see if we need to allocate a new texturecollection block.
  if (tex_memory.blocks_used == 0 ||
      tex_memory.used_in_block >= TEX_BLOCKSIZE) {

    if (tex_memory.blocks_used >= MAX_TEXBLOCKS)
      fatal_error("alloc_texture", "maximum memory size exceeded",
		  "There is a fixed limit on the number of texture blocks\n"
		  "that can be allocated. To change this, alter constants.h\n"
		  "and recompile the system.");

#if DEBUG
    printf("Allocate texture memory block %d\n", tex_memory.blocks_used);
#endif

    the_block = (SingleTexture *) calloc(TEX_BLOCKSIZE,
					 sizeof(SingleTexture));
    if (the_block == NULL)
      fatal_error("alloc_texture", "can't allocate new texture block",
		  "The system has tried to dynamically expand its\n"
		  "working memory and has failed. This could indicate\n"
		  "that you are running severely low on memory or that\n"
		  "there is a memory leak in the system.");

    // Update the array of texture blocks.
    tex_memory.storage[tex_memory.blocks_used++] = the_block;
    tex_memory.used_in_block = 0;
  } else {
    // Find the current TextureCollection block.
    the_block = tex_memory.storage[tex_memory.blocks_used - 1];
  }

  tex_memory.items_used++;
  
  return &(the_block[tex_memory.used_in_block++]);
}
    
TextureCollection *alloc_textureCollection(void)
{
  TextureCollection *the_block;

  // Check to see if we need to allocate a new texturecollection block.
  if (texcol_memory.blocks_used == 0 ||
      texcol_memory.used_in_block >= TEXCOL_BLOCKSIZE) {

    if (texcol_memory.blocks_used >= MAX_TEXCOLBLOCKS)
      fatal_error("alloc_texturecollection", "maximum memory size exceeded",
		  "There is a fixed limit on the number of texcol blocks\n"
		  "that can be allocated. To change this, alter constants.h\n"
		  "and recompile the system.");

#if DEBUG
    printf("Allocate texcol memory block %d\n", texcol_memory.blocks_used);
#endif

    the_block = (TextureCollection *) calloc(TEXCOL_BLOCKSIZE,
					     sizeof(TextureCollection));
    if (the_block == NULL)
      fatal_error("alloc_texturecollection", "can't allocate new texcol block",
		  "The system has tried to dynamically expand its\n"
		  "working memory and has failed. This could indicate\n"
		  "that you are running severely low on memory or that\n"
		  "there is a memory leak in the system.");

    // Update the array of texcol blocks.
    texcol_memory.storage[texcol_memory.blocks_used++] = the_block;
    texcol_memory.used_in_block = 0;
  } else {
    // Find the current TextureCollection block.
    the_block = texcol_memory.storage[texcol_memory.blocks_used - 1];
  }

  texcol_memory.items_used++;
  
  return &(the_block[texcol_memory.used_in_block++]);
}
    
void DeletableObject::mark_for_deletion(void)
{
  DeletableObject **the_block;

  // If we're not marking all the dynamic memory yet, then don't do anything.
  if (!insert_dynamic_memory_into_deletion_list)
    return;

  // Check to see if we need to allocate a new deletionblock.
  if (del_memory.blocks_used == 0 ||
      del_memory.used_in_block >= DEL_BLOCKSIZE) {

    if (del_memory.blocks_used >= MAX_DELBLOCKS)
      fatal_error("mark_for_deletion", "maximum memory size exceeded",
		  "There is a fixed limit on the number of deletionblocks\n"
		  "that can be allocated. To change this, alter constants.h\n"
		  "and recompile the system.");

#if DEBUG
    printf("Allocate deletion memory block %d\n", del_memory.blocks_used);
#endif

    the_block = (DeletableObject **) calloc(DEL_BLOCKSIZE,
					    sizeof(DeletableObject *));
    if (the_block == NULL)
      fatal_error("mark_for_deletion", "can't allocate new deletionblock!",
		  "The system has tried to dynamically expand its\n"
		  "working memory and has failed. This could indicate\n"
		  "that you are running severely low on memory or that\n"
		  "there is a memory leak in the system.");

    // Update the array of deletionblocks.
    del_memory.storage[del_memory.blocks_used++] = the_block;
    del_memory.used_in_block = 0;
  } else {
    // Find the current deletionblock.
    the_block = del_memory.storage[del_memory.blocks_used - 1];
  }

  the_block[del_memory.used_in_block++] = this;

  del_memory.items_used++;
}
    
// Retrieve primnode number (index) from the working memory store.
primnode *retrieve_primnode(unsigned int index)
{
  int block_no, index_no;

  block_no = index / PRIMNODE_BLOCKSIZE;
  index_no = index % PRIMNODE_BLOCKSIZE;

  if (block_no >= primnode_memory.blocks_used ||
      (block_no == primnode_memory.blocks_used - 1 &&
       index_no >= primnode_memory.used_in_block))
    fatal_error("retrieve_primnode", "requested primnode # outside memory",
		"This error should not happen.");

  return &(primnode_memory.storage[block_no][index_no]);
}

// ***********************************************************************
//                LIST, RCSG AND OCTREE ALLOC/DEALLOC CODE
// ***********************************************************************

octree *new_octree(thing *new_thing)
{
  octree *the_block, *the_node;
  int block_no, index_no;
  thing free_thing;

  // Check to see if we need to allocate a new octree block.
  if (octree_memory.next_free_octreenode == NO_FREE_MEMORY) {

    if (octree_memory.blocks_used >= MAX_OCTREEBLOCKS)
      fatal_error("new_octree", "maximum memory size exceeded",
		  "There is a fixed limit on the number of octree blocks\n"
		  "that can be allocated. To change this, alter constants.h\n"
		  "and recompile the system.");

#if DEBUG
    printf("Allocate octree memory block %d\n", octree_memory.blocks_used);
#endif

    the_block = (octree *) calloc(OCTREE_BLOCKSIZE, sizeof(octree));
    if (the_block == NULL)
      fatal_error("new_octree", "can't allocate new octree block!",
		  "The system has tried to dynamically expand its\n"
		  "working memory and has failed. This could indicate\n"
		  "that you are running severely low on memory or that\n"
		  "there is a memory leak in the system.");


    // Update the array of octree blocks.
    block_no = octree_memory.blocks_used++;
    octree_memory.storage[block_no] = the_block;

    // Now for each unused octree node in the block (which is all of them)
    // we set it to point to the next free octree node. We start from the
    // zero node, each one pointing to the next. The last one in the block
    // points to NO_FREE_MEMORY.
    //
    // This pointer (a thing) is stored in child[0] of the node as there
    // is no data being stored in the octree node right now anyway.

    // Thing pointing to free node 1 of the current block.
    free_thing = (OCTREE_BLOCKSIZE * block_no + 1) | THING_OCTREE;
    octree_memory.next_free_octreenode = free_thing;

    for(index_no = 0; index_no < OCTREE_BLOCKSIZE - 1; index_no++) {
      octree_memory.storage[block_no][index_no].child[0] = free_thing;

      free_thing++;
    }

    octree_memory.storage[block_no][OCTREE_BLOCKSIZE - 1].child[0] =
      NO_FREE_MEMORY;

    *new_thing = (OCTREE_BLOCKSIZE * block_no) | THING_OCTREE;

    octree_memory.items_used++;
    
    return octree_memory.storage[block_no];
  } else {
    *new_thing = octree_memory.next_free_octreenode;

    the_node = GET_OCTREE(octree_memory.next_free_octreenode);

    // The next element in the free list is stored in child[0] of this node.
    octree_memory.next_free_octreenode = the_node->child[0];

    octree_memory.items_used++;
    
    return the_node;
  }
}
    
IntersectionList *new_intersection_list(void)
{
  IntersectionList *the_block, *the_list;
  int block_no, index_no;
  IntersectionList *free_list;

  // Check to see if we need to allocate a new list block.
  if (list_memory.next_free_intlist == NO_FREE_LIST_MEMORY) {

    if (list_memory.blocks_used >= MAX_LISTBLOCKS)
      fatal_error("new_intersection_list", "maximum memory size exceeded",
		  "There is a fixed limit on the number of list blocks\n"
		  "that can be allocated. To change this, alter constants.h\n"
		  "and recompile the system.");

#if DEBUG
    printf("Allocate list memory block %d\n", list_memory.blocks_used);
#endif

    the_block = (IntersectionList *) calloc(LIST_BLOCKSIZE,
					    sizeof(IntersectionList));
    if (the_block == NULL)
      fatal_error("new_intersection_list", "can't allocate new list block!",
		  "The system has tried to dynamically expand its\n"
		  "working memory and has failed. This could indicate\n"
		  "that you are running severely low on memory or that\n"
		  "there is a memory leak in the system.");


    // Update the array of list blocks.
    block_no = list_memory.blocks_used++;
    list_memory.storage[block_no] = the_block;

    // Now for each unused list in the block (which is all of them)
    // we set it to point to the next free intlist. We start from the
    // zero list, each one pointing to the next. The last one in the block
    // points to NO_FREE_MEMORY.
    //
    // This pointer (cast to intersection *) is stored in the head pointer
    // of the list as there is no data being stored in the list right now
    // anyway.

    // The next free node from block 0 is block 1 of the current block.
    free_list = &list_memory.storage[block_no][1];
    list_memory.next_free_intlist = free_list;

    for(index_no = 0; index_no < LIST_BLOCKSIZE - 1; index_no++) {
      list_memory.storage[block_no][index_no].head =
	(intersection *) (free_list);

      free_list++;
    }

    list_memory.storage[block_no][LIST_BLOCKSIZE - 1].head =
      (intersection *) (NO_FREE_LIST_MEMORY);

    list_memory.items_used++;

    // Make certain that the list is marked as empty.
    list_memory.storage[block_no]->head = NULL;
    
    return list_memory.storage[block_no];
  } else {
    the_list = list_memory.next_free_intlist;

    // The next element in the free list is stored in the head pointer
    // of the currently free node. We must cast it from intersection * to
    // IntersectionList *.
    list_memory.next_free_intlist = 
      (IntersectionList *) (the_list->head);

    list_memory.items_used++;
    
    // Make certain that the list is marked as empty.
    the_list->head = NULL;
    
    return the_list;
  }
}
    
intersection *alloc_intersection(void)
{
  intersection *the_block, *the_int;
  int block_no, index_no;
  intersection *free_intersect;

  // Check to see if we need to allocate a new intersection block.
  if (intersect_memory.next_free_int == NO_FREE_INT_MEMORY) {

    if (intersect_memory.blocks_used >= MAX_INTBLOCKS)
      fatal_error("alloc_intersection", "maximum memory size exceeded",
		  "There is a fixed limit on the number of intersect-blocks\n"
		  "that can be allocated. To change this, alter constants.h\n"
		  "and recompile the system.");

#if DEBUG
    printf("Allocate intersection block %d\n", intersect_memory.blocks_used);
#endif

    the_block = (intersection *) calloc(INT_BLOCKSIZE, sizeof(intersection));
    if (the_block == NULL)
      fatal_error("alloc_intersection", "can't allocate new intersect block!",
		  "The system has tried to dynamically expand its\n"
		  "working memory and has failed. This could indicate\n"
		  "that you are running severely low on memory or that\n"
		  "there is a memory leak in the system.");


    // Update the array of intersection blocks.
    block_no = intersect_memory.blocks_used++;
    intersect_memory.storage[block_no] = the_block;

    // Now for each unused intersection in the block (which is all of them)
    // we set it to point to the next free intersect. We start from the
    // zero intersect, each one pointing to the next. The last one in the block
    // points to NO_FREE_INT_MEMORY.
    //
    // This pointer is stored in .next of the intersection as there is no
    // data being stored in it right now anyway.

    // The next free node from block 0 is block 1 of the current block.
    free_intersect = &intersect_memory.storage[block_no][1];
    intersect_memory.next_free_int = free_intersect;

    for(index_no = 0; index_no < INT_BLOCKSIZE - 1; index_no++) {
      intersect_memory.storage[block_no][index_no].next = free_intersect;

      free_intersect++;
    }

    intersect_memory.storage[block_no][INT_BLOCKSIZE - 1].next =
      NO_FREE_INT_MEMORY;

    intersect_memory.items_used++;
    
    return intersect_memory.storage[block_no];
  } else {
    the_int = intersect_memory.next_free_int;

    // The next element in the free list is stored in .next of the
    // currently free node.
    intersect_memory.next_free_int = the_int->next;

    intersect_memory.items_used++;
    
    return the_int;
  }
}
    
intersection *new_intersection(uint hit_flags, primnode *node, vector norm,
			       double t_value,
			       int from_empty, int to_empty,
			       Material *mat_from, Material *mat_to,
			       SfxType sfxtype1, SfxType sfxtype2,
			       void *sfx1, void *sfx2,
			       int multitex1, int multitex2,
			       void *tex1, void *tex2,
			       vector ray_from, vector ray_dir,
			       uint ray_flags, primnode *real_primnode,
			       primnode *validate_primnode)
{
  intersection *the_int = alloc_intersection();

  the_int->hit_flags = hit_flags;
  the_int->primitivehit = node;
  the_int->t_value = t_value;
  Vector::set(the_int->normal, norm);

  the_int->info.from.empty = from_empty;
  the_int->info.from.mat = mat_from;
  the_int->info.from.tex_multi = multitex1;
  if (multitex1)
    the_int->info.from.tex_col = (TextureCollection *) tex1;
  else
    the_int->info.from.texture = (SingleTexture *) tex1;
  the_int->info.from.sfxtype = sfxtype1;
  the_int->info.from.sfx = sfx1;

  the_int->info.to.empty = to_empty;
  the_int->info.to.mat = mat_to;
  the_int->info.to.tex_multi = multitex2;
  if (multitex2)
    the_int->info.to.tex_col = (TextureCollection *) tex2;
  else
    the_int->info.to.texture = (SingleTexture *) tex2;
  the_int->info.to.sfxtype = sfxtype2;
  the_int->info.to.sfx = sfx2;

  the_int->info.normal_flipped = FALSE;

  the_int->real_primnode = real_primnode;
  the_int->validate_primnode = validate_primnode;

  return the_int;
}

thing new_rcsg(thing left, thing right, linktype mymode, int num_prims)
{
  rcsgnode *the_block, *the_node;
  int block_no, index_no;
  thing free_thing;

  // Check to see if we need to allocate a new octree block.
  if (rcsg_memory.next_free_rcsgnode == NO_FREE_MEMORY) {

    if (rcsg_memory.blocks_used >= MAX_RCSGBLOCKS)
      fatal_error("new_rcsg", "maximum memory size exceeded",
		  "There is a fixed limit on the number of rcsg blocks\n"
		  "that can be allocated. To change this, alter constants.h\n"
		  "and recompile the system.");

#if DEBUG
    printf("Allocate rcsg memory block %d\n", rcsg_memory.blocks_used);
#endif

    the_block = (rcsgnode *) calloc(RCSG_BLOCKSIZE, sizeof(rcsgnode));
    if (the_block == NULL)
      fatal_error("new_rcsg", "can't allocate new rcsg block!",
		  "The system has tried to dynamically expand its\n"
		  "working memory and has failed. This could indicate\n"
		  "that you are running severely low on memory or that\n"
		  "there is a memory leak in the system.");


    // Update the array of octree blocks.
    block_no = rcsg_memory.blocks_used++;
    rcsg_memory.storage[block_no] = the_block;

    // Now for each unused rcsg node in the block (which is all of them)
    // we set it to point to the next free rcsg node. We start from the
    // zero node, each one pointing to the next. The last one in the block
    // points to NO_FREE_MEMORY.
    //
    // This pointer (a thing) is stored in left_thing of the node as there
    // is no data being stored in the rcsgnode right now anyway.

    // Thing pointing to free node 1 of the current block.
    free_thing = (RCSG_BLOCKSIZE * block_no + 1) | THING_RCSGNODE;
    rcsg_memory.next_free_rcsgnode = free_thing;

    for(index_no = 0; index_no < RCSG_BLOCKSIZE - 1; index_no++) {
      rcsg_memory.storage[block_no][index_no].left_thing = free_thing;

      free_thing++;
    }

    rcsg_memory.storage[block_no][RCSG_BLOCKSIZE - 1].left_thing =
      NO_FREE_MEMORY;

    rcsg_memory.storage[block_no][0].left_thing = left;
    rcsg_memory.storage[block_no][0].right_thing = right;
    rcsg_memory.storage[block_no][0].linkmode = mymode;
    rcsg_memory.storage[block_no][0].prim_count = num_prims;

    rcsg_memory.items_used++;
    
    return (RCSG_BLOCKSIZE * block_no) | THING_RCSGNODE;
  } else {
    free_thing = rcsg_memory.next_free_rcsgnode;

    the_node = GET_RCSGNODE(free_thing);

    // The next element in the free list is stored in left_thing of this node.
    rcsg_memory.next_free_rcsgnode = the_node->left_thing;

    the_node->left_thing = left;
    the_node->right_thing = right;
    the_node->linkmode = mymode;
    the_node->prim_count = num_prims;

    rcsg_memory.items_used++;
    
    return free_thing;
  }
}

// Free an octree or RCSG tree (recursively...)
// Note that this does not free the primnodes that are the leaves of the tree.
// This is because they are only being referred to, not copied.
void free_tree(thing the_thing)
{
  if (IS_OCTREE(the_thing)) {
    octreenode *the_node = GET_OCTREE(the_thing);
    int i;

    for (i = 0; i < 8; i++)
      free_tree(the_node->child[i]);

    free_octreenode(the_thing);
  } else if (IS_RCSG(the_thing)) {
    rcsgnode *the_node = GET_RCSGNODE(the_thing);
    
    free_tree(the_node->left_thing);
    free_tree(the_node->right_thing);

    free_rcsgnode(the_thing);
  }
}

// Free an intersection list.
void free_intersection_list(IntersectionList *the_list)
{
  // Delete the intersections that are presently in this list.
  the_list->delete_intersections();
  
  // Set the recently freed node to point to the
  // previously first node on the free-memory chain.
  the_list->head = (intersection *) list_memory.next_free_intlist;

  // The next free node is now this recently freed one.
  list_memory.next_free_intlist = the_list;

  list_memory.items_used--;
    
  list_memory.items_freed++;
}

// Free an intersection.
void free_intersection(intersection *the_int)
{
  // Set the recently freed node to point to the
  // previously first node on the free-memory chain.
  the_int->next = intersect_memory.next_free_int;

  // The next free node is now this recently freed one.
  intersect_memory.next_free_int = the_int;

  intersect_memory.items_used--;
    
  intersect_memory.items_freed++;
}

// Free an octree node.
void free_octreenode(thing the_thing)
{
  octree *freed_node;

  freed_node = GET_OCTREE(the_thing);

  // Set the recently freed node to point to the
  // previously first node on the free-memory chain.
  freed_node->child[0] = octree_memory.next_free_octreenode;

  // The next free node is now this recently freed one.
  octree_memory.next_free_octreenode = the_thing;

  octree_memory.items_used--;
    
  octree_memory.items_freed++;
}

// Free an RCSG node.
void free_rcsgnode(thing the_thing)
{
  rcsgnode *freed_node;

  freed_node = GET_RCSGNODE(the_thing);

  // Set the recently freed node to point to the
  // previously first node on the free-memory chain.
  freed_node->left_thing = rcsg_memory.next_free_rcsgnode;

  // The next free node is now this recently freed one.
  rcsg_memory.next_free_rcsgnode = the_thing;

  rcsg_memory.items_used--;
    
  rcsg_memory.items_freed++;
}
    
// Retrieve octree node number (index) from the working memory store.
octree *retrieve_octree(unsigned int index)
{
  int block_no, index_no;

  block_no = index / OCTREE_BLOCKSIZE;
  index_no = index % OCTREE_BLOCKSIZE;

  if (block_no > octree_memory.blocks_used)
    fatal_error("retrieve_octree", "requested octree # outside memory",
		"This error should not happen.");

  return &(octree_memory.storage[block_no][index_no]);
}

// Retrieve rcsgnode number (index) from the working memory store.
rcsgnode *retrieve_rcsgnode(unsigned int index)
{
  int block_no, index_no;

  block_no = index / RCSG_BLOCKSIZE;
  index_no = index % RCSG_BLOCKSIZE;

  if (block_no >= rcsg_memory.blocks_used)
    fatal_error("retrieve_rcsgnode", "requested rcsgnode # outside memory",
		"This error should not happen.");

  return &(rcsg_memory.storage[block_no][index_no]);
}


#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif
