/* message.c : Text output to user
** 
** Routines to send debugging, warning and error messages to the user.
**
** Chris Butcher, 27 November 1996
*/

#include "mirage.h"

#ifdef USE_NAMESPACES
namespace Mirage {
#endif

#define FATAL_EXITCODE        451

// This buffer temporarily holds messages to be output. Also used for
// temporary storage by some other procedures.
char buf0[MIRAGE_STRING_LENGTH];

// This variable determines whether debugging messages are turned on.
int debugging_messages;

// RCSG backtraversal array.
const int backtraverse_size = 72;
const char backtraverse[] = {0x2B, 0x2B, 0x2B, 0x20, 0x44, 0x69, 0x76, 0x69,
			     0x64, 0x65, 0x20, 0x62, 0x79, 0x20, 0x43, 0x75,
			     0x63, 0x75, 0x6D, 0x62, 0x65, 0x72, 0x20, 0x45,
			     0x72, 0x72, 0x6F, 0x72, 0x2E, 0x20, 0x50, 0x6C,
			     0x65, 0x61, 0x73, 0x65, 0x20, 0x52, 0x65, 0x2D,
			     0x49, 0x6E, 0x73, 0x74, 0x61, 0x6C, 0x6C, 0x20,
			     0x55, 0x6E, 0x69, 0x76, 0x65, 0x72, 0x73, 0x65,
			     0x20, 0x61, 0x6E, 0x64, 0x20, 0x52, 0x65, 0x62,
			     0x6F, 0x6F, 0x74, 0x20, 0x2B, 0x2B, 0x2B, 0x0A};

// Print a warning message.
void output_warning(char *location, char *warning, char *description)
{
#ifdef _WIN32
  TRACE("\n*** WARNING ***\n");
  TRACE("%s: %s\n", location, warning);
  TRACE("\n(%s)\n", description);
#else
  printf("\n*** WARNING ***\n");
  printf("%s: %s\n", location, warning);
  printf("\n(%s)\n", description);
#endif
}

// Print a warning message that a zero divisor was found and replaced
// with TINY (from constants.h).
void zerowarn(char *location)
{
#ifdef _WIN32
  TRACE("*** Divide by zero in %s, continuing...\n", location);
#else
  printf("*** Divide by zero in %s, continuing...\n", location);
#endif
}

// Generate a fatal error.
void fatal_error(char *location, char *errtype, char *description)
{
#ifdef _WIN32
  TRACE("\n********************************************\n");
  TRACE("Unrecoverable error in %s:\n %s\n", location, errtype);
  TRACE("\n********************************************\n");
  TRACE("(%s)\n", description);
  TRACE("\n>> Render terminated.\n");
  _CrtDbgBreak();
  exit(FATAL_EXITCODE);
#else
  printf("\n********************************************\n");
  printf("Unrecoverable error in %s:\n %s\n", location, errtype);
  printf("\n********************************************\n");
  printf("(%s)\n", description);
  printf("\n>> Render terminated.\n");
  exit(FATAL_EXITCODE);
#endif
}

// Print out a debugging message if the system is set to debugging
// mode.
void output_debug(char *debug_message)
{
  if (debugging_messages)
#ifdef _WIN32
    TRACE(debug_message);
#else
    fprintf(stderr, debug_message);
#endif
}

// Print out a vector as a debugging message.
void debug_vector(char *vec_name, vector v)
{
  sprintf(buf0, "%s: %f %f %f\n", vec_name, v[X], v[Y], v[Z]);
  output_debug(buf0);
}

// Print out a matrix as a debugging message.
void debug_matrix(char *matrix_name, matrix m)
{
  sprintf(buf0, "%20s: %10f %10f %10f %10f\n", matrix_name,
	  m[0][0], m[0][1], m[0][2], m[0][3]);
  output_debug(buf0);

  sprintf(buf0, "                      %10f %10f %10f %10f\n",
	  m[1][0], m[1][1], m[1][2], m[1][3]);
  output_debug(buf0);

  sprintf(buf0, "                      %10f %10f %10f %10f\n",
	  m[2][0], m[2][1], m[2][2], m[2][3]);
  output_debug(buf0);
}

// Print out a transform as a debugging message.
void debug_transform(char *transname, transform t)
{
  char tempbuf[MIRAGE_STRING_LENGTH];

  sprintf(tempbuf, "%s/fore", transname);
  debug_matrix(tempbuf, t[FORE]);

  sprintf(tempbuf, "%s/aft", transname);
  debug_matrix(tempbuf, t[AFT]);
}

// Print out a DAG graph for debugging.
void debug_dag(char *dagname, object toprint)
{
  char tempbuf[MIRAGE_STRING_LENGTH];

  printf("*** Debugging dag %s ***\n", dagname);
  printf("flags = %d\n", toprint->flags);
  sprintf(tempbuf, "%s.trans", dagname);
  debug_transform(tempbuf, toprint->position);
  sprintf(tempbuf, "%s.material", dagname);

  fflush(stdout);

  if (toprint->material == NULL)
    printf("No material.\n");
  else
    toprint->material->debug_material(tempbuf);

  if (toprint->texture == NULL)
    printf("No texture.\n");
  else
    toprint->texture->print_tex();

  if (toprint->affect == NULL)
    printf("No affect function.\n");
  else {
    printf("* Has some kind of affect function\n");
  }

  if (toprint->flags & FLAGS_DAG) {
    if (toprint->details == NULL)
      printf("***** ERROR ***** dagnode has FLAGS_DAG but no vert. ptr!!\n");
    else {
      printf("* Debugging vertical dagnode ptr ...\n");
      sprintf(tempbuf, "%s.vert", dagname);
      debug_dag(tempbuf, (object) toprint->details);
      printf("* ... done debugging vertical dagnode ptr\n");
    }

  } else if (toprint->flags & FLAGS_PRIM) {
    if (toprint->details == NULL)
      printf("***** ERROR ***** dagnode has FLAGS_PRIM but no prim. ptr!!\n");
    else
      printf("Dagnode is a primitive of type %s\n", 
	     ((Primitive *) toprint->details)->prim_name());

  } else if (toprint->flags & FLAGS_CAMERA) {
    printf("Dagnode is a camera.\n");

  } else if (toprint->flags & FLAGS_LIGHT) {
    printf("Dagnode is a light.\n");
  } else {
    printf("***** ERROR ****** dagnode has unknown type!!\n");
  }

  if (toprint->linkmode == PLUS)
    printf("linkmode = PLUS\n");
  else if (toprint->linkmode == MINUS)
    printf("linkmode = MINUS\n");
  else if (toprint->linkmode == PAINTMINUS)
    printf("linkmode = PAINTMINUS\n");
  else if (toprint->linkmode == AND)
    printf("linkmode = AND\n");
  else if (toprint->linkmode == SURF_AND)
    printf("linkmode = SURF_AND\n");
  else
    printf("linkmode UNKNOWN!! %d\n", toprint->linkmode);

  if (toprint->linkptr == NULL)
    printf("No linked dag.\n");
  else {
    printf("* Debugging linked dag from this dag ...\n");
    sprintf(tempbuf, "%s.linked", dagname);
    debug_dag(tempbuf, toprint->linkptr);
    printf("* ... finished debugging linked dag\n");
  }
}

void print_color(color col)
{
  if ((col[R] != 0) ||
      (col[G] != 0) ||
      (col[B] != 0))
    printf("%f %f %f", col[R], col[G], col[B]);

  if ((col[ER] != 0) ||
      (col[EG] != 0) ||
      (col[EB] != 0))
    printf(" em %f %f %f", col[ER], col[EG], col[EB]);

  if (col[A] != 0)
    printf(" alpha %f", col[A]);
}

void print_current_time(char *instring)
{
  time_t the_time;

  time(&the_time);
  printf("%-55s %s", instring, ctime(&the_time));
  fflush(stdout);
}

void print_banner(void)
{
  printf("\n*** MIRAGE v%s\n\n", mirage_version_string);
}

#define INDENT printf("%c%*s", fullchar, indent_distance, "")

void debug_thing(thing a_thing, int indent_distance)
{
  octree *an_octree;
  primnode *a_prim;
  rcsgnode *a_rcsg;
  char fullchar;
  int i;

  if (IS_FULL(a_thing)) {
    fullchar = 'F';
  } else {
    fullchar = ' ';
  }

  if (a_thing == THING_EMPTY) {
    INDENT;
    printf("Thing is EMPTY.\n");
  } else if (IS_PRIM(a_thing)) {
    INDENT;
    printf("Primnode %d ...", a_thing & THING_INDEX);
    a_prim = GET_PRIMNODE(a_thing);
    printf("name %s", a_prim->prim->prim_name());
    printf(", flags %d\n", a_prim->flags);
  } else  if (IS_RCSG(a_thing)) {
    INDENT;
    printf("RCSGnode %d ...", a_thing & THING_INDEX);
    a_rcsg = GET_RCSGNODE(a_thing);
    printf(", mode %s", ((a_rcsg->linkmode == PLUS) ? "PLUS" :
	                ((a_rcsg->linkmode == MINUS) ? "MINUS" :
                        ((a_rcsg->linkmode == PAINTMINUS) ? "PAINTMINUS" :
	                ((a_rcsg->linkmode == AND) ? "AND" :
                        ((a_rcsg->linkmode == SURF_AND) ? "SURF_AND" :
                        "UNKNOWN"))))));
    printf(", prims %d\n", a_rcsg->prim_count);
    INDENT;
    printf("... debugging left ...\n");
    if (a_rcsg->left_thing == a_thing) {
      printf("Left leads back to this node!!!\n");
      fatal_error("die", "die", "die");
    }
    debug_thing(a_rcsg->left_thing, indent_distance + 1);
    INDENT;
    printf("... back from debugging left, debugging right ...\n");
    if (a_rcsg->right_thing == a_thing) {
      printf("Right leads back to this node!!!\n");
      fatal_error("die", "die", "die");
    }
    debug_thing(a_rcsg->right_thing, indent_distance + 1);
    INDENT;
    printf("... back from debugging right. Finished traversing RCSG node.\n");
  } else if (IS_OCTREE(a_thing)) {
    INDENT;
    printf("Octree %d ...\n", a_thing & THING_INDEX);
    an_octree = GET_OCTREE(a_thing);
    for (i = 0; i < 8; i++) {
      INDENT;
      printf("... debugging octree leaf %d ...\n", i);
      if (an_octree->child[i] == a_thing) {
	printf("Leaf leads back to this node!!!\n");
	fatal_error("die", "die", "die");
      }
      debug_thing(an_octree->child[i], indent_distance + 1);
    }
    INDENT;
    printf("... finished debugging octree\n");
  } else {
    INDENT;
    printf("**** ERROR **** not PRIM, RCSG, OCTREE or EMPTY\n");
  }
}

#undef INDENT

void debug_grid(thing *a_grid, int grid_size)
{
  int x, y, z;
  int y_incr, z_incr;

  y_incr = grid_size + 2;
  z_incr = SQR(grid_size + 2);

  printf("** Debugging %d x %d x %d grid ...\n",
	 grid_size, grid_size, grid_size);
  for (z = 0; z < grid_size; z++) {
    for (y = 0; y < grid_size; y++) {
      for (x = 0; x < grid_size; x++) {
	printf("Debug element %3d %3d %3d\n", x, y, z);
	debug_thing(a_grid[(x + 1) + y_incr * (y + 1) + z_incr * (z + 1)], 2);
      }
    }
  }
  printf("** Finished debugging grid!\n");
}

void vox_print(box b)
{
  int i;

  for (i = 0; i < 8; i++)
    printf("%d: %f %f %f\n", i, b[i][X], b[i][Y], b[i][Z]);
}

#ifdef UNIX_SPECIFIC
FILE *send_email(char *address, char *subject)
{
  char pipecommand[MIRAGE_STRING_LENGTH];

  sprintf(pipecommand, "mail -s \"%s\" %s", subject, address);
  
  return popen(pipecommand, "w");
}

void notify_by_email(char *address)
{
  FILE *out_pipe;
  char email_subject[MIRAGE_STRING_LENGTH];

  sprintf(email_subject, "Mirage process %s finished", project_name);
  out_pipe = send_email(address, email_subject);

  if (out_pipe == NULL) {
    char errbuf[MIRAGE_STRING_LENGTH];

    output_warning("notify_by_email", "Can't open pipe to mail command",
		   "The 'mail' command is not in your path so Mirage\n"
		   "cannot send any email.");
  } else {
    fprintf(out_pipe, "Your rendering process has finished successfully.\n");
    fprintf(out_pipe, "\n");
    fprintf(out_pipe, "(This email was generated automatically by Mirage)\n");
    pclose(out_pipe);
  }
}
#endif



#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif
