/* octree.c : Octree traversal routines
** 
** This file contains functions used by all of the different cluster classes
** to cast rays into an octree.
**
** Chris Butcher, 11 January 1997
*/

#include "mirage.h"

#ifdef USE_NAMESPACES
namespace Mirage {
#endif

// This routine prepares variables for the traversal of the octree structure.
// They are stored in the OctreeClusterPrimitive itself.
int OctreeClusterPrimitive::octree_ray_intersect(
			 primnode *cluster_primnode,
			 primnode *validate_primnode,
			 IntersectionList *above_list, vector ray_from,
			 vector ray_dir, uint ray_flags,
			 signature ray_signature)
{
  double ray_dir_mag;
  double dir_x, dir_y, dir_z;
  int cell_no, octree_found;

  if (IS_EMPTY(my_octree))
    return FALSE;

  octreecluster_primnode = cluster_primnode;
  abovecluster_primnode = validate_primnode;
  the_list = above_list;
  this_ray_flags = ray_flags;
  this_ray_signature = ray_signature;

  dir_x = ray_dir[X];
  dir_y = ray_dir[Y];
  dir_z = ray_dir[Z];
  ray_dir_mag = Vector::magnitude(ray_dir);

  // Find the size of the top-level cell (note that since octrees are always
  // divided at least once this is half of the bound size).
  Vector::mult(0.5, bound_size, cell_size[0]);
  
  // Work out which octant the ray is coming from and what the exit
  // values are of X, Y and Z. Also find the absolute values of the ray's
  // direction vector (if -ve, make it +ve).
  ray_entry_octant = 0;
  if (dir_x < 0) {
    ray_entry_octant |= (1 << X);
    exit_value[X] = minbound[X];
    dir_x = -dir_x;
  } else {
    exit_value[X] = maxbound[X];
  }
  if (dir_y < 0) {
    ray_entry_octant |= (1 << Y);
    exit_value[Y] = minbound[Y];
    dir_y = -dir_y;
  } else {
    exit_value[Y] = maxbound[Y];
  }
  if (dir_z < 0) {
    ray_entry_octant |= (1 << Z);
    exit_value[Z] = minbound[Z];
    dir_z = -dir_z;
  } else {
    exit_value[Z] = maxbound[Z];
  }

  // Work out the amount that the ray must traverse to reach the
  // far edge of the bounding volume.
  exit_value[X] = fabs(ray_from[X] - exit_value[X]);
  exit_value[Y] = fabs(ray_from[Y] - exit_value[Y]);
  exit_value[Z] = fabs(ray_from[Z] - exit_value[Z]);

  // Determine values that are necessary for the cell-skipping algorithm.
  //
  // delta_t_cell_x[0], delta_t_cell_y[0] and delta_t_cell_z[0] are the
  //  t-values the ray takes to travel a top-level cell width in X, Y or Z.
  //  (a real top-level cell, not subdivided by one like cell_width[0]).
  //
  // t_next_x, t_next_y and t_next_z are the t-values when the ray will hit
  //  the edge of the bounding volume.

  if (dir_x == 0) {
    delta_t_cell_x[0] = 0.0;     // Note that this will never be used since
                                 // the ray won't cross X boundaries.
    t_next_x = INFINITY;
    ray_dir_ratio_y_x = INFINITY;
    ray_dir_ratio_z_x = INFINITY;
  } else {
    ray_dir_ratio_y_x = dir_y / dir_x;
    ray_dir_ratio_z_x = dir_z / dir_x;

    delta_t_cell_x[0] = 1.0 / dir_x;        // t for ray to cross 1 X unit
    t_next_x = exit_value[X] * delta_t_cell_x[0];

    if (t_next_x >= INFINITY) {
      delta_t_cell_x[0] = 0.0;
      t_next_x = INFINITY;
    } else {
      // Work out how long for the ray to cross the top-level cell
      delta_t_cell_x[0] *= bound_size[X];
    }
  }

  if (dir_y == 0) {
    delta_t_cell_y[0] = 0.0;     // Note that this will never be used since
                                 // the ray won't cross Y boundaries.
    t_next_y = INFINITY;
    ray_dir_ratio_x_y = INFINITY;
    ray_dir_ratio_z_y = INFINITY;
  } else {
    ray_dir_ratio_x_y = dir_x / dir_y;
    ray_dir_ratio_z_y = dir_z / dir_y;

    delta_t_cell_y[0] = 1.0 / dir_y;        // t for ray to cross 1 Y unit
    t_next_y = exit_value[Y] * delta_t_cell_y[0];

    if (t_next_y >= INFINITY) {
      delta_t_cell_y[0] = 0.0;
      t_next_y = INFINITY;
    } else {
      // Work out how long for the ray to cross the top-level cell
      delta_t_cell_y[0] *= bound_size[Y];
    }
  }

  if (dir_z == 0) {
    delta_t_cell_z[0] = 0.0;     // Note that this will never be used since
                                 // the ray won't cross Z boundaries.
    t_next_z = INFINITY;
    ray_dir_ratio_x_z = INFINITY;
    ray_dir_ratio_y_z = INFINITY;
  } else {
    ray_dir_ratio_x_z = dir_x / dir_z;
    ray_dir_ratio_y_z = dir_y / dir_z;

    delta_t_cell_z[0] = 1.0 / dir_z;        // t for ray to cross 1 Z unit
    t_next_z = exit_value[Z] * delta_t_cell_z[0];

    if (t_next_z >= INFINITY) {
      delta_t_cell_z[0] = 0.0;
      t_next_z = INFINITY;
    } else {
      // Work out how long for the ray to cross the top-level cell
      delta_t_cell_z[0] *= bound_size[Z];
    }
  }

  // We have calculated cell_size, delta_t_cell_x etc for cell size 0
  // (top-level cell). Now calculate them for all the smaller cells.
  for (cell_no = 1; cell_no <= subdivision_limit; cell_no++) {
    delta_t_cell_x[cell_no] = (delta_t_cell_x[cell_no - 1] == INFINITY) ?
      INFINITY : delta_t_cell_x[cell_no - 1] * 0.5;

    delta_t_cell_y[cell_no] = (delta_t_cell_y[cell_no - 1] == INFINITY) ?
      INFINITY : delta_t_cell_y[cell_no - 1] * 0.5;

    delta_t_cell_z[cell_no] = (delta_t_cell_z[cell_no - 1] == INFINITY) ?
      INFINITY : delta_t_cell_z[cell_no - 1] * 0.5;

    Vector::mult(0.5, cell_size[cell_no - 1], cell_size[cell_no]);
  }

  // Traverse the ray through the octree structure.
  octree_found = traverse_octree(0, my_octree, exit_value,
				 t_next_x, t_next_y, t_next_z);

  // Remove intersections that lie outside the cluster's bounds.
  tidy_intersections(above_list, cluster_primnode);

  return octree_found;
}

// Traverse the octree structure recursively, using the cell-skipping
// parameters worked out by octree_ray_intersect.
int OctreeClusterPrimitive::traverse_octree(int current_level,
					    thing the_thing,
					    vector parent_cell_exit,
					    double t_next_x,
					    double t_next_y,
					    double t_next_z)
{
  int found_int;
  intersection *list_position;

  if (IS_PRIM(the_thing)) {
    // This might be a FULL primnode, in which case there will be no
    // intersections in this voxel.
    if (IS_FULL(the_thing))
      return FALSE;

    // Cast a ray into this primnode, storing intersections in the_list.
    // Then validate them, removing any unnecessary and sorting out what
    // material each is from/to.
    primnode *the_primnode = GET_PRIMNODE(the_thing);

    if (check_ray_flags(this_ray_flags, the_primnode->flags)) {
      if (the_primnode->last_ray_signature == this_ray_signature) {
	// We've already intersected this ray with this
	// primitive.
	found_int = the_primnode->last_ray_hit;
      } else {
	// Intersect with the primitive.
	list_position = the_list->get_position();
#if 0
	printf("Octree: intersect w/ %s\n", the_primnode->prim->prim_name());
	debug_thing(the_thing, 0);
	DSTmaterial *tmp_mat = (DSTmaterial *) the_primnode->material;
	printf("Primnode's material colour is %f %f %f\n",
	       tmp_mat->diffuse[R], tmp_mat->diffuse[G], tmp_mat->diffuse[B]);
#endif
	found_int = the_primnode->prim->ray_intersect(the_list, the_primnode,
						      octreecluster_primnode,
						      cluster_ray_from,
						      cluster_ray_dir,
						      this_ray_flags,
						      this_ray_signature);
	//	printf("Octree: finished prim intersect\n");
	the_list->set_position(list_position);
      }

      if (found_int) {
	double cell_exit_t = MIN3(t_next_x, t_next_y, t_next_z);

	return validate_intersections(the_list, octreecluster_primnode,
				      abovecluster_primnode,
				      cluster_ray_from, cluster_ray_dir,
				      this_ray_flags,
				      VALIDATE_PRIMNODE, the_thing,
				      cell_entry_t, cell_exit_t);
      } else {
	return FALSE;
      }
    } else {
      return FALSE;
    }

  } else if (IS_RCSG(the_thing)) {
    // Traverse the RCSG structure, storing intersections in the_list.
    // Then validate them, removing any unnecessary and sorting out what
    // material each is from/to.

    list_position = the_list->get_position();
    found_int = rcsg_ray_intersect(the_list, the_thing,
				   octreecluster_primnode,
				   cluster_ray_from,
				   cluster_ray_dir, this_ray_flags,
				   this_ray_signature);
    the_list->set_position(list_position);

    if (found_int) {
      double cell_exit_t = MIN3(t_next_x, t_next_y, t_next_z);

      return validate_intersections(the_list, octreecluster_primnode,
				    abovecluster_primnode,
				    cluster_ray_from, cluster_ray_dir,
				    this_ray_flags,
				    VALIDATE_RCSG, the_thing,
				    cell_entry_t, cell_exit_t);
    } else
      return FALSE;

  } else if (IS_OCTREE(the_thing)) {
    // Traverse this octree node's children recursively if the ray
    // intersects them.
    unsigned char child_no = 0;
    thing next_child, *children_list;
    vector child_cell_exit;
    int child_result, parent_result = FALSE;

    // Get the size of the child cells.
    double cell_x = cell_size[current_level][X],
           cell_y = cell_size[current_level][Y],
           cell_z = cell_size[current_level][Z];

    current_level++;

    // Get a pointer to the children cells.
    children_list = GET_OCTREE(the_thing)->child;

    // Work out which child of this node the ray starts in.
    Vector::set(child_cell_exit, parent_cell_exit);
    
    if (child_cell_exit[X] > cell_x) {
      // We start in the low-X child cell. This means that the next X crossing
      // won't be at the edge of the parent cell but rather at the edge of
      // the low-X child cell.
      child_cell_exit[X] -= cell_x;
      t_next_x -= delta_t_cell_x[current_level];
    } else {
      // We start in the high-X child cell. The next X crossing is at the
      // edge of the parent cell.
      child_no |= (1 << X);
    }

    if (child_cell_exit[Y] > cell_y) {
      // We start in the low-Y child cell. This means that the next Y crossing
      // won't be at the edge of the parent cell but rather at the edge of
      // the low-Y child cell.
      child_cell_exit[Y] -= cell_y;
      t_next_y -= delta_t_cell_y[current_level];
    } else {
      // We start in the high-Y child cell. The next Y crossing is at the
      // edge of the parent cell.
      child_no |= (1 << Y);
    }

    if (child_cell_exit[Z] > cell_z) {
      // We start in the low-Z child cell. This means that the next Z crossing
      // won't be at the edge of the parent cell but rather at the edge of
      // the low-Z child cell.
      child_cell_exit[Z] -= cell_z;
      t_next_z -= delta_t_cell_z[current_level];
    } else {
      // We start in the high-Z child cell. The next Z crossing is at the
      // edge of the parent cell.
      child_no |= (1 << Z);
    }

    // Now we know which child cell the ray starts in. Get this child and
    // recursively traverse it. Note that all of the above calculations
    // are done with absolute values of the distances, etc. and work if
    // the ray is coming from the low-XYZ octant. If this isn't the case,
    // then we are actually indexing a different child to the one found
    // above; we do this by doing a bitwise XOR with the octant that the
    // ray is coming from.
    next_child = children_list[child_no ^ ray_entry_octant];

    if (!IS_EMPTY(next_child)) {
      child_result = traverse_octree(current_level, next_child,
				     child_cell_exit, t_next_x, t_next_y,
				     t_next_z);

      // If we are stopping at the first intersection and we found
      // something [full shadow or a solid intersection], stop.
      if (child_result) {
	if (stop_at_first_intersection) {
	  return child_result;
	} else {
	  parent_result = TRUE;
	}
      }
    }

    // We know which child the ray starts in and have intersected with it.
    // Traverse the rest of the child cells, stopping once we reach the
    // high-XYZ cell or we hit an edge of the parent cell.
    while (child_no != ((1 << X) | (1 << Y) | (1 << Z))) {
      // Work out which direction we will go in next (X, Y or Z)
      if (t_next_x <= t_next_y) {
	if (t_next_x <= t_next_z) {

	  // We will hit an X boundary next.
	  if (child_no & (1 << X)) {
	    // We're already in a high-X child, so we've hit a boundary
	    // of the parent cell. Exit.
	    break;
	  }

	  // We are now in the high-X child.
	  child_no |= (1 << X);

	  // Set cell_entry_t to be the entry of this child cell.
	  // Set the t-value of hitting the next X boundary up by the X
	  // increment.
	  cell_entry_t = t_next_x;
	  t_next_x += delta_t_cell_x[current_level];

	  // Find the new distances to go until the X, Y and Z boundaries.
	  child_cell_exit[Y] -= child_cell_exit[X] * ray_dir_ratio_y_x;
	  child_cell_exit[Z] -= child_cell_exit[X] * ray_dir_ratio_z_x;
	  child_cell_exit[X] = cell_x;
	} else {

	  // We will hit a Z boundary next.
	  if (child_no & (1 << Z)) {
	    // We're already in a high-Z child, so we've hit a boundary
	    // of the parent cell. Exit.
	    break;
	  }

	  // We are now in the high-Z child.
	  child_no |= (1 << Z);

	  // Set cell_entry_t to be the entry of this child cell.
	  // Set the t-value of hitting the next Z boundary up by the Z
	  // increment.
	  cell_entry_t = t_next_z;
	  t_next_z += delta_t_cell_z[current_level];

	  // Find the new distances to go until the X, Y and Z boundaries.
	  child_cell_exit[X] -= child_cell_exit[Z] * ray_dir_ratio_x_z;
	  child_cell_exit[Y] -= child_cell_exit[Z] * ray_dir_ratio_y_z;
	  child_cell_exit[Z] = cell_z;
	}
      } else {
	if (t_next_y <= t_next_z) {
	  // We will hit a Y boundary next.
	  if (child_no & (1 << Y)) {
	    // We're already in a high-Y child, so we've hit a boundary
	    // of the parent cell. Exit.
	    break;
	  }

	  // We are now in the high-Y child.
	  child_no |= (1 << Y);

	  // Set cell_entry_t to be the entry of this child cell.
	  // Set the t-value of hitting the next Y boundary up by the Y
	  // increment.
	  cell_entry_t = t_next_y;
	  t_next_y += delta_t_cell_y[current_level];

	  // Find the new distances to go until the X, Y and Z boundaries.
	  child_cell_exit[X] -= child_cell_exit[Y] * ray_dir_ratio_x_y;
	  child_cell_exit[Z] -= child_cell_exit[Y] * ray_dir_ratio_z_y;
	  child_cell_exit[Y] = cell_y;
	} else {
	  // We will hit a Z boundary next.
	  if (child_no & (1 << Z)) {
	    // We're already in a high-Z child, so we've hit a boundary
	    // of the parent cell. Exit.
	    break;
	  }

	  // We are now in the high-Z child.
	  child_no |= (1 << Z);

	  // Set cell_entry_t to be the entry of this child cell.
	  // Set the t-value of hitting the next Z boundary up by the Z
	  // increment.
	  cell_entry_t = t_next_z;
	  t_next_z += delta_t_cell_z[current_level];

	  // Find the new distances to go until the X, Y and Z boundaries.
	  child_cell_exit[X] -= child_cell_exit[Z] * ray_dir_ratio_x_z;
	  child_cell_exit[Y] -= child_cell_exit[Z] * ray_dir_ratio_y_z;
	  child_cell_exit[Z] = cell_z;
	}
      }

      // Now we know which child cell the ray will hit next. Get this child and
      // recursively traverse it. Note that all of the above calculations
      // are done with absolute values of the distances, etc. and work if
      // the ray is coming from the low-XYZ octant. If this isn't the case,
      // then we are actually indexing a different child to the one found
      // above; we do this by doing a bitwise XOR with the octant that the
      // ray is coming from.
      next_child = children_list[child_no ^ ray_entry_octant];
      
      if (!IS_EMPTY(next_child)) {
	child_result = traverse_octree(current_level, next_child,
				       child_cell_exit, t_next_x, t_next_y,
				       t_next_z);
	
	// If we are stopping at the first intersection and we found
	// something [full shadow or a solid intersection], stop.
	if (child_result) {
	  if (stop_at_first_intersection) {
	    return child_result;
	  } else {
	    parent_result = TRUE;
	  }
	}
      }
    }
    // We have traversed all of the child cells. This means that either
    // we are set to stop at the first intersection and we didn't find
    // anything, or we are set to continue after intersections and we
    // *may* have found something (will be in parent_result).
    return parent_result;
  } else {
    fatal_error("traverse_octree", "traversing non-PRIM/RCSG/OCTREE thing",
		"This error should not happen.");
  }
  return FALSE;
}

#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif
