/* path.c : Functions for motion paths.
** 
** This file defines the functions for the GeoffSpline class.
**
** Chris Butcher, 6 January 1997
*/

#include "mirage.h"

#ifdef USE_NAMESPACES
namespace Mirage {
#endif

GeoffSpline::GeoffSpline(int num_sp_pts, vector *spline_val, double sp_tens,
			 int num_time_pts, vector *time_val, double ti_tens)
{
  int i;

  spline_pts = num_sp_pts;
  time_pts = num_time_pts;
  spline_tension = sp_tens;
  time_tension = ti_tens;

  spline_point = (vector *) calloc(num_sp_pts+2, sizeof(vector));
  if (spline_point == NULL) {
    char errbuf[MIRAGE_STRING_LENGTH];
    sprintf(errbuf, "can't allocate spline data for %d points", num_sp_pts);
    fatal_error("GeoffSpline:constructor", errbuf,
		"Either the requested spline is huge or you are critically\n"
		"low on memory (indicating a memory leak, possibly).");
  }

  spline_tangent = (vector *) calloc(num_sp_pts+2, sizeof(vector));
  if (spline_tangent == NULL) {
    char errbuf[MIRAGE_STRING_LENGTH];
    sprintf(errbuf, "can't allocate spline tangents for %d pts", num_sp_pts);
    fatal_error("GeoffSpline:constructor", errbuf,
		"Either the requested spline is huge or you are critically\n"
		"low on memory (indicating a memory leak, possibly).");
  }

  time_point = (time_pt *) calloc(num_time_pts+2, sizeof(time_pt));
  if (time_point == NULL) {
    char errbuf[MIRAGE_STRING_LENGTH];
    sprintf(errbuf, "can't allocate %d spline time datapoints", num_time_pts);
    fatal_error("GeoffSpline:constructor", errbuf,
		"Either the requested spline is huge or you are critically\n"
		"low on memory (indicating a memory leak, possibly).");
  }

  time_tangent = (time_pt *) calloc(num_time_pts+2, sizeof(time_pt));
  if (time_tangent == NULL) {
    char errbuf[MIRAGE_STRING_LENGTH];
    sprintf(errbuf, "can't allocate %d time tangent pts", num_time_pts);
    fatal_error("GeoffSpline:constructor", errbuf,
		"Either the requested spline is huge or you are critically\n"
		"low on memory (indicating a memory leak, possibly).");
  }

  for(i = 0; i < num_sp_pts; i++)
    Vector::set(spline_point[i + 1], spline_val[i]);

  for(i = 0; i < num_time_pts; i++) {
    time_point[i + 1][POS]  = time_val[i][POS];
    time_point[i + 1][TIME] = time_val[i][TIME];
  }

  // Set up the tangents now that the spline is constructed. Also construct
  // dummy points at either end of the two splines.
  calculate_tangents();

  // Mark this dynamically allocated object for deletion at the end of the
  // frame.
  mark_for_deletion();
}

GeoffSpline::GeoffSpline(char *filename)
{
  FILE *in_file;
  char *new_fname;

  new_fname = filename_oscheck(filename);
  in_file = fopen(new_fname, "rt");
  if (in_file == NULL) {
    char errbuf[MIRAGE_STRING_LENGTH];
    sprintf(errbuf, "can't open spline file %s", new_fname);
    fatal_error("GeoffSpline:constructor", errbuf,
		"The specified file couldn't be found.");
  }
  free(new_fname);

  read_spline_data(in_file);
  fclose(in_file);

  // Mark this dynamically allocated object for deletion at the end of the
  // frame.
  mark_for_deletion();
}

void GeoffSpline::read_spline_data(FILE *input_file)
{
  int to_find, found_pts, i;
  char inputbuf[MIRAGE_STRING_LENGTH], *tempptr;

  if (fgets(inputbuf, MIRAGE_STRING_LENGTH, input_file) == NULL ||
      strncmp(inputbuf, "GS1", 3) != 0)
    fatal_error("GeoffSpline:read_spline_data", "file has no magic number!",
		"Spline path data files must start with the magic\n"
		"number GS1. This doesn't.");

  to_find = 0;       // must first find number of spline points
  found_pts = 0;
  do {
    // Read a line from the file.
    if (fgets(inputbuf, MIRAGE_STRING_LENGTH, input_file) == NULL)
      fatal_error("GeoffSpline:read_spline_data",
		  "unexpected eof while reading spline",
		  "This file probably does not contain a valid spline path.");
    
    // Strip comments from the input line.
    tempptr = strchr(inputbuf, '#');
    if (tempptr != NULL)
      *tempptr = '\0';

    // Skip blank lines.
    for (i = 0; inputbuf[i] != '\0' &&
	   (inputbuf[i] == ' ' || inputbuf[i] == '\t' || inputbuf[i] == '\n');
	 i++)
      ;

    if (inputbuf[i] == '\0')
      continue;

    // Analyse the input line.
    if (to_find == 0) {
      // Parse the spline's tension.
      if (sscanf(inputbuf, "%lf", &spline_tension) != 1)
	fatal_error("GeoffSpline:read_spline_data",
		    "can't parse the spline's tension",
		    "This file could be corrupt or not a valid spline path.");
      to_find = 1;
    } else if (to_find == 1) {
      // Parse the time-spline's tension.
      if (sscanf(inputbuf, "%lf", &time_tension) != 1)
	fatal_error("GeoffSpline:read_spline_data",
		    "can't parse the time-spline's tension",
		    "This file could be corrupt or not a valid spline path.");
      to_find = 2;
    } else if (to_find == 2) {
      // Parse the number of spline points.
      if (sscanf(inputbuf, "%d", &spline_pts) != 1)
	fatal_error("GeoffSpline:read_spline_data",
		    "can't parse the number of spline points",
		    "This file could be corrupt or not a valid spline path.");

      if (spline_pts < 2)
	fatal_error("GeoffSpline:read_spline_data",
		    "spline with < 2 points",
		    "Splines may only be defined with at least two points.");

      spline_point = (vector *) calloc(spline_pts + 2, sizeof(vector));
      if (spline_point == NULL) {
	char errbuf[MIRAGE_STRING_LENGTH];
	sprintf(errbuf, "can't allocate spline data for %d points",
		spline_pts);
	fatal_error("GeoffSpline:constructor", errbuf,
		    "Either the requested spline is huge or you are\n"
		    "critically low on memory (indicating a memory leak,\n"
		    "possibly).");
      }
      spline_tangent = (vector *) calloc(spline_pts + 2, sizeof(vector));
      if (spline_tangent == NULL) {
	char errbuf[MIRAGE_STRING_LENGTH];
	sprintf(errbuf, "can't allocate spline tangent data for %d points",
		spline_pts);
	fatal_error("GeoffSpline:constructor", errbuf,
		    "Either the requested spline is huge or you are\n"
		    "critically low on memory (indicating a memory leak,\n"
		    "possibly).");
      }

      to_find = 3;
      found_pts = 0;
    } else if (to_find == 3) {
      // Read in the points that form the spline, from 1 to spline_pts. Note
      // that points 0 and spline_pts+1 are dummy points, set up later.
      if (sscanf(inputbuf, "%lf %lf %lf", &spline_point[found_pts+1][X],
		 &spline_point[found_pts+1][Y], &spline_point[found_pts+1][Z])
	  != 3)
	fatal_error("GeoffSpline:read_spline_data",
		    "can't parse the spline points",
		    "This file could be corrupt or not a valid spline path.");
      
      found_pts++;
      if (found_pts == spline_pts)
	to_find = 4;
    } else if (to_find == 4) {
      // Parse the number of time points.
      if (sscanf(inputbuf, "%d", &time_pts) != 1)
	fatal_error("GeoffSpline:read_spline_data",
		    "can't parse the number of time points",
		    "This file could be corrupt or not a valid spline path.");

      if (time_pts < 2)
	fatal_error("GeoffSpline:read_spline_data",
		    "spline with < 2 time points",
		    "Splines must have at least two time points.");

      time_point = (time_pt *) calloc(time_pts + 2, sizeof(time_pt));
      if (time_point == NULL) {
	char errbuf[MIRAGE_STRING_LENGTH];
	sprintf(errbuf, "can't allocate spline time data for %d points",
		time_pts);
	fatal_error("GeoffSpline:constructor", errbuf,
		    "Either the requested spline is huge or you are\n"
		    "critically low on memory (indicating a memory leak,\n"
		    "possibly).");
      }

      time_tangent = (time_pt *) calloc(time_pts + 2, sizeof(time_pt));
      if (time_tangent == NULL) {
	char errbuf[MIRAGE_STRING_LENGTH];
	sprintf(errbuf, "can't allocate time tangent data for %d points",
		time_pts);
	fatal_error("GeoffSpline:constructor", errbuf,
		    "Either the requested spline is huge or you are\n"
		    "critically low on memory (indicating a memory leak,\n"
		    "possibly).");
      }

      to_find = 5;
      found_pts = 0;
    } else if (to_find == 5) {
      // Read in the time points that form the spline. Note that points
      // 0 and time_pts+1 are dummy points, set up later.
      if (sscanf(inputbuf, "%lf %lf", &time_point[found_pts+1][POS],
		 &time_point[found_pts+1][TIME]) != 2)
	fatal_error("GeoffSpline:read_spline_data",
		    "can't parse the time points",
		    "This file could be corrupt or not a valid spline path.");
      
      found_pts++;
      if (found_pts == time_pts) {
	// We have finished reading. Calculate the tangents in preparation
	// for the spline to be used. This also sets up the dummy control
	// points 0 and spline_pts+1, and time points 0 and ti_pts+1.
	calculate_tangents();
	return;
      }
    }
  } while(1);
}

// Calculate the tangents to the spline at every control point, and the
// tangents to the time-spline at every time control point. Also construct
// dummy control points at the start and end of the splines.
void GeoffSpline::calculate_tangents(void)
{
  vector offset, dir1, dir2, temp_tangent;
  time_pt time_offset, timedir1, timedir2, temp_time_tangent;
  int i;
  double length1, length2;

  // Make dummy point 0 for the space spline. It is located beyond point 1
  // in the opposite direction and the same distance as point 2 is.
  Vector::sub(spline_point[1], spline_point[2], offset);
  Vector::add(spline_point[1], offset, spline_point[0]);
  Vector::mult(-spline_tension, offset, spline_tangent[0]);

  // Make dummy point n+1 for the space spline. It is located beyond point n
  // in the opposite direction and the same distance as point n-1 is.
  Vector::sub(spline_point[spline_pts], spline_point[spline_pts - 1], offset);
  Vector::add(spline_point[spline_pts], offset, spline_point[spline_pts + 1]);
  Vector::mult(spline_tension, offset, spline_tangent[spline_pts + 1]);

  // Make dummy point 0 for the time spline. It is located beyond point 1
  // in the opposite direction and the same distance as point 2 is.
  TimePt::sub(time_point[1], time_point[2], time_offset);
  TimePt::add(time_point[1], time_offset, time_point[0]);
  TimePt::mult(-time_tension, time_offset, time_tangent[0]);
  
  // Make dummy point n+1 for the time spline. It is located beyond point n
  // in the opposite direction and the same distance as point n-1 is.
  TimePt::sub(time_point[time_pts], time_point[time_pts - 1], time_offset);
  TimePt::add(time_point[time_pts], time_offset, time_point[time_pts + 1]);
  TimePt::mult(-time_tension, time_offset, time_tangent[time_pts + 1]);

  // Compute the tangent for each point on the space spline.
  for (i = 1; i <= spline_pts; i++) {
    Vector::sub(spline_point[i], spline_point[i - 1], dir1);
    Vector::sub(spline_point[i + 1], spline_point[i], dir2);
    length1 = Vector::magnitude(dir1);
    length2 = Vector::magnitude(dir2);
    
    Vector::mult(length1, dir2, dir2);
    Vector::mult(length2, dir1, dir1);
    
    Vector::add(dir1, dir2, temp_tangent);
    Vector::mult(spline_tension / (length1 + length2), temp_tangent,
		 spline_tangent[i]);
  }

  // Compute the tangent for each point on the time spline.
  for (i = 1; i <= time_pts; i++) {
    TimePt::sub(time_point[i], time_point[i - 1], timedir1);
    TimePt::sub(time_point[i + 1], time_point[i], timedir2);
    length1 = TimePt::magnitude(timedir1);
    length2 = TimePt::magnitude(timedir2);
    
    TimePt::mult(length1, timedir2, timedir2);
    TimePt::mult(length2, timedir1, timedir1);

    TimePt::add(timedir1, timedir2, temp_time_tangent);
    TimePt::mult(time_tension / (length1 + length2), temp_time_tangent,
		 time_tangent[i]);
  }
}

// This function just uses the current time of the scene rather than a
// specified time.
void GeoffSpline::get_position(vector obj_pos)
{
  // No time specified, use the current time.
  get_position(obj_pos, worldtime);
}

// These are the cubic functions and their derivatives used for the
// spline's interpolation.
#define c0(t) ( 2 *(t)*(t)*(t) - 3 *(t)*(t)          + 1)
#define c1(t) (-2 *(t)*(t)*(t) + 3 *(t)*(t)             )
#define c2(t) (    (t)*(t)*(t) - 2 *(t)*(t) +    (t)    )
#define c3(t) (    (t)*(t)*(t) - 1 *(t)*(t)             )

#define d0(t) (                  6 *(t)*(t) - 6 *(t)    )
#define d1(t) (                - 6 *(t)*(t) + 6 *(t)    )
#define d2(t) (                  3 *(t)*(t) - 4 *(t) + 1)
#define d3(t) (                  3 *(t)*(t) - 2 *(t)    )

// Get the position of the object on the path.
void GeoffSpline::get_position(vector obj_pos, double obj_time)
{
  double knot_position;
  int knot_base;

  // Find the object's position in terms of the control knots on the
  // spline, using the associated time-spline.
  knot_position = get_knot_position(obj_time);

  // Check for positions outside the bounds of the spline.
  if (knot_position < 1) {
    Vector::set(obj_pos, spline_point[1]);
    return;
  }
  if (knot_position > spline_pts) {
    Vector::set(obj_pos, spline_point[spline_pts]);
    return;
  }
  
  // Find the knots that the position lies between.
  knot_base = 0;
  while(knot_position > 1) {
    knot_position -= 1.0;
    knot_base++;
  }
  
  // Interpolate the object's position from the positions of the
  // control knots around it.
  Vector::mult   (c0(knot_position), spline_point  [knot_base]    , obj_pos);
  Vector::addfrac(c1(knot_position), spline_point  [knot_base + 1], obj_pos); 
  Vector::addfrac(c2(knot_position), spline_tangent[knot_base]    , obj_pos); 
  Vector::addfrac(c3(knot_position), spline_tangent[knot_base + 1], obj_pos); 
}

// Get the position of the object in terms of the control knots from
// the time-spline.
double GeoffSpline::get_knot_position(double obj_time)
{
  int iterations, knot_base;
  double min_time, max_time, knot_position;
  double interval_lo, interval_hi, interval_sample, t, current_time;

  min_time = time_point[1][TIME];
  max_time = time_point[time_pts][TIME];
  
  // Check for times outside the spline's range.
  if (obj_time <= min_time)
    return time_point[1][POS];
  if (obj_time >= max_time)
    return time_point[time_pts][POS];

  // Make an initial guess as to what place on the time-spline will
  // give the desired time.
  interval_sample = obj_time;
  
  // Iterate until we find a point which is close
  // enough to the correct time.
  iterations = 0;
  interval_lo = min_time;
  interval_hi = max_time;
  do {
    iterations++;

    // Guess where we are likely to find the correct time.
    t = (time_pts - 1) * (interval_sample - min_time) / (max_time - min_time);

    // Determine which knots on the time spline this lies between.
    knot_base = 1;
    while (t > 1.0) {
      t -= 1.0;
      knot_base++;
    }

    current_time = c0(t) * time_point  [knot_base  ][TIME] +
                   c1(t) * time_point  [knot_base+1][TIME] +
                   c2(t) * time_tangent[knot_base  ][TIME] +
                   c3(t) * time_tangent[knot_base+1][TIME];

    if (current_time > obj_time) {
      // The guess is too high, look instead at the lower half of the
      // interval.
      interval_hi = interval_sample;
      interval_sample = (interval_hi + interval_lo) / 2;
    } else {
      // The guess is too low, look instead at the upper half of the
      // interval.
      interval_lo = interval_sample;
      interval_sample = (interval_hi + interval_lo) / 2;
    }
  } while (iterations < SPLINE_MAX_ITERATIONS &&
	   fabs(current_time - obj_time) > SPLINE_TIME_TOLERANCE);

  // If we haven't converged to the desired accuracy, then something
  // is wrong.
  
  if (iterations >= SPLINE_MAX_ITERATIONS) {
    char errbuf[MIRAGE_STRING_LENGTH];
    sprintf(errbuf, "can't converge to time %f (bounds %f-%f)",
	    obj_time, min_time, max_time);
    fatal_error("GeoffSpline:get_knot_position",
		errbuf, "This error should not happen. Maybe\n"
		"SPLINE_MAX_ITERATIONS is too low.");
  }
  
  knot_position = c0(t) * time_point  [knot_base  ][POS] +
                  c1(t) * time_point  [knot_base+1][POS] +
                  c2(t) * time_tangent[knot_base  ][POS] +
                  c3(t) * time_tangent[knot_base+1][POS];

  return knot_position;
}

object GeoffSpline::draw(double marker_spacing, double marker_size,
			 Material *my_material)
{
  object thePath;
  object blob, marker, control;
  double min_time, max_time, current_time, timestep;
  int blob_num;
  vector blob_pos;

  min_time = time_point[1][TIME];
  max_time = time_point[time_pts][TIME];
  timestep = marker_spacing / 4.0;
  
  blob = scale(marker_size, new_sphere());
  blob->material = my_material;

  marker = scale(1.1 * marker_size, new_sphere());
  marker->material = black;

  control = scale(1.2 * marker_size, new_sphere());
  control->material = white;

  blob_num = 0;
  thePath = new_nothing();
  for (current_time = min_time; current_time < max_time;
       current_time += timestep) {
    get_position(blob_pos, current_time);

    if (blob_num % 4 == 0) {
      LINK(thePath, PLUS, shift(blob_pos, bind(marker)));
    } else {
      LINK(thePath, PLUS, shift(blob_pos, bind(blob)));
    }

    blob_num++;
  }

  for (blob_num = 1; blob_num <= time_pts; blob_num++) {
    get_position(blob_pos, time_point[blob_num][TIME]);
    LINK(thePath, PLUS, shift(blob_pos, bind(blob)));
  }

  return thePath->bind();
}

void GeoffSpline::dump(FILE *outfile)
{
  int i;

  fprintf(outfile, "DUMPING GEOFFSPLINE ...\n");
  fprintf(outfile, "Spatial tension: %f\n", spline_tension);
  fprintf(outfile, "Time tension: %f\n", time_tension);
  fprintf(outfile, "%d spatial points ...\n", spline_pts);
  for (i = 0; i <= spline_pts + 1; i++)
    fprintf(outfile, "point %d is %f, %f, %f\n", i,
	    spline_point[i][X], spline_point[i][Y], spline_point[i][Z]);
  fprintf(outfile, "%d time points ...\n", time_pts);
  for (i = 0; i <= time_pts + 1; i++)
    fprintf(outfile, "timepoint %d at %f, time %f.\n", i,
	    time_point[i][POS], time_point[i][TIME]);
}

void GeoffSpline::get_velocity(vector obj_vel)
{
  // No time specified, use the current time.
  get_velocity(obj_vel, worldtime);
}

void GeoffSpline::get_velocity(vector obj_vel, double obj_time)
{
  vector obj_pos, obj_delta_pos, pos_diff;

  // Sample the object's position now and an instant in the future
  // and scale the difference by the appropriate value to get an
  // approximation to the object's velocity.
  get_position(obj_pos, obj_time);
  get_position(obj_delta_pos, obj_time + SPLINE_VEL_DELTA_T);

  Vector::sub(obj_delta_pos, obj_pos, pos_diff);
  Vector::mult(1.0 / SPLINE_VEL_DELTA_T, pos_diff, obj_vel);
}

void GeoffSpline::get_orientation(vector obj_fwd, vector obj_up)
{
  // No time specified, use the current time.
  get_orientation(obj_fwd, obj_up, worldtime);
}

void GeoffSpline::get_orientation(vector obj_fwd, vector obj_up,
				  double obj_time)
{
  vector obj_pos, obj_delta_pos, pos_fwd_diff, pos_back_diff;
  double fwd_time, back_time, interp_fraction;
  int found_fwd, found_back;

  get_position(obj_pos, obj_time);
  get_position(obj_delta_pos, obj_time + SPLINE_VEL_DELTA_T);
  Vector::sub(obj_delta_pos, obj_pos, pos_fwd_diff);

  if (Vector::magnitude(pos_fwd_diff) > SPLINE_MIN_DELTA_X) {
    // The object has moved an appreciable amount over the
    // last instant, so we may use this as an estimate of
    // its facing.
    Vector::normalise(pos_fwd_diff, "GeoffSpline:get_orientation: fwd_diff");
    Vector::set(obj_fwd, pos_fwd_diff);
  } else {
    // The object isn't moving appreciably. We backtrack to
    // see what direction it came in from, and also to
    // see what direction it's leaving. Then interpolate
    // these.

    printf("Doing complicated path stuff\n");

    Vector::assign(pos_fwd_diff, 0, 0, 0);
    fwd_time = obj_time;
    found_fwd = 1;
    do {
      if (fwd_time - obj_time > SPLINE_MAX_LOOKAHEAD) {
	// Conclude that it's not going to move.
	found_fwd = 0;
	break;
      }
      fwd_time += SPLINE_DIR_DELTA_T;
      get_position(obj_delta_pos, fwd_time);
      Vector::sub(obj_delta_pos, obj_pos, pos_fwd_diff);
    } while (Vector::magnitude(pos_fwd_diff) < SPLINE_MIN_DELTA_X &&
	     fwd_time - obj_time < SPLINE_MAX_LOOKAHEAD);

    Vector::assign(pos_back_diff, 0, 0, 0);
    back_time = obj_time;
    found_back = 1;
    do {
      if (obj_time - back_time > SPLINE_MAX_LOOKAHEAD) {
	// Conclude that it's not going to move.
	found_back = 0;
	break;
      }
      back_time -= SPLINE_DIR_DELTA_T;
      get_position(obj_delta_pos, back_time);
      Vector::sub(obj_pos, obj_delta_pos, pos_back_diff);
    } while (Vector::magnitude(pos_back_diff) < SPLINE_MIN_DELTA_X);

    // Calculate the object's facing by interpolating between the
    // direction it comes from and the direction it leaves in.
    // If either of these are zero (it doesn't leave or it starts
    // here) then use the other one. If they are both zero, then
    // the object is completely stationary, use a direction along
    // +X.
    if (found_fwd == 0) {
      if (found_back == 0) {
	Vector::assign(obj_fwd, 1, 0, 0);
      } else {
	Vector::mult(1.0 / Vector::magnitude(pos_back_diff), obj_fwd, obj_fwd);
      }
    } else if (found_back == 0) {
      Vector::mult(1.0 / Vector::magnitude(pos_fwd_diff), obj_fwd, obj_fwd);
    } else {
      // Interpolate between the two directions linearly (not perfect,
      // but it may work) so that the object seems to come to a halt,
      // turn around and then leave again.
      interp_fraction = (obj_time - back_time) / (fwd_time - back_time);
      Vector::mult((1.0 - interp_fraction) / Vector::magnitude(pos_fwd_diff),
		   pos_fwd_diff, obj_fwd);
      Vector::addfrac(interp_fraction / Vector::magnitude(pos_back_diff),
		      pos_back_diff, obj_fwd);
    }
  }

  // Now we have found the forward vector. Find the up-vector.
  // This implementation of the spline uses an up-vector along +Z, rather
  // than one derived from the spline itself.
  Vector::assign(obj_up, 0, 0, 1);
}

void GeoffSpline::distance_along_path(double distance, vector end_pos,
				      double *end_time)
{
  // No start time specified, use the current time.
  distance_along_path(worldtime, distance, end_pos, end_time);
}

// Move distance along the path from the given time.
void GeoffSpline::distance_along_path(double start_time, double distance, 
				      vector end_pos, double *end_time)
{
  double t, moved, step_length;
  double amount_back;
  vector obj_pos, new_pos, move_vec;
  
  // Initially we start at the start point.
  t = start_time;
  moved = 0;
  get_position(obj_pos, start_time);

  // Continue stepping along the path until the total distance
  // exceeds the length specified.
  do {
    t += SPLINE_MOVE_DELTA_T;
    get_position(new_pos, t);

    Vector::sub(new_pos, obj_pos, move_vec);
    step_length = Vector::magnitude(move_vec);

    moved += step_length;

    Vector::set(obj_pos, new_pos);
  } while (moved < distance);

  // Calculate how far we overshot in a fraction of the last step.
  amount_back = (moved - distance) / step_length;
  
  // Apply this compensatory term and store the result in the result
  // fields.
  *end_time = t - amount_back * SPLINE_MOVE_DELTA_T;
  get_position(end_pos, *end_time);
}

double GeoffSpline::distance_moved(double start_time, double end_time)
{
  double t, moved, step_length;
  double amount_back;
  vector obj_pos, new_pos, move_vec;

  if (start_time > end_time)
    return distance_moved(end_time, start_time);

  // Initially we start at the start point.
  t = start_time;
  moved = 0;
  get_position(obj_pos, start_time);

  // Continue stepping along the path until we reach the end time.
  do {
    t += SPLINE_MOVE_DELTA_T;
    get_position(new_pos, t);

    Vector::sub(new_pos, obj_pos, move_vec);
    step_length = Vector::magnitude(move_vec);

    moved += step_length;

    Vector::set(obj_pos, new_pos);
  } while (t < end_time);

  // Calculate how far we overshot in a fraction of the last step.
  amount_back = (t - end_time) / SPLINE_MOVE_DELTA_T;
  
  // Apply this compensatory term and return the distance.
  return moved - amount_back * step_length;
}

// Find the closest point on the path to x.
void GeoffSpline::closest_point(vector x, vector path_point, double path_time)
{
  double interval_lo, interval_hi, interval_step, sample_pt;
  double sample_time, min_time, max_time, distance;
  double closest_pt, closest_dist;
  vector sample_pos, dist_vec;

  // Initial interval is the complete spline.
  interval_lo = 0;
  interval_hi = 1;
  interval_step = 1.0 / SPLINE_CLOSEST_SAMPLES;
  min_time = time_point[1][TIME];
  max_time = time_point[time_pts][TIME];
  closest_pt = 0;
  closest_dist = INFINITY;

  do {
    // Take n samples in the current interval ... find the closest one
    // to x.
    for (sample_pt = interval_lo; sample_pt < interval_hi;
	 sample_pt += interval_step) {
      sample_time = min_time + (max_time - min_time) * sample_pt;
      
      get_position(sample_pos, sample_time);

      // Get the distance from x to this point.
      Vector::sub(x, sample_pos, dist_vec);
      distance = Vector::magnitude(dist_vec);
      
      if (distance < closest_dist) {
	closest_pt = sample_pt;
	closest_dist = distance;
      }
    }

    // Create a new interval five times as small, centered around sample_pt.
    interval_lo = closest_pt - interval_step;
    interval_hi = closest_pt + interval_step;
    interval_step /= 5.0;
  } while (interval_step < SPLINE_CLOSEST_PRECISION);

  // Calculate the position and time of this point on the spline.
  if (closest_pt > 1.0)
    closest_pt = 1;
  else if (closest_pt < 0.0)
    closest_pt = 0;

  path_time = min_time + (max_time - min_time) * closest_pt;
  get_position(path_point, path_time);
} 



#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif
