/* path.h : Header file for the Path class and subclasses.
** 
** This file defines the Path and GeoffSpline classes.
**
** Chris Butcher, 6 January 1997
*/

// The Path class holds a motion path for an object. Various details
// about the object's motion may be obtained from the path, for example
// position, orientation, etc.
class Path : public DeletableObject {
public:
  // The object's position at the current or a given time.
  virtual void get_position(vector obj_pos) = 0;
  virtual void get_position(vector obj_pos, double obj_time) = 0;

  // The orientation of the object.
  virtual void get_orientation(vector obj_fwd, vector obj_up) = 0;
  virtual void get_orientation(vector obj_fwd, vector obj_up,
  			       double obj_time) = 0;

  // The velocity of the object.
  virtual void get_velocity(vector obj_vel) = 0;
  virtual void get_velocity(vector obj_vel, double obj_time) = 0;

  // Move a distance along the path from the current or a given time.
  virtual void distance_along_path(double start_time, double distance,
				   vector end_pos, double *end_time) = 0;
  virtual void distance_along_path(double distance, vector end_pos,
				   double *end_time) = 0;

  // Find the distance moved from one time to another.
  virtual double distance_moved(double start_time, double end_time) = 0;

  // Draw the path as an object.
  virtual object draw(double point_size, double point_spacing,
		      Material *point_material) = 0;

  // Find the closest point on the path to a given position.
  virtual void closest_point(vector given_position,
  			     vector path_point, double path_time) = 0;
};

// Time points on a GeoffSpline are stored as pairs of [point, time] where
// the point value indexes a position on the path in terms of the control
// points.
typedef double time_pt[2];

// This class should never be instantiated. It is a toolkit only, for
// the time_pt structure (like a two-value vector).
class TimePt {
public:

  // Adds two timepoints a and b, stores the result in a third timepoint
  // which may be one of a or b.
  static void add(time_pt a, time_pt b, time_pt result)
  {
    // This function inlined for efficiency.
    result[TIME] = a[TIME] + b[TIME];
    result[POS]  = a[POS]  + b[POS];
  }

  // Subtracts two timepoints a and b, stores the result in a third timepoint
  // which may be one of a or b.
  static void sub(time_pt a, time_pt b, time_pt result)
  {
    // This function inlined for efficiency.
    result[TIME] = a[TIME] - b[TIME];
    result[POS]  = a[POS]  - b[POS];
  }

  // Gets the magnitude of a timepoint.
  static double magnitude(time_pt a)
  {
    // This function inlined for efficiency.
    return sqrt(SQR(a[TIME]) + SQR(a[POS]));
  }

  // Multiplies a timepoint a by a scalar s, stores the result in
  // another timepoint which may be a.
  static void mult(double s, time_pt a, time_pt result)
  {
    // This function inlined for efficiency.
    result[TIME] = s * a[TIME];
    result[POS]  = s * a[POS];
  }
};

// The GeoffSpline subclass is a motion path for an object defined
// by a set of (spline_pts) control points in space and a set
// of (time_pts) control points that specify where the object
// is at given times in terms of the control points. The object's
// orientation is defined as facing along the direction of movement
// with the up-vector pointing to +Z.
class GeoffSpline : public Path {
public:
  int spline_pts;
  int time_pts;
  double spline_tension;
  double time_tension;
  vector *spline_point;
  vector *spline_tangent;
  time_pt *time_point;
  time_pt *time_tangent;

  // Calculate the tangents to the spline and its time-spline. Also
  // construct dummy points 0 and pts+1. Called in the constructors.
  void calculate_tangents(void);

  // The object's position at the current or a given time.
  virtual void get_position(vector obj_pos);
  virtual void get_position(vector obj_pos, double obj_time);

  // Get the position of the object in terms of the control points on the
  // path (governed by the time-spline). Used by get_position.
  double get_knot_position(double obj_time);


  // The orientation of the object.
  virtual void get_orientation(vector obj_fwd, vector obj_up);
  virtual void get_orientation(vector obj_fwd, vector obj_up,
  			       double obj_time);

  // The velocity of the object.
  virtual void get_velocity(vector obj_vel);
  virtual void get_velocity(vector obj_vel, double obj_time);

  // Move a distance along the path from the current or a given time.
  virtual void distance_along_path(double start_time, double distance,
				   vector end_pos, double *end_time);
  virtual void distance_along_path(double distance, vector end_pos,
				   double *end_time);

  // Find the distance moved from one time to another.
  virtual double distance_moved(double start_time, double end_time);

  // Find the closest point on the path to a given position.
  virtual void closest_point(vector given_position,
			     vector path_point, double path_time);

  // Read spline data from the file pointer given. (doesn't close file)
  void read_spline_data(FILE *input_file);

  // Draw the spline as an object in the scene (for testing).
  virtual object draw(double marker_spacing, double marker_size,
		      Material *my_material);

  // Dump the spline to console or a file.
  void dump(FILE *out_file);

  // Read the spline from a file.
  GeoffSpline(char *filename);

  // Create the spline from the points given.
  GeoffSpline(int num_sp_pts, vector *spline_val, double spline_tens,
	      int num_time_pts, vector *time_val, double time_tens);

  virtual ~GeoffSpline() {
    free(spline_point);
    free(spline_tangent);
    free(time_point);
    free(time_tangent);
  }
};



