/* pixmap.c : Functions that read and write pixmaps.
** 
** These functions read in pixmaps from a file (e.g. for texture functions)
** and also write pixmaps to files.
**
** Note that although the system outputs in PPM format, it does not use these
** routines because they can be quite slow.
**
** Chris Butcher, 20 December 1996
*/

#include "mirage.h"

#ifdef USE_NAMESPACES
namespace Mirage {
#endif

GenericPixmap::GenericPixmap(char *filename)
{
#ifdef UNIX_SPECIFIC
  char pipecommand[MIRAGE_STRING_LENGTH];
#endif
  char errbuf[MIRAGE_STRING_LENGTH], magic_char, *new_fname;
  FILE *infile;

  storage = NULL;

  new_fname = filename_oscheck(filename);
  infile = fopen(new_fname, "rb");
  if (infile != NULL) {
    // The file exists. Read it in.
    if (!completely_silent)
      printf("Loading uncompressed pixmap %s ...", new_fname);

  } else {
#ifndef UNIX_SPECIFIC
    // We can't load the .gz version even if it is there because we don't
    // have access to gzip on the Macintosh or under Windows.
    sprintf(errbuf, "can't open pixmap file %s", new_fname);
    fatal_error("GenericPixmap::constructor(filename)", errbuf,
	            "The specified filename couldn't be opened. Its\n"
		        "corresponding .gz compressed file isn't there\n"
	            "either.");
#else
    // The file doesn't exist on disk. Check to see if a .gz version exists.
    sprintf(pipecommand, "%s.gz", new_fname);
    infile = fopen(pipecommand, "rb");
    if (infile == NULL) {
      // The .gz version doesn't exist on disk. Can't find the pixmap.
      sprintf(errbuf, "can't open pixmap file %s", new_fname);
      fatal_error("GenericPixmap::constructor(filename)", errbuf,
		  "The specified filename couldn't be opened. Its\n"
		  "corresponding .gz compressed file isn't there\n"
		  "either.");
    }
    
    // The .gz version exists on disk. Open it using a pipe from 
    // gzcat.
    fclose(infile);
    if (!completely_silent)
      printf("Decompressing and loading pixmap %s ...", pipecommand);
    sprintf(pipecommand, "zcat %s.gz", new_fname);
    infile = popen(pipecommand, "rb");
#endif
  }

  // In order to get to here we must have opened the file successfully.
  // Read the first two characters to determine what the magic number
  // of the file is.
  magic_char = fgetc(infile);
  if (magic_char != 'P') {
    sprintf(errbuf, "%s isn't a PNM file, missing the magic number!",
	    new_fname);
    fatal_error("GenericPixmap::constructor(filename)", errbuf,
		"Any PBM, PGM or PPM file is meant to start with the\n"
		"characters \'Px\' where x is the number of the file.\n"
		"This file doesn't, so it can't be read in.\n");
  }
  magic_char = fgetc(infile);
  if (magic_char < '1' || magic_char > '6') {
    sprintf(errbuf, "%s isn't a PNM file, corrupt magic number!", new_fname);
    fatal_error("GenericPixmap::constructor(filename)", errbuf,
		"The magic number of the \'PNM\' file is corrupt,\n"
		"can't tell what kind of file it is, so it can't be\n"
		"read in.");
  }
  if (magic_char < '4') {
    sprintf(errbuf, "%s is an ASCII PNM file, can't read it!", new_fname);
    fatal_error("GenericPixmap::constructor(filename)", errbuf,
		"This file reader only reads binary PNM files,\n"
		"ASCII PNM files are effectively useless as they're\n"
		"huge for no real reason. Convert it using some\n"
		"image program like XV or GraphicConverter.\n");
  }
  if (magic_char == '4')
    GenericPixmap::read_PBM(infile);
  else if (magic_char == '5')
    GenericPixmap::read_PGM(infile);
  else
    GenericPixmap::read_PPM(infile);

  fclose(infile);
  if (!completely_silent)
    printf("read in %d x %d P%cM.\n", width, height,
	   (type == B_W ? 'B' : (type == GRAYSCALE) ? 'G' : 'P'));

  // This will be deleted at the end of the frame.
  mark_for_deletion();
  
  // Free the sanitized filename that was allocated by filename_oscheck().
  free(new_fname);
}

void GenericPixmap::read_PBM(FILE *in_file)
{
  int bytesused, bytesread;

  type = B_W;

  // Read in the width and height.
  read_values(2, in_file);

  // Stored in a packed binary form, bytes used = (width * height) / 8.
  bytesused = width * height;
  if (bytesused % 8 != 0)
    bytesused = (bytesused / 8) + 1;
  else
    bytesused = bytesused / 8;
  storage = (unsigned char *) calloc(bytesused, sizeof(unsigned char));

  bytesread = fread((void *) storage, sizeof(unsigned char),
		    bytesused, in_file);

  if (bytesread != bytesused)
    fatal_error("GenericPixmap::read_PBM", "unexpected EOF!",
		"This file appears to be truncated unexpectedly.\n"
		"It cannot be read in successfully.");

  // Otherwise, we assume that the file was read correctly.
  // (after all, the correct number of bytes were).
}

void GenericPixmap::read_PGM(FILE *in_file)
{
  int bytesused, bytesread;

  type = GRAYSCALE;

  // Read in the width, height and maxval.
  read_values(3, in_file);

  // Stored in a packed binary form, bytes used = width * height.
  bytesused = width * height;
  storage = (unsigned char *) calloc(bytesused, sizeof(unsigned char));

  bytesread = fread((void *) storage, sizeof(unsigned char),
		    bytesused, in_file);

  if (bytesread != bytesused)
    fatal_error("GenericPixmap::read_PGM", "unexpected EOF!",
		"This file appears to be truncated unexpectedly.\n"
		"It cannot be read in successfully.");

  // Otherwise, we assume that the file was read correctly.
  // (after all, the correct number of bytes were).
}

void GenericPixmap::read_PPM(FILE *in_file)
{
  int bytesused, bytesread;

  type = COLOR;

  // Read in the width, height and maxval.
  read_values(3, in_file);

  // Stored in a packed binary form, bytes used = 3 * width * height.
  bytesused = 3 * width * height;
  storage = (unsigned char *) calloc(bytesused, sizeof(unsigned char));

  bytesread = fread((void *) storage, sizeof(unsigned char),
		    bytesused, in_file);

  if (bytesread != bytesused)
    fatal_error("GenericPixmap::read_PPM", "unexpected EOF!",
		"This file appears to be truncated unexpectedly.\n"
		"It cannot be read in successfully.");

  // Otherwise, we assume that the file was read correctly.
  // (after all, the correct number of bytes were).
}

bwPixmap::bwPixmap(int x, int y)
{
  int bytesused;
  
  width = x;
  height = y;
  type = B_W;
  
  // Stored in a packed binary form, bytes used = (x*y) / 8.
  bytesused = x * y;
  if (bytesused % 8 != 0)
    bytesused = (bytesused / 8) + 1;
  else
    bytesused = bytesused / 8;
  storage = (unsigned char *) calloc(bytesused, sizeof(unsigned char));
  
  if (storage == NULL) {
    char tempbuf[MIRAGE_STRING_LENGTH];
    
    sprintf(tempbuf, "can't allocate space for %dx%d bitmap", x, y);
    fatal_error("bwPixmap::constructor", tempbuf,
		"Either the bitmap you have requested is huge"
		"or you are seriously low on memory. This probably"
		"indicates a memory leak of some sort.");
  }

  mark_for_deletion();
}

GrayscalePixmap::GrayscalePixmap(int x, int y)
{
  int bytesused;
  
  width = x;
  height = y;
  type = GRAYSCALE;
  
  // Stored in a packed binary form, bytes used = x*y.
  bytesused = x * y;
  storage = (unsigned char *) calloc(bytesused, sizeof(unsigned char));
  
  if (storage == NULL) {
    char tempbuf[MIRAGE_STRING_LENGTH];
    
    sprintf(tempbuf, "can't allocate space for %dx%d graymap", x, y);
    fatal_error("GrayscalePixmap::constructor", tempbuf,
		"Either the bitmap you have requested is huge\n"
		"or you are seriously low on memory. This probably\n"
		"indicates a memory leak of some sort.");
  }

  mark_for_deletion();
}

ColorPixmap::ColorPixmap(int x, int y)
{
  int bytesused;
  
  width = x;
  height = y;
  type = COLOR;
  
  // Stored in a packed binary form, bytes used = 3*x*y.
  bytesused = 3 * x * y;
  storage = (unsigned char *) calloc(bytesused, sizeof(unsigned char));
  
  if (storage == NULL) {
    char tempbuf[MIRAGE_STRING_LENGTH];
    
    sprintf(tempbuf, "can't allocate space for %dx%d color pixmap", x, y);
    fatal_error("ColorPixmap::constructor", tempbuf,
		"Either the pixmap you have requested is huge\n"
		"or you are seriously low on memory. This probably\n"
		"indicates a memory leak of some sort.");
  }

  mark_for_deletion();
}

void GenericPixmap::read_values(int num_values, FILE *in_file)
{
  char filebuf[MIRAGE_STRING_LENGTH], *buf_ptr;
  int got_width = FALSE, got_height = FALSE;
  int maxval;

  // Gobble whitespace and comments until we find characters.
  do {
    if (fgets(filebuf, MIRAGE_STRING_LENGTH, in_file) == NULL)
      fatal_error("GenericPixmap::read_values", "no width and height!",
		  "EOF found before width and height were read!");

    // Strip any comments starting from the #.
    buf_ptr = strchr(filebuf, '#');
    if (buf_ptr != NULL)
      *buf_ptr = '\0';

    // Keep looping, reading in numeric values.
    buf_ptr = filebuf;
    do {
      // Look for whitespace on the line.
      while (*buf_ptr == ' ' || *buf_ptr == '\t')
	buf_ptr++;
      
      // We have reached \n, \0 or a character.
      if (*buf_ptr == '\n' || *buf_ptr == '\0')
	// Reached the end of the line, no more values on the line.
	break;

      // We have found a non-whitespace, non-comment character.
      if (got_width) {
	if (got_height) {
	  // Only the maxval remains to get.
	  maxval = atoi(buf_ptr);
	  if (maxval == 0)
	    fatal_error("GenericPixmap::read_values", "corrupt maxval!",
			"The header of this file cannot be read.");
	  return;
	} else {
	  // Get the height.
	  height = atoi(buf_ptr);
	  if (height == 0)
	    fatal_error("GenericPixmap::read_values", "corrupt height!",
			"The header of this file cannot be read.");
	  
	  // Have we finished?
	  if (num_values == 2)
	    return;
	  else {
	    got_height = TRUE;
	    // Advance the pointer to the end of the number.
	    while (*buf_ptr != ' ' && *buf_ptr != '\t' &&
		   *buf_ptr != '\n' && *buf_ptr != '\0')
	      buf_ptr++;
	  }
	}
      } else {
	// Get the width.
	width = atoi(buf_ptr);
	if (width == 0)
	  fatal_error("GenericPixmap::read_values", "corrupt width!",
		      "The header of this file cannot be read.");
	
	got_width = TRUE;
	// Advance the pointer to the end of the number.
	while (*buf_ptr != ' ' && *buf_ptr != '\t' &&
	       *buf_ptr != '\n' && *buf_ptr != '\0')
	  buf_ptr++;
      }
    } while(1);
  } while(1);
  exit(0);
}

void GenericPixmap::write_file(char *filename)
{
  FILE *out_file;
  char errbuf[MIRAGE_STRING_LENGTH], *new_fname;
  int byteswritten, bytesused;

  new_fname = filename_oscheck(filename);

  if (!completely_silent)
    printf("Writing pixmap to %s ... ", new_fname);

  out_file = fopen(new_fname, "wb");
  if (out_file == NULL) {
    sprintf(errbuf, "can't write file %s!", new_fname);
    fatal_error("GenericPixmap::write_file", errbuf,
		"Can't open the file to write an image to it.\n"
		"You may be out of disk space or unauthorized to\n"
		"write to this directory.");
  }

  if (type == B_W) {
    // Stored in a packed binary form, bytes used = (width * height) / 8.
    bytesused = width * height;
    if (bytesused % 8 != 0)
      bytesused = (bytesused / 8) + 1;
    else
      bytesused = bytesused / 8;

    fprintf(out_file, "P4\n");
    fprintf(out_file, "# CREATOR: Mirage (pixmap.c)\n");
    fprintf(out_file, "%d %d\n", width, height);
    byteswritten = fwrite((void *) storage, sizeof(unsigned char),
			  bytesused, out_file);
    if (byteswritten != bytesused) {
      sprintf(errbuf, "couldn't write complete image to %s!", new_fname);
      fatal_error("GenericPixmap::write_file", errbuf,
		  "Couldn't write all of the image to this file;\n"
		  "the file writing operation was unable to write all\n"
		  "of the bytes passed to it. You may be out of disk space.");
    }
  } else if (type == GRAYSCALE) {
    // Stored in a packed binary form, bytes used = (width * height).
    bytesused = width * height;

    fprintf(out_file, "P5\n");
    fprintf(out_file, "# CREATOR: Mirage (pixmap.c)\n");
    fprintf(out_file, "%d %d 255\n", width, height);
    byteswritten = fwrite((void *) storage, sizeof(unsigned char),
			  bytesused, out_file);
    if (byteswritten != bytesused) {
      sprintf(errbuf, "couldn't write complete image to %s!", new_fname);
      fatal_error("GenericPixmap::write_file", errbuf,
		  "Couldn't write all of the image to this file;\n"
		  "the file writing operation was unable to write all\n"
		  "of the bytes passed to it. You may be out of disk space.");
    }
  } else if (type == COLOR) {
    // Stored in a packed binary form, bytes used = 3 * (width * height).
    bytesused = 3 * width * height;

    fprintf(out_file, "P6\n");
    fprintf(out_file, "# CREATOR: Mirage (pixmap.c)\n");
    fprintf(out_file, "%d %d 255\n", width, height);
    byteswritten = fwrite((void *) storage, sizeof(unsigned char),
			  bytesused, out_file);
    if (byteswritten != bytesused) {
      sprintf(errbuf, "couldn't write complete image to %s!", new_fname);
      fatal_error("GenericPixmap::write_file", errbuf,
		  "Couldn't write all of the image to this file;\n"
		  "the file writing operation was unable to write all\n"
		  "of the bytes passed to it. You may be out of disk space.");
    }
  } else {
    fatal_error("GenericPixmap::write_file", "Unknown pixmap type!",
		"Pixmap type is not B_W, GRAYSCALE or COLOR. You may be\n"
		"passing an uninitialised GenericPixmap object to\n"
		"write_file().");
  }
  fclose(out_file);
  if (!completely_silent)
    printf("complete.\n");
    
  // Free the memory allocated by filename_oscheck().
  free(new_fname);
}









#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif
