/* pixmap.h : Definition of the GenericPixmap class.
** 
** Chris Butcher, 20 November 1996
*/

// The GenericPixmap class holds an array of some size which may be indexed to
// return a color (or a grayscale value, or a bitflag).
class GenericPixmap : public DeletableObject {
public:
  int width, height;
  pixmap_type type;
  unsigned char *storage;

  // Read in a pixmap from a file.
  GenericPixmap(char *filename);

  // Write out a pixmap to a file.
  void write_file(char *filename);

  // Set up the width and height to null.
  GenericPixmap(void) {
    width = height = 0;
    storage = NULL;

    // Default type for a pixmap is 24-bit color.
    type = COLOR;

    mark_for_deletion();
  }

  ~GenericPixmap(void) {
    if (storage != NULL)
      free(storage);
  }

  // Read in the various kinds of pixmap from a file (or pipe).
  void read_PBM(FILE *in_file);
  void read_PGM(FILE *in_file);
  void read_PPM(FILE *in_file);
  void read_values(int num_values, FILE *in_file);
};

// The bwPixmap stores a black/white bitmap.
class bwPixmap : public GenericPixmap {
public:
  // Constructor allocates some space for the bitmap.
  // Generally is initialised through GenericPixmap::GenericPixmap("x.pbm").
  bwPixmap(int x, int y);

  ~bwPixmap() {
    // Memory is freed in the GenericPixmap constructor.
  }

  // Retrieve the bit from the bitmap at the given location.
  int get(int x, int y) {
    int offset;
    int bitno;

    // Get the bit number in bits from the start of the image.
    offset = (x + y*width);

    // Get the bit number relative to the byte that we want to
    // access in storage. Note that the bits are stored from left
    // to right across the image -> high bit to low bit inside each
    // byte.
    bitno = 8 - (offset % 8);

    // Get the byte number from the start of the image.
    offset /= 8;

    // Read the bit.
    if (storage[offset] & (1 << bitno))
      return 1;
    else
      return 0;
  }

  // Store a bit in the bitmap at the given location.
  void put(int x, int y, int bit) {
    int offset;
    int bitno;

    // Get the bit number in bits from the start of the image.
    offset = (x + y*width);

    // Get the bit number relative to the byte that we want to
    // access in storage. Note that the bits are stored from left
    // to right across the image -> high bit to low bit inside each
    // byte.
    bitno = 8 - (offset % 8);

    // Get the byte number from the start of the image.
    offset /= 8;

    // Set the bit.
    if (bit)
      storage[offset] |= (1 << bitno);
    else
      storage[offset] &= ~(1 << bitno);
  }
};

// The GrayscalePixmap stores a 256-level grayscale map.
class GrayscalePixmap : public GenericPixmap {
public:
  // Constructor allocates some space for the pixmap.
  // Generally is initialised through GenericPixmap::GenericPixmap("x.pgm").
  GrayscalePixmap(int x, int y);

  ~GrayscalePixmap() {
    // Memory is freed in the GenericPixmap constructor.
  }

  // Retrieve the gray value from the bitmap at the given location.
  unsigned char get(int x, int y) {
    // Byte number from the start of the image = x + y*width.
    return storage[x + y*width];
  }

  // Store a gray value in the bitmap at the given location.
  void put(int x, int y, unsigned char grayval) {
    // Byte number from the start of the image = x + y*width.
    storage[x + y*width] = grayval;
  }

  // Bilinearly interpolate from the bitmap at the given location.
  double get_bilinear(double x, double y) {
    int x_base, y_base;
    int x_off, y_off;
    double x_frac, rev_x_frac, y_frac, rev_y_frac;
    double north_col, south_col;

    x_base = (int) x;
    y_base = (int) y;

    x_off = x_base + 1;
    y_off = y_base + 1;

    if (x_off == width)
      x_off = 0;
    if (y_off == height)
      y_off = 0;

    x_frac = x - x_base;
    y_frac = y - y_base;
    rev_x_frac = 1.0 - x_frac;
    rev_y_frac = 1.0 - y_frac;

    north_col = rev_x_frac*get(x_base, y_base) + x_frac*get(x_off, y_base);
    south_col = rev_x_frac*get(x_base, y_off)  + x_frac*get(x_off, y_off);
    
    return rev_y_frac*north_col + y_frac*south_col;
  }  
};

// The ColorPixmap stores a 256-level RGB pixmap.
class ColorPixmap : public GenericPixmap {
public:
  // Constructor allocates some space for the pixmap.
  // Generally is initialised through GenericPixmap::GenericPixmap("x.ppm").
  ColorPixmap(int x, int y);

  ~ColorPixmap() {
    // Memory is freed in the GenericPixmap constructor.
  }

  // Retrieve the color value from the pixmap at the given location.
  void get(int x, int y, color c) {
    long offset;
    double r, g, b;

    // Byte number from the start of the image = 3*(x + y*width).
    offset = 3 * (x + y*width);

    r = storage[offset++] / 255.0;
    g = storage[offset++] / 255.0;
    b = storage[offset]   / 255.0;

    Color::assign(c, r, g, b);
  }

  // Store a color value in the pixmap at the given location.
  void put(int x, int y, color c) {
    long offset;

    // Byte number from the start of the image = 3*(x + y*width).
    offset = 3 * (x + y*width);

    storage[offset++] = (unsigned char) (255 * c[R]);
    storage[offset++] = (unsigned char) (255 * c[G]);
    storage[offset]   = (unsigned char) (255 * c[B]);
  }

  // Bilinearly interpolate from the pixmap at the given location.
  void get_bilinear(double x, double y, color c) {
    int x_base, y_base;
    int x_off, y_off;
    double x_frac, rev_x_frac, y_frac, rev_y_frac;
    color north_col, south_col, temp1col, temp2col;

    x_base = (int) x;
    y_base = (int) y;

    x_off = x_base + 1;
    y_off = y_base + 1;

    if (x_off == width)
      x_off = 0;
    if (y_off == height)
      y_off = 0;

    x_frac = x - x_base;
    y_frac = y - y_base;
    rev_x_frac = 1.0 - x_frac;
    rev_y_frac = 1.0 - y_frac;

    // Linearly interpolate along the north edge of the pixel square.
    get(x_base, y_base, temp1col);
    get(x_off,  y_base, temp2col);
    Color::mult(rev_x_frac, temp1col, temp1col);
    Color::mult(x_frac,     temp2col, temp2col);
    Color::add(temp1col, temp2col, north_col);

    // Linearly interpolate along the south edge of the pixel square.
    get(x_base, y_off, temp1col);
    get(x_off,  y_off, temp2col);
    Color::mult(rev_x_frac, temp1col, temp1col);
    Color::mult(x_frac,     temp2col, temp2col);
    Color::add(temp1col, temp2col, south_col);

    // Linearly interpolate between the north and south edges
    // of the pixel square.
    Color::mult(rev_y_frac, north_col, north_col);
    Color::mult(y_frac,     south_col, south_col);
    Color::add(north_col, south_col, c);
  }  
};




