/* polynomial.c : Polynomial equation manipulation routines.
** 
** This file defines functions to solve polynomial equations.
**
** Chris Butcher, 13 December 1996
*/

#include "mirage.h"

#ifdef USE_NAMESPACES
namespace Mirage {
#endif

// Macro to test for degenerate cases.
#define POLY_EPSILON   1e-60
#define IS_ZERO(x)     (((x) > -POLY_EPSILON) && ((x) < POLY_EPSILON))


int solveQuadric(double *coeff, double *solutions)
{
  // Solve a quadratic (quadric) equation of the form
  // at^2 + bt + c = 0.
  // Uses the familiar formula t = ( -b +/- sqrt(b^2 - 4ac) ) / 2a.

  double disc, a, b, c, two_a;

  a = coeff[2]; b = coeff[1]; c = coeff[0];

  if (IS_ZERO(a)) {
    // Solve the linear equation bt + c = 0.
    if (IS_ZERO(b)) {
      return 0;
    } else {
      solutions[0] = - c / b;
      return 1;
    }
  }
    
  disc = SQR(b) - 4*a*c;

  // If the discriminant is less than or equal to zero,
  // return no solutions. Note that if the discriminant is
  // precisely zero, there is one solution. However, we
  // ignore it (because floating point precision gets in the way).
  if (disc <= 0.0)
    return 0;

  two_a = CHECK0(2*a, "solveQuadric: zero t^2 coefficient");

  // Solve the equation.
  if (b >= 0)
    solutions[0] = (-b - sqrt(disc)) / two_a;
  else
    solutions[0] = (-b + sqrt(disc)) / two_a;
  
  solutions[1] = c / (a * solutions[0]);
  return 2;
}

int solveCubic(double *coeff, double *solutions)
{
  // Solve a cubic equation of the form At^3 + Bt^2 + Ct + D = 0.
  // Uses Cardano's method after reducing the cubic to t^3 + pt + q = 0.
  double a, b, c;
  double a_squared, p_cubed, disc, p, q;
  double temp, phi;
  int i, found;

  if (IS_ZERO(coeff[3]))
    return solveQuadric(coeff, solutions);

  // Convert the cubic into normal form (t^3 + at^2 + bt + c = 0).
  a = coeff[2] / coeff[3];
  b = coeff[1] / coeff[3];
  c = coeff[0] / coeff[3];

  // Perform the substitution x = (t - a/3) to eliminate the t^2 term.
  // Gets the formula x^3 + px + q = 0 at the end.
  a_squared = SQR(a);
  p = 1.0/3.0 * (-(1.0/3.0 ) * a_squared + b);
  q = 1.0/2.0 * ( (2.0/27.0) * a_squared * a - (1.0/3.0)*a*b + c);

  // Apply Cardano's formula to find the discriminant:
  // d = q^2 + p^3
  // If d = 0 and q == 0, there is one triple solution.
  //              q != 0, there is one single and one double solution.
  //    d < 0 this is a 'casus irreducibilis' - three real solutions.
  //    d > 0 there is one real solution and two imaginary solutions.
  p_cubed = p * p * p;
  disc = SQR(q) + p_cubed;

  if (IS_ZERO(disc)) {
    if (IS_ZERO(q)) {
      // Only one solution, at x = 0
      solutions[0] = 0;
      found = 1;
    } else {
      // One single solution and one double solution.
      // At 2*cbrt(-q) and -cbrt(-q).
      temp = cbrt(-q);
      solutions[0] = 2*temp;
      solutions[1] = -temp;
      found = 2;
    }
  } else if (disc < 0) {
    // Casus irreducibilis ... three real solutions.
    phi = (1.0/3.0) * acos(-q / sqrt(-p_cubed));
    temp = 2 * sqrt(-p);

    solutions[0] =  temp*cos(phi);
    solutions[1] = -temp*cos(phi + M_PI/3.0);
    solutions[2] = -temp*cos(phi - M_PI/3.0);
    found = 3;
  } else {
    // disc > 0, one real solution (two imaginary solutions)
    temp = sqrt(disc);

    solutions[0] = cbrt(temp - q) - cbrt(temp + q);
    found = 1;
  }

  // Convert the solutions from x to t space.
  temp = a / 3.0;

  for (i = 0; i < found; i++)
    solutions[i] -= temp;

  return found;
}

int solveQuartic(double *coeff, double *solutions)
{
  // Solve a quartic equation of the form At^4 + Bt^3 + Ct^2 + Dt + E = 0.
  double cubicCoeffs[4];
  double a, b, c, d;
  double a_squared, p, q, r;
  double cubic_sol, u, v, temp;
  int found, cubic_found, i;

  // Check that we really have a quartic.
  if (IS_ZERO(coeff[4])) {
    return solveCubic(coeff, solutions);
  }

  // Convert the equation into normal form
  // t^4 + at^3 + bt^2 + ct + d = 0
  a = coeff[3] / coeff[4];
  b = coeff[2] / coeff[4];
  c = coeff[1] / coeff[4];
  d = coeff[0] / coeff[4];

  // Make the substitution t = x - a/4 to eliminate the term in t^4
  // to give x^4 + px^2 + qx + r = 0
  a_squared = SQR(a);
  p = -(3.0/8.0)*a_squared + b;
  q = (1.0/8.0)*a_squared*a - (1.0/2.0)*a*b + c;
  r = -(3.0/256.0)*a_squared*a_squared + (1.0/16.0)*a_squared*b
      - (1.0/4.0)*a*c + d;
  
  // If r == 0 then the equation degenerates into x(x^3 + px + q) = 0
  // which may be solved as a cubic.
  if (IS_ZERO(r)) {
    cubicCoeffs[0] = q;
    cubicCoeffs[1] = p;
    cubicCoeffs[2] = 0;
    cubicCoeffs[3] = 1;

    found = solveCubic(cubicCoeffs, solutions);
    
    // Add another solution at zero.
    solutions[found++] = 0;
  } else {
    // We solve the resolvent cubic to get one real solution.
    cubicCoeffs[0] = (1.0/2.0)*r*p - (1.0/8.0)*SQR(q);
    cubicCoeffs[1] = -r;
    cubicCoeffs[2] = -(1.0/2.0)*p;
    cubicCoeffs[3] = 1;
    
    cubic_found = solveCubic(cubicCoeffs, solutions);

    if (cubic_found == 0) {
      printf("***** solveQuartic: resolvent cubic is degenerate!\n");
      return 0;
    }

    cubic_sol = solutions[0];

    // The solution to the resolvent cubic allows us to factorise the
    // quartic into two quadratic equations. Solve them.
    u = SQR(cubic_sol) - r;
    v = 2*cubic_sol - p;

    if (IS_ZERO(u))
      u = 0;
    else if (u > 0)
      u = sqrt(u);
    else
      u = sqrt(-u);

    if (IS_ZERO(v))
      v = 0;
    else if (v > 0)
      v = sqrt(v);
    else
      v = sqrt(-v);

    cubicCoeffs[0] = cubic_sol - u;
    cubicCoeffs[1] = (q < 0) ? -v : v;
    cubicCoeffs[2] = 1;

    found = solveQuadric(cubicCoeffs, &solutions[0]);

    cubicCoeffs[0] = cubic_sol + u;
    cubicCoeffs[1] = (q < 0) ? v : -v;
    cubicCoeffs[2] = 1;

    found += solveQuadric(cubicCoeffs, &solutions[2]);
  }

  // Perform the substitution back from x into t-space.
  temp = a/4.0;

  for (i = 0; i < found; i++)
    solutions[i] -= temp;

  return found;
}

#ifdef USE_OWN_CBRT
/*
 * cbrt(x) - return cube root of x
 *
 * ====================================================
 * Copyright (C) 1993 by Sun Microsystems, Inc. All rights reserved.
 *
 * Developed at SunPro, a Sun Microsystems, Inc. business.
 * Permission to use, copy, modify, and distribute this
 * software is freely granted, provided that this notice 
 * is preserved.
 * ====================================================
 */


/* Define n0 1 for little endian, e.g. Intel */
/* Define n0 0 for big endian, e.g. Macintosh */
#ifdef BIG_ENDIAN
#define n0 0
#else
#define n0 1
#endif

double cbrt(double x) 
{
  static const unsigned 
    B1 = 715094163, /* B1 = (682-0.03306235651)*2**20 */
    B2 = 696219795; /* B2 = (664-0.03306235651)*2**20 */

  static const double
    C =  5.42857142857142815906e-01, /* 19/35     = 0x3FE15F15, 0xF15F15F1 */
    D = -7.05306122448979611050e-01, /* -864/1225 = 0xBFE691DE, 0x2532C834 */
    E =  1.41428571428571436819e+00, /* 99/70     = 0x3FF6A0EA, 0x0EA0EA0F */
    F =  1.60714285714285720630e+00, /* 45/28     = 0x3FF9B6DB, 0x6DB6DB6E */
    G =  3.57142857142857150787e-01; /* 5/14      = 0x3FD6DB6D, 0xB6DB6DB7 */

  int hx;
  double r,s,w;
  unsigned sign;
  union {
    double t;
    unsigned pt[2];
  } ut, ux;

  ut.t = 0.0;
  ux.t = x;

  hx = ux.pt[n0];                 	/* high word of x */
  sign = hx&0x80000000; 		/* sign= sign(x) */
  hx ^= sign;
  if (hx >= 0x7ff00000)
    return(x+x); 			/* cbrt(NaN,INF) is itself */
  if( (hx | ux.pt[1-n0]) == 0) 
    return(ux.t);	                /* cbrt(0) is itself */

  ux.pt[n0] = hx;
  
  /* rough cbrt to 5 bits */
  if(hx < 0x00100000) { 		/* subnormal number */
    ut.pt[n0] = 0x43500000; 		/* set t= 2**54 */
	ut.t *= x;
	ut.pt[n0] = ut.pt[n0]/3+B2;
  } else
	ut.pt[n0] = hx/3+B1;

  /* new cbrt to 23 bits, may be implemented in single precision */
  r = ut.t * ut.t/ux.t;
  s = C + r*ut.t;
  ut.t *= G + F/(s + E + D/s);

  /* chopped to 20 bits and make it larger than cbrt(x) */ 
  ut.pt[1-n0] = 0;
  ut.pt[n0] += 0x00000001;


  /* one step newton iteration to 53 bits with error less than 0.667 ulps */
  s = ut.t*ut.t;			/* t*t is exact */
  r = ux.t/s;
  w = ut.t + ut.t;
  r = (r - ut.t) / (w + r);		/* r-s is exact */
  ut.t = ut.t + ut.t*r;

  /* restore the sign bit */
  ut.pt[n0] |= sign;
  return(ut.t);
}

#endif



#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif
