/* prim_bilinear.c : The bilinear primitive
** 
** The bilinear primitive defines a bilinear patch between the points
** (0, 0, 0), (1, 0, 0), (0, 1, 0), (x_point, y_point, 1).
**
** Chris Butcher, 14 July 1997
*/

#include "mirage.h"

#ifdef USE_NAMESPACES
namespace Mirage {
#endif

int bilinearPrimitive::ray_intersect(IntersectionList *the_list,
				     primnode *node, primnode *validate_node, 
				     vector ray_from, vector ray_dir,
				     uint ray_flags, signature ray_signature)
{
  // The coordinates of ray_from in primitive space
  double p_x, p_y, p_z;
  // The direction ray_dir in primitive space
  vector dir;
  // Variables used to form the quadratic equation
  double qx, qy, c_dx, c_dy, c_px, c_py;
  // The coefficients of the quadratic equation in t
  double a, b, c;
  // The discriminant of this equation
  double disc;
  // Have we found an intersection?
  int found = FALSE;
  // The t-values of the intersection points.
  double t_0, t_1;
  // Two vectors along the surface at the hitpoint; dp/ds1 and dp/ds2.
  vector n_s1, n_s2;
  // The normal at the hitpoint.
  vector norm;
  // The points due N, S, E and W of the hitpoint on the patch's edges.
  double n_x, n_y, n_z, s_x, s_y, s_z, e_x, e_y, e_z, w_x, w_y, w_z;
  // Parameters that are used in calculating the normal and hitpoint.
  double hit_z, hit_s1, hit_s2;

  // Store the signature of this ray that we're intersecting with.
  node->last_ray_signature = ray_signature;

  // This apply is an inline function defined in matrix.h
  // Transform the ray startpoint into primitive space
  Matrix::apply(node->position[AFT], ray_from, &p_x, &p_y, &p_z);

  // This apply is also an inline function defined in matrix.h
  // Transform the ray direction into primitive space (note: shift does
  // not apply to this, only apply rotation + stretch sections of the matrix)
  Matrix::applynoshift(node->position[AFT], ray_dir, dir);

  // The patch may be flat. If it is then the equations are different.
  if (is_flat) {
    // The patch is flat. Thus there is only one possible intersection.
    //
    // The intersection is defined as t =  - p_z / dir[Z].
    if (dir[Z] == 0) {
      // No intersections.
      node->last_ray_hit = FALSE;
      return node->last_ray_hit;
    }
    t_0 = - p_z / dir[Z];
    t_1 = -INFINITY; // Not a real intersection; won't be considered.

    qx = 1 - x_point;
    qy = 1 - y_point;
    
  } else {
    // The patch is not flat. Use the normal equations.

    // Form the coefficients of the quadratic equation that defines t.
    // This equation is given by:
    //
    //  qx = 1 - x_point
    //  qy = 1 - y_point
    //
    //  c_dx = dx + qx * dz
    //  c_dy = dy + qy * dz
    //
    //  c_px = px + qx * pz
    //  c_py = py + qy * pz
    //
    //  a = c_dx * c_dy
    //  b = c_dx * c_py + c_dy * c_px - dz
    //  c = c_px * c_py - pz
    
    qx = 1 - x_point;
    qy = 1 - y_point;
    
    c_dx = dir[X] + qx * dir[Z];
    c_dy = dir[Y] + qy * dir[Z];
    
    c_px = p_x + qx * p_z;
    c_py = p_y + qy * p_z;
    
    a = c_dx * c_dy;
    b = c_dx * c_py + c_dy * c_px - dir[Z];
    c = c_px * c_py - p_z;
    
    if (a == 0) {
      // This equation reduces to a linear equation, bt + c = 0.
      // So t = - c / b.
      if (b == 0) {
	// There are no solutions, no intersections.
	node->last_ray_hit = FALSE;
	return node->last_ray_hit;
      }
      
      t_0 = - c / b;
      t_1 = -INFINITY; // Not a real intersection; won't be considered.
    } else {
      
      // Calculate the discriminant of the quadratic equation.
      disc = SQR(b) - 4.0*a*c;
      
      // If the discriminant is less than or equal to zero, then there
      // are no intersections with the bilinear patch (tangent rays are
      // treated as no intersection at all).
      if (disc <= 0) {
	node->last_ray_hit = FALSE;
	return node->last_ray_hit;
      }
      
      // Calculate the intersection t-values. Must have t_0 < t_1 for later
      // calculations to work.
      // Note that we can save a square root operation by saying
      // t_1 = c / a * t_0 ... but this only works if we can be sure that
      // t_0 isn't zero. It may easily be if sqrt(disc) == b or == -b.
      // So, we make sure that the +/-sqrt(disc) is always the same sign
      // as -b to make sure that the subtraction can't produce zero.
      // This makes the code a little long-winded.
      if (a >= 0) {
	if (b >= 0) {
	  t_0 = (-b - sqrt(disc)) / (2 * a);
	  t_1 = c / (a * t_0);
	} else {
	  t_1 = (-b + sqrt(disc)) / (2 * a);
	  t_0 = c / (a * t_1);
	}
      } else {
	if (b >= 0) {
	  t_1 = (-b - sqrt(disc)) / (2 * a);
	  t_0 = c / (a * t_1);
	} else {
	  t_0 = (-b + sqrt(disc)) / (2 * a);
	  t_1 = c / (a * t_0);
	}
      }
    }
  }

  // Insert any positive t-values (valid intersections).
  if (t_0 > 0) {
    hit_z = p_z + t_0*dir[Z];

    // Get the value of s1 (the parameter from point 0 to point 1).
    hit_s1 = (p_x + t_0*dir[X]) + hit_z*qx;

    if ((hit_s1 > 0) && (hit_s1 < 1)) {
      // Get the value of s2 (the parameter from point 0 to point 2).
      hit_s2 = (p_y + t_0*dir[Y]) + hit_z*qy;

      if ((hit_s2 > 0) && (hit_s2 < 1)) {

	// First check to see if the hitpoint is within the Cutout
	// that determines our outline (if we have a Cutout).
	if ((cutout == NULL) || (cutout->is_inside(hit_s1, hit_s2))) {

	  // Find the normal.

	  // Find the east and west points that correspond to the hitpoint.
	  // We do this by interpolating between p1 and p3 for e and
	  // p0 and p2 for w.
	  e_x = 1 * (1 - hit_s2) + x_point * hit_s2;
	  e_y = 0 * (1 - hit_s2) + y_point * hit_s2;
	  e_z = 0 * (1 - hit_s2) +       1 * hit_s2;
	  
	  w_x = 0 * (1 - hit_s2) +       0 * hit_s2;
	  w_y = 0 * (1 - hit_s2) +       1 * hit_s2;
	  w_z = 0 * (1 - hit_s2) +       0 * hit_s2;
	  
	  // Find the north and south points that correspond to the hitpoint.
	  // We do this by interpolating between p2 and p3 for n and
	  // p0 and p1 for s.
	  n_x = 0 * (1 - hit_s1) + x_point * hit_s1;
	  n_y = 1 * (1 - hit_s1) + y_point * hit_s1;
	  n_z = 0 * (1 - hit_s1) +       1 * hit_s1;
	  
	  s_x = 0 * (1 - hit_s2) +       1 * hit_s2;
	  s_y = 0 * (1 - hit_s2) +       0 * hit_s2;
	  s_z = 0 * (1 - hit_s2) +       0 * hit_s2;
	  
	  // Now we construct the vectors n_s1 and n_s2. These are
	  // n_s1 = (dx/ds1 dy/ds1 dz/ds1) and n_s2 = (dx/ds2 dy/ds2 dz/ds2)
	  //
	  // These give us two directions along the surface at the hitpoint.
	  // We take n_s1 x n_s2 = norm.
	  
	  Vector::assign(n_s1, e_x - w_x, e_y - w_y, e_z - w_z);
	  Vector::assign(n_s2, n_x - s_x, n_y - s_y, n_z - s_z);
	  
	  Vector::cross(n_s1, n_s2, norm);
	  
	  // If the ray is approaching from the underside of the patch then
	  // we reverse the calculated normal to get the normal of the
	  // underside.
	  if (Vector::dot(norm, dir) > 0)
	    Vector::mult(-1, norm, norm);
	  
	  if (fabs(t_0) > MIN_TVAL_CUTOFF)
	    t_0 -= MIN_TVAL_CUTOFF * 0.1;
	  
	  insert_entering(the_list, node, validate_node, norm, t_0,
			  ray_from, ray_dir, ray_flags);
	  
	  found = TRUE;
	}
      }
    }
  }

  if (t_1 > 0) {
    hit_z = p_z + t_1*dir[Z];

    // Get the value of s1 (the parameter from point 0 to point 1).
    hit_s1 = (p_x + t_1*dir[X]) + hit_z*qx;

    if ((hit_s1 > 0) && (hit_s1 < 1)) {
      // Get the value of s2 (the parameter from point 0 to point 2).
      hit_s2 = (p_y + t_1*dir[Y]) + hit_z*qy;

      if ((hit_s2 > 0) && (hit_s2 < 1)) {

	// First check to see if the hitpoint is within the Cutout
	// that determines our outline (if we have a Cutout).
	if ((cutout == NULL) || (cutout->is_inside(hit_s1, hit_s2))) {

	  // Find the normal.

	  // Find the east and west points that correspond to the hitpoint.
	  // We do this by interpolating between p1 and p3 for e and
	  // p0 and p2 for w.
	  e_x = 1 * (1 - hit_s2) + x_point * hit_s2;
	  e_y = 0 * (1 - hit_s2) + y_point * hit_s2;
	  e_z = 0 * (1 - hit_s2) +       1 * hit_s2;
	  
	  w_x = 0 * (1 - hit_s2) +       0 * hit_s2;
	  w_y = 0 * (1 - hit_s2) +       1 * hit_s2;
	  w_z = 0 * (1 - hit_s2) +       0 * hit_s2;
	  
	  // Find the north and south points that correspond to the hitpoint.
	  // We do this by interpolating between p2 and p3 for n and
	  // p0 and p1 for s.
	  n_x = 0 * (1 - hit_s1) + x_point * hit_s1;
	  n_y = 1 * (1 - hit_s1) + y_point * hit_s1;
	  n_z = 0 * (1 - hit_s1) +       1 * hit_s1;
	  
	  s_x = 0 * (1 - hit_s2) +       1 * hit_s2;
	  s_y = 0 * (1 - hit_s2) +       0 * hit_s2;
	  s_z = 0 * (1 - hit_s2) +       0 * hit_s2;
	  
	  // Now we construct the vectors n_s1 and n_s2. These are
	  // n_s1 = (dx/ds1 dy/ds1 dz/ds1) and n_s2 = (dx/ds2 dy/ds2 dz/ds2)
	  //
	  // These give us two directions along the surface at the hitpoint.
	  // We take n_s1 x n_s2 = norm.
	  
	  Vector::assign(n_s1, e_x - w_x, e_y - w_y, e_z - w_z);
	  Vector::assign(n_s2, n_x - s_x, n_y - s_y, n_z - s_z);
	  
	  Vector::cross(n_s1, n_s2, norm);
	  
	  // If the ray is approaching from the underside of the patch then
	  // we reverse the calculated normal to get the normal of the
	  // underside.
	  if (Vector::dot(norm, dir) > 0)
	    Vector::mult(-1, norm, norm);
	  
	  if (fabs(t_1) > MIN_TVAL_CUTOFF)
	    t_1 -= MIN_TVAL_CUTOFF * 0.1;
	  
	  insert_leaving(the_list, node, validate_node, norm, t_1,
			 ray_from, ray_dir, ray_flags);
	  
	  found = TRUE;
	}
      }
    }
  }

  if (found)
    node->last_ray_hit = TRUE;
  else
    node->last_ray_hit = FALSE;

  return node->last_ray_hit;
}

int bilinearPrimitive::ray_intersect(vector ray_from, vector ray_dir,
				     rayintersect *int_storage)
{
  // Variables used to form the quadratic equation
  double qx, qy, c_dx, c_dy, c_px, c_py;
  // The coefficients of the quadratic equation in t
  double a, b, c;
  // The discriminant of this equation
  double disc;
  // Have we found an intersection?
  int num_intersections = 0;
  // The t-values of the intersection points.
  double t_0, t_1;
  // Two vectors along the surface at the hitpoint; dp/ds1 and dp/ds2.
  vector n_s1, n_s2;
  // The normal at the hitpoint.
  vector norm;
  // The points due N, S, E and W of the hitpoint on the patch's edges.
  double n_x, n_y, n_z, s_x, s_y, s_z, e_x, e_y, e_z, w_x, w_y, w_z;
  // Parameters that are used in calculating the normal and hitpoint.
  double hit_z, hit_s1, hit_s2;

  // The patch may be flat. If it is then the equations are different.
  if (is_flat) {
    // The patch is flat. Thus there is only one possible intersection.
    //
    // The intersection is defined as t =  - ray_from[Z] / ray_dir[Z].
    if (ray_dir[Z] == 0) {
      // No intersections.
      return 0;
    }
    t_0 = - ray_from[Z] / ray_dir[Z];
    t_1 = -INFINITY; // Not a real intersection; won't be considered.

    qx = 1 - x_point;
    qy = 1 - y_point;
    
  } else {
    // The patch is not flat. Use the normal equations.

    // Form the coefficients of the quadratic equation that defines t.
    // This equation is given by:
    //
    //  qx = 1 - x_point
    //  qy = 1 - y_point
    //
    //  c_dx = dx + qx * dz
    //  c_dy = dy + qy * dz
    //
    //  c_px = px + qx * pz
    //  c_py = py + qy * pz
    //
    //  a = c_dx * c_dy
    //  b = c_dx * c_py + c_dy * c_px - dz
    //  c = c_px * c_py - pz
    
    qx = 1 - x_point;
    qy = 1 - y_point;
    
    c_dx = ray_dir[X] + qx * ray_dir[Z];
    c_dy = ray_dir[Y] + qy * ray_dir[Z];
    
    c_px = ray_from[X] + qx * ray_from[Z];
    c_py = ray_from[Y] + qy * ray_from[Z];
    
    a = c_dx * c_dy;
    b = c_dx * c_py + c_dy * c_px - ray_dir[Z];
    c = c_px * c_py - ray_from[Z];
    
    if (a == 0) {
      // This equation reduces to a linear equation, bt + c = 0.
      // So t = - c / b.
      if (b == 0) {
	// There are no solutions, no intersections.
	return 0;
      }
      
      t_0 = - c / b;
      t_1 = -INFINITY; // Not a real intersection; won't be considered.
    } else {
      
      // Calculate the discriminant of the quadratic equation.
      disc = SQR(b) - 4.0*a*c;
      
      // If the discriminant is less than or equal to zero, then there
      // are no intersections with the bilinear patch (tangent rays are
      // treated as no intersection at all).
      if (disc <= 0) {
	return 0;
      }
      
      // Calculate the intersection t-values. Must have t_0 < t_1 for later
      // calculations to work.
      // Note that we can save a square root operation by saying
      // t_1 = c / a * t_0 ... but this only works if we can be sure that
      // t_0 isn't zero. It may easily be if sqrt(disc) == b or == -b.
      // So, we make sure that the +/-sqrt(disc) is always the same sign
      // as -b to make sure that the subtraction can't produce zero.
      // This makes the code a little long-winded.
      if (a >= 0) {
	if (b >= 0) {
	  t_0 = (-b - sqrt(disc)) / (2 * a);
	  t_1 = c / (a * t_0);
	} else {
	  t_1 = (-b + sqrt(disc)) / (2 * a);
	  t_0 = c / (a * t_1);
	}
      } else {
	if (b >= 0) {
	  t_1 = (-b - sqrt(disc)) / (2 * a);
	  t_0 = c / (a * t_1);
	} else {
	  t_0 = (-b + sqrt(disc)) / (2 * a);
	  t_1 = c / (a * t_0);
	}
      }
    }
  }

  // Insert the intersections.
  hit_z = ray_from[Z] + t_0*ray_dir[Z];

  // Get the value of s1 (the parameter from point 0 to point 1).
  hit_s1 = (ray_from[X] + t_0*ray_dir[X]) + hit_z*qx;
  
  if ((hit_s1 > 0) && (hit_s1 < 1)) {
    // Get the value of s2 (the parameter from point 0 to point 2).
    hit_s2 = (ray_from[Y] + t_0*ray_dir[Y]) + hit_z*qy;
    
    if ((hit_s2 > 0) && (hit_s2 < 1)) {
      
      // First check to see if the hitpoint is within the Cutout
      // that determines our outline (if we have a Cutout).
      if ((cutout == NULL) || (cutout->is_inside(hit_s1, hit_s2))) {
	
	// Find the normal.
	
	// Find the east and west points that correspond to the hitpoint.
	// We do this by interpolating between p1 and p3 for e and
	// p0 and p2 for w.
	e_x = 1 * (1 - hit_s2) + x_point * hit_s2;
	e_y = 0 * (1 - hit_s2) + y_point * hit_s2;
	e_z = 0 * (1 - hit_s2) +       1 * hit_s2;
	
	w_x = 0 * (1 - hit_s2) +       0 * hit_s2;
	w_y = 0 * (1 - hit_s2) +       1 * hit_s2;
	w_z = 0 * (1 - hit_s2) +       0 * hit_s2;
	
	// Find the north and south points that correspond to the hitpoint.
	// We do this by interpolating between p2 and p3 for n and
	// p0 and p1 for s.
	n_x = 0 * (1 - hit_s1) + x_point * hit_s1;
	n_y = 1 * (1 - hit_s1) + y_point * hit_s1;
	n_z = 0 * (1 - hit_s1) +       1 * hit_s1;
	
	s_x = 0 * (1 - hit_s2) +       1 * hit_s2;
	s_y = 0 * (1 - hit_s2) +       0 * hit_s2;
	s_z = 0 * (1 - hit_s2) +       0 * hit_s2;
	
	// Now we construct the vectors n_s1 and n_s2. These are
	// n_s1 = (dx/ds1 dy/ds1 dz/ds1) and n_s2 = (dx/ds2 dy/ds2 dz/ds2)
	//
	// These give us two ray_directions along the surface at the hitpoint.
	// We take n_s1 x n_s2 = norm.
	
	Vector::assign(n_s1, e_x - w_x, e_y - w_y, e_z - w_z);
	Vector::assign(n_s2, n_x - s_x, n_y - s_y, n_z - s_z);
	
	Vector::cross(n_s1, n_s2, norm);
	
	// If the ray is approaching from the underside of the patch then
	// we reverse the calculated normal to get the normal of the
	// underside.
	if (Vector::dot(norm, ray_dir) > 0)
	  Vector::mult(-1, norm, norm);
	
	int_storage[num_intersections].t = t_0;
	int_storage[num_intersections].is_entering = TRUE;
	Vector::set(int_storage[num_intersections].normal, norm);
	num_intersections++;
      }
    }
  }

  hit_z = ray_from[Z] + t_1*ray_dir[Z];

  // Get the value of s1 (the parameter from point 0 to point 1).
  hit_s1 = (ray_from[X] + t_1*ray_dir[X]) + hit_z*qx;
  
  if ((hit_s1 > 0) && (hit_s1 < 1)) {
    // Get the value of s2 (the parameter from point 0 to point 2).
    hit_s2 = (ray_from[Y] + t_1*ray_dir[Y]) + hit_z*qy;
    
    if ((hit_s2 > 0) && (hit_s2 < 1)) {
      
      // First check to see if the hitpoint is within the Cutout
      // that determines our outline (if we have a Cutout).
      if ((cutout == NULL) || (cutout->is_inside(hit_s1, hit_s2))) {
	
	// Find the normal.
	
	// Find the east and west points that correspond to the hitpoint.
	// We do this by interpolating between p1 and p3 for e and
	// p0 and p2 for w.
	e_x = 1 * (1 - hit_s2) + x_point * hit_s2;
	e_y = 0 * (1 - hit_s2) + y_point * hit_s2;
	e_z = 0 * (1 - hit_s2) +       1 * hit_s2;
	
	w_x = 0 * (1 - hit_s2) +       0 * hit_s2;
	w_y = 0 * (1 - hit_s2) +       1 * hit_s2;
	w_z = 0 * (1 - hit_s2) +       0 * hit_s2;
	
	// Find the north and south points that correspond to the hitpoint.
	// We do this by interpolating between p2 and p3 for n and
	// p0 and p1 for s.
	n_x = 0 * (1 - hit_s1) + x_point * hit_s1;
	n_y = 1 * (1 - hit_s1) + y_point * hit_s1;
	n_z = 0 * (1 - hit_s1) +       1 * hit_s1;
	
	s_x = 0 * (1 - hit_s2) +       1 * hit_s2;
	s_y = 0 * (1 - hit_s2) +       0 * hit_s2;
	s_z = 0 * (1 - hit_s2) +       0 * hit_s2;
	
	// Now we construct the vectors n_s1 and n_s2. These are
	// n_s1 = (dx/ds1 dy/ds1 dz/ds1) and n_s2 = (dx/ds2 dy/ds2 dz/ds2)
	//
	// These give us two directions along the surface at the hitpoint.
	// We take n_s1 x n_s2 = norm.
	
	Vector::assign(n_s1, e_x - w_x, e_y - w_y, e_z - w_z);
	Vector::assign(n_s2, n_x - s_x, n_y - s_y, n_z - s_z);
	
	Vector::cross(n_s1, n_s2, norm);
	
	// If the ray is approaching from the underside of the patch then
	// we reverse the calculated normal to get the normal of the
	// underside.
	if (Vector::dot(norm, ray_dir) > 0)
	  Vector::mult(-1, norm, norm);
	  
	int_storage[num_intersections].t = t_1;
	int_storage[num_intersections].is_entering = FALSE;
	Vector::set(int_storage[num_intersections].normal, norm);
	num_intersections++;
      }
    }
  }

  return num_intersections;
}

intersect_result bilinearPrimitive::point_intersect(primnode *node,
						    vector wp,
						    PointInfo *info,
						    uint ray_flags)
{
  // This primitive doesn't have any substance.
  return OUT;
}

intersect_result bilinearPrimitive::point_intersect(vector wp)
{
  // This primitive doesn't have any substance.
  return OUT;
}





// Now define some local macros to make the code length shorter for the
// routine bilinearPrimitive::voxel_intersect.


// Macro CHECK_OUTSIDE_PLUS_PLANE tests to see if the points of the box all
// lie outside the plane P at distance d.
#define CHECK_OUTSIDE_PLUS_PLANE(axis, d) {\
                                           \
  for (i = 0; i < 8; i++) {                \
    if (b[i][axis] < d) {                  \
      break;                               \
    }                                      \
  }                                        \
  if (i == 8) {                            \
    return OUT;                            \
  }                                        \
}


// Macro CHECK_OUTSIDE_MINUS_PLANE tests to see if the points of the box all
// lie outside the plane P at distance -d.
#define CHECK_OUTSIDE_MINUS_PLANE(axis, d) {\
                                            \
  for (i = 0; i < 8; i++)                   \
    if (b[i][axis] > -d)                    \
      break;                                \
                                            \
  if (i == 8) {                             \
    return OUT;                             \
  }                                         \
}

intersect_result bilinearPrimitive::voxel_intersect(box b)
{
  // This routine checks to see if a parallelipiped lies inside, outside,
  // or intersects the bilinear patch. We simply check to see if all the
  // points lie outside any of the planes that bound the patch.
  // (X < 0, X > x_point, Y < 0, Y > y_point, Z < 0, Z > 1)
  //
  // If the voxel is inside any of these planes then we return SURFACE.

  // Loop counter
  int i;

  // Step 1: Check to see if all of the points of the box lie outside
  //         the planes X < 0, X > x_point, Y < 0, Y > y_point, Z < 0, Z > 1.
  //         If so, return OUT.
  CHECK_OUTSIDE_PLUS_PLANE(X, x_point);
  CHECK_OUTSIDE_PLUS_PLANE(Y, y_point);
  if (is_flat) {
    CHECK_OUTSIDE_PLUS_PLANE(Z, 0);
  } else {
    CHECK_OUTSIDE_PLUS_PLANE(Z, 1);
  }
  CHECK_OUTSIDE_MINUS_PLANE(X, 0);
  CHECK_OUTSIDE_MINUS_PLANE(Y, 0);
  CHECK_OUTSIDE_MINUS_PLANE(Z, 0);

  // If we haven't returned OUT by this time then the box intersects
  // the patch's bounding box. Return SURFACE.
  return SURFACE;
}

// Undefine the macros defined above (local only).
#undef CHECK_OUTSIDE_PLUS_PLANE
#undef CHECK_OUTSIDE_MINUS_PLANE

char *bilinearPrimitive::prim_name(void) {
  return "Bilinear Patch";
}

#define PATCH_MESSAGES 0

// A routine to create a new patch and place it so its corners are at
// any four specified points.
object new_bilinear_patch(vector p0, vector p1, vector p2, vector p3)
{
  object patch;
  transform patch_pos;
  double patch_x, patch_y;
  int flat_patch;

  position_patch(p0, p1, p2, p3, &flat_patch, &patch_x, &patch_y, patch_pos);

  patch = new_bilinear_patch(patch_x, patch_y, flat_patch);
  patch->place(patch_pos);

  return bind(patch);
}

// A routine to create a new patch and place it so its corners are at
// any four specified points. Also takes a Cutout parameter.
object new_bilinear_patch(vector p0, vector p1, vector p2, vector p3,
			  Cutout *cut)
{
  object patch;
  transform patch_pos;
  double patch_x, patch_y;
  int flat_patch;

  position_patch(p0, p1, p2, p3, &flat_patch, &patch_x, &patch_y, patch_pos);

  patch = new_bilinear_patch(patch_x, patch_y, flat_patch, cut);
  patch->place(patch_pos);

  return bind(patch);
}

// A routine to create a new patch and place it so its corners are at
// any four specified points. Places a PatchImageTexture on the patch
// with the given GenericPixmap.
object new_bilinear_patch(vector p0, vector p1, vector p2, vector p3,
			  int tex_prec, GenericPixmap *tex_pix,
			  texture_method tex_method, int tex_filter)
{
  object patch;
  transform patch_pos;
  double patch_x, patch_y;
  int flat_patch;

  position_patch(p0, p1, p2, p3, &flat_patch, &patch_x, &patch_y, patch_pos);

  patch = new_bilinear_patch(patch_x, patch_y, flat_patch);
#if 0
  // FIXME: PatchImageTexture does not exist
  patch->texture = new PatchImageTexture(tex_prec, tex_pix, patch_x, patch_y,
					 tex_method, tex_filter);
#endif

  patch = bind(patch);
  patch->place(patch_pos);

  return bind(patch);
}

void position_patch(vector p0, vector p1, vector p2, vector p3, int *zero_z,
		    double *x_param, double *y_param, transform position)
{
  vector shiftedp0, shiftedp1, shiftedp2, shiftedp3, plane_normal;
  vector rotatedp1, rotatedp2, rotatedp3;
  vector x_basis, y_basis, z_basis;
  double x_stretch, y_stretch, z_stretch;
  double x_stretch_inv, y_stretch_inv, z_stretch_inv;
  double shear_factor;

#if PATCH_MESSAGES
  printf("** Beginning bilinear patch placement\n");
  printf("p0: %f %f %f\n", p0[X], p0[Y], p0[Z]);
  printf("p1: %f %f %f\n", p1[X], p1[Y], p1[Z]);
  printf("p2: %f %f %f\n", p2[X], p2[Y], p2[Z]);
  printf("p3: %f %f %f\n", p3[X], p3[Y], p3[Z]);
#endif

  // Shift the p0p1p2p3 space so that p0 is at the origin.
#if PATCH_MESSAGES
  printf("* Shifting by -p0\n");
#endif
  Vector::sub(p0, p0, shiftedp0);
  Vector::sub(p1, p0, shiftedp1);
  Vector::sub(p2, p0, shiftedp2);
  Vector::sub(p3, p0, shiftedp3);

#if PATCH_MESSAGES
  printf("shifted p0: %f %f %f\n", shiftedp0[X], shiftedp0[Y], shiftedp0[Z]);
  printf("shifted p1: %f %f %f\n", shiftedp1[X], shiftedp1[Y], shiftedp1[Z]);
  printf("shifted p2: %f %f %f\n", shiftedp2[X], shiftedp2[Y], shiftedp2[Z]);
  printf("shifted p3: %f %f %f\n", shiftedp3[X], shiftedp3[Y], shiftedp3[Z]);

  printf("* Finding basis system...\n");
#endif

  // We work out the coordinate system defined by the plane p0p1p2 = XY and
  // p0p1 = X axis. First find the cross product of the vectors p0p1 and p0p2
  // to get the normal to this plane ... the Z axis of the basis system.
  Vector::cross(shiftedp1, shiftedp2, plane_normal);
  Vector::mult(1.0 / CHECK0(Vector::magnitude(plane_normal), 
			    "new_bilinear_patch: normal to p0p1p2 plane"),
	       plane_normal, z_basis);

  // Find the X axis of the basis system, a normalised p0p1 vector.
  Vector::mult(1.0 / CHECK0(Vector::magnitude(shiftedp1),
			    "new_bilinear_patch: p0p1 vector"),
	       shiftedp1, x_basis);

  // Find the Y axis of the basis system by taking the cross product of the
  // X and Z axes.
  Vector::cross(z_basis, x_basis, y_basis);
  Vector::normalise(y_basis, "new_bilinear_patch: Y axis of basis system");

#if PATCH_MESSAGES
  printf("x basis: %f %f %f\n", x_basis[X], x_basis[Y], x_basis[Z]);
  printf("y basis: %f %f %f\n", y_basis[X], y_basis[Y], y_basis[Z]);
  printf("z basis: %f %f %f\n", z_basis[X], z_basis[Y], z_basis[Z]);

  printf("* Decomposing p1, p2, p3 into above basis system...\n");
#endif

  // Find the coordinates of p1, p2 and p3 in this rotated basis system
  // which places p1 along the X axis and p2 in the XY plane.
  rotatedp1[X] = Vector::dot(x_basis, shiftedp1);
  rotatedp1[Y] = Vector::dot(y_basis, shiftedp1);
  rotatedp1[Z] = Vector::dot(z_basis, shiftedp1);

  rotatedp2[X] = Vector::dot(x_basis, shiftedp2);
  rotatedp2[Y] = Vector::dot(y_basis, shiftedp2);
  rotatedp2[Z] = Vector::dot(z_basis, shiftedp2);

  rotatedp3[X] = Vector::dot(x_basis, shiftedp3);
  rotatedp3[Y] = Vector::dot(y_basis, shiftedp3);
  rotatedp3[Z] = Vector::dot(z_basis, shiftedp3);

#if PATCH_MESSAGES
  printf("rotated p1: %f %f %f\n", rotatedp1[X], rotatedp1[Y], rotatedp1[Z]);
  printf("rotated p2: %f %f %f\n", rotatedp2[X], rotatedp2[Y], rotatedp2[Z]);
  printf("rotated p3: %f %f %f\n", rotatedp3[X], rotatedp3[Y], rotatedp3[Z]);
#endif

  // We need to stretch the points to place p3 at Z = 1.
  if (fabs(rotatedp3[Z]) < TINY) {
    *zero_z = TRUE;
    z_stretch = z_stretch_inv = 1;
  } else {
    *zero_z = FALSE;
    z_stretch_inv = rotatedp3[Z];
    z_stretch = 1.0 / CHECK0(z_stretch_inv,
			     "new_bilinear_patch: Z stretch factor");
  }

  // We stretch p1, p2 and p3 to place p1 at X = 1, p2 at Y = 1
  // and p3 at Z = 1.
  x_stretch_inv = rotatedp1[X];
  y_stretch_inv = rotatedp2[Y];
  x_stretch = 1.0 / CHECK0(x_stretch_inv,
			   "new_bilinear_patch: X stretch factor");
  y_stretch = 1.0 / CHECK0(y_stretch_inv,
			   "new_bilinear_patch: Y stretch factor");

#if PATCH_MESSAGES
  printf("* Stretching p1, p2, p3 by factors of %f %f %f (inv %f %f %f)\n",
	 x_stretch, y_stretch, z_stretch,
	 x_stretch_inv, y_stretch_inv, z_stretch_inv);
#endif

  rotatedp1[X] *= x_stretch;
  rotatedp1[Y] *= y_stretch;
  rotatedp1[Z] *= z_stretch;

  rotatedp2[X] *= x_stretch;
  rotatedp2[Y] *= y_stretch;
  rotatedp2[Z] *= z_stretch;

  rotatedp3[X] *= x_stretch;
  rotatedp3[Y] *= y_stretch;
  rotatedp3[Z] *= z_stretch;

#if PATCH_MESSAGES
  printf("stretched p1: %f %f %f\n", rotatedp1[X], rotatedp1[Y], rotatedp1[Z]);
  printf("stretched p2: %f %f %f\n", rotatedp2[X], rotatedp2[Y], rotatedp2[Z]);
  printf("stretched p3: %f %f %f\n", rotatedp3[X], rotatedp3[Y], rotatedp3[Z]);
#endif

  // The final component is a shear to transform p2 from its current
  // position at (p2x, 1, 0) to (0, 1, 0).
  shear_factor = - rotatedp2[X];

#if PATCH_MESSAGES
  printf("* Shearing points by shear factor of %f\n", shear_factor);
#endif

  rotatedp1[X] += shear_factor * rotatedp1[Y];
  rotatedp2[X] += shear_factor * rotatedp2[Y];
  rotatedp3[X] += shear_factor * rotatedp3[Y];

#if PATCH_MESSAGES
  printf("sheared p1: %f %f %f\n", rotatedp1[X], rotatedp1[Y], rotatedp1[Z]);
  printf("sheared p2: %f %f %f\n", rotatedp2[X], rotatedp2[Y], rotatedp2[Z]);
  printf("sheared p3: %f %f %f\n", rotatedp3[X], rotatedp3[Y], rotatedp3[Z]);
#endif

  // We now have p0 at the origin, p1 at (1, 0, 0), p2 at (0, 1, 0)
  // and p3 at (x, y, 1). Create a bilinear patch with these coordinates.
#if PATCH_MESSAGES
  printf("* Bilinear patch's parameters are: X %f, Y %f\n",
	 rotatedp3[X], rotatedp3[Y]);
#endif
  *x_param = rotatedp3[X];
  *y_param = rotatedp3[Y];

  // The position transform is initially the unit transform.
  Transform::unit(position);

  // Shear the patch by the shear factor required to place its p2 back at
  // the above (p2x, 1, 0).
  Transform::shear(position, Y, X, -shear_factor);

  // Stretch the patch to place p1, p2 and p3 back at their positions
  // in the rotated basis system.
  Transform::stretch(position, x_stretch_inv, y_stretch_inv, z_stretch_inv);

  // Rotate the patch so that p0p1 lies along x_basis and the plane p0p1p2
  // has a normal pointing along z_basis.
  Transform::rotate(position, x_basis, z_basis);

  // Shift the patch to place its p0 at the original p0.
  Transform::shift(position, p0);

  // The position matrix and parameters that we now have will place
  // the corners of a bilinear patch at p0, p1, p2 and p3.
}

#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif
