/* prim_cluster.c : The cluster primitive
** 
** This file contains functions used by all of the different cluster classes.
**
** Chris Butcher, 5 December 1996
*/

#include "mirage.h"

#define DEBUG_VALIDATE 0

#ifdef USE_NAMESPACES
namespace Mirage {
#endif

// Traverse the DAG that we contain and find absolute locations of
// all labels.
void ClusterPrimitive::find_labels(transform existing_transform)
{
  // If we haven't already found labels in the DAG graph, then
  // do so.
  if (!found_labels) {
    recursive_find_labels(myScene, existing_transform);
  
    found_labels = TRUE;
  }
}

// Traverse the DAG that we contain and create an intermediate RCSG.
void ClusterPrimitive::create_RCSG(primnode *cluster_primnode,
				   transform higher_cluster_world)
{
  transform temp_unit;
  TextureCollection *cluster_texcol;
  SingleTexture *cluster_tex;

  // If we haven't already converted the DAG graph into an RCSG,
  // then create it.
  if (!rcsg_created) {
    // The transform from this cluster's primitive space to worldspace is
    // made up of first this cluster's position relative to the higher
    // cluster's primitive space, and then the higher cluster's
    // primitive->world transform.
    Transform::combine(cluster_primnode->position, higher_cluster_world,
		     my_world_transform);

    // We can only create an RCSG tree from a DAG node, not a PRIM or
    // CAMERA or LIGHT node !!
    if ((myScene->flags & FLAGS_DAG) == 0) {
      fatal_error("ClusterPrimitive::create_RCSG",
		  "clustered object not a dag!",
		  "When creating a cluster, you must pass an object that is\n"
		  "not a primitive, a camera or a light; that is, it must be\n"
		  "a complex object.");
    }

    // Make the initial call to recursive_create_RCSG. All of the position
    // matrices of primitives inside this cluster are relative to a 
    // "worldspace" that isn't the absolute worldspace, but is rather
    // the cluster's primitive space. This is done by supplying a 
    // position matrix that is the unit matrix.
    Transform::unit(temp_unit);
    if (cluster_primnode->flags & FLAGS_MULTITEXTURE) {
      cluster_texcol = cluster_primnode->texcol;
      cluster_tex = NULL;
    } else {
      cluster_tex = cluster_primnode->texture;
      cluster_texcol = NULL;
    }
      
    myRCSG = recursive_create_RCSG(myScene,
				   my_world_transform,
				   cluster_primnode->flags,
				   temp_unit,
				   cluster_primnode->material,
				   cluster_tex, cluster_texcol,
				   cluster_primnode->lightlist,
				   cluster_primnode->light,
				   cluster_primnode->sfxtype,
				   cluster_primnode->sfx);

    rcsg_created = TRUE;
  }
}

// Recurse a DAG structure and create an RCSG tree from it.
thing ClusterPrimitive::recursive_create_RCSG(object theDag,
					      transform cluster_world,
					      uint above_flags,
					      transform above_position,
					      Material *above_material,
					      SingleTexture *above_tex,
					      TextureCollection *above_texcol,
					      lightList *above_lightlist,
					      Light *above_light,
					      SfxType above_sfxtype,
					      void *above_sfx)
{
  // The left + right branches of the dag
  object leftobj, rightobj;
  // If the dagnode is a primitive, rightprim holds the primitive *.
  Primitive *rightprim;
  // The new position of the object after this dag's contribution is used
  transform new_position;
  // The textures given to the lower nodes.
  SingleTexture *new_tex;
  TextureCollection *new_texcol;
  // The left and right branches of the rcsgnode.
  thing left_thing, right_thing;
  // Primnode * that will become the right branch.
  primnode *right_primnode;
  // Number of primitives in the RCSG tree.
  int num_prims;
  // Link type from this dagnode, placed in this RCSG node.
  linktype linkmode;
  // The camera that we have just encountered.
  Camera *the_camera;
  // The light that we have just encountered.
  Light *the_light;
  // The flags value for the new place.
  uint new_flags;
  // The SFX structure for the new primnode(s).
  SfxType new_sfxtype;
  void *new_sfx;

  // We are recursively traversing the dag-graph and turning it into an
  // RCSG tree. If this dagnode has an affect function attached to it,
  // apply it.
  if (theDag->affect != NULL)
    theDag->affect->affect_object(theDag);

  // The RCSG node has the same link type as the dagnode that it is
  // created from.
  linkmode = theDag->linkmode;

  // We don't have any primitives in our tree yet.
  num_prims = 0;

  // Left branch is the linked-in portion of the dag.
  leftobj  = theDag->linkptr;

  if (leftobj == NULL) {
    // There is no left object in the DAG graph, so none in the RCSG.
    left_thing = THING_EMPTY;
  } else {

    // Create an RCSGnode from the linkptr of the dagnode with the same
    // parameters that we were passed.
    left_thing = recursive_create_RCSG(leftobj,
				       cluster_world, above_flags,
				       above_position, above_material,
				       above_tex, above_texcol,
				       above_lightlist, above_light,
				       above_sfxtype, above_sfx);
    if (IS_RCSG(left_thing)) {
      // Add n primitives from the left branch.
      num_prims += GET_RCSGNODE(left_thing)->prim_count;
    } else {
      // Add one primitive.
      num_prims++;
    }
  }

  // Right branch is the actual sub-level of the dag.
  // Must update the parameters we were passed to include the changes
  // made by this dag.
  Transform::combine(theDag->position, above_position, new_position);
  new_flags = above_flags | theDag->flags;

  if (theDag->texture == NULL) {
    // This object has no texture applied to it at this level, use
    // the above one.
    if (above_flags & FLAGS_MULTITEXTURE) {
      new_tex = NULL; // not used
      new_texcol = above_texcol;
    } else {
      new_tex = above_tex;
      new_texcol = NULL;  // not used
    }
  } else {
    // Create a new texture collection combining the above one and the
    // texture at this level.
    if (above_flags & FLAGS_MULTITEXTURE) {
      new_tex = NULL; // not used
      new_texcol = new_textureCollection(above_texcol, theDag->texture,
					 new_position);
    } else {
      // There is 1 or 0 textures above us.
      if (above_tex == NULL) {
	// No previous textures, create a new SimpleTexture struct
	// and use that.
	new_tex = alloc_texture();
	new_tex->tex = theDag->texture;
	Transform::set(new_tex->tex_transform, new_position);
	new_texcol = NULL; // not used
      } else {
	// A previous texture exists, combine the two to make
	// a new texture collection.
	new_texcol = new_textureCollection(above_tex, theDag->texture,
					   new_position);
	new_tex = NULL; // not used
	new_flags |= FLAGS_MULTITEXTURE;
      }
    }
  }

  if (above_sfxtype) {
    new_sfxtype = above_sfxtype;
    new_sfx = above_sfx;
  } else {
    if (theDag->gaseous != NULL) {
      new_sfxtype = GASEOUS;
      new_sfx = (void *) new GasSFX(theDag->gaseous, new_position);

    } else if (theDag->surface_distortion != NULL) {
      new_sfxtype = DISTORTION;
      new_sfx = (void *) new DistortSFX(theDag->surface_distortion,
					new_position);
    } else {
      // No SFX.
      new_sfxtype = NOSFX;
      new_sfx = NULL;
    }
  }   

  if (theDag->flags & FLAGS_DAG) {
    rightobj = (object) theDag->details;

    // Create an RCSGnode from the details of the dagnode with updated
    // parameters.

    right_thing = recursive_create_RCSG(rightobj,
					cluster_world,
					new_flags,
					new_position,
					(above_material) ? above_material
                     					: theDag->material,
					new_tex,
					new_texcol,
					above_lightlist,
					 (above_light) ? above_light
                 					: theDag->light,
					new_sfxtype,
					new_sfx);

    if (IS_PRIM(right_thing)) {
      // Add the one primitive from the primnode.
      num_prims++;
    } else if (IS_RCSG(right_thing)) {
      // Add the n primitives from the right RCSG branch.
      num_prims += GET_RCSGNODE(right_thing)->prim_count;
    }

  } else if (theDag->flags & FLAGS_PRIM) {
    rightprim = (Primitive *) theDag->details;

    // Create a primnode from the details of the dagnode.

    // Allocate a new primnode and put its 'thing' address into right_thing.
    right_primnode = new_primnode(&right_thing);

    right_primnode->flags = new_flags & FLAGS_NOTYPE;
    Transform::set(right_primnode->position, new_position);

    right_primnode->material = (above_material != NULL) ? above_material
                                                        : theDag->material;
    right_primnode->prim = rightprim;
    if (new_flags & FLAGS_MULTITEXTURE)
      right_primnode->texcol = new_texcol;
    else
      right_primnode->texture = new_tex;
    right_primnode->lightlist = above_lightlist;
    right_primnode->light = (above_light != NULL) ? above_light
                                                  : theDag->light;
    right_primnode->sfxtype = new_sfxtype;
    right_primnode->sfx = new_sfx;

    num_prims++;

    if (rightprim->is_cluster) {
      ((ClusterPrimitive *) rightprim)->
	create_RCSG(right_primnode, cluster_world);
    }

  } else if (theDag->flags & FLAGS_CAMERA) {
    // Whenever we encounter a camera object, we set the corresponding
    // camera's world_transform to the correct value so that rays which
    // are generated by the camera look like they're coming from the
    // camera object's position. If the camera object has the
    // ONLY_THROUGH_GATEWAY flag set, then all rays cast from it initially
    // start on the far side of the gateway.
    the_camera = (Camera *) theDag->details;
    Transform::set(the_camera->world_transform, new_position);
    if (new_flags & FLAGS_ONLY_THROUGH_GATEWAY)
      the_camera->primary_ray = PRIMARY_RAY | RAY_THROUGH_GATEWAY;
    else
      the_camera->primary_ray = PRIMARY_RAY;

    right_thing = THING_EMPTY;

  } else if (theDag->flags & FLAGS_LIGHT) {
    // Whenever we encounter a light object, we set the corresponding
    // light's world_transform to the correct value so that this light
    // appears to cast light from the light object's position. Also set
    // the light's flags values.
    the_light = (Light *) theDag->details;
    the_light->light_flags = new_flags & FLAGS_NOTYPE;
    Transform::set(the_light->world_transform, new_position);

    // Add the light into the global lightList.
    globalLightList.add_light(the_light);

    right_thing = THING_EMPTY;
  } else {
    fatal_error("recursive_create_rcsg", "unknown dag type",
		"This error should not happen!");
    exit(-1);
  }

  if (linkmode == PLUS && left_thing == THING_EMPTY) {
    // Just return what the right branch was, not a RCSG node containing
    // it and NULL.
    return right_thing;

  } else if (linkmode == PLUS && right_thing == THING_EMPTY) {
    // Just return what the left branch was, not a RCSG node containing
    // it and EMPTY.
    return left_thing;

  } else {
    // Return an RCSG made up of the left and right branches.
    return new_rcsg(left_thing, right_thing, linkmode, num_prims);
  }
}

int rcsg_ray_intersect(IntersectionList *the_list, thing the_thing,
		       primnode *cluster_primnode,
		       vector ray_from, vector ray_dir, uint ray_flags,
		       signature ray_signature)
{
  // Did we intersect with anything?
  int left_result, right_result, found_int;
  intersection *list_position;

  if (IS_EMPTY(the_thing)) {
    return FALSE;
  } else if (IS_PRIM(the_thing)) {
    // This might be a FULL primnode, in which case there will be no
    // intersections with it in this voxel.
    if (IS_FULL(the_thing))
      return FALSE;

    // The primnode being intersected with.
    primnode *the_primnode = GET_PRIMNODE(the_thing);
    
    // Check to see that the primnode is really in this cell for this ray.
    if (check_ray_flags(ray_flags, the_primnode->flags)) {
      if (the_primnode->last_ray_signature == ray_signature) {
	// We've already intersected this ray with this
	// primitive.
	found_int = the_primnode->last_ray_hit;
      } else {
	// Intersect with the primitive.
	list_position = the_list->get_position();
	found_int = the_primnode->prim->ray_intersect(the_list, the_primnode,
						      cluster_primnode,
						      ray_from, ray_dir,
						      ray_flags,
						      ray_signature);
	the_list->set_position(list_position);
      }

      return found_int;
    } else {
      return FALSE;
    }

  } else {
    // This is an RCSGnode that we're intersecting with.
    rcsgnode *the_rcsg = GET_RCSGNODE(the_thing);

    left_result  = rcsg_ray_intersect(the_list, the_rcsg->left_thing,
				      cluster_primnode,
				      ray_from, ray_dir, ray_flags,
				      ray_signature);
    right_result = rcsg_ray_intersect(the_list, the_rcsg->right_thing,
				      cluster_primnode,
				      ray_from, ray_dir, ray_flags,
				      ray_signature);

    return left_result || right_result;
  }
}

void classify_intersection_in_rcsg(thing the_thing, intersection *the_int,
				   vector hit_point, intersect_result *result,
				   IntersectInfo *int_info, uint ray_flags)
{
  if (IS_PRIM(the_thing)) {
    primnode *the_prim = GET_PRIMNODE(the_thing);

    // Check to see that the primnode is really in this cell for this ray.
    if (check_ray_flags(ray_flags, the_prim->flags)) {
      // If the intersection isn't with this primitive, then we must do a
      // point intersection test.
      if (the_int->primitivehit != the_prim) {
	// Use the primitive's point intersection test to classify the hitpoint
	// as IN or OUT with respect to this primitive.
	//
	// If IN is placed in *result, then the point-intersection will
	// place the primitive's material in *material1.
	// If OUT is placed in *result, *material1 will be untouched.
	*result = the_prim->prim->point_intersect(the_prim, hit_point,
						&(int_info->contents),
						  ray_flags);
	return;
      } else {
	// The intersection is with this primitive. It is a SURFACE
	// intersection.
	//
	// This means that the materials the ray is coming from / going to with
	// respect to this primitive will be in the intersection in the list.
	//
	// Copy the primitive's intersection from the list into the int_info.
	*result = SURFACE;
	copy_intersect_info(int_info, &(the_int->info));
	return;
      }
    } else {
      // The primitive doesn't exist for the purpose of this ray as
      // check_ray_flags failed. The primitive is OUT.
      *result = OUT;
      return;
    }
  } else {
    rcsgnode *the_rcsg = GET_RCSGNODE(the_thing);
    linktype link_mode;
    IntersectInfo rightinfo, leftinfo;
    intersect_result right_result, left_result;

    // This thing isn't a primnode, it's an rcsgnode.

    // Extract the linktype of this RCSG node.
    link_mode = the_rcsg->linkmode;

    // Whatever the linktype is, we need to classify the point with respect
    // to the right branch of the RCSG tree.
    classify_intersection_in_rcsg(the_rcsg->right_thing, the_int, hit_point,
				  &right_result, &rightinfo, ray_flags);

    switch(link_mode) {
    case PLUS : {
      // The CSG table that governs this is ....
      //  A += B
      //
      //  A / B->   | IN(1)     OUT        SURF(1, 2)
      // -----------+--------------------------------
      //   IN(3)    | IN(1)     IN(3)         *
      //   OUT      | IN(1)     OUT        SURF(1, 2)
      // SURF(3,4)  | IN(1)   SURF(3,4)    impossible
      // 
      // * signifies that if 1 = NULL, SURF(3, 2)
      //                     2 = NULL, SURF(1, 3)
      //                     else,     SURF(1, 2)
      
      if (right_result == OUT) {
	// We are adding OUT to some object, result is the previous
	// object (the left branch).
	// Recursively classify the left branch and return it.
	classify_intersection_in_rcsg(the_rcsg->left_thing,
				      the_int, hit_point, &left_result,
				      &leftinfo, ray_flags);
	
	// We return the result of the left node.
	copy_intersect_info(int_info, &leftinfo);
	*result = left_result;
	return;
      } else if (right_result == IN) {
	// We return the right's IN result as it overwrites the left branch.
	copy_point_info(&(int_info->contents), &(rightinfo.contents));
	*result = IN;
	return;
      } else if (right_result == SURFACE) {
	// The right branch has a surface here. We must classify the
	// left branch to determine the return value.
	classify_intersection_in_rcsg(the_rcsg->left_thing,
				      the_int, hit_point, &left_result,
				      &leftinfo, ray_flags);
	
	if (left_result == OUT) {
	  // Return the surface from the right branch.
	  copy_intersect_info(int_info, &rightinfo);
	  *result = right_result;
	  return;
	} else if (left_result == IN) {
	  // If the right branch's surface is into empty space on either
	  // side, then replace this with the left branch's material.
	  // Since the surface is from the right branch, use the right
	  // branch's texture.
	  if (rightinfo.from.empty)
	    copy_point_info(&(int_info->from), &(leftinfo.contents));
	  else
	    copy_point_info(&(int_info->from), &(rightinfo.from));

	  if (rightinfo.to.empty)
	    copy_point_info(&(int_info->to), &(leftinfo.contents));
	  else
	    copy_point_info(&(int_info->to), &(rightinfo.to));

	  int_info->normal_flipped = rightinfo.normal_flipped;

	  // If the final intersection is between two identical materials
	  // then this isn't a real surface intersection. It is instead
	  // IN, containing the material.
	  if (point_info_equal(&(int_info->from), &(int_info->to))) {
	    *result = IN;
	  } else {
	    *result = SURFACE;
	  }
	  return;
	} else if (left_result == SURFACE) {
	  fatal_error("classify_intersection_in_rcsg", "SURFACE x2 conflict",
		      "The left and right branches of the RCSG tree both"
		      "returned a SURFACE classification of the currently"
		      "considered intersection point. This absolutely"
		      "should not happen.");
	} else {
	  fatal_error("classify_intersection_in_rcsg", "unknown classification",
		      "The left branch of the RCSG returned an unknown"
		      "classification of the intersection point (not IN, OUT"
		      "or SURFACE). This error should not happen.");
	}
      } else {
	fatal_error("classify_intersection_in_rcsg", "unknown classification",
		    "The right branch of the RCSG returned an unknown"
		    "classification of the intersection point (not IN, OUT"
		    "or SURFACE). This error should not happen.");
      }
    }
    case MINUS : {
      // The CSG table that governs this is ....
      //  A -= B
      //
      //  A / B->   | IN(1)     OUT        SURF(1, 2)
      // -----------+--------------------------------
      //   IN(3)    |  OUT     IN(3)         *
      //   OUT      |  OUT     OUT          OUT
      // SURF(3,4)  |  OUT   SURF(3,4)    impossible
      // 
      // * signifies that if 1 = NULL, SURF(3, NULL)
      //                     2 = NULL, SURF(NULL, 3)
      //                     else,     OUT
      
      if (right_result == IN) {
	// A is completely subtracted out by B, return OUT.
	*result = OUT;
	return;
      } else if (right_result == OUT) {
	// Recursively classify the left branch.
	classify_intersection_in_rcsg(the_rcsg->left_thing,
				      the_int, hit_point, &left_result,
				      &leftinfo, ray_flags);
	
	// We return the result of the left node.
	copy_intersect_info(int_info, &leftinfo);
	*result = left_result;
	return;
      } else if (right_result == SURFACE) {
	// The right branch has a surface here. We must classify the
	// left branch to determine the return value.
	classify_intersection_in_rcsg(the_rcsg->left_thing,
				      the_int, hit_point, &left_result,
				      &leftinfo, ray_flags);
	
	if (left_result == OUT) {
	  // There is no A here in the first place to be subtracted from,
	  // return OUT.
	  *result = OUT;
	} else if (left_result == IN) {
	  // If the right branch's surface is into empty space on either
	  // side, then we return a surface that has the left branch's
	  // material where the empty space is and nothing where the
	  // right branch's material was. The texture is taken from the
	  // left branch.
	  // Otherwise, if a surface between two non-NULL materials,
	  // we return nothing (OUT).

	  if (rightinfo.from.empty) {
	    copy_point_info(&(int_info->from), &(leftinfo.contents));
	    int_info->to.empty = TRUE;
	    int_info->normal_flipped = !(rightinfo.normal_flipped);
	    *result = SURFACE;
	  } else if (rightinfo.to.empty) {
	    copy_point_info(&(int_info->to), &(leftinfo.contents));
	    int_info->from.empty = TRUE;
	    int_info->normal_flipped = !(rightinfo.normal_flipped);
	    *result = SURFACE;
	  } else {
	    *result = OUT;
	  }
	  return;
	} else if (left_result == SURFACE) {
	  fatal_error("classify_intersection_in_rcsg", "SURFACE x2 conflict",
		      "The left and right branches of the RCSG tree both"
		      "returned a SURFACE classification of the currently"
		      "considered intersection point. This absolutely"
		      "should not happen.");
	} else {
	  fatal_error("classify_intersection_in_rcsg", "unknown classification",
		      "The left branch of the RCSG returned an unknown"
		      "classification of the intersection point (not IN, OUT"
		      "or SURFACE). This error should not happen.");
	}
      } else {
	fatal_error("classify_intersection_in_rcsg", "unknown classification",
		    "The right branch of the RCSG returned an unknown"
		    "classification of the intersection point (not IN, OUT"
		    "or SURFACE). This error should not happen.");
      }
    }
    case PAINTMINUS : {
      // The CSG table that governs this is ....
      //  A p-= B
      //
      //  A / B->   | IN(1)     OUT        SURF(1, 2)
      // -----------+--------------------------------
      //   IN(3)    |  OUT     IN(3)         *
      //   OUT      |  OUT     OUT          OUT
      // SURF(3,4)  |  OUT   SURF(3,4)    impossible
      // 
      // * signifies that if 1 = NULL, SURF(2, NULL)
      //                     2 = NULL, SURF(NULL, 1)
      //                     else,     OUT
      //
      // This is identical to the MINUS operator except that
      // if B (the subtracting object) has a surface in the
      // region, the resulting surface's material is taken from the
      // subtracting object rather than the object subtracted from.
      
      if (right_result == IN) {
	// A is completely subtracted out by B, return OUT.
	*result = OUT;
	return;
      } else if (right_result == OUT) {
	// Recursively classify the left branch.
	classify_intersection_in_rcsg(the_rcsg->left_thing,
				      the_int, hit_point, &left_result,
				      &leftinfo, ray_flags);
      
	// We return the result of the left node.
	copy_intersect_info(int_info, &leftinfo);
	*result = left_result;
	return;
      } else if (right_result == SURFACE) {
	// The right branch has a surface here. We must classify the
	// left branch to determine the return value.
	classify_intersection_in_rcsg(the_rcsg->left_thing,
				      the_int, hit_point, &left_result,
				      &leftinfo, ray_flags);
      
	if (left_result == OUT) {
	  // There is no A here in the first place to be subtracted from,
	  // return OUT.
	  *result = OUT;
	} else if (left_result == IN) {
	  // If the right branch's surface is into empty space on either
	  // side, then we return a surface that has the right branch's
	  // material where the empty space is and nothing where the
	  // right branch's material was. We use the right branch's texture.
	  // Otherwise, if a surface between two non-NULL materials,
	  // we return nothing (OUT).

	  if (rightinfo.from.empty) {
	    copy_point_info(&(int_info->from), &(rightinfo.to));
	    int_info->to.empty = TRUE;
	    int_info->normal_flipped = !(rightinfo.normal_flipped);
	    *result = SURFACE;
	  } else if (rightinfo.to.empty) {
	    copy_point_info(&(int_info->to), &(rightinfo.from));
	    int_info->from.empty = TRUE;
	    int_info->normal_flipped = !(rightinfo.normal_flipped);
	    *result = SURFACE;
	  } else {
	    *result = OUT;
	  }
	  return;
	} else if (left_result == SURFACE) {
	  fatal_error("classify_intersection_in_rcsg", "SURFACE x2 conflict",
		      "The left and right branches of the RCSG tree both"
		      "returned a SURFACE classification of the currently"
		      "considered intersection point. This absolutely"
		      "should not happen.");
	} else {
	  fatal_error("classify_intersection_in_rcsg", "unknown classification",
		      "The left branch of the RCSG returned an unknown"
		      "classification of the intersection point (not IN, OUT"
		      "or SURFACE). This error should not happen.");
	}
      } else {
	fatal_error("classify_intersection_in_rcsg", "unknown classification",
		    "The right branch of the RCSG returned an unknown"
		    "classification of the intersection point (not IN, OUT"
		    "or SURFACE). This error should not happen.");
      }
    }
    case AND : {
      // The CSG table that governs this is ....
      //  A &= B
      //
      //  A / B->   | IN(1)     OUT        SURF(1, 2)
      // -----------+--------------------------------
      //   IN(3)    | IN(1)     OUT      SURF(1, 2)
      //   OUT      |  OUT      OUT         OUT
      // SURF(3,4)  |   *       OUT      impossible
      // 
      // * signifies that if 3 = NULL, SURF(NULL, 1)
      //                     4 = NULL, SURF(2, NULL)
      //                     else,     IN(1)
      
      // We need to classify the left node no matter what the
      // value of the right node.
      classify_intersection_in_rcsg(the_rcsg->left_thing,
				    the_int, hit_point, &left_result,
				    &leftinfo, ray_flags);

      if (right_result == OUT) {
	// No intersection with B, return OUT.
	*result = OUT;
	return;
      }
      
      if (left_result == OUT) {
	// No object to perform an intersection on, return OUT.
	*result = OUT;
	return;
      } else if (left_result == IN) {
	if (right_result == IN) {
	  // The intersection succeeds, return B's material (due to
	  // the asymmetric nature of the intersection).
	  copy_point_info(&(int_info->contents), &(rightinfo.contents));
	  *result = IN;
	  return;
	} else if (right_result == SURFACE) {
	  // A is IN, B has a surface. Return B's surface.
	  copy_intersect_info(int_info, &rightinfo);
	  *result = SURFACE;
	  return;
	} else {
	  fatal_error("classify_intersection_in_rcsg",
		      "unknown classification",
		      "The right branch of the RCSG returned an unknown"
		      "classification of the intersection point (not IN, OUT"
		      "or SURFACE). This error should not happen.");
	}
      } else if (left_result == SURFACE) {
	// A has a surface, B is in. If A's surface is between something
	// and empty space, return a surface with A's material replaced by
	// B's material. Otherwise (if A's surface is between two different
	// materials) return IN with B's material.

	if (leftinfo.from.empty) {
	  int_info->from.empty = TRUE;
	  copy_point_info(&(int_info->to), &(rightinfo.contents));
	  int_info->normal_flipped = leftinfo.normal_flipped;
	  *result = SURFACE;
	  return;
	} else if (leftinfo.to.empty) {
	  copy_point_info(&(int_info->from), &(rightinfo.contents));
	  int_info->to.empty = TRUE;
	  int_info->normal_flipped = leftinfo.normal_flipped;
	  *result = SURFACE;
	  return;
	} else {
	  copy_point_info(&(int_info->contents), &(rightinfo.contents));
	  *result = IN;
	  return;
	}
      } else {
	fatal_error("classify_intersection_in_rcsg", "unknown classification",
		    "The right branch of the RCSG returned an unknown"
		    "classification of the intersection point (not IN, OUT"
		    "or SURFACE). This error should not happen.");
      }
    }
    case SURF_AND : {
      // This preserves the material properties of the left object at
      // any surfaces that belonged to the left object.
      //
      // The CSG table that governs this is ....
      //  A &= B
      //
      //  A / B->   | IN(1)     OUT        SURF(1, 2)
      // -----------+--------------------------------
      //   IN(3)    | IN(1)     OUT      SURF(1, 2)
      //   OUT      |  OUT      OUT         OUT
      // SURF(3,4)  |SURF(3,4)  OUT      impossible
      
      // We need to classify the left node no matter what the
      // value of the right node.
      classify_intersection_in_rcsg(the_rcsg->left_thing,
				    the_int, hit_point, &left_result,
				    &leftinfo, ray_flags);

      if (right_result == OUT) {
	// No intersection with B, return OUT.
	*result = OUT;
	return;
      }
      
      if (left_result == OUT) {
	// No object to perform an intersection on, return OUT.
	*result = OUT;
	return;
      } else if (left_result == IN) {
	if (right_result == IN) {
	  // The intersection succeeds, return B's material (due to
	  // the asymmetric nature of the intersection).
	  copy_point_info(&(int_info->contents), &(rightinfo.contents));
	  *result = IN;
	  return;
	} else if (right_result == SURFACE) {
	  // A is IN, B has a surface. Return B's surface.
	  copy_intersect_info(int_info, &rightinfo);
	  *result = SURFACE;
	  return;
	} else {
	  fatal_error("classify_intersection_in_rcsg",
		      "unknown classification",
		      "The right branch of the RCSG returned an unknown"
		      "classification of the intersection point (not IN, OUT"
		      "or SURFACE). This error should not happen.");
	}
      } else if (left_result == SURFACE) {
	// A has a surface, B is in. Return A's surface.
	copy_intersect_info(int_info, &leftinfo);
	*result = SURFACE;
	return;
      } else {
	fatal_error("classify_intersection_in_rcsg", "unknown classification",
		    "The right branch of the RCSG returned an unknown"
		    "classification of the intersection point (not IN, OUT"
		    "or SURFACE). This error should not happen.");
      }
    }
    default : {
      fatal_error("classify_intersection_in_rcsg", "unknown link type in rcsg",
		  "The link-type in the RCSG graph being traversed does not"
		  "fit any of the defined link types (PLUS, MINUS, PAINTMINUS,"
		  "AND or SURF_AND). This error should not happen.");
    }
    }
  }
}

intersect_result classify_point_in_rcsg(thing the_thing, vector the_point,
					PointInfo *point_info,
					uint ray_flags)
{
  if (IS_PRIM(the_thing)) {
    // This thing is a PRIMNODE.
    primnode *the_prim = GET_PRIMNODE(the_thing);
    
    // Check to see if the primitive isn't really there for this ray.
    if (!check_ray_flags(ray_flags, the_prim->flags))
      return OUT;
    
    // Use the primitive's point intersection test to classify the hitpoint
    // as IN or OUT with respect to this primitive.
    //
    // If the point is IN, the point-intersection routines will place the
    // primitive's material in point_info.
    // If the point is OUT, point_info will be untouched.
    return the_prim->prim->point_intersect(the_prim, the_point,
					   point_info, ray_flags);
  } else {
    // This thing is an RCSGnode.
    rcsgnode *the_rcsg = GET_RCSGNODE(the_thing);
    linktype link_mode;
    PointInfo rightinfo, leftinfo;
    intersect_result right_result, left_result;
    
    // Extract the linktype of this RCSG node.
    link_mode = the_rcsg->linkmode;

    // Whatever the linktype is, we need to classify the point with respect
    // to the right branch of the RCSG tree.
    right_result = classify_point_in_rcsg(the_rcsg->right_thing,
					  the_point, &rightinfo, ray_flags);
    
    switch(link_mode) {
    case PLUS : {
      // The CSG table that governs this is ....
      //  A += B
      //
      //  A / B->   | IN(1)     OUT    
      // -----------+------------------
      //   IN(2)    | IN(1)     IN(2)         
      //   OUT      | IN(1)     OUT    
      // 
      
      if (right_result == IN) {
	// We return IN, material1 == right1 == the material of the right node
	copy_point_info(point_info, &rightinfo);
	return IN;
      } else if (right_result == OUT) {
	// Recursively classify the left branch.
	left_result = classify_point_in_rcsg(the_rcsg->left_thing,
					     the_point, &leftinfo, ray_flags);

	// We return the result of the left node.
	copy_point_info(point_info, &leftinfo);
	return left_result;
      } else {
	fatal_error("classify_point_in_rcsg", "unknown classification",
		    "The right branch of the RCSG returned an unknown"
		    "classification of the point (not IN or OUT)."
		    "This error should not happen.");
      }
    }
    case PAINTMINUS :
      // Note that PAINTMINUS is identical to MINUS for the purposes of
      // in/out tests.
    case MINUS : {
      // The CSG table that governs this is ....
      //  A -= B
      //
      //  A / B->   | IN(1)     OUT    
      // -----------+-----------------
      //   IN(2)    |  OUT     IN(2)         
      //   OUT      |  OUT     OUT     
      // 
      
      if (right_result == IN) {
	// A is completely subtracted out by B, return OUT.
	return OUT;
      } else if (right_result == OUT) {
	// Recursively classify the left branch.
	left_result = classify_point_in_rcsg(the_rcsg->left_thing,
					     the_point, &leftinfo, ray_flags);
	
	// We return the result of the left node.
	copy_point_info(point_info, &leftinfo);
	return left_result;
      } else {
	fatal_error("classify_point_in_rcsg", "unknown classification",
		    "The right branch of the RCSG returned an unknown"
		    "classification of the point (not IN or OUT)."
		    "This error should not happen.");
      }
    }
    case SURF_AND :
      // Note that SURF_AND is identical to AND for the purposes of
      // in/out tests.
    case AND : {
      // The CSG table that governs this is ....
      //  A &= B
      //
      //  A / B->   | IN(1)     OUT 
      // -----------+----------------
      //   IN(3)    | IN(1)     OUT  
      //   OUT      |  OUT      OUT  
      
      // We need to classify the left node no matter what the
      // value of the right node.
      left_result = classify_point_in_rcsg(the_rcsg->left_thing,
					   the_point, &leftinfo, ray_flags);

      if (right_result == OUT) {
	// No intersection with B, return OUT.
	return OUT;
      }
      
      if (left_result == OUT) {
	// No object to perform an intersection on, return OUT.
	return OUT;
      } else if (left_result == IN) {
	if (right_result == IN) {
	  // The intersection succeeds, return B's material (due to
	  // the asymmetric nature of the intersection).
	  copy_point_info(point_info, &rightinfo);
	  return IN;
	} else {
	  fatal_error("classify_point_in_rcsg", "unknown classification",
		      "The right branch of the RCSG returned an unknown"
		      "classification of the point (not IN or OUT)."
		      "This error should not happen.");
	}
      } else {
	fatal_error("classify_point_in_rcsg", "unknown classification",
		    "The right branch of the RCSG returned an unknown"
		    "classification of the point (not IN or OUT)."
		    "This error should not happen.");
      }
    }
    default : {
      fatal_error("classify_point_in_rcsg", "unknown link type in rcsg",
		  "The link-type in the RCSG graph being traversed does not"
		  "fit any of the defined link types (PLUS, MINUS, PAINTMINUS,"
		  "AND, or SURF_AND). This error should not happen.");
    }
    }
    // Should never get here.
    fatal_error("classify_point_in_rcsg", "control logic error",
		"This error should not happen.");
    return OUT;
  }
}

intersect_result classify_voxel_in_rcsg(thing the_thing, box b)
{
  if (IS_PRIM(the_thing)) {
    primnode *the_primnode = GET_PRIMNODE(the_thing);
    box primspace_box;
    int i;

    // Transform the box into the primnode's space.
    for (i = 0; i < 8; i++)
      Matrix::apply(the_primnode->position[AFT], b[i], primspace_box[i]);
    
    return the_primnode->prim->voxel_intersect(primspace_box);
  } else if (IS_RCSG(the_thing)) {

    // This thing is an RCSGnode.
    rcsgnode *the_rcsg = GET_RCSGNODE(the_thing);
    intersect_result right, left;
    
    // Whatever the linktype is, we need to classify the voxel with respect
    // to the right branch of the RCSG tree.
    right = classify_voxel_in_rcsg(the_rcsg->right_thing, b);
    
    switch(the_rcsg->linkmode) {
    case PLUS : {
      // The CSG table that governs this is ....
      //  A += B
      //
      //  A / B  | IN   OUT  S
      // --------+--------------
      //   IN    | IN   IN   S
      //   OUT   | IN   OUT  S
      //   S     | IN   S    S
      
      if (right == OUT)
	return classify_voxel_in_rcsg(the_rcsg->left_thing, b);
      else
	return right;
    }
    case PAINTMINUS :
      // Note that PAINTMINUS is identical to MINUS for the purposes of
      // in/out tests.
    case MINUS : {
      // The CSG table that governs this is ....
      //  A -= B
      //
      //  A / B->| IN   OUT  S   
      // --------+--------------
      //   IN    | OUT  IN   S   
      //   OUT   | OUT  OUT  OUT  
      //    S    | OUT  S    S
      
      if (right == IN) {
	return OUT;
      } else {
	// Recursively classify the left branch.
	left = classify_voxel_in_rcsg(the_rcsg->left_thing, b);

	if (left == OUT)
	  return OUT;
	else if (right == OUT)
	  return left;
	else
	  return SURFACE;
      }
    }
    case SURF_AND :
      // Note that SURF_AND is identical to AND for the purposes of
      // in/out tests.
    case AND : {
      // The CSG table that governs this is ....
      //  A &= B
      //
      //  A / B->| IN   OUT  S
      // --------+---------------
      //   IN    | IN   OUT  S
      //   OUT   | OUT  OUT  OUT
      //   S     | S    OUT  S
      
      if (right == OUT)
	return OUT;
      else {
	left = classify_voxel_in_rcsg(the_rcsg->left_thing, b);
	
	if (left == OUT)
	  return OUT;
	else if (left == SURFACE)
	  return SURFACE;
	else
	  return right;
      }
    }
    default : {
      fatal_error("classify_voxel_in_rcsg", "unknown link type in rcsg",
		  "The link-type in the RCSG graph being traversed does not"
		  "fit any of the defined link types (PLUS, MINUS, PAINTMINUS,"
		  "or AND). This error should not happen.");
    }
    }
    // Should never get here.
    fatal_error("classify_voxel_in_rcsg", "control logic error",
		"This error should not happen.");
    return OUT;
  } else {
    fatal_error("classify_voxel_in_rcsg", "passed a non-PRIM, non-RCSG",
		"This error should not happen.");
    return OUT;
  }
}

void tidy_intersections(IntersectionList *the_list, 
			primnode *cluster_primnode)
{
  intersection *current_intersection, *present_position;

#if DEBUG_VALIDATE
  printf("TIDY: process cluster's intersections\n");
#endif

  // We have reached the end of this cluster. Delete all intersections
  // that were inserted in previous cells of the cluster, but have not
  // been validated.
  present_position = the_list->get_position();

  the_list->restart();
  while (!the_list->empty()) {
    current_intersection = the_list->get_next();
#if DEBUG_VALIDATE
    printf("TIDY: process %f\n", current_intersection->t_value);
#endif
    if (current_intersection->validate_primnode == cluster_primnode) {
#if DEBUG_VALIDATE
      printf("TIDY: remove\n");
#endif
      the_list->remove_intersection(current_intersection);
    }
  }
  the_list->set_position(present_position);
}

int validate_intersections(IntersectionList *the_list,
			   primnode *cluster_primnode,
			   primnode *next_cluster,
			   vector ray_from, vector ray_dir,
			   uint ray_flags, validate_type v_type,
			   thing the_thing,
			   double cell_entry_t, double cell_exit_t)
{
  intersection *current_intersection;
  uint int_flags;
  vector hit_point, normal_in_cluster_space;
  intersect_result int_result;
  IntersectInfo currentinfo;
  double current_t;
  primnode *real_primnode;
  int found_solid;

  found_solid = FALSE;

  // Keep processing intersections until they are all gone.
  while (!the_list->empty()) {
    current_intersection = the_list->get_next();
#if DEBUG_VALIDATE
    printf("VAL: get int %f\n", current_intersection->t_value);
#endif

    if (current_intersection->validate_primnode != cluster_primnode) {
      // This intersection does not need to be validated.
#if DEBUG_VALIDATE
      printf("VAL: unneeded %s %s\n", (current_intersection->validate_primnode == NULL) ? "NULL" : current_intersection->validate_primnode->prim->prim_name(), (cluster_primnode == NULL) ? "NULL" : cluster_primnode->prim->prim_name());
#endif
      continue;
    }

    // Find the current intersection's distance.
    int_flags = current_intersection->hit_flags;
    current_t = current_intersection->t_value;

    // First, we must check to see if this intersection is in front of the
    // current subdivision cell (if it has t < cell_entry_t).
    if (current_t < cell_entry_t) {
      // This intersection is invalid because it's before the
      // current subdivision cell. It's also going to be before any
      // subsequent cells, so we may delete it.
#if DEBUG_VALIDATE
      printf("VAL: remove %f < %f\n", current_t, cell_entry_t);
#endif
      the_list->remove_intersection(current_intersection);
      continue;
    } else if (current_t >= cell_exit_t) {
      // This intersection is invalid because it's after the current
      // subdivision cell. We leave it in the intersection list and
      // don't modify it. Since the intersections are
      // stored in order, all the ones after this must be for the next
      // cell ... we must stop here and leave them all.
#if DEBUG_VALIDATE
      printf("VAL: stop, %f > %f\n", current_t, cell_exit_t);
#endif
      the_list->undo_last_retrieve();
      return found_solid;
    }

    // Get the primnode that we _really_ hit (not the cluster that's generating
    // the intersection).
    real_primnode = current_intersection->real_primnode;

    //    printf("Classifying an intersection with a %s\n",
    //    	   real_primnode->prim->prim_name());

    // If this is an invalid intersection due to the CSG structure of the
    // RCSG tree, then remove it. Note that this is only possible when
    // validating intersections with respect to an RCSG.
    if (v_type == VALIDATE_RCSG) {
      Vector::mult(current_t, ray_dir, hit_point);
      Vector::add(ray_from, hit_point, hit_point);
      //      printf("Classify_intersection_in_rcsg: debug starting RCSG\n");
      //      debug_thing(the_thing, 0);
      classify_intersection_in_rcsg(the_thing, current_intersection,
				    hit_point, &int_result,
				    &currentinfo, ray_flags);
    
      if (int_result != SURFACE) {
	// If the intersection is IN or OUT with respect to the
	// RCSG graph, discard it.
	//      printf("invalid\n");
#if DEBUG_VALIDATE
	printf("VAL: remove invalid\n");
#endif
	the_list->remove_intersection(current_intersection);
	continue;
      }
      //      printf("Classify_intersection_in_rcsg: valid int, debug starting RCSG\n");
      //      debug_thing(the_thing, 0);
      

#if DEBUG_VALIDATE
      printf("VAL: valid wrt RCSG\n");
#endif
    } else {
      // We're classifying intersections with respect to a single primnode
      // in this cell; if we find intersections with other primnodes, we can
      // safely ignore them.
      if (current_intersection->primitivehit != GET_PRIMNODE(the_thing)) {
	// Delete this intersection.
	//	printf("VAL: remove int/n w/ other primnode\n");
	the_list->remove_intersection(current_intersection);
	continue;
      }	

      // We have hit a single primnode, not an RCSG; we don't have to
      // classify the intersection as we know what kind it is already.
#if DEBUG_VALIDATE
      printf("VAL: hit ");
      if (current_intersection->info.to.empty)
	printf("ether / ");
      else {
	DSTmaterial *tmp_mat = (DSTmaterial *) (current_intersection->info.to.mat);
	printf("(%f %f %f) / ", tmp_mat->diffuse[R], tmp_mat->diffuse[G], tmp_mat->diffuse[B]);
      }
      if (current_intersection->info.from.empty)
	printf("ether\n");
      else {
	DSTmaterial *tmp_mat = (DSTmaterial *) (current_intersection->info.from.mat);
	printf("(%f %f %f)\n", tmp_mat->diffuse[R], tmp_mat->diffuse[G], tmp_mat->diffuse[B]);
      }
#endif
      copy_intersect_info(&currentinfo, &(current_intersection->info));
    }

    // First we must transform the normal back from the primitive's
    // primitive space into the cluster's primitive space. We do
    // this using the transpose of the primitive's aft position matrix.
    Matrix::applytranspose(current_intersection->primitivehit->position[AFT],
			   current_intersection->normal,
			   normal_in_cluster_space);

    if (currentinfo.normal_flipped) {
      Vector::mult(-1, normal_in_cluster_space, normal_in_cluster_space);
      currentinfo.normal_flipped = FALSE;
    }

    //    printf("VAL: updating\n");

    // The intersection is valid. Modify it to look like it comes from
    // the cluster primitive, and go on to the next one.
    current_intersection->hit_flags |= cluster_primnode->flags;
    current_intersection->primitivehit = cluster_primnode;
    Vector::set(current_intersection->normal, normal_in_cluster_space);
    copy_intersect_info(&(current_intersection->info), &currentinfo);
    
    // Note that the intersection will next have to be validated by the
    // next higher cluster up the chain.
    current_intersection->validate_primnode = next_cluster;

    // If we find a solid surface when looking for surfaces, or a
    // non-translucent object when looking for shadows, we don't need
    // to keep looking through more RCSG cells and we can stop.
    if (found_solid)
      // FIXME: should this be break? And should it be after the following
      // code paragraph? And should it look at the cluster primitive's
      // stop-at-first?
      continue;

    if ((!point_info_equal(&(currentinfo.from), &(currentinfo.to))) && 
	(((ray_flags & RAY_SHADOW) == 0) ||
	 ((current_intersection->hit_flags & FLAGS_TRANSLUCENT) == 0)))
      found_solid = TRUE;
    
    continue;
  }

  return found_solid;
}

int check_ray_flags(uint ray_flags, uint prim_flags)
{
  if ((ray_flags & RAY_REFLECTED) && (prim_flags & FLAGS_NOREFLECT))
    return FALSE;

  if ((ray_flags & RAY_TRANSMITTED) && (prim_flags & FLAGS_NOTRANSMIT))
    return FALSE;

  if ((ray_flags & RAY_SHADOW) && (prim_flags & FLAGS_NOSHADOW))
    return FALSE;

  if ((ray_flags & RAY_THROUGH_GATEWAY) && 
      (prim_flags & FLAGS_NOT_THROUGH_GATEWAY))
    return FALSE;

  if (((ray_flags & RAY_PRIMARYMASK) == 0) && (prim_flags & FLAGS_NOPRIMARY))
    return FALSE;

  if (((ray_flags & RAY_TRANSMITTED) == 0) &&
      (prim_flags & FLAGS_ONLYTRANSMIT))
    return FALSE;

  if (((ray_flags & RAY_THROUGH_GATEWAY) == 0) &&
      (prim_flags & FLAGS_ONLY_THROUGH_GATEWAY))
    return FALSE;

  return TRUE;
}

void find_point_sfx(primnode *world_object, vector world_point,
		    uint ray_flags, SfxType *sfxtype, void **sfx)
{
  PointInfo pt_info;
  intersect_result pt_result;

  pt_result = world_object->prim->point_intersect(world_object, world_point,
						  &pt_info, ray_flags);
  if ((pt_result == OUT) || (pt_info.empty)) {
    *sfxtype = ether->sfxtype;
    *sfx = ether->sfx;
  } else {
    *sfxtype = pt_info.sfxtype;
    *sfx = pt_info.sfx;
  }
}

#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif
