/* prim_cone.c : The cone primitive
** 
** The cone primitive defines an infinite double right circular cone
** with its vertex at the origin, oriented along the Z axis.
**
** Chris Butcher, 11 December 1996
*/

#include "mirage.h"

#ifdef USE_NAMESPACES
namespace Mirage {
#endif

int conePrimitive::ray_intersect(IntersectionList *the_list, 
				 struct PrimitiveNode *node, 
				 primnode *validate_node, vector ray_from,
				 vector ray_dir, uint ray_flags,
				 signature ray_signature)
{
  // The primitive-space ray.
  double p_x, p_y, p_z, d_x, d_y, d_z;
  // The coefficients of the quadratic equation in t.
  double a, b, c;
  // The discriminant of the quadratic equation.
  double disc;
  // The solutions of the equation.
  double t_0, t_1;
  // The (primitive-space) hitpoint.
  double hit_x, hit_y, hit_z;
  // The normal of the intersection point.
  vector norm;
  // Which half of the cone have we hit?
  double sign;
  // Have we found an intersection?
  int found = 0;
  // The midpoint t-value and corresponding point.
  double mid_t, mid_x, mid_y, mid_z;
  // Is the midpoint in?
  int midpoint_in;
  

  // Store the signature of this ray that we're intersecting with.
  node->last_ray_signature = ray_signature;

  // Get the primitive-space ray.
  Matrix::apply(node->position[AFT], ray_from, &p_x, &p_y, &p_z);
  Matrix::applynoshift(node->position[AFT], ray_dir, &d_x, &d_y, &d_z);

  // Get the quadratic equation that gives the intersections.
  a = SQR(d_x) + SQR(d_y) - SQR(d_z);
  b = 2.0 * (p_x*d_x + p_y*d_y - p_z*d_z);
  c = SQR(p_x) + SQR(p_y) - SQR(p_z);

  // Calculate the discriminant of the quadratic equation.
  disc = SQR(b) - 4.0*a*c;

  // If the discriminant is less than or equal to zero, then there
  // are no intersections with the cone (tangent rays are treated as
  // no intersection at all).
  if (disc <= 0) {
    node->last_ray_hit = FALSE;
    return node->last_ray_hit;
  }

  // Calculate the intersection t-values. Must have t_0 < t_1 for later
  // calculations to work.
  // Note that we can save a square root operation by saying
  // t_1 = c / a * t_0 ... but this only works if we can be sure that
  // t_0 isn't zero. It may easily be if sqrt(disc) == b or == -b.
  // So, we make sure that the +/-sqrt(disc) is always the same sign
  // as -b to make sure that the subtraction can't produce zero.
  // This makes the code a little long-winded.
  if (a >= 0) {
    if (b >= 0) {
      t_0 = (-b - sqrt(disc)) / (2 * a);
      t_1 = c / (a * t_0);
    } else {
      t_1 = (-b + sqrt(disc)) / (2 * a);
      t_0 = c / (a * t_1);
    }
  } else {
    if (b >= 0) {
      t_1 = (-b - sqrt(disc)) / (2 * a);
      t_0 = c / (a * t_1);
    } else {
      t_0 = (-b + sqrt(disc)) / (2 * a);
      t_1 = c / (a * t_0);
    }
  }

  // We must determine whether t_0 is an ENTERING or a LEAVING intersection.
  // We look at the point midway between t_0 and t_1 to see if it is IN or OUT.
  mid_t = (t_0 + t_1) / 2;
  mid_x = p_x + d_x*mid_t;
  mid_y = p_y + d_y*mid_t;
  mid_z = p_z + d_z*mid_t;

  midpoint_in = (SQR(mid_x) + SQR(mid_y) < SQR(mid_z));

  // Insert any positive t-values into the intersection list.
  if (t_0 > MIN_TVAL_CUTOFF) {
    // Calculate the hitpoint and normal. Hitpoint is (p + td)
    // and normal is given by (x * |z|, y * |z|, -(x^2 + y^2)*(z / |z|)).
    hit_x = p_x + d_x*t_0;
    hit_y = p_y + d_y*t_0;
    hit_z = p_z + d_z*t_0;

    if (hit_z >= 0)
      sign = +1.0;
    else
      sign = -1.0;

    norm[X] = hit_x * hit_z * sign;
    norm[Y] = hit_y * hit_z * sign;
    norm[Z] = -sign * (SQR(hit_x) + SQR(hit_y));

    // This is an ENTERING intersection if the midpoint is IN, a LEAVING
    // intersection if the midpoint is OUT.
    if (midpoint_in) {
      insert_entering(the_list, node, validate_node, norm, t_0,
		      ray_from, ray_dir, ray_flags);
    } else {
      insert_leaving(the_list, node, validate_node, norm, t_0,
		     ray_from, ray_dir, ray_flags);
    }
    found = 1;
  }
  if (t_1 > MIN_TVAL_CUTOFF) {
    // Calculate the hitpoint and normal. Hitpoint is (p + td)
    // and normal is given by (x * |z|, y * |z|, -(x^2 + y^2)*(z / |z|)).
    hit_x = p_x + d_x*t_1;
    hit_y = p_y + d_y*t_1;
    hit_z = p_z + d_z*t_1;

    if (hit_z >= 0)
      sign = +1.0;
    else
      sign = -1.0;

    norm[X] = hit_x * hit_z * sign;
    norm[Y] = hit_y * hit_z * sign;
    norm[Z] = -sign * (SQR(hit_x) + SQR(hit_y));

    // This is an ENTERING intersection if the midpoint was OUT, a LEAVING
    // intersection if the midpoint was IN.
    if (midpoint_in) {
      insert_leaving(the_list, node, validate_node, norm, t_1,
		     ray_from, ray_dir, ray_flags);
    } else {
      insert_entering(the_list, node, validate_node, norm, t_1,
		      ray_from, ray_dir, ray_flags);
    }
    found = 1;
  }

  if (found)
    node->last_ray_hit = TRUE;
  else
    node->last_ray_hit = FALSE;

  return node->last_ray_hit;
}

int conePrimitive::ray_intersect(vector ray_from,
				 vector ray_dir, rayintersect *int_storage)
{
  // The coefficients of the quadratic equation in t.
  double a, b, c;
  // The discriminant of the quadratic equation.
  double disc;
  // The solutions of the equation.
  double t_0, t_1;
  // The (primitive-space) hitpoint.
  double hit_x, hit_y, hit_z;
  // The normal of the intersection point.
  vector norm;
  // Which half of the cone have we hit?
  double sign;
  // Have we found an intersection?
  int num_intersections = 0;
  // The midpoint t-value and corresponding point.
  double mid_t, mid_x, mid_y, mid_z;
  // Is the midpoint in?
  int midpoint_in;
  

  // Get the quadratic equation that gives the intersections.
  a = SQR(ray_dir[X]) + SQR(ray_dir[Y]) - SQR(ray_dir[Z]);
  b = 2.0 * (ray_from[X]*ray_dir[X] + ray_from[Y]*ray_dir[Y]
	     - ray_from[Z]*ray_dir[Z]);
  c = SQR(ray_from[X]) + SQR(ray_from[Y]) - SQR(ray_from[Z]);

  // Calculate the discriminant of the quadratic equation.
  disc = SQR(b) - 4.0*a*c;

  // If the discriminant is less than or equal to zero, then there
  // are no intersections with the cone (tangent rays are treated as
  // no intersection at all).
  if (disc <= 0) {
    return 0;
  }

  // Calculate the intersection t-values. Must have t_0 < t_1 for later
  // calculations to work.
  // Note that we can save a square root operation by saying
  // t_1 = c / a * t_0 ... but this only works if we can be sure that
  // t_0 isn't zero. It may easily be if sqrt(disc) == b or == -b.
  // So, we make sure that the +/-sqrt(disc) is always the same sign
  // as -b to make sure that the subtraction can't produce zero.
  // This makes the code a little long-winded.
  if (a >= 0) {
    if (b >= 0) {
      t_0 = (-b - sqrt(disc)) / (2 * a);
      t_1 = c / (a * t_0);
    } else {
      t_1 = (-b + sqrt(disc)) / (2 * a);
      t_0 = c / (a * t_1);
    }
  } else {
    if (b >= 0) {
      t_1 = (-b - sqrt(disc)) / (2 * a);
      t_0 = c / (a * t_1);
    } else {
      t_0 = (-b + sqrt(disc)) / (2 * a);
      t_1 = c / (a * t_0);
    }
  }

  // We must determine whether t_0 is an ENTERING or a LEAVING intersection.
  // We look at the point midway between t_0 and t_1 to see if it is IN or OUT.
  mid_t = (t_0 + t_1) / 2;
  mid_x = ray_from[X] + ray_dir[X]*mid_t;
  mid_y = ray_from[Y] + ray_dir[Y]*mid_t;
  mid_z = ray_from[Z] + ray_dir[Z]*mid_t;

  midpoint_in = (SQR(mid_x) + SQR(mid_y) < SQR(mid_z));

  // Calculate the hitpoint and normal. Hitpoint is (p + td)
  // and normal is given by (x * |z|, y * |z|, -(x^2 + y^2)*(z / |z|)).
  hit_x = ray_from[X] + ray_dir[X]*t_0;
  hit_y = ray_from[Y] + ray_dir[Y]*t_0;
  hit_z = ray_from[Z] + ray_dir[Z]*t_0;
  
  if (hit_z >= 0)
    sign = +1.0;
  else
    sign = -1.0;
  
  norm[X] = hit_x * hit_z * sign;
  norm[Y] = hit_y * hit_z * sign;
  norm[Z] = -sign * (SQR(hit_x) + SQR(hit_y));
  
  // This is an ENTERING intersection if the midpoint is IN, a LEAVING
  // intersection if the midpoint is OUT.
  int_storage[num_intersections].t = t_0;
  int_storage[num_intersections].is_entering = (midpoint_in) ? TRUE : FALSE;
  Vector::set(int_storage[num_intersections].normal, norm);
  num_intersections++;

  // Calculate the hitpoint and normal. Hitpoint is (p + td)
  // and normal is given by (x * |z|, y * |z|, -(x^2 + y^2)*(z / |z|)).
  hit_x = ray_from[X] + ray_dir[X]*t_1;
  hit_y = ray_from[Y] + ray_dir[Y]*t_1;
  hit_z = ray_from[Z] + ray_dir[Z]*t_1;
  
  if (hit_z >= 0)
    sign = +1.0;
  else
    sign = -1.0;
  
  norm[X] = hit_x * hit_z * sign;
  norm[Y] = hit_y * hit_z * sign;
  norm[Z] = -sign * (SQR(hit_x) + SQR(hit_y));
  
  // This is an ENTERING intersection if the midpoint was OUT, a LEAVING
  // intersection if the midpoint was IN.
  int_storage[num_intersections].t = t_1;
  int_storage[num_intersections].is_entering = (midpoint_in) ? FALSE : TRUE;
  Vector::set(int_storage[num_intersections].normal, norm);
  num_intersections++;

  return num_intersections;
}

// Return IN or OUT depending on whether or not a given point is inside
// the cone.
intersect_result conePrimitive::point_intersect(primnode *node, vector wp,
						PointInfo *info,
						uint ray_flags)
{
  double p_x, p_y, p_z;

  // Get the primitive space coordinates of the world-space point wp.
  Matrix::apply(node->position[AFT], wp, &p_x, &p_y, &p_z);
  
  // If (x^2 + y^2 < z^2) then the point is in the right circular cone.
  if (SQR(p_x) + SQR(p_y) < SQR(p_z)) {
    set_point_contents(node, info);
    return IN;
  } else
    return OUT;
}

// Return IN or OUT depending on whether or not a given point is inside
// the cone.
intersect_result conePrimitive::point_intersect(vector wp)
{
  // If (x^2 + y^2 < z^2) then the point is in the right circular cone.
  if (SQR(wp[X]) + SQR(wp[Y]) < SQR(wp[Z]))
    return IN;
  else
    return OUT;
}

// Some local macros to make the code length shorter for the routine
// conePrimitive::voxel_intersect.

// Macro AXIS_SIDE calculates the side on which the Z axis falls of the
// edge defined by the points of the box i and j. The value is stored
// in axis_side[store] for later reference.
#define AXIS_SIDE(i, j, store)                                    \
                                                                  \
(axis_side[store] = ((v[i][Y]*v[j][X] - v[i][X]*v[j][Y]) >= 0.0))
     
// Macro CHECK_EDGE_INTERSECT finds any intersections with the cone
// for the given edge from points i to j by solving a quadratic equation.
// Return SURFACE if any are found.
#define CHECK_EDGE_INTERSECT(i, j)                                \
                                                                  \
px = v[i][X];        py = v[i][Y];        pz = v[i][Z];           \
dx = v[j][X] - px;   dy = v[j][Y] - py;   dz = v[j][Z] - pz;      \
                                                                  \
a = SQR(dx) + SQR(dy) - SQR(dz);  c = SQR(px) + SQR(py) - SQR(pz);\
b = 2.0 * (px*dx + py*dy - pz*dz);  disc = SQR(b) - 4.0*a*c;      \
                                                                  \
if (a == 0) return SURFACE;                                       \
                                                                  \
if (disc >= 0.0) {                                                \
  if (b >= 0.0)                                                   \
    t_0 = (-b - sqrt(disc)) / (2.0 * a);                          \
  else                                                            \
    t_0 = (-b + sqrt(disc)) / (2.0 * a);                          \
                                                                  \
  if (t_0 > 0 && t_0 < 1)                                         \
    return SURFACE;                                               \
                                                                  \
  t_1 = c / (a * t_0);                                            \
                                                                  \
  if (t_1 > 0 && t_1 < 1)                                         \
    return SURFACE;                                               \
}

intersect_result conePrimitive::voxel_intersect(box v)
{
  return cone_voxel_intersect(v);
}

intersect_result cone_voxel_intersect(box v)
{
  // This proceeds in a number of steps.
  // if positive and negative points, and some in, return SURFACE.
  // if some out and some in, return SURFACE.
  // if all eight in and all either positive or negative, return IN.

  // Check five faces of parallelipiped to see they contain the Z axis.
  // If any do, return SURFACE.
  
  // Check each edge of parallelipiped to see if it intersects the cone.
  // If any do, return SURFACE.

  // Otherwise, return OUT.

  int z_pos, z_neg, points_in, points_out, i;
  // Used by the AXIS_SIDE macro.
  double axis_side[12];
  // Used by the CHECK_EDGE_INTERSECT macro.
  double px, py, pz, dx, dy, dz, a, b, c, disc, t_0, t_1;
  
  points_in = points_out = FALSE;
  z_pos = z_neg = FALSE;

  // For each point, test to see whether it is in or out,
  // and in the positive or negative sides of the cone.
  for (i = 0; i < 8; i++) {
    if (SQR(v[i][X]) + SQR(v[i][Y]) < SQR(v[i][Z]))
      points_in = TRUE;
    else
      points_out = TRUE;

    if (v[i][Z] >= 0)
      z_pos = TRUE;
    else
      z_neg = TRUE;
  }

  // If we have both in and out points, return SURFACE. If
  // there are points in both +Z and -Z, then the parallelipiped
  // must cross Z = 0 and thus must have some points outside the
  // cone. Thus if Z+ and Z- and there are some points inside the
  // cone, return SURFACE.
  if ((points_in && points_out) ||
      (z_pos && z_neg && points_in))
    return SURFACE;

  // If all the points are in, return IN. Note that to get to
  // this stage the points must be either all in (+Z), all in (-Z)
  // or all out.
  if (points_in)
    return IN;
  
  // All the points are out. Now check to see if the Z axis passes
  // through the box. If it does then the cone intersects the box,
  // so return SURFACE. We only test 5 out of the 6 faces since if
  // the axis does pass through, it must intersect two faces.

  CHECK_VOXEL_HANDEDNESS(v);

  // This code is taken from prim_cylinder.c as the situation
  // is exactly the same.
  
  // We test each face by returning SURFACE if the centerline is between
  // the N and S edges and also the E and W edges of the face in question.
  // We store the values of the edge tests in axis_side[] for reuse.

  AXIS_SIDE(0, 1, 0);
  AXIS_SIDE(3, 2, 1);
  AXIS_SIDE(0, 2, 8);
  AXIS_SIDE(3, 1, 9);
  
  if ((axis_side[0] == axis_side[1]) &&
      (axis_side[8] == axis_side[9]))
    return SURFACE;

  AXIS_SIDE(5, 4, 3);
  AXIS_SIDE(6, 7, 2);
  AXIS_SIDE(6, 4, 10);
  AXIS_SIDE(5, 7, 11);
  if ((axis_side[3]  == axis_side[2]) &&
      (axis_side[10] == axis_side[11]))
    return SURFACE;

  AXIS_SIDE(4, 0, 4);
  AXIS_SIDE(1, 5, 5);
  if ((axis_side[4] == axis_side[5]) &&
      (axis_side[0] == axis_side[3]))
    return SURFACE;

  AXIS_SIDE(2, 6, 6);
  AXIS_SIDE(7, 3, 7);
  if ((axis_side[6] == axis_side[7]) &&
      (axis_side[1] == axis_side[2]))
    return SURFACE;

  if ((axis_side[11] == axis_side[9]) &&
      (axis_side[5 ] == axis_side[7]))
    return SURFACE;

  // Check each edge of the box to see if it intersects the cone. If
  // it does, then return SURFACE.
  CHECK_EDGE_INTERSECT(0, 1);
  CHECK_EDGE_INTERSECT(3, 2);
  CHECK_EDGE_INTERSECT(6, 7);
  CHECK_EDGE_INTERSECT(5, 4);
  CHECK_EDGE_INTERSECT(4, 0);
  CHECK_EDGE_INTERSECT(1, 5);
  CHECK_EDGE_INTERSECT(2, 6);
  CHECK_EDGE_INTERSECT(7, 3);
  CHECK_EDGE_INTERSECT(0, 2);
  CHECK_EDGE_INTERSECT(3, 1);
  CHECK_EDGE_INTERSECT(4, 6);
  CHECK_EDGE_INTERSECT(7, 5);

  // If none of the points are in, the Z axis doesn't pass through the box,
  // and none of the edges intersect the cone, then the cone lies completely
  // outside the box.

  return OUT;
}

// Undefine the local macros used above.
#undef AXIS_SIDE
#undef CHECK_EDGE_INTERSECT

char *conePrimitive::prim_name(void)
{
  return "Cone";
}


#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif
