/* prim_cube.c : The cube primitive
** 
** The cube primitive defines a cube from -1 to +1 on the X, Y and Z axes.
**
** Chris Butcher, 15 December 1996
*/

#include "mirage.h"

#ifdef USE_NAMESPACES
namespace Mirage {
#endif

int cubePrimitive::ray_intersect(IntersectionList *the_list, primnode *node,
				 primnode *validate_node, vector ray_from, 
				 vector ray_dir,
				 uint ray_flags, signature ray_signature)
{
  // The coordinates of ray_from in primitive space
  double p_x, p_y, p_z;
  // The direction ray_dir in primitive space
  double d_x, d_y, d_z;
  // Which plane is hit first in X, Y and Z ... either +1 or -1.
  vector plane_first;
  // The t values of the first and last hits on the planes in the given
  // direction.
  vector t_first, t_last;
  // The t values of the entry and exit of the ray.
  double t_entry, t_exit;
  // The planes that are the entry and exit of the ray.
  int plane_entry, plane_exit;
  // The normals at the entry and exit points.
  vector norm_entry, norm_exit;
  // Was an intersection found?
  int found = 0;

  // Store the signature of this ray that we're intersecting with.
  node->last_ray_signature = ray_signature;

  // This apply is an inline function defined in matrix.h
  // Transform the ray startpoint into primitive space
  Matrix::apply(node->position[AFT], ray_from, &p_x, &p_y, &p_z);

  // This apply is also an inline function defined in matrix.h
  // Transform the ray direction into primitive space (note: shift does
  // not apply to this, only apply rotation + stretch sections of the matrix)
  Matrix::applynoshift(node->position[AFT], ray_dir, &d_x, &d_y, &d_z);

  // If the ray is outside the cube and heading away from it, then there
  // will be no intersections.
  if (p_x >  1.0 && d_x > 0.0) {
    node->last_ray_hit = FALSE;
    return node->last_ray_hit;
  }
  if (p_x < -1.0 && d_x < 0.0) {
    node->last_ray_hit = FALSE;
    return node->last_ray_hit;
  }
  if (p_y >  1.0 && d_y > 0.0) {
    node->last_ray_hit = FALSE;
    return node->last_ray_hit;
  }
  if (p_y < -1.0 && d_y < 0.0) {
    node->last_ray_hit = FALSE;
    return node->last_ray_hit;
  }
  if (p_z >  1.0 && d_z > 0.0) {
    node->last_ray_hit = FALSE;
    return node->last_ray_hit;
  }
  if (p_z < -1.0 && d_z < 0.0) {
    node->last_ray_hit = FALSE;
    return node->last_ray_hit;
  }

  // Calculate from the sign of dx, dy and dz which plane is hit first in
  // each direction by the ray.
  if (d_x > TINY) {
    plane_first[X] = -1.0;
  } else if (d_x < -TINY) {
    plane_first[X] = +1.0;
  } else {
    plane_first[X] = 0.0;
  }
  if (d_y > TINY) {
    plane_first[Y] = -1.0;
  } else if (d_y < -TINY) {
    plane_first[Y] = +1.0;
  } else {
    plane_first[Y] = 0.0;
  }
  if (d_z > TINY) {
    plane_first[Z] = -1.0;
  } else if (d_z < -TINY) {
    plane_first[Z] = +1.0;
  } else {
    plane_first[Z] = 0.0;
  }

  // Calculate the t-values of the ray's entry and exit on the planes
  // in each direction.
  if (plane_first[X] == 0.0) {
    // The ray is parallel to the planes. It's not going to cross them.
    // So, calculate whether the ray is inside or outside the cube in this
    // direction. If outside, return FALSE. Otherwise, continue.
    if (p_x > -1.0 && p_x < 1.0) {
      t_first[X] = -INFINITY;
      t_last[X] = +INFINITY;
    } else {
      node->last_ray_hit = FALSE;
      return node->last_ray_hit;
    }
  } else {
    t_first[X] = ( plane_first[X] - p_x) / d_x;
    t_last[X]  = (-plane_first[X] - p_x) / d_x;
  }
  if (plane_first[Y] == 0.0) {
    if (p_y > -1.0 && p_y < 1.0) {
      t_first[Y] = -INFINITY;
      t_last[Y] = +INFINITY;
    } else {
      node->last_ray_hit = FALSE;
      return node->last_ray_hit;
    }
  } else {
    t_first[Y] = ( plane_first[Y] - p_y) / d_y;
    t_last[Y]  = (-plane_first[Y] - p_y) / d_y;
  }
  if (plane_first[Z] == 0.0) {
    if (p_z > -1.0 && p_z < 1.0) {
      t_first[Z] = -INFINITY;
      t_last[Z] = +INFINITY;
    } else {
      node->last_ray_hit = FALSE;
      return node->last_ray_hit;
    }
  } else {
    t_first[Z] = ( plane_first[Z] - p_z) / d_z;
    t_last[Z]  = (-plane_first[Z] - p_z) / d_z;
  }

  // We now have the t_first and t_last values for all three directions.
  // The ray does not truly enter the cube until it has entered all three
  // directions, thus t_entry = the greatest of t_first. Also, t_exit
  // = the least of t_last by similar logic.

  // Calculate t_entry = MAX3(t_first).
  if (t_first[X] > t_first[Y]) {
    if (t_first[X] > t_first[Z]) {
      t_entry = t_first[X];
      plane_entry = X;
      Vector::assign(norm_entry, plane_first[X], 0, 0);
    } else {
      t_entry = t_first[Z];
      plane_entry = Z;
      Vector::assign(norm_entry, 0, 0, plane_first[Z]);
    }
  } else {
    if (t_first[Y] > t_first[Z]) {
      t_entry = t_first[Y];
      plane_entry = Y;
      Vector::assign(norm_entry, 0, plane_first[Y], 0);
    } else {
      t_entry = t_first[Z];
      plane_entry = Z;
      Vector::assign(norm_entry, 0, 0, plane_first[Z]);
    }
  }
					      
  // Calculate t_exit = MIN3(t_last).
  if (t_last[X] < t_last[Y]) {
    if (t_last[X] < t_last[Z]) {
      t_exit = t_last[X];
      plane_exit = X;
      Vector::assign(norm_exit, -plane_first[X], 0, 0);
    } else {
      t_exit = t_last[Z];
      plane_exit = Z;
      Vector::assign(norm_exit, 0, 0, -plane_first[Z]);
    }
  } else {
    if (t_last[Y] < t_last[Z]) {
      t_exit = t_last[Y];
      plane_exit = Y;
      Vector::assign(norm_exit, 0, -plane_first[Y], 0);
    } else {
      t_exit = t_last[Z];
      plane_exit = Z;
      Vector::assign(norm_exit, 0, 0, -plane_first[Z]);
    }
  }
					      
  // We now have t-values for the ray entering and exiting the cube. The
  // ray will not truly intersect the cube if t_exit <= t_entry
  if (t_exit <= t_entry) {
    node->last_ray_hit = FALSE;
    return node->last_ray_hit;
  }

  // Insert the entry and exit intersections into the intersection list.
  if (t_entry > MIN_TVAL_CUTOFF) {
    // This is the entry intersection.
    insert_entering(the_list, node, validate_node, norm_entry, t_entry,
		    ray_from, ray_dir, ray_flags);
    found = 1;
  }
  if (t_exit > MIN_TVAL_CUTOFF) {
    // This is the exit intersection.
    insert_leaving(the_list, node, validate_node, norm_exit, t_exit,
		   ray_from, ray_dir, ray_flags);
    found = 1;
  }

  if (found)
    node->last_ray_hit = TRUE;
  else
    node->last_ray_hit = FALSE;

  return node->last_ray_hit;
}

int cubePrimitive::ray_intersect(vector ray_from, 
				 vector ray_dir, rayintersect *int_storage)
{
  // Which plane is hit first in X, Y and Z ... either +1 or -1.
  vector plane_first;
  // The t values of the first and last hits on the planes in the given
  // direction.
  vector t_first, t_last;
  // The t values of the entry and exit of the ray.
  double t_entry, t_exit;
  // The planes that are the entry and exit of the ray.
  int plane_entry, plane_exit;
  // The normals at the entry and exit points.
  vector norm_entry, norm_exit;
  // Was an intersection found?
  int num_intersections = 0;

  // If the ray is outside the cube and heading away from it, then there
  // will be no intersections.
  if (ray_from[X] >  1.0 && ray_dir[X] > 0.0)
    return 0;
  if (ray_from[X] < -1.0 && ray_dir[X] < 0.0)
    return 0;
  if (ray_from[Y] >  1.0 && ray_dir[Y] > 0.0)
    return 0;
  if (ray_from[Y] < -1.0 && ray_dir[Y] < 0.0)
    return 0;
  if (ray_from[Z] >  1.0 && ray_dir[Z] > 0.0)
    return 0;
  if (ray_from[Z] < -1.0 && ray_dir[Z] < 0.0)
    return 0;

  // Calculate from the sign of dx, dy and dz which plane is hit first in
  // each direction by the ray.
  if (ray_dir[X] > TINY) {
    plane_first[X] = -1.0;
  } else if (ray_dir[X] < -TINY) {
    plane_first[X] = +1.0;
  } else {
    plane_first[X] = 0.0;
  }
  if (ray_dir[Y] > TINY) {
    plane_first[Y] = -1.0;
  } else if (ray_dir[Y] < -TINY) {
    plane_first[Y] = +1.0;
  } else {
    plane_first[Y] = 0.0;
  }
  if (ray_dir[Z] > TINY) {
    plane_first[Z] = -1.0;
  } else if (ray_dir[Z] < -TINY) {
    plane_first[Z] = +1.0;
  } else {
    plane_first[Z] = 0.0;
  }

  // Calculate the t-values of the ray's entry and exit on the planes
  // in each direction.
  if (plane_first[X] == 0.0) {
    // The ray is parallel to the planes. It's not going to cross them.
    // So, calculate whether the ray is inside or outside the cube in this
    // direction. If outside, return FALSE. Otherwise, continue.
    if (ray_from[X] > -1.0 && ray_from[X] < 1.0) {
      t_first[X] = -INFINITY;
      t_last[X] = +INFINITY;
    } else {
      return 0;
    }
  } else {
    t_first[X] = ( plane_first[X] - ray_from[X]) / ray_dir[X];
    t_last[X]  = (-plane_first[X] - ray_from[X]) / ray_dir[X];
  }
  if (plane_first[Y] == 0.0) {
    if (ray_from[Y] > -1.0 && ray_from[Y] < 1.0) {
      t_first[Y] = -INFINITY;
      t_last[Y] = +INFINITY;
    } else {
      return 0;
    }
  } else {
    t_first[Y] = ( plane_first[Y] - ray_from[Y]) / ray_dir[Y];
    t_last[Y]  = (-plane_first[Y] - ray_from[Y]) / ray_dir[Y];
  }
  if (plane_first[Z] == 0.0) {
    if (ray_from[Z] > -1.0 && ray_from[Z] < 1.0) {
      t_first[Z] = -INFINITY;
      t_last[Z] = +INFINITY;
    } else {
      return 0;
    }
  } else {
    t_first[Z] = ( plane_first[Z] - ray_from[Z]) / ray_dir[Z];
    t_last[Z]  = (-plane_first[Z] - ray_from[Z]) / ray_dir[Z];
  }

  // We now have the t_first and t_last values for all three directions.
  // The ray does not truly enter the cube until it has entered all three
  // directions, thus t_entry = the greatest of t_first. Also, t_exit
  // = the least of t_last by similar logic.

  // Calculate t_entry = MAX3(t_first).
  if (t_first[X] > t_first[Y]) {
    if (t_first[X] > t_first[Z]) {
      t_entry = t_first[X];
      plane_entry = X;
      Vector::assign(norm_entry, plane_first[X], 0, 0);
    } else {
      t_entry = t_first[Z];
      plane_entry = Z;
      Vector::assign(norm_entry, 0, 0, plane_first[Z]);
    }
  } else {
    if (t_first[Y] > t_first[Z]) {
      t_entry = t_first[Y];
      plane_entry = Y;
      Vector::assign(norm_entry, 0, plane_first[Y], 0);
    } else {
      t_entry = t_first[Z];
      plane_entry = Z;
      Vector::assign(norm_entry, 0, 0, plane_first[Z]);
    }
  }
					      
  // Calculate t_exit = MIN3(t_last).
  if (t_last[X] < t_last[Y]) {
    if (t_last[X] < t_last[Z]) {
      t_exit = t_last[X];
      plane_exit = X;
      Vector::assign(norm_exit, -plane_first[X], 0, 0);
    } else {
      t_exit = t_last[Z];
      plane_exit = Z;
      Vector::assign(norm_exit, 0, 0, -plane_first[Z]);
    }
  } else {
    if (t_last[Y] < t_last[Z]) {
      t_exit = t_last[Y];
      plane_exit = Y;
      Vector::assign(norm_exit, 0, -plane_first[Y], 0);
    } else {
      t_exit = t_last[Z];
      plane_exit = Z;
      Vector::assign(norm_exit, 0, 0, -plane_first[Z]);
    }
  }
					      
  // We now have t-values for the ray entering and exiting the cube. The
  // ray will not truly intersect the cube if t_exit <= t_entry
  if (t_exit <= t_entry) {
    return 0;
  }

  // Insert the entry and exit intersections into the intersection list.
  int_storage[num_intersections].t = t_entry;
  int_storage[num_intersections].is_entering = TRUE;
  Vector::set(int_storage[num_intersections].normal, norm_entry);
  num_intersections++;

  int_storage[num_intersections].t = t_exit;
  int_storage[num_intersections].is_entering = FALSE;
  Vector::set(int_storage[num_intersections].normal, norm_exit);
  num_intersections++;

  return num_intersections;
}

intersect_result cubePrimitive::point_intersect(primnode *node,
						vector wp,
						PointInfo *info,
						uint ray_flags)
{
  // Here wp is the world-point that we are testing for IN/OUT.
  // The values p_x, p_y, p_z are its coordinates in primitive space
  // (the mapping from world to primitive space is given by position.aft)
  double p_x, p_y, p_z;

  // This apply is an inline function defined in matrix.h
  // Transform the worldpoint into primitive space.
  Matrix::apply(node->position[AFT], wp, &p_x, &p_y, &p_z);

  // Check to see if the point lies inside all six planes that bound the
  // cube.
  if (p_x > -1.0 && p_x < 1.0 &&
      p_y > -1.0 && p_y < 1.0 &&
      p_z > -1.0 && p_z < 1.0) {
    set_point_contents(node, info);
    return IN;
  } else
    return OUT;
}

intersect_result cubePrimitive::point_intersect(vector wp)
{
  // Check to see if the point lies inside all six planes that bound the
  // cube.
  if (wp[X] > -1.0 && wp[X] < 1.0 &&
      wp[Y] > -1.0 && wp[Y] < 1.0 &&
      wp[Z] > -1.0 && wp[Z] < 1.0)
    return IN;
  else
    return OUT;
}

intersect_result cubePrimitive::voxel_intersect(box b)
{
  return cube_voxel_intersect(b);
}

// Macro CHECK_EDGE_INTERSECT checks the edge between points i and j
// to see if it intersects the cube. If it does, return SURFACE.
#define CHECK_EDGE_INTERSECT(i, j) {                                    \
  p_x = b[i][X];         p_y = b[i][Y];         p_z = b[i][Z];          \
  d_x = b[j][X] - p_x;   d_y = b[j][Y] - p_y;   d_z = b[j][Z] - p_z;    \
                                                                        \
  t_entry = -INFINITY;                                                  \
  t_exit = +INFINITY;                                                   \
                                                                        \
  if (d_x > 0) {                                                        \
    tval = (-1 - p_x) / d_x;                                            \
    if (tval > t_entry)                                                 \
      t_entry = tval;                                                   \
                                                                        \
    tval = (+1 - p_x) / d_x;                                            \
    if (tval < t_exit)                                                  \
      t_exit = tval;                                                    \
  } else if (d_x < 0) {                                                 \
    tval = (+1 - p_x) / d_x;                                            \
    if (tval > t_entry)                                                 \
      t_entry = tval;                                                   \
                                                                        \
    tval = (-1 - p_x) / d_x;                                            \
    if (tval < t_exit)                                                  \
      t_exit = tval;                                                    \
  } else {                                                              \
    if ((p_x < -1) || (p_x > +1)) {                                     \
      t_entry = +INFINITY;                                              \
      t_exit  = -INFINITY;                                              \
    }                                                                   \
  }                                                                     \
                                                                        \
  if (d_y > 0) {                                                        \
    tval = (-1 - p_y) / d_y;                                            \
    if (tval > t_entry)                                                 \
      t_entry = tval;                                                   \
                                                                        \
    tval = (+1 - p_y) / d_y;                                            \
    if (tval < t_exit)                                                  \
      t_exit = tval;                                                    \
  } else if (d_y < 0) {                                                 \
    tval = (+1 - p_y) / d_y;                                            \
    if (tval > t_entry)                                                 \
      t_entry = tval;                                                   \
                                                                        \
    tval = (-1 - p_y) / d_y;                                            \
    if (tval < t_exit)                                                  \
      t_exit = tval;                                                    \
  } else {                                                              \
    if ((p_y < -1) || (p_y > +1)) {                                     \
      t_entry = +INFINITY;                                              \
      t_exit  = -INFINITY;                                              \
    }                                                                   \
  }                                                                     \
                                                                        \
  if (d_z > 0) {                                                        \
    tval = (-1 - p_z) / d_z;                                            \
    if (tval > t_entry)                                                 \
      t_entry = tval;                                                   \
                                                                        \
    tval = (+1 - p_z) / d_z;                                            \
    if (tval < t_exit)                                                  \
      t_exit = tval;                                                    \
  } else if (d_z < 0) {                                                 \
    tval = (+1 - p_z) / d_z;                                            \
    if (tval > t_entry)                                                 \
      t_entry = tval;                                                   \
                                                                        \
    tval = (-1 - p_z) / d_z;                                            \
    if (tval < t_exit)                                                  \
      t_exit = tval;                                                    \
  } else {                                                              \
    if ((p_z < -1) || (p_z > +1)) {                                     \
      t_entry = +INFINITY;                                              \
      t_exit  = -INFINITY;                                              \
    }                                                                   \
  }                                                                     \
                                                                        \
  if (t_entry < t_exit &&                                               \
      t_entry > 0 &&                                                    \
      t_entry < 1) {                                                    \
    return SURFACE;                                                     \
  }                                                                     \
}

// Macro GENERATE_NORMAL generates the normal to a face f by looking at
// vertices i, j and k that lie on that face.
#define GENERATE_NORMAL(f, i, j, k)                        \
                                                           \
opposite_face = f + 3;                                     \
a_x = b[j][X] - b[i][X];                                   \
a_y = b[j][Y] - b[i][Y];                                   \
a_z = b[j][Z] - b[i][Z];                                   \
b_x = b[k][X] - b[i][X];                                   \
b_y = b[k][Y] - b[i][Y];                                   \
b_z = b[k][Z] - b[i][Z];                                   \
                                                           \
box_normal[f][X] = a_y*b_z - b_y*a_z;                      \
box_normal[f][Y] = a_z*b_x - b_z*a_x;                      \
box_normal[f][Z] = a_x*b_y - b_x*a_y;                      \
box_normal[opposite_face][X] = -box_normal[f][X];          \
box_normal[opposite_face][Y] = -box_normal[f][Y];          \
box_normal[opposite_face][Z] = -box_normal[f][Z];       

// Macro GET_DISTANCE_FROM_PLANE gets the distance of the point px, py, pz from
// the plane p, given by its normal box_normal[p] and a point b[i]. The
// distance is put in dist[p].
#define GET_DISTANCE_FROM_PLANE(p, i) {                                       \
  d_x = p_x - b[i][X];                                                        \
  d_y = p_y - b[i][Y];                                                        \
  d_z = p_z - b[i][Z];                                                        \
                                                                              \
  dist[p] = d_x * box_normal[p][X] +                                          \
            d_y * box_normal[p][Y] + d_z * box_normal[p][Z];                  \
}

// Macro CHECK_POINT_CONTAINED checks to see if the point px, py, pz
// lies inside the voxel by checking to see if it lies on the same side
// of all six planes.
#define CHECK_POINT_CONTAINED(px, py, pz) {           \
  p_x = px;  p_y = py;  p_z = pz;                     \
                                                      \
  GET_DISTANCE_FROM_PLANE(0, 0);                      \
  GET_DISTANCE_FROM_PLANE(1, 0);                      \
  GET_DISTANCE_FROM_PLANE(2, 0);                      \
  GET_DISTANCE_FROM_PLANE(3, 7);                      \
  GET_DISTANCE_FROM_PLANE(4, 7);                      \
  GET_DISTANCE_FROM_PLANE(5, 7);                      \
                                                      \
  if (dist[0] > 0 && dist[1] > 0 && dist[2] > 0 &&    \
      dist[3] > 0 && dist[4] > 0 && dist[5] > 0)      \
    return SURFACE;                                   \
}

// Macro AXIS_SIDE calculates the side on which the ax3 axis falls of the
// edge defined by the points of the box i and j. The value is stored
// in axis_side[store] for later reference.
#define AXIS_SIDE(i, j, ax1, ax2, ax3, store)                     \
                                                                  \
(axis_side[store] = ((b[i][ax2]*b[j][ax1] - b[i][ax1]*b[j][ax2]) >= 0.0))

// Macro TEST_AXISPOINT checks the face defined by points i, j, k and l
// and gets the point on the ax3 axis that this face contains.
// e.g. if ax1 = X, ax2 = Y, ax3 = Z, this will find the point 0, 0, Z
// that this face contains.
//
// Note that this assumes the intersection with the ax3 axis lies inside
// the quadrilateral defined by the face of the cube. You should use
// AXIS_SIDE to work this out.
//
// If this point is inside the cube (-1 to +1), then return SURFACE.
#define TEST_AXISPOINT(i, j, k, l, ax1, ax2, ax3) {            \
  if ((b[j][ax1] != b[i][ax1]) && (b[l][ax1] != b[k][ax1])) {  \
    tval = -b[i][ax1] / (b[j][ax1] - b[i][ax1]);               \
    x1[ax1] = 0.0;                                             \
    x1[ax2] = b[i][ax2] + tval * (b[j][ax2] - b[i][ax2]);      \
    x1[ax3] = b[i][ax3] + tval * (b[j][ax3] - b[i][ax3]);      \
                                                               \
    tval = -b[k][ax1] / (b[l][ax1] - b[k][ax1]);               \
    x2[ax1] = 0.0;                                             \
    x2[ax2] = b[k][ax2] + tval * (b[l][ax2] - b[k][ax2]);      \
    x2[ax3] = b[k][ax3] + tval * (b[l][ax3] - b[k][ax3]);      \
  } else {                                                     \
    tval = -b[i][ax1] / (b[k][ax1] - b[i][ax1]);               \
    x1[ax1] = 0.0;                                             \
    x1[ax2] = b[i][ax2] + tval * (b[k][ax2] - b[i][ax2]);      \
    x1[ax3] = b[i][ax3] + tval * (b[k][ax3] - b[i][ax3]);      \
                                                               \
    tval = -b[j][ax1] / (b[l][ax1] - b[j][ax1]);               \
    x2[ax1] = 0.0;                                             \
    x2[ax2] = b[j][ax2] + tval * (b[l][ax2] - b[j][ax2]);      \
    x2[ax3] = b[j][ax3] + tval * (b[l][ax3] - b[j][ax3]);      \
  }                                                            \
                                                               \
  tval = -x1[ax2] / (x2[ax2] - x1[ax2]);                       \
  axispt[ax1] = 0.0;                                           \
  axispt[ax2] = 0.0;                                           \
  axispt[ax3] = x1[ax3] + tval * (x2[ax3] - x1[ax3]);          \
                                                               \
  if (fabs(axispt[ax3]) < 1)                                   \
    return SURFACE;                                            \
}


intersect_result cube_voxel_intersect(box b)
{
  // This routine checks to see if a parallelipiped lies inside, outside,
  // or intersects the surface of the cube.
  //
  // If all the points lie outside one of the X, Y, Z +/- 1 planes, return
  // OUT. If all the points lie inside the box, return IN.
  //
  // If any of the voxel's edges intersect the cube, return SURFACE.
  //
  // If any of the cube's corners are inside the voxel, return SURFACE.
  //
  // If the voxel intersects the X, Y or Z axes of the cube's primitive
  // space between -1 or +1 then return SURFACE.
  //
  // Otherwise, all the corners are outside, the edges of the voxel are
  // outside and the points of the voxel are outside. We can return OUT.

  int i, points_outside[6] = {0, 0, 0, 0, 0, 0}, planes_inside = 0;
  vector box_normal[6];
  double a_x, a_y, a_z, b_x, b_y, b_z, t_entry, t_exit, tval;
  double p_x, p_y, p_z, d_x, d_y, d_z, dist[6];
  int opposite_face;
  double axis_side[12];
  vector x1, x2, axispt;

  CHECK_VOXEL_HANDEDNESS(b);

  // Count the number of points that lie outside each coordinate plane.
  for (i = 0; i < 8; i++) {

    if (b[i][X] > 1.0)
      points_outside[0]++;
    else if (b[i][X] < -1.0)
      points_outside[1]++;

    if (b[i][Y] > 1.0)
      points_outside[2]++;
    else if (b[i][Y] < -1.0)
      points_outside[3]++;

    if (b[i][Z] > 1.0)
      points_outside[4]++;
    else if (b[i][Z] < -1.0)
      points_outside[5]++;
  }

  // For each plane, check: if all the points lie outside it, return OUT.
  // If all the points lie inside it, then increment the number of planes
  // that the box lies inside. If the box lies inside all six planes,
  // then return IN.
  for (i = 0; i < 6; i++) {
    if (points_outside[i] == 8)
      return OUT;
    if (points_outside[i] == 0)
      planes_inside++;
  }

  if (planes_inside == 6)
    return IN;

  // Check that none of the edges of the voxel intersect the cube.
  CHECK_EDGE_INTERSECT(0, 1);
  CHECK_EDGE_INTERSECT(3, 2);
  CHECK_EDGE_INTERSECT(6, 7);
  CHECK_EDGE_INTERSECT(5, 4);
  CHECK_EDGE_INTERSECT(4, 0);
  CHECK_EDGE_INTERSECT(1, 5);
  CHECK_EDGE_INTERSECT(2, 6);
  CHECK_EDGE_INTERSECT(7, 3);
  CHECK_EDGE_INTERSECT(0, 2);
  CHECK_EDGE_INTERSECT(3, 1);
  CHECK_EDGE_INTERSECT(4, 6);
  CHECK_EDGE_INTERSECT(7, 5);

  // Check that none of the corners of the cube are contained in the voxel.
  GENERATE_NORMAL(0, 0, 1, 2);
  GENERATE_NORMAL(1, 0, 4, 1);
  GENERATE_NORMAL(2, 0, 2, 4);
  CHECK_POINT_CONTAINED(+1, +1, +1);
  CHECK_POINT_CONTAINED(-1, +1, +1);
  CHECK_POINT_CONTAINED(+1, -1, +1);
  CHECK_POINT_CONTAINED(-1, -1, +1);
  CHECK_POINT_CONTAINED(+1, +1, -1);
  CHECK_POINT_CONTAINED(-1, +1, -1);
  CHECK_POINT_CONTAINED(+1, -1, -1);
  CHECK_POINT_CONTAINED(-1, -1, -1);

  // We now check the Z axis ... if any of the faces cube intersect the Z
  // axis, find the point at which they intersect. If this point is within
  // the range +1 to -1, return SURFACE.
  //
  // Each face is checked by finding which side the axis lies on from the
  // N and S sides, and then the E and W sides. If these two pairs agree,
  // then the axis is within the square.

  AXIS_SIDE(0, 1, X, Y, Z, 0);
  AXIS_SIDE(3, 2, X, Y, Z, 1);
  AXIS_SIDE(0, 2, X, Y, Z, 8);
  AXIS_SIDE(3, 1, X, Y, Z, 9);
  
  if ((axis_side[0] == axis_side[1]) &&
      (axis_side[8] == axis_side[9])) {
    TEST_AXISPOINT(0, 1, 2, 3, X, Y, Z);
  }
  
  AXIS_SIDE(5, 4, X, Y, Z, 3);
  AXIS_SIDE(6, 7, X, Y, Z, 2);
  AXIS_SIDE(6, 4, X, Y, Z, 10);
  AXIS_SIDE(5, 7, X, Y, Z, 11);
  if ((axis_side[3]  == axis_side[2]) &&
      (axis_side[10] == axis_side[11])) {
    TEST_AXISPOINT(4, 5, 6, 7, X, Y, Z);
  }
  
  AXIS_SIDE(4, 0, X, Y, Z, 4);
  AXIS_SIDE(1, 5, X, Y, Z, 5);
  if ((axis_side[4] == axis_side[5]) &&
      (axis_side[0] == axis_side[3])) {
    TEST_AXISPOINT(0, 1, 4, 5, X, Y, Z);
  }
  
  AXIS_SIDE(2, 6, X, Y, Z, 6);
  AXIS_SIDE(7, 3, X, Y, Z, 7);
  if ((axis_side[6] == axis_side[7]) &&
      (axis_side[1] == axis_side[2])) {
    TEST_AXISPOINT(2, 3, 6, 7, X, Y, Z);
  }
  
  if ((axis_side[11] == axis_side[9]) &&
      (axis_side[5 ] == axis_side[7])) {
    TEST_AXISPOINT(1, 3, 5, 7, X, Y, Z);
  }
  
  if ((axis_side[8 ] == axis_side[10]) &&
      (axis_side[4 ] == axis_side[6 ])) {
    TEST_AXISPOINT(0, 2, 4, 6, X, Y, Z);
  }





  // We now check the X axis ... if any of the faces cube intersect the X
  // axis, find the point at which they intersect. If this point is within
  // the range +1 to -1, return SURFACE.
  //
  // Each face is checked by finding which side the axis lies on from the
  // N and S sides, and then the E and W sides. If these two pairs agree,
  // then the axis is within the square.

  AXIS_SIDE(0, 1, Y, Z, X, 0);
  AXIS_SIDE(3, 2, Y, Z, X, 1);
  AXIS_SIDE(0, 2, Y, Z, X, 8);
  AXIS_SIDE(3, 1, Y, Z, X, 9);
  
  if ((axis_side[0] == axis_side[1]) &&
      (axis_side[8] == axis_side[9])) {
    TEST_AXISPOINT(0, 1, 2, 3, Y, Z, X);
  }
    
  AXIS_SIDE(5, 4, Y, Z, X, 3);
  AXIS_SIDE(6, 7, Y, Z, X, 2);
  AXIS_SIDE(6, 4, Y, Z, X, 10);
  AXIS_SIDE(5, 7, Y, Z, X, 11);
  if ((axis_side[3]  == axis_side[2]) &&
      (axis_side[10] == axis_side[11])) {
    TEST_AXISPOINT(4, 5, 6, 7, Y, Z, X);
  }
  
  AXIS_SIDE(4, 0, Y, Z, X, 4);
  AXIS_SIDE(1, 5, Y, Z, X, 5);
  if ((axis_side[4] == axis_side[5]) &&
      (axis_side[0] == axis_side[3])) {
    TEST_AXISPOINT(0, 1, 4, 5, Y, Z, X);
  }
  
  AXIS_SIDE(2, 6, Y, Z, X, 6);
  AXIS_SIDE(7, 3, Y, Z, X, 7);
  if ((axis_side[6] == axis_side[7]) &&
      (axis_side[1] == axis_side[2])) {
    TEST_AXISPOINT(2, 3, 6, 7, Y, Z, X);
  }
  
  if ((axis_side[11] == axis_side[9]) &&
      (axis_side[5 ] == axis_side[7])) {
    TEST_AXISPOINT(1, 3, 5, 7, Y, Z, X);
  }
  
  if ((axis_side[8 ] == axis_side[10]) &&
      (axis_side[4 ] == axis_side[6 ])) {
    TEST_AXISPOINT(0, 2, 4, 6, Y, Z, X);
  }



  // We now check the Y axis ... if any of the faces cube intersect the Y
  // axis, find the point at which they intersect. If this point is within
  // the range +1 to -1, return SURFACE.
  //
  // Each face is checked by finding which side the axis lies on from the
  // N and S sides, and then the E and W sides. If these two pairs agree,
  // then the axis is within the square.

  AXIS_SIDE(0, 1, Z, X, Y, 0);
  AXIS_SIDE(3, 2, Z, X, Y, 1);
  AXIS_SIDE(0, 2, Z, X, Y, 8);
  AXIS_SIDE(3, 1, Z, X, Y, 9);
  
  if ((axis_side[0] == axis_side[1]) &&
      (axis_side[8] == axis_side[9])) {
    TEST_AXISPOINT(0, 1, 2, 3, Z, X, Y);
  }
  
  AXIS_SIDE(5, 4, Z, X, Y, 3);
  AXIS_SIDE(6, 7, Z, X, Y, 2);
  AXIS_SIDE(6, 4, Z, X, Y, 10);
  AXIS_SIDE(5, 7, Z, X, Y, 11);
  if ((axis_side[3]  == axis_side[2]) &&
      (axis_side[10] == axis_side[11])) {
    TEST_AXISPOINT(4, 5, 6, 7, Z, X, Y);
  }
  
  AXIS_SIDE(4, 0, Z, X, Y, 4);
  AXIS_SIDE(1, 5, Z, X, Y, 5);
  if ((axis_side[4] == axis_side[5]) &&
      (axis_side[0] == axis_side[3])) {
    TEST_AXISPOINT(0, 1, 4, 5, Z, X, Y);
  }
  
  AXIS_SIDE(2, 6, Z, X, Y, 6);
  AXIS_SIDE(7, 3, Z, X, Y, 7);
  if ((axis_side[6] == axis_side[7]) &&
      (axis_side[1] == axis_side[2])) {
    TEST_AXISPOINT(2, 3, 6, 7, Z, X, Y);
  }
  
  if ((axis_side[11] == axis_side[9]) &&
      (axis_side[5 ] == axis_side[7])) {
    TEST_AXISPOINT(1, 3, 5, 7, Z, X, Y);
  }
  
  if ((axis_side[8 ] == axis_side[10]) &&
      (axis_side[4 ] == axis_side[6 ])) {
    TEST_AXISPOINT(0, 2, 4, 6, Z, X, Y);
  }


  
  // None of the edges of the voxel intersect the cube, none of the corners
  // of the cube are inside the voxel, and all of the points are OUT, so we
  // may safely return OUT.
  return OUT;
}

#undef CHECK_EDGE_INTERSECT
#undef GENERATE_NORMAL
#undef GET_DISTANCE_FROM_PLANE
#undef CHECK_POINT_CONTAINED
#undef TEST_AXISPOINT
#undef AXIS_SIDE

char *cubePrimitive::prim_name(void) {
  return "Cube";
}


#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif
