/* prim_cylinder.c : The cylinder primitive
** 
** The cylinder primitive defines an infinite cylinder of radius 1,
** centred on the Z axis.
**
** Chris Butcher, 11 December 1996
*/

#include "mirage.h"

#ifdef USE_NAMESPACES
namespace Mirage {
#endif

int cylinderPrimitive::ray_intersect(IntersectionList *the_list, 
				     primnode *node, primnode *validate_node, 
				     vector ray_from,
				     vector ray_dir, uint ray_flags,
				     signature ray_signature)
{
  // The coordinates of ray_from in primitive space
  double p_x, p_y;
  // The direction ray_dir in primitive space
  double d_x, d_y;
  // The coefficients of the quadratic equation in t
  double a, b, c;
  // The discriminant of this equation
  double disc;
  // Have we found an intersection?
  int found = 0;
  // The t-values of the intersection points.
  double t_0, t_1;
  // The normal at the hitpoint.
  vector norm;
  // The matrix to transform from world to primitive space.
  matrix *m;

  // Store the intersection of this ray that we are intersecting with.
  node->last_ray_signature = ray_signature;

  // Get the matrix to transform into primitive space.
  m = &(node->position[AFT]);

  // Transform the ray startpoint into primitive space (only the X and
  // Y coordinates because Z is irrelevant).
  p_x = (*m)[0][0]*ray_from[X] + (*m)[0][1]*ray_from[Y]
      + (*m)[0][2]*ray_from[Z] + (*m)[0][3];
  p_y = (*m)[1][0]*ray_from[X] + (*m)[1][1]*ray_from[Y]
      + (*m)[1][2]*ray_from[Z] + (*m)[1][3];

  // Transform the ray direction into primitive space (note: shift does
  // not apply to this, only apply rotation + stretch sections of the matrix)
  d_x = (*m)[0][0]*ray_dir[X] + (*m)[0][1]*ray_dir[Y] + (*m)[0][2]*ray_dir[Z];
  d_y = (*m)[1][0]*ray_dir[X] + (*m)[1][1]*ray_dir[Y] + (*m)[1][2]*ray_dir[Z];

  // Form the coefficients of the quadratic equation that defines t.
  // This equation is |p + td| = 1.0
  // Note that this code is very similar to the sphere intersection test.
  a = SQR(d_x) + SQR(d_y);
  b = 2.0 * (p_x*d_x + p_y*d_y);
  c = SQR(p_x) + SQR(p_y) - 1.0;

  // Calculate the discriminant of the quadratic equation.
  disc = SQR(b) - 4.0*a*c;

  // If the discriminant is less than or equal to zero, then there
  // are no intersections with the cylinder (tangent rays are treated as
  // no intersection at all).
  if (disc <= 0) {
    node->last_ray_hit = FALSE;
    return node->last_ray_hit;
  }

  // a is equal to the square of the magnitude of the direction of the ray.
  // This *must* not be zero.
  if (a == 0)
    fatal_error("cylinderPrimitive::ray_intersect", "ray direction zero",
		"This error should not happen.");

  // Calculate the intersection t-values. Must have t_0 < t_1 for later
  // calculations to work.
  // Note that we can save a square root operation by saying
  // t_1 = c / a * t_0 ... but this only works if we can be sure that
  // t_0 isn't zero. It may easily be if sqrt(disc) == b or == -b.
  // So, we make sure that the +/-sqrt(disc) is always the same sign
  // as -b to make sure that the subtraction can't produce zero.
  // This makes the code a little long-winded.
  if (a >= 0) {
    if (b >= 0) {
      t_0 = (-b - sqrt(disc)) / (2 * a);
      t_1 = c / (a * t_0);
    } else {
      t_1 = (-b + sqrt(disc)) / (2 * a);
      t_0 = c / (a * t_1);
    }
  } else {
    if (b >= 0) {
      t_1 = (-b - sqrt(disc)) / (2 * a);
      t_0 = c / (a * t_1);
    } else {
      t_0 = (-b + sqrt(disc)) / (2 * a);
      t_1 = c / (a * t_0);
    }
  }

  // Insert any positive t-values (valid intersections).
  if (t_0 > MIN_TVAL_CUTOFF) {
    // Get the normal (since this is a unit cylinder, the normal is the
    // same as the primitive-space hitpoint = p + td ... in X and Y at least.
    // The Z component of the normal is zero).
    norm[X] = p_x + t_0*d_x;
    norm[Y] = p_y + t_0*d_y;
    norm[Z] = 0;
    // Note that since t_0 < t_1 always, this intersection is an ENTERING
    // one.
    insert_entering(the_list, node, validate_node, norm, t_0,
		    ray_from, ray_dir, ray_flags);
    found = 1;
  }
  if (t_1 > MIN_TVAL_CUTOFF) {
    // Get the normal (since this is a unit cylinder, the normal is the
    // same as the primitive-space hitpoint = p + td ... in X and Y at least.
    // The Z component of the normal is zero).
    norm[X] = p_x + t_1*d_x;
    norm[Y] = p_y + t_1*d_y;
    norm[Z] = 0;
    // Note that since t_0 < t_1 always, this intersection is a LEAVING
    // one.
    insert_leaving(the_list, node, validate_node, norm, t_1,
		   ray_from, ray_dir, ray_flags);
    found = 1;
  }

  if (found)
    node->last_ray_hit = TRUE;
  else
    node->last_ray_hit = FALSE;

  return node->last_ray_hit;
}

int cylinderPrimitive::ray_intersect(vector ray_from,
				     vector ray_dir, rayintersect *int_storage)
{
  // The coefficients of the quadratic equation in t
  double a, b, c;
  // The discriminant of this equation
  double disc;
  // Have we found an intersection?
  int num_intersections = 0;
  // The t-values of the intersection points.
  double t_0, t_1;
  // The normal at the hitpoint.
  vector norm;

  // Form the coefficients of the quadratic equation that defines t.
  // This equation is |p + td| = 1.0
  // Note that this code is very similar to the sphere intersection test.
  a = SQR(ray_dir[X]) + SQR(ray_dir[Y]);
  b = 2.0 * (ray_from[X]*ray_dir[X] + ray_from[Y]*ray_dir[Y]);
  c = SQR(ray_from[X]) + SQR(ray_from[Y]) - 1.0;

  // Calculate the discriminant of the quadratic equation.
  disc = SQR(b) - 4.0*a*c;

  // If the discriminant is less than or equal to zero, then there
  // are no intersections with the cylinder (tangent rays are treated as
  // no intersection at all).
  if (disc <= 0) {
    return 0;
  }

  // a is equal to the square of the magnitude of the direction of the ray.
  // This *must* not be zero.
  if (a == 0)
    fatal_error("cylinderPrimitive::ray_intersect", "ray direction zero",
		"This error should not happen.");

  // Calculate the intersection t-values. Must have t_0 < t_1 for later
  // calculations to work.
  // Note that we can save a square root operation by saying
  // t_1 = c / a * t_0 ... but this only works if we can be sure that
  // t_0 isn't zero. It may easily be if sqrt(disc) == b or == -b.
  // So, we make sure that the +/-sqrt(disc) is always the same sign
  // as -b to make sure that the subtraction can't produce zero.
  // This makes the code a little long-winded.
  if (a >= 0) {
    if (b >= 0) {
      t_0 = (-b - sqrt(disc)) / (2 * a);
      t_1 = c / (a * t_0);
    } else {
      t_1 = (-b + sqrt(disc)) / (2 * a);
      t_0 = c / (a * t_1);
    }
  } else {
    if (b >= 0) {
      t_1 = (-b - sqrt(disc)) / (2 * a);
      t_0 = c / (a * t_1);
    } else {
      t_0 = (-b + sqrt(disc)) / (2 * a);
      t_1 = c / (a * t_0);
    }
  }


  // Get the normal (since this is a unit cylinder, the normal is the
  // same as the primitive-space hitpoint = p + td ... in X and Y at least.
  // The Z component of the normal is zero).
  norm[X] = ray_from[X] + t_0*ray_dir[X];
  norm[Y] = ray_from[Y] + t_0*ray_dir[Y];
  norm[Z] = 0;
  // Note that since t_0 < t_1 always, this intersection is an ENTERING
  // one.
  int_storage[num_intersections].t = t_0;
  int_storage[num_intersections].is_entering = TRUE;
  Vector::set(int_storage[num_intersections].normal, norm);
  num_intersections++;

  // Get the normal (since this is a unit cylinder, the normal is the
  // same as the primitive-space hitpoint = p + td ... in X and Y at least.
  // The Z component of the normal is zero).
  norm[X] = ray_from[X] + t_1*ray_dir[X];
  norm[Y] = ray_from[Y] + t_1*ray_dir[Y];
  norm[Z] = 0;
  // Note that since t_0 < t_1 always, this intersection is a LEAVING
  // one.
  int_storage[num_intersections].t = t_1;
  int_storage[num_intersections].is_entering = FALSE;
  Vector::set(int_storage[num_intersections].normal, norm);
  num_intersections++;

  return num_intersections;
}

intersect_result cylinderPrimitive::point_intersect(primnode *node,
						    vector wp,
						    PointInfo *info,
						    uint ray_flags)
{
  // Here wp is the world-point that we are testing for IN/OUT.
  // The values p_x, p_y, p_z are its coordinates in primitive space
  // (the mapping from world to primitive space is given by position.aft)
  double p_x, p_y, d_sq;
  matrix *m;

  // Get the matrix to transform from world to primitive space.
  m = &(node->position[AFT]);

  // Transform the worldpoint into primitive space.
  p_x = (*m)[0][0]*wp[X] + (*m)[0][1]*wp[Y] + (*m)[0][2]*wp[Z] + (*m)[0][3];
  p_y = (*m)[1][0]*wp[X] + (*m)[1][1]*wp[Y] + (*m)[1][2]*wp[Z] + (*m)[1][3];

  // d_sq is the point's distance from the cylinder's axis, squared.
  d_sq = SQR(p_x) + SQR(p_y);
  
  // If the distance squared is <= 1 squared, the point is IN.
  if (d_sq <= 1.0) {
    set_point_contents(node, info);
    return IN;
  } else
    return OUT;
}

intersect_result cylinderPrimitive::point_intersect(vector wp)
{
  // If the distance squared is <= 1 squared, the point is IN.
  if (SQR(wp[X]) + SQR(wp[Y]) <= 1.0)
    return IN;
  else
    return OUT;
}





// Now define some local macros to make the code length shorter for the
// routine cylinderPrimitive::voxel_intersect.


// Macro CHECK_OUTSIDE_PLUS_PLANE tests to see if the points of the box all
// lie outside the plane P at distance +1.
#define CHECK_OUTSIDE_PLUS_PLANE(axis) {   \
                                           \
  for (i = 0; i < 8; i++) {                \
    if (b[i][axis] < 1.0) {                \
      break;                               \
    }                                      \
  }                                        \
  if (i == 8)                              \
    return OUT;                            \
}


// Macro CHECK_OUTSIDE_MINUS_PLANE tests to see if the points of the box all
// lie outside the plane P at distance -1.
#define CHECK_OUTSIDE_MINUS_PLANE(axis) {   \
                                            \
  for (i = 0; i < 8; i++)                   \
    if (b[i][axis] > -1.0)                  \
      break;                                \
                                            \
  if (i == 8)                               \
    return OUT;                             \
}


// Macro AXIS_SIDE calculates the side on which the Z axis falls of the
// edge defined by the points of the box i and j. The value is stored
// in axis_side[store] for later reference.
#define AXIS_SIDE(i, j, store)                                    \
                                                                  \
(axis_side[store] = ((b[i][Y]*b[j][X] - b[i][X]*b[j][Y]) >= 0.0))


// Macro CHECK_EDGE_INTERSECT checks the edge between points i and j
// to see if it intersects the unit cylinder. If it does, return SURFACE.
#define CHECK_EDGE_INTERSECT(i, j)                                        \
                                                                          \
p1_x = b[i][X];   p1_y = b[i][Y];   p2_x = b[j][X];   p2_y = b[j][Y];     \
dx = p2_x - p1_x;                   dy = p2_y - p1_y;                     \
                                                                          \
p1_val   = p1_x*dx + p1_y*dy;       p2_val = p2_x*dx + p2_y*dy;           \
p1_cross = p1_y*dx - p1_x*dy;                                             \
                                                                          \
if (((p1_val < 0) != (p2_val < 0)) && (SQR(p1_cross) <= SQR(dx)+SQR(dy))) \
  return SURFACE;





intersect_result cylinderPrimitive::voxel_intersect(box b)
{
  return cylinder_voxel_intersect(b);
}

intersect_result cylinder_voxel_intersect(box b)
{
  // This routine checks to see if a parallelipiped lies inside, outside,
  // or intersects the cylinder. It has five steps:
  // 1) If all of the points lie at X >= 1, the box is outside. Similarly
  //    for X = -1 and for the Y  +/- 1 planes.
  // 2) Count the points inside the cylinder. If there are eight, then the
  //    box is inside. If there are between one and seven, then the box must
  //    cross the surface of the cylinder and so is SURFACE.
  // 3) All the points are outside. Is the centerline pass through any face?
  //    If so, then the box contains the cylinder, return SURFACE.
  // 4) Check to see if any of the edges of the box intersects the cylinder.
  //    If so, return SURFACE.
  // 5) None of the above have returned a SURFACE intersection so the cylinder
  //    lies outside the box, return OUT.

  // Which points are inside the cylinder?
  int i, first_point_in, this_point_in;
  // The side that the Z axis falls on of a given edge.
  double axis_side[12];
  // Variables that are used in CHECK_EDGE_INTERSECT to determine whether
  // a given edge intersects the cylinder.
  double p1_x, p1_y, p2_x, p2_y, dx, dy, p1_val, p2_val, p1_cross;



  // Step 1: Check to see if all of the points of the box lie outside
  //         some unit coordinate plane, (X | Y) = +/- 1. If so,
  //         return OUT.
  CHECK_OUTSIDE_PLUS_PLANE(X);
  CHECK_OUTSIDE_PLUS_PLANE(Y);
  CHECK_OUTSIDE_MINUS_PLANE(X);
  CHECK_OUTSIDE_MINUS_PLANE(Y);



  // Step 2: Count the number of vertices that are inside the cylinder.
  //         If eight, return IN. If zero, continue on, otherwise return
  //         SURFACE.
  //           Note that this means that if we find two that are different,
  //         we can stop and return SURFACE.

  // Work out if the first point is in.
  first_point_in = (SQR(b[0][X]) + SQR(b[0][Y]) < 1.0);

  // For all the rest,
  i = 1;
  do {
    // work out if they are in.
    this_point_in = (SQR(b[i][X]) + SQR(b[i][Y]) < 1.0);
    i++;
    // If the point is on the opposite side of the surface from the first
    // point, then the box must be SURFACE.
    if (this_point_in != first_point_in)
      return SURFACE;
  } while (i < 8);
  
  // If all eight points are checked, they must all be the same. If they
  // are all in, return IN.
  if (this_point_in)
    return IN;




  // Step 3: Check to see if the Z axis passes through the box. We check
  //         each face in turn ... although since the axis will either
  //         intersect 0 or 2 faces, we only test 5 out of the 6 faces.
  //         If it does, return SURFACE.

  CHECK_VOXEL_HANDEDNESS(b);

  // We test each face by returning SURFACE if the centerline is between
  // the N and S edges and also the E and W edges of the face in question.
  // We store the values of the edge tests in axis_side[] for reuse.

  AXIS_SIDE(0, 1, 0);
  AXIS_SIDE(3, 2, 1);
  AXIS_SIDE(0, 2, 8);
  AXIS_SIDE(3, 1, 9);
  
  if ((axis_side[0] == axis_side[1]) &&
      (axis_side[8] == axis_side[9]))
    return SURFACE;

  AXIS_SIDE(5, 4, 3);
  AXIS_SIDE(6, 7, 2);
  AXIS_SIDE(6, 4, 10);
  AXIS_SIDE(5, 7, 11);
  if ((axis_side[3]  == axis_side[2]) &&
      (axis_side[10] == axis_side[11]))
    return SURFACE;

  AXIS_SIDE(4, 0, 4);
  AXIS_SIDE(1, 5, 5);
  if ((axis_side[4] == axis_side[5]) &&
      (axis_side[0] == axis_side[3]))
    return SURFACE;

  AXIS_SIDE(2, 6, 6);
  AXIS_SIDE(7, 3, 7);
  if ((axis_side[6] == axis_side[7]) &&
      (axis_side[1] == axis_side[2]))
    return SURFACE;

  if ((axis_side[11] == axis_side[9]) &&
      (axis_side[5 ] == axis_side[7]))
    return SURFACE;



  // Step 4: Check each edge to see if it intersects the cylinder. If
  //         it does, then return SURFACE.
  CHECK_EDGE_INTERSECT(0, 1);
  CHECK_EDGE_INTERSECT(3, 2);
  CHECK_EDGE_INTERSECT(6, 7);
  CHECK_EDGE_INTERSECT(5, 4);
  CHECK_EDGE_INTERSECT(4, 0);
  CHECK_EDGE_INTERSECT(1, 5);
  CHECK_EDGE_INTERSECT(2, 6);
  CHECK_EDGE_INTERSECT(7, 3);
  CHECK_EDGE_INTERSECT(0, 2);
  CHECK_EDGE_INTERSECT(3, 1);
  CHECK_EDGE_INTERSECT(4, 6);
  CHECK_EDGE_INTERSECT(7, 5);



  // Step 5: All of the intersection tests above have fallen through, the
  //         cylinder is outside the box.
  return OUT;
}

// Undefine the macros defined above (local only).
#undef CHECK_OUTSIDE_PLUS_PLANE
#undef CHECK_OUTSIDE_MINUS_PLANE
#undef AXIS_SIDE
#undef CHECK_EDGE_INTERSECT

char *cylinderPrimitive::prim_name(void) {
  return "Cylinder";
}


#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif
