/* prim_octreecluster.c : The octree cluster primitive
** 
** This file contains the function(s) for
** the OctreeClusterPrimitive.
**
** Chris Butcher, 9 January 1997
*/

#include "mirage.h"

#ifdef USE_NAMESPACES
namespace Mirage {
#endif

// The OctreeClusterPrimitive performs octree subdivision.
void OctreeClusterPrimitive::subdivide(void)
{
  // Subdivide this cluster only if we haven't already
  if (!done_subdivision) {
    // Check for clusters inside the RCSG graph that will need to be
    // subdivided.
    subdivide_clusters_in_RCSG(myRCSG);

    // Convert the RCSG graph into an octree.
    my_octree = octree_subdivide(myRCSG, 0, subdivision_limit,
				 max_prims_in_voxel,
				 minbound[X], minbound[Y], minbound[Z],
				 maxbound[X], maxbound[Y], maxbound[Z]);

    // Compress the octree, looking for empty nodes to save space.
    compress_octree(&my_octree);

    done_subdivision = TRUE;
  }
}

// The ray-intersection routine for OctreeClusterPrimitive traces
// the ray through the octree subdivision ...
int OctreeClusterPrimitive::ray_intersect(IntersectionList *the_list,
					  primnode *node,
					  primnode *validate_node,
					  vector ray_from, vector ray_dir,
					  uint ray_flags,
					  signature ray_signature)
{
  int ray_result;
  vector t_entry, t_exit;
  double p_x, p_y, p_z, d_x, d_y, d_z;
  double overall_entry, overall_exit;

  // Store the signature of this ray that we're intersecting with.
  node->last_ray_signature = ray_signature;

  // Transform the ray into the cluster's primitive space.
  Matrix::apply(node->position[AFT], ray_from, cluster_ray_from);
  Matrix::applynoshift(node->position[AFT], ray_dir, cluster_ray_dir);

  p_x = cluster_ray_from[X];  d_x = cluster_ray_dir[X];
  p_y = cluster_ray_from[Y];  d_y = cluster_ray_dir[Y];
  p_z = cluster_ray_from[Z];  d_z = cluster_ray_dir[Z];

  /*
   * Determine whether or not the ray intersects the octreecluster's
   * bounding box.
   */
  
  // If the ray is outside the bounds and heading away from them, then there
  // will be no intersections.
  if (p_x > maxbound[X] && d_x > 0.0) {
    node->last_ray_hit = FALSE;
    return FALSE;
  }
  if (p_x < minbound[X] && d_x < 0.0) {
    node->last_ray_hit = FALSE;
    return FALSE;
  }
  if (p_y > maxbound[Y] && d_y > 0.0) {
    node->last_ray_hit = FALSE;
    return FALSE;
  }
  if (p_y < minbound[Y] && d_y < 0.0) {
    node->last_ray_hit = FALSE;
    return FALSE;
  }
  if (p_z > maxbound[Z] && d_z > 0.0) {
    node->last_ray_hit = FALSE;
    return FALSE;
  }
  if (p_z < minbound[Z] && d_z < 0.0) {
    node->last_ray_hit = FALSE;
    return FALSE;
  }

  // Determine the low and high t-values where the ray crosses the bounding
  // planes of the cluster.
  if (d_x > TINY) {
    // The ray is heading towards +X. Check to see if it is outside the
    // box where it starts.
    if (p_x > maxbound[X]) {
      node->last_ray_hit = FALSE;
      return FALSE;
    }

    // Determine t-values for entry and exit in X.
    t_entry[X] = (minbound[X] - p_x) / d_x;
    t_exit [X] = (maxbound[X] - p_x) / d_x;
  } else if (d_x < -TINY) {
    // The ray is heading towards -X. Check to see if it is outside the
    // box where it starts.
    if (p_x < minbound[X]) {
      node->last_ray_hit = FALSE;
      return FALSE;
    }

    // Determine t-values for entry and exit in X.
    t_entry[X] = (maxbound[X] - p_x) / d_x;
    t_exit [X] = (minbound[X] - p_x) / d_x;
  } else {
    // The ray isn't moving appreciably in X. Check to see if the starting
    // point is outside the bounding planes.
    if ((p_x < minbound[X]) || (p_x > maxbound[X])) {
      node->last_ray_hit = FALSE;
      return FALSE;
    }

    // The ray is always inside the bounds in X.
    t_entry[X] = -INFINITY;
    t_exit [X] = +INFINITY;
  }

  if (d_y > TINY) {
    // The ray is heading towards +Y. Check to see if it is outside the
    // box where it starts.
    if (p_y > maxbound[Y]) {
      node->last_ray_hit = FALSE;
      return FALSE;
    }

    // Determine t-values for entry and exit in Y.
    t_entry[Y] = (minbound[Y] - p_y) / d_y;
    t_exit [Y] = (maxbound[Y] - p_y) / d_y;
  } else if (d_y < -TINY) {
    // The ray is heading towards -Y. Check to see if it is outside the
    // box where it starts.
    if (p_y < minbound[Y]) {
      node->last_ray_hit = FALSE;
      return FALSE;
    }

    // Determine t-values for entry and exit in Y.
    t_entry[Y] = (maxbound[Y] - p_y) / d_y;
    t_exit [Y] = (minbound[Y] - p_y) / d_y;
  } else {
    // The ray isn't moving appreciably in Y. Check to see if the starting
    // point is outside the bounding planes.
    if ((p_y < minbound[Y]) || (p_y > maxbound[Y])) {
      node->last_ray_hit = FALSE;
      return FALSE;
    }

    // The ray is always inside the bounds in Y.
    t_entry[Y] = -INFINITY;
    t_exit [Y] = +INFINITY;
  }

  if (d_z > TINY) {
    // The ray is heading towards +Z. Check to see if it is outside the
    // box where it starts.
    if (p_z > maxbound[Z]) {
      node->last_ray_hit = FALSE;
      return FALSE;
    }

    // Determine t-values for entry and exit in Z.
    t_entry[Z] = (minbound[Z] - p_z) / d_z;
    t_exit [Z] = (maxbound[Z] - p_z) / d_z;
  } else if (d_z < -TINY) {
    // The ray is heading towards -Z. Check to see if it is outside the
    // box where it starts.
    if (p_z < minbound[Z]) {
      node->last_ray_hit = FALSE;
      return FALSE;
    }

    // Determine t-values for entry and exit in Z.
    t_entry[Z] = (maxbound[Z] - p_z) / d_z;
    t_exit [Z] = (minbound[Z] - p_z) / d_z;
  } else {
    // The ray isn't moving appreciably in Z. Check to see if the starting
    // point is outside the bounding planes.
    if ((p_z < minbound[Z]) || (p_z > maxbound[Z])) {
      node->last_ray_hit = FALSE;
      return FALSE;
    }

    // The ray is always inside the bounds in Z.
    t_entry[Z] = -INFINITY;
    t_exit [Z] = +INFINITY;
  }

  // Determine the overall (maximum) t_entry and overall (minimum) t_exit.
  if (t_entry[X] > t_entry[Y])
    if (t_entry[X] > t_entry[Z])
      overall_entry = t_entry[X];
    else
      overall_entry = t_entry[Z];
  else
    if (t_entry[Y] > t_entry[Z])
      overall_entry = t_entry[Y];
    else
      overall_entry = t_entry[Z];

  if (t_exit[X] < t_exit[Y])
    if (t_exit[X] < t_exit[Z])
      overall_exit = t_exit[X];
    else
      overall_exit = t_exit[Z];
  else
    if (t_exit[Y] < t_exit[Z])
      overall_exit = t_exit[Y];
    else
      overall_exit = t_exit[Z];

  // If there are no intersections then don't traverse the octree.
  if (overall_entry > overall_exit) {
    node->last_ray_hit = FALSE;
    return FALSE;
  }

  /*
   * Otherwise, we DO need to traverse the octree.
   */

  // Set up the entry t value for the first cell. This is given by the
  // t-value of the entry of the octree bounds. However if this is less
  // than zero, we do not want to accept any intersections behind the
  // eyepoint.
  cell_entry_t = MAX(0, overall_entry);

  // Intersect the ray with the octree structure. This traverses the ray
  // through the cells of the octree. In each cell the ray is intersected
  // with the RCSG tree found there. If any valid intersections are found
  // and the stop_at_first_intersection flag is set, then the traversal
  // stops. Otherwise keep looking for more intersections.
  ray_result = octree_ray_intersect(node, validate_node, the_list, 
				    cluster_ray_from, cluster_ray_dir,
				    ray_flags, new_ray_signature());

  node->last_ray_hit = ray_result;
  return ray_result;
}

// This routine must never be called. The OctreeClusterPrimitive is
// tied to Mirage's object structure.
int OctreeClusterPrimitive::ray_intersect(vector ray_from, vector ray_dir,
					  rayintersect *int_storage)
{
  fatal_error("OctreeClusterPrimitive::ray_intersect", "Incorrect r_i call!",
	      "This error should not happen.");
  return 0;
}

intersect_result OctreeClusterPrimitive::point_intersect(primnode *node,
							 vector apoint,
							 PointInfo *info,
							 uint ray_flags)
{
  vector cluster_point;

  // Transform the sample point into the cluster's primitive space.
  Matrix::apply(node->position[AFT], apoint, cluster_point);

  if ((cluster_point[X] < minbound[X]) || (cluster_point[X] > maxbound[X]) ||
      (cluster_point[Y] < minbound[Y]) || (cluster_point[Y] > maxbound[Y]) ||
      (cluster_point[Z] < minbound[Z]) || (cluster_point[Z] > maxbound[Z]))
    return OUT;

  // Classify this point with respect to the primitives in the RCSG.
  return classify_point_in_rcsg(myRCSG, cluster_point, info, ray_flags);
}

// This routine must never be called. The OctreeClusterPrimitive is
// tied to Mirage's object structure.
intersect_result OctreeClusterPrimitive::point_intersect(vector apoint)
{
  fatal_error("OctreeClusterPrimitive::point_intersect", "Incorrect p_i call!",
	      "This error should not happen.");
  return OUT;
}

intersect_result OctreeClusterPrimitive::voxel_intersect(box v)
{
  box cluster_voxel_check;
  int i;
  
  // Check to see if the voxel needs to be intersected with the
  // cluster (see if it intersects the bounding box).
  // We do this by transforming the voxel to be checked so that it
  // lies in a space that maps minbound -> -1 and maxbound -> +1,
  // and then using the cube voxel intersection method.
  for (i = 0; i < 8; i++) {
    cluster_voxel_check[i][X] = -1 + 2*(v[i][X] - minbound[X]) / bound_size[X];
    cluster_voxel_check[i][Y] = -1 + 2*(v[i][Y] - minbound[Y]) / bound_size[Y];
    cluster_voxel_check[i][Z] = -1 + 2*(v[i][Z] - minbound[Z]) / bound_size[Z];
  }

  if (cube_voxel_intersect(cluster_voxel_check) == OUT) {
    // This voxel does not intersect the cluster, so the cluster doesn't need
    // to be intersected with inside it.
    return OUT;
  } else {
    // This voxel either intersects or is inside the cluster. The cluster does
    // need to be intersected with inside it.
    return SURFACE;
  }
}

char *OctreeClusterPrimitive::prim_name(void) {
  return "Octree cluster";
}



#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif
