/* prim_plane.c : The plane primitive
** 
** The plane primitive defines a half-space that occupies all space
** above the plane Z = 0.
**
** Chris Butcher, 11 December 1996
*/

#include "mirage.h"

#ifdef USE_NAMESPACES
namespace Mirage {
#endif

int planePrimitive::ray_intersect(IntersectionList *the_list, primnode *node,
				  primnode *validate_node, 
				  vector ray_from, vector ray_dir,
				  uint ray_flags, signature ray_signature)
{
  // The Z coordinate of ray_from in primitive space
  double p_z;
  // The Z coordinate of ray_dir in primitive space
  double d_z;
  // The t-value of the intersection point.
  double t;
  // The normal at the hitpoint.
  vector norm;
  // The matrix to transform from world to primitive space.
  matrix *m;

  // Store the signature of this ray that we're intersecting with.
  node->last_ray_signature = ray_signature;

  // Get the matrix that defines the transformation from world to primitve
  // space.
  m = &(node->position[AFT]);

  // Transform the ray start point and direction into primitive space.
  // Note that we only require the Z coordinate.
  p_z = (*m)[2][0]*ray_from[X] + (*m)[2][1]*ray_from[Y]
    + (*m)[2][2]*ray_from[Z] + (*m)[2][3];
  d_z = (*m)[2][0]*ray_dir[X] + (*m)[2][1]*ray_dir[Y] + (*m)[2][2]*ray_dir[Z];


  // If the ray is heading away from the plane (if p_z and d_z have
  // the same sign) then no intersection.
  if ((d_z > 0) == (p_z > 0)) {
    node->last_ray_hit = FALSE;
    return node->last_ray_hit;
  }


  // If the ray is parallel to the surface then no intersection.
  if (fabs(d_z) < TINY) {
    node->last_ray_hit = FALSE;
    return node->last_ray_hit;
  }

  // Get the t value of the intersection.
  t = - p_z / d_z;

  // Stop self-intersection in the shadow ray etc. by discarding
  // intersections with very low t value.
  if (t < MIN_TVAL_CUTOFF) {
    node->last_ray_hit = FALSE;
    return node->last_ray_hit;
  }

  // Calculate the normal.
  Vector::assign(norm, 0, 0, -1);

  // Insert the intersection into the intersection list. The ray is
  // ENTERING if d_z > 0, LEAVING if d_z < 0.
  if (d_z > 0) {
    insert_entering(the_list, node, validate_node, norm, t,
		    ray_from, ray_dir, ray_flags);
  } else {
    insert_leaving(the_list, node, validate_node, norm, t,
		   ray_from, ray_dir, ray_flags);
  }

  node->last_ray_hit = TRUE;
  return node->last_ray_hit;
}

int planePrimitive::ray_intersect(vector ray_from, vector ray_dir,
				  rayintersect *int_storage)
{
  // The t-value of the intersection point.
  double t;
  // The normal at the hitpoint.
  vector norm;
  int num_intersections = 0;

  // If the ray is heading away from the plane (if p_z and d_z have
  // the same sign) then no intersection.
  if ((ray_dir[Z] > 0) == (ray_from[Z] > 0)) {
    return 0;
  }


  // If the ray is parallel to the surface then no intersection.
  if (fabs(ray_dir[Z]) < TINY) {
    return 0;
  }

  // Get the t value of the intersection.
  t = - ray_from[Z] / ray_dir[Z];

  // Calculate the normal.
  Vector::assign(norm, 0, 0, -1);

  // Insert the intersection into the intersection list. The ray is
  // ENTERING if ray_dir[Z] > 0, LEAVING if ray_dir[Z] < 0.
  int_storage[num_intersections].t = t;
  int_storage[num_intersections].is_entering = (ray_dir[Z] > 0) ? TRUE : FALSE;
  Vector::set(int_storage[num_intersections].normal, norm);
  num_intersections++;

  return num_intersections;
}

intersect_result planePrimitive::point_intersect(primnode *node,
						 vector wp,
						 PointInfo *info,
						 uint ray_flags)
{
  // Here wp is the world-point that we are testing for IN/OUT.
  // z is its Z-coordinate in primitive space (given by position[AFT])
  double z;
  matrix *m;

  // Get the matrix to transform from world to primitive space.
  m = &(node->position[AFT]);

  // Get the primitive Z coordinate.
  z = (*m)[2][0]*wp[X] + (*m)[2][1]*wp[Y] + (*m)[2][2]*wp[Z] + (*m)[2][3];

  if (z >= 0.0) {
    set_point_contents(node, info);
    return IN;
  } else
    return OUT;
}

intersect_result planePrimitive::point_intersect(vector wp)
{
  if (wp[Z] >= 0.0)
    return IN;
  else
    return OUT;
}

intersect_result planePrimitive::voxel_intersect(box b)
{
  // This routine checks to see if a parallelipiped lies inside, outside,
  // or intersects the plane. 
  // If all eight of the points lie in +Z, the box is inside.
  // If all eight of the points lie in -Z, the box is outside.
  // Otherwise the box is SURFACE.

  // Which points are inside the half-space?
  int i, first_point_in, this_point_in;

  // Work out if the first point is in.
  first_point_in = (b[0][Z] >= 0.0);

  // For all the rest,
  i = 1;
  do {
    // work out if they are in.
    this_point_in = (b[i][Z] >= 0.0);
    i++;
    // If the point is on the opposite side of the surface from the first
    // point, then the box must be SURFACE.
    if (this_point_in != first_point_in)
      return SURFACE;
  } while (i < 8);
  
  // If all eight points are checked, they must all be the same. If they
  // are all in, return IN.
  if (this_point_in)
    return IN;
  else
    return OUT;
}

char *planePrimitive::prim_name(void) {
  return "Plane";
}



#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif
