/* prim_sphere.c : The sphere primitive
** 
** The sphere primitive defines a unit sphere centred on the origin.
**
** Chris Butcher, 3 December 1996
*/

#include "mirage.h"

#ifdef USE_NAMESPACES
namespace Mirage {
#endif

int spherePrimitive::ray_intersect(IntersectionList *the_list, primnode *node,
				   primnode *validate_node, 
				   vector ray_from, vector ray_dir,
				   uint ray_flags, signature ray_signature)
{
  // The coordinates of ray_from in primitive space
  double p_x, p_y, p_z;
  // The direction ray_dir in primitive space
  double d_x, d_y, d_z;
  // The coefficients of the quadratic equation in t
  double a, b, c;
  // The discriminant of this equation
  double disc;
  // Have we found an intersection?
  int found = 0;
  // The t-values of the intersection points.
  double t_0, t_1;
  // The normal at the hitpoint.
  vector norm;

  // Store the signature of this ray that we're intersecting with.
  node->last_ray_signature = ray_signature;

  // This apply is an inline function defined in matrix.h
  // Transform the ray startpoint into primitive space
  Matrix::apply(node->position[AFT], ray_from, &p_x, &p_y, &p_z);

  // This apply is also an inline function defined in matrix.h
  // Transform the ray direction into primitive space (note: shift does
  // not apply to this, only apply rotation + stretch sections of the matrix)
  Matrix::applynoshift(node->position[AFT], ray_dir, &d_x, &d_y, &d_z);

  // Form the coefficients of the quadratic equation that defines t.
  // This equation is |p + td| = 1.0
  a = SQR(d_x) + SQR(d_y) + SQR(d_z);
  b = 2.0 * (p_x*d_x + p_y*d_y + p_z*d_z);
  c = SQR(p_x) + SQR(p_y) + SQR(p_z) - 1.0;

  // Calculate the discriminant of the quadratic equation.
  disc = SQR(b) - 4.0*a*c;

  // If the discriminant is less than or equal to zero, then there
  // are no intersections with the sphere (tangent rays are treated as
  // no intersection at all).
  if (disc <= 0) {
    node->last_ray_hit = FALSE;
    return node->last_ray_hit;
  }

  // a is equal to the square of the magnitude of the direction of the ray.
  // This *must* not be zero.
  if (a == 0)
    fatal_error("spherePrimitive::ray_intersect", "ray direction zero",
		"This error should not happen.");

  // Calculate the intersection t-values. Must have t_0 < t_1 for later
  // calculations to work.
  // Note that we can save a square root operation by saying
  // t_1 = c / a * t_0 ... but this only works if we can be sure that
  // t_0 isn't zero. It may easily be if sqrt(disc) == b or == -b.
  // So, we make sure that the +/-sqrt(disc) is always the same sign
  // as -b to make sure that the subtraction can't produce zero.
  // This makes the code a little long-winded.
  if (a >= 0) {
    if (b >= 0) {
      t_0 = (-b - sqrt(disc)) / (2 * a);
      t_1 = c / (a * t_0);
    } else {
      t_1 = (-b + sqrt(disc)) / (2 * a);
      t_0 = c / (a * t_1);
    }
  } else {
    if (b >= 0) {
      t_1 = (-b - sqrt(disc)) / (2 * a);
      t_0 = c / (a * t_1);
    } else {
      t_0 = (-b + sqrt(disc)) / (2 * a);
      t_1 = c / (a * t_0);
    }
  }

  // Insert any positive t-values (valid intersections).
  if (t_0 > MIN_TVAL_CUTOFF) {
    // Get the normal (since this is a unit sphere, the normal is the
    // same as the primitive-space hitpoint = p + td).
    norm[X] = p_x + t_0*d_x;
    norm[Y] = p_y + t_0*d_y;
    norm[Z] = p_z + t_0*d_z;
    // Note that since t_0 < t_1 always, this intersection is an ENTERING
    // one.
    //    printf("SPH: insert 0\n");
    insert_entering(the_list, node, validate_node, norm, t_0,
		    ray_from, ray_dir, ray_flags);
    found = 1;
  }
  if (t_1 > MIN_TVAL_CUTOFF) {
    // Get the normal (since this is a unit sphere, the normal is the
    // same as the primitive-space hitpoint = p + td).
    norm[X] = p_x + t_1*d_x;
    norm[Y] = p_y + t_1*d_y;
    norm[Z] = p_z + t_1*d_z;
    // Note that since t_0 < t_1 always, this intersection is a LEAVING
    // one.
    //    printf("SPH: insert 1\n");
    insert_leaving(the_list, node, validate_node, norm, t_1,
		   ray_from, ray_dir, ray_flags);
    found = 1;
  }

  if (found)
    node->last_ray_hit = TRUE;
  else
    node->last_ray_hit = FALSE;

  return node->last_ray_hit;
}

int spherePrimitive::ray_intersect(vector ray_from, vector ray_dir,
				   rayintersect *int_storage)
{
  // The coefficients of the quadratic equation in t
  double a, b, c;
  // The discriminant of this equation
  double disc;
  // Have we found an intersection?
  int num_intersections = 0;
  // The t-values of the intersection points.
  double t_0, t_1;
  // The normal at the hitpoint.
  vector norm;

  // Form the coefficients of the quadratic equation that defines t.
  // This equation is |p + td| = 1.0
  a = SQR(ray_dir[X]) + SQR(ray_dir[Y]) + SQR(ray_dir[Z]);
  b = 2.0 * (ray_from[X]*ray_dir[X] + ray_from[Y]*ray_dir[Y]
	     + ray_from[Z]*ray_dir[Z]);
  c = SQR(ray_from[X]) + SQR(ray_from[Y]) + SQR(ray_from[Z]) - 1.0;

  // Calculate the discriminant of the quadratic equation.
  disc = SQR(b) - 4.0*a*c;

  // If the discriminant is less than or equal to zero, then there
  // are no intersections with the sphere (tangent rays are treated as
  // no intersection at all).
  if (disc <= 0) {
    return 0;
  }

  // a is equal to the square of the magnitude of the direction of the ray.
  // This *must* not be zero.
  if (a == 0)
    fatal_error("spherePrimitive::ray_intersect", "ray direction zero",
		"This error should not happen.");

  // Calculate the intersection t-values. Must have t_0 < t_1 for later
  // calculations to work.
  // Note that we can save a square root operation by saying
  // t_1 = c / a * t_0 ... but this only works if we can be sure that
  // t_0 isn't zero. It may easily be if sqrt(disc) == b or == -b.
  // So, we make sure that the +/-sqrt(disc) is always the same sign
  // as -b to make sure that the subtraction can't produce zero.
  // This makes the code a little long-winded.
  if (a >= 0) {
    if (b >= 0) {
      t_0 = (-b - sqrt(disc)) / (2 * a);
      t_1 = c / (a * t_0);
    } else {
      t_1 = (-b + sqrt(disc)) / (2 * a);
      t_0 = c / (a * t_1);
    }
  } else {
    if (b >= 0) {
      t_1 = (-b - sqrt(disc)) / (2 * a);
      t_0 = c / (a * t_1);
    } else {
      t_0 = (-b + sqrt(disc)) / (2 * a);
      t_1 = c / (a * t_0);
    }
  }

  // Get the normal (since this is a unit sphere, the normal is the
  // same as the primitive-space hitpoint = p + td).
  norm[X] = ray_from[X] + t_0*ray_dir[X];
  norm[Y] = ray_from[Y] + t_0*ray_dir[Y];
  norm[Z] = ray_from[Z] + t_0*ray_dir[Z];
  
  // Note that since t_0 < t_1 always, this intersection is an ENTERING
  // one.
  int_storage[num_intersections].t = t_0;
  int_storage[num_intersections].is_entering = TRUE;
  Vector::set(int_storage[num_intersections].normal, norm);
  num_intersections++;

  // Get the normal (since this is a unit sphere, the normal is the
  // same as the primitive-space hitpoint = p + td).
  norm[X] = ray_from[X] + t_1*ray_dir[X];
  norm[Y] = ray_from[Y] + t_1*ray_dir[Y];
  norm[Z] = ray_from[Z] + t_1*ray_dir[Z];
  
  // Note that since t_0 < t_1 always, this intersection is a LEAVING
  // one.
  int_storage[num_intersections].t = t_1;
  int_storage[num_intersections].is_entering = FALSE;
  Vector::set(int_storage[num_intersections].normal, norm);
  num_intersections++;

  return num_intersections;
}

intersect_result spherePrimitive::point_intersect(primnode *node,
						  vector wp,
						  PointInfo *info,
						  uint ray_flags)
{
  // Here wp is the world-point that we are testing for IN/OUT.
  // The values p_x, p_y, p_z are its coordinates in primitive space
  // (the mapping from world to primitive space is given by position.aft)
  double p_x, p_y, p_z, d_sq;

  // This apply is an inline function defined in matrix.h
  // Transform the worldpoint into primitive space.
  Matrix::apply(node->position[AFT], wp, &p_x, &p_y, &p_z);

  // d_sq is the point's distance from the sphere's origin, squared.
  d_sq = SQR(p_x) + SQR(p_y) + SQR(p_z);
  
  // If the distance squared is <= 1 squared, the point is IN.
  if (d_sq <= 1.0) {
    set_point_contents(node, info);
    return IN;
  } else
    return OUT;
}

intersect_result spherePrimitive::point_intersect(vector wp)
{
  // If the distance squared is <= 1 squared, the point is IN.
  if (SQR(wp[X]) + SQR(wp[Y]) + SQR(wp[Z]) <= 1.0)
    return IN;
  else
    return OUT;
}





// Now define some local macros to make the code length shorter for the
// routine spherePrimitive::voxel_intersect.


// Macro CHECK_OUTSIDE_PLUS_PLANE tests to see if the points of the box all
// lie outside the plane P at distance +1.
#define CHECK_OUTSIDE_PLUS_PLANE(axis) {   \
                                           \
  for (i = 0; i < 8; i++) {                \
    if (b[i][axis] < 1.0) {                \
      break;                               \
    }                                      \
  }                                        \
  if (i == 8) {                            \
    return OUT;                            \
  }                                        \
}


// Macro CHECK_OUTSIDE_MINUS_PLANE tests to see if the points of the box all
// lie outside the plane P at distance -1.
#define CHECK_OUTSIDE_MINUS_PLANE(axis) {   \
                                            \
  for (i = 0; i < 8; i++)                   \
    if (b[i][axis] > -1.0)                  \
      break;                                \
                                            \
  if (i == 8) {                              \
    return OUT;                             \
  }                                         \
}


// Macro DISTANCE_FROM_FACE calculates the distance of the origin from
// a face of the box (defined by the three points with indices i, j, k).
#define DISTANCE_FROM_FACE(h, i, j, k)                 \
                                                       \
h = (b[i][X] * (b[j][Y]*b[k][Z] - b[j][Z]*b[k][Y])     \
   - b[i][Y] * (b[j][X]*b[k][Z] - b[j][Z]*b[k][X])     \
   + b[i][Z] * (b[j][X]*b[k][Y] - b[j][Y]*b[k][X]));

     
// Macro CHECK_EDGE_INTERSECT checks the edge between points i and j
// to see if it intersects the unit sphere. If it does, return SURFACE.
#define CHECK_EDGE_INTERSECT(i, j)                                        \
                                                                          \
ij_dot = b[i][X]*b[j][X] + b[i][Y]*b[j][Y] + b[i][Z]*b[j][Z];             \
t1 = radius_sq[i] - ij_dot;                                               \
t2 = radius_sq[j] - ij_dot;                                               \
                                                                          \
if (t1 >= 0 && t2 >= 0 && (1.0 - radius_sq[i] + (SQR(t1)/(t1 + t2)) > 0)) \
  return SURFACE;


// Macro GENERATE_NORMAL generates the normal to a face f by looking at
// two vertices i and j that lie on that face. Two vertices do not define
// a plane, so we use vertices 0, i and j to define the plane.
#define GENERATE_NORMAL(f, i, j)                        \
                                                        \
opposite_face = f + 3;                                  \
a_x = b[i][X] - b[0][X];                                \
a_y = b[i][Y] - b[0][Y];                                \
a_z = b[i][Z] - b[0][Z];                                \
b_x = b[j][X] - b[0][X];                                \
b_y = b[j][Y] - b[0][Y];                                \
b_z = b[j][Z] - b[0][Z];                                \
                                                        \
box_normal[f][X] = a_y*b_z - b_y*a_z;                   \
box_normal[f][Y] = a_z*b_x - b_z*a_x;                   \
box_normal[f][Z] = a_x*b_y - b_x*a_y;                   \
box_normal[opposite_face][X] = -box_normal[f][X];       \
box_normal[opposite_face][Y] = -box_normal[f][Y];       \
box_normal[opposite_face][Z] = -box_normal[f][Z];       


// Macro CHECK_FACE_INTERSECT checks for intersection of the sphere with
// the plane p within the four planes defined by the normals
// i, j, k and l.
#define CHECK_FACE_INTERSECT(p, i, j, k, l)                            \
                                                                       \
normal_sq = SQR(box_normal[p][X]) + SQR(box_normal[p][Y])              \
              + SQR(box_normal[p][Z]);                                 \
                                                                       \
if (normal_sq >= SQR(height[p])) {                                     \
  intersect_pt[X] = box_normal[p][X]*height[p]/normal_sq;              \
  intersect_pt[Y] = box_normal[p][Y]*height[p]/normal_sq;              \
  intersect_pt[Z] = box_normal[p][Z]*height[p]/normal_sq;              \
  if ((Vector::dot(box_normal[i], intersect_pt) - height[i] >= 0) &&   \
      (Vector::dot(box_normal[j], intersect_pt) - height[j] >= 0) &&   \
      (Vector::dot(box_normal[k], intersect_pt) - height[k] >= 0) &&   \
      (Vector::dot(box_normal[l], intersect_pt) - height[l] >= 0)) {   \
    return SURFACE;                                                    \
  }                                                                    \
}

intersect_result spherePrimitive::voxel_intersect(box b)
{
  // This routine checks to see if a parallelipiped lies inside, outside,
  // or intersects the sphere. It has five steps:
  // 1) If all of the points lie at X >= 1, the box is outside. Similarly
  //    for X = -1 and for the Y and Z axes.
  // 2) Count the points inside the sphere. If there are eight, then the
  //    box is inside. If there are between one and seven, then the box must
  //    cross the surface of the sphere and so is SURFACE.
  // 3) All the points are outside. Is the origin inside the box? If so,
  //    then the box contains the sphere, return SURFACE.
  // 4) Check to see if any of the edges of the box intersects the sphere.
  //    If so, return SURFACE.
  // 5) Check to see if the sphere intersects any face by calculating the
  //    center of the circle of intersection of the sphere with each face
  //    of the box. If it lies inside the polygon formed by the four
  //    points on the face, then the sphere must intersect the box and so
  //    return SURFACE.
  // 6) None of the above have returned a SURFACE intersection so the sphere
  //    lies outside the box, return OUT.

  // Loop counter
  int i;
  // Is the first point inside the sphere?
  int first_point_in;
  // Is this point inside the sphere?
  int this_point_in;
  // The distances squared from the origin of the eight points.
  double radius_sq[8];
  // The heights of the origin relative to each box face (> 0 == inside)
  double height[6];
  // The dot product of two vertices (used in CHECK_EDGE_INTERSECT)
  double ij_dot;
  // Two t values used for the edge-intersect code.
  double t1, t2;
  // The normal vectors to each face of the box. Note: these have the right
  // direction, but magnitude isn't guaranteed to be unity.
  vector box_normal[6];
  // The face opposite the face currently being considered (used by
  // GENERATE_NORMAL).
  int opposite_face;
  // Components of the two vectors used to generate the normal vector.
  double a_x, a_y, a_z, b_x, b_y, b_z;
  // Magnitude squared of the normal vector under consideration (used
  // in CHECK_FACE_INTERSECT).
  double normal_sq;
  // Center of the circle of intersection (used in CHECK_FACE_INTERSECT).
  vector intersect_pt;


  // Step 1: Check to see if all of the points of the box lie outside
  //         some unit coordinate plane, (X | Y | Z) = +/- 1. If so,
  //         return OUT.
  CHECK_OUTSIDE_PLUS_PLANE(X);
  CHECK_OUTSIDE_PLUS_PLANE(Y);
  CHECK_OUTSIDE_PLUS_PLANE(Z);
  CHECK_OUTSIDE_MINUS_PLANE(X);
  CHECK_OUTSIDE_MINUS_PLANE(Y);
  CHECK_OUTSIDE_MINUS_PLANE(Z);



  // Step 2: Count the number of points of the box that are inside the sphere.
  //         If eight, return IN. If zero, continue on, otherwise return
  //         SURFACE.
  //           Note that this means that if we find two that are different,
  //         we can stop and return SURFACE.

  // Work out if the first point is in.
  radius_sq[0] = SQR(b[0][X]) + SQR(b[0][Y]) + SQR(b[0][Z]);
  first_point_in = (radius_sq[0] <= 1.0);

  // For all the rest,
  i = 1;
  do {
    // work out if they are in.
    radius_sq[i] = SQR(b[i][X]) + SQR(b[i][Y]) + SQR(b[i][Z]);
    this_point_in = (radius_sq[i] <= 1.0);
    i++;

    // If the point is on the opposite side of the surface from the first
    // point, then the box must be SURFACE.
    if (this_point_in != first_point_in)
      return SURFACE;
  } while (i < 8);
  
  // If all eight points are checked, they must all be the same. If they
  // are all in, return IN.
  if (this_point_in)
    return IN;



  // Step 3: Check to see if the origin lies inside the box. It does if
  //         all of the heights from each box face are greater than zero.
  //         If so, return SURFACE.

  CHECK_VOXEL_HANDEDNESS(b);

  // These heights may be used in step 5 if this step fails, so
  // calculate them anyway.
  DISTANCE_FROM_FACE(height[0], 0, 1, 2);
  DISTANCE_FROM_FACE(height[1], 0, 4, 1);
  DISTANCE_FROM_FACE(height[2], 0, 2, 4);
  DISTANCE_FROM_FACE(height[3], 7, 5, 6);
  DISTANCE_FROM_FACE(height[4], 7, 6, 3);
  DISTANCE_FROM_FACE(height[5], 7, 3, 5);
  if ((height[0] < 0) && (height[1] < 0) && (height[2] < 0) &&
      (height[3] < 0) && (height[4] < 0) && (height[5] < 0))
    return SURFACE;



  // Step 4: Check each edge to see if it intersects the sphere. If
  //         it does, then return SURFACE.
  CHECK_EDGE_INTERSECT(0, 1);
  CHECK_EDGE_INTERSECT(0, 2);
  CHECK_EDGE_INTERSECT(0, 4);
  CHECK_EDGE_INTERSECT(7, 3);
  CHECK_EDGE_INTERSECT(7, 5);
  CHECK_EDGE_INTERSECT(7, 6);
  CHECK_EDGE_INTERSECT(4, 5);
  CHECK_EDGE_INTERSECT(4, 6);
  CHECK_EDGE_INTERSECT(2, 6);
  CHECK_EDGE_INTERSECT(2, 3);
  CHECK_EDGE_INTERSECT(1, 5);
  CHECK_EDGE_INTERSECT(1, 3);


  // Step 5: Check each face to see if it intersects the sphere
  //         by looking at where the center of the circle of intersection
  //         of the sphere with the face's plane is. If the center lies inside
  //         the volume bounded by the four planes that touch this one,
  //         then it is inside the polygon defined by the face's corners,
  //         so return SURFACE.
  GENERATE_NORMAL(0, 1, 2);
  GENERATE_NORMAL(1, 4, 1);
  GENERATE_NORMAL(2, 2, 4);
  CHECK_FACE_INTERSECT(0, 1, 2, 4, 5);
  CHECK_FACE_INTERSECT(3, 1, 2, 4, 5);
  CHECK_FACE_INTERSECT(2, 0, 1, 3, 4);
  CHECK_FACE_INTERSECT(5, 0, 1, 3, 4);
  CHECK_FACE_INTERSECT(1, 0, 2, 3, 5);
  CHECK_FACE_INTERSECT(4, 0, 2, 3, 5);


  
  // Step 6: All of the intersection tests above have fallen through, the
  //         sphere is outside the box.
  return OUT;
}

// Undefine the macros defined above (local only).
#undef CHECK_OUTSIDE_PLUS_PLANE
#undef CHECK_OUTSIDE_MINUS_PLANE
#undef DISTANCE_FROM_FACE
#undef CHECK_EDGE_INTERSECT
#undef GENERATE_NORMAL
#undef CHECK_FACE_INTERSECT

char *spherePrimitive::prim_name(void) {
  return "Sphere";
}


#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif
