/* prim_stupidcluster.c : The stupid cluster primitive
** 
** This file contains the function(s) for the StupidClusterPrimitive.
**
** Chris Butcher, 5 December 1996
*/

#include "mirage.h"

#ifdef USE_NAMESPACES
namespace Mirage {
#endif

// The stupidClusterPrimitive doesn't have any subdivision.
// It must however recurse its RCSG graph looking for cluster primitives
// that might have to be subdivided.
void StupidClusterPrimitive::subdivide(void)
{
  // If we haven't already done so, look for clusters that need to be
  // subdivided.
  if (!done_subdivision) {
    subdivide_clusters_in_RCSG(myRCSG);

    done_subdivision = TRUE;
  }
}

// The ray-intersection routine for StupidClusterPrimitive just
// intersects the ray with all the primitives in the RCSG.
int StupidClusterPrimitive::ray_intersect(IntersectionList *the_list,
					  primnode *node,
					  primnode *validate_node,
					  vector ray_from, vector ray_dir,
					  uint ray_flags,
					  signature ray_signature)
{
  // Ray relative to the cluster's primitive space.
  vector cluster_ray_from, cluster_ray_dir;
  int found_result;
  intersection *list_position;

  // Store the signature of this ray that we're intersecting with.
  node->last_ray_signature = ray_signature;

  // Transform the ray into the cluster's primitive space.
  Matrix::apply(node->position[AFT], ray_from, cluster_ray_from);
  Matrix::applynoshift(node->position[AFT], ray_dir, cluster_ray_dir);

  // Intersect the ray with the RCSG tree.
  list_position = the_list->get_position();
  found_result = rcsg_ray_intersect(the_list, myRCSG, node,
				    cluster_ray_from, cluster_ray_dir,
				    ray_flags, new_ray_signature());
  //  printf("SC: rcsg_ray found %d\n", found_result);
  the_list->set_position(list_position);

  // Validate intersections with the RCSG.
  //
  // Note that since this isn't a cell-based subdivision scheme, we don't
  // want to reject intersections outside the current cell so the min/max
  // distances are -infinity and +2*infinity (2* as the max distance is
  // measured from the min distance, not from zero).
  if (found_result)
    found_result = validate_intersections(the_list, node, validate_node,
					  cluster_ray_from, cluster_ray_dir,
					  ray_flags, VALIDATE_RCSG,
					  myRCSG, -INFINITY, +2*INFINITY);

  //  printf("SC: validated %d\n", found_result);

  // Delete intersections that are outside the current cell (shouldn't
  // be any, but make sure).
  tidy_intersections(the_list, node);

  // Store the result of this ray.
  node->last_ray_hit = found_result;

  return found_result;
}

// This routine must never be called. The StupidClusterPrimitive is
// tied to Mirage's object structure.
int StupidClusterPrimitive::ray_intersect(vector ray_from, vector ray_dir,
					  rayintersect *int_storage)
{
  fatal_error("StupidClusterPrimitive::ray_intersect", "Incorrect r_i call!",
	      "This error should not happen.");
  return 0;
}

intersect_result StupidClusterPrimitive::point_intersect(primnode *node,
							 vector apoint,
							 PointInfo *info,
							 uint ray_flags)
{
  vector cluster_point;

  // Transform the sample point into the cluster's primitive space.
  Matrix::apply(node->position[AFT], apoint, cluster_point);

  // Classify this point with respect to the primitives in the RCSG.
  return classify_point_in_rcsg(myRCSG, cluster_point, info, ray_flags);
}

// This routine must never be called. The StupidClusterPrimitive is
// tied to Mirage's object structure.
intersect_result StupidClusterPrimitive::point_intersect(vector apoint)
{
  fatal_error("StupidClusterPrimitive::ray_intersect", "Incorrect r_i call!",
	      "This error should not happen.");
  return OUT;
}

intersect_result StupidClusterPrimitive::voxel_intersect(box v)
{
  // This cluster has no bounding box. Every ray must be intersected with
  // its objects.
  return SURFACE;
}

char *StupidClusterPrimitive::prim_name(void) {
  return "Stupid cluster";
}



#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif
