/* prim_taper.c : The tapering torus primitive!!
** 
** The tapering torus primitive defines a segment of a torus from
** theta = 0 (along X) to theta = bend_amount. The radius of the torus
** changes from rad_base to rad_top along its length. The torus's defining
** circle has radius 1.
**
** Chris Butcher, 25 June 1997
*/

#include "mirage.h"

#ifdef USE_NAMESPACES
namespace Mirage {
#endif

// A macro to test whether we are inside at a given t value.
// If slightly outside the area on either plane, this returns as if inside
// the plane. Exact clipping on the bounding planes is done by the interval
// arithmetic.
#define TEST_INSIDE(t, inside) {                                \
  in_x = ray_from[X] + t * ray_dir[X];                          \
  in_y = ray_from[Y] + t * ray_dir[Y];                          \
  inside_theta = atan2(in_y, in_x);                             \
                                                                \
  if ((inside_theta < -0.01) ||                                 \
      (inside_theta > bend_amount + 0.01)) {                    \
    inside = FALSE;                                             \
  } else {                                                      \
    inside_radius = rad_base + (rad_top - rad_base)             \
                         * (inside_theta / bend_amount);        \
                                                                \
    in_xy = sqrt(SQR(in_x) + SQR(in_y));                        \
    in_z = ray_from[Z] + t * ray_dir[Z];                        \
    if (SQR(in_xy - 1) + SQR(in_z) < SQR(inside_radius))        \
      inside = TRUE;                                            \
    else                                                        \
      inside = FALSE;                                           \
  }								\
}

int taperPrimitive::ray_intersect(IntersectionList *the_list, 
				  primnode *node, primnode *validate_node, 
				  vector ray_from,
				  vector ray_dir, uint ray_flags,
				  signature ray_signature)
{
  // Have we found any intersections?
  int found, int1_discarded, int2_discarded;
  
  // The coordinates of ray_from in primitive space
  double p_x, p_y, p_z;
  // The direction ray_dir in primitive space
  double d_x, d_y, d_z;
  // Intermediate quadratic expression coefficients (a b c, d e f)
  double a, b, c, d, e, f;
  // The coefficients of the quartic equation in t
  // Note that q_coeff[i] is the coefficient of t^i
  double q_coeff[5];
  // The solutions of the quartic equation.
  double q_sol[5];
  // The number of intersections.
  int q_solutions, i, j, k;

  // The intersections with the planes.
  double base_t, end_t;
  plane_intersect base_intersect, end_intersect;
  double interval_start, interval_end;

  // Used to calculate the end intersection.
  double dir_dot_n, from_dot_n;

  // Variables for sorting intersections.
  double sorted[4];

  // Variables for stepping through intervals.
  double step_t, last_step_t, step_size;
  double interval_min, interval_max, interval_mid;
  int keep_stepping, step_in, last_step_in, min_in, max_in, mid_in, iterations;

  // Variables used by the TEST_INSIDE macro.
  double in_x, in_y, in_xy, in_z, inside_radius, inside_theta;

  // The intervals that we need to search for intersections in.
  double int1_start, int1_end, int2_start, int2_end;
  int int1_start_in, int1_end_in, int2_start_in, int2_end_in;

  // The primitive-space hitpoint.
  double hx, hy, hz;

  // Precalculated values to make the normal calculation nicer.
  double xy_dist, radius_increase, hit_radius, one_over_yx_sq;

  // The normal at the hitpoint.
  vector norm;

  // No intersections yet.
  found = FALSE;

  // Store the signature of this ray that we're intersecting with.
  node->last_ray_signature = ray_signature;

  // This apply is an inline function defined in matrix.h
  // Transform the ray startpoint into primitive space
  Matrix::apply(node->position[AFT], ray_from, &p_x, &p_y, &p_z);

  // This apply is also an inline function defined in matrix.h
  // Transform the ray direction into primitive space (note: shift does
  // not apply to this, only apply rotation + stretch sections of the matrix)
  Matrix::applynoshift(node->position[AFT], ray_dir, &d_x, &d_y, &d_z);

  // First: calculate the intersection with the base of the taper (Y = 0).
  if (fabs(d_y) < TINY) {
    // The ray is parallel to the base. If the ray is above the base then
    // we say it intersects it as an entry at -INFINITY. If the ray is outside
    // the base plane then there are no intersections, we return FALSE.
    if (p_y > 0) {
      base_intersect = ENTRY;
      base_t = -INFINITY;
    } else {
      node->last_ray_hit = FALSE;
      return node->last_ray_hit;
    }
  } else {
    // Intersect with the base.
    base_t = - p_y / d_y;

    if (d_y < 0) {
      // This plane is an EXIT intersection.
      base_intersect = EXIT;
    } else {
      // This plane is an ENTRY intersection.
      base_intersect = ENTRY;
    }
  }

  // Second: calculate the intersection(s) with the end of the taper. This
  // is the plane defined by the point end_p and the normal end_n.
  dir_dot_n = d_x * end_n[X] + d_y * end_n[Y] + d_z * end_n[Z];
  from_dot_n = p_x * end_n[X] + p_y * end_n[Y] + p_z * end_n[Z];
  if (fabs(dir_dot_n) < TINY) {
    // The ray is parallel to the end of the taper. If the ray is below the end
    // then we say it intersects it as an entry as -INFINITY. If the ray is
    // above the end then there are no intersections with the taper, and
    // so we return FALSE.
    if (from_dot_n < 0) {
      // p is below the plane, inside for the ray's whole length.
      end_intersect = ENTRY;
      end_t = -INFINITY;
    } else {
      // p is above the plane, no intersections possible.
      node->last_ray_hit = FALSE;
      return node->last_ray_hit;
    }
  } else {
    // Intersect with the end.
    end_t = - from_dot_n / dir_dot_n;

    if (dir_dot_n > 0) {
      // This plane is an EXIT intersection.
      end_intersect = EXIT;
    } else {
      // This plane is an ENTRY intersection.
      end_intersect = ENTRY;
    }
  }

  // The intersections with the planes give us an interval that we must
  // look for intersections with the taper within.
  if (base_intersect == ENTRY) {
    if (end_intersect == ENTRY) {
      // The two planes must BOTH be entered before we can start looking for
      // valid intersections. There is no end to the interval.
      interval_start = MAX(base_t, end_t);
      interval_end = +INFINITY;
    } else {
      // Enter by the base, exit by the end.
      interval_start = base_t;
      interval_end = end_t;
    }
  } else {
    if (end_intersect == ENTRY) {
      // Enter by the end, exit by the base.
      interval_start = end_t;
      interval_end = base_t;
    } else {
      // The first one crossed of the two EXIT planes spells the end of the
      // interval; there is no start limit.
      interval_start = -INFINITY;
      interval_end = MIN(base_t, end_t);
    }
  }

  // Check to see that we actually have an interval to look in.
  if (interval_start > interval_end) {
    node->last_ray_hit = FALSE;
    return node->last_ray_hit;
  }

  // The equation that defines the torus is
  // (xy_distance)^2 + (z_distance)^2 = (torus_rad)^2.
  // Since the point is given by (p + td), this gives
  //
  // (sqrt((px + tdx)^2 + (py + tdy)^2) - 1.0)^2 + (pz + tdz)^2 = torus_rad^2
  //
  // We expand the xy_distance squared and collect the sqrt terms on one
  // side to give
  //
  // (px + tdx)^2 + (py + tdy)^2 + (pz + tdz)^2 + 1 - torus_rad^2 =
  //                                    2 sqrt((px + tdx)^2 + (py + tdy)^2)
  //
  // Expanding all the terms (p + td) gives us the equation
  //
  // (dx^2 + dy^2 + dz^2)t^2 + 2(pxdx + pydy + pzdz)t +
  //    (px^2 + py^2 + pz^2 + 1 - torus_rad^2) = 2 sqrt( (dx^2 + dy^2)t^2 +
  //                                        2(pxdx + pydy)t + (px^2 + py^2) )
  //
  // This is an equation of the type
  //   at^2 + bt + c = 2 sqrt(dt^2 + et + f)
  //
  // where a = dx^2 + dy^2 + dz^2
  //       b = 2(pxdx + pydy + pzdz)
  //       c = px^2 + py^2 + pz^2 + 1 - torus_rad^2
  //       d = dx^2 + dy^2
  //       e = 2(pxdx + pydy)
  //       f = px^2 + py^2
  //
  // We square both sides to give
  //
  // (a^2)t^4 + (2ab)t^3 + (2ac + b^2 - 4d)t^2 + (2bc - 4e)t + (c^2 - 4f) = 0
  //
  // which can then be solved using the SolveQuartic function.

  // Calculate the coefficients for the quadratic expressions (a b c, d e f)
#if 0
  // Slow version, not used (included for clarity)
  a = SQR(d_x) + SQR(d_y) + SQR(d_z);
  b = 2.0 * (p_x*d_x + p_y*d_y + p_z*d_z);
  c = SQR(p_x) + SQR(p_y) + SQR(p_z) + 1 - SQR(torus_rad);
  d = SQR(d_x) + SQR(d_y);
  e = 2.0 * (p_x*d_x + p_y*d_y);
  f = SQR(p_x) + SQR(p_y);
#else
  // Note that this is the same as above, just optimized slightly.
  // Also note that since b and e will be multiplied by two in the
  // calculation of the quartic coefficients, what we calculate here is
  // really 1/2 b and 1/2 e to save a multiplication.
  d = SQR(d_x) + SQR(d_y);
  e = p_x*d_x + p_y*d_y;
  f = SQR(p_x) + SQR(p_y);

  a = d + SQR(d_z);
  b = e + p_z*d_z;
  c = f + SQR(p_z) + 1 - SQR(rad_base);
#endif

  // Calculate the coefficients of the quartic equation to be solved.
  q_coeff[4] = SQR(a);
  q_coeff[3] = 4*a*b;                      // 2ab = 4a(1/2b)
  q_coeff[2] = 2*a*c + 4*SQR(b) - 4*d;     // b^2 = 4*(1/2b)^2
  q_coeff[1] = 4*b*c - 8*e;                // 2bc = 4c(1/2b) and 4e = 8(1/2e)
  q_coeff[0] = SQR(c) - 4*f;

  // Solve the quartic equation that describes the torus which bounds the
  // outer extent of the taper.
  q_solutions = solveQuartic(q_coeff, q_sol);

  if (q_solutions == 0) {
    // There are no intersections with the torus anywhere along the
    // ray; this is a clean miss.
    node->last_ray_hit = FALSE;
    return node->last_ray_hit;
  } else if (q_solutions == 2) {
    // There is one interval to test. It is bounded by the two solutions to
    // the bounding torus and since the taper doesn't extend that far out,
    // we know that both ends are OUT.
    if (q_sol[0] > q_sol[1]) {
      int1_start = q_sol[1];
      int1_end   = q_sol[0];
    } else {
      int1_start = q_sol[0];
      int1_end   = q_sol[1];
    }
    int1_start_in = FALSE;
    int1_end_in   = FALSE;
    
    if (interval_start > int1_end) {
      // The intersections are both before the interval defined by the
      // planes. No intersections.
      node->last_ray_hit = found;
      return node->last_ray_hit;
    } else if (interval_start > int1_start) {
      // There is an ENTRY intersection with a plane in the middle of the
      // interval. Truncate the interval to this plane.
      int1_start = interval_start;
      TEST_INSIDE(int1_start, int1_start_in);

      // If this plane is a valid entry intersection (i.e. the intersection
      // with the plane is inside the bounds of the taper) then insert it
      // into the intersection list.
      if (int1_start_in && (int1_start > MIN_TVAL_CUTOFF * 10)) {
	found = TRUE;
	if (int1_start == base_t) {
	  // Insert an entry intersection with the base of the taper.
	  Vector::assign(norm, 0, -1, 0);
	  insert_entering(the_list, node, validate_node, norm, int1_start,
			  ray_from, ray_dir, ray_flags);
	} else {
	  // Insert an entry intersection with the end of the taper.
	  insert_leaving(the_list, node, validate_node, end_n, int1_start,
			 ray_from, ray_dir, ray_flags);
	}
      }
    }

    if (interval_end < int1_start) {
      // The intersections are both after the interval defined by the
      // planes. No intersections.
      node->last_ray_hit = found;
      return node->last_ray_hit;
    } else if (interval_end < int1_end) {
      // There is an EXIT intersection with a plane in the middle of the
      // interval. Truncate the interval to this plane.
      int1_end = interval_end;
      TEST_INSIDE(int1_end, int1_end_in);

      // If this plane is a valid exit intersection (i.e. the intersection
      // with the plane is inside the bounds of the taper) then insert it
      // into the intersection list.
      if (int1_end_in && (int1_end > MIN_TVAL_CUTOFF * 10)) {
	found = TRUE;
	if (int1_end == base_t) {
	  // Insert an exit intersection with the base of the taper.
	  Vector::assign(norm, 0, -1, 0);
	  insert_leaving(the_list, node, validate_node, norm, int1_end,
			 ray_from, ray_dir, ray_flags);
	} else {
	  // Insert an exit intersection with the end of the taper.
	  insert_leaving(the_list, node, validate_node, end_n, int1_end,
			 ray_from, ray_dir, ray_flags);
	}
      }
    }

    // Traverse the remaining interval.
    step_size = (int1_end - int1_start) / traversal_num_steps;
    step_in = int1_start_in;
    step_t = int1_start;
    keep_stepping = TRUE;
    iterations = 100;

    do {
      // Store the current position.
      last_step_t = step_t;
      last_step_in = step_in;

      // Calculate the next sample point.
      step_t += step_size;

      // Are we past the end?
      if (step_t > int1_end) {
	// Use the data that we already have about the endpoint.
	step_t = int1_end;
	step_in = int1_end_in;

	keep_stepping = FALSE;

      } else {
	// Calculate whether or not we are inside the taper.
	TEST_INSIDE(step_t, step_in);
      }

      // Have we crossed the surface?
      if (step_in != last_step_in) {
	// Perform a binary search to find the surface. The starting points
	// are the last step and the current position.
	interval_min = last_step_t;       min_in = last_step_in;
	interval_max = step_t;            max_in = step_in;

	// Are we close enough to an answer?
	while (interval_max - interval_min > binary_search_limit) {
	  // Calculate the midpoint and inside-test it.
	  interval_mid = (interval_max + interval_min) / 2;
	  TEST_INSIDE(interval_mid, mid_in);

	  // Contract the interval being searched.
	  if (mid_in == min_in) {
	    interval_min = interval_mid;
	  } else {
	    interval_max = interval_mid;
	  }
	}

	// We have refined the interval to the desired precision. Calculate
	// the normal and insert the intersection.
	interval_mid = (interval_max + interval_min) / 2;

	if (interval_mid > MIN_TVAL_CUTOFF * 10) {
	  // We find X, Y and Z at the hitpoint.
	  hx = p_x + interval_mid * d_x;
	  hy = p_y + interval_mid * d_y;
	  hz = p_z + interval_mid * d_z;
	  
	  // Precalculate some values for calculation of the normal.
	  xy_dist = CHECK0(sqrt(SQR(hx) + SQR(hy)),
			   "xy_dist in taperPrimitive::ray_intersect");
	  radius_increase = (rad_top - rad_base) / bend_amount;
	  hit_radius = rad_base + radius_increase * atan2(hy, hx);
	  one_over_yx_sq = 1 / 
	    (1 + SQR(hy / CHECK0(hx, "hx in taperPrimitive::ray_intersect")));
	  
	  // Calculate the normal. This is the gradient of the equation that
	  // defines the taper ... 
	  //
	  // f(x, y, z) = 0
	  //
	  // f = (sqrt(x^2 + y^2) - 1)^2 + z^2 - hit_rad^2 = 0
	  //
	  // where hit_rad = r1 + (r2 - r1) * (tan-1(y/x) / bend_amount)
	  //
	  // We calculate the three partial derivatives and form the normal
	  // from them.
	  norm[X] = 2 * (xy_dist - 1) * (hx / xy_dist)
	    - 2 * hit_radius * radius_increase
	    * one_over_yx_sq * (hy * (- 1 / SQR(hx)));
	  norm[Y] = 2 * (xy_dist - 1) * (hy / xy_dist)
	    - 2 * hit_radius * radius_increase
	    * one_over_yx_sq * (1 / hx);
	  norm[Z] = 2 * hz;
	  Vector::normalise(norm, "normal in taperPrimitive::ray_intersect");

	  found = TRUE;

	  if (min_in == FALSE) {
	    // Insert an entry intersection with the side of the taper.
	    insert_entering(the_list, node, validate_node, norm, interval_mid,
			    ray_from, ray_dir, ray_flags);
	  } else {
	    // Insert an exit intersection with the side of the taper.
	    insert_leaving(the_list, node, validate_node, norm, interval_mid,
			   ray_from, ray_dir, ray_flags);
	  }
	}
      }

      // Finished processing this step. Keep going.
    } while (keep_stepping && (iterations-- > 0));

  } else if (q_solutions == 4) {
    // There are two intervals to test.
    
    // We sort the four intersections in order of ascending t so that
    // we can order them into two intervals.
    for (i = 0; i < 4; i++) {

      // Find a place for intersection #i.
      for (j = 0; j <= i; j++) {
	if (j == i) {
	  // If we've reached the end successfully, this is the biggest t-value
	  // that we've found yet.
	  sorted[j] = q_sol[i];

	  // Exit from the loop - we've placed this value.
	  break;
	} else {
	  // Check to see if the current value deserves to go here (less than
	  // the current one).
	  if (q_sol[i] < sorted[j]) {
	    // Move all the existing values along.
	    for (k = i; k > j; k--)
	      sorted[k] = sorted[k - 1];

	    // Store the new value.
	    sorted[j] = q_sol[i];

	    // Exit from the loop - we've placed this value.
	    break;
	  }
	}
      }
    }

    int1_start = sorted[0];    int1_start_in = FALSE;
    int1_end   = sorted[1];    int1_end_in   = FALSE;
    int2_start = sorted[2];    int2_start_in = FALSE;
    int2_end   = sorted[3];    int2_end_in   = FALSE;

    /*
     * Truncate and traverse interval 1.
     */
    int1_discarded = FALSE;

    if (interval_start > int1_end) {
      // The intersections are both before the interval defined by the
      // planes. No intersections.
      int1_discarded = TRUE;
    } else if (interval_start > int1_start) {
      // There is an ENTRY intersection with a plane in the middle of the
      // interval. Truncate the interval to this plane.
      int1_start = interval_start;
      TEST_INSIDE(int1_start, int1_start_in);

      // If this plane is a valid entry intersection (i.e. the intersection
      // with the plane is inside the bounds of the taper) then insert it
      // into the intersection list.
      if (int1_start_in && (int1_start > MIN_TVAL_CUTOFF * 10)) {
	found = TRUE;
	if (int1_start == base_t) {
	  // Insert an entry intersection with the base of the taper.
	  Vector::assign(norm, 0, -1, 0);
	  insert_entering(the_list, node, validate_node, norm, int1_start,
			  ray_from, ray_dir, ray_flags);
	} else {
	  // Insert an entry intersection with the end of the taper.
	  insert_entering(the_list, node, validate_node, end_n, int1_start,
			  ray_from, ray_dir, ray_flags);
	}
      }
    }

    if (!int1_discarded) {
      if (interval_end < int1_start) {
	// The intersections are both after the interval defined by the
	// planes. No intersections.
	int1_discarded = TRUE;
      } else if (interval_end < int1_end) {
	// There is an EXIT intersection with a plane in the middle of the
	// interval. Truncate the interval to this plane.
	int1_end = interval_end;
	TEST_INSIDE(int1_end, int1_end_in);
	
	// If this plane is a valid exit intersection (i.e. the intersection
	// with the plane is inside the bounds of the taper) then insert it
	// into the intersection list.
	if (int1_end_in && (int1_end > MIN_TVAL_CUTOFF * 10)) {
	  found = TRUE;
	  if (int1_end == base_t) {
	    // Insert an exit intersection with the base of the taper.
	    Vector::assign(norm, 0, -1, 0);
	    insert_leaving(the_list, node, validate_node, norm, int1_end,
			   ray_from, ray_dir, ray_flags);
	  } else {
	    // Insert an exit intersection with the end of the taper.
	    insert_leaving(the_list, node, validate_node, norm, int1_end,
			   ray_from, ray_dir, ray_flags);
	  }
	}
      }
    }

    if (!int1_discarded) {
      // Traverse the remaining interval.
      step_size = (int1_end - int1_start) / traversal_num_steps;
      step_in = int1_start_in;
      step_t = int1_start;
      keep_stepping = TRUE;
      do {
	// Store the current position.
	last_step_t = step_t;
	last_step_in = step_in;
	
	// Calculate the next sample point.
	step_t += step_size;
	
	// Are we past the end?
	if (step_t > int1_end) {
	  // Use the data that we already have about the endpoint.
	  step_t = int1_end;
	  step_in = int1_end_in;
	  
	  keep_stepping = FALSE;
	} else {
	  // Calculate whether or not we are inside the taper.
	  TEST_INSIDE(step_t, step_in);
	}
	
	// Have we crossed the surface?
	if (step_in != last_step_in) {
	  // Perform a binary search to find the surface. The starting points
	  // are the last step and the current position.
	  interval_min = last_step_t;       min_in = last_step_in;
	  interval_max = step_t;            max_in = step_in;
	  
	  // Are we close enough to an answer?
	  while (interval_max - interval_min > binary_search_limit) {
	    // Calculate the midpoint and inside-test it.
	    interval_mid = (interval_max + interval_min) / 2;
	    TEST_INSIDE(interval_mid, mid_in);
	    
	    // Contract the interval being searched.
	    if (mid_in == min_in) {
	      interval_min = interval_mid;
	    } else {
	      interval_max = interval_mid;
	    }
	  }
	  
	  // We have refined the interval to the desired precision. Calculate
	  // the normal and insert the intersection.
	  interval_mid = (interval_max + interval_min) / 2;
	  
	  if (interval_mid > MIN_TVAL_CUTOFF * 10) {
	    // We find X, Y and Z at the hitpoint.
	    hx = p_x + interval_mid * d_x;
	    hy = p_y + interval_mid * d_y;
	    hz = p_z + interval_mid * d_z;
	    
	    // Precalculate some values for calculation of the normal.
	    xy_dist = CHECK0(sqrt(SQR(hx) + SQR(hy)),
			     "xy_dist in taperPrimitive::ray_intersect");
	    radius_increase = (rad_top - rad_base) / bend_amount;
	    hit_radius = rad_base + radius_increase * atan2(hy, hx);
	    one_over_yx_sq = 1 / 
	      (1 + SQR(hy / CHECK0(hx,
				   "hx in taperPrimitive::ray_intersect")));
	    
	    // Calculate the normal. This is the gradient of the equation that
	    // defines the taper ... 
	    //
	    // f(x, y, z) = 0
	    //
	    // f = (sqrt(x^2 + y^2) - 1)^2 + z^2 - hit_rad^2 = 0
	    //
	    // where hit_rad = r1 + (r2 - r1) * (tan-1(y/x) / bend_amount)
	    //
	    // We calculate the three partial derivatives and form the normal
	    // from them.
	    norm[X] = 2 * (xy_dist - 1) * (hx / xy_dist)
	      - 2 * hit_radius * radius_increase
	      * one_over_yx_sq * (hy * (- 1 / SQR(hx)));
	    norm[Y] = 2 * (xy_dist - 1) * (hy / xy_dist)
	      - 2 * hit_radius * radius_increase
	      * one_over_yx_sq * (1 / hx);
	    norm[Z] = 2 * hz;

	    found = TRUE;
	    
	    if (min_in == FALSE) {
	      // Insert an entry intersection with the side of the taper.
	      insert_entering(the_list, node, validate_node, norm,
			      interval_mid, ray_from, ray_dir, ray_flags);
	    } else {
	      // Insert an exit intersection with the side of the taper.
	      insert_leaving(the_list, node, validate_node, norm,
			     interval_mid, ray_from, ray_dir, ray_flags);
	    }
	  }
	}
	
	// Finished processing this step. Keep going.
      } while (keep_stepping);
      
      // Done traversing interval 1.
    }
    
    /*
     * Truncate and traverse interval 2.
     */
    int2_discarded = FALSE;

    if (interval_start > int2_end) {
      // The intersections are both before the interval defined by the
      // planes. No intersections.
      int2_discarded = TRUE;
    } else if (interval_start > int2_start) {
      // There is an ENTRY intersection with a plane in the middle of the
      // interval. Truncate the interval to this plane.
      int2_start = interval_start;
      TEST_INSIDE(int2_start, int2_start_in);

      // If this plane is a valid entry intersection (i.e. the intersection
      // with the plane is inside the bounds of the taper) then insert it
      // into the intersection list.
      if (int2_start_in && (int2_start > MIN_TVAL_CUTOFF * 10)) {
	found = TRUE;
	if (int2_start == base_t) {
	  // Insert an entry intersection with the base of the taper.
	  Vector::assign(norm, 0, -1, 0);
	  insert_entering(the_list, node, validate_node, norm, int2_start,
			  ray_from, ray_dir, ray_flags);
	} else {
	  // Insert an entry intersection with the end of the taper.
	  insert_entering(the_list, node, validate_node, end_n, int2_start,
			  ray_from, ray_dir, ray_flags);
	}
      }
    }

    if (!int2_discarded) {
      if (interval_end < int2_start) {
	// The intersections are both after the interval defined by the
	// planes. No intersections.
	node->last_ray_hit = found;
	return node->last_ray_hit;
      } else if (interval_end < int2_end) {
	// There is an EXIT intersection with a plane in the middle of the
	// interval. Truncate the interval to this plane.
	int2_end = interval_end;
	TEST_INSIDE(int2_end, int2_end_in);
	
	// If this plane is a valid exit intersection (i.e. the intersection
	// with the plane is inside the bounds of the taper) then insert it
	// into the intersection list.
	if (int2_end_in && (int2_end > MIN_TVAL_CUTOFF * 10)) {
	  found = TRUE;
	  if (int2_end == base_t) {
	    // Insert an exit intersection with the base of the taper.
	    Vector::assign(norm, 0, -1, 0);
	    insert_leaving(the_list, node, validate_node, norm, int2_end,
			   ray_from, ray_dir, ray_flags);
	  } else {
	    // Insert an exit intersection with the end of the taper.
	    insert_leaving(the_list, node, validate_node, end_n, int2_end,
			   ray_from, ray_dir, ray_flags);
	  }
	}
      }
    }

    if (!int2_discarded) {
      // Traverse the remaining interval.
      step_size = (int2_end - int2_start) / traversal_num_steps;
      step_in = int2_start_in;
      step_t = int2_start;
      keep_stepping = TRUE;
      do {
	// Store the current position.
	last_step_t = step_t;
	last_step_in = step_in;
	
	// Calculate the next sample point.
	step_t += step_size;
	
	// Are we past the end?
	if (step_t > int2_end) {
	  // Use the data that we already have about the endpoint.
	  step_t = int2_end;
	  step_in = int2_end_in;
	  
	  keep_stepping = FALSE;
	} else {
	  // Calculate whether or not we are inside the taper.
	  TEST_INSIDE(step_t, step_in);
	}
	
	// Have we crossed the surface?
	if (step_in != last_step_in) {
	  // Perform a binary search to find the surface. The starting points
	  // are the last step and the current position.
	  interval_min = last_step_t;       min_in = last_step_in;
	  interval_max = step_t;            max_in = step_in;
	  
	  // Are we close enough to an answer?
	  while (interval_max - interval_min > binary_search_limit) {
	    // Calculate the midpoint and inside-test it.
	    interval_mid = (interval_max + interval_min) / 2;
	    TEST_INSIDE(interval_mid, mid_in);
	    
	    // Contract the interval being searched.
	    if (mid_in == min_in) {
	      interval_min = interval_mid;
	    } else {
	      interval_max = interval_mid;
	    }
	  }
	  
	  // We have refined the interval to the desired precision. Calculate
	  // the normal and insert the intersection.
	  interval_mid = (interval_max + interval_min) / 2;
	  
	  if (interval_mid > MIN_TVAL_CUTOFF * 10) {
	    found = TRUE;

	    // We find X, Y and Z at the hitpoint.
	    hx = p_x + interval_mid * d_x;
	    hy = p_y + interval_mid * d_y;
	    hz = p_z + interval_mid * d_z;
	    
	    // Precalculate some values for calculation of the normal.
	    xy_dist = CHECK0(sqrt(SQR(hx) + SQR(hy)),
			     "xy_dist in taperPrimitive::ray_intersect");
	    radius_increase = (rad_top - rad_base) / bend_amount;
	    hit_radius = rad_base + radius_increase * atan2(hy, hx);
	    one_over_yx_sq = 1 / 
	      (1 + SQR(hy / CHECK0(hx, 
				   "hx in taperPrimitive::ray_intersect")));
	    
	    // Calculate the normal. This is the gradient of the equation that
	    // defines the taper ... 
	    //
	    // f(x, y, z) = 0
	    //
	    // f = (sqrt(x^2 + y^2) - 1)^2 + z^2 - hit_rad^2 = 0
	    //
	    // where hit_rad = r1 + (r2 - r1) * (tan-1(y/x) / bend_amount)
	    //
	    // We calculate the three partial derivatives and form the normal
	    // from them.
	    norm[X] = 2 * (xy_dist - 1) * (hx / xy_dist)
	      - 2 * hit_radius * radius_increase
	      * one_over_yx_sq * (hy * (- 1 / SQR(hx)));
	    norm[Y] = 2 * (xy_dist - 1) * (hy / xy_dist)
	      - 2 * hit_radius * radius_increase
	      * one_over_yx_sq * (1 / hx);
	    norm[Z] = 2 * hz;
	    
	    if (min_in == FALSE) {
	      // Insert an entry intersection with the side of the taper.
	      insert_entering(the_list, node, validate_node, norm,
			      interval_mid, ray_from, ray_dir, ray_flags);
	    } else {
	      // Insert an exit intersection with the side of the taper.
	      insert_leaving(the_list, node, validate_node, norm, interval_mid,
			     ray_from, ray_dir, ray_flags);
	    }
	  }
	}
	
	// Finished processing this step. Keep going.
      } while (keep_stepping);
      
      // Done traversing interval 2.
    }
    
  } else {
    fatal_error("taperPrimitive::ray_intersect",
		"Quartic has odd number of real solutions!",
		"Obviously the Conjugate Root Theorem was wrong after all...");
    node->last_ray_hit = FALSE;
    return node->last_ray_hit;
  }

  // Return whether or not we found any intersections.
  node->last_ray_hit = found;
  return node->last_ray_hit;
}

int taperPrimitive::ray_intersect(vector ray_from, vector ray_dir,
				  rayintersect *int_storage)
{
  // Have we found any intersections?
  int num_intersections = 0, int1_discarded, int2_discarded;
   
  // Intermediate quadratic expression coefficients (a b c, d e f)
  double a, b, c, d, e, f;
  // The coefficients of the quartic equation in t
  // Note that q_coeff[i] is the coefficient of t^i
  double q_coeff[5];
  // The solutions of the quartic equation.
  double q_sol[5];
  // The number of intersections.
  int q_solutions, i, j, k;

  // The intersections with the planes.
  double base_t, end_t;
  plane_intersect base_intersect, end_intersect;
  double interval_start, interval_end;

  // Used to calculate the end intersection.
  double dir_dot_n, from_dot_n;

  // Variables for sorting intersections.
  double sorted[4];

  // Variables for stepping through intervals.
  double step_t, last_step_t, step_size;
  double interval_min, interval_max, interval_mid;
  int keep_stepping, step_in, last_step_in, min_in, max_in, mid_in, iterations;

  // Variables used by the TEST_INSIDE macro.
  double in_x, in_y, in_xy, in_z, inside_radius, inside_theta;

  // The intervals that we need to search for intersections in.
  double int1_start, int1_end, int2_start, int2_end;
  int int1_start_in, int1_end_in, int2_start_in, int2_end_in;

  // The primitive-space hitpoint.
  double hx, hy, hz;

  // Precalculated values to make the normal calculation nicer.
  double xy_dist, radius_increase, hit_radius, one_over_yx_sq;

  // The normal at the hitpoint.
  vector norm;

  // First: calculate the intersection with the base of the taper (Y = 0).
  if (fabs(ray_dir[Y]) < TINY) {
    // The ray is parallel to the base. If the ray is above the base then
    // we say it intersects it as an entry at -INFINITY. If the ray is outside
    // the base plane then there are no intersections, we return FALSE.
    if (ray_from[Y] > 0) {
      base_intersect = ENTRY;
      base_t = -INFINITY;
    } else {
      return 0;
    }
  } else {
    // Intersect with the base.
    base_t = - ray_from[Y] / ray_dir[Y];

    if (ray_dir[Y] < 0) {
      // This plane is an EXIT intersection.
      base_intersect = EXIT;
    } else {
      // This plane is an ENTRY intersection.
      base_intersect = ENTRY;
    }
  }

  // Second: calculate the intersection(s) with the end of the taper. This
  // is the plane defined by the point end_p and the normal end_n.
  dir_dot_n = ray_dir[X] * end_n[X] + ray_dir[Y] * end_n[Y]
    + ray_dir[Z] * end_n[Z];
  from_dot_n = ray_from[X] * end_n[X] + ray_from[Y] * end_n[Y]
    + ray_from[Z] * end_n[Z];
  if (fabs(dir_dot_n) < TINY) {
    // The ray is parallel to the end of the taper. If the ray is below the end
    // then we say it intersects it as an entry as -INFINITY. If the ray is
    // above the end then there are no intersections with the taper, and
    // so we return FALSE.
    if (from_dot_n < 0) {
      // p is below the plane, inside for the ray's whole length.
      end_intersect = ENTRY;
      end_t = -INFINITY;
    } else {
      // p is above the plane, no intersections possible.
      return 0;
    }
  } else {
    // Intersect with the end.
    end_t = - from_dot_n / dir_dot_n;

    if (dir_dot_n > 0) {
      // This plane is an EXIT intersection.
      end_intersect = EXIT;
    } else {
      // This plane is an ENTRY intersection.
      end_intersect = ENTRY;
    }
  }

  // The intersections with the planes give us an interval that we must
  // look for intersections with the taper within.
  if (base_intersect == ENTRY) {
    if (end_intersect == ENTRY) {
      // The two planes must BOTH be entered before we can start looking for
      // valid intersections. There is no end to the interval.
      interval_start = MAX(base_t, end_t);
      interval_end = +INFINITY;
    } else {
      // Enter by the base, exit by the end.
      interval_start = base_t;
      interval_end = end_t;
    }
  } else {
    if (end_intersect == ENTRY) {
      // Enter by the end, exit by the base.
      interval_start = end_t;
      interval_end = base_t;
    } else {
      // The first one crossed of the two EXIT planes spells the end of the
      // interval; there is no start limit.
      interval_start = -INFINITY;
      interval_end = MIN(base_t, end_t);
    }
  }

  // Check to see that we actually have an interval to look in.
  if (interval_start > interval_end) {
    return 0;
  }

  // The equation that defines the torus is
  // (xy_distance)^2 + (z_distance)^2 = (torus_rad)^2.
  // Since the point is given by (p + td), this gives
  //
  // (sqrt((px + tdx)^2 + (py + tdy)^2) - 1.0)^2 + (pz + tdz)^2 = torus_rad^2
  //
  // We expand the xy_distance squared and collect the sqrt terms on one
  // side to give
  //
  // (px + tdx)^2 + (py + tdy)^2 + (pz + tdz)^2 + 1 - torus_rad^2 =
  //                                    2 sqrt((px + tdx)^2 + (py + tdy)^2)
  //
  // Expanding all the terms (p + td) gives us the equation
  //
  // (dx^2 + dy^2 + dz^2)t^2 + 2(pxdx + pydy + pzdz)t +
  //    (px^2 + py^2 + pz^2 + 1 - torus_rad^2) = 2 sqrt( (dx^2 + dy^2)t^2 +
  //                                        2(pxdx + pydy)t + (px^2 + py^2) )
  //
  // This is an equation of the type
  //   at^2 + bt + c = 2 sqrt(dt^2 + et + f)
  //
  // where a = dx^2 + dy^2 + dz^2
  //       b = 2(pxdx + pydy + pzdz)
  //       c = px^2 + py^2 + pz^2 + 1 - torus_rad^2
  //       d = dx^2 + dy^2
  //       e = 2(pxdx + pydy)
  //       f = px^2 + py^2
  //
  // We square both sides to give
  //
  // (a^2)t^4 + (2ab)t^3 + (2ac + b^2 - 4d)t^2 + (2bc - 4e)t + (c^2 - 4f) = 0
  //
  // which can then be solved using the SolveQuartic function.

  // Calculate the coefficients for the quadratic expressions (a b c, d e f)
#if 0
  // Slow version, not used (included for clarity)
  a = SQR(ray_dir[X]) + SQR(ray_dir[Y]) + SQR(ray_dir[Z]);
  b = 2.0 * (ray_from[X]*ray_dir[X] + ray_from[Y]*ray_dir[Y]
	     + ray_from[Z]*ray_dir[Z]);
  c = SQR(ray_from[X]) + SQR(ray_from[Y]) + SQR(ray_from[Z])
    + 1 - SQR(torus_rad);
  d = SQR(ray_dir[X]) + SQR(ray_dir[Y]);
  e = 2.0 * (ray_from[X]*ray_dir[X] + ray_from[Y]*ray_dir[Y]);
  f = SQR(ray_from[X]) + SQR(ray_from[Y]);
#else
  // Note that this is the same as above, just optimized slightly.
  // Also note that since b and e will be multiplied by two in the
  // calculation of the quartic coefficients, what we calculate here is
  // really 1/2 b and 1/2 e to save a multiplication.
  d = SQR(ray_dir[X]) + SQR(ray_dir[Y]);
  e = ray_from[X]*ray_dir[X] + ray_from[Y]*ray_dir[Y];
  f = SQR(ray_from[X]) + SQR(ray_from[Y]);

  a = d + SQR(ray_dir[Z]);
  b = e + ray_from[Z]*ray_dir[Z];
  c = f + SQR(ray_from[Z]) + 1 - SQR(rad_base);
#endif

  // Calculate the coefficients of the quartic equation to be solved.
  q_coeff[4] = SQR(a);
  q_coeff[3] = 4*a*b;                      // 2ab = 4a(1/2b)
  q_coeff[2] = 2*a*c + 4*SQR(b) - 4*d;     // b^2 = 4*(1/2b)^2
  q_coeff[1] = 4*b*c - 8*e;                // 2bc = 4c(1/2b) and 4e = 8(1/2e)
  q_coeff[0] = SQR(c) - 4*f;

  // Solve the quartic equation that describes the torus which bounds the
  // outer extent of the taper.
  q_solutions = solveQuartic(q_coeff, q_sol);

  if (q_solutions == 0) {
    // There are no intersections with the torus anywhere along the
    // ray; this is a clean miss.
    return 0;
  } else if (q_solutions == 2) {
    // There is one interval to test. It is bounded by the two solutions to
    // the bounding torus and since the taper doesn't extend that far out,
    // we know that both ends are OUT.
    if (q_sol[0] > q_sol[1]) {
      int1_start = q_sol[1];
      int1_end   = q_sol[0];
    } else {
      int1_start = q_sol[0];
      int1_end   = q_sol[1];
    }
    int1_start_in = FALSE;
    int1_end_in   = FALSE;
    
    if (interval_start > int1_end) {
      // The intersections are both before the interval defined by the
      // planes. No intersections.
      return num_intersections;
    } else if (interval_start > int1_start) {
      // There is an ENTRY intersection with a plane in the middle of the
      // interval. Truncate the interval to this plane.
      int1_start = interval_start;
      TEST_INSIDE(int1_start, int1_start_in);

      // If this plane is a valid entry intersection (i.e. the intersection
      // with the plane is inside the bounds of the taper) then insert it
      // into the intersection list.
      if (int1_start_in) {
	if (int1_start == base_t) {
	  // Insert an entry intersection with the base of the taper.
	  int_storage[num_intersections].t = int1_start;
	  int_storage[num_intersections].is_entering = TRUE;
	  Vector::assign(int_storage[num_intersections].normal, 0, -1, 0);
	  num_intersections++;
	} else {
	  // Insert an entry intersection with the end of the taper.
	  int_storage[num_intersections].t = int1_start;
	  int_storage[num_intersections].is_entering = TRUE;
	  Vector::set(int_storage[num_intersections].normal, end_n);
	  num_intersections++;
	}
      }
    }

    if (interval_end < int1_start) {
      // The intersections are both after the interval defined by the
      // planes. No intersections.
      return num_intersections;
    } else if (interval_end < int1_end) {
      // There is an EXIT intersection with a plane in the middle of the
      // interval. Truncate the interval to this plane.
      int1_end = interval_end;
      TEST_INSIDE(int1_end, int1_end_in);

      // If this plane is a valid exit intersection (i.e. the intersection
      // with the plane is inside the bounds of the taper) then insert it
      // into the intersection list.
      if (int1_end_in) {
	if (int1_end == base_t) {
	  // Insert an exit intersection with the base of the taper.
	  int_storage[num_intersections].t = int1_end;
	  int_storage[num_intersections].is_entering = FALSE;
	  Vector::assign(int_storage[num_intersections].normal, 0, -1, 0);
	  num_intersections++;
	} else {
	  // Insert an exit intersection with the end of the taper.
	  int_storage[num_intersections].t = int1_end;
	  int_storage[num_intersections].is_entering = FALSE;
	  Vector::set(int_storage[num_intersections].normal, end_n);
	  num_intersections++;
	}
      }
    }

    // Traverse the remaining interval.
    step_size = (int1_end - int1_start) / traversal_num_steps;
    step_in = int1_start_in;
    step_t = int1_start;
    keep_stepping = TRUE;
    iterations = 100;

    do {
      // Store the current position.
      last_step_t = step_t;
      last_step_in = step_in;

      // Calculate the next sample point.
      step_t += step_size;

      // Are we past the end?
      if (step_t > int1_end) {
	// Use the data that we already have about the endpoint.
	step_t = int1_end;
	step_in = int1_end_in;

	keep_stepping = FALSE;

      } else {
	// Calculate whether or not we are inside the taper.
	TEST_INSIDE(step_t, step_in);
      }

      // Have we crossed the surface?
      if (step_in != last_step_in) {
	// Perform a binary search to find the surface. The starting points
	// are the last step and the current position.
	interval_min = last_step_t;       min_in = last_step_in;
	interval_max = step_t;            max_in = step_in;

	// Are we close enough to an answer?
	while (interval_max - interval_min > binary_search_limit) {
	  // Calculate the midpoint and inside-test it.
	  interval_mid = (interval_max + interval_min) / 2;
	  TEST_INSIDE(interval_mid, mid_in);

	  // Contract the interval being searched.
	  if (mid_in == min_in) {
	    interval_min = interval_mid;
	  } else {
	    interval_max = interval_mid;
	  }
	}

	// We have refined the interval to the desired precision. Calculate
	// the normal and insert the intersection.
	interval_mid = (interval_max + interval_min) / 2;

	// We find X, Y and Z at the hitpoint.
	hx = ray_from[X] + interval_mid * ray_dir[X];
	hy = ray_from[Y] + interval_mid * ray_dir[Y];
	hz = ray_from[Z] + interval_mid * ray_dir[Z];
	
	// Precalculate some values for calculation of the normal.
	xy_dist = CHECK0(sqrt(SQR(hx) + SQR(hy)),
			 "xy_dist in taperPrimitive::ray_intersect");
	radius_increase = (rad_top - rad_base) / bend_amount;
	hit_radius = rad_base + radius_increase * atan2(hy, hx);
	one_over_yx_sq = 1 / 
	  (1 + SQR(hy / CHECK0(hx, "hx in taperPrimitive::ray_intersect")));
	
	// Calculate the normal. This is the gradient of the equation that
	// defines the taper ... 
	//
	// f(x, y, z) = 0
	//
	// f = (sqrt(x^2 + y^2) - 1)^2 + z^2 - hit_rad^2 = 0
	//
	// where hit_rad = r1 + (r2 - r1) * (tan-1(y/x) / bend_amount)
	//
	// We calculate the three partial derivatives and form the normal
	// from them.
	norm[X] = 2 * (xy_dist - 1) * (hx / xy_dist)
	  - 2 * hit_radius * radius_increase
	  * one_over_yx_sq * (hy * (- 1 / SQR(hx)));
	norm[Y] = 2 * (xy_dist - 1) * (hy / xy_dist)
	  - 2 * hit_radius * radius_increase
	  * one_over_yx_sq * (1 / hx);
	norm[Z] = 2 * hz;
	Vector::normalise(norm, "normal in taperPrimitive::ray_intersect");
	
	// Insert an intersection with the side of the taper.
	// If min_in is TRUE, this is a leaving intersection.
	int_storage[num_intersections].t = interval_mid;
	int_storage[num_intersections].is_entering = (min_in) ? FALSE : TRUE;
	Vector::set(int_storage[num_intersections].normal, norm);
	num_intersections++;
      }

      // Finished processing this step. Keep going.
    } while (keep_stepping && (iterations-- > 0));

  } else if (q_solutions == 4) {
    // There are two intervals to test.
    
    // We sort the four intersections in order of ascending t so that
    // we can order them into two intervals.
    for (i = 0; i < 4; i++) {

      // Find a place for intersection #i.
      for (j = 0; j <= i; j++) {
	if (j == i) {
	  // If we've reached the end successfully, this is the biggest t-value
	  // that we've found yet.
	  sorted[j] = q_sol[i];

	  // Exit from the loop - we've placed this value.
	  break;
	} else {
	  // Check to see if the current value deserves to go here (less than
	  // the current one).
	  if (q_sol[i] < sorted[j]) {
	    // Move all the existing values along.
	    for (k = i; k > j; k--)
	      sorted[k] = sorted[k - 1];

	    // Store the new value.
	    sorted[j] = q_sol[i];

	    // Exit from the loop - we've placed this value.
	    break;
	  }
	}
      }
    }

    int1_start = sorted[0];    int1_start_in = FALSE;
    int1_end   = sorted[1];    int1_end_in   = FALSE;
    int2_start = sorted[2];    int2_start_in = FALSE;
    int2_end   = sorted[3];    int2_end_in   = FALSE;

    /*
     * Truncate and traverse interval 1.
     */
    int1_discarded = FALSE;

    if (interval_start > int1_end) {
      // The intersections are both before the interval defined by the
      // planes. No intersections.
      int1_discarded = TRUE;
    } else if (interval_start > int1_start) {
      // There is an ENTRY intersection with a plane in the middle of the
      // interval. Truncate the interval to this plane.
      int1_start = interval_start;
      TEST_INSIDE(int1_start, int1_start_in);

      // If this plane is a valid entry intersection (i.e. the intersection
      // with the plane is inside the bounds of the taper) then insert it
      // into the intersection list.
      if (int1_start_in) {
	if (int1_start == base_t) {
	  // Insert an entry intersection with the base of the taper.
	  int_storage[num_intersections].t = int1_start;
	  int_storage[num_intersections].is_entering = TRUE;
	  Vector::assign(int_storage[num_intersections].normal, 0, -1, 0);
	  num_intersections++;
	} else {
	  // Insert an entry intersection with the end of the taper.
	  int_storage[num_intersections].t = int1_start;
	  int_storage[num_intersections].is_entering = TRUE;
	  Vector::set(int_storage[num_intersections].normal, end_n);
	  num_intersections++;
	}
      }
    }

    if (!int1_discarded) {
      if (interval_end < int1_start) {
	// The intersections are both after the interval defined by the
	// planes. No intersections.
	int1_discarded = TRUE;
      } else if (interval_end < int1_end) {
	// There is an EXIT intersection with a plane in the middle of the
	// interval. Truncate the interval to this plane.
	int1_end = interval_end;
	TEST_INSIDE(int1_end, int1_end_in);
	
	// If this plane is a valid exit intersection (i.e. the intersection
	// with the plane is inside the bounds of the taper) then insert it
	// into the intersection list.
	if (int1_end_in) {
	  if (int1_end == base_t) {
	    // Insert an exit intersection with the base of the taper.
	    int_storage[num_intersections].t = int1_end;
	    int_storage[num_intersections].is_entering = FALSE;
	    Vector::assign(int_storage[num_intersections].normal, 0, -1, 0);
	    num_intersections++;
	  } else {
	    // Insert an exit intersection with the end of the taper.
	    int_storage[num_intersections].t = int1_end;
	    int_storage[num_intersections].is_entering = FALSE;
	    Vector::set(int_storage[num_intersections].normal, end_n);
	    num_intersections++;
	  }
	}
      }
    }

    if (!int1_discarded) {
      // Traverse the remaining interval.
      step_size = (int1_end - int1_start) / traversal_num_steps;
      step_in = int1_start_in;
      step_t = int1_start;
      keep_stepping = TRUE;
      do {
	// Store the current position.
	last_step_t = step_t;
	last_step_in = step_in;
	
	// Calculate the next sample point.
	step_t += step_size;
	
	// Are we past the end?
	if (step_t > int1_end) {
	  // Use the data that we already have about the endpoint.
	  step_t = int1_end;
	  step_in = int1_end_in;
	  
	  keep_stepping = FALSE;
	} else {
	  // Calculate whether or not we are inside the taper.
	  TEST_INSIDE(step_t, step_in);
	}
	
	// Have we crossed the surface?
	if (step_in != last_step_in) {
	  // Perform a binary search to find the surface. The starting points
	  // are the last step and the current position.
	  interval_min = last_step_t;       min_in = last_step_in;
	  interval_max = step_t;            max_in = step_in;
	  
	  // Are we close enough to an answer?
	  while (interval_max - interval_min > binary_search_limit) {
	    // Calculate the midpoint and inside-test it.
	    interval_mid = (interval_max + interval_min) / 2;
	    TEST_INSIDE(interval_mid, mid_in);
	    
	    // Contract the interval being searched.
	    if (mid_in == min_in) {
	      interval_min = interval_mid;
	    } else {
	      interval_max = interval_mid;
	    }
	  }
	  
	  // We have refined the interval to the desired precision. Calculate
	  // the normal and insert the intersection.
	  interval_mid = (interval_max + interval_min) / 2;
	  
	  // We find X, Y and Z at the hitpoint.
	  hx = ray_from[X] + interval_mid * ray_dir[X];
	  hy = ray_from[Y] + interval_mid * ray_dir[Y];
	  hz = ray_from[Z] + interval_mid * ray_dir[Z];
	  
	  // Precalculate some values for calculation of the normal.
	  xy_dist = CHECK0(sqrt(SQR(hx) + SQR(hy)),
			   "xy_dist in taperPrimitive::ray_intersect");
	  radius_increase = (rad_top - rad_base) / bend_amount;
	  hit_radius = rad_base + radius_increase * atan2(hy, hx);
	  one_over_yx_sq = 1 / 
	    (1 + SQR(hy / CHECK0(hx,
				 "hx in taperPrimitive::ray_intersect")));
	  
	  // Calculate the normal. This is the gradient of the equation that
	  // defines the taper ... 
	  //
	  // f(x, y, z) = 0
	  //
	  // f = (sqrt(x^2 + y^2) - 1)^2 + z^2 - hit_rad^2 = 0
	  //
	  // where hit_rad = r1 + (r2 - r1) * (tan-1(y/x) / bend_amount)
	  //
	  // We calculate the three partial derivatives and form the normal
	  // from them.
	  norm[X] = 2 * (xy_dist - 1) * (hx / xy_dist)
	    - 2 * hit_radius * radius_increase
	    * one_over_yx_sq * (hy * (- 1 / SQR(hx)));
	  norm[Y] = 2 * (xy_dist - 1) * (hy / xy_dist)
	    - 2 * hit_radius * radius_increase
	    * one_over_yx_sq * (1 / hx);
	  norm[Z] = 2 * hz;
	  
	  // Insert an intersection with the side of the taper.
	  // If min_in is TRUE, this is a leaving intersection.
	  int_storage[num_intersections].t = interval_mid;
	  int_storage[num_intersections].is_entering = (min_in)
	    ? FALSE : TRUE;
	  Vector::set(int_storage[num_intersections].normal, norm);
	  num_intersections++;
	}
	
	// Finished processing this step. Keep going.
      } while (keep_stepping);
      
      // Done traversing interval 1.
    }
    
    /*
     * Truncate and traverse interval 2.
     */
    int2_discarded = FALSE;

    if (interval_start > int2_end) {
      // The intersections are both before the interval defined by the
      // planes. No intersections.
      int2_discarded = TRUE;
    } else if (interval_start > int2_start) {
      // There is an ENTRY intersection with a plane in the middle of the
      // interval. Truncate the interval to this plane.
      int2_start = interval_start;
      TEST_INSIDE(int2_start, int2_start_in);

      // If this plane is a valid entry intersection (i.e. the intersection
      // with the plane is inside the bounds of the taper) then insert it
      // into the intersection list.
      if (int2_start_in) {
	if (int2_start == base_t) {
	  // Insert an entry intersection with the base of the taper.
	  int_storage[num_intersections].t = int2_start;
	  int_storage[num_intersections].is_entering = TRUE;
	  Vector::assign(int_storage[num_intersections].normal, 0, -1, 0);
	  num_intersections++;
	} else {
	  // Insert an entry intersection with the end of the taper.
	  int_storage[num_intersections].t = int2_start;
	  int_storage[num_intersections].is_entering = TRUE;
	  Vector::set(int_storage[num_intersections].normal, end_n);
	  num_intersections++;
	}
      }
    }

    if (!int2_discarded) {
      if (interval_end < int2_start) {
	// The intersections are both after the interval defined by the
	// planes. No intersections.
	return num_intersections;
      } else if (interval_end < int2_end) {
	// There is an EXIT intersection with a plane in the middle of the
	// interval. Truncate the interval to this plane.
	int2_end = interval_end;
	TEST_INSIDE(int2_end, int2_end_in);
	
	// If this plane is a valid exit intersection (i.e. the intersection
	// with the plane is inside the bounds of the taper) then insert it
	// into the intersection list.
	if (int2_end_in) {
	  if (int2_end == base_t) {
	    // Insert an exit intersection with the base of the taper.
	    int_storage[num_intersections].t = int2_end;
	    int_storage[num_intersections].is_entering = FALSE;
	    Vector::assign(int_storage[num_intersections].normal, 0, -1, 0);
	    num_intersections++;
	  } else {
	    // Insert an exit intersection with the end of the taper.
	    int_storage[num_intersections].t = int2_end;
	    int_storage[num_intersections].is_entering = FALSE;
	    Vector::set(int_storage[num_intersections].normal, end_n);
	    num_intersections++;
	  }
	}
      }
    }

    if (!int2_discarded) {
      // Traverse the remaining interval.
      step_size = (int2_end - int2_start) / traversal_num_steps;
      step_in = int2_start_in;
      step_t = int2_start;
      keep_stepping = TRUE;
      do {
	// Store the current position.
	last_step_t = step_t;
	last_step_in = step_in;
	
	// Calculate the next sample point.
	step_t += step_size;
	
	// Are we past the end?
	if (step_t > int2_end) {
	  // Use the data that we already have about the endpoint.
	  step_t = int2_end;
	  step_in = int2_end_in;
	  
	  keep_stepping = FALSE;
	} else {
	  // Calculate whether or not we are inside the taper.
	  TEST_INSIDE(step_t, step_in);
	}
	
	// Have we crossed the surface?
	if (step_in != last_step_in) {
	  // Perform a binary search to find the surface. The starting points
	  // are the last step and the current position.
	  interval_min = last_step_t;       min_in = last_step_in;
	  interval_max = step_t;            max_in = step_in;
	  
	  // Are we close enough to an answer?
	  while (interval_max - interval_min > binary_search_limit) {
	    // Calculate the midpoint and inside-test it.
	    interval_mid = (interval_max + interval_min) / 2;
	    TEST_INSIDE(interval_mid, mid_in);
	    
	    // Contract the interval being searched.
	    if (mid_in == min_in) {
	      interval_min = interval_mid;
	    } else {
	      interval_max = interval_mid;
	    }
	  }
	  
	  // We have refined the interval to the desired precision. Calculate
	  // the normal and insert the intersection.
	  interval_mid = (interval_max + interval_min) / 2;
	  
	  // We find X, Y and Z at the hitpoint.
	  hx = ray_from[X] + interval_mid * ray_dir[X];
	  hy = ray_from[Y] + interval_mid * ray_dir[Y];
	  hz = ray_from[Z] + interval_mid * ray_dir[Z];
	  
	  // Precalculate some values for calculation of the normal.
	  xy_dist = CHECK0(sqrt(SQR(hx) + SQR(hy)),
			   "xy_dist in taperPrimitive::ray_intersect");
	  radius_increase = (rad_top - rad_base) / bend_amount;
	  hit_radius = rad_base + radius_increase * atan2(hy, hx);
	  one_over_yx_sq = 1 / 
	    (1 + SQR(hy / CHECK0(hx, 
				 "hx in taperPrimitive::ray_intersect")));
	  
	  // Calculate the normal. This is the gradient of the equation that
	  // defines the taper ... 
	  //
	  // f(x, y, z) = 0
	  //
	  // f = (sqrt(x^2 + y^2) - 1)^2 + z^2 - hit_rad^2 = 0
	  //
	  // where hit_rad = r1 + (r2 - r1) * (tan-1(y/x) / bend_amount)
	  //
	  // We calculate the three partial derivatives and form the normal
	  // from them.
	  norm[X] = 2 * (xy_dist - 1) * (hx / xy_dist)
	    - 2 * hit_radius * radius_increase
	    * one_over_yx_sq * (hy * (- 1 / SQR(hx)));
	  norm[Y] = 2 * (xy_dist - 1) * (hy / xy_dist)
	    - 2 * hit_radius * radius_increase
	    * one_over_yx_sq * (1 / hx);
	  norm[Z] = 2 * hz;
	  
	  // Insert an intersection with the side of the taper.
	  // If min_in is TRUE, this is a leaving intersection.
	  int_storage[num_intersections].t = interval_mid;
	  int_storage[num_intersections].is_entering = (min_in)
	    ? FALSE : TRUE;
	  Vector::set(int_storage[num_intersections].normal, norm);
	  num_intersections++;
	}
	
	// Finished processing this step. Keep going.
      } while (keep_stepping);
      
      // Done traversing interval 2.
    }
    
  } else {
    fatal_error("taperPrimitive::ray_intersect",
		"Quartic has odd number of real solutions!",
		"Obviously the Conjugate Root Theorem was wrong after all...");
    return 0;
  }

  // Return the number of intersections.
  return num_intersections;
}

intersect_result taperPrimitive::point_intersect(primnode *node,
						 vector wp,
						 PointInfo *info,
						 uint ray_flags)
{
  // Here wp is the world-point that we are testing for IN/OUT.
  // The values p_x, p_y, p_z are its coordinates in primitive space
  // (the mapping from world to primitive space is given by position.aft)
  double p_x, p_y, p_z, xy_d;
  double theta, radius;
  matrix *m;

  // Get the matrix to transform from world to primitive space.
  m = &(node->position[AFT]);

  // Transform the worldpoint into primitive space.
  p_x = (*m)[0][0]*wp[X] + (*m)[0][1]*wp[Y] + (*m)[0][2]*wp[Z] + (*m)[0][3];
  p_y = (*m)[1][0]*wp[X] + (*m)[1][1]*wp[Y] + (*m)[1][2]*wp[Z] + (*m)[1][3];
  p_z = (*m)[2][0]*wp[X] + (*m)[2][1]*wp[Y] + (*m)[2][2]*wp[Z] + (*m)[2][3];

  if (p_y < 0)
    return OUT;

  theta = atan2(p_y, p_x);

  if (theta < 0)
    return OUT;
  else if (theta > bend_amount)
    return OUT;
  else
    radius = rad_base + (rad_top - rad_base)
      * (theta / bend_amount);

  // xy_d is the point's distance in X-Y space from the circle that defines
  // the torus.
  xy_d = sqrt(SQR(p_x) + SQR(p_y)) - 1.0;
  
  // The world_point's distance from the torus's defining circle is given
  // by Pythagoras' theorem on the X-Y distance and the Z distance (= p_z).
  if (SQR(xy_d) + SQR(p_z) <= SQR(radius)) {
    set_point_contents(node, info);
    return IN;
  } else
    return OUT;
}

intersect_result taperPrimitive::point_intersect(vector wp)
{
  double xy_d;
  double theta, radius;

  if (wp[Y] < 0)
    return OUT;

  theta = atan2(wp[Y], wp[X]);

  if (theta < 0)
    return OUT;
  else if (theta > bend_amount)
    return OUT;
  else
    radius = rad_base + (rad_top - rad_base)
      * (theta / bend_amount);

  // xy_d is the point's distance in X-Y space from the circle that defines
  // the torus.
  xy_d = sqrt(SQR(wp[X]) + SQR(wp[Y])) - 1.0;
  
  // The world_point's distance from the torus's defining circle is given
  // by Pythagoras' theorem on the X-Y distance and the Z distance (= wp[Z]).
  if (SQR(xy_d) + SQR(wp[Z]) <= SQR(radius))
    return IN;
  else
    return OUT;
}


intersect_result taperPrimitive::voxel_intersect(box b)
{
  // The method is:
  //
  // First: test boxes against the Y = 0 plane. Discard any that are OUT.
  //
  // Second: test boxes against the end plane. Discard any that are OUT.
  //
  // Third: test boxes against the bounding torus rad_base. Discard any that
  // are OUT. If SURFACE and the planes' are IN or SURFACE, return SURFACE.
  //
  // Last: test against the inner torus rad_top. If IN and the planes' are
  // IN, return IN. Otherwise, return SURFACE.

  // Note that this is not an exact voxel intersection test. Any box that
  // falls within the area bounded by the inside and outside torii is
  // returned as SURFACE even if it does not intersect the actual primitive.

  int first_point_in, this_point_in;
  int i;
  intersect_result base_plane, top_plane, planes, outside_torus, inside_torus;

  // **** FIRST STEP: classify the box with respect to the plane at Y = 0.

  // Is the first point in? Use this as the result until we find one
  // that disagrees.
  first_point_in = (b[0][Y] > 0);
  if (first_point_in)
    base_plane = IN;
  else
    base_plane = OUT;

  for (i = 1; i < 8; i++) {
    this_point_in = (b[i][Y] > 0);

    if (this_point_in != first_point_in) {
      base_plane = SURFACE;
      break;
    }
  }

  if (base_plane == OUT)
    return OUT;



  // **** SECOND STEP: classify the box with respect to the end plane,
  //                   given by end_n and passing through the origin.

  // Is the first point in? Use this as the result until we find one
  // that disagrees.
  first_point_in = (Vector::dot(b[0], end_n) < 0);
  if (first_point_in)
    top_plane = IN;
  else
    top_plane = OUT;

  for (i = 1; i < 8; i++) {
    this_point_in = (Vector::dot(b[i], end_n) < 0);

    if (this_point_in != first_point_in) {
      top_plane = SURFACE;
      break;
    }
  }

  if (top_plane == OUT)
    return OUT;

  // Combine the results from the two planes.
  if ((base_plane == SURFACE) || (top_plane == SURFACE))
    planes = SURFACE;
  else
    planes = IN;


  // **** THIRD STEP: classify the box with respect to the bounding torus
  //                  (radius rad_base).
  outside_torus = torus_voxel_intersect(rad_base, b);

  if (outside_torus == OUT)
    return OUT;
  else if (outside_torus == SURFACE)
    return SURFACE;
  else if (planes == SURFACE)
    return SURFACE;
  else {
    // Both outside_torus and planes are IN.
  
    // **** FOURTH STEP: classify the box with respect to the inner torus
    //                   (radius rad_top).
    inside_torus = torus_voxel_intersect(rad_top, b);

    if (inside_torus == IN)
      return IN;
    else
      return SURFACE;
  }
}

char *taperPrimitive::prim_name(void) {
  return "Magical Tapering Torus";
}


#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif
