/* prim_torus.c : The torus primitive
** 
** The torus primitive defines a torus generated by sweeping a circle of
** radius torus_rad about a circle of radius 1 in the XY plane.
**
** Chris Butcher, 13 December 1996
*/

#include "mirage.h"

#ifdef USE_NAMESPACES
namespace Mirage {
#endif

int torusPrimitive::ray_intersect(IntersectionList *the_list, 
				  primnode *node, primnode *validate_node, 
				  vector ray_from,
				  vector ray_dir, uint ray_flags,
				  signature ray_signature)
{
  // The coordinates of ray_from in primitive space
  double p_x, p_y, p_z;
  // The direction ray_dir in primitive space
  double d_x, d_y, d_z;
  // Intermediate quadratic expression coefficients (a b c, d e f)
  double a, b, c, d, e, f;
  // The coefficients of the quartic equation in t
  // Note that q_coeff[i] is the coefficient of t^i
  double q_coeff[5];
  // The solutions of the quartic equation.
  double q_sol[5];
  // The number of intersections.
  int q_solutions, i;
  // The primitive-space hitpoint.
  double hit_x, hit_y, hit_z;
  // The distance from the hitpoint to the torus's defining ring
  // at distance 1 from the origin.
  double hit_dist;
  // The normal at the hitpoint.
  vector norm;
  int index, smallest, found = FALSE;

  // Store the signature of this ray that we're intersecting with.
  node->last_ray_signature = ray_signature;

  // This apply is an inline function defined in matrix.h
  // Transform the ray startpoint into primitive space
  Matrix::apply(node->position[AFT], ray_from, &p_x, &p_y, &p_z);

  // This apply is also an inline function defined in matrix.h
  // Transform the ray direction into primitive space (note: shift does
  // not apply to this, only apply rotation + stretch sections of the matrix)
  Matrix::applynoshift(node->position[AFT], ray_dir, &d_x, &d_y, &d_z);


  // The equation that defines the torus is
  // (xy_distance)^2 + (z_distance)^2 = (torus_rad)^2.
  // Since the point is given by (p + td), this gives
  //
  // (sqrt((px + tdx)^2 + (py + tdy)^2) - 1.0)^2 + (pz + tdz)^2 = torus_rad^2
  //
  // We expand the xy_distance squared and collect the sqrt terms on one
  // side to give
  //
  // (px + tdx)^2 + (py + tdy)^2 + (pz + tdz)^2 + 1 - torus_rad^2 =
  //                                    2 sqrt((px + tdx)^2 + (py + tdy)^2)
  //
  // Expanding all the terms (p + td) gives us the equation
  //
  // (dx^2 + dy^2 + dz^2)t^2 + 2(pxdx + pydy + pzdz)t +
  //    (px^2 + py^2 + pz^2 + 1 - torus_rad^2) = 2 sqrt( (dx^2 + dy^2)t^2 +
  //                                        2(pxdx + pydy)t + (px^2 + py^2) )
  //
  // This is an equation of the type
  //   at^2 + bt + c = 2 sqrt(dt^2 + et + f)
  //
  // where a = dx^2 + dy^2 + dz^2
  //       b = 2(pxdx + pydy + pzdz)
  //       c = px^2 + py^2 + pz^2 + 1 - torus_rad^2
  //       d = dx^2 + dy^2
  //       e = 2(pxdx + pydy)
  //       f = px^2 + py^2
  //
  // We square both sides to give
  //
  // (a^2)t^4 + (2ab)t^3 + (2ac + b^2 - 4d)t^2 + (2bc - 4e)t + (c^2 - 4f) = 0
  //
  // which can then be solved using the SolveQuartic function.

  // Calculate the coefficients for the quadratic expressions (a b c, d e f)
#if 0
  // Slow version, not used (included for clarity)
  a = SQR(d_x) + SQR(d_y) + SQR(d_z);
  b = 2.0 * (p_x*d_x + p_y*d_y + p_z*d_z);
  c = SQR(p_x) + SQR(p_y) + SQR(p_z) + 1 - SQR(torus_rad);
  d = SQR(d_x) + SQR(d_y);
  e = 2.0 * (p_x*d_x + p_y*d_y);
  f = SQR(p_x) + SQR(p_y);
#else
  // Note that this is the same as above, just optimized slightly.
  // Also note that since b and e will be multiplied by two in the
  // calculation of the quartic coefficients, what we calculate here is
  // really 1/2 b and 1/2 e to save a multiplication.
  d = SQR(d_x) + SQR(d_y);
  e = p_x*d_x + p_y*d_y;
  f = SQR(p_x) + SQR(p_y);

  a = d + SQR(d_z);
  b = e + p_z*d_z;
  c = f + SQR(p_z) + 1 - SQR(torus_rad);
#endif

  // Calculate the coefficients of the quartic equation to be solved.
  q_coeff[4] = SQR(a);
  q_coeff[3] = 4*a*b;                      // 2ab = 4a(1/2b)
  q_coeff[2] = 2*a*c + 4*SQR(b) - 4*d;     // b^2 = 4*(1/2b)^2
  q_coeff[1] = 4*b*c - 8*e;                // 2bc = 4c(1/2b) and 4e = 8(1/2e)
  q_coeff[0] = SQR(c) - 4*f;

  // Solve the quartic equation.
  q_solutions = solveQuartic(q_coeff, q_sol);

  // Insert the intersections that are valid into the intersection list.
  for (i = 0; i < q_solutions; i++) {
    // Find the intersection with the smallest t-value.
    smallest = 0;
    for (index = 1; index < q_solutions; index++)
      if (q_sol[index] < q_sol[smallest])
	smallest = index;

    if (q_sol[smallest] > MIN_TVAL_CUTOFF * 10) {
      hit_x = p_x + d_x*q_sol[smallest];
      hit_y = p_y + d_y*q_sol[smallest];
      hit_z = p_z + d_z*q_sol[smallest];

      // The normal has an XY component that is proportional to
      // 1 - (1 / dist) where dist is the XY distance of the hitpoint
      // from the origin.
      // Then the Z component is proportional to the hitpoint's
      // distance from the Z = 0 plane. This will give a normal
      // that has length equal to the torus_rad value (not a unit vector).
      // This is not a problem as the normal will get transformed anyway.
      hit_dist = 1.0 - (1.0 / CHECK0(sqrt(SQR(hit_x) + SQR(hit_y)),
		      "torusPrimitive::ray_intersect: zero length normal"));

      norm[X] = hit_x * hit_dist;
      norm[Y] = hit_y * hit_dist;
      norm[Z] = hit_z;

      // This is an entering intersection if it's the first or third
      // (i = 0 or 2).
      if ((i % 2) == 0) {
	// Insert an ENTERING intersection into the intersection list.
	insert_entering(the_list, node, validate_node, norm, q_sol[smallest],
			ray_from, ray_dir, ray_flags);
      } else {
	// Insert a LEAVING intersection into the intersection list.
	insert_leaving(the_list, node, validate_node, norm, q_sol[smallest],
			ray_from, ray_dir, ray_flags);
      }

      found = TRUE;
    }

    // This solution has been processed, don't find it again.
    q_sol[smallest] = +INFINITY;
  }

  node->last_ray_hit = found;

  return node->last_ray_hit;
}

int torusPrimitive::ray_intersect(vector ray_from,
				  vector ray_dir, rayintersect *int_storage)
{
  // Intermediate quadratic expression coefficients (a b c, d e f)
  double a, b, c, d, e, f;
  // The coefficients of the quartic equation in t
  // Note that q_coeff[i] is the coefficient of t^i
  double q_coeff[5];
  // The solutions of the quartic equation.
  double q_sol[5];
  // The number of intersections.
  int q_solutions, i;
  // The primitive-space hitpoint.
  double hit_x, hit_y, hit_z;
  // The distance from the hitpoint to the torus's defining ring
  // at distance 1 from the origin.
  double hit_dist;
  // The normal at the hitpoint.
  vector norm;
  int index, smallest, num_intersections = 0;

  // The equation that defines the torus is
  // (xy_distance)^2 + (z_distance)^2 = (torus_rad)^2.
  // Since the point is given by (p + td), this gives
  //
  // (sqrt((px + tdx)^2 + (py + tdy)^2) - 1.0)^2 + (pz + tdz)^2 = torus_rad^2
  //
  // We expand the xy_distance squared and collect the sqrt terms on one
  // side to give
  //
  // (px + tdx)^2 + (py + tdy)^2 + (pz + tdz)^2 + 1 - torus_rad^2 =
  //                                    2 sqrt((px + tdx)^2 + (py + tdy)^2)
  //
  // Expanding all the terms (p + td) gives us the equation
  //
  // (dx^2 + dy^2 + dz^2)t^2 + 2(pxdx + pydy + pzdz)t +
  //    (px^2 + py^2 + pz^2 + 1 - torus_rad^2) = 2 sqrt( (dx^2 + dy^2)t^2 +
  //                                        2(pxdx + pydy)t + (px^2 + py^2) )
  //
  // This is an equation of the type
  //   at^2 + bt + c = 2 sqrt(dt^2 + et + f)
  //
  // where a = dx^2 + dy^2 + dz^2
  //       b = 2(pxdx + pydy + pzdz)
  //       c = px^2 + py^2 + pz^2 + 1 - torus_rad^2
  //       d = dx^2 + dy^2
  //       e = 2(pxdx + pydy)
  //       f = px^2 + py^2
  //
  // We square both sides to give
  //
  // (a^2)t^4 + (2ab)t^3 + (2ac + b^2 - 4d)t^2 + (2bc - 4e)t + (c^2 - 4f) = 0
  //
  // which can then be solved using the SolveQuartic function.

  // Calculate the coefficients for the quadratic expressions (a b c, d e f)
#if 0
  // Slow version, not used (included for clarity)
  a = SQR(ray_dir[X]) + SQR(ray_dir[Y]) + SQR(ray_dir[Z]);
  b = 2.0 * (ray_from[X]*ray_dir[X] + ray_from[Y]*ray_dir[Y]
	     + ray_from[Z]*ray_dir[Z]);
  c = SQR(ray_from[X]) + SQR(ray_from[Y]) + SQR(ray_from[Z])
    + 1 - SQR(torus_rad);
  d = SQR(ray_dir[X]) + SQR(ray_dir[Y]);
  e = 2.0 * (ray_from[X]*ray_dir[X] + ray_from[Y]*ray_dir[Y]);
  f = SQR(ray_from[X]) + SQR(ray_from[Y]);
#else
  // Note that this is the same as above, just optimized slightly.
  // Also note that since b and e will be multiplied by two in the
  // calculation of the quartic coefficients, what we calculate here is
  // really 1/2 b and 1/2 e to save a multiplication.
  d = SQR(ray_dir[X]) + SQR(ray_dir[Y]);
  e = ray_from[X]*ray_dir[X] + ray_from[Y]*ray_dir[Y];
  f = SQR(ray_from[X]) + SQR(ray_from[Y]);

  a = d + SQR(ray_dir[Z]);
  b = e + ray_from[Z]*ray_dir[Z];
  c = f + SQR(ray_from[Z]) + 1 - SQR(torus_rad);
#endif

  // Calculate the coefficients of the quartic equation to be solved.
  q_coeff[4] = SQR(a);
  q_coeff[3] = 4*a*b;                      // 2ab = 4a(1/2b)
  q_coeff[2] = 2*a*c + 4*SQR(b) - 4*d;     // b^2 = 4*(1/2b)^2
  q_coeff[1] = 4*b*c - 8*e;                // 2bc = 4c(1/2b) and 4e = 8(1/2e)
  q_coeff[0] = SQR(c) - 4*f;

  // Solve the quartic equation.
  q_solutions = solveQuartic(q_coeff, q_sol);

  // Insert the intersections that are valid into the intersection list.
  for (i = 0; i < q_solutions; i++) {
    // Find the intersection with the smallest t-value.
    smallest = 0;
    for (index = 1; index < q_solutions; index++)
      if (q_sol[index] < q_sol[smallest])
	smallest = index;

    hit_x = ray_from[X] + ray_dir[X]*q_sol[smallest];
    hit_y = ray_from[Y] + ray_dir[Y]*q_sol[smallest];
    hit_z = ray_from[Z] + ray_dir[Z]*q_sol[smallest];
    
    // The normal has an XY component that is proportional to
    // 1 - (1 / dist) where dist is the XY distance of the hitpoint
    // from the origin.
    // Then the Z component is proportional to the hitpoint's
    // distance from the Z = 0 plane. This will give a normal
    // that has length equal to the torus_rad value (not a unit vector).
    // This is not a problem as the normal will get transformed anyway.
    hit_dist = 1.0 - (1.0 / CHECK0(sqrt(SQR(hit_x) + SQR(hit_y)),
		"torusPrimitive::ray_intersect: zero length normal"));
    
    norm[X] = hit_x * hit_dist;
    norm[Y] = hit_y * hit_dist;
    norm[Z] = hit_z;
    
    // This is an entering intersection if it's the first or third
    // (i = 0 or 2).
    int_storage[num_intersections].t = q_sol[smallest];
    int_storage[num_intersections].is_entering = ((i % 2) == 0)
      ? TRUE : FALSE;
    Vector::set(int_storage[num_intersections].normal, norm);
    num_intersections++;

    // This solution has been processed, don't find it again.
    q_sol[smallest] = +INFINITY;
  }

  return num_intersections;
}

intersect_result torusPrimitive::point_intersect(primnode *node,
						 vector wp,
						 PointInfo *info,
						 uint ray_flags)
{
  // Here wp is the world-point that we are testing for IN/OUT.
  // The values p_x, p_y, p_z are its coordinates in primitive space
  // (the mapping from world to primitive space is given by position.aft)
  double p_x, p_y, p_z, xy_d;
  matrix *m;

  // Get the matrix to transform from world to primitive space.
  m = &(node->position[AFT]);

  // Transform the worldpoint into primitive space.
  p_x = (*m)[0][0]*wp[X] + (*m)[0][1]*wp[Y] + (*m)[0][2]*wp[Z] + (*m)[0][3];
  p_y = (*m)[1][0]*wp[X] + (*m)[1][1]*wp[Y] + (*m)[1][2]*wp[Z] + (*m)[1][3];
  p_z = (*m)[2][0]*wp[X] + (*m)[2][1]*wp[Y] + (*m)[2][2]*wp[Z] + (*m)[2][3];

  // xy_d is the point's distance in X-Y space from the circle that defines
  // the torus.
  xy_d = sqrt(SQR(p_x) + SQR(p_y)) - 1.0;
  
  // The world_point's distance from the torus's defining circle is given
  // by Pythagoras' theorem on the X-Y distance and the Z distance (= p_z).
  if (SQR(xy_d) + SQR(p_z) <= SQR(torus_rad)) {
    set_point_contents(node, info);
    return IN;
  } else
    return OUT;
}

intersect_result torusPrimitive::point_intersect(vector wp)
{
  double xy_d;

  // xy_d is the point's distance in X-Y space from the circle that defines
  // the torus.
  xy_d = sqrt(SQR(wp[X]) + SQR(wp[Y])) - 1.0;
  
  // The world_point's distance from the torus's defining circle is given
  // by Pythagoras' theorem on the X-Y distance and the Z distance (= p_z).
  if (SQR(xy_d) + SQR(wp[Z]) <= SQR(torus_rad))
    return IN;
  else
    return OUT;
}



intersect_result torusPrimitive::voxel_intersect(box b)
{
  return torus_voxel_intersect(torus_rad, b);
}


// Now define some local macros to make the code length shorter for the
// routine torus_voxel_intersect.


// Macro CHECK_OUTSIDE_PLUS_PLANE tests to see if the points of the box all
// lie outside the plane P at distance x.
#define CHECK_OUTSIDE_PLUS_PLANE(axis, x) {   \
                                              \
  for (i = 0; i < 8; i++) {                   \
    if (b[i][axis] < x) {                     \
      break;                                  \
    }                                         \
  }                                           \
  if (i == 8)                                 \
    return OUT;                               \
}


// Macro CHECK_OUTSIDE_MINUS_PLANE tests to see if the points of the box all
// lie outside the plane P at distance x.
#define CHECK_OUTSIDE_MINUS_PLANE(axis, x) {   \
                                               \
  for (i = 0; i < 8; i++)                      \
    if (b[i][axis] > -x)                       \
      break;                                   \
                                               \
  if (i == 8)                                  \
    return OUT;                                \
}

// Macro AXIS_SIDE calculates the side on which the Z axis falls of the
// edge defined by the points of the box i and j. The value is stored
// in axis_side[store] for later reference.
#define AXIS_SIDE(i, j, store)                                    \
                                                                  \
(axis_side[store] = ((b[i][Y]*b[j][X] - b[i][X]*b[j][Y]) >= 0.0))

// Macro CHECK_EDGE_INTERSECT checks the edge between points i and j
// to see if it intersects the torus. If it does, return SURFACE.
// Note that this is slow because it solves a quartic equation (!).
// Most of this code is taken from the ray_intersect routine above.
#define CHECK_EDGE_INTERSECT(i, j) {                                    \
  p_x = b[i][X];         p_y = b[i][Y];         p_z = b[i][Z];          \
  d_x = b[j][X] - p_x;   d_y = b[j][Y] - p_y;   d_z = b[j][Z] - p_z;    \
                                                                        \
  quad_d = SQR(d_x) + SQR(d_y);                                         \
  quad_e = p_x*d_x + p_y*d_y;                                           \
  quad_f = SQR(p_x) + SQR(p_y);                                         \
                                                                        \
  quad_a = quad_d + SQR(d_z);                                           \
  quad_b = quad_e + p_z*d_z;                                            \
  quad_c = quad_f + SQR(p_z) + 1 - SQR(torus_rad);                      \
                                                                        \
  q_coeff[4] = SQR(quad_a);                                             \
  q_coeff[3] = 4*quad_a*quad_b;                                         \
  q_coeff[2] = 2*quad_a*quad_c + 4*SQR(quad_b) - 4*quad_d;              \
  q_coeff[1] = 4*quad_b*quad_c - 8*quad_e;                              \
  q_coeff[0] = SQR(quad_c) - 4*quad_f;                                  \
                                                                        \
  q_solutions = solveQuartic(q_coeff, q_sol);                           \
                                                                        \
  for (sol_num = 0; sol_num < q_solutions; sol_num++) {                 \
    if (q_sol[sol_num] >= 0 && q_sol[sol_num] <= 1)                     \
      return SURFACE;                                                   \
  }                                                                     \
}

// Macro CHECK_FACE_INTERSECT checks the face defined by points i, j, k and l
// to see if the torus's defining circle intersects it. It does so by looking
// at the pairs (i -> j), (j -> k), (k -> l), (l -> i) to see if they cross
// the Z = 0 plane. If so, the crossing point is found and stored. Either 0
// or 2 crossing points will be found ... if 2 points, then check them in
// X and Y to see if the line defined by them intersects the torus's defining
// circle.
//
// This check is done by forming the quadratic equation
// (p + td)^2 = (1 + torus_rad)^2
//  i.e.
//       (dx^2 + dy^2)t^2 + (2pxdx + 2pydy)t + (px^2 + py^2 - (1 + r)^2) = 0
//
// and then if the discriminant b^2 - 4ac > 0, there is an intersection and
// so we return SURFACE.
#define CHECK_FACE_INTERSECT(i, j, k, l) {                          \
  num_crossings = 0;                                                \
                                                                    \
  FIND_CROSSING(i, j, 0);                                           \
  FIND_CROSSING(j, k, 0);                                           \
  FIND_CROSSING(k, l, 0);                                           \
  FIND_CROSSING(l, i, 0);                                           \
                                                                    \
  if (num_crossings == 2) {                                         \
    p_x = crossing[0][X];          p_y = crossing[0][Y];            \
    d_x = crossing[1][X] - p_x;    d_y = crossing[1][Y] - p_y;      \
                                                                    \
    quad_a = SQR(d_x) + SQR(d_y);                                   \
    quad_b = 2 * (p_x * d_x + p_y * d_y);                           \
    quad_c = SQR(p_x) + SQR(p_y) - SQR(1 + torus_rad);              \
                                                                    \
    if (SQR(quad_b) - 4 * quad_a * quad_c >= 0)                     \
      return SURFACE;                                               \
  }                                                                 \
}

// Macro CHECK_CIRCLE_INTERSECT checks the face defines by points i, j, k and l
// to see if the circle radius 1, Z = z intersects it. It does so by looking
// at the pairs (i -> j), (j -> k), (k -> l), (l -> i) to see if they cross
// the Z = z plane. If so, the crossing point is found and stored. Either 0
// or 2 crossing points will be found ... if 2 points, then check them in
// X and Y to see if the line defined by them intersects the circle with
// radius 1.
#define CHECK_CIRCLE_INTERSECT(i, j, k, l, z) {                     \
  num_crossings = 0;                                                \
                                                                    \
  FIND_CROSSING(i, j, z);                                           \
  FIND_CROSSING(j, k, z);                                           \
  FIND_CROSSING(k, l, z);                                           \
  FIND_CROSSING(l, i, z);                                           \
                                                                    \
  if (num_crossings == 2) {                                         \
    p_x = crossing[0][X];          p_y = crossing[0][Y];            \
    d_x = crossing[1][X] - p_x;    d_y = crossing[1][Y] - p_y;      \
                                                                    \
    quad_a = SQR(d_x) + SQR(d_y);                                   \
    quad_b = 2 * (p_x * d_x + p_y * d_y);                           \
    quad_c = SQR(p_x) + SQR(p_y) - 1;                               \
                                                                    \
    if (SQR(quad_b) - 4 * quad_a * quad_c >= 0)                     \
      return SURFACE;                                               \
  }                                                                 \
}

// Macro CHECK_CIRCLE_INSIDE checks the circle of radius 1, at Z = z,
// to see if it lies entirely inside the box (we already know it doesn't
// intersect any of the faces). It does so by checking the point (1, 0, z)
// to see if it lies inside the box ... we subtract this vector from the box
// and then check to see if the origin lies inside the box.
#define CHECK_CIRCLE_INSIDE(z) {                                        \
  Vector::assign(samplevec, 1, 0, z);                                   \
  for (i = 0; i < 8; i++)                                               \
    Vector::sub(b[i], samplevec, newbox[i]);                            \
                                                                        \
  DISTANCE_FROM_FACE(newbox, height[0], 0, 1, 2);                       \
  DISTANCE_FROM_FACE(newbox, height[1], 0, 4, 1);                       \
  DISTANCE_FROM_FACE(newbox, height[2], 0, 2, 4);                       \
  DISTANCE_FROM_FACE(newbox, height[3], 7, 5, 6);                       \
  DISTANCE_FROM_FACE(newbox, height[4], 7, 6, 3);                       \
  DISTANCE_FROM_FACE(newbox, height[5], 7, 3, 5);                       \
  if ((height[0] < 0) && (height[1] < 0) && (height[2] < 0) &&          \
      (height[3] < 0) && (height[4] < 0) && (height[5] < 0))            \
    return SURFACE;                                                     \
}

// Macro DISTANCE_FROM_FACE calculates the distance of the origin from
// a face of the box bb (defined by the three points with indices i, j, k).
#define DISTANCE_FROM_FACE(bb, h, i, j, k)                  \
                                                            \
h = (bb[i][X] * (bb[j][Y]*bb[k][Z] - bb[j][Z]*bb[k][Y])     \
   - bb[i][Y] * (bb[j][X]*bb[k][Z] - bb[j][Z]*bb[k][X])     \
   + bb[i][Z] * (bb[j][X]*bb[k][Y] - bb[j][Y]*bb[k][X]));


// Macro FIND_CROSSING checks the points i and j to see if they lie on
// opposite sides of the Z = z plane. If so, then the point where their
// defining line crosses this plane is stored in crossing[0] or crossing[1].
#define FIND_CROSSING(i, j, z) {                                        \
  if ((b[i][Z] - z) * (b[j][Z] - z) < 0) {                              \
    tval = (z - b[i][Z]) / (b[j][Z] - b[i][Z]);                           \
    crossing[num_crossings][X] = b[i][X] + tval * (b[j][X] - b[i][X]);  \
    crossing[num_crossings][Y] = b[i][Y] + tval * (b[j][Y] - b[i][Y]);  \
                                                                        \
    num_crossings++;                                                    \
  }                                                                     \
}

// Macro TEST_AXISPOINT checks the face defined by points i, j, k and l
// and gets the point (0, 0, z) on the Z axis that this face contains.
// If this point is outside the torus, then return SURFACE.
#define TEST_AXISPOINT(i, j, k, l) {                  \
  tval = -b[i][X] / (b[j][X] - b[i][X]);              \
  x1[X] = 0.0;                                        \
  x1[Y] = b[i][Y] + tval * (b[j][Y] - b[i][Y]);       \
  x1[Z] = b[i][Z] + tval * (b[j][Z] - b[i][Z]);       \
                                                      \
  tval = -b[k][X] / (b[l][X] - b[k][X]);              \
  x2[X] = 0.0;                                        \
  x2[Y] = b[k][Y] + tval * (b[l][Y] - b[k][Y]);       \
  x2[Z] = b[k][Z] + tval * (b[l][Z] - b[k][Z]);       \
                                                      \
  tval = -x1[Y] / (x2[Y] - x1[Y]);                    \
  axispt[X] = 0.0;                                    \
  axispt[Y] = 0.0;                                    \
  axispt[Z] = x1[Z] + tval * (x2[Z] - x1[Z]);         \
                                                      \
  if (fabs(axispt[Z]) >= torus_cusp)                  \
    return SURFACE;                                   \
}

intersect_result torus_voxel_intersect(double torus_rad, box b)
{
  // This is one of the most complex voxel intersection tests in the system.
  // Not for the faint-hearted.
  //
  // This is an exact intersection test.
  //
  // The method is:
  //
  // First: drop all boxes that lie outside X = +/- (1 + r)
  // or Y = +/- (1 + r) or Z = +/- r. Also drop any boxes that lie
  // entirely within the cylinder of radius (1 - r). These are all OUT.
  // 
  // Then get the number of corners of the voxel that lie inside the torus.
  // If 0, then the box may be OUT. If 8, then the box may be IN. If 1-7,
  // then the box is definitely SURFACE, return this.
  //
  // If the box may be IN, check to see that it doesn't contain the Z axis,
  // which is empty. If it does contain the Z axis, then check the points
  // where the Z axis enters and exits the voxel. If either of them are
  // OUT then return SURFACE.
  //   Otherwise, check all the edges of the voxel... if they intersect the
  // surface of the torus, return SURFACE.
  //   So: we know all the points are in, the Z axis doesn't intersect the
  // voxel (or if it does, it is IN for its entire length in the voxel), and
  // none of the voxel edges intersect the torus's surface. We may return IN.
  //
  // If the box may be OUT, check to see that the outside equator of the
  // torus doesn't intersect the voxel (a definite SURFACE). Next check that
  // the top and bottom circles of the torus don't intersect the faces of the
  // voxel ... it they don't, check they aren't completely contained. Finally,
  // check all the edges of the voxel for intersections with the torus's
  // surface. If there are none, we may return OUT safely.

  double out_radius, in_rad_sq, xy_d, torus_rad_sq;
  double q_coeff[5], q_sol[5], p_x, p_y, p_z, d_x, d_y, d_z;
  double quad_a, quad_b, quad_c, quad_d, quad_e, quad_f;
  vector crossing[4];
  double tval, torus_cusp;
  int sol_num, q_solutions, num_crossings;
  int i, first_point_in, this_point_in;
  double axis_side[12];
  vector x1, x2, axispt;
  double height[6];
  box newbox;
  vector samplevec;

  // **** FIRST STEP: discard any boxes lying completely outside the
  //                  bounding box around the torus in X, Y and Z.

  // Check to see if it lies outside the X, Y or Z planes.
  out_radius = 1 + torus_rad;
  CHECK_OUTSIDE_PLUS_PLANE(X, out_radius);
  CHECK_OUTSIDE_MINUS_PLANE(X, out_radius);
  CHECK_OUTSIDE_PLUS_PLANE(Y, out_radius);
  CHECK_OUTSIDE_MINUS_PLANE(Y, out_radius);
  CHECK_OUTSIDE_PLUS_PLANE(Z, torus_rad);
  CHECK_OUTSIDE_MINUS_PLANE(Z, torus_rad);

  // **** SECOND STEP: discard any boxes lying entirely within the
  //                   cylinder containing the hole in the torus. Only do
  //                   this if the torus's radius is < 1, so there is a
  //                   hole in the middle.

  if (torus_rad < 1) {
    // If all the points lie inside in_radius, then return OUT.
    in_rad_sq = SQR(1 - torus_rad);
    for (i = 0; i < 8; i++) {
      if (SQR(b[i][X]) + SQR(b[i][Y]) > in_rad_sq)
	break;
    }
    if (i == 8)                                  
      return OUT;                                
  }

  // **** THIRD STEP: any boxes with points both inside and outside the
  //                  torus are SURFACE.

  torus_rad_sq = SQR(torus_rad);

  // Work out how far the first point is in X-Y space from the defining circle.
  xy_d = sqrt(SQR(b[0][X]) + SQR(b[0][Y])) - 1.0;

  // Work out if the first point is in.
  first_point_in = ((SQR(xy_d) + SQR(b[0][Z])) < torus_rad_sq);

  for(i = 1; i < 8; i++) {
    // Work out if the ith point is in.
    xy_d = sqrt(SQR(b[i][X]) + SQR(b[i][Y])) - 1.0;

    this_point_in = ((SQR(xy_d) + SQR(b[i][Z])) < torus_rad_sq);

    // If the points are on opposite sides of the surface, return SURFACE.
    if (this_point_in != first_point_in)
      return SURFACE;
  }

  CHECK_VOXEL_HANDEDNESS(b);

  // All the points are either IN or OUT.
  if (first_point_in) {
    // **** FOURTH STEP: if the box contains the Z axis, then
    //                   the box must contain at least one OUT point and so
    //                   is SURFACE. We check all six faces to see if the Z
    //                   axis passes through the face. If it does, we check
    //                   the point of entry. If it is OUT, then we return
    //                   SURFACE.
    //                    If all six faces don't return SURFACE, then the
    //                   Z axis either doesn't intersect the voxel or it
    //                   is IN for the entire length that it is in the
    //                   voxel.

    if (torus_rad <= 1)
      torus_cusp = 0;
    else
      torus_cusp = sqrt(SQR(torus_rad) - 1);
    
    // Code taken from prim_cylinder.c ... tested and valid.
    // We test each face by returning SURFACE if the centerline is between
    // the N and S edges and also the E and W edges of the face in question.
    // We store the values of the edge tests in axis_side[] for reuse.
    
    AXIS_SIDE(0, 1, 0);
    AXIS_SIDE(3, 2, 1);
    AXIS_SIDE(0, 2, 8);
    AXIS_SIDE(3, 1, 9);
    
    if ((axis_side[0] == axis_side[1]) &&
	(axis_side[8] == axis_side[9])) {
      TEST_AXISPOINT(0, 1, 2, 3);
    }
    
    AXIS_SIDE(5, 4, 3);
    AXIS_SIDE(6, 7, 2);
    AXIS_SIDE(6, 4, 10);
    AXIS_SIDE(5, 7, 11);
    if ((axis_side[3]  == axis_side[2]) &&
	(axis_side[10] == axis_side[11])) {
      TEST_AXISPOINT(4, 5, 6, 7);
    }
    
    AXIS_SIDE(4, 0, 4);
    AXIS_SIDE(1, 5, 5);
    if ((axis_side[4] == axis_side[5]) &&
	(axis_side[0] == axis_side[3])) {
      TEST_AXISPOINT(0, 1, 4, 5);
    }
    
    AXIS_SIDE(2, 6, 6);
    AXIS_SIDE(7, 3, 7);
    if ((axis_side[6] == axis_side[7]) &&
	(axis_side[1] == axis_side[2])) {
      TEST_AXISPOINT(2, 3, 6, 7);
    }
    
    if ((axis_side[11] == axis_side[9]) &&
	(axis_side[5 ] == axis_side[7])) {
      TEST_AXISPOINT(1, 3, 5, 7);
    }

    if ((axis_side[8 ] == axis_side[10]) &&
	(axis_side[4 ] == axis_side[6 ])) {
      TEST_AXISPOINT(0, 2, 4, 6);
    }

    // **** FIFTH STEP: all the points are IN and the Z axis (with the hole)
    //                  does not intersect the box, so if the edges are
    //                  all completely IN then the box is also. Check to see
    //                  if any of the edges intersect the surface of the
    //                  torus. If they do, return SURFACE.
    
    CHECK_EDGE_INTERSECT(0, 1);
    CHECK_EDGE_INTERSECT(3, 2);
    CHECK_EDGE_INTERSECT(6, 7);
    CHECK_EDGE_INTERSECT(5, 4);
    CHECK_EDGE_INTERSECT(4, 0);
    CHECK_EDGE_INTERSECT(1, 5);
    CHECK_EDGE_INTERSECT(2, 6);
    CHECK_EDGE_INTERSECT(7, 3);
    CHECK_EDGE_INTERSECT(0, 2);
    CHECK_EDGE_INTERSECT(3, 1);
    CHECK_EDGE_INTERSECT(4, 6);
    CHECK_EDGE_INTERSECT(7, 5);

    // **** SIXTH STEP: the box's points and edges are all completely IN
    //                  and the box does not contain the hole, so the box
    //                  is IN.
    return IN;

  } else {

    // **** FOURTH STEP: the torus might go through a face and not intersect
    //                   any edges. It can only do this if the outside circle
    //                   of the torus goes through the box. If so, the box
    //                   is SURFACE.

    CHECK_FACE_INTERSECT(0, 1, 3, 2);
    CHECK_FACE_INTERSECT(1, 5, 7, 3);
    CHECK_FACE_INTERSECT(7, 5, 4, 6);
    CHECK_FACE_INTERSECT(0, 2, 6, 4);
    CHECK_FACE_INTERSECT(0, 4, 5, 1);
    CHECK_FACE_INTERSECT(2, 3, 7, 6);

    // **** FIFTH STEP: check to see if any of the faces intersect the circles
    //                  of radius 1 and at Z = +/- torus_rad. If they don't,
    //                  then check to see if the circle lies inside the box.
    
    CHECK_CIRCLE_INTERSECT(0, 1, 3, 2, +torus_rad);
    CHECK_CIRCLE_INTERSECT(1, 5, 7, 3, +torus_rad);
    CHECK_CIRCLE_INTERSECT(7, 5, 4, 6, +torus_rad);
    CHECK_CIRCLE_INTERSECT(0, 2, 6, 4, +torus_rad);
    CHECK_CIRCLE_INTERSECT(0, 4, 5, 1, +torus_rad);
    CHECK_CIRCLE_INTERSECT(2, 3, 7, 6, +torus_rad);
    CHECK_CIRCLE_INSIDE(+torus_rad);

    CHECK_CIRCLE_INTERSECT(0, 1, 3, 2, -torus_rad);
    CHECK_CIRCLE_INTERSECT(1, 5, 7, 3, -torus_rad);
    CHECK_CIRCLE_INTERSECT(7, 5, 4, 6, -torus_rad);
    CHECK_CIRCLE_INTERSECT(0, 2, 6, 4, -torus_rad);
    CHECK_CIRCLE_INTERSECT(0, 4, 5, 1, -torus_rad);
    CHECK_CIRCLE_INTERSECT(2, 3, 7, 6, -torus_rad);
    CHECK_CIRCLE_INSIDE(-torus_rad);
    
    // **** SIXTH STEP: all the points are OUT and the tube's center circle
    //                  does not intersect the box, so if the edges are
    //                  all completely OUT then the box is also. Check to see
    //                  if any of the edges intersect the surface of the
    //                  torus. If they do, return SURFACE.
    
    CHECK_EDGE_INTERSECT(0, 1);
    CHECK_EDGE_INTERSECT(3, 2);
    CHECK_EDGE_INTERSECT(6, 7);
    CHECK_EDGE_INTERSECT(5, 4);
    CHECK_EDGE_INTERSECT(4, 0);
    CHECK_EDGE_INTERSECT(1, 5);
    CHECK_EDGE_INTERSECT(2, 6);
    CHECK_EDGE_INTERSECT(7, 3);
    CHECK_EDGE_INTERSECT(0, 2);
    CHECK_EDGE_INTERSECT(3, 1);
    CHECK_EDGE_INTERSECT(4, 6);
    CHECK_EDGE_INTERSECT(7, 5);

    // **** SEVENTH STEP: the box's points and edges are all completely OUT
    //                    and the box does not contain the torus's defining
    //                    circle or the two circles on the top and the bottom,
    //                    so the box is OUT.
    return OUT;
  }
}

// Undefine the macros defined above (local only).
#undef CHECK_OUTSIDE_PLUS_PLANE
#undef CHECK_OUTSIDE_MINUS_PLANE
#undef CHECK_EDGE_INTERSECT
#undef CHECK_FACE_INTERSECT
#undef FIND_CROSSING
#undef AXIS_SIDE
#undef TEST_AXISPOINT
#undef CHECK_CIRCLE_INSIDE
#undef DISTANCE_FROM_FACE
#undef CHECK_CIRCLE_INTERSECT

char *torusPrimitive::prim_name(void) {
  return "Torus";
}


#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif
