/* prim_trimesh.c : The triangle mesh primitive
** 
** A triangle mesh primitive is a set of triangles. May be a shell or solid.
**
** Chris Butcher, 7 April 1998
*/

#include "mirage.h"

#define TRIMESH_MIN_PAD_SIZE     0.00010
#define TRIMESH_MAX_PAD_SIZE     0.00011

#ifdef USE_NAMESPACES
namespace Mirage {
#endif

// Debugging constructor.
triangleMeshPrimitive::triangleMeshPrimitive(void)
{
  printf("TMP: Allocate debugging mesh!\n");
}

// Constructor.
triangleMeshPrimitive::triangleMeshPrimitive(int vnum, vector *vert,
					     int trinum, int *triangles,
					     int subdivision,
					     int do_phong_shading,
					     int stop_at_first,
					     int tris_in_voxel)
{
  int i;
  meshtri *t;
  vector edge1, edge2;

  is_cluster = 0;
  num_tri = trinum;
  num_vertices = vnum;

  //  printf("TMP: Allocate and copy data\n");

  // Allocate vertex data.
  vertex = (trivertex *) calloc(num_vertices, sizeof(trivertex));
  if (vertex == NULL) {
    char errbuf[MIRAGE_STRING_LENGTH];

    sprintf(errbuf, "Can't allocate memory for %d vertices!\n", num_vertices);
    fatal_error("triangleMeshPrimitive::constructor", errbuf,
		"Out of memory.");
  }
  for (i = 0; i < num_vertices; i++) {
    Vector::set(vertex[i].vertex, vert[i]);
    Vector::set(vertex[i].normal, zero_vector);
  }

  // Allocate triangle data.
  tri = (meshtri *) calloc(num_tri, sizeof(meshtri));
  if (tri == NULL) {
    char errbuf[MIRAGE_STRING_LENGTH];

    sprintf(errbuf, "Can't allocate memory for %d triangles!\n", num_tri);
    fatal_error("triangleMeshPrimitive::constructor", errbuf,
		"Out of memory.");
  }
  for (i = 0; i < num_tri; i++) {
    tri[i].v0 = triangles[3*i + 0];
    tri[i].v1 = triangles[3*i + 1];
    tri[i].v2 = triangles[3*i + 2];
  }

  // Allocate triangle bounds data.
  tribound = (vector *) calloc(num_tri*2, sizeof(vector));
  if (tribound == NULL) {
    char errbuf[MIRAGE_STRING_LENGTH];

    sprintf(errbuf, "Can't allocate memory for %d triangle bounds!\n",
	    2*num_tri);
    fatal_error("triangleMeshPrimitive::constructor", errbuf,
		"Out of memory.");
  }
  for (i = 0; i < num_tri; i++) {
    Vector::min(vertex[tri[i].v0].vertex, vertex[tri[i].v1].vertex,
		tribound[2*i + 0]);
    Vector::min(vertex[tri[i].v2].vertex, tribound[2*i + 0],
		tribound[2*i + 0]);

    Vector::max(vertex[tri[i].v0].vertex, vertex[tri[i].v1].vertex,
		tribound[2*i + 1]);
    Vector::max(vertex[tri[i].v2].vertex, tribound[2*i + 1],
		tribound[2*i + 1]);
  }

  // Set up mesh parameters.
  subdivision_limit = subdivision;
  stop_at_first_intersection = stop_at_first;
  phong_shading = do_phong_shading;
  printf("phong shading is %s\n", (phong_shading) ? "ON" : "OFF");
  

  //  printf("TMP: Calc octree variables\n");

  // Allocate variable space. ***NB*** cell_size etc
  // will be indexed from 0 to subdivision_limit, requiring
  // subdivision_limit + 1 elements worth of space.
  cell_size      = (vector *) calloc(subdivision_limit + 1, sizeof(vector));
  delta_t_cell_x = (double *) calloc(subdivision_limit + 1, sizeof(double));
  delta_t_cell_y = (double *) calloc(subdivision_limit + 1, sizeof(double));
  delta_t_cell_z = (double *) calloc(subdivision_limit + 1, sizeof(double));
    
  if (cell_size == NULL || delta_t_cell_x == NULL ||
      delta_t_cell_y == NULL || delta_t_cell_z == NULL) {
    char errbuf[MIRAGE_STRING_LENGTH];
    
    sprintf(errbuf, "can't allocate variable space for subdivision level %d",
	    subdivision_limit);
    fatal_error("triangleMeshPrimitive", errbuf,
		"You are running seriously low on memory (this is not a\n"
		"big request). Possibly there is a memory leak or you\n"
		"are trying to render a huge number of frames at once.");
  }

  //  printf("TMP: Calc bounding box\n");

  // Calculate the triangle mesh's bounding box.
  Vector::assign(minbound, +INFINITY, +INFINITY, +INFINITY);
  Vector::assign(maxbound, -INFINITY, -INFINITY, -INFINITY);
  for (i = 0; i < num_vertices; i++) {
    Vector::min(vertex[i].vertex, minbound, minbound);
    Vector::max(vertex[i].vertex, maxbound, maxbound);
  }

  maxbound[X] += TRIMESH_MAX_PAD_SIZE;
  maxbound[Y] += TRIMESH_MAX_PAD_SIZE;
  maxbound[Z] += TRIMESH_MAX_PAD_SIZE;

  minbound[X] -= TRIMESH_MIN_PAD_SIZE;
  minbound[Y] -= TRIMESH_MIN_PAD_SIZE;
  minbound[Z] -= TRIMESH_MIN_PAD_SIZE;

  Vector::sub(maxbound, minbound, bound_size);

#if 0
  printf("TMP: Bounding box is (%f %f %f) - (%f %f %f)\n",
	 minbound[X], minbound[Y], minbound[Z],
	 maxbound[X], maxbound[Y], maxbound[Z]);
#endif

  //  printf("TMP: Calc normals\n");

  // Calculate normals for each triangle, and use these to calculate the
  // normals for each vertex.
  for (i = 0, t = tri; i < num_tri; i++, t++) {
#if 0
    printf("TMP: v0 %f %f %f\n",
	   vertex[t->v0].vertex[X],
	   vertex[t->v0].vertex[Y],
	   vertex[t->v0].vertex[Z]);
    printf("TMP: v1 %f %f %f\n",
	   vertex[t->v1].vertex[X],
	   vertex[t->v1].vertex[Y],
	   vertex[t->v1].vertex[Z]);
    printf("TMP: v2 %f %f %f\n",
	   vertex[t->v2].vertex[X],
	   vertex[t->v2].vertex[Y],
	   vertex[t->v2].vertex[Z]);
#endif
    Vector::sub(vertex[t->v1].vertex, vertex[t->v0].vertex, edge1);
    Vector::sub(vertex[t->v2].vertex, vertex[t->v0].vertex, edge2);
    //    printf("TMP: edge1 %f %f %f\n", edge1[X], edge1[Y], edge1[Z]);
    //    printf("TMP: edge2 %f %f %f\n", edge2[X], edge2[Y], edge2[Z]);
    Vector::cross(edge1, edge2, t->norm);
    //    printf("TMP: norm %f %f %f\n", t->norm[X], t->norm[Y], t->norm[Z]);

    Vector::normalise(t->norm, "triangle normal in triMeshPrim constructor");
    
    Vector::add(t->norm, vertex[t->v0].normal, vertex[t->v0].normal);
    Vector::add(t->norm, vertex[t->v1].normal, vertex[t->v1].normal);
    Vector::add(t->norm, vertex[t->v2].normal, vertex[t->v2].normal);
  }

  // Normalise each vertex's normals.
  for (i = 0; i < num_vertices; i++)
    Vector::normalise(vertex[i].normal,
		      "triMeshPrim: normalise vertex normal");

  //  printf("TMP: Init storage\n");

  // Set up the octree storage memory.
  init_storage();

  //  printf("TMP: Subdivide\n");

  // Subdivide the triangle mesh.
  subdivide();

  //  printf("TMP: Constructed\n");
}

// Destructor.
triangleMeshPrimitive::~triangleMeshPrimitive(void)
{
  free(tri);
  free(vertex);
  free(cell_size);
  free(delta_t_cell_x);
}

int triangleMeshPrimitive::ray_intersect(IntersectionList *int_list,
					 primnode *node,
					 primnode *validate_node, 
					 vector ray_from, vector ray_dir,
					 uint ray_flags,
					 signature ray_signature)
{
  int ray_result, cell_no;
  vector t_entry, t_exit;
  double p_x, p_y, p_z, d_x, d_y, d_z;
  double overall_entry, overall_exit;
  double ray_dir_mag;

  // Store the signature of this ray that we're intersecting with.
  node->last_ray_signature = ray_signature;

  // Transform the ray into the mesh's primitive space.
  Matrix::apply(node->position[AFT], ray_from, mesh_ray_from);
  Matrix::applynoshift(node->position[AFT], ray_dir, mesh_ray_dir);

  p_x = mesh_ray_from[X];  d_x = mesh_ray_dir[X];
  p_y = mesh_ray_from[Y];  d_y = mesh_ray_dir[Y];
  p_z = mesh_ray_from[Z];  d_z = mesh_ray_dir[Z];

  /*
   * Determine whether or not the ray intersects the mesh's
   * bounding box.
   */
  
  // If the ray is outside the bounds and heading away from them, then there
  // will be no intersections.
  if (p_x > maxbound[X] && d_x > 0.0) {
    node->last_ray_hit = FALSE;
    return FALSE;
  }
  if (p_x < minbound[X] && d_x < 0.0) {
    node->last_ray_hit = FALSE;
    return FALSE;
  }
  if (p_y > maxbound[Y] && d_y > 0.0) {
    node->last_ray_hit = FALSE;
    return FALSE;
  }
  if (p_y < minbound[Y] && d_y < 0.0) {
    node->last_ray_hit = FALSE;
    return FALSE;
  }
  if (p_z > maxbound[Z] && d_z > 0.0) {
    node->last_ray_hit = FALSE;
    return FALSE;
  }
  if (p_z < minbound[Z] && d_z < 0.0) {
    node->last_ray_hit = FALSE;
    return FALSE;
  }

  // Determine the low and high t-values where the ray crosses the bounding
  // planes of the mesh.
  if (d_x > TINY) {
    // The ray is heading towards +X. Check to see if it is outside the
    // box where it starts.
    if (p_x > maxbound[X]) {
      node->last_ray_hit = FALSE;
      return FALSE;
    }

    // Determine t-values for entry and exit in X.
    t_entry[X] = (minbound[X] - p_x) / d_x;
    t_exit [X] = (maxbound[X] - p_x) / d_x;
  } else if (d_x < -TINY) {
    // The ray is heading towards -X. Check to see if it is outside the
    // box where it starts.
    if (p_x < minbound[X]) {
      node->last_ray_hit = FALSE;
      return FALSE;
    }

    // Determine t-values for entry and exit in X.
    t_entry[X] = (maxbound[X] - p_x) / d_x;
    t_exit [X] = (minbound[X] - p_x) / d_x;
  } else {
    // The ray isn't moving appreciably in X. Check to see if the starting
    // point is outside the bounding planes.
    if ((p_x < minbound[X]) || (p_x > maxbound[X])) {
      node->last_ray_hit = FALSE;
      return FALSE;
    }

    // The ray is always inside the bounds in X.
    t_entry[X] = -INFINITY;
    t_exit [X] = +INFINITY;
  }

  if (d_y > TINY) {
    // The ray is heading towards +Y. Check to see if it is outside the
    // box where it starts.
    if (p_y > maxbound[Y]) {
      node->last_ray_hit = FALSE;
      return FALSE;
    }

    // Determine t-values for entry and exit in Y.
    t_entry[Y] = (minbound[Y] - p_y) / d_y;
    t_exit [Y] = (maxbound[Y] - p_y) / d_y;
  } else if (d_y < -TINY) {
    // The ray is heading towards -Y. Check to see if it is outside the
    // box where it starts.
    if (p_y < minbound[Y]) {
      node->last_ray_hit = FALSE;
      return FALSE;
    }

    // Determine t-values for entry and exit in Y.
    t_entry[Y] = (maxbound[Y] - p_y) / d_y;
    t_exit [Y] = (minbound[Y] - p_y) / d_y;
  } else {
    // The ray isn't moving appreciably in Y. Check to see if the starting
    // point is outside the bounding planes.
    if ((p_y < minbound[Y]) || (p_y > maxbound[Y])) {
      node->last_ray_hit = FALSE;
      return FALSE;
    }

    // The ray is always inside the bounds in Y.
    t_entry[Y] = -INFINITY;
    t_exit [Y] = +INFINITY;
  }

  if (d_z > TINY) {
    // The ray is heading towards +Z. Check to see if it is outside the
    // box where it starts.
    if (p_z > maxbound[Z]) {
      node->last_ray_hit = FALSE;
      return FALSE;
    }

    // Determine t-values for entry and exit in Z.
    t_entry[Z] = (minbound[Z] - p_z) / d_z;
    t_exit [Z] = (maxbound[Z] - p_z) / d_z;
  } else if (d_z < -TINY) {
    // The ray is heading towards -Z. Check to see if it is outside the
    // box where it starts.
    if (p_z < minbound[Z]) {
      node->last_ray_hit = FALSE;
      return FALSE;
    }

    // Determine t-values for entry and exit in Z.
    t_entry[Z] = (maxbound[Z] - p_z) / d_z;
    t_exit [Z] = (minbound[Z] - p_z) / d_z;
  } else {
    // The ray isn't moving appreciably in Z. Check to see if the starting
    // point is outside the bounding planes.
    if ((p_z < minbound[Z]) || (p_z > maxbound[Z])) {
      node->last_ray_hit = FALSE;
      return FALSE;
    }

    // The ray is always inside the bounds in Z.
    t_entry[Z] = -INFINITY;
    t_exit [Z] = +INFINITY;
  }

  // Determine the overall (maximum) t_entry and overall (minimum) t_exit.
  if (t_entry[X] > t_entry[Y])
    if (t_entry[X] > t_entry[Z])
      overall_entry = t_entry[X];
    else
      overall_entry = t_entry[Z];
  else
    if (t_entry[Y] > t_entry[Z])
      overall_entry = t_entry[Y];
    else
      overall_entry = t_entry[Z];

  if (t_exit[X] < t_exit[Y])
    if (t_exit[X] < t_exit[Z])
      overall_exit = t_exit[X];
    else
      overall_exit = t_exit[Z];
  else
    if (t_exit[Y] < t_exit[Z])
      overall_exit = t_exit[Y];
    else
      overall_exit = t_exit[Z];

  // If the ray misses the bounding box then don't bother looking at the mesh.
  if (overall_entry > overall_exit) {
    node->last_ray_hit = FALSE;
    return FALSE;
  }

  /*
   * Otherwise, we DO need to traverse the octree.
   */

  // Set up the entry t value for the first cell. This is given by the
  // t-value of the entry of the octree bounds. However if this is less
  // than zero, we do not want to accept any intersections behind the
  // eyepoint. So, we set the entry value to be zero in this case.
  cell_entry_t = MAX(0, overall_entry);

  // Set temporary variables so they don't have to be passed around
  // when traversing the octree.
  this_ray_flags = ray_flags;
  this_ray_signature = ray_signature;
  the_list = int_list;
  mesh_primnode = node;
  validate_primnode = validate_node;

  /*
   * Set up variables for the octree traversal.
   */

  ray_dir_mag = Vector::magnitude(mesh_ray_dir);

  // Find the size of the top-level cell (note that since octrees are always
  // divided at least once this is half of the bound size).
  Vector::mult(0.5, bound_size, cell_size[0]);

  // Work out which octant the ray is coming from and what the exit
  // values are of X, Y and Z. Also find the absolute values of the ray's
  // direction vector (if -ve, make it +ve).
  ray_entry_octant = 0;
  if (d_x < 0) {
    ray_entry_octant |= (1 << X);
    exit_value[X] = minbound[X];
    d_x = -d_x;
  } else {
    exit_value[X] = maxbound[X];
  }
  if (d_y < 0) {
    ray_entry_octant |= (1 << Y);
    exit_value[Y] = minbound[Y];
    d_y = -d_y;
  } else {
    exit_value[Y] = maxbound[Y];
  }
  if (d_z < 0) {
    ray_entry_octant |= (1 << Z);
    exit_value[Z] = minbound[Z];
    d_z = -d_z;
  } else {
    exit_value[Z] = maxbound[Z];
  }
  //  printf("TMP: Ray entry octant is %d\n", ray_entry_octant);

  // Work out the amount that the ray must traverse to reach the
  // far edge of the bounding volume.
  exit_value[X] = fabs(mesh_ray_from[X] - exit_value[X]);
  exit_value[Y] = fabs(mesh_ray_from[Y] - exit_value[Y]);
  exit_value[Z] = fabs(mesh_ray_from[Z] - exit_value[Z]);

  // Determine values that are necessary for the cell-skipping algorithm.
  //
  // delta_t_cell_x[0], delta_t_cell_y[0] and delta_t_cell_z[0] are the
  //  t-values the ray takes to travel a top-level cell width in X, Y or Z.
  //  (a real top-level cell, not subdivided by one like cell_width[0]).
  //
  // t_next_x, t_next_y and t_next_z are the t-values when the ray will hit
  //  the edge of the bounding volume.

  if (d_x == 0) {
    delta_t_cell_x[0] = 0.0;     // Note that this will never be used since
                                 // the ray won't cross X boundaries.
    t_next_x = INFINITY;
    ray_dir_ratio_y_x = INFINITY;
    ray_dir_ratio_z_x = INFINITY;
  } else {
    ray_dir_ratio_y_x = d_y / d_x;
    ray_dir_ratio_z_x = d_z / d_x;

    delta_t_cell_x[0] = 1.0 / d_x;        // t for ray to cross 1 X unit
    t_next_x = exit_value[X] * delta_t_cell_x[0];

    if (t_next_x >= INFINITY) {
      delta_t_cell_x[0] = 0.0;
      t_next_x = INFINITY;
    } else {
      // Work out how long for the ray to cross the top-level cell
      delta_t_cell_x[0] *= bound_size[X];
    }
  }

  if (d_y == 0) {
    delta_t_cell_y[0] = 0.0;     // Note that this will never be used since
                                 // the ray won't cross Y boundaries.
    t_next_y = INFINITY;
    ray_dir_ratio_x_y = INFINITY;
    ray_dir_ratio_z_y = INFINITY;
  } else {
    ray_dir_ratio_x_y = d_x / d_y;
    ray_dir_ratio_z_y = d_z / d_y;

    delta_t_cell_y[0] = 1.0 / d_y;        // t for ray to cross 1 Y unit
    t_next_y = exit_value[Y] * delta_t_cell_y[0];

    if (t_next_y >= INFINITY) {
      delta_t_cell_y[0] = 0.0;
      t_next_y = INFINITY;
    } else {
      // Work out how long for the ray to cross the top-level cell
      delta_t_cell_y[0] *= bound_size[Y];
    }
  }

  if (d_z == 0) {
    delta_t_cell_z[0] = 0.0;     // Note that this will never be used since
                                 // the ray won't cross Z boundaries.
    t_next_z = INFINITY;
    ray_dir_ratio_x_z = INFINITY;
    ray_dir_ratio_y_z = INFINITY;
  } else {
    ray_dir_ratio_x_z = d_x / d_z;
    ray_dir_ratio_y_z = d_y / d_z;

    delta_t_cell_z[0] = 1.0 / d_z;        // t for ray to cross 1 Z unit
    t_next_z = exit_value[Z] * delta_t_cell_z[0];

    if (t_next_z >= INFINITY) {
      delta_t_cell_z[0] = 0.0;
      t_next_z = INFINITY;
    } else {
      // Work out how long for the ray to cross the top-level cell
      delta_t_cell_z[0] *= bound_size[Z];
    }
  }

  // We have calculated cell_size, delta_t_cell_x etc for cell size 0
  // (top-level cell). Now calculate them for all the smaller cells.
  for (cell_no = 1; cell_no <= subdivision_limit; cell_no++) {
    delta_t_cell_x[cell_no] = (delta_t_cell_x[cell_no - 1] == INFINITY) ?
      INFINITY : delta_t_cell_x[cell_no - 1] * 0.5;

    delta_t_cell_y[cell_no] = (delta_t_cell_y[cell_no - 1] == INFINITY) ?
      INFINITY : delta_t_cell_y[cell_no - 1] * 0.5;

    delta_t_cell_z[cell_no] = (delta_t_cell_z[cell_no - 1] == INFINITY) ?
      INFINITY : delta_t_cell_z[cell_no - 1] * 0.5;

    Vector::mult(0.5, cell_size[cell_no - 1], cell_size[cell_no]);
  }

  // Intersect the ray with the octree structure. This traverses the ray
  // through the cells of the octree. In each cell the ray is intersected
  // with the triangle list found there. If any valid intersections are found
  // and the stop_at_first_intersection flag is set, then the traversal
  // stops. Otherwise keep looking for more intersections.
  ray_result = traverse_trioctree(0, my_octree, exit_value,
				  t_next_x, t_next_y, t_next_z);

  node->last_ray_hit = ray_result;
  return ray_result;
}

int triangleMeshPrimitive::ray_intersect(vector ray_from, vector ray_dir,
					 rayintersect *int_storage)
{
  // This routine is just too damn complicated to copy. It's not
  // likely to be needed for the interactive-CSG system, either.
  fatal_error("triangleMeshPrimitive::ray_intersect", "Incorrect r_i call!",
	      "This error should not happen.");
  return 0;
}

intersect_result triangleMeshPrimitive::point_intersect(primnode *node,
							vector wp,
							PointInfo *info,
							uint ray_flags)
{
  vector mesh_point;

  // Transform the worldpoint into meshspace.
  Matrix::apply(node->position[AFT], wp, mesh_point);

  if ((mesh_point[X] < minbound[X]) || (mesh_point[X] > maxbound[X]) ||
      (mesh_point[Y] < minbound[Y]) || (mesh_point[Y] > maxbound[Y]) ||
      (mesh_point[Z] < minbound[Z]) || (mesh_point[Z] > maxbound[Z]))
    return OUT;

  return octreemesh_point_intersect(node, mesh_point, info, ray_flags);
}

intersect_result triangleMeshPrimitive::point_intersect(vector wp)
{
  // This routine is just too damn complicated to copy. It's not
  // likely to be needed for the interactive-CSG system, either.
  fatal_error("triangleMeshPrimitive::point_intersect", "Incorrect p_i call!",
	      "This error should not happen.");
  return OUT;
}

intersect_result triangleMeshPrimitive::voxel_intersect(box b)
{
  box mesh_voxel_check;
  int i;
  
  // Check to see if the voxel needs to be intersected with the
  // mesh (see if it intersects the bounding box).
  // We do this by transforming the voxel to be checked so that it
  // lies in a space that maps minbound -> -1 and maxbound -> +1,
  // and then using the cube voxel intersection method.
  for (i = 0; i < 8; i++) {
    mesh_voxel_check[i][X] = -1 + 2*(b[i][X] - minbound[X]) / bound_size[X];
    mesh_voxel_check[i][Y] = -1 + 2*(b[i][Y] - minbound[Y]) / bound_size[Y];
    mesh_voxel_check[i][Z] = -1 + 2*(b[i][Z] - minbound[Z]) / bound_size[Z];
  }

  if (cube_voxel_intersect(mesh_voxel_check) == OUT) {
    // This voxel does not intersect the mesh, so the mesh doesn't need
    // to be intersected with inside it.
    return OUT;
  } else {
    // This voxel either intersects or is inside the mesh. The mesh does
    // need to be intersected with inside it.
    return SURFACE;
  }
}

char *triangleMeshPrimitive::prim_name(void)
{
  return "TriangleMesh";
}

void triangleMeshPrimitive::init_storage(void)
{
  octree_storage.blocks_used = 0;
  octree_storage.items_used = 0;
  octree_storage.items_freed = 0;
  octree_storage.next_free_octreenode = NO_FREE_TRIMEM;

  trilist_storage.blocks_used = 0;
  trilist_storage.items_used = 0;
  trilist_storage.items_freed = 0;
  trilist_storage.next_free_trilist = NO_FREE_TRIMEM;
}

void triangleMeshPrimitive::dealloc_storage(void)
{
  int i;

  // Deallocate all of the octree blocks.
  for (i = 0; i < octree_storage.blocks_used; i++)
    free(octree_storage.storage[i]);

  // Deallocate all of the trilist blocks.
  for (i = 0; i < trilist_storage.blocks_used; i++)
    free(trilist_storage.storage[i]);
}

trioctree *triangleMeshPrimitive::new_trioctree(trioctindex *new_index)
{
  trioctree *the_block, *the_node;
  int block_no, index_no;
  trioctindex free_index;

  // Check to see if we need to allocate a new octree block.
  if (octree_storage.next_free_octreenode == NO_FREE_TRIMEM) {

    if (octree_storage.blocks_used >= 1000)
      fatal_error("triMeshPrim::new_trioctree", "maximum memory size exceeded",
		  "There is a fixed limit on the number of trioctree blocks\n"
		  "that can be allocated. To change this, alter prim_mesh.h\n"
		  "and recompile Mirage.");

    the_block = (trioctree *) calloc(TRIOCT_BLOCKSIZE, sizeof(trioctree));
    if (the_block == NULL)
      fatal_error("triMeshPrim::new_trioctree",
		  "can't allocate new trioctree block!",
		  "The system has tried to dynamically expand its\n"
		  "working memory and has failed. This could indicate\n"
		  "that you are running severely low on memory or that\n"
		  "there is a memory leak in the system.");

    // Update the array of octree blocks.
    block_no = octree_storage.blocks_used++;
    octree_storage.storage[block_no] = the_block;

    // Now for each unused octree node in the block (which is all of them)
    // we set it to point to the next free octree node. We start from the
    // zero node, each one pointing to the next. The last one in the block
    // points to NO_FREE_TRIMEM.
    //
    // This pointer (a trioctindex) is stored in child[0] of the node as there
    // is no data being stored in the octree node right now anyway.

    // Trioctindex pointing to free node 1 of the current block.
    free_index = (TRIOCT_BLOCKSIZE * block_no + 1) | TRIINDEX_OCTREE;
    octree_storage.next_free_octreenode = free_index;

    for(index_no = 0; index_no < TRIOCT_BLOCKSIZE - 1; index_no++) {
      octree_storage.storage[block_no][index_no].child[0] = free_index;

      free_index++;
    }

    octree_storage.storage[block_no][TRIOCT_BLOCKSIZE - 1].child[0] =
      NO_FREE_TRIMEM;

    if (new_index != NULL)
      *new_index = (TRIOCT_BLOCKSIZE * block_no) | TRIINDEX_OCTREE;

    octree_storage.items_used++;
    
    return octree_storage.storage[block_no];
  } else {
    if (new_index != NULL)
      *new_index = octree_storage.next_free_octreenode;

    the_node = GET_TRIOCT(octree_storage.next_free_octreenode);

    // The next element in the free list is stored in child[0] of this node.
    octree_storage.next_free_octreenode = the_node->child[0];

    octree_storage.items_used++;
    
    return the_node;
  }
}

trilist *triangleMeshPrimitive::new_trilist(trioctindex *new_index)
{
  trilist *the_block, *the_node;
  int block_no, index_no;
  trioctindex free_index;

  // Check to see if we need to allocate a new trilist block.
  if (trilist_storage.next_free_trilist == NO_FREE_TRIMEM) {

    if (trilist_storage.blocks_used >= 1000)
      fatal_error("triMeshPrim::new_trilist", "maximum memory size exceeded",
		  "There is a fixed limit on the number of trilist blocks\n"
		  "that can be allocated. To change this, alter prim_mesh.h\n"
		  "and recompile Mirage.");

    the_block = (trilist *) calloc(TRILIST_BLOCKSIZE, sizeof(trilist));
    if (the_block == NULL)
      fatal_error("triMeshPrim::new_trilist",
		  "can't allocate new trilist block!",
		  "The system has tried to dynamically expand its\n"
		  "working memory and has failed. This could indicate\n"
		  "that you are running severely low on memory or that\n"
		  "there is a memory leak in the system.");

    // Update the array of trilist blocks.
    block_no = trilist_storage.blocks_used++;
    trilist_storage.storage[block_no] = the_block;

    // Now for each unused trilist node in the block (which is all of them)
    // we set it to point to the next free trilist node. We start from the
    // zero node, each one pointing to the next. The last one in the block
    // points to NO_FREE_TRIMEM.
    //
    // This pointer (a trioctindex) is stored in nextlist of the node as there
    // is no data being stored in the node right now anyway.

    // Trioctindex pointing to free node 1 of the current block.
    free_index = (TRILIST_BLOCKSIZE * block_no + 1) | TRIINDEX_LIST;
    trilist_storage.next_free_trilist = free_index;

    for(index_no = 0; index_no < TRILIST_BLOCKSIZE - 1; index_no++) {
      trilist_storage.storage[block_no][index_no].nextlist = free_index;

      free_index++;
    }

    trilist_storage.storage[block_no][TRILIST_BLOCKSIZE - 1].nextlist =
      NO_FREE_TRIMEM;

    if (new_index != NULL)
      *new_index = (TRILIST_BLOCKSIZE * block_no) | TRIINDEX_LIST;

    trilist_storage.items_used++;
    
    return trilist_storage.storage[block_no];
  } else {
    if (new_index != NULL)
      *new_index = trilist_storage.next_free_trilist;

    the_node = GET_TRILIST(trilist_storage.next_free_trilist);

    // The next element in the free list is stored in nextlist of this node.
    trilist_storage.next_free_trilist = the_node->nextlist;

    trilist_storage.items_used++;
    
    return the_node;
  }
}

// Free a triangle octree node.
void triangleMeshPrimitive::free_trioctreenode(trioctindex the_index)
{
  trioctree *freed_node;

  freed_node = GET_TRIOCT(the_index);

  // Set the recently freed node to point to the
  // previously first node on the free-memory chain.
  freed_node->child[0] = octree_storage.next_free_octreenode;

  // The next free node is now this recently freed one.
  octree_storage.next_free_octreenode = the_index;

  octree_storage.items_used--;
    
  octree_storage.items_freed++;
}

// Free a triangle list node.
void triangleMeshPrimitive::free_trilistnode(trioctindex the_index)
{
  trilist *freed_node;

  freed_node = GET_TRILIST(the_index);

  // Set the recently freed node to point to the
  // previously first node on the free-memory chain.
  freed_node->nextlist = trilist_storage.next_free_trilist;

  // The next free node is now this recently freed one.
  trilist_storage.next_free_trilist = the_index;

  trilist_storage.items_used--;
    
  trilist_storage.items_freed++;
}

// Free a triangle-list octree structure recursively.
void triangleMeshPrimitive::free_trilist(trioctindex the_index)
{
  if (INDEX_ISLIST(the_index)) {
    trilist *the_list = GET_TRILIST(the_index);
    
    free_trilist(the_list->nextlist);
    free_trilistnode(the_index);
  }
}

trioctree *triangleMeshPrimitive::get_oct(trioctindex index)
{
  int block_no, index_no;

  block_no = index / TRIOCT_BLOCKSIZE;
  index_no = index % TRIOCT_BLOCKSIZE;

  if (block_no > octree_storage.blocks_used)
    fatal_error("triMeshPrim::get_oct", "requested trioct # outside memory",
		"This error should not happen.");

  return &(octree_storage.storage[block_no][index_no]);  
}

trilist *triangleMeshPrimitive::get_list(trioctindex index)
{
  int block_no, index_no;

  block_no = index / TRILIST_BLOCKSIZE;
  index_no = index % TRILIST_BLOCKSIZE;

  if (block_no > trilist_storage.blocks_used)
    fatal_error("triMeshPrim::get_list", "requested trilist # outside memory",
		"This error should not happen.");

  return &(trilist_storage.storage[block_no][index_no]);  
}

void triangleMeshPrimitive::subdivide(void)
{
  int i;
  trilist *listelem;
  trioctindex last_elem, this_elem;

  //  printf("TMP: Create triangle list\n");

  // Create a triangle list that is all the triangles in the mesh.
  last_elem = TRIINDEX_EMPTY;
  for (i = 0; i < num_tri; i++) {
    listelem = new_trilist(&this_elem);
    
    listelem->tri = i;
    listelem->nextlist = last_elem;

    last_elem = this_elem;
  }

  //  printf("TMP: Initial subdivision\n");

  // Subdivide this list and form an octree.
  my_octree = trioct_subdivide(last_elem, num_tri, 0,
			       minbound[X], minbound[Y], minbound[Z],
			       maxbound[X], maxbound[Y], maxbound[Z]);

  //  printf("TMP: Compress octree\n");

  // Compress the octree, looking for empty nodes to save space.
  compress_trioct(&my_octree);

  //  printf("TMP: Subdivision complete\n");
}

int triangleMeshPrimitive::compress_trioct(trioctindex *index)
{
  // Traverse the octree, turning any nodes that are all
  // empty into a single empty value. This wil remove quite
  // a few octree nodes (sometimes).

  if (INDEX_ISEMPTY(*index)) {
    // This is empty, return so.
    //    printf("TMP: compress empty\n");
    return 1;
  }

  if (INDEX_ISOCTREE(*index)) {
    int node_index, empty_nodes = 0;
    trioctindex *children = GET_TRIOCT(*index)->child;

    //    printf("TMP: checking octree\n");

    for (node_index = 0; node_index < 8; node_index++) {
      //      printf("TMP: compress child %d...\n", node_index);
      empty_nodes += compress_trioct(&children[node_index]);
    }

    if (empty_nodes == 8) {
      // This octree has all eight children empty, compress it.
      //      printf("TMP: octree empty, compresssing\n");
      free_trioctreenode(*index);
      *index = TRIINDEX_EMPTY;
      return 1;
    }
  }

  //  printf("TMP: not an empty octree\n");
  // This octree wasn't compressed.
  return 0;
}

trioctindex triangleMeshPrimitive::trioct_subdivide(trioctindex to_subdivide,
						    int num_tris_in_list,
						    int already_subdivided,
						    double x_min, double y_min,
						    double z_min, double x_max,
						    double y_max, double z_max)
{
  trilist *list;
  trioctree *oct;
  trioctindex oct_index, trimmed[8];
  double x_mid, y_mid, z_mid;
  int tris[8];
  
  //  printf("TMP: Subdivide in\n");
  
  // Empty space returns an empty octree.
  if (INDEX_ISEMPTY(to_subdivide)) {
    //    printf("TMP: Subdivide empty, out\n");
    return TRIINDEX_EMPTY;
  }

  // Full space returns a full octree.
  if (INDEX_ISFULL(to_subdivide)) {
    //    printf("TMP: Subdivide full, out\n");
    return TRIINDEX_FULL;
  }

  // Don't subdivide beyond a certain level.
  if (already_subdivided >= subdivision_limit) {
    //    printf("TMP: Subdivide at max limit, out\n");
    return to_subdivide;
  }

  // If this isn't a triangle list, scream.
  if (!INDEX_ISLIST(to_subdivide))
    fatal_error("triMeshPrim::trioct_subvide",
		"subdivision object not a trilist!",
		"This error should not happen.");

  //  printf("TMP: %d tris in list\n", num_tris_in_list);

  if (num_tris_in_list <= max_tris_in_voxel) {
    // Don't subdivide this list, it is a sufficiently low
    // number of triangles for one voxel.
    //    printf("TMP: less than %d tris in list, out\n", max_tris_in_voxel);
    return to_subdivide;
  }

  // Calculate the midpoints of the voxel to be subdivided.
  x_mid = (x_min + x_max) * 0.5;
  y_mid = (y_min + y_max) * 0.5;
  z_mid = (z_min + z_max) * 0.5;

  list = GET_TRILIST(to_subdivide);
 
  // Form the eight child triangle lists.
  trimmed[0] = trim_triangle_list(list, &tris[0],
				  x_min, y_min, z_min,
				  x_mid, y_mid, z_mid);

  trimmed[1] = trim_triangle_list(list, &tris[1],
				  x_mid, y_min, z_min,
				  x_max, y_mid, z_mid);

  trimmed[2] = trim_triangle_list(list, &tris[2],
				  x_min, y_mid, z_min,
				  x_mid, y_max, z_mid);

  trimmed[3] = trim_triangle_list(list, &tris[3],
				  x_mid, y_mid, z_min,
				  x_max, y_max, z_mid);

  trimmed[4] = trim_triangle_list(list, &tris[4],
				  x_min, y_min, z_mid,
				  x_mid, y_mid, z_max);

  trimmed[5] = trim_triangle_list(list, &tris[5],
				  x_mid, y_min, z_mid,
				  x_max, y_mid, z_max);

  trimmed[6] = trim_triangle_list(list, &tris[6],
				  x_min, y_mid, z_mid,
				  x_mid, y_max, z_max);

  trimmed[7] = trim_triangle_list(list, &tris[7],
				  x_mid, y_mid, z_mid,
				  x_max, y_max, z_max);
				       
  //  printf("TMP: free triangle list\n");

  // Free the parent triangle list, it is no longer necessary.
  free_trilist(to_subdivide);

  //  printf("TMP: new octree\n");

  // Allocate an octree node for this cell.
  oct = new_trioctree(&oct_index);

  // Fill the octree.
  oct->child[0] = trioct_subdivide(trimmed[0], tris[0], already_subdivided + 1,
				   x_min, y_min, z_min,
				   x_mid, y_mid, z_mid);

  oct->child[1] = trioct_subdivide(trimmed[1], tris[1], already_subdivided + 1,
				   x_mid, y_min, z_min,
				   x_max, y_mid, z_mid);

  oct->child[2] = trioct_subdivide(trimmed[2], tris[2], already_subdivided + 1,
				   x_min, y_mid, z_min,
				   x_mid, y_max, z_mid);

  oct->child[3] = trioct_subdivide(trimmed[3], tris[3], already_subdivided + 1,
				   x_mid, y_mid, z_min,
				   x_max, y_max, z_mid);
  
  oct->child[4] = trioct_subdivide(trimmed[4], tris[4], already_subdivided + 1,
				   x_min, y_min, z_mid,
				   x_mid, y_mid, z_max);

  oct->child[5] = trioct_subdivide(trimmed[5], tris[5], already_subdivided + 1,
				   x_mid, y_min, z_mid,
				   x_max, y_mid, z_max);

  oct->child[6] = trioct_subdivide(trimmed[6], tris[6], already_subdivided + 1,
				   x_min, y_mid, z_mid,
				   x_mid, y_max, z_max);

  oct->child[7] = trioct_subdivide(trimmed[7], tris[7], already_subdivided + 1,
				   x_mid, y_mid, z_mid,
				   x_max, y_max, z_max);
  
  // Return the newly formed octree.
  return oct_index;
}

trioctindex triangleMeshPrimitive::trim_triangle_list(trilist *list,
						      int *new_numtri,
			       double x_min, double y_min, double z_min,
			       double x_max, double y_max, double z_max)
{
  trioctindex last_elem, this_elem;
  trilist *newnode, *listptr;
  int newlist_size;

  //  printf("TMP: trim triangle list\n");

  last_elem = TRIINDEX_EMPTY;
  newlist_size = 0;
  listptr = list;

  do {
#if 0
    printf("Trim type %d, index %d\n",
	   ((listptr->nextlist) & TRIINDEX_TYPE) >> 24,
	   (listptr->nextlist) & TRIINDEX_INDEX);
#endif
    if (tri_box_intersect(listptr->tri,
			  x_min, y_min, z_min, x_max, y_max, z_max)) {
      // Add this triangle to the new list.
      newnode = new_trilist(&this_elem);
      
      newnode->tri = listptr->tri;
      newnode->nextlist = last_elem;

      last_elem = this_elem;
      newlist_size++;
    }
    
    if (!INDEX_ISLIST(listptr->nextlist)) {
      // Reached the end of the list, no more triangles.
      break;
    } else {
      listptr = GET_TRILIST(listptr->nextlist);
    }
  } while(1);

  //  printf("TMP: trimmed to %d tris\n", newlist_size);

  *new_numtri = newlist_size;

  // We have finished processing the triangles. Return the list (may be
  // empty).
  return last_elem;
  
  // FIXME: this may be FULL if There are no triangles here. How to determine
  // this?
}

int triangleMeshPrimitive::traverse_trioctree(int current_level,
					      trioctindex index,
					      vector parent_cell_exit,
					      double t_next_x,
					      double t_next_y,
					      double t_next_z)
{
  if (INDEX_ISEMPTY(index)) {
    //    printf("TMP: TTO: traverse empty\n");
    return FALSE;
  } else if (INDEX_ISLIST(index)) {
    // Intersect this ray with the triangles that are present in this
    // octree cell.
    //    printf("TMP: TTO: traverse trilist\n");
    trilist *listptr = GET_TRILIST(index);
    meshtri *t;
    int found_int = FALSE;
    
    do {
      t = &tri[listptr->tri];
      //      printf("TMP: TTO: tri intersect\n");
      tri_ray_intersect(t, &found_int);

      if (!INDEX_ISLIST(listptr->nextlist))
	break;

      listptr = GET_TRILIST(listptr->nextlist);
    } while(1);

    return found_int;
  } else if (INDEX_ISOCTREE(index)) {
    // Traverse this octree node's children recursively if the ray
    // intersects them.
    //    printf("TMP: TTO: traverse octree\n");
    unsigned char child_no = 0;
    trioctindex next_child, *children_list;
    vector child_cell_exit;
    int child_result, parent_result = FALSE;

    // Get the size of the child cells.
    double cell_x = cell_size[current_level][X],
      cell_y = cell_size[current_level][Y],
      cell_z = cell_size[current_level][Z];

    //    printf("TMP: TTO: cell size of child cells is %f %f %f\n",
    //	   cell_x, cell_y, cell_z);

    current_level++;
    
    // Get a pointer to the children cells.
    children_list = GET_TRIOCT(index)->child;

    // Work out which child of this node the ray starts in.
    Vector::set(child_cell_exit, parent_cell_exit);
    
    if (child_cell_exit[X] > cell_x) {
      // We start in the low-X child cell. This means that the next X crossing
      // won't be at the edge of the parent cell but rather at the edge of
      // the low-X child cell.
      child_cell_exit[X] -= cell_x;
      t_next_x -= delta_t_cell_x[current_level];
      //      printf("TMP: TTO: ray in low X\n");
    } else {
      // We start in the high-X child cell. The next X crossing is at the
      // edge of the parent cell.
      child_no |= (1 << X);
      //      printf("TMP: TTO: ray in high X\n");
    }

    if (child_cell_exit[Y] > cell_y) {
      // We start in the low-Y child cell. This means that the next Y crossing
      // won't be at the edge of the parent cell but rather at the edge of
      // the low-Y child cell.
      child_cell_exit[Y] -= cell_y;
      t_next_y -= delta_t_cell_y[current_level];
      //      printf("TMP: TTO: ray in low Y\n");
    } else {
      // We start in the high-Y child cell. The next Y crossing is at the
      // edge of the parent cell.
      child_no |= (1 << Y);
      //      printf("TMP: TTO: ray in high Y\n");
    }

    if (child_cell_exit[Z] > cell_z) {
      // We start in the low-Z child cell. This means that the next Z crossing
      // won't be at the edge of the parent cell but rather at the edge of
      // the low-Z child cell.
      child_cell_exit[Z] -= cell_z;
      t_next_z -= delta_t_cell_z[current_level];
      //      printf("TMP: TTO: ray in low Z\n");
    } else {
      // We start in the high-Z child cell. The next Z crossing is at the
      // edge of the parent cell.
      child_no |= (1 << Z);
      //      printf("TMP: TTO: ray in high Z\n");
    }

    // Now we know which child cell the ray starts in. Get this child and
    // recursively traverse it. Note that all of the above calculations
    // are done with absolute values of the distances, etc. and work if
    // the ray is coming from the low-XYZ octant. If this isn't the case,
    // then we are actually indexing a different child to the one found
    // above; we do this by doing a bitwise XOR with the octant that the
    // ray is coming from.
    next_child = children_list[child_no ^ ray_entry_octant];

    //    printf("TMP: TTO: octree leaf %d ...\n", child_no ^ ray_entry_octant);
    if (!INDEX_ISEMPTY(next_child)) {
      //      printf("TMP: TTO: nonempty leaf, recursive traverse\n");
      child_result = traverse_trioctree(current_level, next_child,
					child_cell_exit, t_next_x, t_next_y,
					t_next_z);

      // If we are stopping at the first intersection and we found
      // something [full shadow or a solid intersection], stop.
      if (child_result) {
	if (stop_at_first_intersection) {
	  return child_result;
	} else {
	  parent_result = TRUE;
	}
      }
    } else {
      //      printf("TMP: TTO: empty leaf, skipping\n");
    }

    // We know which child the ray starts in and have intersected with it.
    // Traverse the rest of the child cells, stopping once we reach the
    // high-XYZ cell or we hit an edge of the parent cell.
    while (child_no != ((1 << X) | (1 << Y) | (1 << Z))) {
      // Work out which direction we will go in next (X, Y or Z)
      if (t_next_x <= t_next_y) {
	if (t_next_x <= t_next_z) {

	  //	  printf("TMP: TTO: ray hits X boundary\n");

	  // We will hit an X boundary next.
	  if (child_no & (1 << X)) {
	    // We're already in a high-X child, so we've hit a boundary
	    // of the parent cell. Exit.
	    //	    printf("TMP: TTO: already high-X, exit\n");
	    break;
	  }

	  // We are now in the high-X child.
	  child_no |= (1 << X);

	  // Set cell_entry_t to be the entry of this child cell.
	  // Set the t-value of hitting the next X boundary up by the X
	  // increment.
	  cell_entry_t = t_next_x;
	  t_next_x += delta_t_cell_x[current_level];

	  // Find the new distances to go until the X, Y and Z boundaries.
	  child_cell_exit[Y] -= child_cell_exit[X] * ray_dir_ratio_y_x;
	  child_cell_exit[Z] -= child_cell_exit[X] * ray_dir_ratio_z_x;
	  child_cell_exit[X] = cell_x;
	} else {

	  //	  printf("TMP: TTO: ray hits Z boundary\n");

	  // We will hit a Z boundary next.
	  if (child_no & (1 << Z)) {
	    // We're already in a high-Z child, so we've hit a boundary
	    // of the parent cell. Exit.
	    //	    printf("TMP: TTO: already high-Z, exit\n");
	    break;
	  }

	  // We are now in the high-Z child.
	  child_no |= (1 << Z);

	  // Set cell_entry_t to be the entry of this child cell.
	  // Set the t-value of hitting the next Z boundary up by the Z
	  // increment.
	  cell_entry_t = t_next_z;
	  t_next_z += delta_t_cell_z[current_level];

	  // Find the new distances to go until the X, Y and Z boundaries.
	  child_cell_exit[X] -= child_cell_exit[Z] * ray_dir_ratio_x_z;
	  child_cell_exit[Y] -= child_cell_exit[Z] * ray_dir_ratio_y_z;
	  child_cell_exit[Z] = cell_z;
	}
      } else {
	if (t_next_y <= t_next_z) {

	  //	  printf("TMP: TTO: ray hits Y boundary\n");

	  // We will hit a Y boundary next.
	  if (child_no & (1 << Y)) {
	    // We're already in a high-Y child, so we've hit a boundary
	    // of the parent cell. Exit.
	    //	    printf("TMP: TTO: already high-Y, exit\n");
	    break;
	  }

	  // We are now in the high-Y child.
	  child_no |= (1 << Y);

	  // Set cell_entry_t to be the entry of this child cell.
	  // Set the t-value of hitting the next Y boundary up by the Y
	  // increment.
	  cell_entry_t = t_next_y;
	  t_next_y += delta_t_cell_y[current_level];

	  // Find the new distances to go until the X, Y and Z boundaries.
	  child_cell_exit[X] -= child_cell_exit[Y] * ray_dir_ratio_x_y;
	  child_cell_exit[Z] -= child_cell_exit[Y] * ray_dir_ratio_z_y;
	  child_cell_exit[Y] = cell_y;
	} else {

	  //	  printf("TMP: TTO: ray hits Z boundary\n");

	  // We will hit a Z boundary next.
	  if (child_no & (1 << Z)) {
	    // We're already in a high-Z child, so we've hit a boundary
	    // of the parent cell. Exit.
	    //	    printf("TMP: TTO: already high-Z, exit\n");
	    break;
	  }

	  // We are now in the high-Z child.
	  child_no |= (1 << Z);

	  // Set cell_entry_t to be the entry of this child cell.
	  // Set the t-value of hitting the next Z boundary up by the Z
	  // increment.
	  cell_entry_t = t_next_z;
	  t_next_z += delta_t_cell_z[current_level];

	  // Find the new distances to go until the X, Y and Z boundaries.
	  child_cell_exit[X] -= child_cell_exit[Z] * ray_dir_ratio_x_z;
	  child_cell_exit[Y] -= child_cell_exit[Z] * ray_dir_ratio_y_z;
	  child_cell_exit[Z] = cell_z;
	}
      }

      // Now we know which child cell the ray will hit next. Get this child and
      // recursively traverse it. Note that all of the above calculations
      // are done with absolute values of the distances, etc. and work if
      // the ray is coming from the low-XYZ octant. If this isn't the case,
      // then we are actually indexing a different child to the one found
      // above; we do this by doing a bitwise XOR with the octant that the
      // ray is coming from.
      next_child = children_list[child_no ^ ray_entry_octant];
      
      //      printf("TMP: TTO: octree leaf %d ...\n", child_no ^ ray_entry_octant);
      if (!INDEX_ISEMPTY(next_child)) {
	//	printf("TMP: TTO: nonempty leaf, recursive traverse\n");
	child_result = traverse_trioctree(current_level, next_child,
					  child_cell_exit, t_next_x, t_next_y,
					  t_next_z);
	
	// If we are stopping at the first intersection and we found
	// something [full shadow or a solid intersection], stop.
	if (child_result) {
	  if (stop_at_first_intersection) {
	    return child_result;
	  } else {
	    parent_result = TRUE;
	  }
	}
      } else {
	//	printf("TMP: TTO: empty leaf, skipping\n");
      }
    }
    // We have traversed all of the child cells. This means that either
    // we are set to stop at the first intersection and we didn't find
    // anything, or we are set to continue after intersections and we
    // *may* have found something (will be in parent_result).
    //    printf("TMP: TTO: octree node traversed\n");
    return parent_result;
  } else {
    fatal_error("triMeshPrim::traverse_trioctree", "unknown trioctindex type!",
		"This error should not happen.");
    return FALSE;
  }
}

void triangleMeshPrimitive::tri_ray_intersect(meshtri *t, int *found)
{
  vector edge1, edge2, pv0, utest, vtest, norm;
  double determ, inv_determ, hit_u, hit_v, hit_t;

  //  printf("TMP: TRI in\n");

  Vector::sub(vertex[t->v1].vertex, vertex[t->v0].vertex, edge1);
  Vector::sub(vertex[t->v2].vertex, vertex[t->v0].vertex, edge2);
  
  // Calculate the determinant (utest is also used to calculate U)
  Vector::cross(mesh_ray_dir, edge2, utest);
  determ = Vector::dot(edge1, utest);

  // determ ~= 0 is a ray that lies in the plane of
  // the triangle. No intersection.
  if (fabs(determ) < TINY) {
    //    printf("TMP: zero det, out\n");
    return;
  }
  inv_determ = 1.0 / determ;

  // Find the ray -> triangle vector.
  Vector::sub(mesh_ray_from, vertex[t->v0].vertex, pv0);

  // Calculate the U parameter of the intersection. Test against the
  // triangle's bounds.
  hit_u = inv_determ * Vector::dot(pv0, utest);
  if ((hit_u < 0) || (hit_u > 1)) {
    //    printf("TMP: u %f, out\n", hit_u);
    return;
  }

  // Calculate the V parameter of the intersection. Test against the
  // triangle's bounds.
  Vector::cross(pv0, edge1, vtest);
  hit_v = inv_determ * Vector::dot(mesh_ray_dir, vtest);
  if ((hit_v < 0) || (hit_u + hit_v > 1)) {
    //    printf("TMP: v %f, out\n", hit_u);
    return;
  }

  // Calculate the T value of the intersection.
  hit_t = inv_determ * Vector::dot(edge2, vtest);

  // Insert an entering intersection.
  if (hit_t > MIN_TVAL_CUTOFF) {
    if (phong_shading) {
#if 0
      printf("Apply phong shading\n");
      printf("U, V %f %f\n", hit_u, hit_v);
      printf("normal 0 %f %f %f\n",
	     vertex[t->v0].normal[X],
	     vertex[t->v0].normal[Y],
	     vertex[t->v0].normal[Z]);
      printf("normal 1 %f %f %f\n",
	     vertex[t->v1].normal[X],
	     vertex[t->v1].normal[Y],
	     vertex[t->v1].normal[Z]);
      printf("normal 2 %f %f %f\n",
	     vertex[t->v2].normal[X],
	     vertex[t->v2].normal[Y],
	     vertex[t->v2].normal[Z]);
#endif
      Vector::mult(1 - hit_u - hit_v, vertex[t->v0].normal, norm);
      Vector::addfrac(hit_u,          vertex[t->v1].normal, norm);
      Vector::addfrac(hit_v,          vertex[t->v2].normal, norm);
    } else {
      Vector::set(norm, t->norm);
    }

    insert_entering(the_list, mesh_primnode, validate_primnode,
		    norm, hit_t, mesh_ray_from, mesh_ray_dir,
		    this_ray_flags);
    *found = TRUE;
    //  printf("TMP: hit at u, v, t = %f, %f, %f\n", hit_u, hit_v, hit_t);
  }
}

#define CHECK_FACE_PLANE(point, outcode) {        \
                                                  \
  if (v[point][X] > x_max)                        \
    outcode |= 0x01;                              \
  if (v[point][X] < x_min)                        \
    outcode |= 0x02;                              \
                                                  \
  if (v[point][Y] > y_max)                        \
    outcode |= 0x04;                              \
  if (v[point][Y] < y_min)                        \
    outcode |= 0x08;                              \
                                                  \
  if (v[point][Z] > z_max)                        \
    outcode |= 0x10;                              \
  if (v[point][Z] < z_min)                        \
    outcode |= 0x20;                              \
                                                  \
  if (outcode == 0) {                             \
    return TRUE;                                  \
  }                                               \
}

#define CHECK_BEVEL_PLANES(point, outcode) {                \
                                                            \
  if (  v[point][X] + v[point][Y] >   x_max + y_max)      \
    outcode |= 0x00100L;                                    \
  if (  v[point][X] - v[point][Y] >   x_max - y_min)      \
    outcode |= 0x00200L;                                    \
  if (- v[point][X] + v[point][Y] > - x_min + y_max)      \
    outcode |= 0x00400L;                                    \
  if (- v[point][X] - v[point][Y] > - x_min - y_min)      \
    outcode |= 0x00800L;                                    \
                                                            \
  if (  v[point][X] + v[point][Z] >   x_max + z_max)      \
    outcode |= 0x01000L;                                    \
  if (  v[point][X] - v[point][Z] >   x_max - z_min)      \
    outcode |= 0x02000L;                                    \
  if (- v[point][X] + v[point][Z] > - x_min + z_max)      \
    outcode |= 0x04000L;                                    \
  if (- v[point][X] - v[point][Z] > - x_min - z_min)      \
    outcode |= 0x08000L;                                    \
                                                            \
  if (  v[point][Y] + v[point][Z] >   y_max + z_max)      \
    outcode |= 0x10000L;                                    \
  if (  v[point][Y] - v[point][Z] >   y_max - z_min)      \
    outcode |= 0x20000L;                                    \
  if (- v[point][Y] + v[point][Z] > - y_min + z_max)      \
    outcode |= 0x40000L;                                    \
  if (- v[point][Y] - v[point][Z] > - y_min - z_min)      \
    outcode |= 0x80000L;                                    \
}

#define CHECK_CORNER_PLANES(point, outcode) {                                 \
                                                                              \
  if (  v[point][X] + v[point][Y] + v[point][Z] >   x_max + y_max + z_max) \
    outcode |= 0x01000000UL;                                                  \
  if (  v[point][X] + v[point][Y] - v[point][Z] >   x_max + y_max - z_min) \
    outcode |= 0x02000000UL;                                                  \
  if (  v[point][X] - v[point][Y] + v[point][Z] >   x_max - y_min + z_max) \
    outcode |= 0x04000000UL;                                                  \
  if (  v[point][X] - v[point][Y] - v[point][Z] >   x_max - y_min - z_min) \
    outcode |= 0x08000000UL;                                                  \
  if (- v[point][X] + v[point][Y] + v[point][Z] > - x_min + y_max + z_max) \
    outcode |= 0x10000000UL;                                                  \
  if (- v[point][X] + v[point][Y] - v[point][Z] > - x_min + y_max - z_min) \
    outcode |= 0x20000000UL;                                                  \
  if (- v[point][X] - v[point][Y] + v[point][Z] > - x_min - y_min + z_max) \
    outcode |= 0x40000000UL;                                                  \
  if (- v[point][X] - v[point][Y] - v[point][Z] > - x_min - y_min - z_min) \
    outcode |= 0x80000000UL;                                                  \
}

#define CHECK_POINT(p0,p1,alpha,mask) {                       \
                                                              \
  alphapt[X] = v[p0][X] + alpha * (v[p1][X] - v[p0][X]);      \
  alphapt[Y] = v[p0][Y] + alpha * (v[p1][Y] - v[p0][Y]);      \
  alphapt[Z] = v[p0][Z] + alpha * (v[p1][Z] - v[p0][Z]);      \
                                                              \
  if (mask & 0x01)                                            \
    if (alphapt[X] < x_max)                                  \
      return TRUE;                                            \
  if (mask & 0x02)                                            \
    if (alphapt[X] > x_min)                                  \
      return TRUE;                                            \
                                                              \
  if (mask & 0x04)                                            \
    if (alphapt[Y] < y_max)                                  \
      return TRUE;                                            \
  if (mask & 0x08)                                            \
    if (alphapt[Y] > y_min)                                  \
      return TRUE;                                            \
                                                              \
  if (mask & 0x10)                                            \
    if (alphapt[Z] < z_max)                                  \
      return TRUE;                                            \
  if (mask & 0x20)                                            \
    if (alphapt[Z] > z_min)                                  \
      return TRUE;                                            \
}  

#define CHECK_LINE(p0,p1) {                                       \
                                                                  \
  out_diff = pointout[p0] | pointout[p1];                         \
                                                                  \
  if (out_diff & 0x01)                                            \
    CHECK_POINT(p0,p1,(x_max-v[p0][X])/(v[p1][X]-v[p0][X]),0x3e);\
  if (out_diff & 0x02)                                            \
    CHECK_POINT(p0,p1,(x_min-v[p0][X])/(v[p1][X]-v[p0][X]),0x3d);\
  if (out_diff & 0x04)                                            \
    CHECK_POINT(p0,p1,(y_max-v[p0][Y])/(v[p1][Y]-v[p0][Y]),0x3b);\
  if (out_diff & 0x08)                                            \
    CHECK_POINT(p0,p1,(y_min-v[p0][Y])/(v[p1][Y]-v[p0][Y]),0x37);\
  if (out_diff & 0x10)                                            \
    CHECK_POINT(p0,p1,(z_max-v[p0][Z])/(v[p1][Z]-v[p0][Z]),0x2f);\
  if (out_diff & 0x20)                                            \
    CHECK_POINT(p0,p1,(z_min-v[p0][Z])/(v[p1][Z]-v[p0][Z]),0x1f);\
                                                                  \
}

#define SIGN3D(v) (((v[X] > 0) ? 0x01 : 0x00) | \
                   ((v[Y] > 0) ? 0x02 : 0x00) | \
                   ((v[Z] > 0) ? 0x04 : 0x00))

#define CHECK_POINT_IN_TRIANGLE(p) {                              \
  if ((p[X] > triboundmin[X]) && (p[X] < triboundmax[X]) &&       \
      (p[Y] > triboundmin[Y]) && (p[Y] < triboundmax[Y]) &&       \
      (p[Z] > triboundmin[Z]) && (p[Z] < triboundmax[Z])) {       \
    if (!done_edgevecs) {                                         \
      Vector::sub(v[1], v[0], edge01);                            \
      Vector::sub(v[2], v[1], edge12);                            \
      Vector::sub(v[0], v[2], edge20);                            \
      done_edgevecs = TRUE;                                       \
    }                                                             \
                                                                  \
    Vector::sub(p, v[0], pv0);                                    \
    Vector::sub(p, v[1], pv1);                                    \
    Vector::sub(p, v[2], pv2);                                    \
                                                                  \
    Vector::cross(edge01, pv0, cross0);                           \
    Vector::cross(edge12, pv1, cross1);                           \
    Vector::cross(edge20, pv2, cross2);                           \
                                                                  \
    sign0 = SIGN3D(cross0);                                       \
    sign1 = SIGN3D(cross1);                                       \
    sign2 = SIGN3D(cross2);                                       \
                                                                  \
    if ((sign0 == sign1) && (sign1 == sign2))                     \
      return TRUE;                                                \
  }                                                               \
}

#define CHECK_DIAGONAL(xsign,ysign,zsign,basex,basey,basez) {       \
                                                                    \
  Vector::assign(thisdiag, xsign*diagonal[X], ysign*diagonal[Y],    \
                 zsign*diagonal[Z]);                                \
  Vector::assign(vecdiff, v[0][X] - basex, v[0][Y] - basey,         \
                 v[0][Z] - basez);                                  \
                                                                    \
  totaldot = Vector::dot(thisdiag, norm);                           \
  diffdot  = Vector::dot(vecdiff,  norm);                           \
                                                                    \
  if (fabs(totaldot) > TINY) {                                      \
    intersect_t = diffdot / totaldot;                               \
                                                                    \
    if ((intersect_t > 0) && (intersect_t < 1)) {                   \
      Vector::assign(int_pt,                                        \
                     basex + intersect_t * xsign * diagonal[X],     \
                     basey + intersect_t * ysign * diagonal[Y],     \
                     basez + intersect_t * zsign * diagonal[Z]);    \
                                                                    \
      CHECK_POINT_IN_TRIANGLE(int_pt);                              \
    }                                                               \
  }                                                                 \
}


int triangleMeshPrimitive::tri_box_intersect(int tri_num,
					     double x_min, double y_min,
					     double z_min, double x_max,
					     double y_max, double z_max)
{
  vector v[3], diagonal, alphapt, thisdiag, vecdiff, norm;
  vector int_pt, triboundmin, triboundmax;
  vector edge01, edge12, edge20, pv0, pv1, pv2, cross0, cross1, cross2;
  int done_edgevecs = FALSE, sign0, sign1, sign2;
  double totaldot, diffdot, intersect_t;
  unsigned long pointout[3], out_diff;

  Vector::set(v[0], vertex[tri[tri_num].v0].vertex);
  Vector::set(v[1], vertex[tri[tri_num].v1].vertex);
  Vector::set(v[2], vertex[tri[tri_num].v2].vertex);
  Vector::set(triboundmin, tribound[2*tri_num + 0]);
  Vector::set(triboundmax, tribound[2*tri_num + 1]);

  pointout[0] = 0;
  pointout[1] = 0;
  pointout[2] = 0;

  // Step 1: Check the triangle's points against each face. We may
  //         trivially discard the triangle if every point lies outside
  //         the same face. Also, if any point lies inside the cube, the
  //         triangle is immediately accepted.
  CHECK_FACE_PLANE(0, pointout[0]);
  CHECK_FACE_PLANE(1, pointout[1]);
  CHECK_FACE_PLANE(2, pointout[2]);

#if 0
  printf("TBI: face outcodes %08lX %08lX %08lX (%08lX)\n",
	 pointout[0], pointout[1], pointout[2],
	 pointout[0] & pointout[1] & pointout[2]);
#endif

  if (pointout[0] & pointout[1] & pointout[2]) {
    //    printf("TBI: face plane -> out\n");
    return FALSE;
  }

  // Step 2: Check the triangle's points against each of the 12 beveled
  //         edge planes. We may trivially discard the triangle if every
  //         point lies outside the same edge plane.
  CHECK_BEVEL_PLANES(0, pointout[0]);
  CHECK_BEVEL_PLANES(1, pointout[1]);
  CHECK_BEVEL_PLANES(2, pointout[2]);

#if 0
  printf("TBI: bevel outcodes %08lX %08lX %08lX (%08lX)\n",
	 pointout[0], pointout[1], pointout[2],
	 pointout[0] & pointout[1] & pointout[2]);
#endif

  if (pointout[0] & pointout[1] & pointout[2]) {
    //    printf("TBI: bevel plane -> out\n");
    return FALSE;
  }

  // Step 3: Check the triangle's points against each of the 8 beveled
  //         corner planes. We may trivially discard the triangle if every
  //         point lies outside the same corner plane.
  CHECK_CORNER_PLANES(0, pointout[0]);
  CHECK_CORNER_PLANES(1, pointout[1]);
  CHECK_CORNER_PLANES(2, pointout[2]);

#if 0
  printf("TBI: corner outcodes %08lX %08lX %08lX (%08lX)\n",
	 pointout[0], pointout[1], pointout[2],
	 pointout[0] & pointout[1] & pointout[2]);
#endif

  if (pointout[0] & pointout[1] & pointout[2]) {
    //    printf("TBI: corner plane -> out\n");
    return FALSE;
  }

  // Step 4: If a pair of vertices do not both lie on the same side of
  //         a plane, check the intersection point of this edge with
  //         the cube's plane. If this intersection point is on the cube's
  //         face, return TRUE.
  CHECK_LINE(0,1);
  CHECK_LINE(0,2);
  CHECK_LINE(1,2);

  // Step 5: Check to see whether the triangle intersects any of the diagonals
  //         of the voxel.
  Vector::assign(diagonal, x_max - x_min, y_max - y_min, z_max - z_min);
  Vector::set(norm, tri[tri_num].norm);

  CHECK_DIAGONAL(+1, +1, +1, x_min, y_min, z_min);
  CHECK_DIAGONAL(-1, +1, +1, x_max, y_min, z_min);
  CHECK_DIAGONAL(+1, -1, +1, x_min, y_max, z_min);
  CHECK_DIAGONAL(-1, -1, +1, x_max, y_max, z_min);

  // The triangle's vertices do not lie inside the cube; the diagonals of the
  // cube do not intersect the triangle. Thus the triangle cannot intersect
  // the cube.
  return FALSE;
}

// Undefine local macros.
#undef CHECK_FACE_PLANE
#undef CHECK_BEVEL_PLANES
#undef CHECK_CORNER_PLANES
#undef CHECK_POINT
#undef CHECK_LINE
#undef SIGN3
#undef CHECK_POINT_IN_TRIANGLE
#undef CHECK_DIAGONAL

intersect_result triangleMeshPrimitive::octreemesh_point_intersect(
						primnode *node,
						vector wp,
						PointInfo *info,
						uint ray_flags)
{
  // FIXME: base this on FULL/EMPTY.
  return OUT;
}


#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif
