/* prim_trimesh.h : The triangle mesh primitive
** 
** A triangle mesh primitive is a set of triangles. May be a shell or solid.
**
** Chris Butcher, 7 April 1998
*/

// Vertex for a triangle mesh.
typedef struct _trivertex {
  vector vertex;
  vector normal;
} trivertex;

// Triangle for a triangle mesh.
typedef struct _meshtri {
  unsigned int v0, v1, v2;
  vector norm;
} meshtri;





// An index in the trimesh octree.
typedef unsigned int trioctindex;

// Bitmask to extract the index component of the trioctindex.
#define TRIINDEX_INDEX           0x00FFFFFF

// Bitmask to extract the type component of the index, and the values it
// can have.
#define TRIINDEX_TYPE            (3 << 24)
#define TRIINDEX_EMPTY           (0 << 24)
#define TRIINDEX_FULL            (1 << 24)
#define TRIINDEX_OCTREE          (2 << 24)
#define TRIINDEX_LIST            (3 << 24)

// Macros for deciding the kind of index.
#define INDEX_ISEMPTY(i)         (i == TRIINDEX_EMPTY)
#define INDEX_ISFULL(i)          (i == TRIINDEX_FULL)
#define INDEX_ISOCTREE(i)        ((i & TRIINDEX_TYPE) == TRIINDEX_OCTREE)
#define INDEX_ISLIST(i)          ((i & TRIINDEX_TYPE) == TRIINDEX_LIST)

// Macros for retrieving an octree or trilist node.
#define GET_TRIOCT(i)            (get_oct( i & TRIINDEX_INDEX))
#define GET_TRILIST(i)           (get_list(i & TRIINDEX_INDEX))

// No free triangle octree memory. Sentinel value.
#define NO_FREE_TRIMEM            ~((unsigned int) 0)

// Size of memory blocks.
#define TRIOCT_BLOCKSIZE         10000
#define TRILIST_BLOCKSIZE        10000

// Each trioctindex points to either a trioctree or a trilist.
typedef struct trioctreenode {
  // The eight children of the octree node.
  trioctindex child[8];
} trioctree;

typedef struct _trilist {
  // Index of the triangle.
  unsigned int tri;
  // Index of the next trilist.
  trioctindex nextlist;
} trilist;

// A primitive that contains a list of triangles and performs octree
// subdivision on them.
class triangleMeshPrimitive : public Primitive {
public:
  // Vertices and triangles in the mesh.
  int num_vertices, num_tri;
  trivertex *vertex;
  meshtri *tri;

  // Mesh parameters.
  vector minbound, maxbound, bound_size;
  vector *tribound;
  int max_tris_in_voxel, subdivision_limit, stop_at_first_intersection;
  int phong_shading, dealloc_data;

  /*
   * Octree storage stuff
   */

  // The octree containing triangles.
  trioctindex my_octree;

  // Octree storage.
  struct _octree_storage {
    int blocks_used, items_used, items_freed;
    trioctindex next_free_octreenode;
    trioctree *storage[1000];
  } octree_storage;

  struct _trilist_storage {
    int blocks_used, items_used, items_freed;
    unsigned int next_free_trilist;
    trilist *storage[1000];
  } trilist_storage;

  // Allocate and deallocate octree storage.
  void init_storage(void);
  void dealloc_storage(void);

  // Get a new storage element.
  trioctree *new_trioctree(trioctindex *new_index);
  trilist *new_trilist(trioctindex *new_index);

  // Free a storage element.
  void free_trilist(trioctindex index);
  void free_trioctreenode(trioctindex oct);
  void free_trilistnode(trioctindex list);

  // Access a storage element.
  trioctree *get_oct(trioctindex index);
  trilist *get_list(trioctindex index);


  // The next block of variables are temporary variables used in
  // traversing the ray through the octree structure.
  vector mesh_ray_from, mesh_ray_dir;
  vector *cell_size, exit_value;
  unsigned char ray_entry_octant;
  double ray_dir_ratio_x_y, ray_dir_ratio_x_z, ray_dir_ratio_y_x,
    ray_dir_ratio_y_z, ray_dir_ratio_z_x, ray_dir_ratio_z_y;
  double t_next_x, t_next_y, t_next_z;
  double *delta_t_cell_x, *delta_t_cell_y, *delta_t_cell_z;
  double cell_entry_t;
  primnode *mesh_primnode, *validate_primnode;
  uint this_ray_flags;
  signature this_ray_signature;
  IntersectionList *the_list;

  // Function prototypes
  virtual int ray_intersect(IntersectionList *the_list, 
			    primnode *node, primnode *validate_node,
			    vector ray_from, vector ray_dir, uint ray_flags,
			    signature ray_signature);
  virtual intersect_result point_intersect(primnode *node, vector apoint,
					   PointInfo *info,
					   uint ray_flags);
  virtual int ray_intersect(vector ray_from,
			    vector ray_dir, rayintersect *int_storage);
  virtual intersect_result point_intersect(vector apoint);
  virtual intersect_result voxel_intersect(box v);
  virtual char *prim_name(void);


  // Octree stuff
  void subdivide(void);
  int traverse_trioctree(int current_level, trioctindex index,
			 vector parent_cell_exit, double t_next_x,
			 double t_next_y, double t_next_z);
  trioctindex trim_triangle_list(trilist *list, int *new_numtri,
				 double x_min, double y_min, double z_min,
				 double x_max, double y_max, double z_max);
  trioctindex trioct_subdivide(trioctindex to_subdivide, int num_tris_in_list,
			       int already_subdivided,
			       double x_min, double y_min, double z_min,
			       double x_max, double y_max, double z_max);
  int compress_trioct(trioctindex *index);

  intersect_result octreemesh_point_intersect(primnode *node,
					      vector wp,
					      PointInfo *info,
					      uint ray_flags);
  void tri_ray_intersect(meshtri *t, int *found);
  int tri_box_intersect(int tri_num,
			double x_min, double y_min,
			double z_min, double x_max,
			double y_max, double z_max);
  


  // Create a new triangle mesh.
  triangleMeshPrimitive(int vnum, vector *vert,
			int trinum, int *triangles,
			int subdivision,
			int do_phong_shading = TRUE,
			int stop_at_first = TRUE,
			int tris_in_voxel = 1);
  
  // Create a new triangle mesh (for debugging ONLY!).
  triangleMeshPrimitive(void);
  
  // Deletion method (invoked at the end of each frame).
  virtual ~triangleMeshPrimitive(void);
};


