/* prim_trunccone.c : The truncated cone primitive
** 
** The truncCone primitive defines a single right circular cone
** with radius 1 at the origin, decreasing towards a vertex at (0, 0, 1)
** but truncated by a plane at Z = zval where zval is somewhere in (0, 1].
**
** Chris Butcher, 31 July 1997
*/

#include "mirage.h"


#ifdef USE_NAMESPACES
namespace Mirage {
#endif

int truncConePrimitive::ray_intersect(IntersectionList *the_list,
				      primnode *node, primnode *validate_node, 
				      vector ray_from,
				      vector ray_dir, uint ray_flags,
				      signature ray_signature)
{
  // The coordinates of ray_from in primitive space
  double p_x, p_y, p_z;
  // The direction ray_dir in primitive space
  double d_x, d_y, d_z;
  // The primitive-space hitpoint
  double hit_x, hit_y, hit_z;
  // The coefficients of the quadratic equation in t
  double a, b, c;
  // The discriminant of this equation
  double disc;
  // Have we found an intersection?
  int found = 0;
  // The t-values of the intersection points with the cylinder.
  double t_0, t_1;
  // The t_values of the intersection points with the top and bottom
  // bounding planes.
  double t_base, t_end;
  // t_lower holds the lesser of these two t values, t_upper holds the
  // larger of them.
  double t_upper, t_lower;
  // Are the intersections with the cone's surface between the planes?
  int valid_intersect_0, valid_intersect_1;
  // The normal at the hitpoint.
  vector norm;
  // Which plane was struck first? TRUE = base, FALSE = end.
  int hit_base_plane_first;

  // Update the last ray that we've intersected with.
  node->last_ray_signature = ray_signature;

  // This apply is an inline function defined in matrix.h
  // Transform the ray startpoint into primitive space
  Matrix::apply(node->position[AFT], ray_from, &p_x, &p_y, &p_z);

  // This apply is also an inline function defined in matrix.h
  // Transform the ray direction into primitive space (note: shift does
  // not apply to this, only apply rotation + stretch sections of the matrix)
  Matrix::applynoshift(node->position[AFT], ray_dir, &d_x, &d_y, &d_z);

  // Check to see if the ray is either above the top and pointed up, or
  // below the bottom and pointed down.
  if (d_z >= 0) {
    if (p_z > zval) {
      // Ray will go above the cone's end.
      node->last_ray_hit = FALSE;
      return node->last_ray_hit;
    }
  } else {
    if (p_z < 0) {
      // Ray will go below the cone's base.
      node->last_ray_hit = FALSE;
      return node->last_ray_hit;
    }
  }

  // Form the coefficients of the quadratic equation that defines t for the
  // cone with a radius 1 base at Z = 0 and vertex at Z = 1.
  //
  // sqrt(x^2 + y^2) + z = 1
  //
  // (px + tdx)^2 + (py + tdy)^2 = (1 - (pz + tdz))^2
  //
  a = SQR(d_x) + SQR(d_y) - SQR(d_z);
  b = 2.0 * (p_x*d_x + p_y*d_y - p_z*d_z + d_z);
  c = SQR(p_x) + SQR(p_y) - SQR(p_z) + 2*p_z - 1;

  // Calculate the discriminant of the quadratic equation.
  disc = SQR(b) - 4.0*a*c;

  // If the discriminant is less than or equal to zero, then there
  // are no intersections with the cone (tangent rays are treated as
  // no intersection at all).
  if (disc <= 0) {
    node->last_ray_hit = FALSE;
    return node->last_ray_hit;
  }

  // a is equal to the square of the magnitude of the direction of the ray.
  // This *must* not be zero.
  if (a == 0)
    fatal_error("truncConePrimitive::ray_intersect", "ray direction zero",
		"This error should not happen.");

  // Calculate the intersection t-values. Must have t_0 < t_1 for later
  // calculations to work.
  // Note that we can save a square root operation by saying
  // t_1 = c / a * t_0 ... but this only works if we can be sure that
  // t_0 isn't zero. It may easily be if sqrt(disc) == b or == -b.
  // So, we make sure that the +/-sqrt(disc) is always the same sign
  // as -b to make sure that the subtraction can't produce zero.
  // This makes the code a little long-winded.
  if (a >= 0) {
    if (b >= 0) {
      t_0 = (-b - sqrt(disc)) / (2 * a);
      t_1 = c / (a * t_0);
    } else {
      t_1 = (-b + sqrt(disc)) / (2 * a);
      t_0 = c / (a * t_1);
    }
  } else {
    if (b >= 0) {
      t_1 = (-b - sqrt(disc)) / (2 * a);
      t_0 = c / (a * t_1);
    } else {
      t_0 = (-b + sqrt(disc)) / (2 * a);
      t_1 = c / (a * t_0);
    }
  }

  // Rays that are parallel to the planes bounding the top and bottom
  // of the cylinder don't need to be tested against
  // them, just insert the t_0 and t_1 intersections directly.
  if (fabs(d_z) < TINY) {

    // If the ray is outside the extent of the cone, no intersections.
    if (p_z < 0 || p_z > zval) {
      node->last_ray_hit = FALSE;
      return node->last_ray_hit;
    }

    // Insert any positive t-values (valid intersections).
    if (t_0 > MIN_TVAL_CUTOFF) {
      // Calculate the hitpoint and normal of the cone.
      hit_x = p_x + d_x*t_0;
      hit_y = p_y + d_y*t_0;
      hit_z = p_z + d_z*t_0;
	
      norm[X] = hit_x * (1 - hit_z);
      norm[Y] = hit_y * (1 - hit_z);
      norm[Z] = SQR(hit_x) + SQR(hit_y);

      // Note that since t_0 < t_1 always, this intersection is an ENTERING
      // one.
      insert_entering(the_list, node, validate_node, norm, t_0,
		      ray_from, ray_dir, ray_flags);

      found = 1;
    }
    if (t_1 > MIN_TVAL_CUTOFF) {
      // Calculate the hitpoint and normal of the cone.
      hit_x = p_x + d_x*t_1;
      hit_y = p_y + d_y*t_1;
      hit_z = p_z + d_z*t_1;
	
      norm[X] = hit_x * (1 - hit_z);
      norm[Y] = hit_y * (1 - hit_z);
      norm[Z] = SQR(hit_x) + SQR(hit_y);

      // Note that since t_0 < t_1 always, this intersection is a LEAVING
      // one.
      insert_leaving(the_list, node, validate_node, norm, t_1,
		     ray_from, ray_dir, ray_flags);
      found = 1;
    }
    
  } else {
    // Ray isn't parallel to the planes, must be checked against them.

    // Calculate the t-values that correspond to the ray hitting the top
    // and bottom planes.
    t_end  = (zval - p_z) / d_z;
    t_base = - p_z / d_z;

    // Which one is struck first?
    if (t_base < t_end) {
      // The ray hits the base plane before the end plane
      // thus t_upper is the end plane, t_lower is the base plane
      hit_base_plane_first = TRUE;
      t_upper = t_end;
      t_lower = t_base;
    } else {
      // The ray hits the end plane before the bottom plane
      // thus t_upper is the base plane, t_lower is the end plane
      hit_base_plane_first = FALSE;
      t_upper = t_base;
      t_lower = t_end;
    }

    // Does the intersection with t_0 lie within the planes?
    if (t_0 > t_lower && t_0 < t_upper)
      valid_intersect_0 = TRUE;
    else
      valid_intersect_0 = FALSE;

    // Does the intersection with t_1 lie within the planes?
    if (t_1 > t_lower && t_1 < t_upper)
      valid_intersect_1 = TRUE;
    else
      valid_intersect_1 = FALSE;

    // Insert the intersection with the base plane if it occurs between
    // t_0 and t_1 (the cone's extent).
    if ((t_base > MIN_TVAL_CUTOFF) && (t_base > t_0) && (t_base < t_1)) {
      Vector::assign(norm, 0, 0, -1);
      if (hit_base_plane_first)
	insert_entering(the_list, node, validate_node, norm, t_base,
			ray_from, ray_dir, ray_flags);
      else
	insert_leaving(the_list, node, validate_node, norm, t_base,
		       ray_from, ray_dir, ray_flags);
      found = 1;
    }

    // Insert the intersection with the end plane if it occurs between
    // t_0 and t_1 (the cone's extent).
    if ((t_end > MIN_TVAL_CUTOFF) && (t_end > t_0) && (t_end < t_1)) {
      Vector::assign(norm, 0, 0, 1);
      if (!hit_base_plane_first)
	insert_entering(the_list, node, validate_node, norm, t_end,
			ray_from, ray_dir, ray_flags);
      else
	insert_leaving(the_list, node, validate_node, norm, t_end,
		       ray_from, ray_dir, ray_flags);

      found = 1;
    }

    // Insert any t-values that are positive and between t_lower and t_upper.
    if (t_0 > MIN_TVAL_CUTOFF && valid_intersect_0) {
      // Calculate the hitpoint and normal of the cone.
      hit_x = p_x + d_x*t_0;
      hit_y = p_y + d_y*t_0;
      hit_z = p_z + d_z*t_0;
	
      norm[X] = hit_x * (1 - hit_z);
      norm[Y] = hit_y * (1 - hit_z);
      norm[Z] = SQR(hit_x) + SQR(hit_y);

      // Note that since t_0 < t_1 always, this intersection is an ENTERING
      // one.
      insert_entering(the_list, node, validate_node, norm, t_0,
		      ray_from, ray_dir, ray_flags);
      found = 1;
    }
    if (t_1 > MIN_TVAL_CUTOFF && valid_intersect_1) {
      // Calculate the hitpoint and normal of the cone.
      hit_x = p_x + d_x*t_1;
      hit_y = p_y + d_y*t_1;
      hit_z = p_z + d_z*t_1;
	
      norm[X] = hit_x * (1 - hit_z);
      norm[Y] = hit_y * (1 - hit_z);
      norm[Z] = SQR(hit_x) + SQR(hit_y);

      // Note that since t_0 < t_1 always, this intersection is a LEAVING
      // one.
      insert_leaving(the_list, node, validate_node, norm, t_1,
		     ray_from, ray_dir, ray_flags);
      found = 1;
    }
  }

  if (found)
    node->last_ray_hit = TRUE;
  else
    node->last_ray_hit = FALSE;

  return node->last_ray_hit;
}

int truncConePrimitive::ray_intersect(vector ray_from, vector ray_dir,
				      rayintersect *int_storage)
{
  // The primitive-space hitpoint
  double hit_x, hit_y, hit_z;
  // The coefficients of the quadratic equation in t
  double a, b, c;
  // The discriminant of this equation
  double disc;
  // Have we found an intersection?
  int num_intersections = 0;
  // The t-values of the intersection points with the cylinder.
  double t_0, t_1;
  // The t_values of the intersection points with the top and bottom
  // bounding planes.
  double t_base, t_end;
  // t_lower holds the lesser of these two t values, t_upper holds the
  // larger of them.
  double t_upper, t_lower;
  // Are the intersections with the cone's surface between the planes?
  int valid_intersect_0, valid_intersect_1;
  // The normal at the hitpoint.
  vector norm;
  // Which plane was struck first? TRUE = base, FALSE = end.
  int hit_base_plane_first;

  // Check to see if the ray is either above the top and pointed up, or
  // below the bottom and pointed down.
  if (ray_dir[Z] >= 0) {
    if (ray_from[Z] > zval) {
      // Ray will go above the cone's end.
      return 0;
    }
  } else {
    if (ray_from[Z] < 0) {
      // Ray will go below the cone's base.
      return 0;
    }
  }

  // Form the coefficients of the quadratic equation that defines t for the
  // cone with a radius 1 base at Z = 0 and vertex at Z = 1.
  //
  // sqrt(x^2 + y^2) + z = 1
  //
  // (px + tdx)^2 + (py + tdy)^2 = (1 - (pz + tdz))^2
  //
  a = SQR(ray_dir[X]) + SQR(ray_dir[Y]) - SQR(ray_dir[Z]);
  b = 2.0 * (ray_from[X]*ray_dir[X] + ray_from[Y]*ray_dir[Y]
	     - ray_from[Z]*ray_dir[Z] + ray_dir[Z]);
  c = SQR(ray_from[X]) + SQR(ray_from[Y]) - SQR(ray_from[Z])
    + 2*ray_from[Z] - 1;

  // Calculate the discriminant of the quadratic equation.
  disc = SQR(b) - 4.0*a*c;

  // If the discriminant is less than or equal to zero, then there
  // are no intersections with the cone (tangent rays are treated as
  // no intersection at all).
  if (disc <= 0) {
    return 0;
  }

  // a is equal to the square of the magnitude of the direction of the ray.
  // This *must* not be zero.
  if (a == 0)
    fatal_error("truncConePrimitive::ray_intersect", "ray direction zero",
		"This error should not happen.");

  // Calculate the intersection t-values. Must have t_0 < t_1 for later
  // calculations to work.
  // Note that we can save a square root operation by saying
  // t_1 = c / a * t_0 ... but this only works if we can be sure that
  // t_0 isn't zero. It may easily be if sqrt(disc) == b or == -b.
  // So, we make sure that the +/-sqrt(disc) is always the same sign
  // as -b to make sure that the subtraction can't produce zero.
  // This makes the code a little long-winded.
  if (a >= 0) {
    if (b >= 0) {
      t_0 = (-b - sqrt(disc)) / (2 * a);
      t_1 = c / (a * t_0);
    } else {
      t_1 = (-b + sqrt(disc)) / (2 * a);
      t_0 = c / (a * t_1);
    }
  } else {
    if (b >= 0) {
      t_1 = (-b - sqrt(disc)) / (2 * a);
      t_0 = c / (a * t_1);
    } else {
      t_0 = (-b + sqrt(disc)) / (2 * a);
      t_1 = c / (a * t_0);
    }
  }

  // Rays that are parallel to the planes bounding the top and bottom
  // of the cylinder don't need to be tested against
  // them, just insert the t_0 and t_1 intersections directly.
  if (fabs(ray_dir[Z]) < TINY) {

    // If the ray is outside the extent of the cone, no intersections.
    if (ray_from[Z] < 0 || ray_from[Z] > zval) {
      return 0;
    }

    // Calculate the hitpoint and normal of the cone.
    hit_x = ray_from[X] + ray_dir[X]*t_0;
    hit_y = ray_from[Y] + ray_dir[Y]*t_0;
    hit_z = ray_from[Z] + ray_dir[Z]*t_0;
    
    norm[X] = hit_x * (1 - hit_z);
    norm[Y] = hit_y * (1 - hit_z);
    norm[Z] = SQR(hit_x) + SQR(hit_y);
    
    // Note that since t_0 < t_1 always, this intersection is an ENTERING
    // one.
    int_storage[num_intersections].t = t_0;
    int_storage[num_intersections].is_entering = TRUE;
    Vector::set(int_storage[num_intersections].normal, norm);
    num_intersections++;

    // Calculate the hitpoint and normal of the cone.
    hit_x = ray_from[X] + ray_dir[X]*t_1;
    hit_y = ray_from[Y] + ray_dir[Y]*t_1;
    hit_z = ray_from[Z] + ray_dir[Z]*t_1;
    
    norm[X] = hit_x * (1 - hit_z);
    norm[Y] = hit_y * (1 - hit_z);
    norm[Z] = SQR(hit_x) + SQR(hit_y);
    
    // Note that since t_0 < t_1 always, this intersection is a LEAVING
    // one.
    int_storage[num_intersections].t = t_1;
    int_storage[num_intersections].is_entering = FALSE;
    Vector::set(int_storage[num_intersections].normal, norm);
    num_intersections++;
    
  } else {
    // Ray isn't parallel to the planes, must be checked against them.

    // Calculate the t-values that correspond to the ray hitting the top
    // and bottom planes.
    t_end  = (zval - ray_from[Z]) / ray_dir[Z];
    t_base = - ray_from[Z] / ray_dir[Z];

    // Which one is struck first?
    if (t_base < t_end) {
      // The ray hits the base plane before the end plane
      // thus t_upper is the end plane, t_lower is the base plane
      hit_base_plane_first = TRUE;
      t_upper = t_end;
      t_lower = t_base;
    } else {
      // The ray hits the end plane before the bottom plane
      // thus t_upper is the base plane, t_lower is the end plane
      hit_base_plane_first = FALSE;
      t_upper = t_base;
      t_lower = t_end;
    }

    // Does the intersection with t_0 lie within the planes?
    if (t_0 > t_lower && t_0 < t_upper)
      valid_intersect_0 = TRUE;
    else
      valid_intersect_0 = FALSE;

    // Does the intersection with t_1 lie within the planes?
    if (t_1 > t_lower && t_1 < t_upper)
      valid_intersect_1 = TRUE;
    else
      valid_intersect_1 = FALSE;

    // Insert the intersection with the base plane if it occurs between
    // t_0 and t_1 (the cone's extent).
    if ((t_base > t_0) && (t_base < t_1)) {
      int_storage[num_intersections].t = t_base;
      int_storage[num_intersections].is_entering = hit_base_plane_first;
      Vector::assign(int_storage[num_intersections].normal, 0, 0, -1);
      num_intersections++;
    }

    // Insert the intersection with the end plane if it occurs between
    // t_0 and t_1 (the cone's extent).
    if ((t_end > t_0) && (t_end < t_1)) {
      int_storage[num_intersections].t = t_end;
      int_storage[num_intersections].is_entering = (!hit_base_plane_first);
      Vector::assign(int_storage[num_intersections].normal, 0, 0, 1);
      num_intersections++;
    }

    // Insert any t-values that are positive and between t_lower and t_upper.
    if (valid_intersect_0) {
      // Calculate the hitpoint and normal of the cone.
      hit_x = ray_from[X] + ray_dir[X]*t_0;
      hit_y = ray_from[Y] + ray_dir[Y]*t_0;
      hit_z = ray_from[Z] + ray_dir[Z]*t_0;
	
      norm[X] = hit_x * (1 - hit_z);
      norm[Y] = hit_y * (1 - hit_z);
      norm[Z] = SQR(hit_x) + SQR(hit_y);

      // Note that since t_0 < t_1 always, this intersection is an ENTERING
      // one.
      int_storage[num_intersections].t = t_0;
      int_storage[num_intersections].is_entering = TRUE;
      Vector::set(int_storage[num_intersections].normal, norm);
      num_intersections++;
    }
    if (valid_intersect_1) {
      // Calculate the hitpoint and normal of the cone.
      hit_x = ray_from[X] + ray_dir[X]*t_1;
      hit_y = ray_from[Y] + ray_dir[Y]*t_1;
      hit_z = ray_from[Z] + ray_dir[Z]*t_1;
	
      norm[X] = hit_x * (1 - hit_z);
      norm[Y] = hit_y * (1 - hit_z);
      norm[Z] = SQR(hit_x) + SQR(hit_y);

      // Note that since t_0 < t_1 always, this intersection is a LEAVING
      // one.
      int_storage[num_intersections].t = t_1;
      int_storage[num_intersections].is_entering = FALSE;
      Vector::set(int_storage[num_intersections].normal, norm);
      num_intersections++;
    }
  }

  return num_intersections;
}

// Return IN or OUT depending on whether or not a given point is inside
// the cone.
intersect_result truncConePrimitive::point_intersect(primnode *node,
						     vector wp,
						     PointInfo *info,
						     uint ray_flags)
{
  double p_x, p_y, p_z;

  // Get the primitive space coordinates of the world-space point wp.
  Matrix::apply(node->position[AFT], wp, &p_x, &p_y, &p_z);
  
  // If (x^2 + y^2 < z^2) then the point is in the right circular cone.
  // The point must also be bounded by Z between 0 and +1.
  if (p_z > 0 && p_z < zval && SQR(p_x) + SQR(p_y) < SQR(1 - p_z)) {
    set_point_contents(node, info);
    return IN;
  } else
    return OUT;
}

// Return IN or OUT depending on whether or not a given point is inside
// the cone.
intersect_result truncConePrimitive::point_intersect(vector wp)
{
  // If (x^2 + y^2 < z^2) then the point is in the right circular cone.
  // The point must also be bounded by Z between 0 and +1.
  if (wp[Z] > 0 && wp[Z] < zval && SQR(wp[X]) + SQR(wp[Y]) < SQR(1 - wp[Z]))
    return IN;
  else
    return OUT;
}

intersect_result truncConePrimitive::voxel_intersect(box b)
{
  // This routine checks to see if a parallelipiped lies inside, outside,
  // or intersects the surface of the unit cone.
  //
  // It does so by calling the cone primitive's intersection routine
  // and checking the bounding planes manually.

  int i, above_top, in_middle, below_bottom;
  intersect_result cone_result, planes_result;

  // Check the points to see which side of the two planes they lie on.
  above_top = below_bottom = in_middle = FALSE;
  planes_result = IN;

  for (i = 0; i < 8; i++) {
    if (b[i][Z] < 0) {
      below_bottom = TRUE;
      if (in_middle || above_top) {
	planes_result = SURFACE;
	break;
      }
    } else if (b[i][Z] > zval) {
      above_top = TRUE;
      if (in_middle || below_bottom) {
	planes_result = SURFACE;
	break;
      }
    } else {
      in_middle = TRUE;
      if (above_top || below_bottom) {
	planes_result = SURFACE;
	break;
      }
    }
  }

  if (i == 8) {
    // We have checked them all without finding two points on opposite
    // sides of a plane.
    if (in_middle)
      planes_result = IN;
    else
      planes_result = OUT;
  }

  // If the box is outside the bounding planes, return OUT.
  if (planes_result == OUT)
    return OUT;
  
  // The points (x, y, z) relative to the upside-down shifted cone are
  // (x, y, 1 - z) relative to the conePrimitive's primitive space.
  //
  // Calculate where the voxel is in the cone's primitive space and
  // perform a voxel-intersection test with it.
  box newbox;

  for (i = 0; i < 8; i++) {
    newbox[i][X] = b[i][X];
    newbox[i][Y] = b[i][Y];
    newbox[i][Z] = 1 - b[i][Z];
  }
  cone_result = cone_voxel_intersect(newbox);

  // If the box is outside the bounding cone, return OUT.
  if (cone_result == OUT)
    return OUT;

  // If we are inside the cone, return IN / SURFACE depending on where the
  // box is relative to the planes.
  if (cone_result == IN) {
    return planes_result;
  } else  {
    // We are on the surface of the cone and we know we are not
    // outside the planes, so return SURFACE.
    return SURFACE;
  }
}

char *truncConePrimitive::prim_name(void)
{
  return "Truncated Cone";
}


#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif
