/* prim_twosidedpatch.c : The two-sided bilinear-patch primitive
** 
** The two-sided bilinear patch is defined by the corner points
** (0, 0, 0), (1, 0, 0), (0, 1, 0), (x_point, y_point, 1). Its normal always
** points toward +Z, rather than back towards the direction of the ray
** that hit the patch. This means that it can be given a material property
** or a texture that makes it look different from the underside.
**
** Surprise, surprise, this makes my life easier when making leaves for
** my tree.
**
** Chris Butcher, 21 August 1997
*/

#include "mirage.h"

#ifdef USE_NAMESPACES
namespace Mirage {
#endif

int twoSidedPatchPrimitive::ray_intersect(IntersectionList *the_list,
					  primnode *node, 
					  primnode *validate_node, 
					  vector ray_from, vector ray_dir,
					  uint ray_flags,
					  signature ray_signature)
{
  // The coordinates of ray_from in primitive space
  double p_x, p_y, p_z;
  // The direction ray_dir in primitive space
  vector dir;
  // Variables used to form the quadratic equation
  double qx, qy, c_dx, c_dy, c_px, c_py;
  // The coefficients of the quadratic equation in t
  double a, b, c;
  // The discriminant of this equation
  double disc;
  // Have we found an intersection?
  int found = FALSE;
  // The t-values of the intersection points.
  double t_0, t_1;
  // Two vectors along the surface at the hitpoint; dp/ds1 and dp/ds2.
  vector n_s1, n_s2;
  // The normal at the hitpoint.
  vector norm;
  // The points due N, S, E and W of the hitpoint on the patch's edges.
  double n_x, n_y, n_z, s_x, s_y, s_z, e_x, e_y, e_z, w_x, w_y, w_z;
  // Parameters that are used in calculating the normal and hitpoint.
  double hit_z, hit_s1, hit_s2;

  // Store the signature of this ray that we're intersecting with.
  node->last_ray_signature = ray_signature;

  // This apply is an inline function defined in matrix.h
  // Transform the ray startpoint into primitive space
  Matrix::apply(node->position[AFT], ray_from, &p_x, &p_y, &p_z);

  // This apply is also an inline function defined in matrix.h
  // Transform the ray direction into primitive space (note: shift does
  // not apply to this, only apply rotation + stretch sections of the matrix)
  Matrix::applynoshift(node->position[AFT], ray_dir, dir);

  // The patch may be flat. If it is then the equations are different.
  if (is_flat) {
    // The patch is flat. Thus there is only one possible intersection.
    //
    // The intersection is defined as t =  - p_z / dir[Z].
    if (dir[Z] == 0) {
      // No intersections.
      node->last_ray_hit = FALSE;
      return node->last_ray_hit;
    }
    t_0 = - p_z / dir[Z];
    t_1 = -INFINITY; // Not a real intersection; won't be considered.

    qx = 1 - x_point;
    qy = 1 - y_point;
    
  } else {
    // The patch is not flat. Use the normal equations.

    // Form the coefficients of the quadratic equation that defines t.
    // This equation is given by:
    //
    //  qx = 1 - x_point
    //  qy = 1 - y_point
    //
    //  c_dx = dx + qx * dz
    //  c_dy = dy + qy * dz
    //
    //  c_px = px + qx * pz
    //  c_py = py + qy * pz
    //
    //  a = c_dx * c_dy
    //  b = c_dx * c_py + c_dy * c_px - dz
    //  c = c_px * c_py - pz
    
    qx = 1 - x_point;
    qy = 1 - y_point;
    
    c_dx = dir[X] + qx * dir[Z];
    c_dy = dir[Y] + qy * dir[Z];
    
    c_px = p_x + qx * p_z;
    c_py = p_y + qy * p_z;
    
    a = c_dx * c_dy;
    b = c_dx * c_py + c_dy * c_px - dir[Z];
    c = c_px * c_py - p_z;
    
    if (a == 0) {
      // This equation reduces to a linear equation, bt + c = 0.
      // So t = - c / b.
      if (b == 0) {
	// There are no solutions, no intersections.
	node->last_ray_hit = FALSE;
	return node->last_ray_hit;
      }
      
      t_0 = - c / b;
      t_1 = -INFINITY; // Not a real intersection; won't be considered.
    } else {
      
      // Calculate the discriminant of the quadratic equation.
      disc = SQR(b) - 4.0*a*c;
      
      // If the discriminant is less than or equal to zero, then there
      // are no intersections with the bilinear patch (tangent rays are
      // treated as no intersection at all).
      if (disc <= 0) {
	node->last_ray_hit = FALSE;
	return node->last_ray_hit;
      }
      
      // Calculate the intersection t-values. Must have t_0 < t_1 for later
      // calculations to work.
      // Note that we can save a square root operation by saying
      // t_1 = c / a * t_0 ... but this only works if we can be sure that
      // t_0 isn't zero. It may easily be if sqrt(disc) == b or == -b.
      // So, we make sure that the +/-sqrt(disc) is always the same sign
      // as -b to make sure that the subtraction can't produce zero.
      // This makes the code a little long-winded.
      if (a >= 0) {
	if (b >= 0) {
	  t_0 = (-b - sqrt(disc)) / (2 * a);
	  t_1 = c / (a * t_0);
	} else {
	  t_1 = (-b + sqrt(disc)) / (2 * a);
	  t_0 = c / (a * t_1);
	}
      } else {
	if (b >= 0) {
	  t_1 = (-b - sqrt(disc)) / (2 * a);
	  t_0 = c / (a * t_1);
	} else {
	  t_0 = (-b + sqrt(disc)) / (2 * a);
	  t_1 = c / (a * t_0);
	}
      }
    }
  }

  // Insert any positive t-values (valid intersections).
  if (t_0 > 0) {
    hit_z = p_z + t_0*dir[Z];

    // Get the value of s1 (the parameter from point 0 to point 1).
    hit_s1 = (p_x + t_0*dir[X]) + hit_z*qx;

    if ((hit_s1 > 0) && (hit_s1 < 1)) {
      // Get the value of s2 (the parameter from point 0 to point 2).
      hit_s2 = (p_y + t_0*dir[Y]) + hit_z*qy;

      if ((hit_s2 > 0) && (hit_s2 < 1)) {

	// First check to see if the hitpoint is within the Cutout
	// that determines our outline (if we have a Cutout).
	if ((cutout == NULL) || (cutout->is_inside(hit_s1, hit_s2))) {

	  // Find the normal.
	  if (is_flat) {
	    Vector::assign(norm, 0, 0, 1);
	  } else {
	    // Find the east and west points that correspond to the hitpoint.
	    // We do this by interpolating between p1 and p3 for e and
	    // p0 and p2 for w.
	    e_x = 1 * (1 - hit_s2) + x_point * hit_s2;
	    e_y = 0 * (1 - hit_s2) + y_point * hit_s2;
	    e_z = 0 * (1 - hit_s2) +       1 * hit_s2;
	    
	    w_x = 0 * (1 - hit_s2) +       0 * hit_s2;
	    w_y = 0 * (1 - hit_s2) +       1 * hit_s2;
	    w_z = 0 * (1 - hit_s2) +       0 * hit_s2;
	    
	    // Find the north and south points that correspond to the hitpoint.
	    // We do this by interpolating between p2 and p3 for n and
	    // p0 and p1 for s.
	    n_x = 0 * (1 - hit_s1) + x_point * hit_s1;
	    n_y = 1 * (1 - hit_s1) + y_point * hit_s1;
	    n_z = 0 * (1 - hit_s1) +       1 * hit_s1;
	    
	    s_x = 0 * (1 - hit_s2) +       1 * hit_s2;
	    s_y = 0 * (1 - hit_s2) +       0 * hit_s2;
	    s_z = 0 * (1 - hit_s2) +       0 * hit_s2;
	    
	    // Now we construct the vectors n_s1 and n_s2. These are
	    // n_s1 = (dx/ds1 dy/ds1 dz/ds1) and n_s2 = (dx/ds2 dy/ds2 dz/ds2)
	    //
	    // These give us two directions along the surface at the hitpoint.
	    // We take n_s1 x n_s2 = norm.
	    
	    Vector::assign(n_s1, e_x - w_x, e_y - w_y, e_z - w_z);
	    Vector::assign(n_s2, n_x - s_x, n_y - s_y, n_z - s_z);
	    
	    Vector::cross(n_s1, n_s2, norm);
	  }

	  if (fabs(t_0) > MIN_TVAL_CUTOFF)
	    t_0 -= MIN_TVAL_CUTOFF * 0.1;

	  insert_entering(the_list, node, validate_node, norm, t_0,
			  ray_from, ray_dir, ray_flags);
	  
	  found = TRUE;
	}
      }
    }
  }

  if (t_1 > 0) {
    hit_z = p_z + t_1*dir[Z];

    // Get the value of s1 (the parameter from point 0 to point 1).
    hit_s1 = (p_x + t_1*dir[X]) + hit_z*qx;

    if ((hit_s1 > 0) && (hit_s1 < 1)) {
      // Get the value of s2 (the parameter from point 0 to point 2).
      hit_s2 = (p_y + t_1*dir[Y]) + hit_z*qy;

      if ((hit_s2 > 0) && (hit_s2 < 1)) {

	// First check to see if the hitpoint is within the Cutout
	// that determines our outline (if we have a Cutout).
	if ((cutout == NULL) || (cutout->is_inside(hit_s1, hit_s2))) {

	  // Find the normal.
	  if (is_flat) {
	    Vector::assign(norm, 0, 0, 1);
	  } else {
	    // Find the east and west points that correspond to the hitpoint.
	    // We do this by interpolating between p1 and p3 for e and
	    // p0 and p2 for w.
	    e_x = 1 * (1 - hit_s2) + x_point * hit_s2;
	    e_y = 0 * (1 - hit_s2) + y_point * hit_s2;
	    e_z = 0 * (1 - hit_s2) +       1 * hit_s2;
	    
	    w_x = 0 * (1 - hit_s2) +       0 * hit_s2;
	    w_y = 0 * (1 - hit_s2) +       1 * hit_s2;
	    w_z = 0 * (1 - hit_s2) +       0 * hit_s2;
	    
	    // Find the north and south points that correspond to the hitpoint.
	    // We do this by interpolating between p2 and p3 for n and
	    // p0 and p1 for s.
	    n_x = 0 * (1 - hit_s1) + x_point * hit_s1;
	    n_y = 1 * (1 - hit_s1) + y_point * hit_s1;
	    n_z = 0 * (1 - hit_s1) +       1 * hit_s1;
	    
	    s_x = 0 * (1 - hit_s2) +       1 * hit_s2;
	    s_y = 0 * (1 - hit_s2) +       0 * hit_s2;
	    s_z = 0 * (1 - hit_s2) +       0 * hit_s2;
	    
	    // Now we construct the vectors n_s1 and n_s2. These are
	    // n_s1 = (dx/ds1 dy/ds1 dz/ds1) and n_s2 = (dx/ds2 dy/ds2 dz/ds2)
	    //
	    // These give us two directions along the surface at the hitpoint.
	    // We take n_s1 x n_s2 = norm.
	    
	    Vector::assign(n_s1, e_x - w_x, e_y - w_y, e_z - w_z);
	    Vector::assign(n_s2, n_x - s_x, n_y - s_y, n_z - s_z);
	    
	    Vector::cross(n_s1, n_s2, norm);
	  }
	  
	  if (fabs(t_1) > MIN_TVAL_CUTOFF)
	    t_1 -= MIN_TVAL_CUTOFF * 0.1;
	  
	  insert_entering(the_list, node, validate_node, norm, t_1,
			  ray_from, ray_dir, ray_flags);
	  
	  found = TRUE;
	}
      }
    }
  }

  node->last_ray_hit = found;
  return node->last_ray_hit;
}

int twoSidedPatchPrimitive::ray_intersect(vector ray_from, vector ray_dir,
					  rayintersect *int_storage)
{
  // Variables used to form the quadratic equation
  double qx, qy, c_dx, c_dy, c_px, c_py;
  // The coefficients of the quadratic equation in t
  double a, b, c;
  // The discriminant of this equation
  double disc;
  // Have we found an intersection?
  int num_intersections = 0;
  // The t-values of the intersection points.
  double t_0, t_1;
  // Two vectors along the surface at the hitpoint; dp/ds1 and dp/ds2.
  vector n_s1, n_s2;
  // The normal at the hitpoint.
  vector norm;
  // The points due N, S, E and W of the hitpoint on the patch's edges.
  double n_x, n_y, n_z, s_x, s_y, s_z, e_x, e_y, e_z, w_x, w_y, w_z;
  // Parameters that are used in calculating the normal and hitpoint.
  double hit_z, hit_s1, hit_s2;

  // The patch may be flat. If it is then the equations are different.
  if (is_flat) {
    // The patch is flat. Thus there is only one possible intersection.
    //
    // The intersection is defined as t =  - ray_from[Z] / ray_dir[Z].
    if (ray_dir[Z] == 0) {
      // No intersections.
      return 0;
    }
    t_0 = - ray_from[Z] / ray_dir[Z];
    t_1 = -INFINITY; // Not a real intersection; won't be considered.

    qx = 1 - x_point;
    qy = 1 - y_point;
    
  } else {
    // The patch is not flat. Use the normal equations.

    // Form the coefficients of the quadratic equation that defines t.
    // This equation is given by:
    //
    //  qx = 1 - x_point
    //  qy = 1 - y_point
    //
    //  c_dx = dx + qx * dz
    //  c_dy = dy + qy * dz
    //
    //  c_px = px + qx * pz
    //  c_py = py + qy * pz
    //
    //  a = c_dx * c_dy
    //  b = c_dx * c_py + c_dy * c_px - dz
    //  c = c_px * c_py - pz
    
    qx = 1 - x_point;
    qy = 1 - y_point;
    
    c_dx = ray_dir[X] + qx * ray_dir[Z];
    c_dy = ray_dir[Y] + qy * ray_dir[Z];
    
    c_px = ray_from[X] + qx * ray_from[Z];
    c_py = ray_from[Y] + qy * ray_from[Z];
    
    a = c_dx * c_dy;
    b = c_dx * c_py + c_dy * c_px - ray_dir[Z];
    c = c_px * c_py - ray_from[Z];
    
    if (a == 0) {
      // This equation reduces to a linear equation, bt + c = 0.
      // So t = - c / b.
      if (b == 0) {
	// There are no solutions, no intersections.
	return 0;
      }
      
      t_0 = - c / b;
      t_1 = -INFINITY; // Not a real intersection; won't be considered.
    } else {
      
      // Calculate the discriminant of the quadratic equation.
      disc = SQR(b) - 4.0*a*c;
      
      // If the discriminant is less than or equal to zero, then there
      // are no intersections with the bilinear patch (tangent rays are
      // treated as no intersection at all).
      if (disc <= 0) {
	return 0;
      }
      
      // Calculate the intersection t-values. Must have t_0 < t_1 for later
      // calculations to work.
      // Note that we can save a square root operation by saying
      // t_1 = c / a * t_0 ... but this only works if we can be sure that
      // t_0 isn't zero. It may easily be if sqrt(disc) == b or == -b.
      // So, we make sure that the +/-sqrt(disc) is always the same sign
      // as -b to make sure that the subtraction can't produce zero.
      // This makes the code a little long-winded.
      if (a >= 0) {
	if (b >= 0) {
	  t_0 = (-b - sqrt(disc)) / (2 * a);
	  t_1 = c / (a * t_0);
	} else {
	  t_1 = (-b + sqrt(disc)) / (2 * a);
	  t_0 = c / (a * t_1);
	}
      } else {
	if (b >= 0) {
	  t_1 = (-b - sqrt(disc)) / (2 * a);
	  t_0 = c / (a * t_1);
	} else {
	  t_0 = (-b + sqrt(disc)) / (2 * a);
	  t_1 = c / (a * t_0);
	}
      }
    }
  }

  hit_z = ray_from[Z] + t_0*ray_dir[Z];

  // Get the value of s1 (the parameter from point 0 to point 1).
  hit_s1 = (ray_from[X] + t_0*ray_dir[X]) + hit_z*qx;
  
  if ((hit_s1 > 0) && (hit_s1 < 1)) {
    // Get the value of s2 (the parameter from point 0 to point 2).
    hit_s2 = (ray_from[Y] + t_0*ray_dir[Y]) + hit_z*qy;
    
    if ((hit_s2 > 0) && (hit_s2 < 1)) {
      
      // First check to see if the hitpoint is within the Cutout
      // that determines our outline (if we have a Cutout).
      if ((cutout == NULL) || (cutout->is_inside(hit_s1, hit_s2))) {
	
	// Find the normal.
	if (is_flat) {
	  Vector::assign(norm, 0, 0, 1);
	} else {
	  // Find the east and west points that correspond to the hitpoint.
	  // We do this by interpolating between p1 and p3 for e and
	  // p0 and p2 for w.
	  e_x = 1 * (1 - hit_s2) + x_point * hit_s2;
	  e_y = 0 * (1 - hit_s2) + y_point * hit_s2;
	  e_z = 0 * (1 - hit_s2) +       1 * hit_s2;
	  
	  w_x = 0 * (1 - hit_s2) +       0 * hit_s2;
	  w_y = 0 * (1 - hit_s2) +       1 * hit_s2;
	  w_z = 0 * (1 - hit_s2) +       0 * hit_s2;
	  
	  // Find the north and south points that correspond to the hitpoint.
	  // We do this by interpolating between p2 and p3 for n and
	  // p0 and p1 for s.
	  n_x = 0 * (1 - hit_s1) + x_point * hit_s1;
	  n_y = 1 * (1 - hit_s1) + y_point * hit_s1;
	  n_z = 0 * (1 - hit_s1) +       1 * hit_s1;
	  
	  s_x = 0 * (1 - hit_s2) +       1 * hit_s2;
	  s_y = 0 * (1 - hit_s2) +       0 * hit_s2;
	  s_z = 0 * (1 - hit_s2) +       0 * hit_s2;
	  
	  // Now we construct the vectors n_s1 and n_s2. These are
	  // n_s1 = (dx/ds1 dy/ds1 dz/ds1) and n_s2 = (dx/ds2 dy/ds2 dz/ds2)
	  //
	  // These give us two directions along the surface at the hitpoint.
	  // We take n_s1 x n_s2 = norm.
	  
	  Vector::assign(n_s1, e_x - w_x, e_y - w_y, e_z - w_z);
	  Vector::assign(n_s2, n_x - s_x, n_y - s_y, n_z - s_z);
	  
	  Vector::cross(n_s1, n_s2, norm);
	}
	
	int_storage[num_intersections].t = t_0;
	int_storage[num_intersections].is_entering = TRUE;
	Vector::set(int_storage[num_intersections].normal, norm);
	num_intersections++;
      }
    }
  }

  hit_z = ray_from[Z] + t_1*ray_dir[Z];
  
  // Get the value of s1 (the parameter from point 0 to point 1).
  hit_s1 = (ray_from[X] + t_1*ray_dir[X]) + hit_z*qx;
  
  if ((hit_s1 > 0) && (hit_s1 < 1)) {
    // Get the value of s2 (the parameter from point 0 to point 2).
    hit_s2 = (ray_from[Y] + t_1*ray_dir[Y]) + hit_z*qy;
    
    if ((hit_s2 > 0) && (hit_s2 < 1)) {
      
      // First check to see if the hitpoint is within the Cutout
      // that determines our outline (if we have a Cutout).
      if ((cutout == NULL) || (cutout->is_inside(hit_s1, hit_s2))) {
	
	// Find the normal.
	if (is_flat) {
	  Vector::assign(norm, 0, 0, 1);
	} else {
	  // Find the east and west points that correspond to the hitpoint.
	  // We do this by interpolating between p1 and p3 for e and
	  // p0 and p2 for w.
	  e_x = 1 * (1 - hit_s2) + x_point * hit_s2;
	  e_y = 0 * (1 - hit_s2) + y_point * hit_s2;
	  e_z = 0 * (1 - hit_s2) +       1 * hit_s2;
	  
	  w_x = 0 * (1 - hit_s2) +       0 * hit_s2;
	  w_y = 0 * (1 - hit_s2) +       1 * hit_s2;
	  w_z = 0 * (1 - hit_s2) +       0 * hit_s2;
	  
	  // Find the north and south points that correspond to the hitpoint.
	  // We do this by interpolating between p2 and p3 for n and
	  // p0 and p1 for s.
	  n_x = 0 * (1 - hit_s1) + x_point * hit_s1;
	  n_y = 1 * (1 - hit_s1) + y_point * hit_s1;
	  n_z = 0 * (1 - hit_s1) +       1 * hit_s1;
	  
	  s_x = 0 * (1 - hit_s2) +       1 * hit_s2;
	  s_y = 0 * (1 - hit_s2) +       0 * hit_s2;
	  s_z = 0 * (1 - hit_s2) +       0 * hit_s2;
	  
	  // Now we construct the vectors n_s1 and n_s2. These are
	  // n_s1 = (dx/ds1 dy/ds1 dz/ds1) and n_s2 = (dx/ds2 dy/ds2 dz/ds2)
	  //
	  // These give us two directions along the surface at the hitpoint.
	  // We take n_s1 x n_s2 = norm.
	  
	  Vector::assign(n_s1, e_x - w_x, e_y - w_y, e_z - w_z);
	  Vector::assign(n_s2, n_x - s_x, n_y - s_y, n_z - s_z);
	  
	  Vector::cross(n_s1, n_s2, norm);
	}
	
	int_storage[num_intersections].t = t_1;
	int_storage[num_intersections].is_entering = TRUE;
	Vector::set(int_storage[num_intersections].normal, norm);
	num_intersections++;
      }
    }
  }

  return num_intersections;
}

intersect_result twoSidedPatchPrimitive::point_intersect(primnode *node,
							 vector wp,
							 PointInfo *info,
							 uint ray_flags)
{
  // This primitive doesn't have any substance.
  return OUT;
}

intersect_result twoSidedPatchPrimitive::point_intersect(vector wp)
{
  // This primitive doesn't have any substance.
  return OUT;
}





// Now define some local macros to make the code length shorter for the
// routine twoSidedPatchPrimitive::voxel_intersect.


// Macro CHECK_OUTSIDE_PLUS_PLANE tests to see if the points of the box all
// lie outside the plane P at distance d.
#define CHECK_OUTSIDE_PLUS_PLANE(axis, d) {\
                                           \
  for (i = 0; i < 8; i++) {                \
    if (b[i][axis] < d) {                  \
      break;                               \
    }                                      \
  }                                        \
  if (i == 8) {                            \
    return OUT;                            \
  }                                        \
}


// Macro CHECK_OUTSIDE_MINUS_PLANE tests to see if the points of the box all
// lie outside the plane P at distance -d.
#define CHECK_OUTSIDE_MINUS_PLANE(axis, d) {\
                                            \
  for (i = 0; i < 8; i++)                   \
    if (b[i][axis] > -d)                    \
      break;                                \
                                            \
  if (i == 8) {                             \
    return OUT;                             \
  }                                         \
}

intersect_result twoSidedPatchPrimitive::voxel_intersect(box b)
{
  // This routine checks to see if a parallelipiped lies inside, outside,
  // or intersects the bilinear patch. We simply check to see if all the
  // points lie outside any of the planes that bound the patch.
  // (X < 0, X > x_point, Y < 0, Y > y_point, Z < 0, Z > 1)
  //
  // If the voxel is inside any of these planes then we return SURFACE.

  // Loop counter
  int i;

  // Step 1: Check to see if all of the points of the box lie outside
  //         the planes X < 0, X > x_point, Y < 0, Y > y_point, Z < 0, Z > 1.
  //         If so, return OUT.
  CHECK_OUTSIDE_PLUS_PLANE(X, x_point);
  CHECK_OUTSIDE_PLUS_PLANE(Y, y_point);
  if (is_flat) {
    CHECK_OUTSIDE_PLUS_PLANE(Z, 0);
  } else {
    CHECK_OUTSIDE_PLUS_PLANE(Z, 1);
  }
  CHECK_OUTSIDE_MINUS_PLANE(X, 0);
  CHECK_OUTSIDE_MINUS_PLANE(Y, 0);
  CHECK_OUTSIDE_MINUS_PLANE(Z, 0);

  // If we haven't returned OUT by this time then the box intersects
  // the patch's bounding box. Return SURFACE.
  return SURFACE;
}

// Undefine the macros defined above (local only).
#undef CHECK_OUTSIDE_PLUS_PLANE
#undef CHECK_OUTSIDE_MINUS_PLANE

char *twoSidedPatchPrimitive::prim_name(void) {
  return "Two-sided Bilinear Patch";
}

// A routine to create a new two-sided patch and place it so its corners are at
// any four specified points.
object new_two_sided_patch(vector p0, vector p1, vector p2, vector p3)
{
  object patch;
  transform patch_pos;
  double patch_x, patch_y;
  int flat_patch;

  position_patch(p0, p1, p2, p3, &flat_patch, &patch_x, &patch_y, patch_pos);

  patch = new_two_sided_patch(patch_x, patch_y, flat_patch);
  patch->place(patch_pos);

  return bind(patch);
}

// A routine to create a new two-sided patch and place it so its corners are at
// any four specified points. Also takes a Cutout parameter.
object new_two_sided_patch(vector p0, vector p1, vector p2, vector p3,
			   Cutout *cut)
{
  object patch;
  transform patch_pos;
  double patch_x, patch_y;
  int flat_patch;

  position_patch(p0, p1, p2, p3, &flat_patch, &patch_x, &patch_y, patch_pos);

  patch = new_two_sided_patch(patch_x, patch_y, flat_patch, cut);
  patch->place(patch_pos);

  return bind(patch);
}


#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif
