/* prim_uniformcluster.c : The uniform-subdivision cluster primitive
** 
** This file contains the function(s) for
** the UniformClusterPrimitive.
**
** Chris Butcher, 14 January 1996
*/

#include "mirage.h"

#ifdef USE_NAMESPACES
namespace Mirage {
#endif

// The UniformClusterPrimitive performs uniform subdivision.
void UniformClusterPrimitive::subdivide(void)
{
  int i, j;
  int low_edge, high_edge;

  // If we haven't already subdivided this cluster, then do so.
  if (!done_subdivision) {
    // Convert the DAG into an RCSG graph.
    subdivide_clusters_in_RCSG(myRCSG);

    // Note that we create an octree of the desired subdivision limit
    // from the RCSG graph. We set the maximum number of primitives to zero
    // so that the subdivision goes to the bottom level always.
    my_octree = octree_subdivide(myRCSG, 0, subdivision_level, 0,
				 minbound[X], minbound[Y], minbound[Z],
				 maxbound[X], maxbound[Y], maxbound[Z]);

    // We now have an octree of level subdivision_level that must
    // be converted from an octree into a three-dimensional array.

    // Allocate space for an array_size cubed array of 'thing' with
    // one cell's padding on each side for the cell-skipping algorithm.

    my_array = (thing *) calloc(CUBE(array_size + 2), sizeof(thing));

    if (my_array == NULL) {
      char errbuf[MIRAGE_STRING_LENGTH];
      sprintf(errbuf, "can't allocate subdivided array for level %d",
	      subdivision_level);
      fatal_error("UniformClusterPrimitive::subdivide", errbuf,
		  "You may be requesting too high a level of subdivision.\n"
		  "Increasing the subdivision level by one increases the\n"
		  "memory used by a factor of eight. You Have Been Warned.\n");
    }

    // Place 'padding' around the edges of the array.
    low_edge = 0;
    high_edge = array_size + 1;
    for (i = 0; i < array_size + 2; i++) {
      for (j = 0; j < array_size + 2; j++) {
	// pad the X planes at X low and X high
	my_array[low_edge  + i * y_step + j * z_step] = THING_CLUSTER_BORDER;
	my_array[high_edge + i * y_step + j * z_step] = THING_CLUSTER_BORDER;

	// pad the Y planes at Y low and Y high
	my_array[i +  low_edge * y_step + j * z_step] = THING_CLUSTER_BORDER;
	my_array[i + high_edge * y_step + j * z_step] = THING_CLUSTER_BORDER;

	// pad the Z planes at Z low and Z high
	my_array[i + j * y_step +  low_edge * z_step] = THING_CLUSTER_BORDER;
	my_array[i + j * y_step + high_edge * z_step] = THING_CLUSTER_BORDER;
      }
    }

    // Work out the size of the cells.
    Vector::mult(1.0 / array_size, bound_size, cell_size);


    // Flatten the octree into the array. (not my terminology! - Chris)
    flatten_octree(my_octree, my_array, 0, 0, 0,
		   array_size, array_size, array_size);

    // It might be wise here to deallocate the octree. If this is done,
    // must remember not to free the RCSG nodes as they are still stored
    // in the uniform grid.

    done_subdivision = TRUE;
  }
}

// The ray-intersection routine for UniformClusterPrimitive traces
// the ray through the uniform subdivision grid ...
int UniformClusterPrimitive::ray_intersect(IntersectionList *the_list,
					   primnode *node,
					   primnode *validate_node,
					   vector ray_from, vector ray_dir,
					   uint ray_flags,
					   signature ray_signature)
{
  vector cluster_ray_from, cluster_ray_dir;
  vector t_entry, t_exit;
  double p_x, p_y, p_z, d_x, d_y, d_z;
  double overall_entry, overall_exit;
  int found_int;

  // Store the signature of this ray that we're intersecting with.
  node->last_ray_signature = ray_signature;

  // Transform the ray into the cluster's primitive space.
  Matrix::apply(node->position[AFT], ray_from, cluster_ray_from);
  Matrix::applynoshift(node->position[AFT], ray_dir, cluster_ray_dir);

  p_x = cluster_ray_from[X];  d_x = cluster_ray_dir[X];
  p_y = cluster_ray_from[Y];  d_y = cluster_ray_dir[Y];
  p_z = cluster_ray_from[Z];  d_z = cluster_ray_dir[Z];

  /*
   * Determine whether or not the ray intersects the uniform cluster's
   * bounding box.
   */
  
  // If the ray is outside the bounds and heading away from them, then there
  // will be no intersections.
  if (p_x > maxbound[X] && d_x > 0.0) {
    node->last_ray_hit = FALSE;
    return FALSE;
  }
  if (p_x < minbound[X] && d_x < 0.0) {
    node->last_ray_hit = FALSE;
    return FALSE;
  }
  if (p_y > maxbound[Y] && d_y > 0.0) {
    node->last_ray_hit = FALSE;
    return FALSE;
  }
  if (p_y < minbound[Y] && d_y < 0.0) {
    node->last_ray_hit = FALSE;
    return FALSE;
  }
  if (p_z > maxbound[Z] && d_z > 0.0) {
    node->last_ray_hit = FALSE;
    return FALSE;
  }
  if (p_z < minbound[Z] && d_z < 0.0) {
    node->last_ray_hit = FALSE;
    return FALSE;
  }

  // Determine the low and high t-values where the ray crosses the bounding
  // planes of the cluster.
  if (d_x > TINY) {
    // The ray is heading towards +X. Check to see if it is outside the
    // box where it starts.
    if (p_x > maxbound[X]) {
      node->last_ray_hit = FALSE;
      return FALSE;
    }

    // Determine t-values for entry and exit in X.
    t_entry[X] = (minbound[X] - p_x) / d_x;
    t_exit [X] = (maxbound[X] - p_x) / d_x;
  } else if (d_x < -TINY) {
    // The ray is heading towards -X. Check to see if it is outside the
    // box where it starts.
    if (p_x < minbound[X]) {
      node->last_ray_hit = FALSE;
      return FALSE;
    }

    // Determine t-values for entry and exit in X.
    t_entry[X] = (maxbound[X] - p_x) / d_x;
    t_exit [X] = (minbound[X] - p_x) / d_x;
  } else {
    // The ray isn't moving appreciably in X. Check to see if the starting
    // point is outside the bounding planes.
    if ((p_x < minbound[X]) || (p_x > maxbound[X])) {
      node->last_ray_hit = FALSE;
      return FALSE;
    }

    // The ray is always inside the bounds in X.
    t_entry[X] = -INFINITY;
    t_exit [X] = +INFINITY;
  }

  if (d_y > TINY) {
    // The ray is heading towards +Y. Check to see if it is outside the
    // box where it starts.
    if (p_y > maxbound[Y]) {
      node->last_ray_hit = FALSE;
      return FALSE;
    }

    // Determine t-values for entry and exit in Y.
    t_entry[Y] = (minbound[Y] - p_y) / d_y;
    t_exit [Y] = (maxbound[Y] - p_y) / d_y;
  } else if (d_y < -TINY) {
    // The ray is heading towards -Y. Check to see if it is outside the
    // box where it starts.
    if (p_y < minbound[Y]) {
      node->last_ray_hit = FALSE;
      return FALSE;
    }

    // Determine t-values for entry and exit in Y.
    t_entry[Y] = (maxbound[Y] - p_y) / d_y;
    t_exit [Y] = (minbound[Y] - p_y) / d_y;
  } else {
    // The ray isn't moving appreciably in Y. Check to see if the starting
    // point is outside the bounding planes.
    if ((p_y < minbound[Y]) || (p_y > maxbound[Y])) {
      node->last_ray_hit = FALSE;
      return FALSE;
    }

    // The ray is always inside the bounds in Y.
    t_entry[Y] = -INFINITY;
    t_exit [Y] = +INFINITY;
  }

  if (d_z > TINY) {
    // The ray is heading towards +Z. Check to see if it is outside the
    // box where it starts.
    if (p_z > maxbound[Z]) {
      node->last_ray_hit = FALSE;
      return FALSE;
    }

    // Determine t-values for entry and exit in Z.
    t_entry[Z] = (minbound[Z] - p_z) / d_z;
    t_exit [Z] = (maxbound[Z] - p_z) / d_z;
  } else if (d_z < -TINY) {
    // The ray is heading towards -Z. Check to see if it is outside the
    // box where it starts.
    if (p_z < minbound[Z]) {
      node->last_ray_hit = FALSE;
      return FALSE;
    }

    // Determine t-values for entry and exit in Z.
    t_entry[Z] = (maxbound[Z] - p_z) / d_z;
    t_exit [Z] = (minbound[Z] - p_z) / d_z;
  } else {
    // The ray isn't moving appreciably in Z. Check to see if the starting
    // point is outside the bounding planes.
    if ((p_z < minbound[Z]) || (p_z > maxbound[Z])) {
      node->last_ray_hit = FALSE;
      return FALSE;
    }

    // The ray is always inside the bounds in Z.
    t_entry[Z] = -INFINITY;
    t_exit [Z] = +INFINITY;
  }

  // Determine the overall (maximum) t_entry and overall (minimum) t_exit.
  if (t_entry[X] > t_entry[Y])
    if (t_entry[X] > t_entry[Z])
      overall_entry = t_entry[X];
    else
      overall_entry = t_entry[Z];
  else
    if (t_entry[Y] > t_entry[Z])
      overall_entry = t_entry[Y];
    else
      overall_entry = t_entry[Z];

  if (t_exit[X] < t_exit[Y])
    if (t_exit[X] < t_exit[Z])
      overall_exit = t_exit[X];
    else
      overall_exit = t_exit[Z];
  else
    if (t_exit[Y] < t_exit[Z])
      overall_exit = t_exit[Y];
    else
      overall_exit = t_exit[Z];

  // If there are no intersections then don't traverse the uniform subdivision.
  if (overall_entry > overall_exit) {
    node->last_ray_hit = FALSE;
    return FALSE;
  }

  /*
   * Otherwise, we DO need to traverse the uniform cluster.
   */

  // Intersect the ray with the uniform grid structure. This traverses the ray
  // through the cells of the grid. In each cell the ray is intersected
  // with the RCSG tree found there. If any valid intersections are found
  // and the stop_at_first_intersection flag is set, then the traversal
  // stops. Otherwise keep looking for more intersections.
  found_int = grid_ray_intersect(node, validate_node, the_list,
				 cluster_ray_from, cluster_ray_dir,
				 ray_flags, new_ray_signature());

#if 0
  if (found_int)
    printf("UCP: r_i true\n");
  else
    printf("UCP: r_i false\n");
#endif

  node->last_ray_hit = found_int;

  return found_int;
}

// This routine must never be called. The UniformClusterPrimitive is
// tied to Mirage's object structure.
int UniformClusterPrimitive::ray_intersect(vector ray_from, vector ray_dir,
					   rayintersect *int_storage)
{
  fatal_error("UniformClusterPrimitive::ray_intersect", "Incorrect r_i call!",
	      "This error should not happen.");
  return 0;
}

intersect_result UniformClusterPrimitive::point_intersect(primnode *node,
							  vector apoint,
							  PointInfo *info,
							  uint ray_flags)
{
  vector cluster_point;

  // Transform the sample point into the cluster's primitive space.
  Matrix::apply(node->position[AFT], apoint, cluster_point);

  // Classify this point with respect to the primitives in the RCSG.
  return classify_point_in_rcsg(myRCSG, cluster_point, info, ray_flags);
}

// This routine must never be called. The UniformClusterPrimitive is
// tied to Mirage's object structure.
intersect_result UniformClusterPrimitive::point_intersect(vector apoint)
{
  fatal_error("UniformClusterPrimitive::point_intersect",
	      "Incorrect p_i call!",
	      "This error should not happen.");
  return OUT;
}

intersect_result UniformClusterPrimitive::voxel_intersect(box v)
{
  box cluster_voxel_check;
  int i;
  
  // Check to see if the voxel needs to be intersected with the
  // cluster (see if it intersects the bounding box).
  // We do this by transforming the voxel to be checked so that it
  // lies in a space that maps minbound -> -1 and maxbound -> +1,
  // and then using the cube voxel intersection method.
  for (i = 0; i < 8; i++) {
    cluster_voxel_check[i][X] = -1 + 2*(v[i][X] - minbound[X]) / bound_size[X];
    cluster_voxel_check[i][Y] = -1 + 2*(v[i][Y] - minbound[Y]) / bound_size[Y];
    cluster_voxel_check[i][Z] = -1 + 2*(v[i][Z] - minbound[Z]) / bound_size[Z];
  }

  if (cube_voxel_intersect(cluster_voxel_check) == OUT) {
    // This voxel does not intersect the cluster, so the cluster doesn't need
    // to be intersected with inside it.
    return OUT;
  } else {
    // This voxel either intersects or is inside the cluster. The cluster does
    // need to be intersected with inside it.
    return SURFACE;
  }
}

char *UniformClusterPrimitive::prim_name(void) {
  return "Uniform cluster";
}



#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif
