/* prim_unitcyl.c : The unit cylinder primitive
** 
** The cylinder primitive defines a cylinder of radius 1,
** from Z = 0 to Z = +1.
**
** Chris Butcher, 14 December 1996
*/

#include "mirage.h"

#ifdef USE_NAMESPACES
namespace Mirage {
#endif

int unitCylinderPrimitive::ray_intersect(IntersectionList *the_list, 
					 primnode *node, 
					 primnode *validate_node, 
					 vector ray_from,
					 vector ray_dir, uint ray_flags,
					 signature ray_signature)
{
  // The coordinates of ray_from in primitive space
  double p_x, p_y, p_z;
  // The direction ray_dir in primitive space
  double d_x, d_y, d_z;
  // The coefficients of the quadratic equation in t
  double a, b, c;
  // The discriminant of this equation
  double disc;
  // Have we found an intersection?
  int found = 0;
  // The t-values of the intersection points with the cylinder.
  double t_0, t_1;
  // The t_values of the intersection points with the top and bottom
  // bounding planes. t_lower holds the lesser of these two t values,
  // t_upper holds the larger of them.
  double t_lower, t_upper;
  // The normal at the hitpoint.
  vector norm;

  node->last_ray_signature = ray_signature;

  // This apply is an inline function defined in matrix.h
  // Transform the ray startpoint into primitive space
  Matrix::apply(node->position[AFT], ray_from, &p_x, &p_y, &p_z);

  // This apply is also an inline function defined in matrix.h
  // Transform the ray direction into primitive space (note: shift does
  // not apply to this, only apply rotation + stretch sections of the matrix)
  Matrix::applynoshift(node->position[AFT], ray_dir, &d_x, &d_y, &d_z);

  // Check to see if the ray is either above the top and pointed up, or
  // below the bottom and pointed down.
  if (d_z >= 0) {
    if (p_z > 1.0) {
      // Ray will go above the cylinder.
      node->last_ray_hit = FALSE;
      return node->last_ray_hit;
    }
  } else {
    if (p_z < 0) {
      // Ray will go below the cylinder.
      node->last_ray_hit = FALSE;
      return node->last_ray_hit;
    }
  }

  // Form the coefficients of the quadratic equation that defines t.
  // This equation is |p + td| = 1.0
  // Note that this code is very similar to the sphere intersection test.
  a = SQR(d_x) + SQR(d_y);
  b = 2.0 * (p_x*d_x + p_y*d_y);
  c = SQR(p_x) + SQR(p_y) - 1.0;

  // Calculate the discriminant of the quadratic equation.
  disc = SQR(b) - 4.0*a*c;

  // If the discriminant is less than or equal to zero, then there
  // are no intersections with the cylinder (tangent rays are treated as
  // no intersection at all).
  if (disc <= 0) {
    node->last_ray_hit = FALSE;
    return node->last_ray_hit;
  }

  // a is equal to the square of the magnitude of the direction of the ray.
  // This *must* not be zero.
  if (a == 0)
    fatal_error("cylinderPrimitive::ray_intersect", "ray direction zero",
		"This error should not happen.");

  // Calculate the intersection t-values. Must have t_0 < t_1 for later
  // calculations to work.
  // Note that we can save a square root operation by saying
  // t_1 = c / a * t_0 ... but this only works if we can be sure that
  // t_0 isn't zero. It may easily be if sqrt(disc) == b or == -b.
  // So, we make sure that the +/-sqrt(disc) is always the same sign
  // as -b to make sure that the subtraction can't produce zero.
  // This makes the code a little long-winded.
  if (a >= 0) {
    if (b >= 0) {
      t_0 = (-b - sqrt(disc)) / (2 * a);
      t_1 = c / (a * t_0);
    } else {
      t_1 = (-b + sqrt(disc)) / (2 * a);
      t_0 = c / (a * t_1);
    }
  } else {
    if (b >= 0) {
      t_1 = (-b - sqrt(disc)) / (2 * a);
      t_0 = c / (a * t_1);
    } else {
      t_0 = (-b + sqrt(disc)) / (2 * a);
      t_1 = c / (a * t_0);
    }
  }

  // Rays that are parallel to the planes bounding the top and bottom
  // of the cylinder don't need to be tested against
  // them, just insert the t_0 and t_1 intersections directly.
  if ((d_z > -TINY) && (d_z < TINY)) {

    // If the ray is outside the extent of the cylinder, no intersections.
    if (p_z < 0 || p_z > 1) {
      node->last_ray_hit = FALSE;
      return node->last_ray_hit;
    }

    // Insert any positive t-values (valid intersections).
    if (t_0 > MIN_TVAL_CUTOFF) {
      // Get the normal (since this is a unit cylinder, the normal is the
      // same as the primitive-space hitpoint = p + td ... in X and Y at least.
      // The Z component of the normal is zero).
      norm[X] = p_x + t_0*d_x;
      norm[Y] = p_y + t_0*d_y;
      norm[Z] = 0;
      // Note that since t_0 < t_1 always, this intersection is an ENTERING
      // one.
      insert_entering(the_list, node, validate_node, norm, t_0,
		      ray_from, ray_dir, ray_flags);
      found = 1;
    }
    if (t_1 > MIN_TVAL_CUTOFF) {
      // Get the normal (since this is a unit cylinder, the normal is the
      // same as the primitive-space hitpoint = p + td ... in X and Y at least.
      // The Z component of the normal is zero).
      norm[X] = p_x + t_1*d_x;
      norm[Y] = p_y + t_1*d_y;
      norm[Z] = 0;
      // Note that since t_0 < t_1 always, this intersection is a LEAVING
      // one.
      insert_leaving(the_list, node, validate_node, norm, t_1,
		     ray_from, ray_dir, ray_flags);
      found = 1;
    }
    
  } else {
    // Ray isn't parallel to the planes, must be checked against them.

    // Calculate the t-values that correspond to the ray hitting the top
    // and bottom planes.
    if (d_z > 0) {
      // The ray hits the bottom plane before the top plane
      // thus t_upper is the top plane, t_lower is the bottom plane
      t_upper = (1 - p_z) / d_z;
      t_lower = - p_z / d_z;

      // The normal for the first intersection (t_lower) points down
      norm[Z] = -1.0;
    } else {
      // The ray hits the top plane before the bottom plane
      // thus t_upper is the bottom plane, t_lower is the top plane
      t_lower = (1 - p_z) / d_z;
      t_upper = - p_z / d_z;

      // The normal for the first intersection (t_lower) points up
      norm[Z] = +1.0;
    }

    // Insert the intersections with the top and bottom planes if they
    // occur at positive t, and between t_0 and t_1 (the cylinder's extent).
    norm[X] = 0;
    norm[Y] = 0;
    if (t_lower > MIN_TVAL_CUTOFF && t_lower > t_0 && t_lower < t_1) {
      // t_lower is less than t_upper so this is an ENTERING intersection.
      insert_entering(the_list, node, validate_node, norm, t_lower,
		      ray_from, ray_dir, ray_flags);
      found = 1;
    }
    if (t_upper > MIN_TVAL_CUTOFF && t_upper > t_0 && t_upper < t_1) {
      // The normal for t_upper points in the opposite Z direction to the
      // normal for t_lower.
      norm[Z] = -norm[Z];
      // t_lower is less than t_upper so this is a LEAVING intersection.
      insert_leaving(the_list, node, validate_node, norm, t_upper,
		     ray_from, ray_dir, ray_flags);
      found = 1;
    }

    // Insert any t-values that are positive and between t_lower and t_upper.
    if (t_0 > MIN_TVAL_CUTOFF && t_0 > t_lower && t_0 < t_upper) {
      // Get the normal (since this is a unit cylinder, the normal is the
      // same as the primitive-space hitpoint = p + td ... in X and Y at least.
      // The Z component of the normal is zero).
      norm[X] = p_x + t_0*d_x;
      norm[Y] = p_y + t_0*d_y;
      norm[Z] = 0;
      // Note that since t_0 < t_1 always, this intersection is an ENTERING
      // one.
      insert_entering(the_list, node, validate_node, norm, t_0,
		      ray_from, ray_dir, ray_flags);
      found = 1;
    }
    if (t_1 > MIN_TVAL_CUTOFF && t_1 > t_lower && t_1 < t_upper) {
      // Get the normal (since this is a unit cylinder, the normal is the
      // same as the primitive-space hitpoint = p + td ... in X and Y at least.
      // The Z component of the normal is zero).
      norm[X] = p_x + t_1*d_x;
      norm[Y] = p_y + t_1*d_y;
      norm[Z] = 0;
      // Note that since t_0 < t_1 always, this intersection is a LEAVING
      // one.
      insert_leaving(the_list, node, validate_node, norm, t_1,
		     ray_from, ray_dir, ray_flags);
      found = 1;
    }
  }

  if (found)
    node->last_ray_hit = TRUE;
  else
    node->last_ray_hit = FALSE;

  return node->last_ray_hit;
}

int unitCylinderPrimitive::ray_intersect(vector ray_from, vector ray_dir,
					 rayintersect *int_storage)
{
  // The coefficients of the quadratic equation in t
  double a, b, c;
  // The discriminant of this equation
  double disc;
  // Have we found an intersection?
  int num_intersections = 0;
  // The t-values of the intersection points with the cylinder.
  double t_0, t_1;
  // The t_values of the intersection points with the top and bottom
  // bounding planes. t_lower holds the lesser of these two t values,
  // t_upper holds the larger of them.
  double t_lower, t_upper;
  // The normal at the hitpoint.
  vector norm;

  // Check to see if the ray is either above the top and pointed up, or
  // below the bottom and pointed down.
  if (ray_dir[Z] >= 0) {
    if (ray_from[Z] > 1.0) {
      // Ray will go above the cylinder.
      return 0;
    }
  } else {
    if (ray_from[Z] < 0) {
      // Ray will go below the cylinder.
      return 0;
    }
  }

  // Form the coefficients of the quadratic equation that defines t.
  // This equation is |p + td| = 1.0
  // Note that this code is very similar to the sphere intersection test.
  a = SQR(ray_dir[X]) + SQR(ray_dir[Y]);
  b = 2.0 * (ray_from[X]*ray_dir[X] + ray_from[Y]*ray_dir[Y]);
  c = SQR(ray_from[X]) + SQR(ray_from[Y]) - 1.0;

  // Calculate the discriminant of the quadratic equation.
  disc = SQR(b) - 4.0*a*c;

  // If the discriminant is less than or equal to zero, then there
  // are no intersections with the cylinder (tangent rays are treated as
  // no intersection at all).
  if (disc <= 0) {
    return 0;
  }

  // a is equal to the square of the magnitude of the direction of the ray.
  // This *must* not be zero.
  if (a == 0)
    fatal_error("cylinderPrimitive::ray_intersect", "ray direction zero",
		"This error should not happen.");

  // Calculate the intersection t-values. Must have t_0 < t_1 for later
  // calculations to work.
  // Note that we can save a square root operation by saying
  // t_1 = c / a * t_0 ... but this only works if we can be sure that
  // t_0 isn't zero. It may easily be if sqrt(disc) == b or == -b.
  // So, we make sure that the +/-sqrt(disc) is always the same sign
  // as -b to make sure that the subtraction can't produce zero.
  // This makes the code a little long-winded.
  if (a >= 0) {
    if (b >= 0) {
      t_0 = (-b - sqrt(disc)) / (2 * a);
      t_1 = c / (a * t_0);
    } else {
      t_1 = (-b + sqrt(disc)) / (2 * a);
      t_0 = c / (a * t_1);
    }
  } else {
    if (b >= 0) {
      t_1 = (-b - sqrt(disc)) / (2 * a);
      t_0 = c / (a * t_1);
    } else {
      t_0 = (-b + sqrt(disc)) / (2 * a);
      t_1 = c / (a * t_0);
    }
  }

  // Rays that are parallel to the planes bounding the top and bottom
  // of the cylinder don't need to be tested against
  // them, just insert the t_0 and t_1 intersections directly.
  if ((ray_dir[Z] > -TINY) && (ray_dir[Z] < TINY)) {

    // If the ray is outside the extent of the cylinder, no intersections.
    if (ray_from[Z] < 0 || ray_from[Z] > 1) {
      return 0;
    }

    // Get the normal (since this is a unit cylinder, the normal is the
    // same as the primitive-space hitpoint = p + td ... in X and Y at least.
    // The Z component of the normal is zero).
    norm[X] = ray_from[X] + t_0*ray_dir[X];
    norm[Y] = ray_from[Y] + t_0*ray_dir[Y];
    norm[Z] = 0;
    
    // Note that since t_0 < t_1 always, this intersection is an ENTERING
    // one.
    int_storage[num_intersections].t = t_0;
    int_storage[num_intersections].is_entering = TRUE;
    Vector::set(int_storage[num_intersections].normal, norm);
    num_intersections++;

    // Get the normal (since this is a unit cylinder, the normal is the
    // same as the primitive-space hitpoint = p + td ... in X and Y at least.
    // The Z component of the normal is zero).
    norm[X] = ray_from[X] + t_1*ray_dir[X];
    norm[Y] = ray_from[Y] + t_1*ray_dir[Y];
    norm[Z] = 0;
    
    // Note that since t_0 < t_1 always, this intersection is a LEAVING
    // one.
    int_storage[num_intersections].t = t_1;
    int_storage[num_intersections].is_entering = FALSE;
    Vector::set(int_storage[num_intersections].normal, norm);
    num_intersections++;
    
  } else {
    // Ray isn't parallel to the planes, must be checked against them.

    // Calculate the t-values that correspond to the ray hitting the top
    // and bottom planes.
    if (ray_dir[Z] > 0) {
      // The ray hits the bottom plane before the top plane
      // thus t_upper is the top plane, t_lower is the bottom plane
      t_upper = (1 - ray_from[Z]) / ray_dir[Z];
      t_lower = - ray_from[Z] / ray_dir[Z];

      // The normal for the first intersection (t_lower) points down
      norm[Z] = -1.0;
    } else {
      // The ray hits the top plane before the bottom plane
      // thus t_upper is the bottom plane, t_lower is the top plane
      t_lower = (1 - ray_from[Z]) / ray_dir[Z];
      t_upper = - ray_from[Z] / ray_dir[Z];

      // The normal for the first intersection (t_lower) points up
      norm[Z] = +1.0;
    }

    // Insert the intersections with the top and bottom planes if they
    // occur at positive t, and between t_0 and t_1 (the cylinder's extent).
    norm[X] = 0;
    norm[Y] = 0;
    if (t_lower > t_0 && t_lower < t_1) {
      // t_lower is less than t_upper so this is an ENTERING intersection.
      int_storage[num_intersections].t = t_lower;
      int_storage[num_intersections].is_entering = TRUE;
      Vector::set(int_storage[num_intersections].normal, norm);
      num_intersections++;
    }
    if (t_upper > t_0 && t_upper < t_1) {
      // The normal for t_upper points in the opposite Z direction to the
      // normal for t_lower.
      norm[Z] = -norm[Z];
      // t_lower is less than t_upper so this is a LEAVING intersection.
      int_storage[num_intersections].t = t_upper;
      int_storage[num_intersections].is_entering = FALSE;
      Vector::set(int_storage[num_intersections].normal, norm);
      num_intersections++;
    }

    // Insert any t-values that are positive and between t_lower and t_upper.
    if (t_0 > t_lower && t_0 < t_upper) {
      // Get the normal (since this is a unit cylinder, the normal is the
      // same as the primitive-space hitpoint = p + td ... in X and Y at least.
      // The Z component of the normal is zero).
      norm[X] = ray_from[X] + t_0*ray_dir[X];
      norm[Y] = ray_from[Y] + t_0*ray_dir[Y];
      norm[Z] = 0;

      // Note that since t_0 < t_1 always, this intersection is an ENTERING
      // one.
      int_storage[num_intersections].t = t_0;
      int_storage[num_intersections].is_entering = TRUE;
      Vector::set(int_storage[num_intersections].normal, norm);
      num_intersections++;
    }
    if (t_1 > t_lower && t_1 < t_upper) {
      // Get the normal (since this is a unit cylinder, the normal is the
      // same as the primitive-space hitpoint = p + td ... in X and Y at least.
      // The Z component of the normal is zero).
      norm[X] = ray_from[X] + t_1*ray_dir[X];
      norm[Y] = ray_from[Y] + t_1*ray_dir[Y];
      norm[Z] = 0;

      // Note that since t_0 < t_1 always, this intersection is a LEAVING
      // one.
      int_storage[num_intersections].t = t_1;
      int_storage[num_intersections].is_entering = FALSE;
      Vector::set(int_storage[num_intersections].normal, norm);
      num_intersections++;
    }
  }

  return num_intersections;
}

intersect_result unitCylinderPrimitive::point_intersect(primnode *node,
							vector wp,
							PointInfo *info,
							uint ray_flags)
{
  // Here wp is the world-point that we are testing for IN/OUT.
  // The values p_x, p_y, p_z are its coordinates in primitive space
  // (the mapping from world to primitive space is given by position.aft)
  double p_x, p_y, p_z, d_sq;

  // This apply is an inline function defined in matrix.h
  // Transform the worldpoint into primitive space.
  Matrix::apply(node->position[AFT], wp, &p_x, &p_y, &p_z);

  // Test the point against the planes that bound the cylinder.
  if (p_z < 0 || p_z > 1)
    return OUT;

  // d_sq is the point's distance from the cylinder's axis, squared.
  d_sq = SQR(p_x) + SQR(p_y);
  
  // If the distance squared is <= 1 squared, the point is IN.
  if (d_sq <= 1.0) {
    set_point_contents(node, info);
    return IN;
  } else
    return OUT;
}

intersect_result unitCylinderPrimitive::point_intersect(vector wp)
{
  double d_sq;

  // Test the point against the planes that bound the cylinder.
  if (wp[Z] < 0 || wp[Z] > 1)
    return OUT;

  // d_sq is the point's distance from the cylinder's axis, squared.
  d_sq = SQR(wp[X]) + SQR(wp[Y]);
  
  // If the distance squared is <= 1 squared, the point is IN.
  if (d_sq <= 1.0)
    return IN;
  else
    return OUT;
}

intersect_result unitCylinderPrimitive::voxel_intersect(box b)
{
  // This routine checks to see if a parallelipiped lies inside, outside,
  // or intersects the surface of the cylinder.
  //
  // It does so by calling the cylinder primitive's intersection routine
  // and checking the bounding planes manually.

  int i, above_top, in_middle, below_bottom;
  intersect_result cyl_result, planes_result;

  CHECK_VOXEL_HANDEDNESS(b);

  // Check the points to see which side of the two planes they lie on.
  above_top = below_bottom = in_middle = FALSE;
  planes_result = IN;

  for (i = 0; i < 8; i++) {
    if (b[i][Z] < 0) {
      below_bottom = TRUE;
      if (in_middle || above_top) {
	planes_result = SURFACE;
	break;
      }
    } else if (b[i][Z] > 1) {
      above_top = TRUE;
      if (in_middle || below_bottom) {
	planes_result = SURFACE;
	break;
      }
    } else {
      in_middle = TRUE;
      if (above_top || below_bottom) {
	planes_result = SURFACE;
	break;
      }
    }
  }

  if (i == 8) {
    // We have checked them all without finding two points on opposite
    // sides of a plane.
    if (in_middle)
      planes_result = IN;
    else
      planes_result = OUT;
  }

  // If the box is outside the bounding planes, return OUT.
  if (planes_result == OUT)
    return OUT;
  
  // Intersect with the bounding cylinder.
  cyl_result = cylinder_voxel_intersect(b);

  // If the box is outside the bounding cylinder, return OUT.
  if (cyl_result == OUT)
    return OUT;

  // If we are inside the cylinder, return IN / SURFACE depending on the planes
  if (cyl_result == IN) {
    return planes_result;
  } else  {
    // We are on the surface of the cylinder and we know we are not
    // outside the planes, so return SURFACE.
    return SURFACE;
  }
}

char *unitCylinderPrimitive::prim_name(void) {
  return "Unit Cylinder";
}


#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif
