/* primitives.h : A list of all the primitives
** 
** This file has prototypes for all of the subclasses of Primitive
** (including the ClusterPrimitives).
**
** Chris Butcher, 28 November 1996
*/

// rayintersect structs are used to return information from the alternative
// ray_intersect routine.
typedef struct _rayintersect {
  double t;
  int is_entering;
  vector normal;
} rayintersect;

// Primitives are defined by three procedures: the intersection procedures
// for rays, voxels and points. This allows us to create things like
// spheres, cubes, cones etc. very easily and things like volkswagens not
// so easily. This may be done by using a cluster, a special type of
// primitive that implements its ray_intersect procedure by traversing
// some kind of internal structure containing another level of objects
// [typically a (possibly subdivided) RCSG tree].
class Primitive : public DeletableObject {
public:
  int is_cluster;

  // These functions define the primitive - how it behaves, what it looks
  // like, what shape it is etc.
  virtual int ray_intersect(class IntersectionList *the_list,
			    primnode *node, primnode *validate_node,
			    vector ray_from, vector ray_dir,
			    uint ray_flags, signature ray_signature) = 0;
  virtual int ray_intersect(vector ray_from, vector ray_dir,
			    rayintersect *int_storage) = 0;
  virtual intersect_result point_intersect(primnode *node,
					   vector apoint,
					   PointInfo *info,
					   uint ray_flags) = 0;
  virtual intersect_result point_intersect(vector apoint) = 0;
  virtual intersect_result voxel_intersect(box v) = 0;
  
  // Return the name of the primitive (for debugging).
  virtual char *prim_name(void) = 0;
};

// This prototype is needed for many cluster primitives.
IntersectionList *new_intersection_list(void);

// This macro checks that the voxel which we are passed is a right-handed
// coordinate system. If it is not, then it reverses the Z order of the points
// so that it is.
#define CHECK_VOXEL_HANDEDNESS(vox) {                 \
  vector handed_x, handed_y, handed_z, right_handed;  \
  vector swap;                                        \
  double dot_product;                                 \
                                                      \
  Vector::sub(vox[1], vox[0], handed_x);              \
  Vector::sub(vox[2], vox[0], handed_y);              \
  Vector::sub(vox[4], vox[0], handed_z);              \
                                                      \
  Vector::cross(handed_x, handed_y, right_handed);    \
                                                      \
  dot_product = Vector::dot(right_handed, handed_z);  \
                                                      \
  if (dot_product < 0) {                              \
    Vector::set(swap, vox[0]);                        \
    Vector::set(vox[0], vox[4]);                      \
    Vector::set(vox[4], swap);                        \
                                                      \
    Vector::set(swap, vox[1]);                        \
    Vector::set(vox[1], vox[5]);                      \
    Vector::set(vox[5], swap);                        \
                                                      \
    Vector::set(swap, vox[2]);                        \
    Vector::set(vox[2], vox[6]);                      \
    Vector::set(vox[6], swap);                        \
                                                      \
    Vector::set(swap, vox[3]);                        \
    Vector::set(vox[3], vox[7]);                      \
    Vector::set(vox[7], swap);                        \
  }                                                   \
}

// All cluster primitives are a subclass of ClusterPrimitive.
class ClusterPrimitive : public Primitive {
public:
  object myScene;
  thing myRCSG;
  int done_subdivision, rcsg_created, found_labels;
  transform my_world_transform;

  // Traverse the DAG that we contain and find absolute locations of
  // all labels.
 void find_labels(transform existing_transform);

  // Traverse the DAG that we contain and create an intermediate RCSG.
  void create_RCSG(primnode *cluster_primnode,
		   transform higher_cluster_world);

  // Recursively create the RCSG ... propagate the above values of flags,
  // position, material, texcol, lightList, light etc. etc. through the RCSG.
  // Also pass the dagnode * that we want to create an RCSG from, plus
  // a the cluster's primitive space transform to be used to get the
  // absolute coordinate locations of cameras and lights.
  thing recursive_create_RCSG(object theDag,
			      transform cluster_world,
			      uint above_flags,
			      transform above_position,
			      Material *above_material,
			      SingleTexture *above_tex,
			      TextureCollection *above_texcol,
			      lightList *above_lightlist,
			      Light *above_light,
			      SfxType above_sfxtype,
			      void *above_sfx);

  // Subdivide the RCSG according to the cluster's subdivision scheme.
  virtual void subdivide(void) = 0;
};

// The most basic cluster primitive; no subdivision routines. Every ray
// is intersected with every primitive in the cluster.
class StupidClusterPrimitive : public ClusterPrimitive {
public:
  
  virtual int ray_intersect(IntersectionList *the_list, 
			    primnode *node, primnode *validate_node,
			    vector ray_from, vector ray_dir, uint ray_flags,
			    signature ray_signature);
  virtual intersect_result point_intersect(primnode *node, vector apoint,
					   PointInfo *info,
					   uint ray_flags);
  virtual intersect_result voxel_intersect(box v);
  virtual char *prim_name(void);
  virtual int ray_intersect(vector ray_from,
			    vector ray_dir, rayintersect *int_storage);
  virtual intersect_result point_intersect(vector apoint);

  // Don't do anything in here.
  void subdivide(void);

  // Create a new primitive to encapsulate the given object graph.
  StupidClusterPrimitive(object x) {
    is_cluster = 1;
    done_subdivision = FALSE;
    rcsg_created = FALSE;
    found_labels = FALSE;
    myScene = x;
    mark_for_deletion();
  }

  // Deletion method (invoked at the end of each frame).
  virtual ~StupidClusterPrimitive(void) {
  }
};

// A cluster primitive that performs octree subdivision.
class OctreeClusterPrimitive : public ClusterPrimitive {
public:
  thing my_octree;
  vector minbound, maxbound, bound_size;
  int max_prims_in_voxel, subdivision_limit, stop_at_first_intersection;

  // The next block of variables are temporary variables used in
  // traversing the ray through the octree structure.
  vector cluster_ray_from, cluster_ray_dir;
  vector *cell_size, exit_value;
  unsigned char ray_entry_octant;
  double ray_dir_ratio_x_y, ray_dir_ratio_x_z, ray_dir_ratio_y_x,
    ray_dir_ratio_y_z, ray_dir_ratio_z_x, ray_dir_ratio_z_y;
  double t_next_x, t_next_y, t_next_z;
  double *delta_t_cell_x, *delta_t_cell_y, *delta_t_cell_z;
  double cell_entry_t;
  primnode *octreecluster_primnode, *abovecluster_primnode;
  uint this_ray_flags;
  signature this_ray_signature;
  IntersectionList *the_list; // NB not the cluster's list, temp variable

  // Function prototypes
  virtual int ray_intersect(IntersectionList *the_list, 
			    primnode *node, primnode *validate_node,
			    vector ray_from, vector ray_dir, uint ray_flags,
			    signature ray_signature);
  virtual intersect_result point_intersect(primnode *node, vector apoint,
					   PointInfo *info,
					   uint ray_flags);
  virtual intersect_result voxel_intersect(box v);
  virtual int ray_intersect(vector ray_from,
			    vector ray_dir, rayintersect *int_storage);
  virtual intersect_result point_intersect(vector apoint);
  virtual char *prim_name(void);

  void subdivide(void);
  int octree_ray_intersect(primnode *cluster_primnode,
			   primnode *validate_primnode,
			   IntersectionList *above_list, vector ray_from,
			   vector ray_dir, uint ray_flags,
			   signature ray_signature);
  int traverse_octree(int current_level, thing the_thing,
		      vector parent_cell_exit, double t_next_x,
		      double t_next_y, double t_next_z);

  // Create a new cluster to encapsulate the given object graph.
  OctreeClusterPrimitive(object x, vector my_min, vector my_max,
			 int my_subdivision, int my_max_prims,
			 int should_stop) {
    is_cluster = 1;
    done_subdivision = FALSE;
    rcsg_created = FALSE;
    found_labels = FALSE;
    myScene = x;
    Vector::set(minbound, my_min);
    Vector::set(maxbound, my_max);
    Vector::sub(maxbound, minbound, bound_size);
    subdivision_limit = my_subdivision;
    max_prims_in_voxel = my_max_prims;
    stop_at_first_intersection = should_stop;
    mark_for_deletion();

    // Allocate temporary variable space. ***NB*** cell_size etc
    // will be indexed from 0 to subdivision_limit, requiring
    // subdivision_limit + 1 elements worth of space.
    cell_size      = (vector *) calloc(subdivision_limit + 1, sizeof(vector));
    delta_t_cell_x = (double *) calloc(subdivision_limit + 1, sizeof(double));
    delta_t_cell_y = (double *) calloc(subdivision_limit + 1, sizeof(double));
    delta_t_cell_z = (double *) calloc(subdivision_limit + 1, sizeof(double));
    
    if (cell_size == NULL || delta_t_cell_x == NULL ||
	delta_t_cell_y == NULL || delta_t_cell_z == NULL) {
      char errbuf[MIRAGE_STRING_LENGTH];
      
      sprintf(errbuf, "can't allocate variable space for subdivision level %d",
	      subdivision_limit);
      fatal_error("OctreeClusterPrimitive", errbuf,
		  "You are running seriously low on memory (this is not a\n"
		  "big request). Possibly there is a memory leak or you\n"
		  "are trying to render a huge number of frames at once.");
    }
  }

  // Create a new cluster to encapsulate the given object graph.
  // Same as the above constructor except passed x, y, z, values
  // rather than vectors.
  OctreeClusterPrimitive(object x,
			 double x_min, double y_min, double z_min,
			 double x_max, double y_max, double z_max,
			 int my_subdivision, int my_max_prims,
			 int should_stop) {
    is_cluster = 1;
    done_subdivision = FALSE;
    rcsg_created = FALSE;
    found_labels = FALSE;
    myScene = x;
    Vector::assign(minbound, x_min, y_min, z_min);
    Vector::assign(maxbound, x_max, y_max, z_max);
    Vector::sub(maxbound, minbound, bound_size);
    subdivision_limit = my_subdivision;
    max_prims_in_voxel = my_max_prims;
    stop_at_first_intersection = should_stop;
    mark_for_deletion();

    // Allocate temporary variable space. ***NB*** cell_size etc
    // will be indexed from 0 to subdivision_limit, requiring
    // subdivision_limit + 1 elements worth of space.
    cell_size      = (vector *) calloc(subdivision_limit + 1, sizeof(vector));
    delta_t_cell_x = (double *) calloc(subdivision_limit + 1, sizeof(double));
    delta_t_cell_y = (double *) calloc(subdivision_limit + 1, sizeof(double));
    delta_t_cell_z = (double *) calloc(subdivision_limit + 1, sizeof(double));
    
    if (cell_size == NULL || delta_t_cell_x == NULL ||
	delta_t_cell_y == NULL || delta_t_cell_z == NULL) {
      char errbuf[MIRAGE_STRING_LENGTH];
      
      sprintf(errbuf, "can't allocate variable space for subdivision level %d",
	      subdivision_limit);
      fatal_error("OctreeClusterPrimitive", errbuf,
		  "You are running seriously low on memory (this is not a\n"
		  "big request). Possibly there is a memory leak or you\n"
		  "are trying to render a huge number of frames at once.");
    }
  }

  // Deletion method (invoked at the end of each frame).
  virtual ~OctreeClusterPrimitive(void) {
    free(cell_size);
    free(delta_t_cell_x);
  }
};

// A cluster primitive that performs uniform-grid subdivision.
class UniformClusterPrimitive : public ClusterPrimitive {
public:
  thing my_octree, *my_array;
  vector minbound, maxbound, bound_size, cell_size;
  int subdivision_level, stop_at_first_intersection;
  int array_size, y_step, z_step;

  // Function prototypes
  virtual int ray_intersect(IntersectionList *the_list, 
			    primnode *node, primnode *validate_node,
			    vector ray_from, vector ray_dir, uint ray_flags,
			    signature ray_signature);
  virtual intersect_result point_intersect(primnode *node, vector apoint,
					   PointInfo *info,
					   uint ray_flags);
  virtual intersect_result voxel_intersect(box v);
  virtual int ray_intersect(vector ray_from,
			    vector ray_dir, rayintersect *int_storage);
  virtual intersect_result point_intersect(vector apoint);
  virtual char *prim_name(void);

  void subdivide(void);

  int grid_ray_intersect(primnode *cluster_primnode, primnode *val_primnode,
			 IntersectionList *above_list, vector ray_from,
			 vector ray_dir, uint ray_flags,
			 signature ray_signature);
  void flatten_octree(thing the_thing, thing *the_array,
		      int x_min, int y_min, int z_min,
		      int x_max, int y_max, int z_max);

  // Create a new cluster to encapsulate the given object graph.
  UniformClusterPrimitive(object x, vector my_min, vector my_max,
			  int my_subdivision, int should_stop) {
    is_cluster = 1;
    done_subdivision = FALSE;
    rcsg_created = FALSE;
    found_labels = FALSE;
    myScene = x;
    my_octree = THING_EMPTY;
    my_array = NULL;
    Vector::set(minbound, my_min);
    Vector::set(maxbound, my_max);
    Vector::sub(maxbound, minbound, bound_size);
    subdivision_level = my_subdivision;
    array_size = 1 << subdivision_level;
    
    if (subdivision_level > 10)
      fatal_error("UniformClusterPrimitive",
		  "uniform subdivision cannot be more than 10",
		  "The array that stores the cells in the uniform\n"
		  "subdivision is indexed with an unsigned int (32\n"
		  "bits). This means that you cannot exceed an array\n"
		  "size of the cube root of 32, i.e. 10 bits. To fix\n"
		  "this, rewrite the cell-skipping code to use a long\n"
		  "integer instead [will only allow up to 20 subdivision].");

    // Set up the increments to go from one cell of the array to the next.
    // Must allow for one cell's padding on each side of the cluster so that
    // the grid-traversal algorithm knows when to terminate.
    y_step = array_size + 2;
    z_step = SQR(array_size + 2);
       
    stop_at_first_intersection = should_stop;
    mark_for_deletion();
  }

  // Create a new cluster to encapsulate the given object graph.
  // Same as the above constructor except passed x, y, z values
  // instead of a vector.
  UniformClusterPrimitive(object x,
			  double x_min, double y_min, double z_min,
			  double x_max, double y_max, double z_max,
			  int my_subdivision, int should_stop) {
    is_cluster = 1;
    done_subdivision = FALSE;
    rcsg_created = FALSE;
    found_labels = FALSE;
    myScene = x;
    my_octree = THING_EMPTY;
    my_array = NULL;
    Vector::assign(minbound, x_min, y_min, z_min);
    Vector::assign(maxbound, x_max, y_max, z_max);
    Vector::sub(maxbound, minbound, bound_size);
    subdivision_level = my_subdivision;
    array_size = 1 << subdivision_level;
    
    if (subdivision_level > 10)
      fatal_error("UniformClusterPrimitive",
		  "uniform subdivision cannot be more than 10",
		  "The array that stores the cells in the uniform\n"
		  "subdivision is indexed with an unsigned int (32\n"
		  "bits). This means that you cannot exceed an array\n"
		  "size of the cube root of 32, i.e. 10 bits. To fix\n"
		  "this, rewrite the cell-skipping code to use a long\n"
		  "integer instead [will only allow up to 20 subdivision].");

    // Set up the increments to go from one cell of the array to the next.
    // Must allow for one cell's padding on each side of the cluster so that
    // the grid-traversal algorithm knows when to terminate.
    y_step = array_size + 2;
    z_step = SQR(array_size + 2);
       
    stop_at_first_intersection = should_stop;
    mark_for_deletion();
  }

  // Deletion method (invoked at the end of each frame).
  virtual ~UniformClusterPrimitive(void) {
    if (my_array != NULL)
      free(my_array);
  }
};

// Primitive 'full' : just full space (everything is IN).
class fullPrimitive : public Primitive {
public:
  virtual int ray_intersect(IntersectionList *the_list, 
			    primnode *node, primnode *validate_node,
			    vector ray_from, vector ray_dir, uint ray_flags,
			    signature ray_signature);
  virtual intersect_result point_intersect(primnode *node, vector apoint,
					   PointInfo *info,
					   uint ray_flags);
  virtual intersect_result voxel_intersect(box v);
  virtual int ray_intersect(vector ray_from,
			    vector ray_dir, rayintersect *int_storage);
  virtual intersect_result point_intersect(vector apoint);
  virtual char *prim_name(void);

  // This primitive takes no parameters when created.
  fullPrimitive(void) {
    is_cluster = 0;
    mark_for_deletion();
  }

  virtual ~fullPrimitive(void) {
  }
};

// Primitive 'nothing' : just empty space (everything is OUT).
class nothingPrimitive : public Primitive {
public:
  virtual int ray_intersect(IntersectionList *the_list, 
			    primnode *node, primnode *validate_node,
			    vector ray_from, vector ray_dir, uint ray_flags,
			    signature ray_signature);

  virtual intersect_result point_intersect(primnode *node, vector apoint,
					   PointInfo *info,
					   uint ray_flags);
  virtual intersect_result voxel_intersect(box v);
  virtual int ray_intersect(vector ray_from,
			    vector ray_dir, rayintersect *int_storage);
  virtual intersect_result point_intersect(vector apoint);
  virtual char *prim_name(void);

  // This primitive takes no parameters when created.
  nothingPrimitive(void) {
    is_cluster = 0;
    mark_for_deletion();
  }

  virtual ~nothingPrimitive(void) {
  }
};

// Primitive 'sphere' : a unit sphere centred on the origin.
class spherePrimitive : public Primitive {
public:
  virtual int ray_intersect(IntersectionList *the_list, 
			    primnode *node, primnode *validate_node,
			    vector ray_from, vector ray_dir, uint ray_flags,
			    signature ray_signature);

  virtual intersect_result point_intersect(primnode *node, vector apoint,
					   PointInfo *info,
					   uint ray_flags);
  virtual intersect_result voxel_intersect(box v);
  virtual int ray_intersect(vector ray_from,
			    vector ray_dir, rayintersect *int_storage);
  virtual intersect_result point_intersect(vector apoint);
  virtual char *prim_name(void);

  // This primitive takes no parameters when created.
  spherePrimitive(void) {
    is_cluster = 0;
    mark_for_deletion();
  }

  virtual ~spherePrimitive(void) {
  }
};

// Primitive 'cylinder' : an infinite cylinder of radius 1,
// centred on the Z axis.
class cylinderPrimitive : public Primitive {
public:
  virtual int ray_intersect(IntersectionList *the_list, 
			    primnode *node, primnode *validate_node,
			    vector ray_from, vector ray_dir, uint ray_flags,
			    signature ray_signature);

  virtual intersect_result point_intersect(primnode *node, vector apoint,
					   PointInfo *info,
					   uint ray_flags);
  virtual intersect_result voxel_intersect(box v);
  virtual int ray_intersect(vector ray_from,
			    vector ray_dir, rayintersect *int_storage);
  virtual intersect_result point_intersect(vector apoint);
  virtual char *prim_name(void);

  // This primitive takes no parameters when created.
  cylinderPrimitive(void) {
    is_cluster = 0;
    mark_for_deletion();
  }

  virtual ~cylinderPrimitive(void) {
  }
};

// Primitive 'plane' : the half-space defined by Z > 0.
class planePrimitive : public Primitive {
public:
  virtual int ray_intersect(IntersectionList *the_list, 
			    primnode *node, primnode *validate_node,
			    vector ray_from, vector ray_dir, uint ray_flags,
			    signature ray_signature);
  virtual intersect_result point_intersect(primnode *node, vector apoint,
					   PointInfo *info,
					   uint ray_flags);
  virtual intersect_result voxel_intersect(box v);
  virtual int ray_intersect(vector ray_from,
			    vector ray_dir, rayintersect *int_storage);
  virtual intersect_result point_intersect(vector apoint);
  virtual char *prim_name(void);

  // This primitive takes no parameters when created.
  planePrimitive(void) {
    is_cluster = 0;
    mark_for_deletion();
  }

  virtual ~planePrimitive(void) {
  }
};

// Primitive 'cone' : a double right-circular cone oriented along Z.
class conePrimitive : public Primitive {
public:
  virtual int ray_intersect(IntersectionList *the_list, 
			    primnode *node, primnode *validate_node,
			    vector ray_from, vector ray_dir, uint ray_flags,
			    signature ray_signature);

  virtual intersect_result point_intersect(primnode *node, vector apoint,
					   PointInfo *info,
					   uint ray_flags);
  virtual intersect_result voxel_intersect(box v);
  virtual int ray_intersect(vector ray_from,
			    vector ray_dir, rayintersect *int_storage);
  virtual intersect_result point_intersect(vector apoint);
  virtual char *prim_name(void);

  // This primitive takes no parameters when created.
  conePrimitive(void) {
    is_cluster = 0;
    mark_for_deletion();
  }

  virtual ~conePrimitive(void) {
  }
};

// Primitive 'unitCylinder' : a cylinder of radius 1, centered on the Z
// axis and bounded by planes at Z = 0 and Z = +1.
class unitCylinderPrimitive : public Primitive {
public:
  virtual int ray_intersect(IntersectionList *the_list, 
			    primnode *node, primnode *validate_node,
			    vector ray_from, vector ray_dir, uint ray_flags,
			    signature ray_signature);

  virtual intersect_result point_intersect(primnode *node, vector apoint,
					   PointInfo *info,
					   uint ray_flags);
  virtual intersect_result voxel_intersect(box v);
  virtual int ray_intersect(vector ray_from,
			    vector ray_dir, rayintersect *int_storage);
  virtual intersect_result point_intersect(vector apoint);
  virtual char *prim_name(void);

  // This primitive takes no parameters when created.
  unitCylinderPrimitive(void) {
    is_cluster = 0;
    mark_for_deletion();
  }

  virtual ~unitCylinderPrimitive(void) {
  }
};

// Primitive 'torus' : a torus with defining circle in the XY plane with
// radius 1; cross-section has radius torus_rad (passed upon primitive
// creation).
class torusPrimitive : public Primitive {
public:
  double torus_rad;

  virtual int ray_intersect(IntersectionList *the_list, 
			    primnode *node, primnode *validate_node,
			    vector ray_from, vector ray_dir, uint ray_flags,
			    signature ray_signature);

  virtual intersect_result point_intersect(primnode *node, vector apoint,
					   PointInfo *info,
					   uint ray_flags);
  virtual intersect_result voxel_intersect(box v);
  virtual int ray_intersect(vector ray_from,
			    vector ray_dir, rayintersect *int_storage);
  virtual intersect_result point_intersect(vector apoint);
  virtual char *prim_name(void);

  // When created, this primitive requires a cross-section radius.
  torusPrimitive(double my_rad) {
    is_cluster = 0;
    torus_rad = my_rad;
    mark_for_deletion();
  }

  virtual ~torusPrimitive(void) {
  }
};

// Primitive 'cube' : a cube at the origin from X, Y, Z +1 to -1.
class cubePrimitive : public Primitive {
public:
  virtual int ray_intersect(IntersectionList *the_list, 
			    primnode *node, primnode *validate_node,
			    vector ray_from, vector ray_dir, uint ray_flags,
			    signature ray_signature);

  virtual intersect_result point_intersect(primnode *node, vector apoint,
					   PointInfo *info,
					   uint ray_flags);
  virtual intersect_result voxel_intersect(box v);
  virtual int ray_intersect(vector ray_from,
			    vector ray_dir, rayintersect *int_storage);
  virtual intersect_result point_intersect(vector apoint);
  virtual char *prim_name(void);

  // This primitive takes no parameters when created.
  cubePrimitive(void) {
    is_cluster = 0;
    mark_for_deletion();
  }

  virtual ~cubePrimitive(void) {
  }
};

// Primitive 'truncCone' : a single right-circular cone with radius 1 at
// the origin, decreasing towards a vertex at (0, 0, 1) but truncated
// by a plane at Z = zval where zval is somewhere in (0, 1].
class truncConePrimitive : public Primitive {
public:
  double zval;

  virtual int ray_intersect(IntersectionList *the_list, 
			    primnode *node, primnode *validate_node,
			    vector ray_from, vector ray_dir, uint ray_flags,
			    signature ray_signature);

  virtual intersect_result point_intersect(primnode *node, vector apoint,
					   PointInfo *info,
					   uint ray_flags);
  virtual intersect_result voxel_intersect(box v);
  virtual int ray_intersect(vector ray_from,
			    vector ray_dir, rayintersect *int_storage);
  virtual intersect_result point_intersect(vector apoint);
  virtual char *prim_name(void);

  truncConePrimitive(double z) {
    zval = z;
    if ((z <= 0) || (z > 1))
      fatal_error("truncConePrimitive::constructor",
		  "Z value must be in (0, 1] !",
		  "The truncated cone primitive can only represent cones\n"
		  "that taper from one value to another, not any of the\n"
		  "self-intersecting or upside-down forms. You may be\n"
		  "passing incorrect parameters to the new_trunc_cone\n"
		  "routine.");
    is_cluster = 0;
    mark_for_deletion();
  }

  virtual ~truncConePrimitive(void) {
  }
};

// Primitive 'taper' : a tapering segment of a torus from theta = 0 (along X)
// to theta = bend_amount. The radius of the torus changes from rad_base to
// rad_top along its length. The torus's defining circle has radius 1.
class taperPrimitive : public Primitive {
public:
  double bend_amount;
  double rad_base, rad_top;
  vector end_p, end_n;
  int traversal_num_steps;
  double binary_search_limit;

  virtual int ray_intersect(IntersectionList *the_list, 
			    primnode *node, primnode *validate_node,
			    vector ray_from, vector ray_dir, uint ray_flags,
			    signature ray_signature);

  virtual intersect_result point_intersect(primnode *node, vector apoint,
					   PointInfo *info,
					   uint ray_flags);
  virtual intersect_result voxel_intersect(box v);
  virtual int ray_intersect(vector ray_from,
			    vector ray_dir, rayintersect *int_storage);
  virtual intersect_result point_intersect(vector apoint);
  virtual char *prim_name(void);

  // When created, this primitive requires two radii, an angle and
  // an approximation count.
  taperPrimitive(double bend, double r1, double r2, int steps, double limit) {
    is_cluster = 0;
    bend_amount = bend;
    rad_base = r1;
    rad_top = r2;

    traversal_num_steps = steps;
    binary_search_limit = limit;
    
    // Calculate a point at the end of the taper and the normal to the plane
    // that forms its end.
    Vector::assign(end_p, cos(bend_amount), sin(bend_amount), 0);
    Vector::assign(end_n, -end_p[Y], end_p[X], 0);
    mark_for_deletion();
  }

  virtual ~taperPrimitive(void) {
  }
};

// The Cutout class is basically a wrapper for a function that determines
// whether or not a given point is inside an object's outline.
//
// It is used by the bilinearPrimitive and twoSidedPatchPrimitive.
class Cutout : public DeletableObject {
public:

  virtual int is_inside(double x, double y) = 0;
};

// An ImageCutout is a Cutout that uses a GenericPixmap image to
// create the outline of an object.
class ImageCutout : public Cutout {
public:
  GenericPixmap *bound_image;

  // Is a given point inside the image?
  virtual int is_inside(double x, double y);

  // The constructors for the ImageCutout.
  ImageCutout(char *fname);
  ImageCutout(GenericPixmap *pix);
};

// Primitive 'bilinear' : a bilinear patch defined by the points (0, 0, 0),
// (1, 0, 0), (0, 1, 0), (x_point, y_point, 1).
//
// The bilinear primitive can also take a Cutout parameter that determines
// the outline of the patch.
class bilinearPrimitive : public Primitive {
public:
  double x_point, y_point;
  int is_flat;
  Cutout *cutout;

  virtual int ray_intersect(IntersectionList *the_list, 
			    primnode *node, primnode *validate_node,
			    vector ray_from, vector ray_dir, uint ray_flags,
			    signature ray_signature);

  virtual intersect_result point_intersect(primnode *node, vector apoint,
					   PointInfo *info,
					   uint ray_flags);
  virtual intersect_result voxel_intersect(box v);
  virtual int ray_intersect(vector ray_from,
			    vector ray_dir, rayintersect *int_storage);
  virtual intersect_result point_intersect(vector apoint);
  virtual char *prim_name(void);

  // When created, this primitive requires X and Y coordinates for point 3.
  bilinearPrimitive(double x, double y, int flat) {
    is_cluster = 0;
    x_point = x;
    y_point = y;
    is_flat = flat;

    cutout = NULL;

    mark_for_deletion();
  }

  // When created, this primitive requires X and Y coordinates for point 3.
  // It also takes a Cutout parameter to determine its outline.
  bilinearPrimitive(double x, double y, int flat, Cutout *cut) {
    is_cluster = 0;
    x_point = x;
    y_point = y;
    is_flat = flat;

    cutout = cut;

    mark_for_deletion();
  }

  virtual ~bilinearPrimitive(void) {
  }
};

// A routine to create a new patch and place it so its corners are at
// any four specified points.
object new_bilinear_patch(vector p0, vector p1, vector p2, vector p3);

// A routine to create a new patch and place it so its corners are at
// any four specified points. Also has a Cutout parameter.
object new_bilinear_patch(vector p0, vector p1, vector p2, vector p3,
			  Cutout *cut);

// A routine to create a new patch and place it so its corners are at
// any four specified points. Places a PatchImageTexture on the patch
// with the given GenericPixmap.
object new_bilinear_patch(vector p0, vector p1, vector p2, vector p3,
			  int tex_prec, GenericPixmap *tex_pix,
			  texture_method tex_method, int tex_filter);

// Routine to determine a patch's parameters and position matrix
// to place it in the right place.
void position_patch(vector p0, vector p1, vector p2, vector p3,
		    int *flat_patch, double *x_param, double *y_param,
		    transform position);

// The twoSidedPatch primitive : a bilinear patch whose normal isn't flipped to
// always point towards the incoming ray. Thus you can use a material
// property that has two distinct sides to it.
//
// The twoSidedPatch can also take a Cutout parameter that determines
// the outline of the patch.
class twoSidedPatchPrimitive : public Primitive {
public:
  double x_point, y_point;
  int is_flat;
  Cutout *cutout;

  virtual int ray_intersect(IntersectionList *the_list, 
			    primnode *node, primnode *validate_node,
			    vector ray_from, vector ray_dir, uint ray_flags,
			    signature ray_signature);

  virtual intersect_result point_intersect(primnode *node, vector apoint,
					   PointInfo *info,
					   uint ray_flags);
  virtual intersect_result voxel_intersect(box v);
  virtual int ray_intersect(vector ray_from,
			    vector ray_dir, rayintersect *int_storage);
  virtual intersect_result point_intersect(vector apoint);
  virtual char *prim_name(void);

  // When created, this primitive requires X and Y coordinates for point 3.
  twoSidedPatchPrimitive(double x, double y, int flat) {
    is_cluster = 0;
    x_point = x;
    y_point = y;
    is_flat = flat;

    cutout = NULL;

    mark_for_deletion();
  }

  // When created, this primitive requires X and Y coordinates for point 3.
  // It also takes a Cutout parameter to determine its outline.
  twoSidedPatchPrimitive(double x, double y, int flat, Cutout *cut) {
    is_cluster = 0;
    x_point = x;
    y_point = y;
    is_flat = flat;

    cutout = cut;

    mark_for_deletion();
  }

  virtual ~twoSidedPatchPrimitive(void) {
  }
};

// A routine to create a new 2-sided patch and place it so its corners are at
// any four specified points.
object new_two_sided_patch(vector p0, vector p1, vector p2, vector p3);

// A routine to create a new 2-sided patch and place it so its corners are at
// any four specified points. Also takes a Cutout parameter.
object new_two_sided_patch(vector p0, vector p1, vector p2, vector p3,
			   Cutout *cut);


