/* prototypes.h : Function prototypes for the Mirage.
** 
** This file holds all of the function prototypes, organized by file.
**
** Chris Butcher, 27 November 1996
*/

// main.c
// Main routines.
void init_system(void);
void init_system(char *keyfilename);
void init_system(int argc, char **argv);
void init_system(int argc, char **argv, char *keyfilename);
void parse_commandline(int argc, char **argv);
void process_keyfile(char *keyfilename);
void shutdown_mirage(void);

void create_animation(void);
void create_animation(char *file_template);

void set_worldmaker(object (*worldmaker)(object *active_camera));
void print_usage(char *name);
void RCSG_backtraverse(void);
#ifdef UNIX_SPECIFIC
void find_username(void);
void torment_user(void);
#endif



// volutils.c
// Routines to create volume primitives of different types; extrusions,
// revolutions.

Volume *create_file_extrusion(char *filename, int stop_at_first);
Volume *create_revolution_surface(int x_res, int y_res, int z_res,
				  double t_min, double t_max,
				  double (*func)(double tval),
				  int should_stop);


// message.c
// Routines to send debugging, warning and error messages to the user.

void output_warning(char *location, char *warning, char *description);
void zerowarn(char *location);
// fatal_error is in constants.h as it needs to be included early on;
// lots of header files use it.
void output_debug(char *debug_message);
void debug_vector(char *vec_name, vector v);
void debug_matrix(char *matrix_name, matrix m);
void debug_transform(char *transname, transform t);
void debug_dag(char *dagname, object todebug);
void print_current_time(char *event);
void print_color(color col);
void print_banner(void);
void debug_thing(thing the_thing, int indent_distance);
void debug_grid(thing *a_grid, int grid_size);
void vox_print(box b);
#ifdef UNIX_SPECIFIC
FILE *send_email(char *address, char *subject);
void notify_by_email(char *address);
#endif

// keyfile.c
// Routines to read in a key file and to extract the data items from it.

void read_keyfile(char *keyfilename);
// get_keyfile_data is in constants.h as it needs to be included early on;
// lots of header files use it.
int get_keyfile_toggle(char *key_item, int *to_get);
int get_keyfile_number(char *key_item, int *to_get);
int get_keyfile_float(char *key_item, double *to_get);
int get_keyfile_vector(char *key_item, vector to_get);


// write.c
// Routines to write output files.

char *filename_oscheck(char *fname);
void writeOut(Storage *myStorage, char *file_template, filetype file_type,
	      int frameNumber, char *postprocess_command);
void writeOut(Storage *myStorage, char *fname, filetype file_type,
	      char *postprocess_command);
void write_ppm_to_file(FILE *out_file, ColorStorage *aStorage);
FILE *open_output_file(char *filename);

#ifdef PNG_ON
void write_png_to_file(FILE *out_file, ColorStorage *aStorage);
#endif



// scene.c
// Routines to prepare the scene for rendering.

void scene_find_labels(object theScene);
void scene_create_RCSG(object theScene, primnode *convertedScene);
void scene_subdivide(primnode *theScene);


// prim_cluster.c
// Routines for cluster primitives to use.

int rcsg_ray_intersect(IntersectionList *the_list, thing the_thing,
		       primnode *cluster_primnode,
		       vector ray_from, vector ray_dir, uint ray_flags,
		       signature ray_signature);
void classify_intersection_in_rcsg(thing the_thing, intersection *the_int,
				   vector hit_point, intersect_result *result,
				   IntersectInfo *int_info, uint ray_flags);
intersect_result classify_point_in_rcsg(thing the_thing, vector the_point,
					PointInfo *point_info,
					uint ray_flags);
intersect_result classify_voxel_in_rcsg(thing the_thing, box b);
int check_ray_flags(uint ray_flags, uint prim_flags);
void tidy_intersections(IntersectionList *the_list, 
			primnode *cluster_primnode);
int validate_intersections(IntersectionList *the_list,
			   primnode *cluster_primnode,
			   primnode *next_cluster,
			   vector ray_from, vector ray_dir,
			   uint ray_flags, validate_type v_type,
			   thing the_thing,
			   double cell_entry_t, double cell_exit_t);
void find_point_sfx(primnode *world_object, vector world_point,
		    uint ray_flags, SfxType *sfxtype, void **sfx);


// noise.c
// Routines for filtered pseudo-random noise.

void initialise_noise(unsigned int extent);
int noisept_index(double noisept);
double noise_1(double samplept);
double noise_1grad(double samplept, double *gradient);
double noise_2(vector samplept);
double noise_2grad(vector samplept, vector gradient);
double noise_3(vector samplept);
double noise_3grad(vector samplept, vector gradient);
void vector_noise3(vector samplept, vector noiseval);


// polynomial.c
// Routines for finding roots of polynomial equations.
int solveQuadric(double *q_coeff, double *q_sol);
int solveCubic(double *q_coeff, double *q_sol);
int solveQuartic(double *q_coeff, double *q_sol);
#ifdef USE_OWN_CBRT
double cbrt(double x);
#endif


// camera.c
// Routines related to the camera code.
object create_default_camera(int x_size, int y_size, int antialiasing,
			     double field_of_view);
object create_default_camera(int antialiasing, double field_of_view);


// label.c
// Routines for finding labels recursively in the object graph.
void recursive_find_labels(object theDag, transform existing_transform);


// raytrace.c
// Routines for tracing rays.
signature new_ray_signature(void);


// prim_cube, cone, cylinder etc.
// Routines for primitive intersection procedures to use.
intersect_result cylinder_voxel_intersect(box b);
intersect_result cone_voxel_intersect(box b);
intersect_result cube_voxel_intersect(box b);
intersect_result torus_voxel_intersect(double torus_rad, box b);

#ifdef USE_OWN_DRAND48

// drand48 routines' prototypes. Only necessary if drand48 does not exist on
// the target system.
void srand48(long seed);
unsigned short *seed48(unsigned short xseed[3]);
long nrand48(unsigned short xseed[3]);
long mrand48(void);
long lrand48(void);
void lcong48(unsigned short p[7]);
long jrand48(unsigned short xseed[3]);
double erand48(unsigned short xseed[3]);
double drand48(void);
void __mirage_dorand48(unsigned short xseed[3]);

#endif // USE_OWN_DRAND48

