/* quaternion.h : Quaternion manipulation utilities
** 
** The Quaternion class is a toolkit of quaternion functions that is used by
** the rendering system. They are accessed as Quaternion::function(a, b, c)
** e.g. Quaternion::mult(rot1, rot2, total);
**
** Chris Butcher, 5 March 1997
*/

// Define the quaternion storage. Note that a 'quaternion' is a different
// thing to the 'Quaternion' toolkit class ...
// Quaternions are stored as an array of 4 doubles (1, i, j, k).
typedef double quaternion[4];

// This class should never be instantiated. It is a toolkit only.
// Most of these are inline functions so will be compiled into tolerably
// quick code.
class Quaternion {
public:

  // Adds two quaternions a and b, stores the result in a third quaternion
  // which may be one of a or b.
  static void add(quaternion a, quaternion b, quaternion result)
  {
    result[X] = a[X] + b[X];
    result[Y] = a[Y] + b[Y];
    result[Z] = a[Z] + b[Z];
    result[W] = a[W] + b[W];
  }

  // Add a fraction of quaternion a to some quaternion b.
  static void addfrac(double frac, quaternion a, quaternion b)
  {
    b[X] += frac * a[X];
    b[Y] += frac * a[Y];
    b[Z] += frac * a[Z];
    b[W] += frac * a[W];
  }

  // Subtracts two quaternions a and b, stores the result in a third
  // quaternion which may be one of a or b.
  static void sub(quaternion a, quaternion b, quaternion result)
  {
    result[X] = a[X] - b[X];
    result[Y] = a[Y] - b[Y];
    result[Z] = a[Z] - b[Z];
    result[W] = a[W] - b[W];
  }

  // Sets quaternion a to quaternion b.
  static void set(quaternion a, quaternion b)
  {
    a[X] = b[X];
    a[Y] = b[Y];
    a[Z] = b[Z];
    a[W] = b[W];
  }

  // Assigns the given values to the quaternion a.
  static void assign(quaternion a, double sval,
		     double xval, double yval, double zval)
  {
    a[W] = sval;
    a[X] = xval;
    a[Y] = yval;
    a[Z] = zval;
  }

  // Returns the magnitude of the quaternion a.
  static double magnitude(quaternion a)
  {
    return sqrt(SQR(a[X]) + SQR(a[Y]) + SQR(a[Z]) + SQR(a[W]));
  }

  // Scales a quaternion a by the given scalar quantity s, stores it in result.
  inline static void mult(double s, quaternion a, quaternion result)
  {
    result[X] = s * a[X];
    result[Y] = s * a[Y];
    result[Z] = s * a[Z];
    result[W] = s * a[W];
  }

  // Scales a quaternion a by the given scalar quantity s, stores it in result.
  static void scale(double s, quaternion a, quaternion result)
  {
    result[X] = s * a[X];
    result[Y] = s * a[Y];
    result[Z] = s * a[Z];
    result[W] = s * a[W];
  }

  // Normalises a quaternion.
  static void normalise(quaternion q, char *msg)
  {
    Quaternion::scale(1.0 / CHECK0(Quaternion::magnitude(q), msg), q, q);
  }

  // Get the multiplicative inverse of a quaternion.
  static void inverse(quaternion q, quaternion inv, char *msg)
  {
    double m_inv = 1.0 / CHECK0(Quaternion::magnitude(q), msg);

    // Scalar component
    inv[W] = +m_inv * q[W];

    // Vector components
    inv[X] = -m_inv * q[X];
    inv[Y] = -m_inv * q[Y];
    inv[Z] = -m_inv * q[Z];
  }

  // Multiply a quaternion by another.
  static void mult(quaternion a, quaternion b, quaternion result)
  {
    // Temporary variable so we don't mangle either of the source quaternions
    // (e.g. if a == result or b == result).
    quaternion temp_result;

    // Scalar portion of the result is a_s * b_s - a_v . b_v
    temp_result[W] = a[W] * b[W] - Vector::dot(a, b);

    // Vector portion of the result is a_s * a_v + b_s * b_v + a_v x b_v
    Vector::cross(a, b, temp_result);
    Vector::addfrac(b[W], a, temp_result);
    Vector::addfrac(a[W], b, temp_result);

    // Copy the result into its final location.
    Quaternion::set(result, temp_result);
  }

  // Apply a quaternion to a vector.
  static void apply(quaternion q, vector v, vector result)
  {
    quaternion new_v, q_inv;

    // Form the quaternion with 0 scalar, vector == v.
    Quaternion::assign(new_v, 0, v[X], v[Y], v[Z]);

    Quaternion::inverse(q, q_inv, "Zero quaternion to inverse");

    // We apply the quaternion as q * v * q-1
    Quaternion::mult(q, new_v, new_v);
    Quaternion::mult(new_v, q_inv, new_v);

    // Move the rotated quaternion into the result vector.
    Vector::set(result, new_v);
  }

  // Convert a quaternion into a matrix representation.
  static void form_matrix(quaternion q, matrix m)
  {
    // We look at the quaternion's effects on the unit basis vectors.
    // This gives us the rotation matrix for q = [S, X, Y, Z]:
    //
    //  (1 - 2Y^2 - 2Z^2     2XY - 2SZ         2XZ + 2SY  )
    //  (   2XY + 2SZ     1 - 2X^2 - 2Z^2      2YZ - 2SX  )
    //  (   2XZ - 2SY        2YZ + 2SX     1 - 2X^2 - 2Y^2)
    double qs = q[W], qx = q[X], qy = q[Y], qz = q[Z];
      
    Matrix::assign(m,
		   1-2*SQR(qy)-2*SQR(qz), 2*qx*qy-2*qs*qz, 2*qx*qz+2*qs*qy, 0,
		   2*qx*qy+2*qs*qz, 1-2*SQR(qx)-2*SQR(qz), 2*qy*qz-2*qs*qx, 0,
		   2*qx*qz-2*qs*qy, 2*qy*qz+2*qs*qx, 1-2*SQR(qx)-2*SQR(qy), 0);
  }
  
  // Convert a quaternion into a transform representation.
  static void form_transform(quaternion q, transform t)
  {
    quaternion q_inv;

    // Get the inverse quaternion.
    Quaternion::inverse(q, q_inv, "Zero quaternion forming transform");

    // Convert the quaternion pair into a transform.
    Quaternion::form_matrix(q,     t[FORE]);
    Quaternion::form_matrix(q_inv, t[AFT]);
  }

  // Set the quaternion to the multiplicative identity.
  static void unit(quaternion q)
  {
    Quaternion::assign(q, 1, 0, 0, 0);
  }

  static double dot(quaternion a, quaternion b)
  {
    return (a[W]*b[W] + a[X]*b[X] + a[Y]*b[Y] + a[Z]*b[Z]);
  }
    
  static void interpolate(quaternion a, double frac, quaternion b,
			  quaternion result)
  {
    double theta, sintheta, fwd, back, costheta;
    quaternion temp_result;

    // We first find the angle between the quaternions to be interpolated.
    costheta = Quaternion::dot(a, b);

    if (1.0 + costheta > -1e-06) {
      // cos(theta) is not close to -1.
      
      if (1.0 - costheta > 1e-06) {
	// cos(theta) is not close to +1.

	// Thus we can use the normal method of interpolation.
	theta = acos(Quaternion::dot(a,b));

	// Interpolate using the formula
	// (sin ((1-t)*theta) / sin (theta))*a + (sin ((t*theta) / sin (theta))*b
	sintheta = sin(theta);

	fwd  = sin(     frac  * theta) / sintheta;
	back = sin((1 - frac) * theta) / sintheta;
      } else {
	// cos(theta) is close to +1. Thus the vectors are very close together
	// and we should use linear interpolation.
	fwd  = frac;
	back = 1 - frac;
      }

      Quaternion::mult(fwd, b, temp_result);
      Quaternion::addfrac(back, a, temp_result);
    } else {
      // cos(theta) is close to -1; the vectors are nearly opposite.
      // We use an alternative interpolation method.

      Quaternion::assign(temp_result, a[Z], -a[Y], a[X], -a[W]);
      
      fwd = sin(frac * PI);
      back = sin((0.5 - frac) * PI);

      Quaternion::mult(fwd, temp_result, temp_result);
      Quaternion::addfrac(back, a, temp_result);
    }

    // Copy from the temp result into the final quaternion.
    Quaternion::set(result, temp_result);
  }
    
  // Raise a *unit* quaternion to a power. This is used for multiplying
  // a rotation by some amount ... e.g. rotate it 2.4 times as far.
  //
  // Note that power(q, -1, q') is the same as inverse(q, q').
  static void power(quaternion a, double power, quaternion result)
  {
    double theta, sintheta, fwd, back;
    quaternion temp;

    Quaternion::set(temp, a);
    if (fabs(temp[W]) > 1)
      // temp MUST be a unit quaternion or else acos() will return NaN and
      // Bad Things will happen.
      Quaternion::normalise(temp, "Quaternion::power arithmetic error");

    // We first find the angle between the quaternion and the unit quaternion.
    theta = acos(temp[W]);

    sintheta = sin(theta);

    if (sintheta == 0) {
      // Formula does not apply for the unit quaternion or its inverse
      // (-1, 0, 0, 0).
      Quaternion::set(result, temp);
    } else { 
      // Interpolate from the unit vector to a using the formula as above
      // in Quaternion::interpolate.
      fwd  = sin(     power  * theta) / sintheta;
      back = sin((1 - power) * theta) / sintheta;
      
      Quaternion::mult(fwd, temp, result);
      result[W] += back;
    }
  }

  // Premultiply a fraction of a quaternion with another quaternion. Used for
  // taking an original rotation and adding in a fraction of another
  // rotation.
  static void multfrac(double frac, quaternion a, quaternion total)
  {
    quaternion temp;

    Quaternion::power(a, frac, temp);
    Quaternion::mult(temp, total, total);
  }

  // Turn a rotation matrix into a quaternion.
  static void from_matrix(matrix m, quaternion q)
  {
    axis i, j, k;
    double s;

    // Work out which entry along the trace is the greatest.
    // This makes the method more numerically stable.
    if (m[X][X] > m[Y][Y]) {
      if (m[X][X] > m[Z][Z]) {
	i = X; j = Y; k = Z;
      } else {
	i = Z; j = X; k = Y;
      }
    } else {
      if (m[Y][Y] > m[Z][Z]) {
	i = Y; j = Z; k = X;
      } else {
	i = Z; j = X; k = Y;
      }
    }
    
    // Since this is a rotation matrix we are guaranteed that
    // m[i][i] etc will be from [0, 1]. And we know m[i][i] is the
    // biggest of the three -> s >= 0.
    s = sqrt((m[i][i] - (m[j][j] + m[k][k])) + 1.0);

    if (s == 0) {
      // m[i][i] == m[j][j] == m[k][k] == 1 or the matrix is not a
      // pure rotation.
      Quaternion::unit(q);
    } else {
      q[i] = s * 0.5;

      s = 0.5 / s;
      
      q[W] = (m[k][j] - m[j][k]) * s;
      q[j] = (m[j][i] + m[i][j]) * s;
      q[k] = (m[k][i] + m[i][k]) * s;
    }
  }

  // Construct a unit rotation quaternion from theta and phi values in radians.
  static void theta_phi_quaternion(double theta, double phi,
				   quaternion q)
  {
    double sinphi = sin(phi/2);

    // NB: a rotation towards theta has an axis that is 90 degrees
    // further around anticlockwise. So we add PI / 2 to theta to compute
    // the axis vector.
    Quaternion::assign(q, cos(phi/2),
		       sinphi*cos(theta + PI/2), sinphi*sin(theta + PI/2), 0);
  }
};

