/* raytrace.c : The StandardRayTracer class.
** 
** This file defines the functions necessary to trace rays.
**
** Chris Butcher, 4 December 1996
*/

#include "mirage.h"

#ifdef USE_NAMESPACES
namespace Mirage {
#endif

#define DEBUG_RT         0

// The point-contents that are everywhere that objects aren't.
PointInfo etherstruct, *ether = &etherstruct;

// The background colour that is used if there isn't
// a background_texture. Default is nasty green.
color background_color = {0.3, 1.0, 0.7};

// Ray counting variables.
int counter_primary_rays, counter_secondary_rays, counter_shadow_rays;
int counter_initial_rays, counter_antialiasing_rays;
signature next_ray_signature = 1;

// Return a new ray signature.
signature new_ray_signature(void)
{
  return next_ray_signature++;
}

void StandardRayTracer::traceRay(primnode *world_object,
				 vector ray_from, vector ray_dir,
				 uint ray_flags, signature ray_signature,
				 SfxType start_sfxtype, void *start_sfx,
				 double light_t, color ray_contribution,
				 primnode **hit_object,
				 double *hit_t, vector hit_normal,
				 vector hit_point, color hit_color,
				 int max_ray_depth)
{
  int ray_is_shadow, delete_frommat, delete_tomat, i;
  uint int_flags;
  primnode *node;
  double sfx_start_t = 0;
  Material *fromMaterial, *toMaterial;
  IntersectionList *intersect_list;
  intersection *current_intersection, *list_position;
  SfxType current_shadow_sfxtype = start_sfxtype;
  void *current_shadow_sfx = start_sfx;

  // Is this a shadow ray?
  ray_is_shadow = ray_flags & RAY_SHADOW;

#if DEBUG_RT
  printf("RT in\n");
#endif

  if (ray_is_shadow) {
    // Initially all light reaches the object in question.
    for (i = 0; i < num_color_components; i++)
      hit_color[i] = 1.0;
  } else {
    // If this object ray isn't going to make an appreciable difference to the
    // image, then don't bother tracing it. We return the background instead.
    if (Vector::dot(ray_contribution, ray_contribution)
	< SQR(RAY_CONTRIBUTION_CUTOFF)) {

      // The ray hits the background.
      Color::set(hit_color, background_color);
      
      // Call the background texture if it exists.
      if (apply_textures_to_objects && (background_texture != NULL))
	background_texture->apply(ray_from, ray_dir, ray_flags,
				  hit_color, this);

      return;
    }
  }

  // Update the primary / secondary / shadow ray counters.
  if ((ray_flags & RAY_PRIMARYMASK) == 0)
    counter_primary_rays++;
  else {
    counter_secondary_rays++;
    if (ray_is_shadow) {
      counter_shadow_rays++;
    }
  }

  // Intersect with the worldscene ... place all intersctions with
  // necessary primitives in the intersect_list.
  intersect_list = new_intersection_list();
  intersect_list->init();

  list_position = intersect_list->get_position();

#if DEBUG_RT
  printf("RT trace %s\n", world_object->prim->prim_name());
#endif

  world_object->prim->ray_intersect(intersect_list, world_object, NULL,
				    ray_from, ray_dir, ray_flags,
				    ray_signature);

#if DEBUG_RT
  printf("RT traced\n");
#endif

  intersect_list->set_position(list_position);

  // Loop around all the intersections.
  while (1) {

#if DEBUG_RT
    printf("RT loop\n");
#endif

    // Check to see if there are more intersections.
    if (intersect_list->empty()) {
      // The ray hits the background.
      *hit_object = NULL;
      *hit_t = HIT_BACKGROUND;

      if (!ray_is_shadow) {
	// Set the background colour and then call the background texture,
	// if it exists.
	Color::set(hit_color, background_color);
	
	if (background_texture != NULL)
	  background_texture->apply(ray_from, ray_dir, ray_flags,
				    hit_color, this);
      }
      
      // Apply any gaseous effects along the ray.
      if (ray_is_shadow) {
	if (current_shadow_sfxtype == GASEOUS) {
	  GasSFX *current_gas = (GasSFX *) current_shadow_sfx;
	  
	  // Apply gaseous effect from the light to the start of the
	  // gaseous effect (at sfx_start_t).
	  current_gas->gas->traverse_gas(current_gas->gas_transform,
					 ray_from, ray_dir, hit_color,
					 sfx_start_t, light_t);
	}
      } else {
	if (start_sfxtype == GASEOUS) {
	  GasSFX *current_gas = (GasSFX *) start_sfx;
	  
	  // Apply gaseous effect from the ray start point to the background.
	  current_gas->gas->traverse_gas(current_gas->gas_transform,
					 ray_from, ray_dir, hit_color,
					 0, HIT_BACKGROUND);
	}
      }
      
      // No more intersections.
      break;
    }
    
#if DEBUG_RT
    printf("RT get\n");
#endif

    // Get the next intersection from the list.
    current_intersection = intersect_list->get_next();

#if DEBUG_RT
    printf("RT got tval %f\n", current_intersection->t_value);
#endif

    node = current_intersection->primitivehit;
    int_flags = current_intersection->hit_flags;

    // If we are a shadow ray and the t value for this intersection is greater
    // than the t value of the light (passed in light_t) then we have reached
    // the light, stop raytracing.
    if (ray_is_shadow && current_intersection->t_value > light_t) {
      // Apply current gaseous SFX out to light_t.
      if (current_shadow_sfxtype == GASEOUS) {
	GasSFX *current_gas = (GasSFX *) current_shadow_sfx;

	current_gas->gas->traverse_gas(current_gas->gas_transform,
				       ray_from, ray_dir, hit_color,
				       sfx_start_t, light_t);
      }
#if DEBUG_RT
      printf("RT at light_t, breaking\n");
#endif
      break;
    }

    // If we are a shadow ray and we encounter a solid object which isn't
    // translucent, then it blocks all the light. Return no light.
    if (ray_is_shadow && ((int_flags & FLAGS_TRANSLUCENT) == 0)) {
      for (i = 0; i < num_color_components; i++)
	hit_color[i] = 0.0;
      // We are a shadow ray, the only thing we want to return is hit_color.
      // So we don't need to set primnode, t-value etc, just break.
#if DEBUG_RT
      printf("RT found solid at %f, shadowed by %s.\n",
	     current_intersection->t_value,
	     current_intersection->real_primnode->prim->prim_name());
#endif
      break;
    }

#if DEBUG_RT
    printf("RT values\n");
#endif

    // Set the correct values in hit_object, hit_t, hit_normal, hit_point.
    // This is necessary for passing to texture functions.
    *hit_object = node;
    
    *hit_t = current_intersection->t_value;
    Matrix::applytranspose(node->position[AFT], current_intersection->normal,
			   hit_normal);
    Vector::mult(*hit_t, ray_dir, hit_point);
    Vector::add(ray_from, hit_point, hit_point);

#if DEBUG_RT
    printf("RT ether\n");
#endif

    // If either the from contents or the to contents are 'empty' then we
    // set them to be the 'ether' point contents, which is defined to
    // be present everywhere that objects aren't.
    if (current_intersection->info.from.empty) {
#if DEBUG_RT
      printf("RT: from ether\n");
#endif
      copy_point_info(&(current_intersection->info.from), ether);
    }
    if (current_intersection->info.to.empty) {
#if DEBUG_RT
      printf("RT: to ether\n");
#endif
      copy_point_info(&(current_intersection->info.to),   ether);
    }

    // We haven't allocated any Material classes yet.
    delete_frommat = delete_tomat = FALSE;

#if DEBUG_RT
    printf("RT texture\n");
#endif

    // If there are textures on the materials, apply them.
    if (apply_textures_to_objects) {
      // Apply the from texture.
      if (current_intersection->info.from.tex_multi) {
	// There is more than one texture on this object, in
	// a collection.
#if DEBUG_RT
	printf("RT from-multitex\n");
#endif
	if (current_intersection->info.from.tex_col != NULL) {
#if DEBUG_RT
	  printf("RT apply tex\n");
#endif
	  // Make a copy of the material properties and apply the texture
	  // functions in the TextureCollection attached to this primitive.
	  current_intersection->info.from.mat->copy(&fromMaterial);
	  current_intersection->info.from.tex_col->apply(ray_from, ray_dir,
							 ray_flags,
							 fromMaterial,
							 hit_point,
							 hit_normal, this);
	  delete_frommat = TRUE;
	} else {
	  // We want to apply the unchanged material.
	  fromMaterial = current_intersection->info.from.mat;
	}
      } else {
#if DEBUG_RT
	printf("RT from-singletex\n");
#endif
	if (current_intersection->info.from.texture != NULL) {
#if DEBUG_RT
	  printf("RT apply tex\n");
#endif
	  // Make a copy of the material properties and apply the texture
	  // functions attached to this primitive.
	  current_intersection->info.from.mat->copy(&fromMaterial);
	  current_intersection->info.from.texture->tex->apply(ray_from,
							      ray_dir,
							      ray_flags,
							      fromMaterial,
							      hit_point,
							      hit_normal, this,
		current_intersection->info.from.texture->tex_transform);
	  delete_frommat = TRUE;
	} else {
	  // We want to apply the unchanged material.
	  fromMaterial = current_intersection->info.from.mat;
	}
      }
      
    // Apply the to texture.
      if (current_intersection->info.to.tex_multi) {
#if DEBUG_RT
	printf("RT to-multitex\n");
#endif
	// There is more than one texture on this object, in
	// a collection.
	if (current_intersection->info.to.tex_col != NULL) {
#if DEBUG_RT
	  printf("RT apply tex\n");
#endif
	  // Make a copy of the material properties and apply the texture
	  // functions in the TextureCollection attached to this primitive.
	  current_intersection->info.to.mat->copy(&toMaterial);
	  current_intersection->info.to.tex_col->apply(ray_from, ray_dir,
						       ray_flags,
						       toMaterial, hit_point,
						       hit_normal, this);
	  delete_tomat = TRUE;
	} else {
	  // We want to apply the unchanged material.
	  toMaterial = current_intersection->info.to.mat;
	}
      } else {
#if DEBUG_RT
	printf("RT to-singletex\n");
#endif
	if (current_intersection->info.to.texture != NULL) {
#if DEBUG_RT
	  printf("RT apply tex\n");
#endif
	  // Make a copy of the material properties and apply the texture
	  // functions attached to this primitive.
	  current_intersection->info.to.mat->copy(&toMaterial);
	  current_intersection->info.to.texture->tex->apply(ray_from, ray_dir,
							    ray_flags,
							    toMaterial,
							    hit_point,
							    hit_normal, this,
			current_intersection->info.to.texture->tex_transform);
	  delete_tomat = TRUE;
	} else {
	  // We want to apply the unchanged material.
	  toMaterial = current_intersection->info.to.mat;
	}
      }
    } else {
      // We want to apply the unchanged materials.
      fromMaterial = current_intersection->info.from.mat;
      toMaterial   = current_intersection->info.to.mat;
    }

#if DEBUG_RT
    printf("RT done texture\n");
#endif
    
    // If this is a shadow ray, adjust the hit_color to take account of the
    // amount transmitted through the material. This is only done for the
    // 'to' material so it is applied once per object.
    if (ray_is_shadow) {
      toMaterial->apply_transmissive(hit_color);
    
      // If hit_color is zero, no light, don't keep checking.
      if (hit_color[R] == 0 && hit_color[G] == 0 && hit_color[B] == 0) {
	// We are a shadow ray, the only return value that we need to do
	// is hit_color, which has already been set.
	break;

      } else {
	// We've applied the transmitted light, so continue checking
	// for more intersections.
	//
	// Note that we ideally want to apply gaseous effects in order
	// from nearest light -> nearest object. However this requires a
	// reverse traverse of the intersection list and a lot of fancy
	// footwork. We compromise by applying in the opposite direction
	// instead. sfx_start_t is the start of the current SFX along
	// the shadow ray, so apply current_shadow_sfx from there to here.
	if (current_shadow_sfxtype == GASEOUS) {
	  GasSFX *current_gas = (GasSFX *) current_shadow_sfx;
	  
	  current_gas->gas->traverse_gas(current_gas->gas_transform,
					 ray_from, ray_dir, hit_color,
					 sfx_start_t, *hit_t);
	}
	
	// Store the new SFX and where we started it, and continue
	// checking intersections.
	current_shadow_sfxtype = current_intersection->info.to.sfxtype;
	current_shadow_sfx     = current_intersection->info.to.sfx;
	sfx_start_t = *hit_t;
	continue;
      }
    } else {
      // If not a shadow ray, then apply the material properties of the
      // primitive to get its color (takes account of lighting, reflected and
      // transmitted rays by casting secondary rays).

      // FIXME: how do we square this with LEAVING intersections?
      // Or should the determineColor routines do this? Yeah that's it,
      // put off the problem...

#if DEBUG_RT
      printf("RT material\n");
#endif
      
      toMaterial->determineColor(hit_color, hit_point,
				 current_intersection->hit_flags, hit_normal,
				 ray_dir, current_intersection->real_primnode,
				 this, world_object, ray_flags,
				 ray_contribution,
				 fromMaterial, toMaterial,
				 &(current_intersection->info), max_ray_depth);
      
#if DEBUG_RT
      printf("RT delete\n");
#endif

      // If we've copied the material properties of this object temporarily
      // we've allocated memory in doing so. This isn't placed in the
      // deletionlist for deletion at the end of the frame because then
      // we would build up hundreds of megabytes of properties information.
      // Deallocate the material(s) after they have been used.
      if (delete_frommat)
	delete fromMaterial;
      if (delete_tomat)
	delete toMaterial;

      // Apply any gaseous effects along the ray's length from the
      // ray's start to where it hits this next intersection..
      if (start_sfxtype == GASEOUS) {
	GasSFX *current_gas = (GasSFX *) start_sfx;
	  
	// Apply gaseous effect from the ray start point to the point we hit.
	current_gas->gas->traverse_gas(current_gas->gas_transform,
				       ray_from, ray_dir, hit_color,
				       0, *hit_t);
      }

      // Found a solid intersection, stop raytracing.
      break;
    }
  }

#if DEBUG_RT
  printf("RT out\n");
#endif

  // Deallocate the intersection list.
  free_intersection_list(intersect_list);

  // We've found the colour of the ray. Now, if the ray was a primary ray
  // (i.e. came from the camera directly), apply the foreground texture
  // to it.
  if (IS_PRIMARY(ray_flags) && (foreground_texture != NULL))
	foreground_texture->apply(ray_from, ray_dir, ray_flags, hit_color,
				  this);
}

void insert_entering(IntersectionList *the_list, primnode *node,
		     primnode *validate_node, vector normal, double t_val,
		     vector ray_from, vector ray_dir, uint ray_flags)
{
#if 0
  DSTmaterial *tmp_mat = (DSTmaterial *) node->material;
  printf("Insert entering, mat %f %f %f\n",
	 tmp_mat->diffuse[R], tmp_mat->diffuse[G], tmp_mat->diffuse[B]);
#endif

  if (node->flags & FLAGS_MULTITEXTURE)
    the_list->insert(node->flags, node, normal, t_val,
		     TRUE, FALSE,
		     NULL, node->material,
		     NOSFX, node->sfxtype,
		     NULL, node->sfx,
		     FALSE, TRUE,
		     NULL, node->texcol,
		     ray_from, ray_dir,
		     ray_flags, node, validate_node);
  else
    the_list->insert(node->flags, node, normal, t_val,
		     TRUE, FALSE,
		     NULL, node->material,
		     NOSFX, node->sfxtype,
		     NULL, node->sfx,
		     FALSE, FALSE,
		     NULL, node->texture,
		     ray_from, ray_dir,
		     ray_flags, node, validate_node);
}

void insert_leaving(IntersectionList *the_list, primnode *node,
		     primnode *validate_node, vector normal, double t_val,
		     vector ray_from, vector ray_dir, uint ray_flags)
{
#if 0
  DSTmaterial *tmp_mat = (DSTmaterial *) node->material;
  printf("Insert leaving, mat %f %f %f\n",
	 tmp_mat->diffuse[R], tmp_mat->diffuse[G], tmp_mat->diffuse[B]);
#endif

  if (node->flags & FLAGS_MULTITEXTURE)
    the_list->insert(node->flags, node, normal, t_val,
		     FALSE, TRUE,
		     node->material, NULL,
		     node->sfxtype, NOSFX, 
		     node->sfx, NULL,
		     TRUE, FALSE,
		     node->texcol, NULL,
		     ray_from, ray_dir,
		     ray_flags, node, validate_node);
  else
    the_list->insert(node->flags, node, normal, t_val,
		     FALSE, TRUE,
		     node->material, NULL,
		     node->sfxtype, NOSFX, 
		     node->sfx, NULL,
		     FALSE, FALSE,
		     node->texture, NULL,
		     ray_from, ray_dir,
		     ray_flags, node, validate_node);
}

#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif
