/* subdivide.c : Subdivision routines
** 
** This file contains functions used by all of the different cluster classes
** to perform subdivision.
**
** Chris Butcher, 9 January 1997
*/

#include "mirage.h"

#ifdef USE_NAMESPACES
namespace Mirage {
#endif

thing octree_subdivide(thing to_subdivide, int already_subdivided,
		       int subdivision_limit, int max_prims_in_voxel,
		       double x_min, double y_min, double z_min,
		       double x_max, double y_max, double z_max)
{
  rcsgnode *rnode;
  double x_mid, y_mid, z_mid;
  thing *children, divided_thing;
  int index;

  // Empty things (this includes NULL pointers which are zero and so are
  // counted as EMPTY) return an empty octree.
  if (IS_EMPTY(to_subdivide)) {
    return THING_EMPTY;
  }

  // Primnodes have no reason to be subdivided further, it won't increase
  // the efficiency. Stop and return them.
  if (IS_PRIM(to_subdivide)) {
    return to_subdivide;
  }

  // Don't subdivide beyond a certain level.
  if (already_subdivided >= subdivision_limit) {
    return to_subdivide;
  }

  // In order to get this far we must be subdividing an rcsgnode. If there
  // aren't enough primitives to bother, then don't subdivide it.
  if (IS_RCSG(to_subdivide)) {
    // We subdivide the RCSG normally.
  } else if (IS_OCTREE(to_subdivide)) {
    printf("SUBDIVIDE: OCTREE!\n");
    printf("Octree number %d\n", (to_subdivide & THING_INDEX));
    fatal_error("octree_subdivide", "trying to subdivide an octree!",
		"This error *really* should not happen.");
  } else {
    printf("SUBDIVIDE: UNKNOWN TYPE!\n");
    printf("Thing type %d, number %d\n", (to_subdivide & (~THING_INDEX)) >> 24,
  	   to_subdivide & THING_INDEX);
    fatal_error("octree_subdivide", "corrupt thing encoding",
		"This error should not happen.");
  }
  rnode = GET_RCSGNODE(to_subdivide);

  if (rnode->prim_count <= max_prims_in_voxel) {
    // Don't subdivide any further; this is a sufficiently low number
    // of primitives.
    return to_subdivide;
  }

  // Call create_octree_from_rcsg to cull unnecessary primitives out of
  // this RCSG tree and in the process subdivide into eight voxels.
  divided_thing = create_octree_from_rcsg(to_subdivide,
					  x_min, y_min, z_min,
					  x_max, y_max, z_max);

  // Free the previous approximation to the subdivision (unless it was
  // the thing passed to us, which would be the toplevel RCSG tree,
  // in which case we want to keep it in the cluster).
  if (already_subdivided != 0)
    free_tree(to_subdivide);

  // Calculate the midpoints of the voxel's sides.
  x_mid = (x_min + x_max) * 0.5;
  y_mid = (y_min + y_max) * 0.5;
  z_mid = (z_min + z_max) * 0.5;

  children = GET_OCTREE(divided_thing)->child;

  for (index = 0; index < 8; index++) {
    // Subdivide each leaf of the octree further if it's necessary.
    children[index] = octree_subdivide(children[index], already_subdivided + 1,
				       subdivision_limit, max_prims_in_voxel,
				       (index & 1) ? x_mid : x_min,
				       (index & 2) ? y_mid : y_min,
				       (index & 4) ? z_mid : z_min,
				       (index & 1) ? x_max : x_mid,
				       (index & 2) ? y_max : y_mid,
				       (index & 4) ? z_max : z_mid);
  }
   
  return divided_thing;
}

thing create_octree_from_rcsg(thing the_thing,
			      double x_min, double y_min, double z_min,
			      double x_max, double y_max, double z_max)
{
  thing combined_thing;

  if (IS_EMPTY(the_thing)) {
    // Empty RCSGnodes create empty octrees.
    return THING_EMPTY;

  } else if (IS_PRIM(the_thing)) {
    // Primnodes are not subdivided at all.
    return the_thing;

  } else if (IS_RCSG(the_thing)) {
    // This is an RCSG node.
    rcsgnode *the_rcsg = GET_RCSGNODE(the_thing);
    thing left_thing, right_thing;

    // Create an octree for the left rcsg branch using create_octree_from_rcsg
    left_thing = create_octree_from_rcsg(the_rcsg->left_thing,
					 x_min, y_min, z_min,
					 x_max, y_max, z_max);

    // Create an octree for the right rcsg branch using create_octree_from_rcsg
    right_thing = create_octree_from_rcsg(the_rcsg->right_thing,
					  x_min, y_min, z_min,
					  x_max, y_max, z_max);

    // Combine the two octrees according to the RCSGnode's linkmode.
    combined_thing =  combine_things(the_rcsg->linkmode,
			  left_thing, right_thing,
			  x_min, y_min, z_min,
			  x_max, y_max, z_max);

    return combined_thing;
  } else {
    printf("** Unknown thing type in create_octree_from_rcsg\n");
    printf("passed type %d, number %d\n", (the_thing & (~THING_INDEX)) >> 24,
    	   the_thing & THING_INDEX);
    fatal_error("create_octree_from_rcsg", "corrupt thing encoding",
		"This error should not happen.");
    exit(-1);
    return 0;
  }
}

thing create_octree_from_prim(thing the_thing,
			      double x_min, double y_min, double z_min,
			      double x_max, double y_max, double z_max)
{
  box world_voxel, top_voxel;
  box sub_voxels[8];
  primnode *the_primnode;
  octree *created_octree;
  thing *child, octree_thing;
  int index;
  intersect_result in_out;

  if (IS_PRIM(the_thing)) {
    the_primnode = GET_PRIMNODE(the_thing);

  } else {
    printf("CREATE_OCTREE_FROM_PRIM, passed type %d number %d, not a prim\n",
    	   (the_thing & (~THING_INDEX)) >> 24, the_thing & THING_INDEX);
    fatal_error("create_octree_from_prim", "passed non-primnode thing",
		"This error should not happen.");
    exit(-1);
  }

  // We have a primnode that we want to subdivide ... find the coordinates
  // of the voxel in primitivespace, then construct the eight
  // sub-voxels and use the primitive's voxel intersection test to check them.
  for (index = 0; index < 8; index++) {
    world_voxel[index][X] = (index & 1) ? x_max : x_min;
    world_voxel[index][Y] = (index & 2) ? y_max : y_min;
    world_voxel[index][Z] = (index & 4) ? z_max : z_min;
    Matrix::apply(the_primnode->position[AFT],
		  world_voxel[index], top_voxel[index]);
  }

  make_subvoxels(sub_voxels, top_voxel);

  // Allocate new octree space.
  created_octree = new_octree(&octree_thing);
  child = created_octree->child;

  for (index = 0; index < 8; index++) {
    in_out = the_primnode->prim->voxel_intersect(sub_voxels[index]);

    switch(in_out) {
    case OUT: {
      // This leaf of the octree is empty.
      child[index] = THING_EMPTY;
      break;
    }
    case IN: {
      // This leaf of the octree is full of this primnode.
      child[index] = the_thing | THING_FULL;
      break;
    }
    case SURFACE: {
      // This leaf of the octree contains this primnode's surface.
      child[index] = the_thing;
      break;
    }
    default: {
      fatal_error("create_octree_from_prim", "prim returned unknown voxel_in",
		  "This error should not happen.");
    }
    }
  }

  return octree_thing;
}

thing combine_things(linktype the_link, thing left_thing, thing right_thing,
		     double x_min, double y_min, double z_min,
		     double x_max, double y_max, double z_max)
{
  thing left_octree, right_octree;
  thing *left_child, *right_child;
  thing combined, left, right;
  int index, num_prims;

  switch(the_link) {
  case PLUS: {
    if (IS_OCTREE(left_thing)) {
      left_octree = left_thing;
    } else {
      // This is a primnode; we need an octree. Subdivide the primnode
      // to see which cells it is present in.
      left_octree = create_octree_from_prim(left_thing,
					    x_min, y_min, z_min,
					    x_max, y_max, z_max);
    }
      
    if (IS_OCTREE(right_thing)) {
      right_octree = right_thing;
    } else {
      // This is a primnode; we need an octree. Subdivide the primnode
      // to see which cells it is present in.
      right_octree = create_octree_from_prim(right_thing,
					     x_min, y_min, z_min,
					     x_max, y_max, z_max);
    }

    left_child = GET_OCTREE(left_octree)->child;
    right_child = GET_OCTREE(right_octree)->child;

    for (index = 0; index < 8; index++) {
      // If left or right are empty then take the other one.
      // If right is full then take that one.
      // Else combine the two things into an RCSG node.
      if (IS_EMPTY(left_child[index]))
	combined = right_child[index];
      else if (IS_EMPTY(right_child[index]))
	combined = left_child[index];
      else if (IS_FULL(right_child[index]))
	combined = right_child[index];
      else {
	left = left_child[index];
	right = right_child[index];
	num_prims  = (IS_RCSG(left))  ? GET_RCSGNODE(left )->prim_count : 1;
	num_prims += (IS_RCSG(right)) ? GET_RCSGNODE(right)->prim_count : 1;
	combined = new_rcsg(left, right, PLUS, num_prims);
      }
      
      left_child[index] = combined;
    }

    // We place the combined children in the left octree node and
    // discard the right octree node.
    free_octreenode(right_octree);

    return left_octree;
  }
  case PAINTMINUS: // Identical to the MINUS operator for subdivision purposes
  case MINUS: {
    if (IS_OCTREE(left_thing)) {
      left_octree = left_thing;
    } else {
      // This is a primnode; we need an octree. Subdivide the primnode
      // to see which cells it is present in.
      left_octree = create_octree_from_prim(left_thing,
					    x_min, y_min, z_min,
					    x_max, y_max, z_max);
    }
      
    if (IS_OCTREE(right_thing)) {
      right_octree = right_thing;
    } else {
      // This is a primnode; we need an octree. Subdivide the primnode
      // to see which cells it is present in.
      right_octree = create_octree_from_prim(right_thing,
					     x_min, y_min, z_min,
					     x_max, y_max, z_max);
    }

    left_child = GET_OCTREE(left_octree)->child;
    right_child = GET_OCTREE(right_octree)->child;

    for (index = 0; index < 8; index++) {
      // If left is empty return EMPTY.
      // If right is empty return left one.
      // If right is full return EMPTY.
      // Else combine the two things into an RCSG node.
      if (IS_EMPTY(left_child[index])) {
	combined = THING_EMPTY;
      } else if (IS_EMPTY(right_child[index])) {
	combined = left_child[index];
      } else if (IS_FULL(right_child[index])) {
	combined = THING_EMPTY;
      } else {
	left = left_child[index];
	right = right_child[index];
	num_prims  = (IS_RCSG(left))  ? GET_RCSGNODE(left )->prim_count : 1;
	num_prims += (IS_RCSG(right)) ? GET_RCSGNODE(right)->prim_count : 1;
	combined = new_rcsg(left_child[index], right_child[index],
			    the_link, num_prims);
      }
      
      left_child[index] = combined;
    }

    // We place the combined children in the left octree node and
    // discard the right octree node.
    free_octreenode(right_octree);

    return left_octree;
  }
  case SURF_AND: // Identical to the AND operator for subdivision purposes
  case AND: {
    if (IS_OCTREE(left_thing)) {
      left_octree = left_thing;
    } else {
      // This is a primnode; we need an octree. Subdivide the primnode
      // to see which cells it is present in.
      left_octree = create_octree_from_prim(left_thing,
					    x_min, y_min, z_min,
					    x_max, y_max, z_max);
    }
      
    if (IS_OCTREE(right_thing)) {
      right_octree = right_thing;
    } else {
      // This is a primnode; we need an octree. Subdivide the primnode
      // to see which cells it is present in.
      right_octree = create_octree_from_prim(right_thing,
					     x_min, y_min, z_min,
					     x_max, y_max, z_max);
    }

    left_child = GET_OCTREE(left_octree)->child;
    right_child = GET_OCTREE(right_octree)->child;

    for (index = 0; index < 8; index++) {
      // If left is empty return EMPTY.
      // If right is empty return EMPTY.
      // If left is full return right.
      // Else combine the two things into an RCSG node.
      if (IS_EMPTY(left_child[index]))
	combined = THING_EMPTY;
      else if (IS_EMPTY(right_child[index]))
	combined = THING_EMPTY;
      else if (IS_FULL(left_child[index]))
	combined = right_child[index];
      else {
	left = left_child[index];
	right = right_child[index];
	num_prims  = (IS_RCSG(left))  ? GET_RCSGNODE(left )->prim_count : 1;
	num_prims += (IS_RCSG(right)) ? GET_RCSGNODE(right)->prim_count : 1;
	combined = new_rcsg(left_child[index], right_child[index],
			    the_link, num_prims);
      }
     
      left_child[index] = combined;
    }

    // We place the combined children in the left octree node and
    // discard the right octree node.
    free_octreenode(right_octree);

    return left_octree;
  }
  default: {
    fatal_error("combine_things", "unknown linktype!",
		"This error should not happen.");
    return THING_EMPTY;
  }
  }
}
  
// This calculates the midpoint of voxel_corners[a - c] and places it
// in voxel_corners[b].
#define MID_POINT(a, b, c)                                                   \
    voxel_corners[b][X] = (voxel_corners[a][X] + voxel_corners[c][X]) * 0.5; \
    voxel_corners[b][Y] = (voxel_corners[a][Y] + voxel_corners[c][Y]) * 0.5; \
    voxel_corners[b][Z] = (voxel_corners[a][Z] + voxel_corners[c][Z]) * 0.5;

// Sets the voxel v to have points p0, p1, ... p7.
#define WRITE_VOX(v, p0, p1, p2, p3, p4, p5, p6, p7)                         \
  Vector::set(v[0], voxel_corners[p0]);                                      \
  Vector::set(v[1], voxel_corners[p1]);                                      \
  Vector::set(v[2], voxel_corners[p2]);                                      \
  Vector::set(v[3], voxel_corners[p3]);                                      \
  Vector::set(v[4], voxel_corners[p4]);                                      \
  Vector::set(v[5], voxel_corners[p5]);                                      \
  Vector::set(v[6], voxel_corners[p6]);                                      \
  Vector::set(v[7], voxel_corners[p7]);

void make_subvoxels(box *sub_voxels, box top_voxel)
{
  // voxel_corners holds all 27 corners of the subdivided voxel.
  // They are numbered as xval + 3*yval + 9*zval where xval goes from
  // 0 at x_min to 1 at the midpoint to 2 at x_max and similarly for
  // yval and zval.
  // Draw a diagram, it'll help.
  vector voxel_corners[27];

  // Set the eight points at the corners of the main voxel.
  Vector::set(voxel_corners[0],  top_voxel[0]);
  Vector::set(voxel_corners[2],  top_voxel[1]);
  Vector::set(voxel_corners[6],  top_voxel[2]);
  Vector::set(voxel_corners[8],  top_voxel[3]);
  Vector::set(voxel_corners[18], top_voxel[4]);
  Vector::set(voxel_corners[20], top_voxel[5]);
  Vector::set(voxel_corners[24], top_voxel[6]);
  Vector::set(voxel_corners[26], top_voxel[7]);

  // Find the midpoints along the four X edges of the main voxel.
  MID_POINT(0,  1,  2);
  MID_POINT(6,  7,  8);
  MID_POINT(18, 19, 20);
  MID_POINT(24, 25, 26);
  
  // Find the midpoints along the four Y edges of the main voxel.
  MID_POINT(0,  3,  6);
  MID_POINT(2,  5,  8);
  MID_POINT(18, 21, 24);
  MID_POINT(20, 23, 26);
  
  // Find the midpoints along the four Z edges of the main voxel.
  MID_POINT(0,  9,  18);
  MID_POINT(2,  11, 20);
  MID_POINT(6,  15, 24);
  MID_POINT(8,  17, 26);

  // Find the centers of the two X faces of the main voxel.
  MID_POINT(9, 12, 15);
  MID_POINT(11, 14, 17);
  
  // Find the centers of the two Y faces of the main voxel.
  MID_POINT(9,  10, 11);
  MID_POINT(15, 16, 17);
  
  // Find the centers of the two Z faces of the main voxel.
  MID_POINT(3,  4,  5);
  MID_POINT(21, 22, 23);
  
  // Find the center of the main voxel.
  MID_POINT(12, 13, 14);

  // Write these points into the individual voxels.
  // If n is the low-XYZ corner, then the indices of the eight
  // corners are n, n+1, n+3, n+1+3, n+9, n+1+9, n+3+9, n+1+3+9.
  WRITE_VOX(sub_voxels[0],  0,  1,  3,  4,  9, 10, 12, 13);
  WRITE_VOX(sub_voxels[1],  1,  2,  4,  5, 10, 11, 13, 14);
  WRITE_VOX(sub_voxels[2],  3,  4,  6,  7, 12, 13, 15, 16);
  WRITE_VOX(sub_voxels[3],  4,  5,  7,  8, 13, 14, 16, 17);
  WRITE_VOX(sub_voxels[4],  9, 10, 12, 13, 18, 19, 21, 22);
  WRITE_VOX(sub_voxels[5], 10, 11, 13, 14, 19, 20, 22, 23);
  WRITE_VOX(sub_voxels[6], 12, 13, 15, 16, 21, 22, 24, 25);
  WRITE_VOX(sub_voxels[7], 13, 14, 16, 17, 22, 23, 25, 26);
}

// Look for clusters that may be in this RCSG tree. If they are there,
// subdivide them.
void subdivide_clusters_in_RCSG(thing a_thing)
{
  primnode *a_primnode;
  rcsgnode *a_rcsgnode;

  if (IS_EMPTY(a_thing))
    return;
  else if (IS_PRIM(a_thing)) {
    a_primnode = GET_PRIMNODE(a_thing);
    if (a_primnode->prim->is_cluster)
      ((ClusterPrimitive *) (a_primnode->prim))->subdivide();
  } else if (IS_RCSG(a_thing)) {
    a_rcsgnode = GET_RCSGNODE(a_thing);
    subdivide_clusters_in_RCSG(a_rcsgnode->left_thing);
    subdivide_clusters_in_RCSG(a_rcsgnode->right_thing);
  } else {
    fatal_error("subdivide_clusters_in_RCSG", "unknown type of thing",
		"This error should not happen.");
  }
}

int compress_octree(thing *a_thing)
{
  // Traverse the octree, turning any nodes that are all
  // empty into a single empty value. This wil remove quite
  // a few octree nodes (sometimes).

  if (IS_EMPTY(*a_thing)) {
    // This is empty, return so.
    return 1;
  }

  if (IS_OCTREE(*a_thing)) {
    int node_index, empty_nodes = 0;
    thing *children = GET_OCTREE(*a_thing)->child;

    for (node_index = 0; node_index < 8; node_index++)
      empty_nodes += compress_octree(&children[node_index]);

    if (empty_nodes == 8) {
      // This octree has all eight children empty, compress it.
      free_octreenode(*a_thing);
      *a_thing = THING_EMPTY;
      return 1;
    }
  }

  // This octree wasn't compressed.
  return 0;
}





#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif
