/* subdivide.h : Definitions of subdivision data structures
** 
** This file contains definitions of the data structures used for
** subdivision, and prototypes for the subdivision routines.
**
** Chris Butcher, 9 January 1997
*/

// A 'thing' is a convenient way of representing a number of pointers to
// different classes of structures in one data type. This is because in
// a node of the octree, we may want to store pointers to other octree nodes,
// pointers to primitive nodes or to RCSG nodes.
// We do this by storing an index to the given node (relative to the working
// memory) and also a set of flags to say what kind of node it is.
// There are also several special cases; an octree may be EMPTY, it may be
// something else with the FULL flag or it may be the border of the cluster
// (CLUSTER_BORDER).
typedef unsigned int thing;

// An octree node stores eight 'things' ... the 'thing' is
// defined in subdivide.h and is an encoded pointer to
// either an octree, an rcsgnode, or a primnode.
typedef struct octreenode {
  // The eight children of the octree node.
  thing child[8];
} octree;

// An RCSG node is equivalent to a dagnode except that all the information
// about lights, textures, positions etc. is propagated down into the
// primnodes.
typedef struct RCSGnode {
  // The number of primitives in the RCSG tree.
  int prim_count;

  // The link mode of this node (PLUS, MINUS, PAINTMINUS, AND, SURF_AND)
  linktype linkmode;

  // The left_thing and right_thing values index the left and right branches
  // of the RCSG tree ... they may be either primnodes or rcsgnodes.
  // For more information on the 'thing' data type, see subdivide.h.
  //
  // The left branch is equivalent to the object that is linked in to
  // the dagnode, the right branch is equivalent to the object that it
  // was linked to. i.e. the command
  // LINK(object, PLUS, newbit)
  // makes a dagnode that will translate to an RCSG node with the
  // left branch pointing to newbit, the right branch pointing to
  // object. object is then set to point to this new node.
  thing left_thing, right_thing;
} rcsgnode;

// A primnode holds all of the information about a primitive in the
// rendering system - where it is, what its material properties are,
// what its textures are etc.
// It is found in RCSG trees and also in the intermediate stages of
// building an octree.
typedef struct PrimitiveNode {
  // flags holds information about the primitive's behaviour with respect
  // to different kinds of rays (reflection, transparency, etc.)
  uint flags;

  // position holds transforms to go from worldspace to primitive space and
  // back again.
  transform position;

  // The primitive that is represented by this primnode.
  class Primitive *prim;

  // material holds the properties of the object and a method for
  // applying these material properties by casting additional rays.
  class Material *material;

  // Primnodes may have one or many textures on them. This is determined by
  // the MULTITEXTURE flag.
  union {
    // The texture applied to this object, as a SingleTexture structure.
    struct _SingleTexture *texture;

    // texcol is a collection of textures, each with their own transform,
    // precedence and routines.
    class TextureCollection *texcol;
  };

  // lightlist holds a pointer to the list of Lights that are tested
  // on this primitive. At present this is the entire global lightList,
  // but in future it may be wise to add a voxel in/out test to lights
  // and use this to determine which lights should be applied to each
  // primnode.
  lightList *lightlist;
  
  // light holds a pointer to a Light which is applied in a different manner
  // to this primitive (as this primitive is the source of this light).
  Light *light;

  // The sfx parameter refers to special effects that may be applied
  // to the primitive - gaseous or surface-distortion presently.
  SfxType sfxtype;
  void *sfx;

  // The signature of the last ray to be intersected with this primnode.
  // This is stored so that we don't intersect the same ray more than once
  // with the same primnode.
  signature last_ray_signature;

  // Did the ray specified in last_ray_signature actually hit the primnode?
  // Each subsequent call to ray_intersect will just return this value,
  // as the intersection is already in the intersection list.
  int last_ray_hit;
} primnode;

// Bitmask to extract the index component of the thing.
#define THING_INDEX           0x00FFFFFF

// Bitmask to extract the type component of the thing, and the various types.
#define THING_TYPE            (3 << 24)
#define THING_OCTREE          (1 << 24)
#define THING_PRIMNODE        (2 << 24)
#define THING_RCSGNODE        (3 << 24)

// The special kinds of thing.
#define THING_EMPTY           (0 << 24)
#define THING_FULL            (4 << 24)
#define THING_CLUSTER_BORDER  (5 << 24)

// Macros for deciding the kind of thing, and retrieving the original
// pointer that is stored there.
#define IS_OCTREE(t)          ((t & THING_TYPE) == THING_OCTREE)
#define IS_PRIM(t)            ((t & THING_TYPE) == THING_PRIMNODE)
#define IS_RCSG(t)            ((t & THING_TYPE) == THING_RCSGNODE)
#define IS_EMPTY(t)           (t == THING_EMPTY)
#define IS_FULL(t)            (t & THING_FULL)
#define IS_CLUSTER_BORDER(t)  (t == THING_CLUSTER_BORDER)

#define GET_OCTREE(t)         (retrieve_octree(t & THING_INDEX))
#define GET_PRIMNODE(t)       (retrieve_primnode(t & THING_INDEX))
#define GET_RCSGNODE(t)       (retrieve_rcsgnode(t & THING_INDEX))

thing create_octree_from_prim(thing the_thing,
			      double x_min, double y_min, double z_min,
			      double x_max, double y_max, double z_max);
thing create_octree_from_rcsg(thing the_thing,
			      double x_min, double y_min, double z_min,
			      double x_max, double y_max, double z_max);
thing octree_subdivide(thing to_subdivide, int already_subdivided,
		       int subdivision_limit, int max_prims_in_voxel,
		       double x_min, double y_min, double z_min,
		       double x_max, double y_max, double z_max);
thing combine_things(linktype the_link, thing left_thing, thing right_thing,
		     double x_min, double y_min, double z_min,
		     double x_max, double y_max, double z_max);
void make_subvoxels(box *sub_voxels, box top_voxel);
void subdivide_clusters_in_RCSG(thing a_thing);
int compress_octree(thing *a_thing);
