/* texture.c : Contains code for the TextureCollection class.
** 
** This has the TextureCollection class and code to add and apply textures
** from a texture collection.
**
** Chris Butcher, 3 December 1996
*/

#include "mirage.h"

#ifdef USE_NAMESPACES
namespace Mirage {
#endif

// Create a new texture collection consisting of one texture.
TextureCollection *new_textureCollection(Texture *newTex, transform newTrans)
{
  TextureCollection *new_tc;

  new_tc = alloc_textureCollection();

  // Initialise the TextureCollection.
  new_tc->num_textures = 0;
  new_tc->add(newTex, newTrans);

  return new_tc;
}

// Create a new texture collection from the old texture plus a new texture.
TextureCollection *new_textureCollection(SingleTexture *oldTex,
					 Texture *newTex, transform newTrans)
{
  TextureCollection *new_tc;

  // Initialise the TextureCollection.
  new_tc = alloc_textureCollection();
  new_tc->num_textures = 0;

  if (oldTex != NULL)
    new_tc->add(oldTex->tex, oldTex->tex_transform);

  new_tc->add(newTex, newTrans);

  return new_tc;
}

// Create a new texture collection from the old texture collection,
// with a new texture added.
TextureCollection *new_textureCollection(TextureCollection *oldTex,
					 Texture *newTex, transform newTrans)
{
  TextureCollection *new_tc;
  int i;

  new_tc = alloc_textureCollection();

  // Initialise the TextureCollection.
  if (oldTex == NULL) {
    // No old textures to add!
    new_tc->num_textures = 0;
  } else {
    // Copy the old texture collection before adding the new texture.
    new_tc->num_textures = oldTex->num_textures;
    for (i = 0; i < new_tc->num_textures; i++) {
      new_tc->texStorage[i] = oldTex->texStorage[i];
      Transform::set(new_tc->transStorage[i], oldTex->transStorage[i]);
    }
  }

  new_tc->add(newTex, newTrans);

  return new_tc;
}

// Add a texture to the current collection.
void TextureCollection::add(Texture *newTex, transform newTrans)
{
  int i;

  if (num_textures >= MAX_OVERLAID_TEXTURES) {
    char errbuf[MIRAGE_STRING_LENGTH];
    
    sprintf(errbuf, "exceeded maximum of %d textures", MAX_OVERLAID_TEXTURES);
    fatal_error("TextureCollection::add", errbuf,
		"There is a limit to the number of textures that you\n"
		"can overlay on an object; reduce the number of textures\n"
		"that you want to overlay or change the limit\n"
		"MAX_OVERLAID_TEXTURES and recompile the rendering system.");
  }

  // If the list is empty, add it in position 1.
  if (num_textures == 0) {
    num_textures = 1;
    texStorage[0] = newTex;
    Transform::set(transStorage[0], newTrans);
    return;
  }

  // If the current texture has a negative precedence,
  if (texStorage[0]->precedence < 0) {
    // any new textures with lower precedence are ignored (the old one
    // takes priority).
    if (newTex->precedence > texStorage[0]->precedence)
      return;
    
    // Otherwise the new texture has a stronger negative precedence, so
    // replace the old texture.
    texStorage[0] = newTex;
    Transform::set(transStorage[0], newTrans);
    return;
  }

  // Otherwise the list has some number of textures which have positive
  // precedence (i.e. they are overlaying each other). A negative precedence
  // on the added texture will remove them all.
  if (newTex->precedence < 0) {
    num_textures = 1;
    texStorage[0] = newTex;
    Transform::set(transStorage[0], newTrans);
  }

  // So there are positive-precedence overlaid textures in the list
  // and we're adding a new positive-precedence texture. Go through the
  // list sorting the textures correctly.
  i = num_textures;
  // while the newtex still has power to displace the old textures,
  while (i != 0 && newTex->precedence < texStorage[i - 1]->precedence) {
    // move the old texture up one in the list.
    texStorage[i] = texStorage[i - 1];
    Transform::set(transStorage[i], transStorage[i - 1]);
    i--;
  }
  // We've either reached the front of the list or we've reached
  // an old texture that is stronger than the new texture,
  // so put the new one in just after it in the list.
  texStorage[i] = newTex;
  Transform::set(transStorage[i], newTrans);
  num_textures++;
}

// Apply the textures in order of precedence.
void TextureCollection::apply(vector ray_from, vector ray_dir,
			      uint ray_flags, Material *properties,
			      vector hit_point, vector hit_normal,
			      RayTracer *engine)
{
  int i;

  // If only one texture, apply it.
  if (num_textures == 1) {
    texStorage[0]->apply(ray_from, ray_dir, ray_flags, properties, hit_point,
			 hit_normal, engine, transStorage[0]);
    return;
  }
  
  if (num_textures == 0)
    fatal_error("TextureCollection::apply", "null collection",
		"This error shouldn't happen.");

  // Apply the textures from lowest (most positive) to highest (least
  // positive) priority.
  for (i = num_textures - 1; i >= 0; i--) {
    texStorage[i]->apply(ray_from, ray_dir, ray_flags, properties,
			 hit_point, hit_normal, engine, transStorage[i]);
  }
}
    
    

#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif
