/* texture.c : Contains definitions for the texturing system.
** 
** This file defines the Texture, BackgroundTexture and TextureCollection
** classes.
**
** Chris Butcher, 26 January 1998
*/

// The Texture class stores a function that modifies the material
// properties of the object based on certain properties of the ray
// that hit it (hit point, ray start point, ray direction, surface normal,
// etc.)
// Textures also have a precedence associated with them. Higher-precedence
// textures are called first (appear behind later textures). To make a texture
// override those before it, give it a negative precedence. To create a
// texture, subclass Texture to, say, WoodTexture and add some parameters
// to the class to control the appearance.
class Texture : public DeletableObject {
public:
  int precedence;

  // Alter the material properties of the object according to some
  // function.
  virtual void apply(vector ray_from, vector ray_dir, uint ray_flags,
		     Material *properties, vector hit_point, vector hit_normal,
		     RayTracer *engine, transform tex_transform) = 0;

  // Print out the texture (for debugging).
  virtual void print_tex(void) = 0;
};

// The BackgroundTexture class stores a function that provides a background
// colour depending on where the ray hit the background. An object of this
// type can also be used as a foreground texture as well, applied to the
// return value of every ray and used to simulate things like lensflares.
class BackgroundTexture : public DeletableObject {
public:

  // Return a colour depending on where the ray hit.
  virtual void apply(vector ray_from, vector ray_dir, uint ray_flags,
		     vector hit_color, RayTracer *engine) = 0;

  // Print out the texture (for debugging).
  virtual void print_tex(void) = 0;
};

// A SingleTexture structure holds a textures and its transform.
typedef struct _SingleTexture {
  Texture *tex;

  transform tex_transform;
} SingleTexture;

// A TextureCollection holds a collection of textures and their respective
// transforms + precedences.
class TextureCollection {
public:
  Texture *texStorage[MAX_OVERLAID_TEXTURES];
  transform transStorage[MAX_OVERLAID_TEXTURES];
  int num_textures;

  // Add another texture to the collection.
  void add(Texture *newTex, transform newTrans);

  // Apply the textures in order of precedence.
  void apply(vector ray_from, vector ray_dir, uint ray_flags,
	     Material *properties, vector hit_point, vector hit_normal,
	     RayTracer *engine);
};

// Routines for initialising newly created TextureCollections.
TextureCollection *new_textureCollection(Texture *newTex, transform newTrans);
TextureCollection *new_textureCollection(TextureCollection *oldTex,
					 Texture *newTex, transform newTrans);
TextureCollection *new_textureCollection(SingleTexture *oldTex,
					 Texture *newTex, transform newTrans);
