/* useful.c : Useful functions for scenes.
** 
** This file holds a collection of functions that may be useful when
** creating scenes. Please add to it!
**
** Chris Butcher, 15 January 1997
*/

#include "mirage.h"

#ifdef USE_NAMESPACES
namespace Mirage {
#endif

// The default up-vector for cameras and other objects.
vector plus_Z = {0.0, 0.0, 1.0};

// A zero-vector, handy for many things.
vector zero_vector = {0.0, 0.0, 0.0};

// The multiplicative identity quaternion (zero rotation).
quaternion identity_quaternion = {0.0, 0.0, 0.0, 1.0};

// Interpolates a value from lowval to highval by fraction (from 0 - 1)
// using the Geoff function (a cubic filter) 3t^2 - 2t^3.
//
// Values outside the range are clamped.
double cubicInterpolate(double lowval, double highval, double fraction)
{
  CLAMP(0, fraction, 1);
  
  return lowval + (highval - lowval) * (3*SQR(fraction) - 2*CUBE(fraction));
}

// Returns the position of the highest bit set in an unsigned long as
// an integer (0 - 31), or -1 if it is zero.
int highest_bit(unsigned long theLong)
{
  int i;
  unsigned long bitmask;

  bitmask = 0x80000000UL;
  for (i = 31; ((theLong & bitmask) == 0) && i >= 0;  i--, theLong <<= 1)
    ;

  return i;
}

// Convert a vector into spherical coordinates rho, theta and phi. Theta and
// phi are computed in radians.
void spherical_coordinates(vector v, double *rho, double *theta, double *phi)
{
  vector xz_vector;

  if ((v[X] == 0) && (v[Y] == 0)) {
    // The vector points directly along Z.
    if (v[Z] > 0) {
      // Vector points along +Z.
      *theta = 0;
      *phi = 0;
      *rho = v[Z];
      return;
    } else {
      // Vector points along -Z.
      *theta = 0;
      *phi = PI;
      *rho = -v[Z];
      return;
    }
  } else {
    // First we determine which direction the XY components of the vector
    // are pointing. Theta measures this, from 0 to 2pi.
    *theta = atan2(v[Y], v[X]);
    if (*theta < 0) {
      *theta += 2 * PI;
    }

    // Rotate the vector so that it lies in the XZ plane.
    Vector::set(xz_vector, v);
    Vector::rotate(X, Y, -RAD2DEG(*theta), xz_vector);

    // Phi is the angle between the vector and the +Z axis.
    *phi = fabs(atan2(-xz_vector[X], xz_vector[Z]));

    // Rho is the length of the vector.
    *rho = Vector::magnitude(v);
  }
}   

// Apply the Noisetexture to an object's diffuse colour.
void Noisetexture::apply(vector ray_from, vector ray_dir, uint ray_flags,
			 Material *properties, vector hit_point,
			 vector hit_normal, RayTracer *engine,
			 transform tex_transform)
{
  double noiseparam;
  vector primitive_hitpoint;
  DiffuseComponent *diffuse;

  // Find the hitpoint in primitive space.
  Matrix::apply(tex_transform[AFT], hit_point, primitive_hitpoint);
  Vector::mult(scalefactor, primitive_hitpoint, primitive_hitpoint);

  // Sample at the hitpoint in primitive space.
  noiseparam = noise_3(primitive_hitpoint);

  // Vary the intensity from 1 - noise_amount to 1.
  noiseparam = noise_amount * noiseparam + (1 - noise_amount);

  // Apply this factor to the diffuse color.
  diffuse = get_diffuse_component(properties);
  Color::mult(noiseparam, diffuse->diffuse_color, diffuse->diffuse_color);
}

// Code for the PointAtAffect that points one object at another.
void PointAtAffect::affect_object(object the_obj)
{
  int find_result;
  object from_obj, to_obj;
  transform from_trans, to_trans;
  vector direction;

  // Find the world-transforms of the from-object and the to-object.
  find_result = world_label_list->find(fromname, &from_obj, from_trans);
  if (find_result == FALSE) {
    char tempbuf[MIRAGE_STRING_LENGTH];

    sprintf(tempbuf, "can't find the from-label %s", fromname);
    fatal_error("PointAtAffect::affect", tempbuf,
		"You have to give an object a label if you want a\n"
		"PointAtAffect function to point from it to something\n"
		"else.");
  }

  find_result = world_label_list->find(toname, &to_obj, to_trans);
  if (find_result == FALSE) {
    char tempbuf[MIRAGE_STRING_LENGTH];

    sprintf(tempbuf, "can't find the to-label %s", toname);
    fatal_error("PointAtAffect::affect", tempbuf,
		"You have to give an object a label if you want a\n"
		"PointAtAffect function to point something at it.");
  }

  // Given the from-object and to-object's worldtransforms, the origins
  // of the two objects are from_trans[i][3] and to_trans[i][3]. We
  // then find the direction from one to the other.
  direction[X] = to_trans[FORE][X][3] - from_trans[FORE][X][3];
  direction[Y] = to_trans[FORE][Y][3] - from_trans[FORE][Y][3];
  direction[Z] = to_trans[FORE][Z][3] - from_trans[FORE][Z][3];
  if ((fabs(direction[X]) < TINY) &&
      (fabs(direction[Y]) < TINY)) {
    if (fabs(direction[Z]) < TINY) {
      output_warning("PointAtAffect::affect_object",
		     "From and to objects are in the same place",
		     "The from object must be able to point at the to\n"
		     "object so they cannot be coincident.");
    } else {
      // We are looking virtually straight down or straight up.
      the_obj->rotate(Z, Y, 90);
      if (direction[Z] > 0)
	the_obj->rotate(X, Z, 90);
      else
	the_obj->rotate(Z, X, 90);
    }
  } else {
    Vector::normalise(direction, "PointAtAffect::affect_object ... error in"
		      "zero checking, notify administrator");

    // Rotate the object that we're affecting so that it points in this
    // direction, with the right up-vector. The object must be pointing
    // along +X with its up-vector along +Z.
    the_obj->rotate(direction, up_vector);
  }
}

// Code for the HugHeightfieldAffect that makes objects hug heightfields.
void HugHeightfieldAffect::affect_object(object the_obj)
{
  int find_result;
  object hf_obj, this_obj;
  transform hf_trans, hf_prim_trans, obj_trans, obj_prim_trans;
  vector obj_pos;
  double hf_height, shift_val;
  Heightfield *hf_primitive;

  // Find the world-transform of the heightfield.
  find_result = world_label_list->find(hfname, &hf_obj, hf_trans);
  if (find_result == FALSE) {
    char tempbuf[MIRAGE_STRING_LENGTH];

    sprintf(tempbuf, "can't find the heightfield \"%s\"", hfname);
    fatal_error("HugHeightfield::affect", tempbuf,
		"The heightfield for objects to hug has to be labeled.");
  }

  find_result = world_label_list->find(objname, &this_obj, obj_trans);
  if (find_result == FALSE) {
    char tempbuf[MIRAGE_STRING_LENGTH];

    sprintf(tempbuf, "can't find the label of this object \"%s\"", objname);
    fatal_error("HugHeightfieldAffect::affect", tempbuf,
		"You have to give an object a label if you want it\n"
		"to use a HugHeightfieldAffect.");
  }

  // Work out the position of this object's origin in worldspace.
  Matrix::apply(obj_trans[FORE], zero_vector, obj_pos);

  // Where is this relative to the heightfield?
  Matrix::apply(hf_trans[AFT], obj_pos, obj_pos);

  // Find the heightfield primitive.
  if ((hf_obj->flags & FLAGS_PRIM) == 0)
    fatal_error("HugHeightfield::affect",
		"'Heightfield' object is not a primitive object!",
		"The second object whose label is given in the constructor\n"
		"for HugHeightfieldAffect must directly refer to a\n"
		"Heightfield, i.e. it should be either returned from\n"
		"new_heightfield or created by Dagnode::newPrimitive()\n"
		"from a Heightfield object!");
  hf_primitive = (Heightfield *) hf_obj->details;

  // Determine the height of the heightfield at this position.
  hf_height = hf_primitive->get_height(obj_pos[X], obj_pos[Y]);

  // Determine how the object needs to be shifted in Z to lie on the
  // heightfield's surface.
  shift_val = hf_height - obj_pos[Z];

  // We need to find out what the shift in heightfield-space translates
  // to in worldspace. The steps to shift a world-point are:
  //
  //  1) Transform it into the heightfield's primitive space
  //  2) Shift it in Z
  //  3) Transform it back into worldspace.
  //
  // The steps to shifting the object are thus:
  //  1) Transform it into worldspace
  //  2) Shift it in worldspace (the 3 steps above)
  //  3) Transform it back into the object's primitive space.

  Transform::invert(hf_trans,  hf_prim_trans);
  Transform::invert(obj_trans, obj_prim_trans);

  Transform::combine(the_obj->position, obj_trans,      the_obj->position);
  Transform::combine(the_obj->position, hf_prim_trans,  the_obj->position);
  Transform::shift  (the_obj->position, 0, 0, shift_val);
  Transform::combine(the_obj->position, hf_trans,       the_obj->position);
  Transform::combine(the_obj->position, obj_prim_trans, the_obj->position);
}


#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif
