/* userinterface.c : Functions to provide a nice interface to Mirage
**                   for the user. This includes creating new objects easily
**                   and K******-like operations on objects.
** 
** Chris Butcher, 15 July 1997
*/

#include "mirage.h"

#ifdef USE_NAMESPACES
namespace Mirage {
#endif

// Some primitive objects that are used by the default primitive
// creation routines. Initialised by init_primitives().
object prim_full, prim_nothing, prim_sphere, prim_cyl, prim_plane;
object prim_cone, prim_cube, prim_unitcyl;

int primobjects_initialised = FALSE;
fullPrimitive *fullprim;
nothingPrimitive *nothingprim;
spherePrimitive *sphereprim;
cylinderPrimitive *cylprim;
planePrimitive *planeprim;
conePrimitive *coneprim;
cubePrimitive *cubeprim;
unitCylinderPrimitive *unitcylprim;

// Some alternative interfaces for the transformation functions.
// More like K******.
object bind(object x) {
  return x->bind();
}

object place(transform t, object x) {
  x->place(t);
  return x;
}

object rotate(axis around, double angle, object x) {
  x->rotate(around, angle);
  return x;
}

object rotate(axis from, axis to, double angle, object x) {
  x->rotate(from, to, angle);
  return x;
}

object rotate(vector fwd, vector up, object x) {
  x->rotate(fwd, up);
  return x;
}

object shift(double xshift, double yshift, double zshift, object x) {
  x->shift(xshift, yshift, zshift);
  return x;
}

object shift(vector shift_vec, object x) {
  x->shift(shift_vec);
  return x;
}

object reflect(axis the_axis, object x) {
  x->reflect(the_axis);
  return x;
}

object stretch(double xstretch, double ystretch, double zstretch, object x) {
  x->stretch(xstretch, ystretch, zstretch);
  return x;
}

object stretch(vector stretchvec, object x) {
  x->stretch(stretchvec);
  return x;
}

object scale(double scalefac, object x) {
  x->scale(scalefac);
  return x;
}

object shear(axis from, axis to, double amount, object x) {
  x->shear(from, to, amount);
  return x;
}

object set_label(char *labelname, object x) {
  x->label(labelname);
  return x;
}

object give_material(Material *m, object x) {
  x->material = m;
  return x;
}

// Set up the default primitive objects.
void init_primitives(void)
{
  prim_full     = Dagnode::newPrimitive(fullprim    = new fullPrimitive);
  prim_nothing  = Dagnode::newPrimitive(nothingprim = new nothingPrimitive);
  prim_sphere   = Dagnode::newPrimitive(sphereprim  = new spherePrimitive);
  prim_cyl      = Dagnode::newPrimitive(cylprim     = new cylinderPrimitive);
  prim_plane    = Dagnode::newPrimitive(planeprim   = new planePrimitive);
  prim_cone     = Dagnode::newPrimitive(coneprim    = new conePrimitive);
  prim_cube     = Dagnode::newPrimitive(cubeprim    = new cubePrimitive);
  prim_unitcyl  = Dagnode::newPrimitive(unitcylprim = new unitCylinderPrimitive);

  primobjects_initialised = TRUE;
}

// Delete the default primitive objects.
void deallocate_primitives(void)
{
  if (primobjects_initialised) {
    delete fullprim;
    delete nothingprim;
    delete sphereprim;
    delete cylprim;
    delete planeprim;
    delete coneprim;
    delete cubeprim;
    delete unitcylprim;
  }
}


// Routines for returning a new primitive object.
object new_full(void) {
  return prim_full->bind();
}

object new_nothing(void) {
  return prim_nothing->bind();
}

object new_sphere(void) {
  return prim_sphere->bind();
}

object new_cyl(void) {
  return prim_cyl->bind();
}

object new_plane(void) {
  return prim_plane->bind();
}

object new_cone(void) {
  return prim_cone->bind();
}

object new_cone(double radius_1, double radius_2, double height) {
  if (radius_1 > radius_2) {
    object the_cone;

    the_cone = stretch(1, 1, height / (radius_1 - radius_2),
		       shift(0, 0, radius_1, bind(prim_cone)));
    LINK(the_cone, MINUS, reflect(Z, new_plane()));
    LINK(the_cone, MINUS, shift(0, 0, height, new_plane()));
    
    return bind(the_cone);

  } else if (radius_1 < radius_2) {
    object the_cone;

    the_cone = stretch(1, 1, height / (radius_2 - radius_1),
		       shift(0, 0, -radius_1, bind(prim_cone)));
    LINK(the_cone, MINUS, reflect(Z, new_plane()));
    LINK(the_cone, MINUS, shift(0, 0, height, new_plane()));
    
    return bind(the_cone);

  } else {
    return stretch(radius_1, radius_1, height, bind(prim_unitcyl));
  }
}

object new_cube(void) {
  return prim_cube->bind();
}

object new_torus(double r) {
  return Dagnode::newPrimitive(new torusPrimitive(r));
}

object new_torus(double circle_r, double tube_r) {
  object t;

  if (circle_r == 0)
    return new_nothing();
  else {
    t = Dagnode::newPrimitive(new torusPrimitive(tube_r / circle_r));
    t->scale(circle_r);

    return t;
  }
}

object new_taper(double t, double r1, double r2) {
  return Dagnode::newPrimitive(new taperPrimitive(t, r1, r2,
						  10, MIN_TVAL_CUTOFF * 10));
}

object new_taper(double t, double r1, double r2, int steps, double limit) {
  return Dagnode::newPrimitive(new taperPrimitive(t, r1, r2, steps, limit));
}

object new_taper(double circle_r, double theta,
		 double bottom_r, double top_r) {
  object t;

  if (circle_r == 0)
    return new_nothing();
  else {
    t = new_taper(theta, bottom_r / circle_r, top_r / circle_r);
    t->scale(circle_r);

    return t;
  }
}

object new_unitcyl(void) {
  return prim_unitcyl->bind();
}

object new_stupid_cluster(object obj) {
  return Dagnode::newPrimitive(new StupidClusterPrimitive(obj));
}

object new_uniform_cluster(object obj, vector min, vector max,
			   int level, int stop_at_first) {
  return Dagnode::newPrimitive(new UniformClusterPrimitive(obj, min, max,
						 level, stop_at_first));
}

object new_uniform_cluster(object obj,
			   double x_min, double y_min, double z_min,
			   double x_max, double y_max, double z_max,
			   int level, int stop_at_first) {
  return Dagnode::newPrimitive(new UniformClusterPrimitive(obj,
                                                  x_min, y_min, z_min,
                                                  x_max, y_max, z_max,
						  level, stop_at_first));
}

object new_octree_cluster(object obj, vector min, vector max,
			  int limit, int max_prims, int stop_at_first) {
  return Dagnode::newPrimitive(new OctreeClusterPrimitive(obj, min, max,
					   limit, max_prims, stop_at_first));
}

object new_heightfield(int smooth_normals, int should_stop, int do_jitter,
		       char *filename) {
  return Dagnode::newPrimitive(new Heightfield(smooth_normals, should_stop,
					       do_jitter, filename));
}

object new_heightfield(int smooth_normals, int should_stop,
		       int x_size, int y_size, height_val *field_data) {
  return Dagnode::newPrimitive(new Heightfield(smooth_normals, should_stop,
					       x_size, y_size, field_data));
}

object new_volume(int smooth_normals, int stop_at_first,
		  int xres, int yres, int zres,
		  isosurface_val *field_data, isosurface_val threshold)
{
  return Dagnode::newPrimitive(new Volume(smooth_normals, stop_at_first,
					  xres, yres, zres,
					  field_data, threshold, FALSE));
}

object new_octree_cluster(object obj,
			  double x_min, double y_min, double z_min,
			  double x_max, double y_max, double z_max,
			  int limit, int max_prims, int stop_at_first) {
  return Dagnode::newPrimitive(new OctreeClusterPrimitive(obj,
                                           x_min, y_min, z_min,
                                           x_max, y_max, z_max,
					   limit, max_prims, stop_at_first));
}

// Routines to return new light objects.
object new_point_light(double r, double g, double b) {
  return Dagnode::newLight(new PointLight(r, g, b));
}

object new_directional_light(double r, double g, double b,
			     double x, double y, double z) {
  return Dagnode::newLight(new DirectionalLight(r, g, b, x, y, z));
}

object new_ambient_light(double r, double g, double b) {
  return Dagnode::newLight(new AmbientLight(r, g, b));
}

// Routine to return a new camera object.
object new_camera(OpticsModel *anOptics, RayTracer *anEngine,
		  PictureSampler *aSampler, Storage *aStorage,
		  PostProcessor *aProcess) {
  return Dagnode::newCamera(new Camera(anOptics, anEngine,
				       aSampler, aStorage, aProcess));
}

// Create an extrusion from a bitmap file. Uses the Volume primitive.
object new_extrusion(char *filename, int stop_at_first)
{
  return Dagnode::newPrimitive(create_file_extrusion(filename, stop_at_first));
}

// Create a surface of revolution. Uses the Volume primitive.
object new_revolution(int x_res, int y_res, int z_res,
		      double t_min, double t_max,
		      double (*func)(double tval), int should_stop)
{
  return Dagnode::newPrimitive(create_revolution_surface(x_res, y_res, z_res,
							 t_min, t_max,
							 func, should_stop));
}

// Create a bilinear patch in its primitive space. The version of
// new_bilinear_patch that takes four points is more often directly used.
object new_bilinear_patch(double x, double y, int flat) {
  return Dagnode::newPrimitive(new bilinearPrimitive(x, y, flat));
}

// Create a bilinear patch in its primitive space. Also takes a
// Cutout parameter.
object new_bilinear_patch(double x, double y, int flat, Cutout *cut) {
  return Dagnode::newPrimitive(new bilinearPrimitive(x, y, flat, cut));
}

// Create a two-sided bilinear patch in its primitive space. The version of
// new_bilinear_patch that takes four points is more often directly used.
object new_two_sided_patch(double x, double y, int flat) {
  return Dagnode::newPrimitive(new twoSidedPatchPrimitive(x, y, flat));
}

// Create a two-sided bilinear patch in its primitive space. Also takes a
// Cutout parameter.
object new_two_sided_patch(double x, double y, int flat, Cutout *cut) {
  return Dagnode::newPrimitive(new twoSidedPatchPrimitive(x, y, flat, cut));
}

object new_trunc_cone(double z) {
  return Dagnode::newPrimitive(new truncConePrimitive(z));
}

object new_trunc_cone(double radius_1, double radius_2, double height)
{
  double zval;
  object the_cone;

  // Error check.

  if (radius_1 == 0)
    return new_nothing();

  if (radius_2 > radius_1) {
    fatal_error("new_trunc_cone",
		"The base radius must be bigger or equal to the end radius!",
		"Choose radii wisely, grasshopper.");
  } else if (radius_2 == radius_1) {
    return stretch(radius_1, radius_1, height, bind(prim_unitcyl));
  }

  // A sensible case, then.
  zval = 1 - radius_2 / radius_1;

  the_cone = new_trunc_cone(zval);
  return stretch(radius_1, radius_1, height / zval, the_cone);
}

object new_unitcone(void)
{
  object the_cone;

  the_cone = new_cone();
  LINK(the_cone, MINUS, new_plane());
  LINK(the_cone, MINUS, shift(0, 0, -1, reflect(Z, new_plane())));

  return shift(0, 0, 1, bind(the_cone));
}

object new_limited_light(double r, double g, double b,
			 double limit, dropoff_type light_type)
{
  return Dagnode::newLight(new LimitedLight(r, g, b, limit, light_type));
}

#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif
