/* vectors.h : Vector manipulation utilities
** 
** The Vector class is a toolkit of vector functions that is used by
** the rendering system. They are accessed as Vector::function(a, b, c) e.g.
** Vector::sub(to, from, dir);
**
** Chris Butcher, 27 November 1996
*/

// Define the vector storage. Note that a 'vector' is a different thing to the
// 'Vector' toolkit class ...
// Vectors are stored as an array of 3 doubles (no homogenous value W)
typedef double vector[3];

// Define the parallelipiped storage class 'box' as a set of 8 vectors
typedef vector box[8];

// This class should never be instantiated. It is a toolkit only.
// Most of these are inline functions so will be compiled into tolerably
// quick code.
class Vector {
public:

  // Is a vector zero?
  static int is_zero(vector v)
  {
    return ((v[X] == 0) && (v[Y] == 0) && (v[Z] == 0));
  }

  // Adds two vectors a and b, stores the result in a third vector
  // which may be one of a or b.
  static void add(vector a, vector b, vector result)
  {
    result[X] = a[X] + b[X];
    result[Y] = a[Y] + b[Y];
    result[Z] = a[Z] + b[Z];
  }

  // Add a fraction of vector a to some vector b.
  static void addfrac(double frac, vector a, vector b)
  {
    b[X] += frac * a[X];
    b[Y] += frac * a[Y];
    b[Z] += frac * a[Z];
  }

  // Subtracts two vectors a and b, stores the result in a third
  // vector which may be one of a or b.
  static void sub(vector a, vector b, vector result)
  {
    result[X] = a[X] - b[X];
    result[Y] = a[Y] - b[Y];
    result[Z] = a[Z] - b[Z];
  }

  // Calculates the scalar dot product of two vectors a and b
  static double dot(vector a, vector b)
  {
    return (a[X] * b[X] + a[Y] * b[Y] + a[Z] * b[Z]);
  }

  // Sets vector a to vector b.
  static void set(vector a, vector b)
  {
    a[X] = b[X];
    a[Y] = b[Y];
    a[Z] = b[Z];
  }

  // Assigns the given values to the vector a.
  static void assign(vector a, double xval, double yval, double zval)
  {
    a[X] = xval;
    a[Y] = yval;
    a[Z] = zval;
  }

  // Returns the magnitude of the vector a.
  static double magnitude(vector a)
  {
    return sqrt(SQR(a[X]) + SQR(a[Y]) + SQR(a[Z]));
  }

  // Finds the minimum of two vectors on a value-by-value basis.
  inline static void min(vector a, vector b, vector result)
  {
    result[X] = MIN(a[X], b[X]);
    result[Y] = MIN(a[Y], b[Y]);
    result[Z] = MIN(a[Z], b[Z]);
  }

  // Finds the maximum of two vectors on a value-by-value basis.
  inline static void max(vector a, vector b, vector result)
  {
    result[X] = MAX(a[X], b[X]);
    result[Y] = MAX(a[Y], b[Y]);
    result[Z] = MAX(a[Z], b[Z]);
  }

  // Scales a vector a by the given scalar quantity s, stores it in result.
  // (precisely the same as Vector::scale, just an alias)
  inline static void mult(double s, vector a, vector result)
  {
    result[X] = s * a[X];
    result[Y] = s * a[Y];
    result[Z] = s * a[Z];
  }

  // Scales a vector a by the given scalar quantity s, stores it in result.
  static void scale(double s, vector a, vector result)
  {
    result[X] = s * a[X];
    result[Y] = s * a[Y];
    result[Z] = s * a[Z];
  }

  // Normalises a vector.
  static void normalise(vector v, char *msg)
  {
    Vector::scale(1.0 / CHECK0(Vector::magnitude(v), msg), v, v);
  }

  // Takes the cross product of two vectors (a x b), stores it in result.
  static void cross(vector a, vector b, vector result)
  {
    vector temp;

    // Must use a temp variable rather than direct to result because otherwise
    // we might overwrite values that are needed later in the calculation.
    // This function is safe to call with a or b == result.
    temp[X] = a[Y] * b[Z] - a[Z] * b[Y];
    temp[Y] = a[Z] * b[X] - a[X] * b[Z];

    result[Z] = a[X] * b[Y] - a[Y] * b[X];
    result[Y] = temp[Y];
    result[X] = temp[X];
  }

  // Stretch a vector; same format as the object's stretch routine.
  static void stretch(double x, double y, double z, vector v) {
    v[X] *= x;
    v[Y] *= y;
    v[Z] *= z;
  }

  // Shift a vector; same format as the object's shift routine.
  static void shift(double x, double y, double z, vector v) {
    v[X] += x;
    v[Y] += y;
    v[Z] += z;
  }

  // Rotate a vector; same format as the object's rotate routine.
  static void rotate(axis from, axis to, double angle, vector v) {
    double radians, sine, cosine;
    vector old_v;
    axis around;

    around = (axis) ((from + 1) % 3);
    if (around == to)
      around = (axis) ((to + 1) % 3);

    radians = DEG2RAD(angle);
    sine = sin(radians);
    cosine = cos(radians);

    // Apply the rotation matrix      cos  0 -sin  0
    //                                 0   1   0   0
    //                                sin  0  cos  0  
    Vector::set(old_v, v);
    
    v[from]   = cosine*old_v[from] - sine*old_v[to];
    v[to]     = cosine*old_v[to]   + sine*old_v[from];
    v[around] = old_v[around];
  }

  // Rotate a vector; same format as the object's rotate routine.
  static void rotate(axis around, double angle, vector v) {
    vector old_v;
    int from, to;
    double radians, sine, cosine;

    from = (axis) ((around + 1) % 3);
    to = (axis) ((around + 2) % 3);

    radians = DEG2RAD(angle);
    sine = sin(radians);
    cosine = cos(radians);

    // Apply the rotation matrix      cos  0 -sin  0
    //                                 0   1   0   0
    //                                sin  0  cos  0  
    Vector::set(old_v, v);
    
    v[from]   = cosine*old_v[from] - sine*old_v[to];
    v[to]     = cosine*old_v[to]   + sine*old_v[from];
    v[around] = old_v[around];
  }

  // Rotate a vector; same format as the object's rotate routine.
  static void rotate(vector given_fwd, vector given_up, vector v) {
    vector new_fwd, new_up, new_left, old_v;

    Vector::set(new_fwd, given_fwd);
    Vector::set(new_up, given_up);

    Vector::cross(new_up, new_fwd, new_left);
    Vector::normalise(new_fwd, "Vector::rotate(vec, vec, r), zero fwd vector");

    Vector::cross(new_fwd, new_left, new_up);
    Vector::normalise(new_up, "Vector::rotate(vec, vec, r), zero up-vector");
    
    Vector::normalise(new_left, 
		      "Vector::rotate(vec, vec, r), zero left-vector");

    // Apply the matrix whose columns are the vectors (fwd), (left), (up).
    Vector::set(old_v, v);
    
    v[X] = new_fwd[X]*old_v[X] + new_left[X]*old_v[Y] + new_up[X]*old_v[Z];
    v[Y] = new_fwd[Y]*old_v[X] + new_left[Y]*old_v[Y] + new_up[Y]*old_v[Z];
    v[Z] = new_fwd[Z]*old_v[X] + new_left[Z]*old_v[Y] + new_up[Z]*old_v[Z];
  }

  // Shear a vector; same format as the object's shear routine.
  static void shear(axis from, axis to, double amount, vector v) {
    v[to] += amount * v[from];
  }
};






