/* volume.c : The Volume primitive.
** 
** This file defines the functions for the Volume primitive.
**
** Chris Butcher, 20 February 1997
*/

#include "mirage.h"

#ifdef USE_NAMESPACES
namespace Mirage {
#endif

int Volume::ray_intersect(IntersectionList *the_list, 
			  struct PrimitiveNode *node,
			  struct PrimitiveNode *cluster_node,
			  vector ray_from, vector ray_dir,
			  uint ray_flags, signature ray_signature)
{
  double p_x, p_y, p_z, d_x, d_y, d_z;
  double t_x_entry, t_x_exit, t_y_entry, t_y_exit, t_z_entry, t_z_exit;
  double scaled_x, scaled_y, scaled_z;
  int entry_x, entry_y, entry_z, exit_x, exit_y, exit_z;
  vector norm;
  int entry_plane, exit_plane;
  int found_intersection = FALSE;
  double initial_value = 0, exit_value;

  validate_node = cluster_node;

  // Store the signature of this ray that we're intersecting with.
  node->last_ray_signature = ray_signature;

  // This apply is an inline function defined in matrix.h
  // Transform the ray startpoint into primitive space
  Matrix::apply(node->position[AFT], ray_from, &p_x, &p_y, &p_z);

  // This apply is also an inline function defined in matrix.h
  // Transform the ray direction into primitive space (note: shift does
  // not apply to this, only apply rotation + stretch sections of the matrix)
  Matrix::applynoshift(node->position[AFT], ray_dir, &d_x, &d_y, &d_z);

  if (d_x == 0) {
    // The ray has no X direction component.
    if (p_x <= 0 || p_x >= 1) {
      // The ray is outside the volume in X always.
      node->last_ray_hit = FALSE;
      return node->last_ray_hit;
    } else {
      // The ray is inside the volume in X always.
      t_x_entry = -INFINITY;
      t_x_exit = +INFINITY;
    }
  } else if (d_x > 0) {
    // The ray is travelling in +X.
    t_x_entry = (0 - p_x) / d_x;
    t_x_exit  = (1 - p_x) / d_x;
  } else {
    // The ray is travelling in -X.
    t_x_entry = (1 - p_x) / d_x;
    t_x_exit  = (0 - p_x) / d_x;
  }

  if (d_y == 0) {
    // The ray has no Y direction component.
    if (p_y <= 0 || p_y >= 1) {
      // The ray is outside the volume in Y always.
      node->last_ray_hit = FALSE;
      return node->last_ray_hit;
    } else {
      // The ray is inside the volume in Y always.
      t_y_entry = -INFINITY;
      t_y_exit = +INFINITY;
    }
  } else if (d_y > 0) {
    // The ray is travelling in +Y.
    t_y_entry = (0 - p_y) / d_y;
    t_y_exit  = (1 - p_y) / d_y;
  } else {
    // The ray is travelling in -Y.
    t_y_entry = (1 - p_y) / d_y;
    t_y_exit  = (0 - p_y) / d_y;
  }

  if (d_z == 0) {
    // The ray has no Z direction component.
    if (p_z <= 0 || p_z >= 1) {
      // The ray is outside the volume in Z always.
      node->last_ray_hit = FALSE;
      return node->last_ray_hit;
    } else {
      // The ray is inside the volume in Z always.
      t_z_entry = -INFINITY;
      t_z_exit = +INFINITY;
    }
  } else if (d_z > 0) {
    // The ray is travelling in +Z.
    t_z_entry = (0 - p_z) / d_z;
    t_z_exit  = (1 - p_z) / d_z;
  } else {
    // The ray is travelling in -Z.
    t_z_entry = (1 - p_z) / d_z;
    t_z_exit  = (0 - p_z) / d_z;
  }

  // Work out when the ray enters and exits the heightfield's bounding box.
  if (t_x_entry > t_y_entry) {
    if (t_x_entry > t_z_entry) {
      t_entry = t_x_entry;
      entry_plane = X;
    } else {
      t_entry = t_z_entry;
      entry_plane = Z;
    }
  } else {
    if (t_y_entry > t_z_entry) {
      t_entry = t_y_entry;
      entry_plane = Y;
    } else {
      t_entry = t_z_entry;
      entry_plane = Z;
    }
  }

  if (t_x_exit < t_y_exit) {
    if (t_x_exit < t_z_exit) {
      t_exit = t_x_exit;
      exit_plane = X;
    } else {
      t_exit = t_z_exit;
      exit_plane = Z;
    }
  } else {
    if (t_y_exit < t_z_exit) {
      t_exit = t_y_exit;
      exit_plane = Y;
    } else {
      t_exit = t_z_exit;
      exit_plane = Z;
    }
  }

  if (t_entry > t_exit) {
    // The ray never enters the bounding volume.
    node->last_ray_hit = FALSE;
    return node->last_ray_hit;
  }

  // Classify the ray's entry intersection with the bounding volume.

  // The entry plane intersection has to be checked to see
  // if it is a valid intersection (i.e. value above the surface threshold).
  // In the course of this we also work out what initial_value is.
  if (entry_plane == X) {
    // Work out the place on the X plane where the ray enters.
    scaled_y = (p_y + t_entry * d_y) * (y_res - 1);
    scaled_z = (p_z + t_entry * d_z) * (z_res - 1);

    // Is this a low-X or high-X entry?
    if (d_x > 0) {
      // This enters the volume through the -X bounding plane. Determine
      // the field value at this point.
      entry_x = 0;
      initial_value = x_face_val(entry_x, scaled_y, scaled_z);
      
      // If this is above the threshold and not too close to the from point,
      // then we need to insert an entry intersection.
      if ((initial_value > surf_val) && (t_entry > MIN_TVAL_CUTOFF)) {
	norm[X] = -1;
	norm[Y] = 0;
	norm[Z] = 0;

	found_intersection = TRUE;

	insert_entering(the_list, node, validate_node, norm, t_entry,
			ray_from, ray_dir, ray_flags);
      }
    } else {
      // This enters the volume through the +X bounding plane. Determine
      // the field value at this point.
      entry_x = x_res - 1;
      initial_value = x_face_val(entry_x, scaled_y, scaled_z);
      
      // If this is above the threshold and not too close to the from point,
      // then we need to insert an entry intersection.
      if ((initial_value > surf_val) && (t_entry > MIN_TVAL_CUTOFF)) {
	norm[X] = +1;
	norm[Y] = 0;
	norm[Z] = 0;

	found_intersection = TRUE;

	insert_entering(the_list, node, validate_node, norm, t_entry,
			ray_from, ray_dir, ray_flags);
      }
    }
  } else if (entry_plane == Y) {
    // Work out the place on the Y plane where the ray enters.
    scaled_x = (p_x + t_entry * d_x) * (x_res - 1);
    scaled_z = (p_z + t_entry * d_z) * (z_res - 1);

    // Is this a low-Y or high-Y entry?
    if (d_y > 0) {
      // This enters the volume through the -Y bounding plane. Determine
      // the field value at this point.
      entry_y = 0;
      initial_value = y_face_val(scaled_x, entry_y, scaled_z);
      
      // If this is above the threshold and not too close to the from point,
      // then we need to insert an entry intersection.
      if ((initial_value > surf_val) && (t_entry > MIN_TVAL_CUTOFF)) {
	norm[X] = 0;
	norm[Y] = -1;
	norm[Z] = 0;

	found_intersection = TRUE;

	insert_entering(the_list, node, validate_node, norm, t_entry,
			ray_from, ray_dir, ray_flags);
      }
    } else {
      // This enters the volume through the +Y bounding plane. Determine
      // the field value at this point.
      entry_y = y_res - 1;
      initial_value = y_face_val(scaled_x, entry_y, scaled_z);
      
      // If this is above the threshold and not too close to the from point,
      // then we need to insert an entry intersection.
      if ((initial_value > surf_val) && (t_entry > MIN_TVAL_CUTOFF)) {
	norm[X] = 0;
	norm[Y] = +1;
	norm[Z] = 0;

	found_intersection = TRUE;

	insert_entering(the_list, node, validate_node, norm, t_entry,
			ray_from, ray_dir, ray_flags);
      }
    }
  } else if (entry_plane == Z) {
    // Work out the place on the Z plane where the ray enters.
    scaled_x = (p_x + t_entry * d_x) * (x_res - 1);
    scaled_y = (p_y + t_entry * d_y) * (y_res - 1);

    // Is this a low-Z or high-Z entry?
    if (d_z > 0) {
      // This enters the volume through the -Z bounding plane. Determine
      // the field value at this point.
      entry_z = 0;
      initial_value = z_face_val(scaled_x, scaled_y, entry_z);
      
      // If this is above the threshold and not too close to the from point,
      // then we need to insert an entry intersection.
      if ((initial_value > surf_val) && (t_entry > MIN_TVAL_CUTOFF)) {
	norm[X] = 0;
	norm[Y] = 0;
	norm[Z] = -1;

	found_intersection = TRUE;

	insert_entering(the_list, node, validate_node, norm, t_entry,
			ray_from, ray_dir, ray_flags);
      }
    } else {
      // This enters the volume through the +Z bounding plane. Determine
      // the field value at this point.
      entry_z = z_res - 1;
      initial_value = z_face_val(scaled_x, scaled_y, entry_z);
      
      // If this is above the threshold and not too close to the from point,
      // then we need to insert an entry intersection.
      if ((initial_value > surf_val) && (t_entry > MIN_TVAL_CUTOFF)) {
	norm[X] = 0;
	norm[Y] = 0;
	norm[Z] = +1;

	found_intersection = TRUE;

	insert_entering(the_list, node, validate_node, norm, t_entry,
			ray_from, ray_dir, ray_flags);
      }
    }
  }

  // If we've found an intersection, stop if necessary.
  if (stop_at_first_intersection && found_intersection) {
    node->last_ray_hit = TRUE;
    return node->last_ray_hit;
  }

  // Intersect with the isosurface here.
  if (intersect_with_isosurface(p_x, p_y, p_z, d_x, d_y, d_z,
				t_entry, t_exit, initial_value,
				the_list, node, ray_from, ray_dir,
				ray_flags, ray_signature))
    found_intersection = TRUE;
  
  // If we've found an intersection, stop if necessary.
  if (stop_at_first_intersection && found_intersection) {
    node->last_ray_hit = TRUE;
    return node->last_ray_hit;
  }

  // Classify the ray's exit intersection with the bounding volume.

  // The exit plane intersection has to be checked to see
  // if it is a valid intersection (i.e. value above the surface threshold).
  if (exit_plane == X) {
    // Work out the place on the X plane where the ray exits.
    scaled_y = (p_y + t_exit * d_y) * (y_res - 1);
    scaled_z = (p_z + t_exit * d_z) * (z_res - 1);

    // Is this a low-X or high-X exit?
    if (d_x < 0) {
      // This exits the volume through the -X bounding plane. Determine
      // the field value at this point.
      exit_x = 0;
      exit_value = x_face_val(exit_x, scaled_y, scaled_z);
      
      // If this is above the threshold and not too close to the from point,
      // then we need to insert an exit intersection.
      if ((exit_value > surf_val) && (t_exit > MIN_TVAL_CUTOFF)) {
	norm[X] = -1;
	norm[Y] = 0;
	norm[Z] = 0;

	found_intersection = TRUE;

	insert_leaving(the_list, node, validate_node, norm, t_exit,
		       ray_from, ray_dir, ray_flags);
      }
    } else {
      // This exits the volume through the +X bounding plane. Determine
      // the field value at this point.
      exit_x = x_res - 1;
      exit_value = x_face_val(exit_x, scaled_y, scaled_z);
      
      // If this is above the threshold and not too close to the from point,
      // then we need to insert an exit intersection.
      if ((exit_value > surf_val) && (t_exit > MIN_TVAL_CUTOFF)) {
	norm[X] = +1;
	norm[Y] = 0;
	norm[Z] = 0;

	found_intersection = TRUE;

	insert_leaving(the_list, node, validate_node, norm, t_exit,
		       ray_from, ray_dir, ray_flags);
      }
    }
  } else if (exit_plane == Y) {
    // Work out the place on the Y plane where the ray exits.
    scaled_x = (p_x + t_exit * d_x) * (x_res - 1);
    scaled_z = (p_z + t_exit * d_z) * (z_res - 1);

    // Is this a low-Y or high-Y exit?
    if (d_y < 0) {
      // This exits the volume through the -Y bounding plane. Determine
      // the field value at this point.
      exit_y = 0;
      exit_value = y_face_val(scaled_x, exit_y, scaled_z);
      
      // If this is above the threshold and not too close to the from point,
      // then we need to insert an exit intersection.
      if ((exit_value > surf_val) && (t_exit > MIN_TVAL_CUTOFF)) {
	norm[X] = 0;
	norm[Y] = -1;
	norm[Z] = 0;

	found_intersection = TRUE;

	insert_leaving(the_list, node, validate_node, norm, t_exit,
		       ray_from, ray_dir, ray_flags);
      }
    } else {
      // This exits the volume through the +Y bounding plane. Determine
      // the field value at this point.
      exit_y = y_res - 1;
      exit_value = y_face_val(scaled_x, exit_y, scaled_z);
      
      // If this is above the threshold and not too close to the from point,
      // then we need to insert an exit intersection.
      if ((exit_value > surf_val) && (t_exit > MIN_TVAL_CUTOFF)) {
	norm[X] = 0;
	norm[Y] = +1;
	norm[Z] = 0;

	found_intersection = TRUE;

	insert_leaving(the_list, node, validate_node, norm, t_exit,
		       ray_from, ray_dir, ray_flags);
      }
    }
  } else if (exit_plane == Z) {
    // Work out the place on the Z plane where the ray exits.
    scaled_x = (p_x + t_exit * d_x) * (x_res - 1);
    scaled_y = (p_y + t_exit * d_y) * (y_res - 1);

    // Is this a low-Z or high-Z exit?
    if (d_z < 0) {
      // This exits the volume through the -Z bounding plane. Determine
      // the field value at this point.
      exit_z = 0;
      exit_value = z_face_val(scaled_x, scaled_y, exit_z);
      
      // If this is above the threshold and not too close to the from point,
      // then we need to insert an exit intersection.
      if ((exit_value > surf_val) && (t_exit > MIN_TVAL_CUTOFF)) {
	norm[X] = 0;
	norm[Y] = 0;
	norm[Z] = -1;

	found_intersection = TRUE;

	insert_leaving(the_list, node, validate_node, norm, t_exit,
		       ray_from, ray_dir, ray_flags);
      }
    } else {
      // This exits the volume through the +Z bounding plane. Determine
      // the field value at this point.
      exit_z = z_res - 1;
      exit_value = z_face_val(scaled_x, scaled_y, exit_z);
      
      // If this is above the threshold and not too close to the from point,
      // then we need to insert an exit intersection.
      if ((exit_value > surf_val) && (t_exit > MIN_TVAL_CUTOFF)) {
	norm[X] = 0;
	norm[Y] = 0;
	norm[Z] = +1;

	found_intersection = TRUE;

	insert_leaving(the_list, node, validate_node, norm, t_exit,
		       ray_from, ray_dir, ray_flags);
      }
    }
  }

  node->last_ray_hit = found_intersection;
  return node->last_ray_hit;
}

int Volume::ray_intersect(vector ray_from, vector ray_dir,
			  rayintersect *int_storage)
{
  // This routine is just too damn complicated to copy. It's not
  // likely to be needed for the interactive-CSG system, either.
  fatal_error("Volume::ray_intersect", "Incorrect r_i call!",
	      "This error should not happen.");
  return 0;
}

int Volume::intersect_with_isosurface(double p_x, double p_y, double p_z,
				      double d_x, double d_y, double d_z,
				      double t_entry, double t_exit,
				      double initial_value,
				      IntersectionList *the_list,
				      primnode *node, vector ray_from,
				      vector ray_dir, uint ray_flags,
				      signature ray_signature)
{
  double entry_x, entry_y, entry_z;
  double start_cell_x, start_cell_y, start_cell_z;
  double cell_frac_x, cell_frac_y, cell_frac_z;
  double ray_length, t_next_x, t_next_y, t_next_z, cell_entry_t, cell_exit_t;
  double delta_t_per_x_cell, delta_t_per_y_cell, delta_t_per_z_cell;
  int x_cell_no, y_cell_no, z_cell_no, x_incr, y_incr, z_incr;
  int x_leave_side, y_leave_side, z_leave_side;
  double cell_entry_val, cell_exit_val;
  int found_intersection, this_cell_found, ray_starts_inside;

  // Find the position where the ray enters the volume (alternatively,
  // if the ray starts inside the heightfield, the position where the ray
  // starts).
  if (t_entry > 0) {
    // The ray starts outside the volume and hits it in +t.
    ray_starts_inside = FALSE;
    entry_x = p_x + t_entry * d_x;
    entry_y = p_y + t_entry * d_y;
    entry_z = p_z + t_entry * d_z;
  } else if (t_exit > MIN_TVAL_CUTOFF) {
    // The ray starts inside the volume and exits it in +t.
    ray_starts_inside = TRUE;
    entry_x = p_x;
    entry_y = p_y;
    entry_z = p_z;
  } else {
    // The ray starts in front of the volume (t_entry, t_exit both -ve).
    // No intersections.
    return FALSE;
  }

  // Get the coordintes of the ray entry in terms of cells.
  start_cell_x = entry_x * (x_res - 1);
  start_cell_y = entry_y * (y_res - 1);
  start_cell_z = entry_z * (z_res - 1);

  x_cell_no = (int) start_cell_x;
  y_cell_no = (int) start_cell_y;
  z_cell_no = (int) start_cell_z;

  cell_frac_x = start_cell_x - x_cell_no;
  cell_frac_y = start_cell_y - y_cell_no;
  cell_frac_z = start_cell_z - z_cell_no;

  // Ensure that we start in the right volume cell and not in a cell
  // just outside the volume's extent.
  if (x_cell_no < 0) {
    x_cell_no = 0;
    cell_frac_x = 0;
  } else if (x_cell_no > x_res - 2) {
    x_cell_no = x_res - 2;
    cell_frac_x = 1;
  }

  if (y_cell_no < 0) {
    y_cell_no = 0;
    cell_frac_y = 0;
  } else if (y_cell_no > y_res - 2) {
    y_cell_no = y_res - 2;
    cell_frac_y = 1;
  }

  if (z_cell_no < 0) {
    z_cell_no = 0;
    cell_frac_z = 0;
  } else if (z_cell_no > z_res - 2) {
    z_cell_no = z_res - 2;
    cell_frac_z = 1;
  }

  // Set up the t values to the next X, Y and Z cell boundaries. Also determine
  // the delta-t that the ray takes to cross an X, Y or Z cell.
  ray_length = sqrt(SQR(d_x) + SQR(d_y) + SQR(d_z));

  if (d_x == 0) {
    // The ray is parallel to the X boundaries and never crosses them.
    t_next_x = INFINITY;
    delta_t_per_x_cell = INFINITY;
  } else if (d_x > 0) {
    // We are travelling from low X to high X across the cells.

    // Work out how long in t it takes to cross an X cell.
    delta_t_per_x_cell = 1.0 / (d_x * (x_res - 1));

    // Work out how long in t until we cross the next boundary.
    t_next_x = (1.0 - cell_frac_x) * delta_t_per_x_cell;

    // If it's going to be a looong time, flag it as being so.
    if (t_next_x >= INFINITY) {
      t_next_x = INFINITY;
      delta_t_per_x_cell = INFINITY;
    }

  } else {
    // We are travelling from high X to low X across the cells.

    // Work out how long in t it takes to cross an X cell.
    delta_t_per_x_cell = 1.0 / (-d_x * (x_res - 1));

    // Work out how long in t until we cross the next boundary.
    t_next_x = cell_frac_x * delta_t_per_x_cell;

    // If it's going to be a looong time, flag it as being so.
    if (t_next_x >= INFINITY) {
      t_next_x = INFINITY;
      delta_t_per_x_cell = INFINITY;
    }
  }

  if (d_y == 0) {
    // The ray is parallel to the Y boundaries and never crosses them.
    t_next_y = INFINITY;
    delta_t_per_y_cell = INFINITY;
  } else if (d_y > 0) {
    // We are travelling from low Y to high Y across the cells.

    // Work out how long in t it takes to cross a Y cell.
    delta_t_per_y_cell = 1.0 / (d_y * (y_res - 1));

    // Work out how long in t until we cross the next boundary.
    t_next_y = (1.0 - cell_frac_y) * delta_t_per_y_cell;

    // If it's going to be a looong time, flag it as being so.
    if (t_next_y >= INFINITY) {
      t_next_y = INFINITY;
      delta_t_per_y_cell = INFINITY;
    }
  } else {
    // We are travelling from high Y to low Y across the cells.

    // Work out how long in t it takes to cross a Y cell.
    delta_t_per_y_cell = 1.0 / (-d_y * (y_res - 1));

    // Work out how long in t until we cross the next boundary.
    t_next_y = cell_frac_y * delta_t_per_y_cell;

    // If it's going to be a looong time, flag it as being so.
    if (t_next_y >= INFINITY) {
      t_next_y = INFINITY;
      delta_t_per_y_cell = INFINITY;
    }
  }

  if (d_z == 0) {
    // The ray is parallel to the Z boundaries and never crosses them.
    t_next_z = INFINITY;
    delta_t_per_z_cell = INFINITY;
  } else if (d_z > 0) {
    // We are travelling from low Z to high Z across the cells.

    // Work out how long in t it takes to cross a Z cell.
    delta_t_per_z_cell = 1.0 / (d_z * (z_res - 1));

    // Work out how long in t until we cross the next boundary.
    t_next_z = (1.0 - cell_frac_z) * delta_t_per_z_cell;

    // If it's going to be a looong time, flag it as being so.
    if (t_next_z >= INFINITY) {
      t_next_z = INFINITY;
      delta_t_per_z_cell = INFINITY;
    }
  } else {
    // We are travelling from high Z to low Z across the cells.

    // Work out how long in t it takes to cross a Z cell.
    delta_t_per_z_cell = 1.0 / (-d_z * (z_res - 1));

    // Work out how long in t until we cross the next boundary.
    t_next_z = cell_frac_z * delta_t_per_z_cell;

    // If it's going to be a looong time, flag it as being so.
    if (t_next_z >= INFINITY) {
      t_next_z = INFINITY;
      delta_t_per_z_cell = INFINITY;
    }
  }

  // We are currently measuring t_next_x, t_next_y and t_next_z relative to the
  // point where traversal starts. If the ray starts outside the bounds
  // of the heightfield, we must add t_entry.
  if (!ray_starts_inside) {
    t_next_x += t_entry;
    t_next_y += t_entry;
    t_next_z += t_entry;
    cell_entry_t = t_entry;
  } else {
    // The ray starts inside the volume. We must recalculate the initial_value
    // to work out whether the start point is inside or outside the surface.
    initial_value = interpolate_value(p_x * (x_res - 1),
				      p_y * (y_res - 1),
				      p_z * (z_res - 1));
    cell_entry_t = 0;
  }

  // Get the directions to step in.
  x_incr = (d_x < 0) ? -1 : +1;
  y_incr = (d_y < 0) ? -1 : +1;
  z_incr = (d_z < 0) ? -1 : +1;

  // Get the sides of the current cell to check when leaving.
  x_leave_side = (d_x < 0) ? 0 : 1;
  y_leave_side = (d_y < 0) ? 0 : 1;
  z_leave_side = (d_z < 0) ? 0 : 1;

  found_intersection = FALSE;

  cell_entry_val = initial_value;

  // Setup complete.

  // ****************************************************************
  
  // Step through cells looking for intersections.

  while (1) {

    // Work out the next cell that we will go to ... and process the current
    // cell for intersections if we find ourselves on the opposite side of the
    // surface.
    if (t_next_x <= t_next_y) {
      if (t_next_x <= t_next_z) {
	// The next step is in X.
	cell_exit_t = t_next_x;
	
	// When we leave the cell, what is the surface value?
	cell_exit_val = x_face_val(x_cell_no + x_leave_side,
				   (p_y + cell_exit_t * d_y) * (y_res - 1),
				   (p_z + cell_exit_t * d_z) * (z_res - 1));
	
	// If we've crossed the surface value, intersect with it in the current
	// cell. We must not do this if we started on the surface (interpolated
	// value within 1% of the (surface - exit difference) of the surface
	// value) and the cell entry t was 0 (ray started in this cell).
	// Otherwise we may get erroneous self-shadowing.
	if (((cell_entry_t != 0) ||
	     (fabs(cell_entry_val - surf_val) >
	               0.1 * fabs(cell_exit_val - cell_entry_val))) &&
	    (((cell_entry_val <  surf_val) && (cell_exit_val >= surf_val)) ||
	    ((cell_entry_val >= surf_val) && (cell_exit_val <  surf_val)))) {

	  this_cell_found = get_intersection(x_cell_no, y_cell_no, z_cell_no,
					     cell_entry_t, cell_exit_t,
					     cell_entry_val, cell_exit_val,
					     p_x, p_y, p_z, d_x, d_y, d_z, 
					     the_list, node, ray_from, ray_dir,
					     ray_flags, ray_signature);
	  if (this_cell_found)
	    found_intersection = TRUE;

	  if (stop_at_first_intersection && found_intersection)
	    return TRUE;
	}
    
	t_next_x += delta_t_per_x_cell;
	x_cell_no += x_incr;
      } else {
	// The next step is in Z.
	cell_exit_t = t_next_z;
	
	// When we leave the cell, what is the surface value?
	cell_exit_val = z_face_val((p_x + cell_exit_t * d_x) * (x_res - 1),
				   (p_y + cell_exit_t * d_y) * (y_res - 1),
				   z_cell_no + z_leave_side);
	
	// If we've crossed the surface value, intersect with it in the current
	// cell. We must not do this if we started on the surface (interpolated
	// value within 1% of the (surface - exit difference) of the surface
	// value) and the cell entry t was 0 (ray started in this cell).
	// Otherwise we may get erroneous self-shadowing.
	if (((cell_entry_t != 0) ||
	     (fabs(cell_entry_val - surf_val) >
	               0.1 * fabs(cell_exit_val - cell_entry_val))) &&
	    (((cell_entry_val <  surf_val) && (cell_exit_val >= surf_val)) ||
	    ((cell_entry_val >= surf_val) && (cell_exit_val <  surf_val)))) {

	  this_cell_found = get_intersection(x_cell_no, y_cell_no, z_cell_no,
					     cell_entry_t, cell_exit_t,
					     cell_entry_val, cell_exit_val,
					     p_x, p_y, p_z, d_x, d_y, d_z, 
					     the_list, node, ray_from, ray_dir,
					     ray_flags, ray_signature);
	  
	  if (this_cell_found)
	    found_intersection = TRUE;
	  
	  if (stop_at_first_intersection && found_intersection)
	    return TRUE;
	}
    
	t_next_z += delta_t_per_z_cell;
	z_cell_no += z_incr;
      }
    } else {
      if (t_next_y <= t_next_z) {
	// The next step is in Y.
	cell_exit_t = t_next_y;
	
	// When we leave the cell, what is the surface value?
	cell_exit_val = y_face_val((p_x + cell_exit_t * d_x) * (x_res - 1),
				   y_cell_no + y_leave_side,
				   (p_z + cell_exit_t * d_z) * (z_res - 1));
	
	// If we've crossed the surface value, intersect with it in the current
	// cell. We must not do this if we started on the surface (interpolated
	// value within 1% of the (surface - exit difference) of the surface
	// value) and the cell entry t was 0 (ray started in this cell).
	// Otherwise we may get erroneous self-shadowing.
	if (((cell_entry_t != 0) ||
	     (fabs(cell_entry_val - surf_val) >
	               0.1 * fabs(cell_exit_val - cell_entry_val))) &&
	    (((cell_entry_val <  surf_val) && (cell_exit_val >= surf_val)) ||
	    ((cell_entry_val >= surf_val) && (cell_exit_val <  surf_val)))) {

	  this_cell_found = get_intersection(x_cell_no, y_cell_no, z_cell_no,
					     cell_entry_t, cell_exit_t,
					     cell_entry_val, cell_exit_val,
					     p_x, p_y, p_z, d_x, d_y, d_z, 
					     the_list, node, ray_from, ray_dir,
					     ray_flags, ray_signature);
	  
	  if (this_cell_found)
	    found_intersection = TRUE;
	  
	  if (stop_at_first_intersection && found_intersection)
	    return TRUE;
	}
    
	t_next_y += delta_t_per_y_cell;
	y_cell_no += y_incr;
      } else {
	// The next step is in Z.
	cell_exit_t = t_next_z;
	
	// When we leave the cell, what is the surface value?
	cell_exit_val = z_face_val((p_x + cell_exit_t * d_x) * (x_res - 1),
				   (p_y + cell_exit_t * d_y) * (y_res - 1),
				   z_cell_no + z_leave_side);
	
	// If we've crossed the surface value, intersect with it in the current
	// cell. We must not do this if we started on the surface (interpolated
	// value within 1% of the (surface - exit difference) of the surface
	// value) and the cell entry t was 0 (ray started in this cell).
	// Otherwise we may get erroneous self-shadowing.
	if (((cell_entry_t != 0) ||
	     (fabs(cell_entry_val - surf_val) >
	               0.1 * fabs(cell_exit_val - cell_entry_val))) &&
	    (((cell_entry_val <  surf_val) && (cell_exit_val >= surf_val)) ||
	    ((cell_entry_val >= surf_val) && (cell_exit_val <  surf_val)))) {

	  this_cell_found = get_intersection(x_cell_no, y_cell_no, z_cell_no,
					     cell_entry_t, cell_exit_t,
					     cell_entry_val, cell_exit_val,
					     p_x, p_y, p_z, d_x, d_y, d_z, 
					     the_list, node, ray_from, ray_dir,
					     ray_flags, ray_signature);
	  
	  if (this_cell_found)
	    found_intersection = TRUE;
	  
	  if (stop_at_first_intersection && found_intersection)
	    return TRUE;
	}
    
	t_next_z += delta_t_per_z_cell;
	z_cell_no += z_incr;
      }
    }

    // Check to see if we just left the volume. If so, stop traversing.
    if ((x_cell_no < 0         && x_incr < 0) ||
	(y_cell_no < 0         && y_incr < 0) ||
	(z_cell_no < 0         && z_incr < 0) ||
	(x_cell_no > x_res - 2 && x_incr > 0) ||
	(y_cell_no > y_res - 2 && y_incr > 0) ||
	(z_cell_no > z_res - 2 && z_incr > 0)) {
      return found_intersection;
    }

    // Set up for the traversal of the next cell.
    cell_entry_t = cell_exit_t;
    cell_entry_val = cell_exit_val;
  }

  // We never get here.
  return FALSE;
}

int Volume::get_intersection(int x_index, int y_index, int z_index,
			     double cell_entry_t, double cell_exit_t,
			     double cell_entry_val, double cell_exit_val,
			     double from_x, double from_y, double from_z,
			     double dir_x, double dir_y, double dir_z, 
			     IntersectionList *the_list,
			     primnode *node, vector ray_from,
			     vector ray_dir, uint ray_flags,
			     signature ray_signature)
{
  double hit_x, hit_y, hit_z, px, py, pz, dx, dy, dz;
  double a, b, c, d, e, f, g, h, c_coeff[4], c_sols[4];
  double val0, val1, val2, val3, val4, val5, val6, val7;
  int array_index, num_sols, current_sol, found_intersection = FALSE;;
  double intersect_t;
  vector norm;

  array_index = x_index * x_step + y_index * y_step + z_index * z_step;

  // Get the values of the field at the vertices.
  val0 = surf_data[array_index                           ];
  val1 = surf_data[array_index + x_step                  ];
  val2 = surf_data[array_index          + y_step         ];
  val3 = surf_data[array_index + x_step + y_step         ];
  val4 = surf_data[array_index                   + z_step];
  val5 = surf_data[array_index + x_step          + z_step];
  val6 = surf_data[array_index          + y_step + z_step];
  val7 = surf_data[array_index + x_step + y_step + z_step];

  // Get the from point and direction vector in the 'primitive space' of the
  // cell ... i.e. the edges of the cell map to 0-1 X, Y and Z.
  px = from_x * (x_res - 1) - x_index;
  py = from_y * (y_res - 1) - y_index;
  pz = from_z * (z_res - 1) - z_index;
  dx = dir_x * (x_res - 1);
  dy = dir_y * (y_res - 1);
  dz = dir_z * (z_res - 1);

  // Work out coefficients that describe the surface in the cell.
  a = - val0 + val1 + val2 - val3 + val4 - val5 - val6 + val7;
  b = val0 - val1 - val2 + val3;
  c = val0 - val1 - val4 + val5;
  d = val0 - val2 - val4 + val6;
  e = - val0 + val1;
  f = - val0 + val2;
  g = - val0 + val4;
  h = val0 - surf_val;

  // Form the cubic equation.
  c_coeff[3] = a * dx * dy * dz;
  c_coeff[2] = dx * dy * (a * pz + b) +
               dx * dz * (a * py + c) +
               dy * dz * (a * px + d);
  c_coeff[1] = dx * (a * py * pz + b * py + c * pz + e) +
               dy * (a * px * pz + b * px + d * pz + f) +
               dz * (a * px * py + c * px + d * py + g);
  c_coeff[0] = a * px * py * pz +
               b * px * py + c * px * pz + d * py * pz +
               e * px + f * py + g * pz +
               h;

  // Solve the cubic equation.
  num_sols = solveCubic(c_coeff, c_sols);

  for (current_sol = 0; current_sol < num_sols; current_sol++) {
    intersect_t = c_sols[current_sol];

    if ((intersect_t >= cell_entry_t) &&
	(intersect_t <= cell_exit_t) &&
	(intersect_t > MIN_TVAL_CUTOFF)) {

      // Get the hitpoint in the cell's primitive space.
      hit_x = px + dx * intersect_t;
      hit_y = py + dy * intersect_t;
      hit_z = pz + dz * intersect_t;

      if (interpolate_normals) {
	// Get the hitpoint in terms of cells.
	hit_x += x_index;
	hit_y += y_index;
	hit_z += z_index;

	// Get the interpolated gradient of the field at this point.
	get_field_gradient(hit_x, hit_y, hit_z, norm);
      } else {
	// Get the gradient of the pseudo-linear field in this
	// single cell.
	norm[X] = a * hit_y * hit_z + b * hit_y + c * hit_z + e;
	norm[Y] = a * hit_x * hit_z + b * hit_x + d * hit_z + f;
	norm[Z] = a * hit_x * hit_y + c * hit_x + d * hit_y + g;
      }

      // The normal is the inverse of the gradient.
      Vector::mult(-1, norm, norm);

      if (cell_entry_val < surf_val) {
	// We started this cell outside the surface.
	insert_entering(the_list, node, validate_node, norm, intersect_t,
			ray_from, ray_dir, ray_flags);
      } else {
	// We started this cell inside the surface.
	insert_leaving(the_list, node, validate_node, norm, intersect_t,
		       ray_from, ray_dir, ray_flags);
      }

      found_intersection = TRUE;
    }
  }

  return found_intersection;
}

// Test to see if a single point is inside the volume (a point in
// primitive space).
intersect_result Volume::point_intersect(primnode *node, vector apoint,
					 PointInfo *info,
					 uint ray_flags)
{
  double p_x, p_y, p_z;
  double volume_val;

  // Transform the point into primitivespace.
  Matrix::apply(node->position[AFT], apoint, &p_x, &p_y, &p_z);

  // Check to see if the point is inside the volume's bounding box.
  if (p_z > 0 && p_z < 1 &&
      p_x > 0  && p_x < 1 &&
      p_y > 0  && p_y < 1) {
    
    // The point lies inside the volume. Sample the field at this point.
    volume_val = interpolate_value(p_x * (x_res - 1),
				   p_y * (y_res - 1),
				   p_z * (z_res - 1));

    if (volume_val < surf_val) {
      // Return an OUT result.
      return OUT;
    } else {
      // Return an IN result.
      set_point_contents(node, info);
      return IN;
    }
  }
  
  // The point is outside the volume's bounding box.
  return OUT;
}

intersect_result Volume::point_intersect(vector wp)
{
  // This routine is just too damn complicated to copy. It's not
  // likely to be needed for the interactive-CSG system, either.
  fatal_error("Volume::point_intersect", "Incorrect p_i call!",
	      "This error should not happen.");
  return OUT;
}

// Calculate the value of the surface at a point.
double Volume::interpolate_value(double px, double py, double pz)
{
  int cell_x, cell_y, cell_z;
  double f_x, f_y, f_z, b_x, b_y, b_z;
  int array_index;
  double found_val;

  // Calculate the cell that we are in.
  cell_x = (int) px;
  cell_y = (int) py;
  cell_z = (int) pz;

  // Calculate where we are in the cell (forward and backward fractions).
  f_x = px - cell_x;
  f_y = py - cell_y;
  f_z = pz - cell_z;
  b_x = 1.0 - f_x;
  b_y = 1.0 - f_y;
  b_z = 1.0 - f_z;

  if (cell_x < 0) {
    cell_x = 0;
    f_x = 0;
    b_x = 1;
  } else if (cell_x >= x_res - 1) {
    cell_x = x_res - 2;
    f_x = 1;
    b_x = 0;
  }

  if (cell_y < 0) {
    cell_y = 0;
    f_y = 0;
    b_y = 1;
  } else if (cell_y >= y_res - 1) {
    cell_y = y_res - 2;
    f_y = 1;
    b_y = 0;
  }

  if (cell_z < 0) {
    cell_z = 0;
    f_z = 0;
    b_z = 1;
  } else if (cell_z >= z_res - 1) {
    cell_z = z_res - 2;
    f_z = 1;
    b_z = 0;
  }

  // Calculate the array index at the base of the cell (SWB corner).
  array_index = cell_x * x_step + cell_y * y_step + cell_z * z_step;

  // Interpolate the values from the eight vertices of the current cell.
  found_val  = b_x * b_y * b_z * surf_data[array_index];
  found_val += f_x * b_y * b_z * surf_data[array_index + x_step];
  found_val += b_x * f_y * b_z * surf_data[array_index + y_step];
  found_val += f_x * f_y * b_z * surf_data[array_index + x_step + y_step];
  found_val += b_x * b_y * f_z * surf_data[array_index + z_step];
  found_val += f_x * b_y * f_z * surf_data[array_index + x_step + z_step];
  found_val += b_x * f_y * f_z * surf_data[array_index + y_step + z_step];
  found_val += f_x * f_y * f_z * surf_data[array_index + x_step + y_step
                                                                + z_step];

  return found_val;
}

double Volume::x_face_val(int cell_x, double py, double pz)
{
  int cell_y, cell_z;
  double f_y, f_z, b_y, b_z;
  int array_index;
  double found_val;

  // Calculate the cell that we are in.
  cell_y = (int) py;
  cell_z = (int) pz;

  // Calculate where we are in the cell (forward and backward fractions).
  f_y = py - cell_y;
  f_z = pz - cell_z;
  b_y = 1.0 - f_y;
  b_z = 1.0 - f_z;

  if (cell_y < 0) {
    cell_y = 0;
    f_y = 0;
    b_y = 1;
  } else if (cell_y >= y_res - 1) {
    cell_y = y_res - 2;
    f_y = 1;
    b_y = 0;
  }

  if (cell_z < 0) {
    cell_z = 0;
    f_z = 0;
    b_z = 1;
  } else if (cell_z >= z_res - 1) {
    cell_z = z_res - 2;
    f_z = 1;
    b_z = 0;
  }

  // Calculate the array index at the base of the face (SB corner).
  array_index = cell_x * x_step + cell_y * y_step + cell_z * z_step;

  // Interpolate the values from the four vertices of the current face.
  found_val  = b_y * b_z * surf_data[array_index];
  found_val += f_y * b_z * surf_data[array_index + y_step];
  found_val += b_y * f_z * surf_data[array_index + z_step];
  found_val += f_y * f_z * surf_data[array_index + y_step + z_step];

  return found_val;
}

double Volume::y_face_val(double px, int cell_y, double pz)
{
  int cell_x, cell_z;
  double f_x, f_z, b_x, b_z;
  int array_index;
  double found_val;

  // Calculate the cell that we are in.
  cell_x = (int) px;
  cell_z = (int) pz;

  // Calculate where we are in the cell (forward and backward fractions).
  f_x = px - cell_x;
  f_z = pz - cell_z;
  b_x = 1.0 - f_x;
  b_z = 1.0 - f_z;

  if (cell_x < 0) {
    cell_x = 0;
    f_x = 0;
    b_x = 1;
  } else if (cell_x >= x_res - 1) {
    cell_x = x_res - 2;
    f_x = 1;
    b_x = 0;
  }

  if (cell_z < 0) {
    cell_z = 0;
    f_z = 0;
    b_z = 1;
  } else if (cell_z >= z_res - 1) {
    cell_z = z_res - 2;
    f_z = 1;
    b_z = 0;
  }

  // Calculate the array index at the base of the face (SB corner).
  array_index = cell_x * x_step + cell_y * y_step + cell_z * z_step;

  // Interpolate the values from the four vertices of the current face.
  found_val  = b_x * b_z * surf_data[array_index];
  found_val += f_x * b_z * surf_data[array_index + x_step];
  found_val += b_x * f_z * surf_data[array_index + z_step];
  found_val += f_x * f_z * surf_data[array_index + x_step + z_step];

  return found_val;
}

double Volume::z_face_val(double px, double py, int cell_z)
{
  int cell_x, cell_y;
  double f_x, f_y, b_x, b_y;
  int array_index;
  double found_val;

  // Calculate the cell that we are in.
  cell_x = (int) px;
  cell_y = (int) py;

  // Calculate where we are in the cell (forward and backward fractions).
  f_x = px - cell_x;
  f_y = py - cell_y;
  b_x = 1.0 - f_x;
  b_y = 1.0 - f_y;

  if (cell_x < 0) {
    cell_x = 0;
    f_x = 0;
    b_x = 1;
  } else if (cell_x >= x_res - 1) {
    cell_x = x_res - 2;
    f_x = 1;
    b_x = 0;
  }

  if (cell_y < 0) {
    cell_y = 0;
    f_y = 0;
    b_y = 1;
  } else if (cell_y >= y_res - 1) {
    cell_y = y_res - 2;
    f_y = 1;
    b_y = 0;
  }

  // Calculate the array index at the base of the face (SB corner).
  array_index = cell_x * x_step + cell_y * y_step + cell_z * z_step;

  // Interpolate the values from the four vertices of the current face.
  found_val  = b_x * b_y * surf_data[array_index];
  found_val += f_x * b_y * surf_data[array_index + x_step];
  found_val += b_x * f_y * surf_data[array_index + y_step];
  found_val += f_x * f_y * surf_data[array_index + x_step + y_step];

  return found_val;
}

void Volume::get_field_gradient(double px, double py, double pz, vector grad)
{
  int cell_x, cell_y, cell_z;
  double f_x, f_y, f_z, b_x, b_y, b_z;
  vector vertex_normal[8];

  // Determine which cell of the volume we are in.
  cell_x = (int) px;
  cell_y = (int) py;
  cell_z = (int) pz;

  // Determine the gradient of the field at each of the eight vertices of the
  // current cell.
  get_vertex_gradient(cell_x,     cell_y,     cell_z,     vertex_normal[0]);
  get_vertex_gradient(cell_x + 1, cell_y,     cell_z,     vertex_normal[1]);
  get_vertex_gradient(cell_x,     cell_y + 1, cell_z,     vertex_normal[2]);
  get_vertex_gradient(cell_x + 1, cell_y + 1, cell_z,     vertex_normal[3]);
  get_vertex_gradient(cell_x,     cell_y,     cell_z + 1, vertex_normal[4]);
  get_vertex_gradient(cell_x + 1, cell_y,     cell_z + 1, vertex_normal[5]);
  get_vertex_gradient(cell_x,     cell_y + 1, cell_z + 1, vertex_normal[6]);
  get_vertex_gradient(cell_x + 1, cell_y + 1, cell_z + 1, vertex_normal[7]);

  // Calculate where we are in the cell (forward and backward fractions).
  f_x = px - cell_x;
  f_y = py - cell_y;
  f_z = pz - cell_z;
  b_x = 1.0 - f_x;
  b_y = 1.0 - f_y;
  b_z = 1.0 - f_z;

  // Initialise the gradient to zero.
  Vector::assign(grad, 0, 0, 0);

  // Interpolate the gradient from the eight vertex gradients.
  Vector::addfrac(b_x * b_y * b_z, vertex_normal[0], grad);
  Vector::addfrac(f_x * b_y * b_z, vertex_normal[1], grad);
  Vector::addfrac(b_x * f_y * b_z, vertex_normal[2], grad);
  Vector::addfrac(f_x * f_y * b_z, vertex_normal[3], grad);
  Vector::addfrac(b_x * b_y * f_z, vertex_normal[4], grad);
  Vector::addfrac(f_x * b_y * f_z, vertex_normal[5], grad);
  Vector::addfrac(b_x * f_y * f_z, vertex_normal[6], grad);
  Vector::addfrac(f_x * f_y * f_z, vertex_normal[7], grad);
}

void Volume::get_vertex_gradient(int vx, int vy, int vz, vector grad)
{
  int array_index;

  // Determine the position in the surface data array.
  array_index = vx * x_step + vy * y_step + vz * z_step;
  
  // Calculate an approximation to the partial derivative of the field
  // in X. We go one cell on either side unless this would take us outside
  // the bounds of the volume.
  if (vx == 0)
    grad[X] = surf_data[array_index + x_step] - surf_data[array_index];
  else if (vx == (x_res - 1))
    grad[X] = surf_data[array_index] - surf_data[array_index - x_step];
  else
    grad[X] = (surf_data[array_index + x_step]
                   - surf_data[array_index - x_step]) / 2.0;

  // Calculate an approximation to the partial derivative of the field
  // in Y. We go one cell on either side unless this would take us outside
  // the bounds of the volume.
  if (vy == 0)
    grad[Y] = surf_data[array_index + y_step] - surf_data[array_index];
  else if (vy == (y_res - 1))
    grad[Y] = surf_data[array_index] - surf_data[array_index - y_step];
  else
    grad[Y] = (surf_data[array_index + y_step]
                   - surf_data[array_index - y_step]) / 2.0;

  // Calculate an approximation to the partial derivative of the field
  // in Z. We go one cell on either side unless this would take us outside
  // the bounds of the volume.
  if (vz == 0)
    grad[Z] = surf_data[array_index + z_step] - surf_data[array_index];
  else if (vz == (z_res - 1))
    grad[Z] = surf_data[array_index] - surf_data[array_index - z_step];
  else
    grad[Z] = (surf_data[array_index + z_step]
                   - surf_data[array_index - z_step]) / 2.0;
}

// Test to see if a voxel intersects with the volume's bounding
// volume (X, Y, Z from 0 - 1).
intersect_result Volume::voxel_intersect(box v)
{
  box cubeprim_box;
  int i;
  intersect_result cube_result;

  // Calculate the equivalent box relative to a -1 - +1 cube in
  // X, Y and Z.
  for (i = 0; i < 8; i++) {
    cubeprim_box[i][X] = 2*v[i][X] - 1;
    cubeprim_box[i][Y] = 2*v[i][Y] - 1;
    cubeprim_box[i][Z] = 2*v[i][Z] - 1;
  }

  cube_result = cube_voxel_intersect(cubeprim_box);

  // If the voxel doesn't intersect the bounding volume, return OUT.
  if (cube_result == OUT)
    return OUT;

  // Otherwise, the volume will need to be tested against.
  // This is not a real voxel intersection test; ideally this would actually
  // check the voxel against the cells of the volume to see if it
  // intersected the surface.
  return SURFACE;
}





#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif
