/* volume.h : Header file for the Volume primitive.
** 
** This file defines the prototypes and definitions for the volume
** (isosurface) primitive.
**
** Chris Butcher, 20 February 1997
*/

//   The Volume primitive is defined by a field in the region (0, 0, 0) to
// (1, 1, 1). The surface of the object is obtained as an isosurface (a
// three-dimensional contour line) at a certain value of the field.
//   Currently the only method of field storage supported is a uniform array
// of field values stored at the vertices of cells.
class Volume : public Primitive {
public:
  // Variables that hold the data for the Volume primitive.
  isosurface_val *surf_data;
  isosurface_val surf_val;
  int x_res, y_res, z_res;
  int x_step, y_step, z_step;
  int stop_at_first_intersection;
  int interpolate_normals;

  // Stores the primnode which will validate our intersections.
  struct PrimitiveNode *validate_node;

  // Do we need to deallocate the field data when destroyed?
  int deallocate_field_data;

  // Entry/exit points of the ray under consideration.
  double t_entry, t_exit;

  // Intersection interface (subclassed from Primitive)
  virtual int ray_intersect(IntersectionList *the_list, 
			    struct PrimitiveNode *node,
			    struct PrimitiveNode *cluster_node,
			    vector ray_from, vector ray_dir, uint ray_flags,
			    signature ray_signature);

  virtual intersect_result point_intersect(primnode *node, vector apoint,
					   PointInfo *info,
					   uint ray_flags);
  virtual int ray_intersect(vector ray_from,
			    vector ray_dir, rayintersect *int_storage);
  virtual intersect_result point_intersect(vector apoint);
  virtual intersect_result voxel_intersect(box v);

  // Internal routines used for intersection.
  int intersect_with_isosurface(double p_x, double p_y, double p_z,
				double d_x, double d_y, double d_z,
				double t_entry, double t_exit,
				double initial_value,
				IntersectionList *the_list,
				primnode *node, vector ray_from,
				vector ray_dir, uint ray_flags,
				signature ray_signature);
  int get_intersection(int x_index, int y_index, int z_index,
		       double cell_entry_t, double cell_exit_t,
		       double cell_entry_val, double cell_exit_val,
		       double from_x, double from_y, double from_z,
		       double dir_x, double dir_y, double dir_z, 
		       IntersectionList *the_list,
		       primnode *node, vector ray_from,
		       vector ray_dir, uint ray_flags,
		       signature ray_signature);

  double interpolate_value(double px, double py, double pz);
  double x_face_val(int cell_x, double py, double pz);
  double y_face_val(double px, int cell_y, double pz);
  double z_face_val(double px, double py, int cell_z);
  void get_field_gradient(double px, double py, double pz, vector grad);
  void get_vertex_gradient(int vx, int vy, int vz, vector grad);

  // Return the name of the primitive.
  virtual char *prim_name(void) {
    return "Volume";
  }

  // This primitive takes a size and array of isosurface_val when created.
  Volume(int smooth_normals, int should_stop,
	 int x_size, int y_size, int z_size,
	 isosurface_val *field_data, isosurface_val threshold,
	 int dealloc_field_data) {

    interpolate_normals = smooth_normals;
    stop_at_first_intersection = should_stop;
    x_res = x_size;
    y_res = y_size;
    z_res = z_size;
    surf_data = field_data;
    surf_val = threshold;

    deallocate_field_data = dealloc_field_data;

    // Work out increments for stepping in Y and Z.
    x_step = 1;
    y_step = x_res;
    z_step = y_step * y_res;

    is_cluster = 0;
    mark_for_deletion();
  }

  virtual ~Volume(void) {
    if (deallocate_field_data)
      free(surf_data);
  }
}; 


