/* volutils.c : Utility routines that use the Volume primitive.
** 
** This file defines functions that use the Volume primitive to create
** various effects.
**
** Chris Butcher, 24 February 1997
*/

#include "mirage.h"

#ifdef USE_NAMESPACES
namespace Mirage {
#endif

Volume *create_file_extrusion(char *filename, int should_stop)
{
  int x_res, y_res, z_res;
  int i, j, k, array_index;
  isosurface_val *field_data, current_val;
  GenericPixmap *fileImage;
  unsigned char *gray_ptr;

  // Read in the pixmap file.
  fileImage = new GenericPixmap(filename);
  x_res = fileImage->width;
  y_res = fileImage->height;
  z_res = 2;

  field_data = (isosurface_val *) calloc(x_res * y_res * z_res,
					 sizeof(isosurface_val));
  if (field_data == NULL)
    fatal_error("create_file_extrusion", "can't allocate field memory",
		"Allocation of memory to store the extrusion field values in\n"
		"failed. The extrusion may be huge, or else you are\n"
		"critically low on memory.");

  if (fileImage->type != GRAYSCALE)
    fatal_error("create_file_extrusion", "Image must be grayscale",
		"The pixmap that the extrusion is constructed from must\n"
		"be a grayscale PNM file (e.g. extrude1.pgm). This is a\n"
		"different kind of PNM and can't be read.");
  
  // Copy data from the pixmap into the volume array. Note that we must
  // flip the pixmap in Y because the X, Y coordinate system on an image
  // is different to that in X, Y in primitivespace.
  gray_ptr = fileImage->storage;
  
  for (j = 0; j < y_res; j++) {
    for (i = 0; i < x_res; i++) {
      current_val = (isosurface_val) ((*gray_ptr++ + 0.5) *
				      (MAX_SHORT / 256.0));
      for (k = 0; k < z_res; k++) {
	array_index = i + ((y_res - 1) - j) * x_res + k * x_res * y_res;
	
	field_data[array_index] = current_val;
      }
    }
  }
  
  return new Volume(TRUE, should_stop, x_res, y_res, z_res,
		      field_data, 32767, TRUE);
}

Volume *create_revolution_surface(int x_res, int y_res, int z_res,
				  double t_min, double t_max,
				  double (*func)(double), int should_stop)
{
  int i, j, k, array_index;
  double mid_x, mid_y, max_dist;
  double current_dist, surface_dist, current_t;
  isosurface_val *field_data, current_val;

  field_data = (isosurface_val *) calloc(x_res * y_res * z_res,
					 sizeof(isosurface_val));
  if (field_data == NULL)
    fatal_error("create_revolution_surface", "can't allocate field memory",
		"Allocation of memory to store the revolution field data in\n"
		"failed. The revolution may be huge, or else you are\n"
		"critically low on memory.");

  // Find the central axis of the surface.
  mid_x = (x_res - 1) / 2.0;
  mid_y = (x_res - 1) / 2.0;
  max_dist = MIN(mid_x, mid_y);

  // Calculate field values from the revolution-surface function.
  for (k = 0; k < z_res; k++) {
    // Sample the revolution function.
    current_t = t_min + (t_max - t_min) * (k / ((double) z_res - 1));
    surface_dist = fabs(func(current_t));

    for (i = 0; i < x_res; i++) {
      for (j = 0; j < y_res; j++) {
	current_dist = sqrt(SQR(i - mid_x) + SQR(j - mid_y)) / max_dist;

	if (surface_dist == 0) {
	  // There is no surface here.
	  current_val = 0;
	} else {
	  current_dist /= surface_dist;
	  CLAMP(0, current_dist, 2);

	  // We now have the distance from 0 to 2 where 1 is the surface
	  // position. We want field value = max * (1 - d/2)
	  current_val = (isosurface_val) (MAX_SHORT * (1 - current_dist/2));
	}

	array_index = i + ((y_res - 1) - j) * x_res + k * x_res * y_res;

	field_data[array_index] = current_val;
      }
    }
  }
  
  return new Volume(TRUE, should_stop, x_res, y_res, z_res,
		      field_data, 32767, TRUE);
}  



#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif
