/* write.c : Writes an image file.
** 
** This file defines the functions that interact to write image files.
** Currently supported formats:   PPM, PPM.GZ, TIF, PNG
**
** Chris Butcher, 2 December 1996
*/

#include "mirage.h"

#ifdef USE_NAMESPACES
namespace Mirage {
#endif

#ifdef PNG_ON
#include <png.h>
#endif

// This routine checks file / path names to check that they are the right
// format for the given OS. If Macintosh, it will convert
// "maps/oasis_hf.pgm" to ":maps:oasis_hf.pgm" and if UNIX, vice versa.
char *filename_oscheck(char *fname)
{
#ifdef macintosh
  char separator = ':', badseparator = '/';
#else
  char separator = '/', badseparator = ':';
#endif
  char *stringptr, *newptr, *newstring, thischar;
	
  newstring = (char *) malloc(strlen(fname) + 2);
  if (newstring == NULL)
    fatal_error("filename_oscheck",
		"Can't allocate memory for copy of string!",
		"Out of memory.");
  
  newptr = newstring;
  stringptr = fname;
#ifdef macintosh
  // Prefix : to relative path names to make them work. Only if we know
  // we're converting a relative pathname from UNIX.
  if (strchr(fname, '/') != NULL)
    *newptr++ = separator;
#else
  // If there is a : at the start then this is a relative pathname.
  if (*stringptr == badseparator)
    stringptr++;
#endif
  
  for (; (thischar = *stringptr) != '\0'; stringptr++)
    *newptr++ = (thischar == badseparator) ? separator : thischar;
    
  *newptr = '\0';
    
  return newstring;	
}

// This routine writes out a file based on the frame number given and the
// given filename template. The type of file to write is determined by
// the extension of the filename template unless overridden by file_type.
void writeOut(Storage *myStorage, char *file_template, filetype file_type,
	      int frameNumber, char *postprocess_command)
{
  char *tempptr;
  char outfilename[MIRAGE_STRING_LENGTH], extension[10];

  // If we need to, work out what kind of file to write.
  if (file_type == NONE) {
    // Different classes of Storage will store different things - most commonly
    // a color storage only. However we might want to generate a depth map
    // instead or something eclectic.
    // Each subclass of Storage has its own magic number; this allows us
    // to determine which kind of file to output and how to interpret the data
    // in the Storage object.
    // Some kinds of Storage (for example, combined color / normal) may be
    // able to be output in different ways; we choose which one to use
    // by reading in values from the key file.
    
    if (myStorage->magic == COLORSTORAGE_MAGIC) {
      // Write out a colorStorage to a file. There's only one way to do so
      // ... write the colors. Unfortunately many formats to support.

      // Determine the type to write (based on the extension of the template).
      strcpy(extension, "");
      tempptr = strrchr(file_template, '.');
      if (tempptr != NULL) {
	// If we find a valid extension, remove it from the template.
	if (strncmp(tempptr, ".gz", 3) == 0) {
	  //  This is a 'something' gzipped.
	  *tempptr = '\0';

	  // Analyse the rest of the string to work out what kind of
	  // image file it is before being gzipped.
	  tempptr = strrchr(file_template, '.');
	  if (tempptr != NULL) {
	    if (strncmp(tempptr, ".ppm", 4) == 0) {
	      // This is a gzipped ppm file.
	      strcpy(extension, "ppm.gz");
	      *tempptr = '\0';
	    } else {
	      sprintf(outfilename,
		      "extension %s not a valid gzipped image type", tempptr);
	      fatal_error("writeOut", outfilename,
			  "The filename template must either not contain\n"
			  "any full stops or have an image file extension.\n"
			  "Only supported gzipped image type is ppm.gz");
	    }
	  } else {
	    fatal_error("writeOut", ".gz extension on filename but no image"
			"type!", "To output as a gzipped image file, you"
			"must specify an image\nformat as well, for example"
			"new%03d.ppm.gz.");
	  }
	} else if (strncmp(tempptr, ".ppm", 4) == 0) {
	  strcpy(extension, "ppm");
	  *tempptr = '\0';
	} else if (strncmp(tempptr, ".tiff", 5) == 0) {
	  strcpy(extension, "tiff");
	  *tempptr = '\0';
	} else if (strncmp(tempptr, ".tif", 4) == 0) {
	  strcpy(extension, "tif");
	  *tempptr = '\0';
#ifdef PNG_ON
	} else if (strncmp(tempptr, ".png", 4) == 0) {
	  strcpy(extension, "png");
	  *tempptr = '\0';
#endif
	} else {
	  sprintf(outfilename, "extension %s not a valid image type", tempptr);
	  fatal_error("writeOut", outfilename,
		      "The filename template must either not contain\n"
		      "any full stops or have an image file extension\n"
		      "Valid examples are:\n"
		      "island.frame%04d.ppm     new%04d\n"
		      "Invalid example:\n"
		      "geoffs.image.file%04d\n"
		      "\n"
		      "Currently supported image types:\n"
		      "ppm, ppm.gz, tif, tiff"
#ifdef PNG_ON
		      ", png"
#endif
		      );
	}
      } else {
	if (!get_keyfile_data("image-format", extension)) {
	  fatal_error("writeOut", "no image format specified!",
		      "You must either specify an image format by using\n"
		      "a valid file extension in the filename template\n"
		      "or by using the keyfile parameter image-format.");
	}
      }

      // Now we have an extension. Convert this to a file type and add it
      // back onto the filename template.
      if (strcmp(extension, "ppm.gz") == 0) {
	//	printf("setting filetype to PPM_GZ\n");
	file_type = PPM_GZ;
	strcat(file_template, ".ppm.gz");

      } else if (strcmp(extension, "ppm") == 0) {
	//	printf("setting filetype to PPM\n");
	file_type = PPM;
	strcat(file_template, ".ppm");

      } else if (strcmp(extension, "tif") == 0) {
	//	printf("setting filetype to TIFF\n");
	file_type = TIFF;
	strcat(file_template, ".tif");

      } else if (strcmp(extension, "tiff") == 0) {
	//	printf("setting filetype to TIFF\n");
	file_type = TIFF;
	strcat(file_template, ".tiff");

#ifdef PNG_ON
      } else if (strcmp(extension, "png") == 0) {
	//	printf("setting filetype to PNG\n");
	file_type = PNG;
	strcat(file_template, ".png");
#endif

      } else {
	sprintf(outfilename, "unrecognized image format %s", extension);
	fatal_error("writeOut", outfilename,
		    "Currently supported image formats are:\n"
		    "ppm, ppm.gz, tif"
#ifdef PNG_ON
		    ", png"
#endif
		    );
      }
    
    } else if (myStorage->magic == TRACKSTORAGE_MAGIC) {
      fatal_error("writeOut()", "TrackStorage cannot be written to a file.",
		  "TrackStorage must be converted into a ColorStorage in\n"
		  "order to write an image file.");
    } else {
      // Don't recognize the Storage class. Scream and die.
      fatal_error("writeOut()", "unrecognized Storage subclass",
		  "This error should not happen.");
    }
  }
  // *** End of filetype determination ***


  // We know what type of file to write. Compose the filename.
  sprintf(outfilename, file_template, frameNumber);

  // Write the file.
  writeOut(myStorage, outfilename, file_type, postprocess_command);
}

void writeOut(Storage *myStorage, char *fname, filetype file_type,
	      char *postprocess_command)
{
#ifdef UNIX_SPECIFIC
  FILE *out_pipe;
  char tmpfile[MIRAGE_STRING_LENGTH];
#endif
  FILE *out_file, *temp_file = NULL;
  char commandbuf[MIRAGE_STRING_LENGTH];

  if (!completely_silent) {
    sprintf(commandbuf, "> Writing file %s", fname);
    print_current_time(commandbuf);
  }

  // Write out the file.
  switch(file_type) {
  case PPM:
    out_file = open_output_file(fname);
    write_ppm_to_file(out_file, (ColorStorage *) myStorage);
    fclose(out_file);
    break;

#ifdef PNG_ON
  case PNG:
    out_file = open_output_file(fname);
    write_png_to_file(out_file, (ColorStorage *) myStorage);
    fclose(out_file);
    break;
#endif

  case PPM_GZ:
  case TIFF:
#ifndef UNIX_SPECIFIC
    fatal_error("writeOut",
		".ppm.gz, .tiff formats only available on UNIX.",
		"Use a supported file format instead.");
    break;
#else
    // Write out the image as PPM.
    if ((tmpnam(tmpfile) == NULL) ||
	((temp_file = open_output_file(tmpfile)) == NULL))
      fatal_error("writeOut", "Cannot open temporary file!",
		  "This error should not happen.");

    write_ppm_to_file(temp_file, (ColorStorage *) myStorage);
    fclose(temp_file);

    // Set up the conversion command.
    if (file_type == PPM_GZ)
      sprintf(commandbuf, "gzip -c %s > %s", tmpfile, fname);
    else if (file_type == TIFF)
      sprintf(commandbuf, "(pnmtotiff %s > %s) >& /dev/null", tmpfile, fname);
    else
      fatal_error("writeOut", "Invalid tempfile processing type!",
		  "This error should not happen.");

    // Convert the image and delete the old file.
    system(commandbuf);
    unlink(tmpfile);
    break;
#endif
  default:
    fatal_error("writeOut", "no file type set!",
		"This error should not happen.");
  }

  // Perform the post-processing command if there is one.
  if (strcmp(postcommand, "") != 0) {
    sprintf(commandbuf, postprocess_command, fname);
    system(commandbuf);
  }
}

void write_ppm_to_file(FILE *outfile, ColorStorage *aStorage)
{
  int data_bytes, x, y;
  unsigned char *pic;

  x = aStorage->xsize;
  y = aStorage->ysize;
  pic = aStorage->colorStorage;

  // Write PPM header
  fprintf(outfile, "P6 %d %d 255\n", aStorage->xsize, aStorage->ysize);

  // Write the pixel values into the PPM file (RAWBITS compaction)
  data_bytes = 3 * x * y;
  fwrite((void *) pic, sizeof(unsigned char), data_bytes, outfile);
}

#ifdef PNG_ON
// Write out the image as a 24-bit RGB PNG file.
void write_png_to_file(FILE *outfile, ColorStorage *aStorage)
{
  int image_height, image_width, row_no;
  png_bytep image_values, *row_ptrs;
  png_structp png_write_struct;
  png_infop pngfile_info;

  // Read the size of the image.
  image_width = aStorage->xsize;
  image_height = aStorage->ysize;
  
  // Get a png_byte pointer to the image data in bytes.
  image_values = (png_bytep) aStorage->colorStorage;

  // Initialise the PNG write struct. Check that the versions of the library
  // and the header png.h correspond and set the default error handling
  // method (using setjmp and stderr).
  png_write_struct = png_create_write_struct(PNG_LIBPNG_VER_STRING,
					     NULL, NULL, NULL);
  if (png_write_struct == NULL) {
    fatal_error("write_png_to_file", "can't create png write struct",
		"The versions of png.h and libpng.a may be incorrect, or\n"
		"you may be critically low on memory.");
  }

  pngfile_info = png_create_info_struct(png_write_struct);
  if (pngfile_info == NULL) {
    // Deallocate the png writing structure.
    png_destroy_write_struct(&png_write_struct, (png_infopp) NULL);

    fatal_error("write_png_to_file", "can't create png info struct",
		"You may be critically low on memory. Regardless, this\n"
		"error should not happen.");
  }

  // Setup for error handling using setjmp to specify that this condition
  // should be entered if an error occurs using a longjmp routine.
  if (setjmp(png_write_struct->jmpbuf)) {
    // An error has occurred. Deallocate memory and die.
    png_destroy_write_struct(&png_write_struct, (png_infopp) NULL);

    fatal_error("write_png_to_file", "PNG library error writing file",
		"An unknown error in the PNG library has prevented the\n"
		"file being written.");
  }

  // Set up the output control to the file pointer.
  png_init_io(png_write_struct, outfile);

  // Set up and write the file info header.
  pngfile_info->width          = image_width;
  pngfile_info->height         = image_height;
  pngfile_info->color_type     = PNG_COLOR_TYPE_RGB;
  pngfile_info->channels       = 3;
  pngfile_info->bit_depth      = 8;
  pngfile_info->interlace_type = 0;
  pngfile_info->sig_bit.red    = 8;
  pngfile_info->sig_bit.green  = 8;
  pngfile_info->sig_bit.blue   = 8;
  png_write_info(png_write_struct, pngfile_info);

  // Set up an array of pointers to the start of each row of the image.
  row_ptrs = (png_bytep *) calloc(image_height, sizeof(png_bytep));
  if (row_ptrs == NULL) {
    fatal_error("write_png_to_file", "can't allocate row pointers for image",
		"You may be running critically low on memory. Regardless,\n"
		"this error shouldn't happen ... it's not a large amount\n"
		"of memory.");
  }
  for(row_no = 0; row_no < image_height; row_no++)
    row_ptrs[row_no] = &image_values[row_no * 3 * image_width];

  // Write the image to the PNG file.
  png_write_image(png_write_struct, row_ptrs);

  // Finish up and write the file end chunk.
  png_write_end(png_write_struct, pngfile_info);

  // Deallocate the write struct.
  png_destroy_write_struct(&png_write_struct, (png_infopp) NULL);
}
#endif // PNG_ON

FILE *open_output_file(char *filename)
{
  FILE *outfile;
  char *new_fname;

  new_fname = filename_oscheck(filename);
  outfile = fopen(new_fname, "wb");
  if (outfile == NULL) {
    char errbuf[MIRAGE_STRING_LENGTH];
    
    sprintf(errbuf, "Can't open output file %s", new_fname);
    fatal_error("open_output_file", errbuf,
		"The specified output file might be an invalid filename,\n"
		"or possibly there is no room left on the drive.");
  }

  free(new_fname);

  return outfile;
}

#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif
