/* xwindows.c : X Windows interface.
** 
** This file defines the functions for the XDisplay class, which handles
** creating and updating the X-Windows progressive display that is done
** by the DisplayAntialiaser.
**
** Chris Butcher, 21 January 1997
*/

#include "mirage.h"

#ifdef USE_NAMESPACES
namespace Mirage {
#endif

#ifdef XWINDOWS_ON
// All of the following code deals with the X11 interface. If this
// is not working on your system, you may wish to remove it by commenting
// out the XWINDOWS_ON definition in mirage.h.

// Constructor for the XDisplay object; initialises an X window to draw in.
XDisplay::XDisplay(int img_width, int img_height, Display *use_display)
{
  int theXScreen;
  Colormap theXColormap;
  Window theRootWindow;
  XSetWindowAttributes theWindowAttrib;
  XGCValues theGCValues;
  Pixmap dummyPixmap;
  char window_name[MIRAGE_STRING_LENGTH];

  win_width = img_width;
  win_height = img_height;

  if (use_display != NULL) {
    // Use the display passed to us.
    theXDisplay = use_display;
  } else {
    // Open the X display.
    theXDisplay = XOpenDisplay(NULL);
    if (theXDisplay == NULL) {
      output_warning("XDisplay::constructor", "can't open X display",
		     "The DISPLAY environment variable may be set\n"
		     "incorrectly or else the target display is rejecting\n"
		     "X connections. Cannot display rendering progress.");
      return this;
    }
  }

  // Get information about the X display.
  theXScreen          = DefaultScreen(theXDisplay);
  theRootWindow       = RootWindow   (theXDisplay, theXScreen);
  theXVisual          = DefaultVisual(theXDisplay, theXScreen);
  num_bitplanes       = DisplayPlanes(theXDisplay, theXScreen);

  // Check that we have a true-color display.
  if (theXVisual->c_class != TrueColor) {
    // The default visual isn't TrueColor. Try to find one that is.
    XVisualInfo requestedVisualInfo, *foundVisualInfo;
    long request_mask;
    int num_found, selected_visual, best_depth, i;

    // We would like a 24-plane TrueColor visual on the current screen.
    requestedVisualInfo.c_class = TrueColor;
    requestedVisualInfo.screen = theXScreen;
    requestedVisualInfo.depth = 24;
    request_mask = VisualClassMask | VisualScreenMask | VisualDepthMask;

    // Look for matching Visuals.
    foundVisualInfo = XGetVisualInfo(theXDisplay, request_mask,
				     &requestedVisualInfo, &num_found);

    // If we can't find a 24-plane TrueColor visual, give up.
    if ((!foundVisualInfo) || (num_found == 0)) {
      // Free the visual list.
      if (foundVisualInfo != NULL)
	XFree((char *) foundVisualInfo);
	
      output_warning("XDisplay::constructor", "X display is not TrueColor",
		     "The display code can only display on a true-color\n"
		     "display, not an indexed or colourmapped one. Also,\n"
		     "24 bitplanes of TrueColor is required. This\n"
		     "is because new colours would have to be allocated\n"
		     "on the fly as the render progressed if the display\n"
		     "wasn't true-color ... and this is difficult to do.\n"
		     "X-windows display not possible.");
    
      // Flag that display will not take place.
      theXDisplay = NULL;

      return this;
    }

    // We've found at least one 24-bit TrueColor visual. Use the first one.
    theXVisual = foundVisualInfo[0].visual;
    num_bitplanes = foundVisualInfo[0].depth;
    
    // Free the visual list.
    if (foundVisualInfo != NULL)
      XFree((char *) foundVisualInfo);
  }

  // Get the correct colormap.
  theXColormap = XCreateColormap(theXDisplay, theRootWindow, theXVisual,
				 AllocNone);

  // Set up a dummy (1x1) pixmap of the correct number of bitplanes for
  // use as the border pixmap of the window. This is the essential step
  // according to Dean McRobie ... at least, without this I have had no
  // success.
  dummyPixmap = XCreatePixmap(theXDisplay, theRootWindow, 1, 1, num_bitplanes);
  
  // Set some attributes of the window that we will create.
  theWindowAttrib.background_pixel = BlackPixel(theXDisplay, theXScreen);
  theWindowAttrib.border_pixel     = BlackPixel(theXDisplay, theXScreen);
  theWindowAttrib.border_pixmap    = dummyPixmap;
  theWindowAttrib.colormap         = theXColormap;
  theWindowAttrib.event_mask       = ExposureMask;

  // Flush the X output stream.
  XFlush(theXDisplay);
  XSync(theXDisplay, False);

  // Create a new window on the given display, with parent as the root
  // window. Coordinates 0,0, size width x height, border width 2 pixels,
  // an input and output window, using the true-color visual,
  // taking the background_pixel, border_pixel, border_pixmap, colormap
  // and event_mask values from theWindowAttrib.
  theXWindow = XCreateWindow(theXDisplay, theRootWindow,
			     0, 0, win_width, win_height, 2,
			     num_bitplanes, InputOutput, theXVisual,
			     CWBackPixel | CWBorderPixel | CWBorderPixmap |
			     CWColormap | CWEventMask, &theWindowAttrib);

  if (!theXWindow) {
    output_warning("XDisplay::constructor", "cannot create X window",
		   "The XCreateWindow call has failed and so rendering\n"
		   "progress cannot be displayed. This should not happen.");

    // Flag that display will not take place.
    theXDisplay = NULL;
    
    return this;
  }

  // Create the window name.
  sprintf(window_name, "%s: rendering at %d x %d", project_name,
	  win_width, win_height);
  window_name[0] = toupper(window_name[0]);
  XStoreName(theXDisplay, theXWindow, window_name);

  // Place the window on the display.
  XMapWindow(theXDisplay, theXWindow);

  // Create a graphics context for the window.
  theXGraphicsContext = XCreateGC(theXDisplay, theXWindow, 
				  0, &theGCValues);

  // Initialise an XImage to display in the window (initially black).
  theXImage = initialise_XImage(win_width, win_height);

  // Display the image from (0, 0) - (width-1, height-1).
  XPutImage(theXDisplay, theXWindow, theXGraphicsContext, theXImage,
	    0, 0, 0, 0, win_width, win_height);
}

// Create a blank XImage to store the window's image in.
XImage *XDisplay::initialise_XImage(int width, int height)
{
  XImage *theImage;
  
  // Set up the color shifting masks appropriate to this Visual.
  redmask = theXVisual->red_mask;
  grnmask = theXVisual->green_mask;
  blumask = theXVisual->blue_mask;

  redshift = highest_bit(redmask) - 7;
  grnshift = highest_bit(grnmask) - 7;
  blushift = highest_bit(blumask) - 7;

  // Create the XImage.
  theImage = XCreateImage(theXDisplay, theXVisual, num_bitplanes,
			  ZPixmap, 0, NULL, win_width, win_height,
			  32, 0);
  if (theImage == NULL) {
    fatal_error("XDisplay::initialise_XImage", "cannot create XImage",
		"The call to XCreateImage has failed, possibly because\n"
		"of very low memory. This is a small amount of memory,\n"
		"though, and should not cause a problem.");
    return NULL;
  }

  bytes_per_line = theImage->bytes_per_line;
  byte_pack_order = theImage->byte_order;
  bits_per_pixel = theImage->bits_per_pixel;

  // Check that we have a supported number of bits per pixel.
  if (bits_per_pixel == 32)
    bytes_per_pixel = 4;
  else {
    char errbuf[MIRAGE_STRING_LENGTH];

    sprintf(errbuf, "TrueColor with %d bits per pixel (not supported)",
	    bits_per_pixel);
    fatal_error("XDisplay::initialise_XImage", errbuf,
		"Currently only TrueColor with 32 bits per pixel is\n"
		"supported by this X display code. Other bits/pixel\n"
		"combinations could be easily implemented but aren't.");
    return NULL;
  }

  // Allocate data to store the image colours in.
  theImage->data = (char *) malloc((size_t) (height * bytes_per_line));
  if (theImage->data == NULL) {
    fatal_error("XDisplay::initialise_XImage", "cannot allocate image memory",
		"Can't allocate memory to store the XImage data in. This\n"
		"is a small amount of memory; its failure likely indicates\n"
		"either a huge image or a memory leak somewhere.");
    return NULL;
  }

  // Set the XImage to black initially.
  memset((void *) theImage->data, 0, height * bytes_per_line);

  return theImage;
}


// Colour one pixel in the window the specified colour.
void XDisplay::colour_pixel(int pix_x, int pix_y, int r, int g, int b)
{
  byte *image_ptr;
  unsigned long red, green, blue, x_color;

  if (theXDisplay == NULL) {
    // We aren't drawing due to some initialisation error.
    return;
  }

  // Get the place to store data in.
  image_ptr = (byte *) theXImage->data + pix_x*bytes_per_pixel
                                       + pix_y*bytes_per_line;

  // Determine the colour value that X expects.
  red   = (unsigned long) r;
  green = (unsigned long) g;
  blue  = (unsigned long) b;

  if (redshift > 0)
    red <<= redshift;
  else
    red >>= (-redshift);

  if (grnshift > 0)
    green <<= grnshift;
  else
    green >>= (-grnshift);

  if (blushift > 0)
    blue <<= blushift;
  else
    blue >>= (-blushift);
  
  x_color = (red | green | blue);

  // Store the colour value in the XImage.
  if (bits_per_pixel == 32) {
    if (byte_pack_order == MSBFirst) {
      *image_ptr++ = (x_color >> 24) & 0xFF;
      *image_ptr++ = (x_color >> 16) & 0xFF;
      *image_ptr++ = (x_color >> 8 ) & 0xFF;
      *image_ptr++ =  x_color        & 0xFF;
    } else {  // byte_pack_order == LSBFirst
      *image_ptr++ =  x_color        & 0xFF;
      *image_ptr++ = (x_color >> 8 ) & 0xFF;
      *image_ptr++ = (x_color >> 16) & 0xFF;
      *image_ptr++ = (x_color >> 24) & 0xFF;
    }
  }

  // Draw the pixel on the window.
  XPutImage(theXDisplay, theXWindow, theXGraphicsContext,
	    theXImage, pix_x, pix_y, pix_x, pix_y, 1, 1);

  // Update and flush events.
  XFlush(theXDisplay);
  XSync(theXDisplay, False);

  // Check to see if the window has been exposed and needs updating.
  check_exposed();
}
    
// Colour a block of pixels with one colour.
void XDisplay::colour_block(int x_min, int y_min, int x_max, int y_max,
			    int r, int g, int b)
{
  byte *image_ptr;
  unsigned long red, green, blue, x_color;
  int x, y;

  if (theXDisplay == NULL) {
    // We aren't drawing due to some initialisation error.
    return;
  }

  // Find the start of the block.
  image_ptr = (byte *) theXImage->data + x_min*bytes_per_pixel
                                       + y_min*bytes_per_line;

  // Determine the colour value that X expects.
  red   = (unsigned long) r;
  green = (unsigned long) g;
  blue  = (unsigned long) b;

  if (redshift > 0)
    red <<= redshift;
  else
    red >>= (-redshift);

  if (grnshift > 0)
    green <<= grnshift;
  else
    green >>= (-grnshift);

  if (blushift > 0)
    blue <<= blushift;
  else
    blue >>= (-blushift);
  
  x_color = (red | green | blue);

  // Store the colour value in the XImage for each pixel in the block.
  for (y = y_min; y < y_max; y++) {
    image_ptr = (byte *) theXImage->data + x_min*bytes_per_pixel
                                         + y    *bytes_per_line;
    for (x = x_min; x < x_max; x++) {
      if (bits_per_pixel == 32) {
	if (byte_pack_order == MSBFirst) {
	  *image_ptr++ = (x_color >> 24) & 0xFF;
	  *image_ptr++ = (x_color >> 16) & 0xFF;
	  *image_ptr++ = (x_color >> 8 ) & 0xFF;
	  *image_ptr++ =  x_color        & 0xFF;
	} else {  // byte_pack_order == LSBFirst
	  *image_ptr++ =  x_color        & 0xFF;
	  *image_ptr++ = (x_color >> 8 ) & 0xFF;
	  *image_ptr++ = (x_color >> 16) & 0xFF;
	  *image_ptr++ = (x_color >> 24) & 0xFF;
	}
      }
    }
  }

  // Write the block to the display.
  XPutImage(theXDisplay, theXWindow, theXGraphicsContext,
	    theXImage, x_min, y_min, x_min, y_min,
	    (x_max - x_min), (y_max - y_min));

  // Flush all output events.
  XFlush(theXDisplay);
  XSync(theXDisplay, False);

  // Check to see if the window has been exposed and needs updating.
  check_exposed();
}
    
void XDisplay::check_exposed(void)
{
  XEvent the_event;
  Bool is_event;

  // Checks to see if there are any Expose events in the queue to
  // be processed (for this window).

  do {
    is_event = XCheckWindowEvent(theXDisplay, theXWindow,
				 ExposureMask, &the_event);
    
    if (is_event == True) {
      switch(the_event.type) {
      case Expose: {
	XExposeEvent *exp_event = (XExposeEvent *) &the_event;
	
	// Update the area of the screen that needs updating.
	XPutImage(theXDisplay, theXWindow, theXGraphicsContext, theXImage,
		  exp_event->x, exp_event->y, exp_event->x, exp_event->y,
		  exp_event->width, exp_event->height);
	break;
      }
      }
    }
  } while(is_event == True);
}
			     
// End of X11-specific code.
#endif

#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif


